const
  MAX_CHARACTER_SIZE = 12;

type
{  TCgeFont = class; //forward declaration

  TCgeWord = class
  private
    wS: UnicodeString;
    wF: TCgeFont;
    wZ: integer;
    wB, wI: boolean;
    wHigh, wLow, wWidth: array[boolean] of integer;
    cX, cBase: array[boolean] of array of integer;
    cGlyph: array[boolean] of array of PImageData;
  public
    constructor Create;
    destructor Destroy; override;
    procedure SetString(s: PUnicodeChar);
    function SetOpts(font, size: integer; bold, italic: boolean): integer;
    function CalcWidth(condensed: boolean): integer;
    function CalcHigh(condensed: boolean): integer;
    function CalcLow(condensed: boolean): integer;
    procedure Render (var image: TImageData; condensed: boolean);
  end;  }
  TGlyphPar = (gpX, gpY, gpW, gpH, gpBase);
  TGlyphData = array [TGlyphPar] of smallint;
  PGlyphData = ^TGlyphData;

  TFaceRec = record
    Subs: array [0..3] of boolean;
    cGlyph: array of TImageData;
    char: array of TGlyphData;
    Kerning: TImageData;
    Texture,
    ShadowTexture,
    TextureWidth,
    TextureHeight: glUint;
  end;
  
  TCgeFont = class
  private
    Face: array of array [boolean, boolean, boolean] of TFaceRec;
{    cBase:  array of array of array [boolean, boolean, boolean] of integer;
    cGlyph: array of array of array [boolean, boolean, boolean] of TImageData;
    cGlyphData: array of array [boolean, boolean, boolean] of array of array[TGlyphPar] of integer;
    Kerning:array of array [boolean, boolean, boolean] of TImageData;
    faceSubs: array of array [boolean, boolean, boolean, 0..3] of boolean;
    Tex: array of array [boolean, boolean, boolean, 0..3] of glUint; //0 name 1 width 2 height 3 shadow texture
}
    SzMin, szMax: integer;
    cSize:  array of integer;
    cIndex: array [UnicodeChar] of integer;
    sIndex: array [0..MAX_CHARACTER_SIZE] of integer;
  public
    Name: string;
    procedure Load(fn: string);
    function GetBase(A: UnicodeChar; size: integer; bold, italic, condensed: boolean): integer;
    function GetKerning(A, B: UnicodeChar; size: integer; bold, italic, condensed: boolean): integer;
    function GetGlyph(A: UnicodeChar; size: integer; bold, italic, condensed: boolean): PImageData;
    function GetTexture(size: integer; bold, italic, condensed: boolean; sizex, sizey: PInteger; shadow: pglUint): glUint;

    function FigureFace (size: integer; var bold, italic, condensed: boolean): integer; //return size index
    //converts to supported bold/italic/condensed set
    
    //these versions use the already conversed ad thus definitely valid bold/italic/condensed values:
    function _GetBase(A: PUnicodeChar; si: integer; bold, italic, condensed: boolean): integer;
    function _GetKerning(A, B: PUnicodeChar; si: integer; bold, italic, condensed: boolean): integer;
    function _GetGlyph(A: PUnicodeChar; si: integer; bold, italic, condensed: boolean): PImageData;
    function _GetGlyphData(A: PUnicodeChar; si: integer; bold, italic, condensed: boolean): PGlyphData;

    Constructor Create;
    Destructor Destroy; override;

    function CalcWord (chars: PUnicodeChar; len, size: integer; bold, italic, condensed: boolean;
                       word: PiCharArray; texture, shadow: pglUint): integer;
    function GetSupportedSize(i: integer): integer;
  end;
