{
    This file is part of Chentrah,
    Copyright (C) 2004-2008 Anton Rzheshevski (chebmaster@mail.ru).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see http://www.gnu.org/licenses/

 **********************************************************************
 
    This unit is a placeholder for most of the mother module functions
      composed C-style from various include files

 **********************************************************************}

{$macro on}
{$longstrings on}
{$if FPC_FULLVERSION>=20700}
  {$modeswitch unicodestrings}
{$endif}
{$codepage utf-8}

unit cge; //stands for Cheb's Game Engine

{$define cge}
{$define cgekernel} // should be "mothermodule" now but I'm too lazy to change it.


{$ifdef windows}
  {$define use_dynlibs}
{$endif}


interface
   uses
       {$ifdef unix}
         cthreads,
       {$endif unix}
         SysUtils
       , typinfo
       , ctypes
        {$ifdef unix}
//           , libc //for only one function, to determine the phys. RAM size. What a shame
           , baseunix
           , unix
           , x
           , xlib
           , xutil
           , dl
           , keysym
        {$endif}
        {$ifdef windows}
          , windows
          //, registry
          , cl_windows_missing
          , cl_winedetect
          , process
        {$endif}
      // , cl_clamscan //The fuck was I wasting my time on making a built-in antivirus scanner based on ClamAV?
       , classes
       , syncobjs
       {$ifdef fix_fpc3_unicode}
         , contnrs
       {$else}
         , IniFiles//fixed_IniFiles
       {$endif}
       , zstream
       , math
       , md5
       , un_process
       , cl_strings
       , cl_builtinmodule


       (* 2015: SCRAPED.
       {$ifdef buildmein} //2014: I suppose this will eventually become the "release mode" //Jan 2010: Ain't need no this shit...?
        , cl_internalmodule //no any DLLs, only one module built into the executable (a.k.a. Release Mode)
       {$else}
       //, cl_seh_hack //It also performs a small task of determining process handle.
       {$endif}
        *)
        {$ifdef win32}
          , cl_seh_hack_win32  // necessary to catch AVs in the module DLL
         // without this, any exception in a module thread would lead to a
         // silent crash-to-desktop
        {$endif}
        {$ifdef win64}
          , cl_seh_hack_win64  // likewise
        {$endif}


       , Imaging  //Vampyre Imaging Library by Marek Mauder
       , ImagingTypes
       , ImagingFormats
       ;

 //WTF this is defined for the ARM platform ?! {$ifdef CPUI386}
 // {$define ASSEMBLER_ALLOWED}
//{$endif}

procedure Run();

procedure PerformCgeShutdown;

// ---------=========******* ГЛОБАЛЬНЫЕ ПЕРЕМЕННЫЕ *******========-------

CONST
//  UnixIcon = BaseImagesPath + 'icon.xbm';
  BuildNumber = {$include build.h};

  {$include un_texsages.h} //enumerated type for the message container
    MI_CORE_MIDs_END ); 
  
var
  CGEString: UnicodeString = 'Chentrah';

  {$include un_typedefs.h} // all important types and global variables here *****************


  {$ifdef fix_fpc3_unicode}
     {$include un_fpc_3-0-0_fix.h} // Use custom mods of TFileStream, TStringList and TIniFile
  {$endif}
  {$if defined(windows) and (FPC_FULLVERSION<30000)}
     {$include un_fpc_2-6-4_fix.h} // Use fixed TFileStream and copy-pasted everything that depends on it
  {$endif}

  {$include un_lineinfo.h}
  
var
  Mother: PMother; // the great and terrible interface hub connecting the mother executable with the module dll

  {$include un_unicode.h}

  {$INCLUDE cl_texsages.h} //message container class header + built-in message strings
  
  {$include cl_preinityells.inc} // pre-initialization error message strings
  
//  {$include cl_dyna.h}   //this shit causes instability!!!

  {$include cl_dynareplacer.h}
  
  {$include cl_default_conf.h} // a built-in config .ini file
  
  procedure ResetMotherErrorState();


type
  //float = single;
  EFake = class(Exception);
  EDying = class(Exception);

var
(*
  OpenJpegDll: string = //can be overridden in the main ini file
    {$ifdef unix}
      {$ifdef darwin}
        {$fatal No OpenJpeg support for you!}
      {$else}
        '*I/3rdparty/libopenjpeg.so'
      {$endif}
    {$else}
      '*I\3rdparty\OpenJPEG.dll'
    {$endif}
      ;
 *)
  CppSupportLibrary: TFileNameString = 'libstdc++.so.5'; //must be dynamically loaded for openjpeg to work (Linux)
  LocalStoreStr: TFileNameString = 'temporary-files';
  TranslationsDir: TFileNameString = 'modules' + DS + 'chentrah' + DS + 'translations' + DS;
  SourcesDir: TFileNameString = 'modules' + DS + 'chentrah' + DS + 'src' + DS;
  ConfigFilePath: TFileNameString = 'conf' + DS + 'conf.ini';
  LogsDir: TFileNameString = 'logs' + DS;
  ImagesDir: TFileNameString = 'modules' + DS + 'chentrah' + DS + 'img' + DS;
  SoundsDir: TFileNameString = 'modules' + DS + 'chentrah' + DS + 'snd' + DS;
  //State.RestartRequested: boolean = No; //restart from the scratch. dont rewrite the old log (moved to the MotherState interface record)
  NowRestarting: boolean = No;
//moved to motherstate  State.State.ExitRequested: boolean = No; //is used to initiate program shutdown

  FirstRestart: boolean = Yes;
  XServerBully: boolean = No; //set when hacking around the force shutdown by X server
  ReceivedSigterm: boolean = No; //set when requested to terminate by the operating system
//  EmergencyShutdown: boolean;// signals that application goes down due to some svere failure. (OpenGL may be unavailable)

  
  HeartbeatMinMiliseconds: integer = 35;

  MainIni: TMemIniFile;
  {StartDir, chInstallPath, LocalStorePath, FoLocalStoreDir,} ModulesBasePath {,SessionPath, SavePath, chHomePath}
  : ansistring; {all use the default system 8-bit locale!
                 (on modern OSes it's usually utf-8)}

  AppNick: ansistring;
  LastHeartbeatMoment: TDateTime;
  AbortHotKey, RestartHotKey, ConsoleToggleHotkey, ConsoleScrollUpHotkey,
    ConsoleScrollDownHotkey, LanguageHotKey, SessionRollbackHotkey,
    QualityFactorManualOverrideToggleHotkey,
    FullscreenHotkey
     : TIntegerArray;
  ModuleHotKey : array[0..0{MAX_MODULE}] of TIntegerArray;

  CgeStartTime, SecondStageInitTime: TDateTime;
  CgeStartingTick: int64;//longint; then it ran it on a laptop that wasn't rebooted in ages... And hello, overflow error!
  DontWriteALog: boolean;
  AnotherInstanceIsRunning: boolean = No;
  NoHardwareAcceleration: boolean = No;

  CurrentFrame: int64 = 0;
  CurrentFrameDate: TDateTime = 0.0;
type
  THomePathWarning = (hfwNone, hfwTriedIUseH, hfwTriedHuseI, hfwUseI, hfwWin9x);
var
  HomePathWarning: THomePathWarning = hfwNone;
  
type
  TBimMessage = record
    inipar: string;
    capMID, msgMID: TMessageID;
    capStr, msgStr: UnicodeString;
    params: array of UnicodeString;
    writeval: integer;
    isyellow: boolean;
  end;
var
  StopMessage, WarningMessage: array of TBimMessage;
  ModuleStopMessage: TBimMessage;

//  WarningQueue: TUnicodeStringContainer;//TAOW;

type
  TEngineObjects = class(TObjectContainer)
    Function AddObj(i: TObject): integer;
  end;
  
  TCgeProcess = class(TExProcess)
    procedure OnReadLn(ss: string); override;
  end;
  
  TProcedure = procedure; cdecl;
  TTryWrapperProcedure = procedure(p: pointer); cdecl;

var
  EngineObject: TEngineObjects;
  EverythingIsOk: boolean = Yes;

const
  MAX_STACK_FRAMES_DUMP = 5;//in error messages, unless the debug mode is on

  OGL_REQ_HI = 1{2}; //Now each module decides if it needs 2.0 or 1.2
  OGL_REQ_MED = 2{0};

  OGL_REQ_HI2 = 2; //OpenGl 2.0 required
  OGL_REQ_MED2 = 0;


  OAL_BUFFERS = 2; //OpenAL sources and buffers (swap at receiving new data)
  OAL_DEVICE_OPEN_TIMEOUT = 10.0 / SecondsPerDay; //OAL sometimes locks up when opening a device

  MODULE_STOP_TIMEOUT = 15.0 / SecondsPerDay; {If the module dawdles longer
    than that when it is unloaded, it causes application restart.}

// 2014: Moved this to un_typedefs.h as QualityFactor is now controlled by the module.
//  QF_MIN = 10;
//  QF_MAX = 999;
  FPS_COUNTER_SIZE = 0.10; //of the window size
  FPS_COUNTER_MAX_FONT_SIZE = 18;
  FPS_MEASURING_INTERVAL = 0.33; //seconds

  FADE_IN_TIME = 1.33; //seconds. Screen from black.

  //resources
  NOT_A_MODULE = -1;
  MAX_MODULE_LOAD_ATTEMPTS = 6; //since an attempt to load it while it's still being compiled
    // is destined to fail but the module manager does that as soon as the
    //file date changes, we have to try several times before giving up
  
    //deprecated! As of 2012, module always continues running on at least 30 Timer.FPS.
    FRAMES_TO_CONTINUE_RENDER_AFTER_LOSING_FOCUS = 4; {While not having focus, the
    window updates only when it receives a message (e.g. you pass your mouse
    pointer above it). This doesn't just affect render, it stops calling the
    module's Pulse procedure when inactive.}
  //sleep this interval each frame if our window does not have the input focus
  NO_FOCUS_SLEEP_INTERVAL = 150;

//moved to un_typedefs.h
//  FPS_CULL_DEFAULT = 30; //also, values of Mother^.Timer.CullFrameRate below this are ignored
  FPS_CULL_DEFAULT_MOUSE_MOVED = 60; //raise if mouse moved. Without this, the OpenGL mouse pointer looks sluggish.
  FPS_CULL_DEFAULT_MOUSE_DELTA_THRESHOLD = 5; //pixels per frame (mouse movement speed)



  //image handling
  CGE_RESIZE_FILTER: TResizeFilter = rfBicubic;
  CGE_RESAMPLE_FILTER: TSamplingFilter = sfLanczos;
  
 //Two values control the mouse pointer mode: either the full-color one via OpenGL
 //  or the 2-bit black and white image set as the standard system cursor
 //  using the obsolete mechanisms of Win32 / X.
 //The upside of this primitivism is that it's compatible
 //  with both Windows 95 and any obscure X server with no randR extensions.
 //The mode is controlled from TCLWindow.SwapBuffers (cl_winman.inc),
 //  which also renders the OpenGL cursor.
  HW_POINTER_LOW_FPS_THRESHOLD = 20;
  HW_POINTER_HIGH_FPS_THRESHOLD = 25;
  HW_POINTER_ALPHA_THRESHOLD = 0.5; //where to cut off while creating the 2-bit version.
  HW_POINTER_DITHER_FACTOR = 0.6;

  
  //must match cl_cursor_normal.png
  CGE_STD_CURSOR_HOTSPOT_X = 3;
  CGE_STD_CURSOR_HOTSPOT_Y = 3;
  
  //must match cl_cursor_error.png
  CGE_ERR_CURSOR_HOTSPOT_X = 3;
  CGE_ERR_CURSOR_HOTSPOT_Y = 3;
  
  //must match cl_cursor_hourglass.png
  CGE_HRGLASS_CURSOR_HOTSPOT_X = 15;
  CGE_HRGLASS_CURSOR_HOTSPOT_Y = 15;

  fixedfont_min_qf_to_use_hq_texture = 15;
  fixedfont_max_qf_to_use_lq_texture = 20;

  // Timer
  TSCMeasuringInterval = 10.0;//seconds
  INITIAL_BLACKOUT = 0.3; //seconds before the fade in from black begins.
  LOGO_ANIMATION_LENGTH = 4.0; //seconds;
{
var
  Resources: TAssetRecordContainer;
}
  //ActiveOwner: integer = 0;
//  InExportedProc: boolean = false;

  procedure InitMotherState;

  function Tick: longint; //miliseconds since CGE start;

  Procedure Die(YellID :TMessageID; Param: array of const); overload;
  Procedure Die(YellID :TMessageID); overload;
  Procedure Die(AnsiYell :AnsiString; Param: array of const); overload;
  Procedure Die(WideYell :UnicodeString; Param: array of const); overload;
  Procedure Die(Ru, En :UnicodeString; Param: array of const); overload;
  Procedure Die(AnsiYell: AnsiString); overload;
  Procedure Die(WideYell: UnicodeString); overload;
  procedure DieReraisingExceptionScreenedByModule(msg: UnicodeString);
  procedure DieUnsupportedGLMode;

  procedure StartDying(u: UnicodeString); {Does NOT raise an exception,
    only sets the states. Can be safely called to add a message to queue}
  function StopDying(pes: PMotherSehState = nil): UnicodeString; {resets the states and clears the queue
    returning the accumulated error messages detritus as one string
    (in backward order, High() to 0)
  }
//  Procedure Warning(YellID :TMessageID; Param: array of const); overload;

  Procedure AddStopMessage(YellID :TMessageID; Param: array of const); overload;
  Procedure AddStopMessage(Yell: UnicodeString; Param: array of const); overload;
  Procedure AddStopMessage(Yell: UnicodeString); overload;
  
  procedure AddMStopMessage(itisfatal: boolean);
  
  procedure GiveWarning(_inipar: string; Cap, Yell :TMessageID; Param: array of const); overload;
  Procedure GivePeriodicWarning(DaysToRepeat: float; _inipar: string; Cap, Yell :TMessageID; Param: array of const);
  Procedure GiveWarningNotYellow(_inipar: string; Cap, Yell :TMessageID; Param: array of const);
  procedure GiveWarning(_inipar: string; Cap, Yell :UnicodeString; Param: array of const); overload;
  procedure AddWarning(Cap: TMessageId; Yell :UnicodeString; Param: array of const); overload;
  procedure AddWarning(Cap, Yell :UnicodeString; Param: array of const); overload;
  procedure GiveWarnings();


  function ExtractDyingYell(pes: PMotherSehState = nil): UnicodeString;
  function GetExceptionState(): PMotherSehState;
  procedure CollateWarningsFromAllThreads;

  
  procedure ProcessGuardedException; cdecl;
  procedure CheckForGuardedException;

//  Procedure ClearWarningQueue;
//  Function IsWarningQueueEmpty: boolean;
  Function PullWarnings: UnicodeString; //collates and clears error messages for *all* threads


  Procedure PreInitDie(Yell: UnicodeString; p: array of const);
  procedure DisplayDyingYells(DyingYell: UnicodeString = '');
  function CallStack(IgnoreLastN: integer; Padding: UnicodeString): UnicodeString;

  {$include un_api_common.h}

  function StrOrUndefined(U: UnicodeString): UnicodeString;
  
  function LoadUnicodeText(FileName: string): TUnicodeStringContainer;

//  Procedure AddLog(S: AnsiString); OVERLOAD;
//  Procedure VerboseLog(S: AnsiString); OVERLOAD;
  Procedure AddLog(U: UnicodeString); OVERLOAD;
  Procedure AddLog(mID: TMessageID); OVERLOAD;
//  Procedure AddLog(S: AnsiString; Param: array of const); OVERLOAD;
  Procedure VerboseLog(U: UnicodeString; Param: array of const); OVERLOAD;
  Procedure VerboseLog(U: UnicodeString); OVERLOAD;
  Procedure VerboseLogOk;
  Procedure AddLog(U: UnicodeString; Param: array of const); OVERLOAD;
  Procedure AddLog(mID: TMessageID; Param: array of const); OVERLOAD;
  procedure VerboseLog(mID: TMessageID; Param: array of const); OVERLOAD;
  procedure VerboseLog(mID: TMessageID); OVERLOAD;
//  Procedure AddLogComment(S: AnsiString; Param: array of const); OVERLOAD;
  Procedure AddLogComment(U: UnicodeString; Param: array of const); OVERLOAD;
  Procedure AddLogComment(mID: TMessageID; Param: array of const); OVERLOAD;
//  Procedure AddLogComment(S: AnsiString); OVERLOAD;
  Procedure AddLogComment(U: UnicodeString); OVERLOAD;
  Procedure AddLogComment(mID: TMessageID); OVERLOAD;

  procedure _LogGoDown;
  procedure LogGoDown(s: UnicodeString);
  procedure LogGoDownWM(s: UnicodeString);
  procedure LogGoDownM(s: UnicodeString);

  Procedure DbgSayS(Yell: AnsiString);
  Procedure DbgSay(Yell: UnicodeString);
  procedure LogProcessAddressSpace; cdecl;

  Procedure Say(s: UnicodeString);


  function _RunningInWindows98: boolean;
  {$ifdef win32}
  function CheckIfWindows98HasUnicodeSupportInstalled: boolean;
  {$endif}
  function _RunningInWindowsNT: boolean;
  procedure DetectOSType;
  procedure LogOSVersion;
  function ThisIsAnOnlyInstance: boolean;

  procedure RestartMyself(param: ansistring);
  procedure KillMyself(exitcode: integer = 0);
  
  procedure GetHotKeysFromConfig;
  function GetHotkeyName(var HK: TIntegerArray): string;
  procedure InitCTimer;
  procedure StartVampyre;
//  procedure InitResourceManager;

  function RuEn(Ru, En: UnicodeString): UnicodeString;
  function VersionToStr(major, minor, build: integer): AnsiString;
  function RectToStr(r: TWindowManagerRect) : UnicodeString;

  function CGEPath(s: TFileNameString): TFileNameString;
  procedure DetermineInstallPath;
  procedure DetermineHomePath;
  function ValidateWritePath (path: TFileNameString): boolean;
  function ValidateFileIsWriteable(FileName: TFileNameString): boolean;
  
  procedure chCopyFile(n1, n2: TFilenameString);

{  Remains of me working in Delphi 5  many years ago. It was nice stuff
     but the exception information in FPC/win32 doesn't support the
     exception address (not the code, but the exact address accessing which
     caused the AV). Still, it was nice to know  which of the multitude of DLLs
     had allocated the memory block where this happened.
   Anyway, I doubt Linux has anything similar to WinAPI's VirtualQuery().}
{ function ToldMemRgn(ptr: pointer): UnicodeString;
  function ExpCallerAddress(): UnicodeString; cdecl; }

  function TellException(E: Exception): UnicodeString;
  function ExpExpAddress(addr: pointer): UnicodeString;
  function ShortExpAddr(addr: pointer): UnicodeString;

  function GetDLLFileName(h: {$ifdef unix} pointer {$else} THandle {$endif}): TFileNameString;

  procedure HiResSleep(ms: longint); cdecl;

  {$include un_threadloadind.h}

   function DoesModulehaveThreadsRunning: boolean; { Ideally, module itself
     should answer this question. But! If the module's main thread is
     incapacitated, there is no other way but parse thread exceptuion states
     and find if all module threads have marked themselves as finished
   }

   procedure GetModuleStopMessage(var caption, text: UnicodeString);
    { 2015 : With recent cxhanges, sometimes the stop screen cannot be shown.
      Must save to config or show in a MessageBox. }


 {$include cl_confman.h}

 {$include un_al.h}  {
     Alas, the OpenAL header should be custom-made too.
       blame the vastly imperfect initialisation/dynamic linking routines
       of the default FPC packages. I will NOT tolerate a unit
       that causes my program to crash silently or with an unitelligible
       error message if some desired library is missing on the user machine.

     My code of honor demands that my program must fail graciously,
       giving a detailed explanation of the problem and some suggestions
       as how to alleviate it (e.g. "Install fresh drivers, ya stupid noob!")
   }

 {$include un_gles_from_fpc_2_6_4.h} {  OpenGL ES header which the game module shares  with the mother module }
 {$ifndef glesonly} //for debugging purposes only (compatibility testing): cut off all support for regular opengl
   {$include un_gl_from_fpc_2_6_4.h}  {  OpenGL header which the game module shares  with the mother module }
 {$endif}
 {$include un_gl_common.h} { a slightly higher level library for working transparently with all OpenGL flavors  }

 {$ifdef unix}  { platform-dependent OpenGL stuff
                    which the game module doesn't need to know about }
   {$ifdef darwin}
     {$include cl_agl.h}
   {$else}
     {$include cl_glx.h}
   {$endif}
 {$else}
   {$include cl_wgl.h}
 {$endif}
 
 function GetGLProcAddress(ProcName: ansistring; Suffix: ansistring = ''): Pointer;
 function GetGLProcAddress2(ProcName, Suffix1, Suffix2: ansistring): Pointer;
 function GetGLProcAddressIgnoreMissing(ProcName: ansistring): Pointer;
 function _GetGLProcAddress(ProcName: ansistring): Pointer;
 function GetEGLProcAddress(ProcName: ansistring): Pointer;
 procedure InitGlProcAddresses(Load2: system.boolean = false);
 procedure CheckGLVersion;
 procedure CheckGLESVersion(majo, mino: EGLint);
 procedure CreateESFFPEmulationProgram;

 function GetALProcAddress(ProcName: ansistring): pointer;

 {$include framework.h}

var
  Framework: TAbstractFramework;


type
  TSmallIntArray = array of smallint;
  function DecodeWav(inp: TStream): TSmallIntArray;

  procedure InitOpenGL;
  procedure LoadOpenGLFunctions;
  procedure CloseOpenGL;
  Procedure InitOpenAL;
  Procedure CloseOpenAL;


  procedure GetOpenGLVersion(var hi, med: integer);

var
  DoShowFps: boolean;
  prevfpssec: double = 0;

  procedure MeasureFps;
  procedure RenderFpsCounter;
  procedure InitFpsCounter;


  procedure InitFixedFont;

  procedure InitMotherAssets;

 {$ifdef unix}
  procedure SetSigtermHandler;
 {$endif}

 {$define cgekernel}
 { define header}
 {$define cgekernel}
   { include cl_exported_func.h}
 { undef header}


 {$include cl_mother_api.h}

 {$include un_fileutil.h}


 
 {$include cl_module.h}
 
 {$include cl_console.h}

  procedure PreloadDefaultPointerImages;
 
  //***TODO: clean up this!
  procedure LoadFixedFontTexture;

type
  FixedFont = class
  protected
    class var ffColor: TVector4f;
  public
    class procedure SizeToFace(size: integer; face: Pinteger; zoom: PGLfloat );
    class function GetStringWidth(Size, numChars: integer): GLfloat;
    class function GetStringHeight(Size, numLines: integer): GLfloat;
    class procedure SetRenderState (noshadow: boolean);
    class procedure FitStringIntoRectangle (
       Size, numChars, numLines, Width, Height: integer;  fit: PStringFitRec);
    class procedure RenderString (fit: PStringFitRec; Left, Top: glFloat; w: PUnicodeChar);

    class procedure Color3f(r, g, b: GLfloat);
    class procedure Color4f(r, g, b, a: GLfloat);
    class procedure Color3fv(p: PGLfloat);
    class procedure Color4fv(p: PGLfloat);
  end;

  procedure ResampleImg (var i: TImageData; newwidth, newheight: integer);

  {$include cl_lockupguard.h}

//  procedure SafeInitSound; cdecl;
//  procedure SafeInitVideo; cdecl;

var
  BottleNeckCounter1, BottleNeckCounter2: int64;

  function NumeralFormEn (n: integer; wSingular, wPlural: UnicodeString): UnicodeString;
  function NumeralFormRu (n: integer; wSingular, wPlural2_4, wPlural: UnicodeString): UnicodeString;

  function SanateStringForFileName(s: AnsiString): AnsiString;


// A non-linear function for screen fade in from black.
  function FancyFunc(f: double): glFloat;

  procedure ErrorMessageBox(Title, Body: UnicodeString);

  {$ifdef unix}{$ifndef darwin}procedure DetectLinuxVersionString;{$endif}{$endif}

  function IniCompatibleStringForm(a: AnsiString): AnsiString;


  function NewInputEvent: integer;

var
  StubTexture: array[0..1] of ptruint; //glEnum
  CurrentStubTextureInd: integer = 0;

 procedure InitStubTexture;

 procedure SwapStubTexture;

 procedure LoadQF;
 procedure SaveQF;

 {$ifndef cpuarm}
  function RecheckTSCFreq(period: double = TSCMeasuringInterval): longbool;
 {$endif}

 function PCharToString(P: PAnsiChar): Utf8String;

var
  LogStream: TFileStream;














IMPLEMENTATION //=====================================================================================



















//uses chebs_heap_brutal_trace;

uses
{$ifndef cpu32}
  lineinfo,
{$endif}
  un_filelist

{$ifdef windows}
   //, cl_framework_winapi
   , cl_splashscreen_win32
   , framework_winapi
{$endif}
{$ifdef linux}
   , cl_splashscreen_linux
{$endif}

;







    procedure CheckForGenericDyingYells(var U: UnicodeString;
       var _crash_details: UnicodeString; omit_frames: integer = 0); forward;
  



  {$include cl_translit.inc} // converts russian text for output using latin alphabet
  

  {$ifdef windows}
    {$include un_errmode.inc} //prevents XP from popping up its own error messages
  {$endif}

  function GoingDownHard: boolean; cdecl;
  begin
    Result:=//EmergencyShutdown or
      XServerBully or ReceivedSigterm;
  end;

  function Tick: longint; //Время в милисекундах от запуска программы
  begin
    //{$ifdef windows}
    // Result:=GetTickCount() - CgeStartingTick;
    //{$endif}

    // Rough. Rounds to 15..16 ms.
    Result:=round((Now() - CgeStartTime) * 86400000.0);
  end;

  procedure HiResSleep(ms: longint); cdecl;
  var
    q, q2: qword;
    cn: integer = 0;
    oto: TThreadLoadKind;
  begin
    oto:= NoteThreadOccupation(tlk_Sleeping);
    {$ifdef unix}
      Sleep(ms);
    {$else}
      if
        //(Mother^.Os in WindowsEmulators) or
        (0 = Mother^.Timer.RDTSCFrequency)
        or not Mother^.Timer.ActiveTBP1
         then Sleep(ms)
      else begin
        //use high resolution sleep
        UsecDelta(@q);
        q2:= q;
        repeat
          if cn = 0 then Sleep(0) else Sleep(2);
          q:= q2;
          Inc(cn);
        until UsecDelta(@q) >= (ms * 1000.0);
      end;
    {$endif}
    NoteThreadOccupation(oto);
  end;
  
  procedure LoadQF;
  var
    i: integer;
    s: ansistring;
    w: dword;
  begin
    i:= Config.Int['videocard', 'quality_factor'];
    if i = 0 then i:=100;
    i:= max(QF_MIN, min(QF_MAX, i));
    Mother^.Timer.QF:= i;
    DoShowFps:= Config.Bool['video', 'show_fps'] or Mother^.Debug.Enabled or Mother^.State.DeveloperMode;
    Config.GetSet(@Mother^.GAPI.Quirks, typeInfo(TGAPIQuirksEnum), 'videocard', 'quirks');
    Config.GetSet(@Mother^.GAPI.QuirksTested, typeInfo(TGAPIQuirksEnum), 'videocard', 'tested_quirks');
  end;
  
  procedure SaveQF;
  begin
    Config.Int['videocard', 'quality_factor'] := round(Mother^.Timer.QF);
    Config.SetSet(@Mother^.GAPI.Quirks, typeInfo(TGAPIQuirksEnum), 'videocard', 'quirks');
    Config.SetSet(@Mother^.GAPI.QuirksTested, typeInfo(TGAPIQuirksEnum), 'videocard', 'tested_quirks');
  end;
  

  var
    _dsay: UnicodeString;
  Procedure DbgSay(Yell: UnicodeString);
  begin
    EnterCriticalSectionWv(Mother^.Core.DbgSayCriticalSection);
    _dsay:= Yell;
    {$ifdef debug}
      WriteLn; WriteLn('-=', IntToHex(GetCurrentThreadId(), 4),'=- ',WideToConsole(yell));
    {$endif}
    if Mother^.Debug.Verbose then AddLog(yell);
    LeaveCriticalSection(Mother^.Core.DbgSayCriticalSection);
  end;

  var sayinit: boolean = false;
  Procedure Say(s: UnicodeString);
  begin


  end;

var
  TimeStamp: Int64 = 0;

  function PCharToString(P: PAnsiChar): Utf8String;
  var
    i: integer;
    p2: PAnsiChar;
  begin
    if not Assigned(p) then Result:=''
    else try
      p2:=p;
      i:=0;
      While p2^ <> #0 do begin
        inc(p2);
        inc(i);
      end;
      SetLength(Result, i);
      if i > 0 then Move(p^, Result[1], i);
    except raise Exception.Create(Utf8Encode('AV in PCharToString() : ' + inttohex(ptruint(p),8) + 'h, count = '+ IntToStr(i))) end;
  end;

  function PUnicodeCharToUnicodeString(P: PUnicodeChar): UnicodeString;
  var
    i: integer;
    p2: PUnicodeChar;
  begin
    if not Assigned(p) then Result:=''
    else begin
      p2:=p;
      i:=0;
      While p2^ <> #0 do begin
        inc(p2);
        inc(i);
      end;
      SetLength(Result, i);
      if i > 0 then Move(p^, Result[1], i * SizeOf(UnicodeChar));
    end;
  end;

  function PFileNameCharToFileNameString(P: PFileNameChar): TFileNameString;
  var
    i: integer;
    p2: PFileNameChar;
  begin
    if not Assigned(p) then Result:=''
    else begin
      p2:=p;
      i:=0;
      While p2^ <> #0 do begin
        inc(p2);
        inc(i);
      end;
      SetLength(Result, i);
      if i > 0 then Move(p^, Result[1], i * SizeOf(p^));
    end;
  end;

 {$ifdef windows}
 {***TODO parse whole 256TB if Windows NT 6.2 or higher
 he first versions of Windows for x64 did not even use the full 256 TB;
 they were restricted to just 8 TB of user space and 8 TB of kernel space.[18]
 Windows did not support the entire 48-bit address space until Windows 8.1,
 which was released in October 2013.}
  procedure LogProcessAddressSpace; cdecl;
  var
    current, next: pointer;
    Des: UnicodeString;
    S: UnicodeString;
    mm, mm2: _MEMORY_BASIC_INFORMATION;
    A, step: ptruint;
    TST: array[0..Max_Path + 1] of AnsiChar;
    Protect: ptruint;
    gapsz, maxgapsz, gapstep: ptruint;
  type
    TMMEnum = (meReserved, meCommitted, meFree, meUnknown, meShared, meGuarded, meNoCache);
  var
    MTSize: array[TMMEnum] of ptruint;
    MTName: array[TMMEnum] of UnicodeString;
    mi, mimax: TMMEnum;
  const
    MaxAddr: ptruint = {$ifdef cpu32}$FFFFFFE0{$else}$000007FFFFFFFFE0{$endif};
  begin
    if Mother^.Debug.Verbose then AddLog(RuEn(
      '  Адресное пространство процесса:',
      '  The address space of the process:'));
    current:= nil;
    FillChar(MTSize, sizeof(MTSize), 0);
    repeat
      if VirtualQuery(current, @mm, SizeOf(mm)) <> SizeOf(mm) then begin
        gapstep:= 64 * 1024;
        maxgapsz:= MaxAddr - gapstep - ptruint(mm.BaseAddress);
        gapsz:= 0;
        next:= nil;
        repeat
          {$ifdef cpu64}
          if gapsz >  $0fffffff then gapstep:= Math.min(maxgapsz - gapsz - 1, 1024*1024);
          //if gapsz > $fffffffff then gapstep:= Math.min(maxgapsz - gapsz, 1024*1024*1024);
          {$endif}
          if (gapstep > maxgapsz) or (gapsz > (maxgapsz - gapstep))
            then break;
          gapsz+= gapstep;
          if VirtualQuery(current, @mm, SizeOf(mm)) = SizeOf(mm) then begin
            next:= mm.BaseAddress;
            gapsz:= ptruint(next) - ptruint(current);
            break;
          end;
        until false;
        if Mother^.Debug.Verbose then AddLog('    %0..%1 (%2K): %3',
          [current, (current - 1) + gapsz, gapsz div 1024, RuEn(
          'неизвестно (отказано в доступе)', 'unknown (access denied)')]);
        MTSize[meUnknown]+= gapsz;
        if not Assigned(next) then break;
        current:= next;
      end
      else begin
        Des:= '';
        Case mm.State of
         MEM_RESERVE: begin
             Des:=RuEn('зарезервирована', 'reserved');
             MTSize[meReserved]+= mm.RegionSize;
           end;
         MEM_FREE: begin
             Des:= RuEn('свободна', 'free');
             MTSize[meFree]:= MTSize[meFree] + mm.RegionSize;
           end;
         MEM_COMMIT: begin
            MTSize[meCommitted]+= mm.RegionSize;
            if
              {$ifdef che_unicode}
                GetModuleFileNameW
              {$else}
                GetModuleFileName
              {$endif}
                  (THandle(mm.AllocationBase), @TST, SizeOf(TST) div SizeOf(TST[0]))
                <> 0
              then Des:= PervertedFormat(RuEn('выделена %0','committed by %0'), [PFileNameCharToFileNameString(@TST)])
              else Des:= PervertedFormat(RuEn('выделена (база %0)','committed (base %0)'), [mm.Allocationbase]);

            If (mm._Type <> MEM_PRIVATE) then begin
              Des+= RuEn(', общая',', shared');
              MTSize[meShared]+= mm.RegionSize;
            end;
            If (Protect and PAGE_NOCACHE > 0) then begin
              Des+= RuEn(', некэшируемая',', non-cached');
              MTSize[meNoCache]+= mm.RegionSize;
            end;
            If (Protect and PAGE_GUARD > 0) then begin
              Des+= RuEn(', охраняемая',', guarded');
              MTSize[meGuarded]+= mm.RegionSize;
            end;
            Des+= ', ';//RuEn(', доступ: ',', access: ');
            Protect:= mm.Protect and not (PAGE_NOCACHE + PAGE_GUARD);
            case Protect of
              PAGE_READONLY: Des+= RuEn('только чтение','read only');
              PAGE_READWRITE: Des+= RuEn('чтение/запись','read/write');
              PAGE_WRITECOPY: Des+= RuEn('копируется при записи','copy-on-write');
              PAGE_EXECUTE: Des+= RuEn('только выполнение кода','execute only');
              PAGE_EXECUTE_READ: Des+= RuEn('чтение и выполнение кода','read and execute');
              PAGE_EXECUTE_READWRITE: Des+= RuEn('полный','full');
              PAGE_EXECUTE_WRITECOPY: Des+= RuEn('выполнение кода/копируется при записи','execute/copy-on-write');
              PAGE_NOACCESS: Des+= RuEn('запрещён','denied');
            else
              Des+= RuEn('неизвестен','unknown');
            end;
          end;
        end;

        if Mother^.Debug.Verbose then AddLog('    %0..%1 (%2K): %3',
          [mm.BaseAddress, mm.BaseAddress + mm.RegionSize - 1, mm.RegionSize div 1024, Des]);
        if pointer(MaxAddr) - mm.RegionSize <= mm.BaseAddress then break;
      end;

      current+= mm.RegionSize;
    until false;
    MTName[meReserved]:= RuEn('Зарезервировано','Reserved');
    MTName[meCommitted]:= RuEn('Выделеннная память','Committed');
    MTName[meFree]:= RuEn('Свободно','Free');
    MTName[meUnknown]:= RuEn('Нет данных, отказано в доступе','Unknown, access denied');
    MTName[meShared]:= RuEn('Память, общая с другими процессами','Shared memory');
    MTName[meGuarded]:= RuEn('Охраняемая память','Guarded memory');
    MTName[meNoCache]:= RuEn('Некешируемая память','No-cache memory');
    Des:= '';
    AddLog(RuEn(
    {$ifdef cpu32}
      '  4 Гб адресного пространства делятся следующим образом:',
      '  Of the 4 Gb total adress space, there are:'
    {$else}
      '  8 Тб адресного пространства делятся следующим образом:',
      '  Of the 8 Tb total adress space, there are:'
    {$endif}
    ));
    if Mother^.Debug.Verbose
      then mimax:= High(TMMEnum)
      else mimax:= meFree;
    for mi:= Low(TMMEnum) to mimax do
      if MTSize[mi] > 0 then AddLog(RuEn(
        '    %0: %1 Гб',
        '    %0: %1 Gb'),
        [MTName[mi], FormatFloat('0.00', MTSize[mi] / (1024*1024*1024))]);
  end;
 {$else windows}
  procedure LogProcessAddressSpace; cdecl;
  begin
    if Mother^.Debug.Verbose then AddLog(RuEn(
      '  Адресное пространство процесса:'#10#13'    Нет данных, поддерживается только для MS Windows.',
      '  The address space of the process:'#10#13'    Sorry, this is only supported for MS Windows.'));
  end;
 {$endif windows}



 // The image files used by the mother module:
 {$include iim_00.inc}
// {$include cl_j2k_test.inc}
 {$include cl_bgimg_error.inc}
// {$include cl_console_font.inc}
 {$include cl_cursor_cyan.inc}
 {$include cl_cursor_yellow.inc}
 {$include cl_cursor_hourglass.inc}

 {$include cl_mother_api.inc}  //the functions the mother exe exports into the game dll


  Procedure PreInitDie(Yell: UnicodeString; p: array of const);
  //all error messages given before MessageContainer is initialized
  // are in English, so there's no need for unicode...
  begin
    Yell:=PervertedFormat(Yell, p);
   {$ifdef unix}
    WriteLn(#10#13#10#13#10#13 + ExtractFileName(CGEString + ' crashed at startup!'));
    Writeln(#10#13 + Yell + #10#13);
   {$endif}
    ErrorMessageBox(CGEString + ' crashed при запуске!', Yell);
    Halt(0);//1);
  end;

  function ExtractDyingYell(pes: PMotherSehState = nil): UnicodeString;
  var
    i: integer;
  begin
    Result:='';
    if not Assigned(pes) then pes:= GetExceptionState();
    For i:= High(pes^.Warnings) downto 0 do begin
      Result:=Result + pes^.Warnings[i];
      if i > 0 then Result:=Result + #10#13#10#13;
    end;
    SetLength(pes^.Warnings, 0);
  end;

  procedure DisplayDyingYells(DyingYell: UnicodeString = '');
  var
    tit: UnicodeString;
    titA: AnsiString;
  begin
    if DyingYell = '' then begin
      DyingYell:=ExtractDyingYell();
      AddLog(MI_ERROR_MESSAGE_IS, [DyingYell]);
    end;
    {$ifdef unix}
     WriteLn(#10#13#10#13#10#13 + ExtractFileName(ParamStr(0) + ' crashed!'));
     Writeln(#10#13 + DyingYell + #10#13);
    {$endif}
     if Mother^.State.IsServer then begin
       tit:= MessageContainer[MI_CGE_TITLE] + RuEn(' (режим сервера',' (server mode');
       if Assigned(Module) and (Mother^.Module.NameW <> '')
         then tit+= RuEn(', модуль ',', module ') + Mother^.Module.NameW;
       tit+= ')';
     end
     else
       tit:=MessageContainer[MI_CGE_TITLE];
     ErrorMessageBox(tit, DyingYell);
  end;

  {$include un_lineinfo.inc} //moved this one waaaaay back here
  // because DetectWindowsNTVersion -> LoadProcAddress requires GetModuleByAddr

(* DOESNT WORK, GIVES FALSE NEGATIVE
  {$ifdef windows}
  function TryWin98LacksUnicodeSupport(): boolean;
  var
    a: ansistring;
    w: UnicodeString;
    i: integer;
  begin
    SetLength(a, 1);
    a[1]:='i';
    i:=MultiByteToUnicodeChar(CP_ACP, MB_PRECOMPOSED, @a[1], length(a), nil, 0);
    if i > 0 then begin
      SetLength(w, i);
      MultiByteToUnicodeChar(CP_ACP, MB_PRECOMPOSED, @a[1], length(a), @w[1], i);
      Result:= (w[1] <> 'i');
    end
    else Result:= Yes;
  end;
 {$endif} *)


  function _RunningInWindows98: boolean;
  begin
   {$ifdef windows}
    Result := (Win32Platform = VER_PLATFORM_WIN32_WINDOWS) {and
      (Win32MajorVersion < 5)} {and (WIN32MinorVersion <= 1)}
   {$else}
    Result:=No;
   {$endif}
  end;
  
  function _RunningInWindowsNT: boolean;
  begin
   {$ifdef windows}
    Result := (Win32Platform =  VER_PLATFORM_WIN32_NT)
   {$else}
    Result:=No;
   {$endif}
  end;


  function _RunningInMacOSX: boolean;
  begin
   {$ifdef unix}
    Result:= No; //********** IMPLEMENT ME *********************
   {$else}
    Result:=No;
   {$endif}
  end;

  {$ifdef windows}
  function GetLocalAppDataDir: TFileNameString; forward;

  function DetectWindowsNTVersion(): TOSType;

  {
    Dec. 2016: the behavior of Windows changed again,
    now WIN32MajorVersion really stopped working, gives me Windows 8
    even with a proper manifest

    October 2016: And again Micro$oft makes me do unnecessary work
    by changing their specification back and forth on the fly.
    They promised to obsolete the standard OS version detection,
    then backed off on their promise.

    WIN32MinorVersion / WIN32MajorVersion still work, even as
    you need a proper manifest to detect Win10.

  }


 {
https://msdn.microsoft.com/en-us/library/windows/desktop/ms724451%28v=vs.85%29.aspx
[GetVersionEx may be altered or unavailable for releases after Windows 8.1.
Instead, use the Version Helper APIs]

With the release of Windows 8.1, the behavior of the GetVersionEx API has
changed in the value it will return for the operating system version. The value
returned by the GetVersionEx function now depends on how the application
is manifested.

Applications not manifested for Windows 8.1 or Windows 10 will return the
Windows 8 OS version value (6.2). Once an application is manifested for a given
operating system version, GetVersionEx will always return the version that the
application is manifested for in future releases.

    }

{$ifndef nowin10}
  {$if not defined(OSVersionInfoEX)}
     type
      OSVersionInfoEX = record
          dwOSVersionInfoSize : DWORD;
          dwMajorVersion      : DWORD;
          dwMinorVersion      : DWORD;
          dwBuildNumber       : DWORD;
          dwPlatformId        : DWORD;
          szCSDVersion        : array[0..127] of AnsiChar;
          wServicePackMajor   : WORD;
          wServicePackMinor   : WORD;
          wSuiteMask          : WORD;
          wProductType        : BYTE;
          wReserved           : BYTE;
        end;
  {$endif}
  var
    productName: UnicodeString;
  {$if defined(win32) and (FPC_FULLVERSION<20700)}
    var
      VerifyVersionInfo: function(
        var lpVersionInfo:  OSVersionInfoEX;
        dwTypeMask: dword;
        dwlConditionMask: qword
      ): boolean; stdcall;
      VerSetConditionMask: function(
        dwlConditionMask: qword;
        dwTypeBitMask: dword;
        dwConditionMask: byte
      ): qword; stdcall;
//    ULONGLONG WINAPI VerSetConditionMask(
//      _In_ ULONGLONG dwlConditionMask,
//      _In_ DWORD     dwTypeBitMask,
//      _In_ BYTE      dwConditionMask
//    );
      //BOOL WINAPI VerifyVersionInfo(
      //  _In_ LPOSVERSIONINFOEX lpVersionInfo,
      //  _In_ DWORD             dwTypeMask,
      //  _In_ DWORDLONG         dwlConditionMask
      //);

      dll: THandle;
      p: pointer;
    const
      VER_MAJORVERSION = $0000002;
      VER_MINORVERSION = $0000001;
      VER_GREATER_EQUAL = 3;



      function LoadProcAddress(hdll: THandle; ProcName: ansistring) : pointer;
      var
        baseaddr: pointer;
        exename: UnicodeString;
      begin
        Result:= GetProcAddress(hdll, PAnsiChar(ProcName));
        if Assigned(Result) then begin
          GetModuleByAddr(Result, baseaddr, exename);
          VerboseLog('      %0() at %1 in %2', [ProcName, Result, exename]);
        end
        else
          VerboseLog('      %0() not found!', [ProcName]);
      end;
  {$endif}
    var
      IsWindows8Point1OrGreater: function(): boolean; stdcall;
      IsWindows10OrGreater: function(): boolean; stdcall;
{$endif}
    var gv_win10CrapLoaded: boolean = false;

    {$if not defined(REG_QWORD)}
    const REG_QWORD = 11;
    {$endif}

    function IsOrGreater(wMajorVersion, wMinorVersion: Word): Boolean;
    {$ifndef nowin10}
    var
      osvi: OSVersionInfoEX;
      dwlConditionMask: qword;
    {$endif}
    begin
     {$if defined(win32) and not defined(nowin10)}
      if gv_win10CrapLoaded then begin
        FillChar(osvi, sizeof(osvi), 0);
        osvi.dwOSVersionInfoSize := SizeOf(osvi);
        osvi.dwMajorVersion := wMajorVersion;
        osvi.dwMinorVersion := wMinorVersion;
        dwlConditionMask := VerSetConditionMask(0, VER_MAJORVERSION, VER_GREATER_EQUAL);
        dwlConditionMask := VerSetConditionMask(dwlConditionMask, VER_MINORVERSION, VER_GREATER_EQUAL);
        Result := VerifyVersionInfo(osvi, VER_MAJORVERSION or VER_MINORVERSION, dwlConditionMask);
        VerboseLog(  '    %0.%1 or greater: %2',[wMajorVersion, wMinorVersion, Result]);
      end
      else
     {$endif}
        Result:= (WIN32MajorVersion > wMajorVersion)
          or ((WIN32MajorVersion = wMajorVersion) and (WIN32MinorVersion >= wMinorVersion))
    end;

    function urootname(root: UnicodeString): HKey;
    begin
      root:=uppercase(root);
      if root = 'HKEY_LOCAL_MACHINE' then Result:= HKEY_LOCAL_MACHINE
      else
      if root = 'HKEY_CLASSES_ROOT' then Result:= HKEY_CLASSES_ROOT
      else
      if root = 'HKEY_CURRENT_USER' then Result:= HKEY_CURRENT_USER
      else
      if root = 'HKEY_USERS' then Result:= HKEY_USERS
      else
      if root = 'HKEY_CURRENT_CONFIG' then Result:= HKEY_CURRENT_CONFIG
      else
        Result:= HKEY_LOCAL_MACHINE
    end;

    function GetRegKey(key: UnicodeString)
      : String; //I think it may be either AnsiString or UnicodeString, depending on FPC version
    var
      K: HKey;
      a: UnicodeStringArray;
      root, value: UnicodeString;
      i: integer;
      slen, typ: dword;

    begin
      Result:= '';
      try
        VerboseLog('      Querying registry key %0 ...',[key]);

        a:= Explode(UnicodeString('\'), key);
        root:= a[0];
        value:= a[High(a)];
        for i:= 1 to High(a) - 1 do a[i - 1]:= a[i]; //where's my array_shift(), bitch?!
        SetLength(a, Length(a) - 2);
        key:= Implode(UnicodeString('\'), a);

        if RegOpenKeyEx(urootname(root), PChar(String(key)), 0, KEY_QUERY_VALUE, K) = ERROR_SUCCESS then begin
          slen:= 10000;
          SetLength(Result, slen + 64);
          if RegQueryValueEx(K, PChar(String(value)), nil, @typ, @Result[1], @slen) <> ERROR_SUCCESS
          then begin
            VerboseLog('        Failed to read.');
            Result:= '';
          end
          else begin
            case typ of
              REG_SZ, REG_EXPAND_SZ: begin
                SetLength(Result, slen);
                if (Length(Result) > 0) and (Result[Length(Result)] = #0)
                  then SetLength(Result, Length(Result) - 1);
              end;
              REG_DWORD: Result:= IntToHex(pdword(@Result[1])^, 8);
              REG_QWORD: Result:= IntToHex(pqword(@Result[1])^, 16);
            else
              VerboseLog('        Not supporting parsing of this type of value');
              Result:= '';
            end;
            VerboseLog('        = "%0"', [Result]);
          end;
          RegCloseKey(K);
        end
        else begin
          VerboseLog('        Failed to open.');
        end;
      except
      end;
    end;

    function StringScan(s, sig: UnicodeString): boolean;
    var
      i: integer;
    begin
      Result:= false;
      VerboseLog('      ..scanning for signature string "%0"...', [sig]);
      for i:=1 to length(s) - length(Sig) do
        if copy (s, i, length(Sig)) = Sig then begin
          Result:=True;
          VerboseLog('        ..FOUND at offset %0!', [i]);
          Break;
        end;
    end;
  begin
    Result:= ostWin2k;
    try
      VerboseLog(  '  Testing Windows NT version...');
      {$ifndef nowin10}
        {$if defined(win32) and (FPC_FULLVERSION<20700)}
          gv_win10CrapLoaded:= true;
          VerboseLog('    loading kernel32.dll...');
          dll:= LoadLibrary('kernel32.dll');
          p:=  LoadProcAddress(dll, 'VerifyVersionInfoA');
          if not Assigned(p) then begin
            VerboseLog('    Falling back to legacy mechanisms.');
            gv_win10CrapLoaded:= false;
          end
          else
            pointer(VerifyVersionInfo):= p;
          if gv_win10CrapLoaded then begin
            p:=  LoadProcAddress(dll, 'VerSetConditionMask');
            if not Assigned(p) then begin
              VerboseLog('    Falling back to legacy mechanisms.');
              gv_win10CrapLoaded:= false;
            end
            else
              pointer(VerSetConditionMask):= p;
          end;
        {$else}
          // FPC 3.x has it included in the RTL
          gv_win10CrapLoaded:= true;
        {$endif}
      {$endif}
      if IsOrGreater(5, 1) then Result:= ostWinXP;
      if IsOrGreater(6, 0) then Result:= ostVista;
      if IsOrGreater(6, 1) then Result:= ostWin7;
      if IsOrGreater(6, 2) then Result:= ostWin8;
      if IsOrGreater(6, 3) then Result:= ostWin81;
      if IsOrGreater(10, 0) then Result:= ostWin10;
      if Result = ostWin8 then begin
        VerboseLog('    Seems like Windows 8. Let''s use dirty hacks to double-check this "fact"...');
        productName:= GetRegKey('HKEY_LOCAL_MACHINE\SOFTWARE\Microsoft\Windows NT\CurrentVersion\ProductName');
        if StringScan(productName, 'Windows 8.1') then begin
          VerboseLog('      HA! Take that, Micro$oft!');
          Result:= ostWin81;
        end;
        if StringScan(productName, 'Windows 10') then begin
          VerboseLog('      HA! Take that, Micro$oft!');
          Result:= ostWin10;
        end;
      end;
      VerboseLog(  '    Result: %0', [GetEnumName(typeinfo(TOSType), ord(result))]);
    except
    end;
  end;

  {$endif}



  procedure DetectOSType;


  begin
    VerboseLog('Detecting operating system...');
   {$ifdef unix}
     {$ifdef darwin}
      Mother^.State.OS:= ostMacOSX;
     {$else}
      Mother^.State.OS:= {$ifdef cpuarm}ostLinuxARM{$else}ostLinux{$endif};
     {$endif}
   {$else}
     if RunningInWine then begin
       Mother^.State.OS:= ostWine;
       Mother^.State.DriveZIsRoot:= GetZIsRoot();
       if Mother^.State.DriveZIsRoot then begin
         //otherwise we're unable to tell Linux from MacOS X
         if FileExists('z:\usr\lib\libc.dylib')
           then Mother^.State.OS:= ostDarwine;
       end;
     end
     else
       if _RunningInWindows98 then begin
         Mother^.State.OS:= ostWin98; //98 or Me
         VerboseLog('  It''s Windows 98 or Millennium, v%0.%1.', [Win32MajorVersion, WIN32MinorVersion]);
       end
       else begin
         VerboseLog('  It''s Windows NT %0.%1.', [Win32MajorVersion, WIN32MinorVersion]);
         case Win32MajorVersion of
           3,4: Mother^.State.OS:= ostWin2k;  //Windows NT 3.x and 4.x are treated as Windows 2000. Chentrah probably won't event start there
{           5: if WIN32MinorVersion = 0
                then Mother^.State.OS:= ostWin2k
                else Mother^.State.OS:= ostWinXP; //or server 2003, doesn't matter
           6: case WIN32MinorVersion of
                0: Mother^.State.OS:= ostVista;
                1: Mother^.State.OS:= ostWin7;
                2: Mother^.State.OS:= ostWin8;
              else
                Mother^.State.OS:= ostWin8;
              end;
}
         else
           Mother^.State.OS:= DetectWindowsNTVersion();
         end;
       end;

    Mother^.State.CSIDL_AppDataDir:= GetLocalAppDataDir;
    VerboseLog('  Local app data directory is %0', [GetLocalAppDataDir]);

    Mother^.State.Windows98.DoesNotSupportUnicode:= {$ifdef cpu64} false {$else} not CheckIfWindows98HasUnicodeSupportInstalled() {$endif} ;
   {$endif}
    Mother^.State.OstID:= GetEnumName(typeinfo(TOSType), ord(Mother^.State.OS));
   {$ifdef windows}
    if Mother^.State.OS = ostWine
      then Mother^.State.OSName:= WineVersionString
      else
   {$endif}
//        if (Win32MajorVersion < 5) or (Win32MajorVersion > 6)
//        or ((Win32MajorVersion = 6) and (WIN32MinorVersion > 1))
//          then Mother^.State.OSName:= 'Windows NT ' + IntToStr(Win32MajorVersion) + '.' + IntToStr(WIN32MinorVersion)
//          else
        Mother^.State.OSName:= OSDisplayedName[Mother^.State.OS];

   {$ifdef unix}
    {$ifndef darwin}
     DetectLinuxVersionString;
    {$endif}
   {$endif}
  end;

  var
    CachedVerboseLog: array of UnicodeString = nil;

  //key words to remove from the processor name:
  const ccpun_keywords: array [0..8] of ansistring = ('@', '(r)', 'amd', 'cpu', '(tm)', 'intel', 'dual core', 'dual-core', 'processor');

  function CleanCPUName (s: ansistring): ansistring;
  var
    i, n, k: integer;
    prevc: ansichar;
    b: boolean;
    o: ansistring;
  begin
    if s = '' then exit;
    o:= lowercase(s);
    //first, erase the keywords:
    for i:= 1 to length(s) do
      for k:= 0 to high(ccpun_keywords) do begin
        b:= true;
        for n:= 1 to length(ccpun_keywords[k]) do
          if ((i + n - 1) > length (s)) or (o[i + n - 1] <> ccpun_keywords[k][n]) then begin
            b:=false;
            break;
          end;
        if b then for n:= 1 to length(ccpun_keywords[k]) do
          s[i + n - 1]:= ' ';
      end;
    //then remove all the extra unnecessary spaces
    Result:= '';
    prevc:= ' ';
    for i:= 1 to length(s) do begin
      if (prevc > ' ') or (s[i] > ' ') then Result+= s[i];
      prevc:= s[i];
    end;
    Result:= Trim(Result);
  end;


  {$if defined(CPU386) or defined (CPUX86_64)}

  function GetProcessorInfo: string;
  var
    A, ffa, ffb, ffd, family, model, stepping, lvl, elvl: dword;
    vstr: string[12];
    dstr: string[48] = '';
   {$ifdef windows}
    si: SYSTEM_INFO;
   {$endif}
  begin
    setlength(vstr, 12);
    asm
      mov eax, 0
      cpuid
      mov lvl, eax
      mov dword[vstr + 1], ebx
      mov dword[vstr + 5], edx
      mov dword[vstr + 9], ecx

      mov eax, 1
      cpuid
      mov [ffa], eax
      mov [ffb], ebx
      mov [ffd], edx

      mov eax, 80000000h
      cpuid
      mov elvl, eax
    end ['eax', 'ebx', 'ecx', 'edx'];
    if elvl >= $80000004 then begin
      setLength(dstr, 48);
      asm
        mov eax, 80000002h
        cpuid
        mov dword[dstr + 1], eax
        mov dword[dstr + 5], ebx
        mov dword[dstr + 9], ecx
        mov dword[dstr + 13], edx

        mov eax, 80000003h
        cpuid
        mov dword[dstr + 17], eax
        mov dword[dstr + 21], ebx
        mov dword[dstr + 25], ecx
        mov dword[dstr + 29], edx

        mov eax, 80000004h
        cpuid
        mov dword[dstr + 33], eax
        mov dword[dstr + 37], ebx
        mov dword[dstr + 41], ecx
        mov dword[dstr + 45], edx
      end;
    end;
    if elvl >= $80000006 then begin
      asm
        mov eax, 80000006h
        cpuid
        mov [A], ecx
      end;
      Mother^.CPU.L2CacheSize:= A shr 16;
      Mother^.CPU.L2CacheLineSize:= A and $ff;
    end;
    if elvl >= $80000007 then begin
      asm
        mov eax, 80000007h
        cpuid
        mov [A], edx
      end;
      Mother^.CPU.TSCInvariance:= ((A shr 8) and $00000001) > 0;
    end;
    if (ffd and (1 shl 28)) > 0  then begin
      Mother^.CPU.Count:= (ffb shr 16) and $00ff;
      (*
      if (Mother^.CPU.Count > 1)
        and (vstr = 'GenuineIntel')
        then Mother^.CPU.Count:= Mother^.CPU.Count div 2;
        {
          Fuck. There's no *sane* way to get the number of PHYSICAL processors on Intel CPUs.
          So I simply divide the number of logical cores by 2 if HT is supported.
          I'm not checking if it is really enabled (do I have to? does Windows ever disable it?)
        }
      *)
      Mother^.CPU.Count:= Math.Min(32, Math.Max(1, Mother^.CPU.Count)); //sanity check
      if (Mother^.CPU.L2CacheSize > 0)
      and ((Mother^.CPU.L2CacheSize div Mother^.CPU.Count) < 256)
      and (Mother^.CPU.Count > 2) then begin
        Mother^.CPU.Count:= Mother^.CPU.Count div 2;
      end;
    end
    else
      Mother^.CPU.Count:= 1;
    family:= (ffa shr 8) and $0f;
    if family = $0f then family+= ((ffa shr 20) and $ff);
    model:= (ffa shr 4) and $0f;
    if model = $0f then model+= ((ffa shr 16) and $0f);
    stepping:= ffa and $0f;
//    Mother^.State.SSE2Available:= ((ffd and (1 shl 26)) > 0);
    Result:= CleanCPUName(dstr);
    if Result = '' then begin
      if vstr = 'GenuineIntel'
        then Result:='Intel'
        else
          if vstr = 'AuthenticAMD'
            then result:='AMD'
            else result:= vstr;
      result+= ' ' + IntToStr(family)+'.'+IntToStr(model)+'.'+IntToStr(stepping);
    end;
    Mother^.CPU.Name:= Result;
    {$ifdef windows}
      windows.GetSystemInfo(@si);
      if (si.dwNumberOfProcessors > 0) and (si.dwNumberOfProcessors < Mother^.CPU.Count) then begin
        Mother^.CPU.Count:= si.dwNumberOfProcessors;
      end;
    {$else}
{
        Re: Проблема в многопоточности

        Непрочитанное сообщение Vadim » 17.02.2015 02:39:36
        qivi
        Или, если речь идёт о Линуксе - заглянуть в файл /proc/cpuinfo.
        Количество названий процессора будет ровно столько, сколько у него ядер.
}
    {$endif}

{
        Re: Проблема в многопоточности

        Непрочитанное сообщение Sergei I. Gorelkin » 19.02.2015 18:32:27
        Штатное появилось в 3.1.1:
        class property TThread.ProcessorCount: longword;

        Но реализация на чем-то экзотическом остается под вопросом.
}

  end;

{$endif} //intel CPUs

{$ifdef cpuarm}
  function GetProcessorInfo: string;
    var
      sl: TStringList;
      st: TFileStream;
      a: array of Utf8String;
      s: Utf8String;
      x: integer;
    function Find(ind: Utf8String; FindLast: boolean = false): Utf8String;
    var
      k: integer;
      m: array of Utf8String;
    begin
      Result:= '';
      k:= 0;
      while k < sl.Count do begin
        m:= Explode(Utf8String(':'), sl[k], true);
        if Length(m) = 2 then begin
          if m[0] = ind then begin
            Result:= m[1];
            if not FindLast then Break;
          end;
        end;
        Inc(k);
      end;
    end;
    function FindInt(ind: Utf8String; default: integer; FindLast: boolean = false): integer;
    var g: Utf8String;
    begin
      g:= Find(ind, FindLast);
      if g = '' then Exit(default);
      try
        Result:= StrToInt(g);
      except
        Exit(default);
      end;
    end;

  begin
    sl:= TStringList.Create;
    try
      try
        st:= TFileStream.Create('/proc/cpuinfo', fmOpenRead);
        sl.LoadFromStream(st);
        st.Free;

        Result:= CleanCPUName(Find('model name'));
        if Result = '' then Result:= 'Unknown ' + {$I %FPCTARGETCPU%};

        Mother^.CPU.Count:= FindInt('cpu cores', 1 + FindInt('processor', 1, true));

        a:= Explode(Utf8String(' '), Find('cache size'), true);
        if (Length(a) >= 2) and (a[1] = 'KB') then try
          Mother^.CPU.L2CacheSize:= StrToInt(a[0]);
        except
        end;

        Mother^.CPU.L2CacheLineSize:= FindInt('cache_alignment', 0);


      except
        Result:= '<UNKNOWN ' + {$I %FPCTARGETCPU%} + '>';
        Mother^.CPU.Count:= 2;
      end;
    finally
      sl.Free;
    end;
    Mother^.CPU.Name:= Result;
  end;

{$endif} // CPU ARM

  function NumeralFormRu (n: integer; wSingular, wPlural2_4, wPlural: UnicodeString): UnicodeString;
  begin
    n:= n mod 100; //that's right, the singular form is used for 21, 101 and such
    if n > 19 then n:= n mod 10;
    case n of
      1: Result:= wSingular;
      2..4: Result:= wPlural2_4;
    else
      Result:= wPlural;
    end;
  end;

  function NumeralFormEn (n: integer; wSingular, wPlural: UnicodeString): UnicodeString;
  begin
    if n = 1
      then Result:= wSingular
      else Result:= wPlural;
  end;


  procedure LogOsVersion;
  var
    w, oops: UnicodeString;
    nc, nl, ni: UnicodeString;
  begin
    w:= PervertedFormat(MessageContainer[MI_OS_VERSION], [Mother^.State.OSName]);
  {$ifdef windows}
    if Mother^.State.Windows98.DoesNotSupportUnicode
      then w+= RuEn(' (без поддержки Unicode)',' (with no Unicode support)');
  {$endif}
    AddLog(w);
    GetProcessorInfo();
    AddLog(RuEn('Пользователь: %0','User name: %0%'), [Mother^.State.UserNameUCS16]);
    if Mother^.CPU.L2CacheSize > 0
      then nc:= IntToStr(Mother^.CPU.L2CacheSize) + RuEn(' Кб',' Kbytes')
      else nc:= RuEn(' размер неизвестен',' unknown size');
    if Mother^.CPU.L2CacheLineSize > 0
      then nl:= IntToStr(Mother^.CPU.L2CacheLineSize)
        + RuEn(NumeralFormRu(Mother^.CPU.L2CacheLineSize, ' байт',' байта',' байт'),' bytes')
      else nl:= RuEn(' неизвестно',' unknown');
    {$ifndef cpuarm}
    if Mother^.CPU.TSCInvariance
      then ni:= RuEn('есть','yes')
      else ni:= RuEn('не поддерживается','not supported');
    {$endif}
    {$ifdef cpuarm}
    AddLog(RuEn(
     'ЦП %0'#10#13'  x%1 '
         + NumeralFormRu(Mother^.CPU.Count,'логическое ядро','логических ядра','логических ядер')
         + #10#13'  кэш второго уровня:%2, линейка%3',
     'CPU %0'#10#13'  x%1 '
         + NumeralFormEn(Mother^.CPU.Count, 'logical core', 'logical cores')
         + #10#13'  level 2 cache: %2, line size %3'
      ),
      [ Mother^.CPU.Name,
        Mother^.CPU.Count,
        nc,
        nl]);
    {$else cpuarm}
    AddLog(RuEn(
     'ЦП %0'#10#13'  x%1 '
         + NumeralFormRu(Mother^.CPU.Count,'логическое ядро','логических ядра','логических ядер')
         + #10#13'  кэш второго уровня: %3, линейка %4'
         + #10#13'  инвариантность TSC: %5'
         + #10#13'  частота TSC: %2 ГГц.',
     'CPU %0'#10#13'  x%1 '
         + NumeralFormEn(Mother^.CPU.Count, 'logical core', 'logical cores')
         + #10#13'  level 2 cache: %3, line size %4'
         + #10#13'  TSC invariancy: %5'
         + #10#13'  TSC frequency: %2 GHz)'
      ),
      [ Mother^.CPU.Name,
        Mother^.CPU.Count,
        format('%1.2f', [Mother^.Timer.RDTSCFrequency / 1000000000.0]),
        nc,
        nl,
        ni]);

    {$endif}
    AddLog(RuEn('ОЗУ: %0 Мб','RAM: %0 Mb'), [Mother^.State.PhysMemory]);

    LogProcessAddressSpace;

    oops:= '';

{    if not Mother^.State.SSE2Available
      then oops+= RuEn(
       ' - не поддерживается набор команд SSE2'#10#13#10#13,
       ' -- no support for the SSE2 instruction set.'#10#13#10#13);
}

    if Mother^.CPU.Count < 2
      then oops+= RuEn(
       ' - только одно ядро'#10#13#10#13,' -- has only one core'#10#13#10#13);
{   Bull. Shit. I was using faulty cache detection methods.

    if Mother^.CPU.L2CacheSize < 512
      then oops+= RuEn(
      ' - размер кеша второго уровня меньше 512 килобайт (имеется '+ nc +')'#10#13#10#13,
      ' -- level 2 cache is smaller than 512 kilobytes (its size is '+nc+')'#10#13#10#13);
}
    if oops <> '' then
      GiveWarning('_cpu_is_feature_weak', RuEn('МОЖЕТ НЕ ПОТЯНУТЬ.','NOT GOOD.'),
         RuEn(
           'Процессор не поддерживает некоторые необязательные,'#10#13 +
           '  но влияющие на производительность технологии:',
           'Your CPU lacks some features'#10#13 +
           '  that are optional but still desired for better performance:')
         + #10#13#10#13 + oops, []);
  end;
  
{$ifdef unix}
  var
    Fn: string;
    Di: boolean = false;

  function ThisIsAnOnlyInstance: boolean;
  var
    i: integer;
    t: Text;
  begin
    Result:=True;
    Fn:= '/tmp/.' + ChangeFileExt(ExtractFileName(ParamStr(0)),'') + 'SingleInstanceMutex';
    if Mother^.State.IsServer then Fn+= '_srv_' + Mother^.State.ServerModuleName;
    Try
      if FileExists(Fn) then begin
        i:=0;
        AssignFile(t, Fn);
        Reset(t);
        Readln(t, i);
        CloseFile(t);
        if i = fpGetPid() then Exit
        else
          if DirectoryExists('/proc/' + IntToStr(i))
            and (
              (ExtractFileName(fpReadLink ('/proc/' + IntToStr(i) + '/exe')) = ExtractFileName(ParamStr(0)))
              //safety check:
              // if we can't determine if the PID belongs to the same app,
              // then we assume it does.
              or not FileExists('/proc/' + IntToStr(i) + '/exe') //different unix
              or ((Length(ParamStr(0)) >=8) and (copy(ParamStr(0),1, 8) = '/tmp/upx'))) //executable is upx'ed
            then
              Exit(False);
      end;
      AssignFile(t, Fn);
      Rewrite(t);
      Writeln(t, fpGetPid());
      CloseFile(t);
    Except
    End;
    Di:= Result;
  end;
{$else}
  var
    M: THandle;
    Di: boolean = false;

  function ThisIsAnOnlyInstance: boolean;
  var N: Utf8String;
  begin
    Result:=True;

    N:= Utf8Encode(ChangeFileExt(ExtractFileName(ParamStr(0)),'') + 'SingleInstanceMutex');
    if Mother^.State.IsServer then N+= Utf8Encode('_srv_' + Mother^.State.ServerModuleName);
    M:= OpenMutex(MUTEX_MODIFY_STATE, False, PAnsiChar(N));
    if M = 0 then M:= CreateMutex(nil, True, PAnsiChar(N))
    else begin
      if WaitForSingleObject(M, 0) <> WAIT_ABANDONED then Result:=False;
    end;
    Di:=Result;
  end;
{$endif}


  Procedure DbgSayS(Yell: AnsiString);
  begin
   {$ifdef windows}
    MessageBox(0, PAnsiChar(Yell), PAnsiChar(AnsiString('Ы')), MB_ICONINFORMATION + MB_OK);
   {$else}
//    UnixPMMessageBox(AnsiToWide(Yell), '---Ok---');
   {$endif}
  end;

 {$include un_unicode.inc}

  function LoadUnicodeText(FileName: string): TUnicodeStringContainer;
  var
    i: integer;
    s: TStringList;
  begin
    if not FileExists(FileName)
    then
      Die(MI_ERROR_FILE_NOT_FOUND, [FileName]);
    Try
      s:=TStringList.Create;
      s.LoadFromFile(FileName);
      Result:= TUnicodeStringContainer.Create;
      For i:=0 to s.Count - 1 do begin
        //if (Result.high < 0) and (Trim(s[i]) = '') then continue;
        Result.Add({chTRimRightSpaces}(Utf8Decode(s[i])));
//        if (Result.high = 0) and (Result.last[1] = #$fffe)
//          then Result.last:=copy(Result.last, 2, length(Result.last) - 1);
        if (Result.last = '.') then Result[Result.High]:= '';
      end;
      s.Free;
    Except
      Die(MI_ERROR_INVALID_UTF8_TEXT,[FileName])
    End;
  end;

  {$include cl_texsages.inc}

  {$include un_api_common.inc}
  
  function StrOrUndefined(U: UnicodeString): UnicodeString;
  begin
    if U='' then Result:=MessageContainer[MI_UNDEFINED]
            else Result:=U;
  end;

 {$ifdef win32}
  function Win32LastError: UnicodeString;
  var
    M: Cardinal;
    u: UnicodeString;
  begin
    M:=GetLastError;
    if M = ERROR_SUCCESS then Result:=''
    else begin
      if Mother^.State.OS = ostWin98
        then u:=AnsiToWide(SysErrorMessage(M)) //for Win 98, always assume cp1251
        else u:=SysErrorMessage(M);
      Result:=#10#13 + Result + #10#13 + PervertedFormat(
        MessageContainer[MI_WIN32_EXPLAINS],
        [IntToHex(M, 8), u]);
    end;
  end;
 {$endif}

    function ExceptObjectisException: boolean;
    begin
//      Try
        Result:=Assigned(ExceptObject) and (ExceptObject is Exception);
//      Except
//        Result:=No;
//      End;
    end;

(*  var
    EChar: string;
  procedure SetEChar;
  begin
    //1 - "естественное" исключение (не Die() )
    //2 - в модуле
    //3 - внутри экспортируемой процедуры
    //4 - модуль не загружен
    EChar:='----';
    if not Mother^.State.NowDying and ExceptObjectisException
       and not (ExceptObject is EFake) then EChar[1]:='*';
    if Mother^.Module.CurrentlyExecutingInd <> NOT_A_MODULE then EChar[2]:='*';
    if ActiveOwner = NOT_A_MODULE then EChar[4]:='*';
    if InExportedProc then EChar[3]:='*';
        {$ifdef win32}
//    if GetLastError <> 0 then EChar[3]:='*';
        {$endif}
  end;
*)


var
  zz0p: pointer = nil;

  {$include cl_die.inc}

  procedure StartDying(u: UnicodeString);
  var
    pes: PMotherSehState;
  begin
    pes:= GetExceptionState();
    if not pes^.NowDying then begin
      AddLog(MI_SHIT_HAPPENS_SEE_BELOW, ['----']); //*TODO: remember what the fuck this ['----'] was for.
      pes^.NowDying:= true;
    end;
//bgSay('* StartDying, pes='+IntToHex(dword(pes), 8)+' str='+ u);

    SetLength(pes^.Warnings, Length(pes^.Warnings) + 1);
    pes^.Warnings[High(pes^.Warnings)]:= u;
//bgSay('  Length(pes^.Warnings)=' + IntToStr(Length(pes^.Warnings)));
    //WarningQueue.Add(u);
  end;
  
{  Procedure Warning(YellID :TMessageID; Param: array of const); OVERLOAD;
  var
    U: UnicodeString;
    pes: PMotherSehState;
  begin
    pes:= GetExceptionState();
    U:= PervertedFormat(MessageContainer[YellID], Param);
    SetLength(pes^.Warnings, Length(pes^.Warnings) + 1);
    pes^.Warnings[High(pes^.Warnings)]:= u;
//    WarningQueue.Add(U);
  end;
}

  type
    TWstrArray = array of UnicodeString;

  function VarrecsToWstrs(param: array of const): TWstrArray;
  var
    i: integer;
  begin
    SetLength(Result,length(param));
    for i:=0 to high(param) do Result[i]:= VarRecToWide(param[i]);
  end;

  Procedure AddStopMessage(YellID :TMessageID; Param: array of const);
  begin
    SetLength(StopMessage, length(StopMessage) + 1);
    with StopMessage[high(StopMessage)] do begin
      capMID:= MI_STOP_MESSAGE_CAPTION;
      capStr:= '';
      msgMID:= YellID;
      msgStr:= '';
      params:= VarrecsToWstrs(Param);
    end;
  end;
  
  Procedure AddStopMessage(Yell: UnicodeString; Param: array of const);
  begin
    SetLength(StopMessage, length(StopMessage) + 1);
    with StopMessage[high(StopMessage)] do begin
      capMID:= MI_STOP_MESSAGE_CAPTION;
      capStr:= '';
      msgStr:= Yell;
      params:= VarrecsToWstrs(Param);
    end;
  end;
  
  Procedure AddStopMessage(Yell: UnicodeString);
  begin
    SetLength(StopMessage, length(StopMessage) + 1);
    with StopMessage[high(StopMessage)] do begin
      capMID:= MI_STOP_MESSAGE_CAPTION;
      capStr:= '';
      msgStr:= Yell;
      params:= nil;
    end;
  end;
  
  procedure AddMStopMessage(itisfatal: boolean);
  begin
    bimState:= bimsModuleError;
    with ModuleStopMessage do begin
      capStr:= PervertedFormat(MessageContainer[MI_MODULE_CRASHED_ONELINER], [Mother^.Module.NameW]);
      msgStr:= StopDying();
{      if not itisfatal
        then msgStr+= #10#13#10#13'······································'#10#13#10#13
        + MessageContainer[MI_MODULE_CRASHED_EXPLAIN];
}
//AddLog(capStr + #10#13 + MsgStr);
    end;
  end;
  
  
  Procedure AddWarning(Cap: TMessageId; Yell :UnicodeString; Param: array of const);
  begin
    SetLength(WarningMessage, length(WarningMessage) + 1);
    with WarningMessage[high(WarningMessage)] do begin
      capMID:= Cap;
      capStr:= '';
      msgStr:= Yell;
      params:= VarrecsToWstrs(Param);
      inipar:= '';
    end;
  end;

  Procedure AddWarning(Cap, Yell :UnicodeString; Param: array of const);
  begin
    SetLength(WarningMessage, length(WarningMessage) + 1);
    with WarningMessage[high(WarningMessage)] do begin
      capStr:= Cap;
      msgStr:= Yell;
      params:= VarrecsToWstrs(Param);
      inipar:= '';
    end;
  end;


  
  Procedure GiveWarning(_inipar: string; Cap, Yell :TMessageID; Param: array of const);
  begin
    if Config.Int['warnings', _inipar] = 1  then exit;
    SetLength(WarningMessage, length(WarningMessage) + 1);
    with WarningMessage[high(WarningMessage)] do begin
      capMID:= Cap;
      capStr:= '';
      msgMID:= Yell;
      msgStr:= '';
      params:= VarrecsToWstrs(Param);
      inipar:= _inipar;
      writeval:= 1;
      isyellow:= true;
    end;
  end;

  Procedure GiveWarningNotYellow(_inipar: string; Cap, Yell :TMessageID; Param: array of const);
  begin
    if Config.Int['warnings', _inipar] = 1  then exit;
    SetLength(WarningMessage, length(WarningMessage) + 1);
    with WarningMessage[high(WarningMessage)] do begin
      capMID:= Cap;
      capStr:= '';
      msgMID:= Yell;
      msgStr:= '';
      params:= VarrecsToWstrs(Param);
      inipar:= _inipar;
      writeval:= 1;
      isyellow:= false;
    end;
  end;



  procedure GiveWarning(_inipar: string; Cap, Yell :UnicodeString; Param: array of const);
  begin
    if Config.Int['warnings', _inipar] = 1  then exit;
    SetLength(WarningMessage, length(WarningMessage) + 1);
    with WarningMessage[high(WarningMessage)] do begin
      capStr:= Cap;
      msgStr:= Yell;
      params:= VarrecsToWstrs(Param);
      inipar:= _inipar;
      writeval:= 1;
      isyellow:= true;
    end;
  end;


  
  Procedure GivePeriodicWarning(DaysToRepeat: float; _inipar: string; Cap, Yell :TMessageID; Param: array of const);
  var
    dd: integer;
  begin
    dd:= round(Now() / daysToRepeat);
    if Config.Int['warnings', _inipar] = dd  then exit;
    SetLength(WarningMessage, length(WarningMessage) + 1);
    with WarningMessage[high(WarningMessage)] do begin
      capMID:= Cap;
      capStr:= '';
      msgMID:= Yell;
      msgStr:= '';
      params:= VarrecsToWstrs(Param);
      inipar:= _inipar;
      writeval:= dd;
      isyellow:= true;
    end;
  end;
  //function StopDying(pes: PMotherSehState = nil): UnicodeString;
//    moved to cl_die.inc

    var uckn: array [0..255] of string;

  function GetHotkeyName(var HK: TIntegerArray): string;
  var
    i: integer;
    kn: string;
  begin
    Result:='';
    For i:=0 to high(HK) do begin
      if i > 0 then Result += '+';
      kn:= GetEnumName(typeinfo(TKey), HK[i]);
      kn:= Copy(kn, 5, length(kn) - 4); //remove the 'KEY_' part
      Result += kn;
//addlog('---> %0  %1   %2  %3', [Result, kn, i, hk[i]]);
    end;
  end;

  procedure GetHotKeyFromConfig(par: string; var HK: TIntegerArray);
  var
    str, sp: string;
    i, j, n: integer;
  begin
    Try
      str:=Config.Str['main', par];
      i:=1;
      While StrParm(str, i, '+') <> '' do begin
        System.SetLength(HK, i);
        sp:=uppercase (Trim(StrParm(str, i, '+')));
        n:=0;
        For j:=1 to 255 do begin
          if uckn[j] = ''
            then uckn[j]:= uppercase(GetEnumName(typeinfo(TKey), j));
          if (sp = uckn[j] ) or ('KEY_' + sp = uckn[j] )
          then begin
            n:=j;
            break;
          end;
        end;
        if n = 0 then Die(MI_INVALID_HOST_HOTKEY_NAME, [uppercase(sp), Mother^.State.InstallPath]);
        HK[i - 1]:= n;
        inc(i);
      end;
    Except
      Die(MI_INVALID_HOST_HOTKEY_RECORD, [Config.Wherefoundname, 'main', par, str]);
    End;
  end;

  procedure GetHotKeysFromConfig;
  var
    i: integer;
  begin
    GetHotKeyFromConfig('Fullscreen_Hotkey', FullscreenHotkey);
    GetHotKeyFromConfig('Abort_Hotkey', AbortHotKey);
    GetHotKeyFromConfig('Restart_Hotkey', RestartHotKey);
    GetHotKeyFromConfig('Session_rollback_Hotkey', SessionRollbackHotKey);
    GetHotKeyFromConfig('Language_Hotkey', LanguageHotKey);
    GetHotKeyFromConfig('Console_Toggle_Hotkey', ConsoleToggleHotkey);
    GetHotKeyFromConfig('Console_Scroll_Up_Hotkey', ConsoleScrollUpHotkey);
    GetHotKeyFromConfig('Console_Scroll_Down_Hotkey', ConsoleScrollDownHotkey);
    GetHotKeyFromConfig('QualityFactorManualOverrideToggleHotkey', QualityFactorManualOverrideToggleHotkey);

    For i:=0 to 0 {MAX_MODULE} do GetHotKeyFromConfig('Module_'+IntToStr(i) + '_Hotkey', ModuleHotKey[i]);
  end;
  
  function TEngineObjects.AddObj(i: TObject): integer;
  begin
//byte(nil^):= 0;
    Result:= Self.Add(i);
    if Mother^.Debug.Enabled then AddLog('Added engine object #%0 (%1)', [Result + 1, AnsiString(i.ClassName)]);
  end;

  
  procedure VyRussifyResourceStrings;
  Var I,J : Longint;
      S, S1 : UnicodeString;
  begin
// FPC 2.2.0: /home/cheb/chentrah/modules/chentrah/src/cge.pp(1162,41)
// Error: Identifier not found "ResourceStringTableCount"

(*    For I:=0 to ResourceStringTableCount-1 do
      For J:=0 to ResourceStringCount(i)-1 do
        begin
          S:=AnsiToWide(GetResourceStringDefaultValue(I,J));
          S1:=S;
          if S = 'Vampyre Imaging Library' then S:='Vampyre Imaging Library';
          if S = 'Exception Message' then S:='Сообщение';
          if S = 'All Images' then S:='Все изображения';
          if S = 'Unknown and unsupported format' then S:='Неизвестный и неподдерживаемый формат';
          if S = 'Error while freeing image. %s' then S:='Ошибка при освобождении изображения. %s';
          if S = 'Error while cloning image. %s' then S:='Ошибка при клонировании изображения. %s';
{
          if S = 'Error while flipping image. %s' then S:='';
          if S = 'Error while mirroring image. %s' then S:='';
          if S = 'Error while resizing image.  %s' then S:='';
          if S = 'Error while swapping channels of image. %s' then S:='';
          if S = 'Image Format "%s" does not support loading images.' then S:='';
}
          if S = 'Image Format "%s" does not support saving images.' then S:='Для формата "%s" не поддерживается запись.';
{
          if S = 'Error while creating image data with params: Width=%d ' +
            'Height=%d Format=%s.' then S:='';
}
          if S = 'Error while converting image to format "%s". %s' then S:='Ошибка при преобразовании изображения в формат "%s". %s';
{
          if S = 'Image @%p info: Width = %dpx, Height = %dpx, ' +
            'Format = %s, Size = %.0nKiB, Bits @%p, Palette @%p.' then S:='';
}
          if S = 'Access violation encountered when getting info on ' +
            'image at address %p.' then S:='Нарушение прав доступа к памяти (Access violation) при получении информации о изображении по адресу %p.';
          if S = 'File "%s" is not valid image in "%s" format.' then S:='Файл "%s" не является исправным изображением формата "%s".';
{
          if S = 'Stream %p does not contain valid image in "%s" format.' then S:='';
          if S = 'Memory %p (%d Bytes) does not contain valid image ' +
            'in "%s" format.' then S:='';
}
          if S = 'Error while loading images from file "%s" (file format: %s).' then S:='Ошибка при загрузке изображения из файла "%s" (формат файла: %s)';
          if S = 'Error while saving images to file "%s" (file format: %s).' then S:='Ошибка при сохранении изображения в файл "%s" (формат файла: %s)';
{
          if S = 'Error while loading images from stream %p (file format: %s).' then S:='';
          if S = 'Error while loading images from memory %p (%d Bytes) (file format: %s).' then S:='';
          if S = 'Error while saving images to stream %p (file format: %s).' then S:='';
          if S = 'Error while saving images to memory %p (%d Bytes) (file format: %s).' then S:='';
          if S = 'Error while finding color in palette @%p with %d entries.' then S:='';
          if S = 'Error while filling grayscale palette @%p with %d entries.' then S:='';
          if S = 'Error while filling custom palette @%p with %d entries.' then S:='';
          if S = 'Error while swapping channels of palette @%p with %d entries.' then S:='';
          if S = 'Error while reducing number of colors of image to %d. %s' then S:='';
          if S = 'Error while generating %d mipmap levels for image %s' then S:='';
          if S = 'One or more images are not valid.' then S:='';
          if S = 'Error while copying rect from image %s to image %s.' then S:='';
          if S = 'Error while mapping image %s to palette.' then S:='';
          if S = 'Error while filling rectangle X:%d Y:%d W:%d H:%d in image %s' then S:='';
          if S = 'Error while splitting image %s to %dx%d sized chunks.' then S:='';
          if S = 'Error while making %d color palette for %d images.' then S:='';
          if S = 'Error while creating new palette with %d entries' then S:='';
          if S = 'Error while freeing palette @%p' then S:='';
          if S = 'Error while copying %d entries from palette @%p to @%p' then S:='';
          if S = 'Error while replacing colors in rectangle X:%d Y:%d W:%d H:%d of image %s' then S:='';
          if S = 'Error while rotating image %s by %d degrees' then S:='';
          if S = 'Error while stretching rect from image %s to image %s.' then S:='';
}
          If S1 <> S then SetResourceStringValue(I,J,WideToAnsi(S));
        end;
*)
  end;

  procedure StartVampyre;
  var
    I: TImageData;
    dllname: string;
  begin
//addlog(' sizeof(opj_image_comp) = %0',[sizeof(opj_image_comp)]);
//addlog(' sizeof(OPJ_COMPONENT_TYPE) = %0',[sizeof(OPJ_COMPONENT_TYPE)]);
    if MessageContainer.CurrentLanguage = 0//'RUSSIAN'
      then VyRussifyResourceStrings;
//J2k support removed
(*
    if (Mother^.State.InstallPath <> '') and Config.Bool['main', 'use_OpenJpeg']
    then try
      dllname:= Config.Path['OpenJPEG', 'j2k' + SystemSuffix];
      if dllname = '' then dllname:= UnmangleFileName(OpenJpegDll);
      if (ExtractFilePath(dllname) <> '') and not FileExists(dllname)
        then Die(MI_ERROR_FILE_NOT_FOUND, [OpenJpegDll])
      else;// CpsInitOpenJpeg(dllname);
      try
        InitImage(I);
        LoadImageFromMemory(@j2k_test_image, sizeof(j2k_test_image), I);
        FreeImage(I);
      except
        Die('Сбой при чтении пробного изображения',
              'Failed to load the test image',[]);
      end;
    except
      try
        Die(
         'Не удалось запустить поддержку jpeg2000' ,
         'Unable to init the jpeg2000 support',[]);
      except end;
      AddStopMessage(StopDying());
    end;
 *)
  end;

  function CGEPath(s: TFileNameString): TFileNameString;
  begin
    Result:=StrReplace(StrReplace(s, '{$PLATFORM}', SystemSuffix), '{$HOME}', Mother^.State.HomePath);
    if  (Result = '')
      or ({$ifdef unix}Result[1] <> '/'{$else}ExtractFileDrive(Result) = '' {$endif})
      then Result:=Mother^.State.InstallPath + Result;
    Result:=OptiPath(Result) + ExtractFileName(Result);
  end;
  
  function RuEn(Ru, En: UnicodeString): UnicodeString;
  begin
    if Mother^.State.IsRussian
      then Result:=Ru
      else Result:=En;
  end;

  function VersionToStr(major, minor, build: integer): AnsiString;
  begin
    Result:=format('%d.%.2d.%.4d' ,[Major, minor, build]);
  end;
  

  {$ifdef windows}


    function GetSpecialFolder(f: integer): TFileNameString;
    var
      i: integer;
    begin
      if Mother^.State.Windows98.DoesNotSupportUserDirectory then Result:= '<Error>'
      else begin
//writeln(inttohex(ptruint(pointer(SHGetFolderPathA)),8));
        SetLength(Result, MAX_PATH);
        FillChar(Result[1], Length(Result), 0);
//writeln('=:(');
        {$ifdef che_unicode}
          SHGetFolderPathW(0, f {or CSIDL_FLAG_CREATE}, 0, SHGFP_TYPE_CURRENT, PUnicodeChar(Result));
        {$else}
          SHGetFolderPathA(0, f {or CSIDL_FLAG_CREATE}, 0, SHGFP_TYPE_CURRENT, PAnsiChar(Result));
        {$endif}
//writeln(':)');
        for i:=1 to Length(Result) do
          if Result[i] = #0 then begin
            SetLength(Result, i - 1);
            break;
          end;
        if Result > '' then begin
          Result[1]:= UpCase(Result[1]);
          Result:=IncludeTrailingPathDelimiter(Result);
        end;

      end;
    end;

    function GetProgramFilesDir: TFileNameString;
    begin
      Result:=GetSpecialFolder(CSIDL_PROGRAM_FILES);
    end;

    function GetLocalAppDataDir: TFileNameString;
    begin
      Result:=GetSpecialFolder(CSIDL_APPDATA);
    end;

    function GetUserName: UnicodeString;
    var L: cardinal;
    begin
      if Mother^.State.Windows98.DoesNotSupportUserDirectory then Result:='<Unknown>'
      else begin
        SetLength(Result, 1000);
        L:=Length(Result);
        if not windows.GetUserNameW(@Result[1], @L)
          then Exit('<ERROR>');
        SetLength(Result, max(0, L - 1)); //don't forget the trailing zero!
      end;
    end;


    const
      SECURITY_NT_AUTHORITY: TSIDIdentifierAuthority = (Value: (0, 0, 0, 0, 0, 5)) ;
      SECURITY_BUILTIN_DOMAIN_RID = $00000020;
      DOMAIN_ALIAS_RID_ADMINS = $00000220;

    function GetRunningAsAdmin: longbool;
    var
      hAccessToken: THandle;
      ptgGroups: PTokenGroups;
      dwInfoBufferSize: DWORD;
      psidAdministrators: PSID;
      g: Integer;
      bSuccess: BOOL;
    begin
      Result := False;
      if Mother^.State.Windows98.DoesNotSupportUserDirectory then exit;
      if not OpenThreadToken(GetCurrentThread, TOKEN_QUERY, True, hAccessToken)
      then begin
        if (GetLastError <> ERROR_NO_TOKEN)
         or not OpenProcessToken(GetCurrentProcess, TOKEN_QUERY, hAccessToken)
           then Exit(false);
      end;
      GetMem(ptgGroups, 1024) ;
      if GetTokenInformation(hAccessToken, TokenGroups, ptgGroups, 1024, dwInfoBufferSize)
      then begin
        CloseHandle(hAccessToken) ;
        AllocateAndInitializeSid(SECURITY_NT_AUTHORITY, 2, SECURITY_BUILTIN_DOMAIN_RID, DOMAIN_ALIAS_RID_ADMINS, 0, 0, 0, 0, 0, 0, psidAdministrators) ;
        for g := 0 to ptgGroups^.GroupCount - 1 do
          if EqualSid(psidAdministrators, ptgGroups^.Groups[g].Sid) then
          begin
            Result := True;
            Break;
          end;
        FreeSid(psidAdministrators) ;
      end;
      FreeMem(ptgGroups) ;
    end;

  {$endif}
  
  function BacktrackSymlink(a: ansistring): ansistring;
  var
    b: string;
  begin
    {$ifdef unix}
     while true do begin
       b:= fpReadLink(a);
       if b = '' then break;
       a:= b;
     end;
    {$endif}
     Result:=a;
  end;

//  var State.ExePath, StdLocation, AltLocation, StdHome: ansistring;

  procedure CheckIfUserNameValid(deadly: boolean);
  var
    i: integer;
    s: ansistring;
    oops: boolean = No;
  begin
   {$ifdef windows}
    if (Mother^.State.OS in WindowsEmulators)
      or (Mother^.State.Windows98.DoesNotSupportUserDirectory) then Exit;
    if Mother^.State.UserNameUCS16 <> AnsiToWide(WideToAnsi(Mother^.State.UserNameUCS16))
       then oops:= Yes;
    s:= GetLocalAppDataDir();
    for i:=1 to length(s) do if s[i] = '?' then oops:= Yes;
    if oops then begin
      if deadly then AddStopMessage(MI_INVALID_USER_NAME,[Mother^.State.UserNameUCS16, s])
                else GiveWarning(
                       Utf8Encode(Mother^.State.UserNameUCS16) + '-invalid-name',
                            MI_WARN_WARNING,
                            MI_INVALID_USER_NAME, [Mother^.State.UserNameUCS16, s]);
    end;
   {$endif}
  end;
  
  {$ifdef unix}
{  Dis one glitched
  function GetPhysicalMemorySize: integer;
  //wtf? they seem to be defined in libc.pp!
  const
    _SC_UIO_MAXIOV = 60;
    _SC_PHYS_PAGES = (_SC_UIO_MAXIOV)+25;
  var
    p: TLoggedProcess;
    s2: string = '';
    i,j: integer;
    procedure SayOops;
    begin
      WriteLn('FAILED TO DETERMINE RAM SIZE, ASSUMING 1024 Mb');
    end;
  begin
    p:= TLoggedProcess.Create('head -n20 /proc/meminfo');
    p.Execute;
    for i:=0 to p.log.Count-1 do
      if (copy (p.log[i], 1, 9) = 'MemTotal:') then begin
        for j:=1 to length(p.log[i]) do
          if p.log[i][j] in ['0'..'9'] then s2+= p.log[i][j];
        break;
      end;
    if s2 = '' then begin
      SayOops;
      Result:= 1024;
    end
    else begin
      try
        Result:= StrToInt(s2) div 1024;
      except
        WriteLn((ExceptObject as Exception).ClassName, '  ', (ExceptObject as Exception).Message);
        SayOops;
        Result:= 1024;
      end;
      if Result < 90 then begin
        WriteLn('RAM SIZE SUSPICIOUSLY SMALL (', Result, ' Mb). ASSUMING 256 Mb');
        Result:= 256;
      end;
    end;
    p.Free;
    //old version, relied on the libc unit.
    //Result:= sysconf(_SC_PHYS_PAGES) * 4096//sysconf(_SC_PAGE_SIZE)//can't find numeric value for it
     //         div (1024*1024);
  end;
 }





  function GetPhysicalMemorySize: integer;
    var
      sl: TStringList;
      st: TFileStream;
      a: array of Utf8String;
      s: Utf8String;
      x: integer;
    function Find(ind: UnicodeString; FindLast: boolean = false): Utf8String;
    var
      k: integer;
      m: array of Utf8String;
    begin
      Result:= '';
      k:= 0;
      while k < sl.Count do begin
        m:= Explode(Utf8String(':'), sl[k], true);
        if Length(m) = 2 then begin
          if m[0] = ind then begin
            Result:= m[1];
            Break;
          end;
        end;
        Inc(k);
      end;
//WriteLn;
//WriteLn('Find: ', ind, ' : ', Result);
    end;
  begin
    sl:= TStringList.Create;
    try
      try
        st:= TFileStream.Create('/proc/meminfo', fmOpenRead);
        sl.LoadFromStream(st);
        st.Free;
//WriteLn;
//WriteLn(sl.Text);
        x:= 0;
        a:= Explode(Utf8String(' '), Find('MemTotal'), true);
        if (Length(a) >= 2) and (a[1] = 'kB') then try
          x:= StrToInt(a[0]);
        except
        end;

        if x = 0 then begin
          WriteLn('FAILED TO DETERMINE RAM SIZE, ASSUMING 1 Gb');
          Result:= 1000;
        end
        else Result:= x div 1024;


      except
        WriteLn('CRASHED TRYING TO DETERMINE RAM SIZE, ASSUMING 1 Gb');
        Result:= 1000;
      end;
    finally
      sl.Free;
    end;
  end;

  {$ifndef darwin}
  procedure DetectLinuxVersionString;
  var
    p: TLoggedProcess;
    i: integer;
    w: UnicodeString = '';
    a: Utf8String;
    S: TFileStream;
  begin
    p:= TLoggedProcess.Create('lsb_release -d');
    p.Execute;
    for i:=0 to p.log.Count-1 do
      if (copy (p.log[i], 1, 13) = 'Description:'#$09) then begin
        w:= Utf8Decode(trim(copy(p.log[i], 14, length(p.log[i]))));
        break;
      end;
    p.Free;
    if w <> '' then Mother^.State.OSName:= w
    else try
      S:=TFileStream.Create('/etc/issue', fmOpenRead);
      i:= min(S.Size, 300);
      SetLength(a, i);
      S.ReadBuffer(a[1], i);
      S.Free;
      a:= StringReplace(a, '\n', ' ', [rfReplaceAll, rfIgnoreCase]);
      a:= StringReplace(a, '\l', ' ', [rfReplaceAll, rfIgnoreCase]);
      a:= StringReplace(a, #10, ' ', [rfReplaceAll, rfIgnoreCase]);
      a:= StringReplace(a, #13, ' ', [rfReplaceAll, rfIgnoreCase]);
      a:= trim(a);
      if a <> '' then Mother^.State.OSName:= Utf8Decode(a);
    except
      if Mother^.Debug.Verbose
        then AddLog('Cannot detect Linux version: %0',[StopDying()])
        else StopDying();
    end;
  end;
  {$endif}

  {$else}
  function GetPhysicalMemorySize: integer; // in Mbytes
  var
    Status : TMemoryStatus;
  begin
    Status.dwLength := sizeof( TMemoryStatus ) ;
    GlobalMemoryStatus( Status ) ;
    Result:=Status.dwTotalPhys div (1024*1024);
  end;
  {$endif}

  var
    HomepathThatBeenNonWriteable: ansistring = '';
    
  procedure DetermineInstallPath;
    function IsProgramDir(a: TFilenameString): boolean;
    begin
      Result:= FileExists(IncludeTrailingPathDelimiter(a) + MyAppName + '.ini');
//WriteLn;WriteLn;
//WriteLn(' IPD: ', a);
//WriteLn('    : ', IncludeTrailingPathDelimiter(a));
//WriteLn('    : ', ord(Result));
      if Result then Mother^.State.InstallPath:= a;
    end;
  begin
    with Mother^ do begin
      State.ProgramDirIsWriteable:= No;
      State.ExePath:=ExtractFilePath({BacktrackSymlink}(ParamStr(0)));
      State.InstallPath:='';
    {$ifdef unix}
      {$ifdef darwin}
        {$fatal Oopsie... Not ported yet.}
      {$else}
        State.UserName:= Trim(StrReplace( TFileNameString(GetEnvironmentVariable('USERNAME')), TFileNameString('%'), TFileNameString('')));
        if State.Username = '' then State.UserName:=Trim(StrReplace(TFileNameString(GetEnvironmentVariable('USER')), TFileNameString('%'), TFileNameString('')));
        State.UserNameUCS16:= Utf8Decode(State.UserName);
        if not IsProgramDir(State.ExePath)
           and not IsProgramDir('/usr/share/' + MyAppName +'/')
           and not IsProgramDir('/usr/local/share/' + MyAppName +'/')
          then AddStopMessage(MI_INSTALL_DIR_NOT_FOUND, ['/usr/share/' + MyAppName +'/', State.ExePath]);
//        if (State.UserName = 'root') or (Copy (Mother^.State.HomePath, 1, 6) = '/root/')
//          then AddStopMessage(MI_BAD_USER_NO_ROOT, [CgeString]);
      {$endif}
    {$else}
      State.UserNameUCS16:= GetUserName;
      State.UserName:= WideToAnsi(State.UserNameUCS16);
      if not IsProgramDir(State.ExePath)
        and not IsProgramDir(GetProgramFilesDir + MyAppName + '\')
        then AddStopMessage(MI_INSTALL_DIR_NOT_FOUND, [GetProgramFilesDir + MyAppName + '\', State.ExePath]);
    {$endif}
      State.HomePath:= '';
//WriteLn(' IP=', Mother^.State.InstallPath);
    end;
  end;

  function ValidateFileIsWriteable(FileName: TFileNameString): boolean;
  var s: TFileStream = nil;
    attempt: integer;
  begin
    VerboseLog('  Checking if file is writeable: %0',[FileName]);
    Result:= false;
    for Attempt:= 0 to 2 do begin
      if Result then break;
//  WriteLn('VFW: "', FileName, '"');
//  WriteLn(' FileExists: ', FileExists(FileName));
      if not FileExists(FileName) then begin
        Result:= true;
        VerboseLog('    Not found.');
        break;
      end;
      if Attempt > 0 then begin
        VerboseLog('    Sleeping 500 miliseconds.');
        Sleep(500);
      end;
      try
  //      try
          s:= TFileStream.Create(FileName, fmOpenReadWrite);
          Result:= true;
//writeln('  file handle=', s.Handle);
  //      finally
  //if ExceptObject is Exception then writeln('Exception 1: ', (ExceptObject as Exception).Message);
          s.Free;
  //      end;
      except
//  writeln('Exception 2: ', (ExceptObject as Exception).Message);
        VerboseLog('    ..whoops. %0: %1',
          [string((ExceptObject as Exception).ClassName), (ExceptObject as Exception).Message]);
      end;
    end;
    VerboseLog('    Result = %0',[Result]);
  end;

  procedure DetermineHomePath;
    var
      H: boolean = true;
      I: boolean = true;
      mainreason: boolean;
    procedure TryHomePath;
    begin
      with Mother^ do begin
        VerboseLog('  ..trying the default home path %0 ...', [State.StdHomepath]);
        State.HomePath:= State.StdHomepath;
        if not ValidateWritePath(State.HomePath) or not ValidateFileIsWriteable(State.HomePath + ConfigFilePath) then begin
          HomepathThatBeenNonWriteable:= State.HomePath;
          H:= false;
        end;
      end;
    end;
    procedure TryInstallPath;
    begin
      VerboseLog('  ..trying the install path...');
      I:= ValidateWritePath(Mother^.State.InstallPath) and ValidateFileIsWriteable(Mother^.State.InstallPath + ConfigFilePath);
      Mother^.State.ProgramDirIsWriteable:= I;
      if I then Mother^.State.HomePath:= Mother^.State.InstallPath;
    end;
  begin
    VerboseLog('Determining the home path...');
      if Mother^.State.ExePath = '' then exit;
      Mother^.State.StdHomepath:= Config.Path[Mother^.State.OstID, 'save-path'];
      Mother^.State.IsPortable:= Config.Bool['main','portable'];
      Mother^.State.ProgramDirIsWriteable:= No;
      mainreason:= Config.Bool[Mother^.State.OstID, 'try-program-dir-first'] or Mother^.State.DeveloperMode or Mother^.State.IsPortable;
      if mainreason {$ifdef windows} or Mother^.State.Windows98.DoesNotSupportUserDirectory {$endif}
      then begin
        if Mother^.State.Windows98.DoesNotSupportUserDirectory and not mainreason
          then HomePathWarning:= hfwWin9x
          else
            if not Mother^.State.IsPortable then HomePathWarning:= hfwUseI;
        TryInstallPath;
        if not I then begin
          if Mother^.State.Windows98.DoesNotSupportUserDirectory then begin
            AddStopMessage(MI_LEGACY_WIN_USING_I,[AnsiToWide(Mother^.State.InstallPath)]);
            AddStopMessage(MI_INSTALL_DIR_NOT_WRITEABLE,[AnsiToWide(Mother^.State.InstallPath), Mother^.State.UserNameUCS16]);
            Mother^.State.HomePath:='';
            Exit;
          end
          else begin
            HomePathwarning:= hfwTriedIUseH;
            TryHomePath;
          end;
        end;
        if not H then begin
          AddStopMessage(MI_INSTALL_DIR_NOT_WRITEABLE,[AnsiToWide(Mother^.State.InstallPath), Mother^.State.UserNameUCS16]);
          AddStopMessage(MI_HOME_PATH_NOT_FOUND,[AnsiToWide(Mother^.State.HomePath), Mother^.State.UserNameUCS16]);
          Mother^.State.HomePath:= '';
        end;
      end
      else begin
        TryHomePath;
        if not H then begin
          Mother^.State.HomePath:='';
          HomePathwarning:= hfwTriedHuseI;
          if Config.Bool[Mother^.State.OstID, 'try-program-dir-if-save-path-not-writeable']
            then begin
            TryInstallPath;
            if not I then begin
              AddStopMessage(MI_HOME_PATH_NOT_FOUND,[AnsiToWide(Mother^.State.HomePath), Mother^.State.UserNameUCS16]);
              AddStopMessage(MI_INSTALL_DIR_NOT_WRITEABLE,[AnsiToWide(Mother^.State.InstallPath), Mother^.State.UserNameUCS16]);
              Mother^.State.HomePath:= '';
            end;
          end
          else
            AddStopMessage(MI_HOME_PATH_NOT_FOUND,[AnsiToWide(Mother^.State.HomePath), Mother^.State.UserNameUCS16]);
        end;
      end;
      if Mother^.State.HomePath <> '' then Mother^.State.HomePath:= IncludeTrailingpathDelimiter(Mother^.State.HomePath);
  end;


   procedure chCopyFile(n1, n2: TFileNameString);
   var
     s1, s2: TFileStream;
     f: file;
     sz: int64;
   begin
     try
       MakeSurePathExists(ExtractFilePath(n2));
       CheckForGuardedException;
       if FileExists(n2) then try
         AssignFile(f, n2);
         Erase(f);
       except
         Die(MI_ERROR_CANNOT_DESTROY_FILE, [n2]);
       end;
       s1:=TFileStream.Create(n1, fmOpenRead);
       s2:=TFileStream.Create(n2, fmCreate);
       sz:= s2.CopyFrom(s1, s1.Size);
       if sz <> s1.Size then Die(RuEn(
          'Файл скопировался не полностью',
          'File did not copy completely'));
     except
       s1.Free;
       s2.Free;
       Die(MI_CANNOT_COPY_FILE, [n1, n2]);
     end;
     s1.Free;
     s2.Free;
   end;
   

  var
    Wgiven: boolean = No;

  procedure GiveWarnings;
  begin
    if Wgiven then exit;
    GiveWarningNotYellow('gpl', MI_WARN_GNU_GPL_CAPTION, MI_WARN_GNU_GPL,
      [
       RuEn(CgeStringRu, CgeStringEn),
       AnsiToWide(Mother^.State.InstallPath) + 'license' + PathSlash +'gpl.txt',
       RuEn(CopyrightRu, CopyrightEn),
       RuEn(CopyrightLibsRu, CopyrightLibsEn)
      ]);

    if Mother^.State.DeveloperMode then GiveWarning('developer_mode', MI_WARN_WARNING, MI_WARN_DEVELOPER_MODE,
      [
        CgeString,
        AnsiToWide(Mother^.State.InstallPath) + MyAppName + '.ini'
      ]);
      
   {$ifdef windows}
    if Mother^.State.Windows98.DoesNotSupportUnicode
      then GiveWarning('win98-no-unicode', MI_WARN_WARNING, MI_WINDOWS_98_NO_UNICODE_SUPPORT, []);
   {$endif}

    if not (Mother^.State.OS in SupportedOSes)
      then GiveWarning(
         OSDisplayedName[Mother^.State.OS] + '-unsupported', MI_WARN_WARNING,
          MI_UNSUPPORTED_OS, [CgeString, OSDisplayedName[Mother^.State.OS]]);
          
    if not Mother^.State.IsPortable then
      case HomePathWarning of
        hfwNone:;
        hfwTriedIUseH: GiveWarning(
          Utf8Encode(Mother^.State.UserNameUCS16) + '-' + Mother^.State.OstID + '-install-path-not-writeable-use-home-path',
          MI_WARN_WARNING,
          MI_I_PATH_NON_WRITEABLE_USING_H,
          [AnsiToWide(Mother^.State.HomePath), AnsiToWide(Mother^.State.InstallPath),
           Mother^.State.UserNameUCS16, OSDisplayedName[Mother^.State.OS]]);
        hfwTriedHuseI: GiveWarning(
          Utf8Encode(Mother^.State.UserNameUCS16) + '-' + Mother^.State.OstID + '-home-path-not-writeable-use-install-path',
          MI_WARN_WARNING,
          MI_H_PATH_NON_WRITEABLE_USING_I,
          [AnsiToWide(Mother^.State.StdHomepath), AnsiToWide(Mother^.State.InstallPath),
           Mother^.State.UserNameUCS16, OSDisplayedName[Mother^.State.OS]]);
        hfwUseI: GiveWarning(
          Utf8Encode(Mother^.State.UserNameUCS16) + '-' + Mother^.State.OstID + '-using-install-path',
          MI_WARN_WARNING,
          MI_USING_I,
          [AnsiToWide(Mother^.State.StdHomepath), AnsiToWide(Mother^.State.InstallPath),
           OSDisplayedName[Mother^.State.OS]]);
        hfwWin9x:  GiveWarning(
          Mother^.State.OstID + '-no-multiuser-support-using-install-path',
          MI_WARN_WARNING,
          MI_LEGACY_WIN_USING_I,
          [AnsiToWide(Mother^.State.InstallPath)]);
      else
        Die('missing case branch for HomePathWarning %0 in cl_main.inc', [GetEnumName(TypeInfo(THomePathWarning),ord(HomePathWarning))]);
      end;

    if Mother^.State.PhysMemory < OsReqMem[Mother^.State.OS]
      then GiveWarning(OSDisplayedName[Mother^.State.OS] + '--phys-mem', MI_WARN_WARNING,
        MI_TOO_LITTLE_PHYSICAL_MEMORY, [OSDisplayedName[Mother^.State.OS],
        OsReqMemRep[Mother^.State.OS], Mother^.State.PhysMemory]);


    if Mother^.Debug.Enabled then GiveWarning('debug_mode', MI_WARN_WARNING, MI_WARN_DEBUG_MODE,
      [
        CgeString,
        AnsiToWide(Mother^.State.InstallPath) + MyAppName + '.ini'
      ]);

    Wgiven:= Yes;
  end;

  {$ifdef unix}
  function GetDLLFileName(h: pointer): TFileNameString;
  var
    dlinfo: dl_info;
  begin
    FillChar(dlinfo, sizeof(dlinfo), 0);
    dladdr(h, @dlinfo);
    Result:= Utf8ToFileName(Utf8String(dlinfo.dli_fname));
  end;
  {$else}
  function GetDLLFileName(h: THandle): TFileNameString;
  var
    TST: array[0..Max_Path] of {$ifdef che_unicode} UnicodeChar {$else} AnsiChar {$endif};
  begin
    {$ifdef che_unicode}
      GetModuleFileNameW
    {$else}
      GetModuleFileName
    {$endif}
        (h, TST, SizeOf(TST) div SizeOf(TST[0]));
    Result:= TFileNameString(PFileNameChar(@TST));
  end;
  {$endif}

  function FancyFunc(f: double): glFloat;
  {A non-linear function for screen fade in from black.
   The input is time scaled to 0.0..1.0 (but may go beyond 1.0)}
  begin
    if f < 0.0 then Result:= 0
    else
      if f > 1.0 then Result:= 1
      else
        if f < 0.5
          then Result:= 0.5 * sqr(f * 2)
          else Result:= 1 - 0.5 * sqr((1 - f) * 2);
  end;

  function IniCompatibleStringForm(a: AnsiString): AnsiString;
  var i: integer;
  begin
    a:= LowerCase(a);
    Result:= '';
    for i:=1 to length(a) do
      if a[i] in ['a'..'z', '0'..'9', '-'] then Result+= a[i]
                                           else Result+= '_';
  end;

  function NewInputEvent: integer;
  begin
    Result:= length(Mother^.Input.Events) - 1;
    SetLength(Mother^.Input.Events, Result + 2);
    FillChar(Mother^.Input.Events[Result + 1], Sizeof(TInputEvent), 0);
    if Result < 0 then Result:= NewInputEvent(); //a perverted way to keep one ahead
  end;
  
  function GetAppNameEventOverride: string;
  begin
    Result:=MyAppName;
  end;

  procedure CheckForGuardedException;
  {This used after mother itself calls the guarded function, to restart
    the exception chain.}
  var
    pes: PMotherSehState;
  begin
    pes:= GetExceptionState();
    if pes^.NowDying
      then raise Exception.Create('');
  end;
  
  procedure TCgeProcess.OnReadLn(ss: string);
  begin
    AddLogComment(#10#13'  ' + s);
//    _cgeHeartBeat;
  end;

  procedure DisplayLastWill;
  var
    i: integer;
    //F: File;
    s: TFileStream;
    u: AnsiString;
  begin
    For i:=1 to ParamCount do
      if copy(lowercase(ParamStr(i)), 1, 11) = '--last-will' then begin
        try
          s:= TFileStream.Create(UnmangleFileName('*H/logs/last-last-will.txt'), fmOpenRead);
//          AssignFile(F, UnmangleFileName('*H/logs/last-last-will.txt'));
//          Reset(F, 1);
          SetLength(u, s.Size);//FileSize(F));
//          BlockRead(F, u[1], FileSize(F));
//          CloseFile(F);
          s.Read(u[1], s.Size);
          s.Free;
        except
          Halt(0);
          {Die(RuEn(
            'Не удалось прочитать файл с завещанием!',
            'Failed to read the last will file!'))}
        end;
{        AddLog(
          RuEn(
            'Получено завещание, вывожу диалоговое окно:#10#13#10#13%0',
            'Received a last will, displaying the dialog window:#10#13#10#13%0'),
          [utf8Decode(u)]
        );
}
        ErrorMessageBox(CGEString, utf8Decode(u));
        Halt(0);
      end;
  end;
 {
  function BoolChars(b: boolean): UnicodeString;
  begin
    if b then result:='true' else result:='false';
  end;
 }
  const CharsTolerableInFileName: set of AnsiChar = ['_', '-', 'a'..'z', 'A'..'Z', '0'..'9', '(', ')', '[', ']'];
  function SanateStringForFileName(s: AnsiString): AnsiString;
  var i: integer;
  begin
    SetLength(Result, Length(s));
    for i:=1 to length(s) do if s[i] in CharsTolerableInFileName then Result[i]:= s[i] else Result[i]:= '_';
  end;

  function CgeNeedToKeepModuleResources(): boolean; cdecl;
  begin
    Result:= Assigned(Module) and not Mother^.Module.ApplicationExiting
      //(Mother^.Module.State  <> ms_Exiting); // (Module.IndexToLoad <> Module.Index);
  end;

  procedure InitMotherState;
  begin
    {$include cl_init_mother.inc}
  end;

  {$ifdef windows}
  var
   rstp: TProcess;
  {$endif}

  procedure RestartMyself(param: ansistring);
  begin
    {$ifdef unix}
      fpExecL(ParamStr(0), [param]);
    {$else}
      rstp:= TProcess.Create(nil);
      rstp.CommandLine:= ParamStr(0) + ' ' + param;
      rstp.Options:=[poNoConsole];
      rstp.Execute;
//      ShellExecute(0, 'open', PChar(ParamStr(0)), '--restart', '', SW_SHOWNORMAL);
    {$endif}
    halt(0);
  end;

  procedure KillMyself(exitcode: integer = 0);
  begin
 //   halt(exitcode);


    //seriously, there could be some threads locked up, so terminating self requires a decisive approach.
    {$ifdef unix}
      if Mother^.State.ProcessID > 0 then FpKill(Mother^.State.ProcessID, SIGKILL);
    {$else}
      if {$ifdef cpu32}Assigned(TerminateProcess) and {$endif}
        (Mother^.State.ExeHandle <> 0) and (Mother^.State.OS <> ostWine)
      then TerminateProcess(Mother^.State.ExeHandle, exitcode);//hara-kiri!
    {$endif}
    halt(exitcode);
  end;

  procedure ResetMotherErrorState();
  var
    pes, pnext, pprev: PMotherSehState;
  begin
    EnterCriticalsection(Mother^.SehCriticalSection);

    //remove all states belonging to module threads
    pnext:= @Mother^.ExceptionState;
    pes:= nil;
    while Assigned(pnext) do begin
      pprev:= pes;
      pes:= pnext;
      pnext:= pes^.Next;
      if pes^.IsModuleThread then begin
        pes^._StopDyingString:= '';
        pes^.ThreadTitle:= '';
        SetLength(pes^.Warnings, 0);
        Dispose(pes);
        if Assigned(pprev)
          then pprev^.Next:= pnext
          else Mother^.ExceptionState.Next:= pnext;
        pes:= pprev;
      end;
    end;

    Mother^.ExceptionState.NowDying:= false;
    Mother^.ExceptionState.CallStackLogged:= false;
    {$ifdef windows}
      Mother^.ExceptionState.ExceptionCode:= 0;
      Mother^.ExceptionState.ExceptionAddress:= 0;
    {$endif}
    LeaveCriticalsection(Mother^.SehCriticalSection);
  end;



  procedure GetModuleStopMessage(var caption, text: UnicodeString);
  { 2015 : With recent cxhanges, sometimes the stop screen cannot be shown.
    Must save to config or show in a MessageBox. }
  begin
    with ModuleStopMessage do begin
      if (ord(capMID) = 0) and (capStr = '') then begin
        caption:= '';
        text:= '';
      end
      else begin
        if capStr = ''
          then caption:= MessageContainer[capMID]
          else caption:= capStr;
        if msgStr = ''
          then text:= PervertedFormat(MessageContainer[msgMID], params)
          else text:= PervertedFormat(msgStr, params);
      end;
    end;
  end;

  procedure InitMotherAssets;
  begin
    if Mother^.Debug.Verbose then AddLog('Loading standard pointer images from embedded image files...');
    PreloadDefaultPointerImages;

    if Mother^.Debug.Enabled then AddLog('Loading the fixed font...');
    InitFixedFont;

    if Mother^.Debug.Enabled then AddLog('Loading the stub texture...');
    InitStubTexture;
  end;

  {$include framework.inc}

  {$include un_typedefs.inc}

  {$include cl_seh.inc}

  {$ifdef unix}
    {$include cl_sigterm.inc} // intercepting sigterm to allow saving the session
  {$endif}

  {$ifdef win32}
    {$include cl_win98_u_test.inc} //test if windows 98 has unicode support installed
  {$endif}

//  {$include cl_dyna.inc} //a bunch of legacy dynamic array classes

  {$include cl_dynareplacer.inc}
  
//  {$include cl_stream.inc}

  {$include cl_confman.inc} //configuration manager

//both moved to cl_framework_winapi.pp or cl_framework_x11.pp
  //{$include cl_winman.inc} //the window manager basis, mostly platform-dependent
  //{$include cl_window.inc} //the higher, platform-independent window manager logic


  {$include un_talesteller.inc} //composes a detailed error message from the exception address

  {$include un_fileutil.inc}

  {$ifndef glesonly} //for debugging purposes only (compatibility testing): cut off all support for regular opengl
    {$include un_gl_from_fpc_2_6_4.inc} //dynamic loading opengl functions
  {$endif}
  {$include un_gles_from_fpc_2_6_4.inc} //dynamic loading opengl es functions

  {$include un_gl_common.inc} //a slightly higher level library for working transparently with all OpenGL flavors

  {$include cl_gl_init.inc} //loading opengl dll and getting functions from it

  {$include un_al.inc} //dynamic loading openal functions

  {$include cl_al_init.inc} //loading openal dll and getting functions from it

  {$ifdef windows}

  {$endif}

  {$include cl_module.inc}

 //loading and interacting with the actual game module dll
  
//  {$include cl_container.inc}
  
  {$include cl_console.inc} //opengl log console, also renders background and logo
  
  {$include cl_main.inc} // procedure main()
  
  {$include cl_fixed_font.inc} //handling the bitmap font and rendering text using OpenGL FFP

  {$include cl_stub_texture.inc} //loading a stub texture image
  
  {$include cl_fps_counter.inc} //measure FPS and render the counter

  {$include cl_msgbox.inc} //platform-independent wrapper for system error message boxes

  {$include cl_lockupguard.inc} //a separate thread to commit harakiri if processing one frame takes more than 30 seconds

  {$include cl_log.inc}

  {$include un_wav_format.inc}

  {$include un_threadloadind.inc}

  {$ifdef fix_fpc3_unicode}
     {$include un_fpc_3-0-0_fix.inc} // Use custom mods of TFileStream, TStringList and TIniFile
  {$endif}
  {$if defined(windows) and (FPC_FULLVERSION<30000) }
     {$include un_fpc_2-6-4_fix.inc} // Use fixed TFileStream and copy-pasted everything that depends on it
  {$endif}



initialization
  pointer(OnGetApplicationName):= @GetAppNameEventOverride;

  {$include cl_msg_ru.inc}
  {$include cl_msg_en.inc}

finalization
  Try
{$ifdef unix}
    if Di and FileExists(Fn) then DeleteFile(Fn);
{$else}
    if Di then ReleaseMutex(M);
{$endif}
  Except End;
end.


