{
    This file is part of the Cheb's persistence system (Chepersy),
    Copyright (c) 2004-2008 by Anton Rzheshevski (chebmaster@mail.ru),
      and contains most of the persistency system source code.

    See the file COPYING.CPS, included in this distribution,
    for details about the copyright.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

 **********************************************************************}
{
 ToDo:
  - DROPPED: Add all loaded scraped objects to the graveyard!
  - DROPPED: cant read previous versions' saves: Basket optimization (caching, header UIDs et cetera)
  - BUG: when a pointer type field is not marked as skipped (an array of pointer,  in this case),
        the resulting error message is misleading: it says the control value of tthe
        size in the file is 16 while the current size is 4.
        It should either skip the field properly, or refuse to register such a field
        if it is not skipped!
  - BROKEN: in a record with mixed 32-bit and 64-bit fields, fails to determine
        proper alignment. Considers the record packed when it is not.
  - OPTIMIZATION: compress the index arrays at registration end, to make them cache-friendly

05.07.12: corrected tkWstring/tkUstring problem. Now works in Linux but requires fpc >= 2.4
}

{$include chepersy_defs.inc}


{$rangechecks off}
{$ifndef fpc}
   Ы
   //Free Pascal only!
{$endif}
{$codepage utf-8} //sets the source code code page
{$ifopt C+} //if assertions are turned on...
  {$define safeloading} //..compile with all the safeguards turned on
{$endif}

{$if defined(CPUi386) or defined(CPUX86_64)}
  {$asmmode intel} //RDTSC used for garbage collector timing
{$endif}

{$ifdef cpu64}
  {$define cps_no_record_validation} { Currently, development that catches all
  the mismatches could only be done in 32 bits. Will revisit later needs research}
  {$undef chepersy_assert_64_fields_are_aligned}
{$endif}
unit chepersy;

interface
  uses
    //chebs_heap_brutal_trace,
    SysUtils, Classes, typinfo, math, md5
       {$ifdef windows}
         , windows //needs VirtualQuery for custom stack checking
       {$endif}
       {$ifdef go32v2}
         , fake_cse
       {$else}
         , syncobjs
       {$endif}
       {$ifdef cge}
         , mo_hub //originally Chepersy is an integral part of Chentrah, aka Cheb's Game Engine (cge)
       {$endif}
       ;

const
  ChepersyVersion = '1.1.20';
  ChepersyVersionMajor = 1;
  ChepersyVersionMinor = 1;
  ChepersyVersionSubVer = 20;

(*
Changelog 0.9.19 -> 1.1.20:
  1. DROPPED the "managed" part. No more graveyard, no more garbage collection. Backward compatibility is dropped again.
  2. Renamed TManagedObject to TChepersyObject
  3. TChepersyObject is now class(TObject, IUnknown) that ONLY uses InterlockedIncrement in its _AddRef if stars are right (see implementation)
  4. Added (saveable) reference count, type longint
  5. Plan firmly to create a specialized memory manager and evict the cache-unfriendly CpsIndex and f_cpsmask fields into detached cache-friendly pools.

Changelog 0.8.99 -> 0.9.19:
  1. DROPPED BACKWARD COMPATIBILITY of data files
  2. removed support for registering types by string IDs (was used in ancient Delphi to support types lacking RTTI)
  3. backward incompatible API changes: sets and dynamic arrays are now registered by proviting one typeinfo (the base type is inferred from the RTTI). All the obsolete API functions were removed.
  4. It is now possible to pass types that are aliases of floats (for example, TDateTime is silently resolved to Double)
  5. Added a basis for supporting UnicodeString

Changelog 0.8.98 -> 0.8.99:
  1. stream format updated, with header md5 now saved before the header
  2. header and scenario caching, based on the md5 checksums. For each distinct checksum, header is loaded and parsed only once per program execution.
  3. support for headerless streams
  4. TManagedObject.Clone() and TManagedObject.CloneAnother()
  5. Fixed the bug with doubling the field lists in the log.
  6. TManagedObject.Resurrect()
  7. new optional parameter OutputList in CpsLoad()
  8. Scrape made virtual
  9. Added 64-bit fields validation that raises an error if these are not aligned to a 64-bit boundary
 10. Added stubs for future support of the "fixed32" type.
 11. Introduced buffering, as I found TFileStream horrendously inefficient (was causing 20x - 30x decrease in performance). Should help with *any* streams.
 12. Introduced built-in benchmarking for easier bottleneck serach. Controlled via a battery of commented out defines.
 13. Added method Exists, encompassing both Assigned and not Scraped. Yes, you cal safely invoke TManagedObject(nil).Exists. It's for this exat purpose
 14. Added constant CPS_MANAGED_OBJECT, to be used during registration instead of typeinfo(TMyClass). Reason: a) easier than implementing circular class references. b) There is *no* type checking anyway. All fields that are class instances are treated as TManagedObject regardless of their true type.

TODO list for 0.9:
  1. Make sure the alignment prediction and validation does work without the forced packing with {$packrecords 4}
  (done? auto-resolving to base type) 2.
  3. Implement the fixed32 type and its converters.

TODO list for future versions:
  1. As soon as FPC 3.0 is out, implement handling strings that have a variable code page
  2. As soon as FPC 3.0 is out, make sure {$optimization noorderfields} works as intended
  NOTE: not "as soon as it is out" but "as soon as it stops crashing with internal errors on my sources"

*)

{$ifndef fpc}
  {$ifndef oldhacky}
    {$define delphiworkaround}
  {$endif}
{$endif}
const
  //bits of CpsMask reserved for the garbage collector. Do *NOT* use.
  CPS_NO_INTERLOCK_MASK_BIT = 31; {If set, use Inc instead of InterlockedIncrement to work with the reference count}
  CPS_CLONE_MASK_BIT = 30;
  CPS_NO_INTERLOCK_MASK = dword(1) shl CPS_NO_INTERLOCK_MASK_BIT;  // $80000000;
  CPS_CLONE_MASK = dword(1) shl CPS_CLONE_MASK_BIT;      // $40000000;
  CPS_RESERVED_MASK = CPS_NO_INTERLOCK_MASK or CPS_CLONE_MASK;
  CPS_LEAST_NON_RESERVED_BIT = 29;

  CpsBufferSize = 32768; //(bytes) Internal buffer to maximize streaming performance
  //See http://freepascal.ru/forum/viewtopic.php?f=9&t=1137&p=83396#p83396 (Russian)

// DOES NOT WORK :(((((  CpsCustomStackCheckPanicThreshold = 3000000; //(bytes) As of 0.9.00, only works in Windows.
  //Trivia: with its heavy recursion, Chepersy is likely to cause stack overflow
  // when it encounters huge, inoptimal class instance chains.
  //But the standard mechanisms of error processing mostly fail in the case of a
  // stack overflow, often even preventing an error message from being shown

{$if not declared(RawByteString)} //Free Pascal below 3.0
  type
    RawByteString = AnsiString;
{$endif}

type
  TChepersyObject = class(TObject, IUnknown)
  protected
    //***TODO: in 1.2 move index to the compact companion arrays in the new memory manager,
    //  to make parsing the object tree much more cache-friendly!

    CpsIndex: longint; // Do not touch it. Period.
    f_cpsmask: dword; //used to filter objects at saving. Gets saved.
    f_RefCount: longint;
    function QueryInterface({$IFDEF FPC_HAS_CONSTREF}constref{$ELSE}const{$ENDIF} iid : tguid;out obj) : longint;{$IFNDEF WINDOWS}cdecl{$ELSE}stdcall{$ENDIF};
    function _AddRef : longint;{$IFNDEF WINDOWS}cdecl{$ELSE}stdcall{$ENDIF};
    function _Release : longint;{$IFNDEF WINDOWS}cdecl{$ELSE}stdcall{$ENDIF};
  public
    procedure OnZeroReferenceCount; virtual; { by default just calls Destroy.
      Some objects may not want to die, relegating themselves to some manager
      - for example to not free an OpenGL handle in a wrong thread
    }
    procedure AfterConstruction; override;
    procedure BeforeDestruction; override;
    class function NewInstance : TObject; override;
    property RefCount : longint read f_RefCount;
  published
  {$ifdef oldhacky}
    { This isn't a method. This is a dirty trick to store integer value
        directly in the class VMT.
      WARNING: can trigger panic attack in Windows Data Execution Prevention
        or some vigilant antivirus software, as this in fact borders on
        self-modifying code. Not possible in Delphi, btw. as it needs
        calling VirtualProtect() to remove write protection from the block
        of memory contaning the VMTs
      Why class fields would not work:
        a class field is *shared* by all descendants,
        while we need unique storage per class here!
      The metod is not abstract to avoid warnings about
        "constructing a class with abstract methods".}
  {
http://support.microsoft.com/kb/127904
 Follow the steps in this article to create self-modifying code; that is, to modify code pages while they are in memory and execute them there.

 NOTE: Self-modifying code is not advised, but there are cases where you may wish to use it.
 Step-by-Step Example

     Call VirtualProtect() on the code pages you want to modify, with the PAGE_WRITECOPY protection.
     Modify the code pages.
     Call VirtualProtect() on the modified code pages, with the PAGE_EXECUTE protection.
     Call FlushInstructionCache().

 All four steps are required. The reason for calling FlushInstructionCache() is to make sure that your changes are executed. As processors get faster, the instruction caches on the chips get larger. This allows more out of order prefetching to be done. If you modify your code, but do not call FlushInstructionCache(), the previous instructions may already be in the cache and your changes will not be executed.
 Properties
 Article ID: 127904 - Last Review: February 24, 2014 - Revision: 2.1
}
    procedure _MyClassIndex(); virtual;
  {$endif}
  public
    class function ClassIndex(): longint;
    function ClassNameMd5(): TMd5Digest;
  protected
    class procedure _RegisterMyIndex(index: integer);
    //class procedure _SetScenarioIndex(index: integer);
    class procedure _SetMyFieldsList(F: TObject);//TFieldsList);
   constructor Register; //creates a temporary instance during registration (just allocate a chunk of memory)
  {$ifdef fpc}
 //  class function NewInstance() : tobject; override;
  {$endif}
//   function _GetScraped: longbool;
   {$ifdef safeloading}
     procedure _SetCpsMask(w: dword);
   {$endif}
 public

   //*TODO: This should *not* be a constructor.
   //Make sure that a custom NewInstance/InitInstance is called instead
   //constructor Generate; //create an instance *without* initializing it (just allocate a chunk of memory)
   class function Generate: pointer;//TChepersyObject;

   //obsoleted in 0.9. It was a horrible, buggy idea to make it a destructor
   // destructor TechnicalDestroy;
    destructor Destroy; override;

//    procedure FreeInstance; override; //gets intercepted by the garbage collector
//    procedure CpsFreeInstance; virtual; //plays its role instead

    // **** Truly public stuff. For the users of Chepersy.
//    function Exists: longbool; //Assigned and not Scraped. You cal safely invoke TChepersyObject(nil).Exists. It's for this exat purpose

    // Necessary. Needs to call all an inherited method first (see the manual)
    procedure RegisterFields; virtual;

    //      Allow to perform special actions on these two events
    procedure BeforeSaving(); virtual;
    procedure AfterLoading(); virtual; //is called after *all* the objects are loaded.

    //cloning the class instance;
    function Clone(): TChepersyObject; virtual; { everything gets cloned, even
      the skipped fields -- except fields of type Pointer.}
    function CloneAnother(another: TChepersyObject): longbool; virtual; { copy
      another object into self. All fields, like the function above. }


    //managed object functions (new in 0.8.99)
//    procedure Scrape; virtual; //marks self as scraped and adds self to the garbage collector graveyard. To be used instead of Free.
//    procedure Resurrect; //clears the Scraped state and removes the object from graveyard. To be used in the rollback manager.
//    property Scraped: longbool read _GetScraped;
    property CpsMask: dword read f_cpsmask
      {$ifdef safeloading}
        write _SetCpsMask
      {$else}
        write f_cpsmask
      {$endif}
      ;
  end;

  CChepersyObject = class of TChepersyObject;
  PArrayOfChepersyObjects = ^TArrayOfChepersyObjects;
  TArrayOfChepersyObjects = array of TChepersyObject;
  T2DArrayOfChepersyObjects = array of TArrayOfChepersyObjects;


const
  ExtraInfoOffset = {$ifdef cpu64} 12 {$else} 8 {$endif};
  { offset of the first field in TChepersyObject.
    First, there is the metaclass pointer (derived from TObject) - 4 bytes,
    then there is the CpsIndex field - again, 4 bytes. 8 in total.}

var
  ForceInterlock: boolean = false; { Ignore no interlock mask }

  CpsCriticalSection: TCriticalSection;

  { in FPC these are always utf-8 (see the Utf8Encode/Utf8Decode functions in SysUtils),
    in Delphi it's the result of assigning WideString to AnsiString.}
  CpsParserWarnings,
  CpsLog,
  CpsError: TStringList;
  
  CpsValid: boolean = false; //set to true when Chepersy is initialized
  CpsFinalized: boolean = false; //set to true when classes registration is finished

{$ifndef cge}
  CpsUseWriteLnInAddLog: boolean = false; //used for debugging and in the test example
{$endif}

  CpsStream: TStream = nil;

  CpsStackLimit: pointer = nil;

type
  TChepersyDatabaseHeader = class; //forward declaration
var
  CpsDatabaseHeaders: array of TChepersyDatabaseHeader;

// **************** SAVING / LOADING ROUTINES **********************

type
  TArrayOfPointer = array of pointer;
  PArrayOfPointer = ^TArrayOfPointer;

  function CpsStore(
    o: TChepersyObject;
    {$ifdef cge}m: UnicodeString;{$endif} //for error messages
    Target: TStream;
    XorMask: dword = $ffffffff; //if (((o.CpsMask xor XorMask) and AndMask) = 0) then skip o
    AndMask: dword = $ffffffff;
    Headerless: longbool = false
  ): longbool;


  function CpsLoad(
    {$ifdef cge}m: UnicodeString;{$endif}
    Source: TStream;
    POutputList: PArrayOfPointer = nil; //PArrayOfChepersyObjects = nil;
    DatabaseHeader: TChepersyDatabaseHeader = nil //if set, treats the stream as headerless!
  ): TChepersyObject;


  //added in 0.9.19 for fast cloning of a part of the game world.
  function CpsClone(
    o: TChepersyObject;
    {$ifdef cge}m: UnicodeString;{$endif}
    XorMask: dword = $ffffffff; //if (((o.CpsMask xor XorMask) and AndMask) = 0) then skip o
    AndMask: dword = $ffffffff
  ): TChepersyObject;


    
    //added in 0.8.2
   //deprecated in 0.9.0
  function ExportChepersyClassList(): TStringList; //class names per index

type
  TFieldPropertiesEnum = (
    //prefixes are added to field names in the call to ListFields(). Any combination.

    fpro_Skipped,   // prefix '-'. The field is skipped at saving and filled with
      //zeroes at loading. Does not affect cloning.

    fpro_CloneByIncreasingRefCount, //prefix '%'. Only works on dynamic arrays,
      //strings are *always* copied by increasing refcount
      //!will stop with an error if a string or object is found at any depth in the type.

    fpro_NeverClone, //prefix '#', affects only objects, but at any depth in the type.
      //will not be cloned even if the clone mask bit is set
      //example: global game world referrals

    fpro_AlwaysClone //prefix '@', affects only objects but at any depth in the type.
      //will always be cloned, even if the clone mask bit is not set.
      //example: character inventory and stats
  );
  TFieldPropertiesSet = set of TFieldPropertiesEnum; //must firt in a dword, or else.

  //examples of field name as passed to ListFields:
  //   '{Skipped,NeverClone}myfield'
  //   '-#myfield'

  //added in 0.9.0
type
  TExportClassDesc = record
    Name: AnsiString;
    Metaclass: CChepersyObject;
    Index: integer;
  end;

  TArrayOfExportClassDesc = array of TExportClassDesc;

  function ExportClassesList(OnlyDescendantsOf: CChepersyObject = nil): TArrayOfExportClassDesc;

  //added in 0.9.0
  function IsDescendantOf(Ancestor, Descendant: CChepersyObject): longbool; overload;
  function IsDescendantOf(Ancestor: CChepersyObject; DescendantClassIndex: integer): longbool; overload;
  //eliminates the need to have class instances (without them the *is* operator doesn't work)
  function IsRegistered(pinfo: PTypeInfo): longbool; overload;
  function IsRegistered(c: CChepersyObject): longbool; overload;


// ************** GRAPH WALKING FUNCTIONS **************************
  type
    TcustomWalkProc = procedure(o: pointer); //TChepersyObject);
  
// ***TODO add skip mask
  function CpsMarkupGraph(o: pointer//TChepersyObject
    ; SetMaskBits, ClearMaskBits: dword): longbool;
  function CpsWalkGraph(o: pointer//TChepersyObject
    ; proc: TcustomWalkProc): longbool;

//  function CpsGetNameHash(name: string): longint;
//  function CpsGetField(NameHash: integer): Variant;


  {$ifdef cge}
    //{$include mo_typedefs.h}
  {$else cge}
    //simplified replacements for the parts of Cheb's Game Engine
    type
      float = single;
      boolean = longbool;
      {$ifndef fpc}
        //delphi
        winbool = longbool;
        dword = cardinal;
      {$endif}
     const
        Yes = True;
        No = False;
      {$ifdef chepersy_align_4}
        AlignGranularity = 4;
        BasicUnitSize = 4;
      {$else}
        {$ifdef chepersy_align_8}
          {$fatal 64-bit alignment support not implemented yet!}
          //AlignGranularity = 8;
          //BasicUnitSize = 4;
        {$endif}
      {$endif}

    procedure InitStubs;

    type
      TMother = record
        Debug: record
          Verbose: boolean;
        end;
        State: record
          IsRussian: boolean; //controls the log and error messages
        end;
        Timer: record
          UsecDelta: function (ptsc: pqword): double; cdecl;
          RdtscFrequency: double;
        end;
        Core: record
          AddLogOK:  procedure (); cdecl;
        end;
      end;
    PMother = ^TMother;
    var
      _Mother: TMother;
      Mother: PMother = @_Mother;
    {$include mo_cps_tmid_public.h}
  {$endif cge}

type
    {$include mo_cps_dyna.h}
(*
type
  TCpsGcAutoRun = record
    Enabled: boolean;
    NumObjectsToActivate,
    NumObjectsToIgnoreTimeConstraint: integer;
    TimeConstraint: int64;
  end;

  TGarbageCollector = class
  protected
    f_Graveyard: TAOMO;
    f_previousCollectInterrupted: boolean;
    f_prevcollectind: integer;
  public
    {This counter increases with each orphaned object destroyed.
       reset it to zero if you need.}
    ObjectsCollected: int64;
    ThreadSafe: boolean; // false by default
    AutoRun: TCpsGcAutorun;

    {root used for walking the graph. When Nil, Collect()
      does only destroy objects marked prior as orphaned.
      Autorun doesn't work too.}
    GraphRoot: TChepersyObject;
    constructor Create;

    {The Collect function will interrupt its work if it takes more
      CPU clock cycles than TimeConstraint, as per RDTSC command.
      Obviously, walking the graph is most often required,
      which could *not* be interrupted, but the actual
      deleting of objects and emptying the graveyard could,
      so the next time Collect is called it will clean up the
      previously found orphaned objects first, then walk the
      graph again if there is time left.}
    procedure Collect(TimeConstraint: int64 = 0); virtual;
    procedure ForceFlush; virtual;
    procedure Add(o: TChepersyObject); //doesn't do anything to its scraped state!
    function GraveyardCount: integer; virtual;
  end;

  var
    GarbageCollector: TGarbageCollector;
*)


const
  SignatureString: UnicodeString = 'Chepersy Database File';

type
  TFieldOperation = (fio_Load, fio_Save, fio_Clone, fio_Skip, fio_Walk);
//Removed in 0.8.96  TCustomClassProc = procedure (var o: TObject; op: TFieldOperation);

// ****************** INIT PROCEDURES ********************************
  procedure ClassesRegistrationStart;
  procedure ClassesRegistrationEnd;

{   RegClass must be called for each descendant of TChepersyObject.
      The best place to do it is the initialization procedure of the module.
      Automatically calls the "Register" constructor of a class.
}
  procedure RegClass(C: CChepersyObject);

  // custom classes (any other than the TChepersyObject descendants)
 // introduced in 0.8.95, removed in 0.8.96 due to various problems
 // procedure RegCustomClass(C: TClass; proc: TCustomClassProc);

  //use if your custom class contains any TChepersyObjects
//removed in 0.8.96  procedure CpsProcessObject(var o: TObject; op: TFieldOperation);

// ************** NEW FUNCTIONS IN v0.9 TOTALLY REPLACING THE OLD API *********

  procedure ListFields(v: array of const);

  //class, enum, set or dynamic array
  procedure RegType(info: PTypeInfo);  overload;

  //dynamic array (deprecated) or set (deprecated)
//  procedure RegType(info, base: PTypeInfo);  overload;
  
  //record
  //procedure RegType(tname: Utf8string; Size: longint; v: array of const);overload;
{$ifdef fpc}
  procedure RegType(info: PTypeInfo; Size: longint; v: array of const); overload;
{$endif}

  //enumerated array
  procedure RegType(Info, BaseType, IndEnum: PTypeInfo);  overload;
  //procedure RegType(Info: PTypeInfo; BaseType: Utf8string; IndEnum: PTypeInfo);overload;
  //procedure RegType(Info, BaseType: Utf8string; IndEnum: PTypeInfo);overload;


  //static array
  //procedure RegType(tname: Utf8string; base: PTypeInfo; lohi: array of const); overload;
  //procedure RegType(tname, base: Utf8string; lohi: array of const); overload;
{$ifdef fpc}
  procedure RegType(info, base: PTypeInfo; lohi: array of const); overload;
//  procedure RegType(info: PTypeInfo; base: Utf8string; lohi: array of const);  overload;
{$endif}

  //dynamic array
  procedure RegType(info: PTypeInfo; n: integer; base: PTypeInfo); overload;
  //procedure RegType(info: PTypeInfo; n: integer; base: Utf8string);  overload;
  //procedure RegType(info: PTypeInfo; base: Utf8string);  overload;

  function GetStackLimit(): pointer; inline; //because syste.StackBottom contains some totally bullshit values

  function Md5String(s: RawByteString): TMd5Digest; overload;
  function Md5String(s: UnicodeString): TMd5Digest; overload;

  function GetTypeInfoByName(n: RawByteString): PTypeInfo;
  function GetNameByTypeInfo(p: PTypeInfo): RawByteString;


type
  TArrayofDword = array of dword;
  T2dArrayOfDword = array of array of dword;

  TArrayOfLongint = array of longint;
  T2dArrayOfLongint = array of TArrayOfLongint;
  
  TArrayofByte = array of byte;

 {$ifdef fpc}
  //Delphi has no such type.
  //A fake info is added to the registry to emulate it
  // (sorry, but just defining it as a mirror of int64 is not possible).
  TArrayofQword = array of Qword;
  T2dArrayOfQword = array of array of Qword;
 {$endif}

  TArrayOfint64 = array of int64;
  T2dArrayOfint64 = array of int64;
  

const
  CPS_POINTER = 7777;
  CPS_METACLASS = 7778;
  CPS_MANAGED_OBJECT = 7779;
  CPS_MAX_MULTIDIM_ARRAY = 25; //over this, you cannot register
  


{   This bunny registers array types.
      Due to lack of RTTI it goes via string IDs.}
{DEPRECATED} procedure ObsoleteRegTypeStaticArray(StringID: utf8string; BaseType: PTypeInfo; Size: Integer);

  {$ifdef cge}
    {$include mo_fixedpoint.h}
  {$endif}

type
{$ifdef support0899}
  TFieldInfo = packed record
    offset,
    name,
    Reserved4, //HasPP,  //ancient history by now
    Tind: longint; // the index of its type in the global type registry

    case integer of
      2: (
        DeprecatedSkip: LongBool; //deprecated in 0.9.19, left here for backward compatibility
      );
      3: (
        Reserved3: Longint; //the former Skip field is now free for grabs
        RFind, // the number of this field in the global class registry
        NewTind: longint; //this one is used only at loading,
                          //  in the file it is always zero:
        properties: TFieldPropertiesSet; //new in 0.9.19 //Reserved2: longint;
      );
  end;
{$else}
  TFieldInfo = packed record
    offset,
    name,
    Reserved4,
    Tind, // the index of its type in the global type registry
    Reserved3,
    RFind, // the number of this field in the global class registry
    NewTind: longint;
    properties: TFieldPropertiesSet; //new in 0.9.19 //Reserved2: longint;
  end;
{$endif}

  PFieldInfo = ^TFieldInfo;
  TFieldDesc = record
    Size: longint;
    Field: array of TFieldInfo;
  end;

  {$define RP_CallingConvention:= pascal} //This is a macro :))))

    TCustomTypeProcessingProc = procedure
          (PField: pointer; Op: TFieldOperation; Tind: longint;
            FieldProperties: TFieldPropertiesSet); RP_CallingConvention;
    TCustomConverterProc = procedure
          (PField: pointer; Tind1, Tind2: longint); register;


  TFieldKind =
  //do NOT change the order, or the data files will become incompatible!
  ( fk_notsupported, fk_binary, fk_integer, fk_float,
    fk_string, //UnicodeString
    fk_set, fk_enum, fk_Class,
        fk_Complex, //not used anymore

    fk_8bit_string, //AnsiString / RawByteString

        fk_CustomClass, //ORLY? Not used anymore
        fk_Array, //obsolete, not used.
    fk_enumInd_Array, //array indexed by an enumerated type.
    fk_record, fk_dynamic_array, fk_packed_record, fk_static_array,
        fk_09_custom_class, //not used anymore
    fk_metaclass
  );



    TTypeRecord = record
      name: integer; //index in the global name space table
      size: ptruint;
      Info: PTypeInfo;
      Data: PTypeData;
      Proc: TCustomTypeProcessingProc;
  //    CustomClassProc: TCustomClassProc;
      ClassName: longint;
      TypeInfoMd5: TMd5Digest;
      _class: CChepersyObject;
      Kind: TFieldKind;
      ElementTypeInd,
      IndEnumInd,
      Length: integer;
      IsEnumIndArray: boolean;
      IsRecord: boolean;
     // RecFieldName, RecFieldTInd: array of integer;
      SetLen,
      DynArrayLevel,
      DynArrayElementInd: integer;
//      IsArrayOfTTP: boolean; // is some level array of TChepersyObject or its descendants
      StaticLow: integer;
      RecordInd: integer;
      //hasDiscardAtLoading: boolean; //to avoid an extra call if that virual metod wasn't redefined
      hasBeforeSaving: boolean; //to avoid an extra call if that virual metod wasn't redefined
      hasAfterLoading: boolean; //to avoid an extra call if that virual metod wasn't redefined
      recordAlignment: integer; //8 or 4. Obviously, the compiler aligns a pair of 32-bit fields differently from one 64-bit field

      //These apply to records, static arrays and to everything containing them up the chain
      containsComplexTypes: boolean; // does contain dynamic arrays or strings at some depth
      containsChepersyObjects: boolean; // does contain TChepersyObject at some depth
    end;

  var TestInstanceOfTChepersyObject: TChepersyObject = nil;

type

  {$include mo_cps_spaia.h}

  {$include mo_cps_spala.h}

  {$include mo_cps_dyna_1.h}


type
  TStreamSignature = packed record
    sign: array [0..2] of AnsiChar;
    ver: AnsiChar;
  end;

{$ifdef support08x}
(*
 TOldCGEHeader = packed record
    case integer of
      0: (
        stub0, StreamSize: int64;
        glUintSize,
        glFloatSize,
        PointerSize,
        AlignGranularity: shortint);
      1: (Stub1: packed array[0..31 - 1] of DWORD;) //128 bytes size
  end;
*)
{$endif}

  TNameSpace = class (TAOS)
  public
    Alias: array of integer;
    function Ind(i_s: string; AssumedInd: integer): integer;
    function Translate(RegNS: TNameSpace): TAOI;
    function Add (a_s: Utf8String): integer; overload; override;
    function Add(s1, s2: Utf8string): integer; overload;
  end;


// Alas, no amount of classes could allow it perform two reading/writion operations at once. Nothing could. :(
  TChepersyDatabaseHeader = class
  protected
    class var C: record
      Sig: TStreamSignature;
     {$ifdef support08x}
      Head: TOldCGEHeader;
     {$endif}
      Read: boolean;
      Opened: boolean;
      FN: AnsiString;
    end;
    class procedure CgeFilePropByType(t: AnsiChar; name: UnicodeString);
    procedure WriteSignatureToStream;
    class procedure ReadSignatureFromStream (var version: AnsiChar);
    {new in 0.9.0. Determines the stream format at reading:
      0 for the old one (v0.7x..0.8.x),
      1 for the new one (0.9.0 and up).
      At writing the format is always 1.}
    class var
      StreamVersion: integer;
      VersionOfLoadedFile: AnsiChar;
      ReadChecksum,
      CurrentVersionChecksum: TMd5Digest;
      ReadHeaderSize,
      CurrentVersionHeaderSize: dword;

    constructor CreateForLoad; //uses global variable CpsStream

  public
    Checksum: TMd5Digest;
    HeaderSize: dword;
    //For reading. Reads stream header to determine checksum.
    //Then either calls the constructor or returns a pre-existing class instance from  CpsDatabaseHeaders array.
    class function GetInstance: TChepersyDatabaseHeader;  //uses global variable CpsStream



    constructor CreateDefaultInstance;

    procedure ReadHeader;
    procedure WriteHeader;
    procedure SkipHeader;

    procedure Parse();

    //slapdash converting from a procedural/global variable mess to classes
    var
      B_Enums, EnumConvTable: T2DAOI;
      B_Fields: TFieldInfos;
      B_NameSpace: TNameSpace;
      NameTT,
      NaToT,
      TNames,
      TSizes,
      TKinds,
      B_TBaseInds,
      B_TIndInds,
      CNames,
      B_RecBTinds,
      B_StaticLow: TAOI;
      B_ToGtind,
      B_FromGtind,
      B_Lrecind: array of integer;
      B_RecKinds: array of TFieldKind;
      TypeChanged: array of boolean;
      CAncestorNames: T2DAOI;
      CNums: array of integer; //local class/record index to global class/record index
      EnumIdentic: array of boolean;
      B_Classes: array of CChepersyObject;
      b_ScenarioIndex: array of integer;
  end;

  function GetTypeIndex(info: PTypeInfo): integer;
  function GetTypeInfoMd5(p: PTypeInfo): TMd5Digest; { is calculated once,
    during registration.
    Passing unregistered type returns zero.
    Some types may not have md5 checksum. Initially, intended only
    for classes and enumerated types }

  
implementation


{$include mo_cps_body.inc}


  function CpsStore(o: TChepersyObject; {$ifdef cge}m: UnicodeString;{$endif} Target: TStream;
              XorMask: dword = $ffffffff; AndMask: dword = $ffffffff; Headerless: longbool = false): longbool;
  var
    ContName: UnicodeString;
    q, qT: qword;
    t: double;
    n: integer;
  begin
    {$ifdef benchmarking}
      {$include un_bottleneck_start}
    {$endif}
    Mother^.Timer.UsecDelta(@qT);
    {$ifdef cge}
      AddLog(MI_SAVING, [m]);
    {$endif}
    ValidateChepersyState('CpsStore()');
    Result:= false;
    CpsCriticalSection.Enter;
    InitBuffer;
    {$ifndef cge}
    CpsError.Clear;
    {$endif}
    mXorMask:= XorMask;
    mAndMask:= AndMask;
    try
      try
        PrepareToSave(Target); //assigns CpsStream
        CpsStackLimit:= GetStackLimit();

        DH:= CpsDatabaseHeaders[0];
        if not Headerless then begin
          DH.WriteSignatureToStream;
          DH.WriteHeader;
        end;
    //     {$include un_bottleneck_start.inc}
         Mother^.Timer.UsecDelta(@q);
         WritePersistent(o, []);
         t:= Mother^.Timer.UsecDelta(@q);
         n:= Obj.High;
         if Mother^.Debug.Verbose then AddLog('Chepersy: saved %0 objects in %1 microseconds (%2 nanoseconds per object)', [n, round(t), round(t * 1000 / max(1, n))]);
    //     {$include un_bottleneck_stop.inc}
         DoneSaving(No);
         Result:= true;
      except
        if Target is TFileStream
          then ContName:=(Target as TFileStream).FileName
          else Contname:=Target.ClassName;
        Try EmergencyCloseCgeFile Except End;
       {$ifdef cge}
        Die(MI_ERROR_SAVING, [m, ContName]);
       {$else}
        try Die(MI_ERROR_SAVING, [ContName]); except end;
      {$endif}
      end;
    finally
      CpsCriticalSection.Leave;
    end;
    {$ifdef benchmarking}
      {$include un_bottleneck_stop}
    {$endif}
    Mother^.Core.AddLogOK;
    if Mother^.Debug.Verbose then AddLog('Chepersy: finished saving, total time %0 microseconds', [round(Mother^.Timer.UsecDelta(@qT))]);
  end;


  function CpsLoad(
    {$ifdef cge}m: UnicodeString;{$endif}
    Source: TStream;
    POutputList: PArrayOfPointer = nil;//PArrayOfChepersyObjects = nil;
    DatabaseHeader: TChepersyDatabaseHeader = nil //if set, treats the stream as headerless!
  ): TChepersyObject;
  var
//    t: integer;
    ContName: UnicodeString;
    q, qT: qword;
    t: double;
    n: integer;
  begin
    {$ifdef benchmarking}
      {$include un_bottleneck_start}
    {$endif}
    Mother^.Timer.UsecDelta(@qT);
   {$ifdef cge}
    AddLog(MI_LOADING, [m]);
   {$endif}
    ValidateChepersyState('CpsLoad()');
    Result:= nil;
    CpsCriticalSection.Enter;
    InitBuffer;

    {$ifndef cge}
    CpsError.Clear;
    {$endif}
    try
      try
  //t:=tick();
         //{$include un_bottleneck_start.inc}
         PrepareToLoad(Source);

         CpsStackLimit:= GetStackLimit();

         if not Assigned(DatabaseHeader)
           then databaseHeader:= TChepersyDatabaseHeader.GetInstance() {
             loads/skips header from stream as needed
             Sets DH:= Self in the process!
         }
         //  else Fuck, got NIL here :(
          ;
            DH:= DatabaseHeader;


         if Mother^.Debug.Verbose then AddLog('Chepersy: loading objects...');
         Mother^.Timer.UsecDelta(@q);
//bgsay('  obj=' + Inttohex(ptruint(pointer(obj)), 8));
//bgsay('  DH=' + Inttohex(ptruint(pointer(DH)), 8));
//bgsay(   'Length(DH.Cnums)=' + IntToStr(Length(DH.Cnums)));
         ReadPersistent(pointer(Result), []);
         t:= Mother^.Timer.UsecDelta(@q);
         n:= Obj.High;
         if Mother^.Debug.Verbose then AddLog('Chepersy: loaded %0 objects in %1 microseconds (%2 nanoseconds per object)', [n, round(t), round(t * 1000 / max(1, n))]);
  //       n:=NumObjects;
  //if State.VerboseLog then AddLog('Loading %0 objects from the file took %1 ms (%3 ns per object).'#10#13'FileFormat is %2',[n, Tick()-t, 'CGE' + VersionOfLoadedFile, int(1000000.0*((Tick()-t)/n))]);

         if Mother^.Debug.Verbose then AddLog('Chepersy: performing post-load actions...');
         Mother^.Timer.UsecDelta(@q);
         DoneLoading(No, POutputList);
         t:= Mother^.Timer.UsecDelta(@q);
         if Mother^.Debug.Verbose then AddLog('Chepersy: processing took %0 microseconds (%1 nanoseconds per object)', [round(t), round(t * 1000 / max(1,n))]);
//AddLog('CpsLoad(): high(OutputList) = %0', [high(OutputList)]);
        //{$include un_bottleneck_stop.inc}
      except
        if Source is TFileStream
          then ContName:=(Source as TFileStream).FileName
          else Contname:=Source.ClassName;
        Try EmergencyCloseCgeFile Except End;
        {$ifdef cge}
          Die(MI_ERROR_LOADING, [m, ContName]);
        {$else}
          try Die(MI_ERROR_LOADING, [ContName]); except end;
          result:= nil;
        {$endif}
      end;
      if Mother^.Debug.Verbose then AddLog('Chepersy: finished loading, total time %0 microseconds', [round(Mother^.Timer.UsecDelta(@qT))]);
    finally
      CpsCriticalSection.Leave;
    end;
    {$ifdef benchmarking}
      {$include un_bottleneck_stop}
    {$endif}
    Mother^.Core.AddLogOK;
  end;

{
  function TChepersyObject.Exists: boolean;
  begin
    if not Assigned(Self) then Exit(false);
    Result:= RefCount > 0; //not Scraped;
  end;
}

  function CpsClone(
      o: TChepersyObject; {$ifdef cge}m: UnicodeString;{$endif}
       XorMask: dword = $ffffffff; AndMask: dword = $ffffffff): TChepersyObject;
  var
    ContName: UnicodeString;
    q, qT: qword;
    t: double;
    n: integer;
  begin
    {$ifdef benchmarking}
      {$include un_bottleneck_start}
    {$endif}
    Mother^.Timer.UsecDelta(@qT);
    {$ifdef cge}
      if Mother^.Debug.Verbose then AddLog('Chepersy is cloning...');
    {$endif}
    ValidateChepersyState('CpsClone()');
    Result:= nil;
    CpsCriticalSection.Enter;
    {$ifndef cge}
    CpsError.Clear;
    {$endif}
    mXorMask:= XorMask;
    mAndMask:= AndMask;
    try
      try
        PrepareToClone;

    //     {$include un_bottleneck_start.inc}
         Mother^.Timer.UsecDelta(@q);
         ClonePersistent(pointer(o), []);
         t:= Mother^.Timer.UsecDelta(@q);
         n:= Obj.High;
         if Mother^.Debug.Verbose then AddLog('Chepersy: cloned %0 objects in %1 μs (%2 ns per object)', [n, round(t), round(t * 1000 / max(1, n))]);
    //     {$include un_bottleneck_stop.inc}
         Result:= TChepersyObject(Obj[1]);
         DoneCloning;
      except
        Die(RuEn(
          'Сбой при клонировании части графа объектов',
          'Failure to clone a part of the object graph'));
      end;
    finally
      CpsCriticalSection.Leave;
    end;
    {$ifdef benchmarking}
      {$include un_bottleneck_stop}
    {$endif}
    if Mother^.Debug.Verbose then AddLog('Chepersy: finished cloning, total time %0 μs', [round(Mother^.Timer.UsecDelta(@qT))]);
  end;


end.
