{$ifndef fpc} ## FREE PASCAL ONLY! {$endif}

//Uncomment to allow compiler to optimize field placement of classes and records.
//Warning: as of fpc 2.6.4 this leads to suboptimal memory usage
//as the instance sizes get rounded up, leading to unused memory blocks
// in cases of long inheritance chains!
//Also, there is no guarantee that field placement algorithms stay the same
//in the future versions. It's all guesswork, after all.
//{$define chepersy_no_packing}


//Will raise errors at the registration stage if any 64-bit fields
//aren't aligned to 64-bit boundary. Otherwise they could be not,
//due to the packrecords 4 directive
{$define chepersy_assert_64_fields_are_aligned}

//since 0.9.0 the data format has changed.
//  comment this out to disable the support for reading the old format
//{$define support08x}

//since 0.9.19 the data format has changed.
//  comment this out to disable the support for reading the old format
//{$define support0899}


//uncomment to get comprehensible messages on corrupt data files (slows down)
// It activates automatically when you switch {$assertions on}
//WARNING! Huge performance impact! Slows down 2 to 3 times.
{$define safeloading}

//uncomment to use the old class indexing method with a VMT hack.
//  It may be a bit faster but less compatible and reliable: if user's
//   antivirus software is too paranoidal or if future Windows enforces
//   a paranoidal Data Execution Prevention policy, your application WON'T RUN!
//{$define oldhacky} //not tested in v0.9 yet!




{$mode objfpc} //sorry I have to have my own operators now. Which doesn't work in Delphi mode.//  {$mode delphi}
{$macro on}
{$inline on}

{$if (FPC_FULLVERSION>=20700) and (FPC_FULLVERSION<=30100)}
  {$inline off}
{$endif}

{$if FPC_FULLVERSION>=20700}
  {$optimization noorderfields}
  {$modeswitch unicodestrings}
{$endif}

{$ifndef chepersy_no_packing}
  {$packrecords 4} //yes, even on a 64-bit cpu.
  {$define chepersy_align_4}
{$endif}

{$longstrings on}
{$minenumsize 4}
{$typeinfo on}

{$ifdef safeloading}
  {$rangechecks on}
  {$overflowchecks on}
{$endif}



//For finding bottlenecks in code. Requires the four includes un_bottleneck_XXX.inc
// to be in the compiler path
//Outputs results via AddLog during saving and loading.
//Calculates using Asm includes working directly with RDTSC,
//so resol;ution is VERY high (most CPUs nowadays have TSC invariancy,
//e.g. the counter is the same for all cores and always runs at a fixed
//frequency. Mine is always 2.5 GHz even as i5 fluctuates between 0.8 and 2.9)

{define benchmarking}

{$ifdef benchmarking}
  {$ifndef cpu386}
    {$fatal Currently only supported on 32-bit i386!}
  {$endif}

  //Preferrably, only uncomment one at a time!

  //calling the ClassIndex class method
  //Only applicable to saving
  //0.1%
  {define bench_classindex}

  //setting the CpsIndex field to mark already saved objects (may be cache-unfriendly)
  //Only applicable to saving and walking
  //0.1%
  {define bench_setting_object_index}

  //calling Obj.Add(o);
  //1%
  {define bench_adding_object_to_array}

  //Total time writing from/to stream
  // valid for saving and loading
  //{$define bench_streaming}

  //This one should catch CPU cache misses, in theory
  //valid for saving
  //0.5%
  {define bench_first_access_to_object}

  //This one is for sanity checking only. Shoud give near 100%.
  //NOT compatible with any others
  //valid for saving and loading
  //50% reading, 60% writing. WTF? Should be 99.9% or the like!
  {define bench_read_persistent}

  //How much do NewInstance calls eat up?
  //valid for loading
  //2%
  {define bench_constructor_calls}

{$endif}




