{
    This file is part of Chentrah,
    Copyright (C) 2004-2008 Anton Rzheshevski (chebmaster@mail.ru).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see http://www.gnu.org/licenses/

 ********************************************************************** 

    This file contains the configuration manager class implementation.

 **********************************************************************}


type
  TConfLimRec = record
    s: UnicodeString;
    min, max: integer;
  end;

var
  {format: 'SECTION.ID', necessarily in uppercase}
  ConfigLimits: array [0..0] of TConfLimRec = (
    (s:'NONE.NONE'; min:0; max:1)
{    (s:'VIDEO.HEIGHT'; min:480; max:1536),
    (s:'VIDEO.WIDTH'; min:640; max:2048) }
  );

 function LimitsValid(sect, id: UnicodeString; val: integer): boolean;
 var
   j: integer;
   u: UnicodeString;
 begin
   Result:= Yes;
   u:=UpperCase(sect + '.' + id);
   For j:=0 to High(ConfigLimits) do
     with ConfigLimits[j] do
       if (s = u) and ((val < min) or (val > max)) then begin
         Result:= No;
         Exit;
       end;
 end;

 function TConfigManager._Find(s, i: UnicodeString): UnicodeString;
 var
   j, maxj: integer;
   us: UnicodeString;
   found: boolean;
 begin
   try
     CS.Enter;
     VerboseLog ('  Reading from config [%0].%1 ...', [s,i]); //can only use the
     // basic/emergency logging that supports caching, yet

     Result:='';
     wherefound:=0;
     us:= lowercase(s);

     if Assigned(ini[2]) and (us = 'videocard') then begin
       wherefound:= 2;
       //it is only assigned much further into initialization where we can safely use normal logging
       if Mother^.Debug.Verbose then AddLogComment(' redirected to the per-videocard INI file... ');
       Result:= ini[2].ReadString(s, i, '');
       if Mother^.Debug.Verbose then AddLogComment(' = "' + Result + '"');
     end
     else begin
       found:= false;
       for j:=1 downto 0 do begin
         Result:= ini[j].ReadString(s, i, '');
         if Result <> '' then begin
           if j = 1
             then VerboseLog('    .. = "' + Result + '", found in the user INI file ' + ini[1].FileName)
             else VerboseLog('    .. = "' + Result + '", found in the main INI file ' + ini[0].FileName);
           wherefound:=j;
           if (j = 1) and ( us = 'main')
             then Die(MI_ERROR_CONFIG_MAIN_NO_OVERRIDE, [ini[1].FileName, ini[0].FileName, lowercase(i)]);
           found:= true;
           break;
         end;
       end;
       if not found then VerboseLog('    .. not found.');
     end;
   finally
     CS.Leave;
   end;
 end;
 
 function TConfigManager._ReadStr(s, i: UnicodeString): UnicodeString;
 begin
   Result:=_Find(LowerCase(s), LowerCase(i));
 end;

Function OptiPath(s: UnicodeString): UnicodeString;
var
  sl1, sl2: TStringList;
  j: integer;
begin
  sl1:=TStringList.Create;
  sl1.Text:=(StrReplace(StrReplace(ExtractFilepath(s), '\', #13), '/', #13));
  sl2:=TStringList.Create;
  For j:=0 to sl1.Count - 1 do begin
    if sl1[j]='..' then sl2.Delete(sl2.Count - 1)
                   else sl2.Add(sl1[j]);
  end;
  sl1.Free;
  Result:='';
  For j:=0 to sl2.Count - 1 do Result:=Result + sl2[j] + PathSlash;
  sl2.Free;
   {$ifdef win32}
     Result:=StrReplace(Result, '/', '\');
   {$else}
     Result:=StrReplace(Result, '\', '/');
   {$endif}
end;

 function TConfigManager._ReadPath(s, i: UnicodeString): TFileNameString;
 begin
   Result:=
   {$ifdef windows}
     StrReplace(
   {$endif}
     StrReplace(StrReplace(StrReplace(StrReplace(UnicodeToFileName(_ReadStr(s, i)),
       '{$PLATFORM}', SystemSuffix),
         '{$HOMEDIR}', Mother^.State.HomePath),
            '{$INSDIR}', Mother^.State.InstallPath),
              '{$State.UserName}', Mother^.State.UserName)
            
   {$ifdef windows}
     , '{$CSIDL_APPDATA}', Mother^.State.CSIDL_AppDataDir)
   {$endif}
     ;
   Result:= OptiPath(Result) + ExtractFileName(Result);
   {$ifdef windows}
     Result:=StrReplace(Result, '/', '\');
   {$else}
     Result:=StrReplace(Result, '\', '/');
   {$endif}
   VerboseLog('    .. = "%0"', [Result]);
 end;
 
 procedure TConfigManager._WritePath(s,i: UnicodeString; d: TFileNameString);
 begin
   _WriteStr(OptiPath(s), i, d);
 end;

 function TConfigManager._ReadBool(s, i: UnicodeString): boolean;
 var j: integer;
 begin
   if StrictLimitsCheck then j:=_ReadInt(s, i, 0, 1)
                        else j:=_ReadInt(s, i);
   Result:=(j = 1);
 end;

 function TConfigManager._ReadInt(s, i: UnicodeString): integer;
 var v: UnicodeString;
 begin
   v:=_ReadStr(s, i);
   if v='' then v:='0';
   Try
     Result:=StrToInt(v);
   Except
     Die (MI_ERROR_INVALID_CONFIG_PARAMETER_NO_LIM, [Ini[wherefound].FileName, s, i, v])
   End;
   VerboseLog('    .. = %0', [Result]);
 end;

 function TConfigManager._ReadInt(s, i: UnicodeString; min, max: integer): integer;
 var v: UnicodeString;
 begin
   if not StrictLimitsCheck then begin
     Result:=_ReadInt(s, i);
     if Result < min then Result:=min;
     if Result > max then Result:=max;
   end
   else begin
     v:=_ReadStr(s, i);
     if v='' then v:='0';
     Try
       Result:=StrToInt(v);
       if (Result < min) or (Result > max) then raise Exception.Create('Value is out ot bounds!');
     Except
       Die (MI_ERROR_INVALID_CONFIG_PARAMETER, [Ini[wherefound].FileName, s, i, v, min, max])
     End;
   end;
   VerboseLog('    .. = %0', [Result]);
 end;


function TConfigManager.GetFloat(sect, id: UnicodeString; defval: float = 0.0) : float;
var v: UnicodeString;
begin
  v:= _ReadStr(sect, id);
  if v = '' then Result:= defval
  else begin
    Try
      Result:=StrToFloat(v);
    Except
      Die (MI_ERROR_INVALID_CONFIG_PARAMETER_NO_LIM, [Ini[wherefound].FileName, sect, id, v])
    End;
  end;
  VerboseLog('    .. = %0', [Result]);
end;

function TConfigManager._ReadFloat(s, i: UnicodeString): float;
var v: UnicodeString;
begin
  v:=_ReadStr(s, i);
  if v='' then v:='0';
  Try
    Result:=StrToFloat(v);
  Except
    Die (MI_ERROR_INVALID_CONFIG_PARAMETER_NO_LIM, [Ini[wherefound].FileName, s, i, v])
  End;
  VerboseLog('    .. = %0', [Result]);
end;
 
 function TConfigManager._ReadIntChk(s, i: UnicodeString; min,max: integer): integer;
 begin
   Result:=_ReadInt(s, i);
   if (Result < min) or (Result > max) then
     ScreamLimitsRead(wherefound, s, i, Result, min, max);
 end;

 function TConfigManager._ReadFloatChk(s, i: UnicodeString; min,max: float): float;
 begin
   Result:= _ReadFloat(s, i);
   if (Result < min) or (Result > max) then
     ScreamLimitsRead(wherefound, s, i, Result, min, max);
 end;


 procedure TConfigManager._WriteStr(s,i,d: UnicodeString);
 var
   FileName2: TFileNameString;
   which: TCustomIniFile;
   discard: boolean;
 begin

   if Mother^.Debug.Verbose then AddLog('  Writing to config, [%0].%1 = "%2"...', [s,i,d]);
   s:= LowerCase(s);
   if s = 'main'
     then Die(MI_ERROR_PROGRAMMER_NO_BAKA,
       [PervertedFormat('attempt to write [main].%0 = %1 to the config.', [i, d])]);
//addlog('--_WriteStr("%0", "%1, "%2")',[s, i, d]);

   try
     CS.Enter;
     discard:= false;
     which:= ini[1];
     if s = 'videocard' then begin
       if Mother^.Debug.Verbose then AddLogComment(' redirected to the per-videocard INI file...');
       if not Assigned (ini[2]) then begin
         if Mother^.GAPI.RendererAlias <> '' then begin
           if Mother^.State.IsPortable
             then FileName2:= Mother^.State.HomePath + ExtractFilePath(ConfigFilePath)
               + 'video_'
               + SanateStringForFileName({$I %FPCTARGETOS%} + '-' + {$I %FPCTARGETCPU%})
               + '_' + Mother^.GAPI.RendererAlias + '.ini' { a portable application
                 will have mostly separate settings for separate machines
                 but will lose them at every drivers update }
             else FileName2:= 'video.ini'; { for an installed application
               it is better to keep settings despite driver updates }
           if Mother^.Debug.Verbose then AddLog('   Attaching or creating %0 ...', [FileName2]);
           ini[2]:= TIniFile.Create(FileName2);
           if Mother^.Debug.Verbose then AddLogComment(' Ok');
         end
         else begin
           if Mother^.Debug.Verbose then AddLog('   DISCARDED: OpenGL not initialized yet, device name unknown.');
           //CS.Leave;
           //Exit;
           discard:= true;
         end;
       end;
       which:= ini[2];
     end;

     if not discard then which.WriteString(
       {$ifdef che_unicode}
         s, i, d
       {$else}
         Utf8Encode(s), Utf8Encode(i), Utf8Encode(d)
       {$endif}
     );
     if Mother^.Debug.Verbose then AddLogComment(' Ok');
   finally
     CS.Leave;
   end;
 end;

 procedure TConfigManager._WriteInt(s,i: UnicodeString; v:integer);
 begin
   _WriteStr(s, i, IntToStr(v));
 end;

procedure TConfigManager._WriteFloat(s,i: UnicodeString; v:float);
begin
  _WriteStr(s, i, FloatToStr(v));
end;
 
 procedure TConfigManager._WriteIntChk(s,i: UnicodeString; min,max,v:integer);
 begin
   if (v < min) or (v > max) then begin
     if Mother^.Debug.Verbose then AddLog('Attempt to write invalid configuration data. Section ='+s+' par='+i+' min='+IntToStr(min)+' max='+IntToStr(max)+' value='+IntToStr(v)+'. Value truncated.');
   end;
   if v < min then v:=min;
   if v > max then v:=max;
   _WriteStr(s, i, IntToStr(v));
 end;

 procedure TConfigManager._WriteFloatChk(s,i: UnicodeString; min, max, v:float);
 begin
    if (v < min) or (v > max) then begin
     if Mother^.Debug.Verbose then AddLog('Attempt to write invalid configuration data. Section ='+s+' par='+i+' min='+FloatToStr(min)+' max='+FloatToStr(max)+' value='+FloatToStr(v)+'. Value truncated.');
   end;
   if v < min then v:=min;
   if v > max then v:=max;
   _WriteStr(s, i, FloatToStr(v));
 end;

 procedure TConfigManager._WriteBool(s,i: UnicodeString; b: boolean);
 begin
   if b then _WriteStr(s, i, '1') else _WriteStr(s, i, '0');
 end;

 Constructor TConfigManager.Create;
 var
   sl: TStringList;
 begin
   CS:= TProfiledCriticalSection.Create;
   //initially, load it from the built-in constant:

   VerboseLog('Creating temporary main INI file based on the string embedded in the executable...');
   sl:=TStringList.Create;
   sl.Text:=DefaultConfString;
   ini[0]:= TMemIniFile.Create('', false);
   (ini[0] as TMemIniFile).SetStrings(sl);
   sl.Free;

   VerboseLog('Creating temporary fake user INI file to avoid crashing on writes to config...');
   ini[1]:= TMemIniFile.Create('', false); { temporary, to allow functionig without
     the actual file to write to }
 end;

 Destructor TConfigManager.Destroy;
 var j: integer;
 begin
   if ini[1].ClassType <> TMemIniFile then ini[1].UpdateFile;
   ini[1].Free;
   if Assigned(ini[2]) then begin
     if ini[1].ClassType <> TMemIniFile then ini[2].UpdateFile;
     ini[2].Free;
   end;
   CS.Free;
 end;

 procedure TConfigManager.ScreamLimitsRead(fnum: integer; s,i: UnicodeString; val, min, max: integer);
 begin
   Die(MI_ERROR_INVALID_CONFIG_PARAMETER,[Ini[fnum].FileName, s, i, val, min, max]);
 end;

procedure TConfigManager.ScreamLimitsRead(fnum: integer; s,i: UnicodeString; val, min, max: float);
begin
  Die(MI_ERROR_INVALID_CONFIG_PARAMETER,[Ini[fnum].FileName, s, i, val, min, max]);
end;
 
 procedure TConfigManager.LoadMainFile;
 var fn: TFileNameString;
 begin
   fn:= Mother^.State.InstallPath +  MyAppName + '.ini';
   VerboseLog('Opening the real main INI file ' + fn + ' ...');
   ini[0].Free;
//WriteLn('IFC: ', Mother^.State.InstallPath +  MyAppName + '.ini');
   ini[0]:= TIniFile.Create(fn);
 end;

 procedure TConfigManager.LoadUserFile;
 var fn: TFileNameString;
 begin
   fn:= Mother^.State.HomePath + ConfigFilePath;
   VerboseLog('Attaching or creating the user INI file ' + fn + ' ...');
   if Mother^.State.HomePath = '' then Exit;
   ValidateWritePath(Mother^.State.HomePath + ExtractFilePath(ConfigFilePath));
   ini[1].Free;
//WriteLn('IFC: ', Mother^.State.HomePath + ConfigFilePath);
   ini[1]:= TIniFile.Create(fn);
   //a strange bug in Debian 7 64-bit with wine 32-bit:
   //  Chentrah crashes with EFileCreate if the ini file already exists
   //P.S. COPY-PASTED the entire IniFiles unid adding a few lines to avoid using
   //  the broken TStrings.LoadFromFile and TStrings.SaveToFile
   //P.P.S. RESOLVED: it's a bug of the RTL, TFileStream.Destroy fails
   //  if Handle is 4.
 end;
 

 function TConfigManager.Wherefoundname: UnicodeString;
 begin
   Result:=AnsiToWide(ini[wherefound].FileName);
 end;


 procedure TConfigManager.GetSet(penum: pointer; ti: ptypeinfo; sect, id: UnicodeString);
 var
   v: UnicodeString;
   e, l: array of UnicodeString;
   i, j, n: integer;
   D: PTypeData;
 begin
   if not (ti^.Kind in [tkEnumeration, tkInteger]) then  Die(MI_ERROR_PROGRAMMER_NO_BAKA, [RuEn('приведённая TypeInfo не принадлежит перечислимому типу', 'the TypeInfo provided os not of an enumerated type')]);
   D:= GetTypeData(ti);
   if D^.MinValue <> 0 then Die(MI_ERROR_PROGRAMMER_NO_BAKA, [RuEn('не рассчитано на поддиапазонные типы', 'not designed for subrange types')]);
   if D^.MaxValue > 31 then Die(MI_ERROR_PROGRAMMER_NO_BAKA, [RuEn('рассчитано только на типы, множества от которых приводимы к dword (не более 32 элементов)', 'only designed for enums with derivative sets type-castable to dword (no more than 32 entries)')]);
   dword(penum^):= 0;
   v:= Self._ReadStr(sect, id);
   if v = '' then Exit;
   SetLength(e, D^.MaxValue + 1);
   for i:= 0 to High(e) do
     if ti^.Kind = tkEnumeration
       then e[i]:= GetEnumName(ti, i)
       else e[i]:= IntToStr(i);
   l:= Explode(UnicodeString(','), v, true);
   for i:=0 to High(l) do begin
     n:= -1;
     for j:= 0 to High(e) do
       if e[j] = l[i] then begin
         n:= j;
         break;
       end;
     if n < 0 then Die(MI_ERROR_INVALID_CONFIG_PARAMETER_ENUM, [Ini[1].FileName, sect, id, l[i]]);
     dword(penum^):= dword(penum^) or (dword(1) shl n);
   end;
 end;


procedure TConfigManager.SetSet(penum: pointer; ti: ptypeinfo; sect, id: UnicodeString);
 var
   e: array of UnicodeString = nil;
   i: integer;
   D: PTypeData;
 begin
   if not (ti^.Kind in [tkEnumeration, tkInteger]) then  Die(MI_ERROR_PROGRAMMER_NO_BAKA, [IntToStr(D^.MinValue) + RuEn('приведённая TypeInfo не принадлежит перечислимому типу', 'the TypeInfo provided os not of an enumerated type')]);
   D:= GetTypeData(ti);
   if D^.MinValue <> 0 then Die(MI_ERROR_PROGRAMMER_NO_BAKA, [RuEn(
    'не рассчитано на поддиапазонные типы',
    'not designed for subrange types')]);
   if D^.MaxValue > 31 then Die(MI_ERROR_PROGRAMMER_NO_BAKA, [RuEn(
    'рассчитано только на перечислимые типы, множества от которых приводимы к dword (не более 32 элементов)',
    'only designed for enums with derivative sets type-castable to dword (no more than 32 entries)')]);
   for i:= 0 to D^.MaxValue do begin
     if (dword(penum^) and (dword(1) shl i)) = 0 then continue;
     SetLength(e, Length(e) + 1);
     if ti^.Kind = tkEnumeration
       then e[High(e)]:= GetEnumName(ti, i)
       else e[High(e)]:= IntToStr(i);
   end;
   Self._WriteStr(sect, id, Implode(',', e));
 end;

