{
    This file is part of Chentrah,
    Copyright (C) 2004-2008 Anton Rzheshevski (chebmaster@mail.ru).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see http://www.gnu.org/licenses/

 ********************************************************************** 

   This file contains the graphical console class implementation.

 **********************************************************************}


 { TConsole }

procedure TConsole.SetGLStatesForGUI;
begin
  case  Mother^.GAPI.Mode of
    gapi_GL21, gapi_DeprecatedImmediateGL: begin
      glViewport(0, 0, Mother^.Display.WindowClientRect.Width, Mother^.Display.WindowClientRect.Height);
      glMatrixMode(GL_PROJECTION);
      glLoadIdentity();
      glMatrixMode(GL_MODELVIEW);
      glLoadIdentity();
      glOrtho(0, Mother^.Display.WindowClientRect.Width, Mother^.Display.WindowClientRect.Height, 0, 0 ,1);
    end;
  else
    DieUnsupportedGLMode;
  end;
end;

 procedure TConsole.Fade();
 begin
   f_vis:=False;
   if not(Assigned(Module) and (Module.Index = 0)) //.BimLoaded))
     then f_lasttime:=Now()
     else f_lasttime:=Now() - 1; //1 day ago ;)
 end;

 procedure TConsole._SetVisible(b: boolean);
 begin
   if not b then f_scrolled_back:=0;
   if b = f_vis then exit;
   f_vis:=b;
  // f_lasttime:=Now();
 end;

 constructor TConsole.Create;
 begin
   inherited Create;
   f_Limit:=1000;
   f_vis:= Mother^.State.DeveloperMode or Mother^.Debug.Enabled;
   Add('*');
 end;
 
 procedure TConsole.SetDefaultBg (bg: TDefaultBGs);
 begin
   //SetBg(BaseImagesPath + DefaultBgNames[bg, GetIsRussian()]);
   if Mother^.State.InstallPath = ''
     then begin
       SetBgPtr(@error_bg_img, sizeof(error_bg_img));
       _CgeErrorCursor;
     end
     else
       Case bg of
         dbTitle: begin
           SetBg(Mother^.State.InstallPath + ImagesDir + 'cl_background_default.jpg');
           CgeStandardCursor;
         end;
         dbError: begin
           SetBg(Mother^.State.InstallPath + ImagesDir + 'cl_background_error.jpg');
           _CgeErrorCursor;
         end;
         dbWarning: begin
           SetBg(Mother^.State.InstallPath + ImagesDir + 'cl_background_warning.jpg');
           CgeStandardCursor;
         end;
         dbInstalling: begin
           SetBg(Mother^.State.InstallPath + ImagesDir + 'cl_background_default.jpg');
           CgeHourglassCursor;
         end;
         dbShutdown: begin
           SetBg(Mother^.State.InstallPath + ImagesDir +'shutdown.jpg');
           CgeHourglassCursor;
         end;
       else
         Die(MI_ERROR_PROGRAMMER_NO_BAKA, ['FIXME: unknown constant ' + GetEnumName(typeinfo(TDefaultBgs), ord(bg)) + ' in TConsole.SetDefaultBg()']);
       end;
 end;
 
 
  procedure ResampleImg (var i: TImageData; newwidth, newheight: integer);
  var
    i2: TImageData;
  begin
    NewImage(newwidth, newheight, i.Format, i2);
    StretchResample(i, 0, 0, i.Width, i.Height, i2, 0, 0, i2.Width, i2.Height,
      CGE_RESAMPLE_FILTER, No);
    FreeImage(i);
    i:=i2;
  end;


 procedure TConsole._ApplyBg(var b: TImageData);
 var
   w, h: integer;
 begin
   if f_bgtex <> 0 then glDeleteTextures(1, @f_bgtex);
   glGenTextures(1, @f_bgtex);
   glBindTexture(GL_TEXTURE_2D, f_bgtex);
   glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
   glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
   if b.Format <> ifR8G8B8 then ConvertImage(b, ifR8G8B8);
   w:= GetHigherPowerOf2(b.Width) shr f_bgmip;
   h:= GetHigherPowerOf2(b.Height) shr f_bgmip;
//writeln('width=', b.Width, ' height=', b.height);
//AddLog('%0', [f_bgmip]);
   if (b.Width <> w) or (b.Height <> h)
     then ResampleImg(b, w, h);
   glTexImage2D (GL_TEXTURE_2D, 0, GL_RGB, w, h, 0, {$ifndef glesonly}GL_BGR{$else}GL_RGB{$endif}, GL_UNSIGNED_BYTE, b.Bits);
   FreeImage(b);
 end;


 procedure TConsole.SetBg (name: TFileNameString);
 var
   b: TImageData;
 begin
   if (name = f_bgname) and (f_bgmip = _RequiredBgmip()) then exit;
   try
     fillchar(b, sizeof(b), 0);
     f_bgname:= CgePath(name);
     f_bgmip:= _RequiredBgmip();
     LoadImageFromFile(f_bgname, b);
//writeln('* "', name, '", "', f_bgname, '"');
     _ApplyBg(b);
   except end;
   f_bgptr:= nil;
 end;

 function TConsole._RequiredBgmip(): integer;
  //with high Timer.QF, background will be oversized even for small window sizes
  // with low 1f, background will be low-res
 var
   wr, hr : integer;
   delta: float;
 begin
   Result:= 0;
   //assumed background image size is 2048x1024
{
   FUCK THIS!

   wr:= 2048 shr f_bgmip;
   hr:= 1024 shr f_bgmip;
   with Mother^
     do delta:= math.min(Display.WindowClientWidth / wr , Display.WindowClientHeight / hr ) * (Timer.QF / 100);
   if delta > 0.9 then Result:= math.max(0, f_bgmip - 1);
   if delta < (0.9 / 2.6) then Result:= math.min(3, f_bgmip + 1);
}
 end;

 procedure TConsole.SetBgPtr (p: pointer; size: integer);
 var
   b: TImageData;
 begin
   if p = f_bgptr then exit;
   try
     LoadImageFromMemory(p, size, b);
     _ApplyBg(b);
   except end;
   f_bgname:='';
 end;


 procedure TConsole.InitGL;
 var
   name: integer;
   p: pointer;
   i: TImageData;
   x, y: integer;
   file_name: TFileNameString = 'consolefont.png';

 begin
   Try
     f_limit:=Config.IntChk['console','MaxLines',150,10000];
     f_hidetime:=Config['console','MessageDisplayTime'];
     f_fadetime:=Config['console','MessageFadeOutTime'];
     if f_hidetime = 0 then f_hidetime:=1; //division by zero is bad
     if f_fadetime = 0 then f_fadetime:=1;
     f_tostretch:= Config.Int['console','DoubleSizeIfResolutionHighetThan'];
     f_tosquish:= Config.Int['console', 'UseCondensedFontIfResolutionLowerThan'];

     //if (Mother^.State.InstallPath = '') //Not possible now. The program stops with an error.

     file_name:= Mother^.State.InstallPath + ImagesDir + 'consolefont.png';
     if not FileExists(file_name)
       then Die(MI_ERROR_FILE_NOT_FOUND, [file_name]);
     p:= LoadPic(PFileNameChar(file_name));
     CheckForGuardedException;
     glGenTextures(1, @f_texture);
     glBindTexture(GL_TEXTURE_2D, f_texture);
     glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
     glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
     glTexImage2D (GL_TEXTURE_2D, 0, GL_LUMINANCE_ALPHA{GL_RGBA}, GetPicWidth(p), GetPicHeight(p), 0, GetPicFmt(p), GL_UNSIGNED_BYTE, GetPicPixels(p));
     DeletePic(p);
     CheckForGuardedException;

     file_name:= Mother^.State.InstallPath + ImagesDir + 'consolefont-half.png';
     if not FileExists(file_name)
       then Die(MI_ERROR_FILE_NOT_FOUND, [file_name]);
     p:= LoadPic(PFileNameChar(file_name));
     CheckForGuardedException;
     glGenTextures(1, @f_halftexture);
     glBindTexture(GL_TEXTURE_2D, f_halftexture);
     glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
     glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
     glTexImage2D (GL_TEXTURE_2D, 0, GL_LUMINANCE_ALPHA{GL_RGBA}, GetPicWidth(p), GetPicHeight(p), 0, GetPicFmt(p), GL_UNSIGNED_BYTE, GetPicPixels(p));
     DeletePic(p);
     CheckForGuardedException;


     file_name:= Mother^.State.InstallPath + ImagesDir + TFileNameString('bsod.png');
     if not FileExists(file_name)
       then Die(MI_ERROR_FILE_NOT_FOUND, [file_name]);
     p:= LoadPic(PFileNameChar(file_name));
     CheckForGuardedException;
     glGenTextures(1, @f_bsodtexture);
     glBindTexture(GL_TEXTURE_2D, f_bsodtexture);
     glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
     glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
     glTexImage2D (GL_TEXTURE_2D, 0, GL_LUMINANCE, GetPicWidth(p), GetPicHeight(p), 0, GetPicFmt(p), GL_UNSIGNED_BYTE, GetPicPixels(p));
     CheckForGuardedException;
     DeletePic(p);
     CheckForGuardedException;

     glGenTextures(1, @f_noisetexture);

     SetDefaultBg(dbTitle);

     InitAnimatedLogo;

   Except
     Die(MI_DIED_INITIALIZING, [RuEn('консоли','console')]);
   End;
 end;
 



 function CCLd(a: glFloat; b: integer): glFloat;
 begin
   Case b of
     1: if a < 0 then Result:=1.0 else Result:=sqr(1.0 - a);
     2: if a < 0 then Result:=1.0 else Result:=sqr(1.0 - a);
     3: if a < 0
          then Result:=1.0
          else
            if a < 1.0 then Result:=sqrt(1.0 - a)
                       else Result:=0;
     4: if a < 0 then Result:=1.0 else Result:=(1.0 - a);
   end;
 end;

   function SLines(s:string; _m: integer): integer;
   begin
     Result:=1 + ((Length(s) - 1) div (_m - 1));
   end;

 procedure randoby(var b: byte); inline;
 var c: integer;
 begin
   c:=Random(101) - 50;
   if (integer (b) + c < 0) or (integer(b) + c > 255)
     then dec(b, c)
     else inc(b, c);
 end;

 procedure TConsole.DrawBSODBackground;
 var
   bsodvoffs, ntdx, ntdy: GLfloat;
   x, y: integer;
   r1, g1, b1, r, g, b: float;
   Mesh: TDumbUniMesh;
 begin
   f_vis:= true;
   f_fade:= 1;

   if not Mother^.Display.CanRender then Exit;

   SetGLStatesForGUI;

   //procedural noise texture
   for y:=0 to (ntz div 4) - 1 do
     for x:=0 to ntz - 1 do
       f_noisearray[y,x]:= random(256);
      // randoby(f_noisearray[x, y]);


   if Mother^.State.StateTrashedRestartRequired
   then begin
     r1:= Mother^.Display.FadeIn * 0.35; g1:=0; b1:=0;//red
     r:= 1; g:= 0.4; b:= 0;
   end
   else begin
     r1:= 0; g1:= 0; b1:= Mother^.Display.FadeIn; //blue
     r:= 1; g:= 1; b:= 1;
   end;


   if MessageContainer.CurrentLanguage = 0
   { Russian has hard-coded index 0, English - 1.
     One half of the texture is Russian "blue screen of death",
       another half is Latin "BSOD"}
     then bsodvoffs:= 0.5
     else bsodvoffs:= 0;
   glBindTexture(GL_TEXTURE_2D, f_bsodtexture);
   Mesh:= TDumbUniMesh.Create;
   Mesh.AddQuad2d(
     r1, g1, b1, 1,
     0, 0, Mother^.Display.WindowClientRect.Width, Mother^.Display.WindowClientRect.Height,
     0, 0 + bsodvoffs, 1, 0.5 + bsodvoffs);
   TGAPIAbstractionLayer.SetGLStatesForGUI;
   glDisable(GL_BLEND);
  case  Mother^.GAPI.Mode of
   {$ifndef glesonly}
    gapi_GL21, gapi_DeprecatedImmediateGL :begin
      glDisable(GL_ALPHA_TEST);
    end;
   {$endif glesonly}
    gapi_GLES2: begin
    end;
  else
    DieUnsupportedGLMode;
  end;

   Mesh.Render;
   Mesh.Free;

{
   glBegin(GL_QUADS);
     glTexCoord2f(0, 0.5 + bsodvoffs);
     glVertex2f(0, Mother^.Display.WindowClientRect.Height); //bottom left corner
     glTexCoord2f(0, 0 + bsodvoffs);
     glVertex2f(0, 0);
     glTexCoord2f(1, 0 + bsodvoffs);
     glVertex2f(Mother^.Display.WindowClientRect.Width, 0);
     glTexCoord2f(1, 0.5 + bsodvoffs);
     glVertex2f(Mother^.Display.WindowClientRect.Width, Mother^.Display.WindowClientRect.Height);
   glEnd;
}

   glBindTexture(GL_TEXTURE_2D, f_noisetexture);
   glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
   glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

   ntdx:= 0;
   ntdy:= random();
   glTexImage2D (GL_TEXTURE_2D, 0, GL_LUMINANCE, ntz, ntz div 4, 0, GL_LUMINANCE, GL_UNSIGNED_BYTE, @f_noisearray[0,0]);

   Mesh:= TDumbUniMesh.Create;
   Mesh.AddQuad2d(
     Mother^.Display.FadeIn * r,  Mother^.Display.FadeIn * g, Mother^.Display.FadeIn * b, 0.3,
     0, 0, Mother^.Display.WindowClientRect.Width, Mother^.Display.WindowClientRect.Height,
     0 + ntdx, 0 + ntdy, 1 + ntdx, 5 + ntdy);

   TGAPIAbstractionLayer.SetGLStatesForGUI;
   glEnable(GL_BLEND);
   glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
   Mesh.Render;
   Mesh.Free;

{
   glNormal3f( 0.0, 0.0, 1.0);
  // glColor3f(0.5, 0.5, 0.35 + 0.35 * sin(Now * 100000.0) / 2);
   glColor4f(
     1 * Mother^.Display.FadeIn * r ,
     1 * Mother^.Display.FadeIn * g,
     1 * Mother^.Display.FadeIn * b, 0.3);
   glBegin(GL_QUADS);
     glTexCoord2f(0 + ntdx, 5 + ntdy);
     glVertex2f(0, 0); //bottom left corner
     glTexCoord2f(0 + ntdx, 0 + ntdy);
     glVertex2f(0, Mother^.Display.WindowClientRect.Height);
     glTexCoord2f(1 + ntdx, 0 + ntdy);
     glVertex2f(Mother^.Display.WindowClientRect.Width, Mother^.Display.WindowClientRect.Height);
     glTexCoord2f(1 + ntdx, 5 + ntdy);
     glVertex2f(Mother^.Display.WindowClientRect.Width, 0);
   glEnd;
}
 end;








 procedure TConsole.Draw();
 var
   xmax, zms, beginfrom, endat, viewportbottom: integer;
   condensed: boolean;
   liT, Red, Green, Blue, Alpha: glFloat;
   currentYTop, yOffset: integer;
   //processing from bottom to top until the next line is completely off-screen
   function DrawOneLine() : boolean;
   var
     xp, yp, x, y, c, cc, L2, y2, SubLines, tlen: integer;
     ts, tt: GLfloat;
   begin
     if f_currentLineTop < endat then Exit(false);
     tlen:= Length(Self.Strings[f_currentLineTop]);
     if (tlen > 1)
       then SubLines:= 1 + ((tlen - 1 - 1) div (xmax))
       else SubLines:= 1;
     currentYTop -= Sublines * trunc(ConsoleCharHeight * f_zoom);
     if (currentYTop + Sublines * ConsoleCharHeight * f_zoom) < 0 //completely off screen?
       then Exit(false);
     x:=0;
     y:=0;
     c:=1;
     While c < tlen - 1 do begin
       //draw character;
       xp:= trunc(ConsoleBorders) + x * trunc(ConsoleCharWidth * f_zoom);
       yp:= y * trunc(ConsoleCharHeight * f_zoom);
       cc:=Ord(Self.Strings[f_currentLineTop][c]);
       if cc >= 32 then begin //ignore special characters
         ts:=((cc mod 16) * ConsoleCharXstep + ConsoleCharOffsetS) / ConsoleImageWidth;
         tt:=((cc div 16) * 16 + ConsoleCharOffsetT) / 256.0;
         f_mesh.AddQuad2d(
           Red, Green, Blue, Alpha,
           xp, yp + currentYTop,
             xp + (ConsoleCharWidth + 1) * f_zoom, yp + currentYTop + ConsoleCharHeight * f_zoom,
           ts, tt, ts + (ConsoleCharWidth + 1) / ConsoleImageWidth, tt + ConsoleCharHeight / 256.0 );
       end;

       inc (c);
       inc (x);
       if x > xmax then begin
         x:=1;
         inc(y);
       end;
     end;
     f_meshHeight+= SubLines * trunc(ConsoleCharHeight * f_zoom);
     f_currentLineTop-= 1;
     Result:= true;
   end;
 begin
   if not Mother^.Display.CanRender then Exit;

   zms:= Round((Now() - f_lasttime) * 1000.0 * SECONDS_PER_DAY );

   LastHeartbeatMoment:=Now();
   If not (f_vis
           or (zms < (f_hidetime + f_fadetime)))
   then begin
     f_fadeline:= -1;
     f_fade:= 1.0;
     Exit;
   end;

   if f_vis then begin
     liT:=0;
     f_fadeline:= -1;
   end
   else begin
     if zms < f_hidetime then liT:= - (f_hidetime - zms) / f_hidetime
                         else liT:=(zms - f_hidetime) / f_fadetime;
   end;

   if f_scrolled_back < 0 then f_scrolled_back:=0;
   if not f_vis then f_scrolled_back:=0;

//  f_fade:= 1.0;

   Alpha:= CCLd(liT, 4) * Mother^.Display.FadeIn;

   condensed:= ConsoleUseCondensedFont or (Mother^.Display.WindowClientRect.Width < f_tosquish);

   if not Assigned(f_mesh)
     or f_change.Yes
     or (f_change.lit <> trunc(Alpha * 1000))
     or (f_change.sb <> f_scrolled_back)
     or (f_change.dw <> Mother^.Display.WindowClientRect.Width)
     or (f_change.dh <> Mother^.Display.WindowClientRect.Height)
     or (f_change.condensed <> condensed)
     or (f_change.fadeline <> f_fadeline)
   then begin
     if Assigned(f_mesh) then begin f_mesh.Free; f_mesh:= nil end;

     if Self.Count < 1 then Exit;

     f_change.Yes := No;
     f_change.dw := Mother^.Display.WindowClientRect.Width;
     f_change.dh := Mother^.Display.WindowClientRect.Height;
     f_change.sb := f_scrolled_back;
     f_change.lit := trunc(Alpha * 1000);
     f_change.condensed:= condensed;
     f_change.fadeline:= f_fadeline;

      Red:= CCLd(liT, 1);
      Green:= CCLd(liT, 2);
      Blue:= CCLd(liT, 3);

     if condensed then begin
       ConsoleCharWidth:= 4;
       ConsoleCharXstep:= 128 div 16;
       ConsoleImageWidth:= 128.0;
       ConsoleCharOffsetS:= 2;
     end
     else begin
       ConsoleCharWidth:= 9;
       ConsoleCharXstep:= 256 div 16;
       ConsoleImageWidth:= 256.0;
       ConsoleCharOffsetS:= 4;
     end;

      f_zoom:=1.0;

      if Mother^.Display.WindowClientRect.Width > f_tostretch then f_zoom:= 2.0;
      xmax := -1 + ((Mother^.Display.WindowClientRect.Width
        - trunc(2 * ConsoleBorders * f_zoom ) ) div  trunc(f_zoom * ConsoleCharWidth));

      currentYTop:= Mother^.Display.WindowClientRect.Height - trunc(ConsoleBorders);
      beginfrom:= Self.Count - 1;
      if f_fadeline > 0 then endat:= f_fadeline else endat:= 0;

      f_currentLineTop:= max(endat, beginfrom - f_scrolled_back);

      f_mesh:= TDumbUniMesh.Create;

      f_meshHeight:= trunc(0.7 * ConsoleCharHeight * f_zoom);// 0;
      while DrawOneLine() do;
   end;

   if (f_currentLineTop < 1) and (f_scrolled_back > 0)
     and (f_meshHeight < Mother^.Display.WindowClientRect.Height
                                          - trunc(ConsoleCharHeight * f_zoom))
       then Dec(f_scrolled_back);

   f_viewBottom:= min(f_meshHeight, Mother^.Display.WindowClientRect.Height);

   if Assigned(f_mesh) then begin
     SetGLStatesForGUI;
     TGAPIAbstractionLayer.SetGLStatesForGUI;
     if condensed
       then glBindTexture(GL_TEXTURE_2D, f_halftexture)
       else glBindTexture(GL_TEXTURE_2D, f_texture);
     glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
     glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
     yOffset:= max(0, Mother^.Display.WindowClientRect.Height - f_meshHeight);
     case  Mother^.GAPI.Mode of
       gapi_GL21, gapi_DeprecatedImmediateGL: begin
         glViewport(0, max(0, Mother^.Display.WindowClientRect.Height - f_viewBottom),//max(0, Mother^.Display.WindowClientRect.Height - f_meshHeight),
           Mother^.Display.WindowClientRect.Width, Mother^.Display.WindowClientRect.Height);
       end;
     else
       DieUnsupportedGLMode;
     end;


     f_mesh.Render;
   end;
 end;

 procedure TConsole.SetScrolledBack(i: integer);
 begin
   if (f_scrolled_back < i)
     and (f_currentLineTop < 1) and (f_scrolled_back >= 0)
     and (f_meshHeight < Mother^.Display.WindowClientRect.Height
                                            - trunc(ConsoleCharHeight * f_zoom))
       then Exit;

   f_scrolled_back:= i;
 end;

{{
           glBegin(GL_QUADS);
             glTexCoord2f(ts, tt); //top left
             glVertex2f(xp, Mother^.Display.WindowClientRect.Height - (Rdh - yp));
             glTexCoord2f(ts + (ConsoleCharWidth + 1) / ConsoleImageWidth, tt); //top right
             glVertex2f(xp + (ConsoleCharWidth + 1) * zoom, Mother^.Display.WindowClientRect.Height - (Rdh - yp));
             glTexCoord2f(ts + (ConsoleCharWidth + 1) / ConsoleImageWidth, tt + ConsoleCharHeight / 256.0); //bottom right
             glVertex2f(xp + (ConsoleCharWidth + 1) * zoom, Mother^.Display.WindowClientRect.Height - (Rdh - yp - ConsoleCharHeight * zoom));
             glTexCoord2f(ts, tt + ConsoleCharHeight / 256.0); //bottom left
             glVertex2f(xp, Mother^.Display.WindowClientRect.Height - (Rdh - yp - ConsoleCharHeight * zoom));
           glEnd; }  }
 
 procedure TConsole.DrawBackground();
 var f: glFloat;
   Mesh: TDumbUniMesh;
 begin
   if not Mother^.Display.CanRender then Exit;

   RecalculateFade;
   Self.SetGLStatesForGUI;

   if f_bgmip <> _RequiredBgmip() then SetBg(f_bgname);

   f:=f_fade;

   Mesh:= TDumbUniMesh.Create;
   Mesh.AddQuad2d(
     f * Mother^.Display.FadeIn, f * Mother^.Display.FadeIn, f * Mother^.Display.FadeIn, 1,
     0, 0, Mother^.Display.WindowClientRect.Width, Mother^.Display.WindowClientRect.Height,
     0, 0, 1, 1);
   TGAPIAbstractionLayer.SetGLStatesForGUI;
   case  Mother^.GAPI.Mode of
    {$ifndef glesonly}
     gapi_GL21, gapi_DeprecatedImmediateGL : begin
       glDisable(GL_ALPHA_TEST);
     end;
    {$endif glesonly}
     gapi_GLES2: begin
     end;
   else
     DieUnsupportedGLMode;
   end;

   glDisable(GL_BLEND);
   glEnable(GL_TEXTURE_2D);
   glBindTexture(GL_TEXTURE_2D, f_bgtex);
   Mesh.Render;
   Mesh.Free;
{
   glColor4f(f * Mother^.Display.FadeIn, f * Mother^.Display.FadeIn, f * Mother^.Display.FadeIn, 1);
   glBindTexture(GL_TEXTURE_2D, f_bgtex);
   glBegin(GL_QUADS);
     glTexCoord2f(0, 0);
     glVertex2f(0, 0); //bottom left corner
     glTexCoord2f(0, 1);
     glVertex2f(0, Mother^.Display.WindowClientRect.Height);
     glTexCoord2f(1, 1);
     glVertex2f(Mother^.Display.WindowClientRect.Width, Mother^.Display.WindowClientRect.Height);
     glTexCoord2f(1, 0);
     glVertex2f(Mother^.Display.WindowClientRect.Width, 0);
   glEnd;
   glEnable(GL_ALPHA_TEST);
   glEnable(GL_BLEND);
}
 end;
 
 procedure TConsole.BeginBlackout(delta: double);
 begin
   Mother^.Display.BlackoutBegin:= Now() + delta / SECONDS_PER_DAY;
   RecalculateFade;
 end;

 procedure TConsole.RecalculateFade;
 begin
   f_fade:= FancyFunc( (Now() - Mother^.Display.BlackoutBegin) * 84600 / FADE_IN_TIME);
   Mother^.Display.FadeIn:= f_fade;
 end;

 procedure TConsole.DrawLogo();
 var
   f: integer;
   s, a, n: glFloat;
   cx, cy: glFloat;
   phase, animphase: glFloat;
   Mesh: TDumbUniMesh;
 begin
   phase:= (Now() - Mother^.Display.LogoAnimationStart) * SecondsPerDay / LOGO_ANIMATION_LENGTH;
   if phase > 1.0 then Mother^.Display.LogoAnimationPlaying:= No;
   if not Mother^.Display.LogoAnimationPlaying then Exit;

   animphase:= (phase - logo_fadein) / (1.0 - logo_fadeout - logo_fadein);
   f:= round(animphase * length(f_logotex));
   f:= min (f,  high(f_logotex));
   f:= max (0, f);

   if phase < logo_fadein then begin
     //fade in
     n:= max(0.0, 1 - (phase / logo_fadein));
     s:= logo_size + 3500.0 * n;
     a:= sqr(sqr( 1 - n));
   end
   else
     if phase < (1.0 - logo_fadeout) then begin
       //animate
       s:=logo_size;
       a:=1;
     end
     else begin
       //fade out
       s:= logo_size;
       a:= sqr((1.0 - phase) / logo_fadeout);
     end;

   if not Assigned(f_logotex) then Exit;

   Mesh:= TDumbUniMesh.Create;
   s:=s/2;
   cx:= Trunc(Mother^.Display.WindowClientRect.Width /2);
   cy:= Trunc(Mother^.Display.WindowClientRect.Height /2);
   Mesh.AddQuad2d(
     a * 2, a * 3, 1, sqr(a),
     cx - s, cy - s, cx + s, cy + s,
     0, 0, 1, 1);
   TGAPIAbstractionLayer.SetGLStatesForGUI;
   glEnable(GL_TEXTURE_2D);
   glBindTexture(GL_TEXTURE_2D, f_logotex[high(f_logotex) - f]);
   Mesh.Render;
   Mesh.Free;

{   SetGlStates2d;

   glColor4f(a*2, a*3, 1, sqr(a));
//   glBlendFunc(GL_DST_COLOR, GL_SRC_COLOR);
   s:=s/2;
   cx:= Trunc(Mother^.Display.WindowClientRect.Width /2);
   cy:= Trunc(Mother^.Display.WindowClientRect.Height /2);
   if Assigned(f_logotex) then begin
     glEnable(GL_TEXTURE_2D);
     glBindTexture(GL_TEXTURE_2D, f_logotex[high(f_logotex) - f]);
     glBegin(GL_QUADS);
       glTexCoord2f(0, 1);
       glVertex2f(cx - s, cy + s); //bottom left corner
       glTexCoord2f(0, 0);
       glVertex2f(cx - s, cy - s);
       glTexCoord2f(1, 0);
       glVertex2f(cx + s, cy - s);
       glTexCoord2f(1, 1);
       glVertex2f(cx + s, cy + s);
     glEnd;
   end;
}
 end;

 
 procedure TConsole.InitAnimatedLogo;
 var
   i: integer;
   logo: TDynImageDataArray;
   f: string;
   {$ifdef win32} msg:TMSG;{$endif}
 begin
   if NowRestarting or (Mother^.State.InstallPath = '') then Exit;
   Try
     LoadMultiImageFromFile(Mother^.State.InstallPath + ImagesDir + 'chentrah-logo.mng', logo);
     
     SetLength(f_logotex, Length(logo));
//addlog('==== %0 logo frames',[length(logo)]);
     glEnable(GL_TEXTURE_2D);
     glGenTextures(length(logo), @f_logotex[0]);
     For i:=0 to high(logo) do begin
       if logo[i].Format <> ifGray8 then ConvertImage(logo[i], ifGray8);
       glBindTexture(GL_TEXTURE_2D, f_logotex[i]);
       glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
       glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
       glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
       glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
       glTexImage2D (GL_TEXTURE_2D, 0, GL_ALPHA, 256, 256, 0, GL_ALPHA, GL_UNSIGNED_BYTE, logo[i].Bits);
     end;
     FreeImagesInArray(logo);
   except
     AddLog(RuEn('Не удалось загрузить логотип!'#10#13'%0','Unable to load the logo!'#10#13'%0'), [StopDying()])
   end;
 end;

 procedure TConsole.Add(s: AnsiString);
 var
   i: integer;
 begin
   //recursion. Scary.
   i:=Pos(#13, s);
   if (i > 0) and (i < Length(s)) then begin
     if i <= 2 then s:= Copy(s, i + 1, Length(s) - i)
     else begin
       s[i]:=#0;//don't draw it
       Add(Copy(s, 1, i));
       Add(Copy(s, i, Length(s) - i + 1));
       Exit;
     end;
   end;
   if (s <> '') and (s[1]<>#0) then s:=#127 + s;
   inherited Add(s);
   if f_fadeline < 0 then f_fadeline:=Count - 1;
   if Count > f_limit then begin
     if f_fadeline > 0 then dec(f_fadeline);
     Delete(0);
   end;
   f_change.Yes:= Yes;
     {$ifdef debug_build}
       //The console window plays that role.
       f_lasttime:=Now() - 1 //a day ago
     {$else}
       if (Assigned(Module) and (Module.Index = 0))//Module.BimLoaded))
        or not (Mother^.State.DeveloperMode or Mother^.Debug.Enabled)
         then f_lasttime:=Now() - 1 //a day ago
         else f_lasttime:=Now();
     {$endif}
 end;
 
 procedure TConsole.AddComment(S: AnsiString);
 var
   i: integer;
 begin
   if Count = 0 then Add(S)
   else begin
     i:=Pos(#13, s);
     if (i > 0) and (i < Length(s)) then begin
       s[i]:=#0;//don't draw it
       if i <= 2
       then Add(Copy(s, i + 1, Length(s) - i))
       else begin
         AddComment(Copy(s, 1, i));
         Add(Copy(s, i, Length(s) - i + 1));
       end;
       Exit;
     end;
     Strings[Count - 1] := Strings[Count - 1] + S;
     if (Assigned(Module) and (Module.Index = 0)) //.BimLoaded))
       or not (Mother^.State.DeveloperMode or Mother^.Debug.Enabled)
       then f_lasttime:=Now() - 1 //a day ago
       else f_lasttime:=Now();
     f_change.Yes:= Yes;
   end;
 end;

 procedure TConsole.Delete(i: integer);
 begin
   inherited Delete(i);
 end;

