type
  TThreadLIRec = record
    Abra: ShortString3;
    KLus: array[TThreadLoadKind] of double;
  end;

var
  tLIRs: array of TThreadLIRec;


  procedure MeasureThreadLoads(delta: double);
  var
    tsca: array[TThreadLoadKind] of int64;
    total: int64;
    i: integer;
    k: TThreadLoadKind;
    pes: PMotherSehState;
  begin
    pes:= @Mother^.ExceptionState;
    SetLength(tLIRs, 0);
    while Assigned(pes) do begin
      if not (pes^.ThreadFinished)
      then begin
        SetLength(tLIRs, Length(tLIRs) + 1);
        with tLIRs[High(tLIRs)] do begin
          Abra:= pes^.AbbrTitleForIndicator;
          MOVE(Pes^.tli, tsca, SizeOf(tsca));
          //NO locking, so this here may give a glitch if the thread adds at the same moment
          total:= 0;
          for k in TThreadLoadKind do total+= tsca[k];
          if total > round (0.8 * delta * Mother^.Timer.RDTSCFrequency)
            //some threads may sleep for *longer* intervals than the FPS measuring interval!
            then FillChar(Pes^.tli, SizeOf(Pes^.tli), 0);

          if total > 0 then begin
            for k in TThreadLoadKind do
              KLus[k]:= 1000000000.0 * tsca[k] / total;
          end
          else begin
            // if all are zero, use current one as 100% (again, thread switches states too slowly)
            FillChar(KLus[k], SizeOf(KLus[k]), 0);
            KLus[pes^.tlic]:= 1000000000.0;
          end;
        end;
      end;
      pes:= pes^.Next;
    end;
  end;


  function DoesModulehaveThreadsRunning: boolean; { Ideally, module itself
     should answer this question. But! If the module's main thread is
     incapacitated, there is no other way but parse thread exceptuion states
     and find if all module threads have marked themselves as finished
  }
  var
    pes: PMotherSEHState;
  begin
    Result:= false;
    pes:= @Mother^.ExceptionState;
    while Assigned(pes) do begin
      Result:= Result or (pes^.IsModuleThread and not pes^.ThreadFinished);
      pes:= pes^.Next;
    end;
  end;


var
  framecount,
  swb_framecount: integer;
  swb_accum: double = 0;

  procedure MeasureFps;
  var
    n: double;
    dt: double;
  begin
    inc(framecount);
    if not Mother^.Timer.SuspendSwapBuffersMeasurement then begin
      inc (swb_framecount);
      swb_accum+= Mother^.Timer.LastSwapBuffersDuration;
    end;
    n:= Second();
    if prevfpssec =0 then prevfpssec:= n;
    dt:= (n - prevfpssec);
    if dt < FPS_MEASURING_INTERVAL then exit;

    Mother^.Timer.FPS:= framecount / max(dt, 0.0011);

    //Must suspend measurement during operations like creating FBOs to avoid a feedback loop
    if swb_framecount > 0 then Mother^.Timer.AverageSwapBuffersDuration:= swb_accum / swb_framecount;

    swb_accum:= 0;
    prevfpssec:= n;
    framecount:=0;
    swb_framecount:= 0;

    MeasureThreadLoads(dt);
  end;


  procedure InitFpsCounter;
  var
    par: TFpsCounterParam;
    parn: string;
  begin
    for par in TFpsCounterParam do begin
      parn:= GetEnumName(typeinfo(TFpsCounterParam), ord(par));
      parn:= copy(parn, 4 + 1, length(parn) - 4);
      if Config.Bool['FpsCounter', parn]
        then Mother^.Timer.FpsDisplayingParams += [par]
    end;
  end;

  procedure RenderFpsCounter;
  var
    fit: TStringFitRec;
    left, top: integer;
    zoomX, zoomY: GLfloat;
    ws: UnicodeString;
    par: TFpsCounterParam;
    Mesh: TDumbUniMesh;
    procedure OutputString(ws: UnicodeString);
    begin
      FixedFont.RenderString (@fit, left, top, PUnicodeChar(ws));
      top-= fit.ActualHeight;
    end;
    procedure RenderBarDiag(var tLIR: TThreadLIRec; left, top, width, height: integer);
    var
      x, x2: integer;
      k: TThreadLoadKind;
      tot: double;
    begin
      glDisable(GL_TEXTURE_2D);
      Mesh:= TDumbUniMesh.Create;
      x:= 0;
      tot:= 0;
      for k in TThreadLoadKind do begin
        if k = tlk_Sleeping then continue;
        tot+= tLIR.KLus[k];
        x2:= round(tot * width / 1000000000);
        Mesh.AddQuad2d(
          ThreadLoadKindColor[k, 0], ThreadLoadKindColor[k, 1], ThreadLoadKindColor[k, 2], 1,
          left + x, top, left + x2, top + height,
          0,0,1,1);
        {
        glColor3fv(@ThreadLoadKindColor[k]);
        glBegin(GL_QUADS);
          glVertex2f(left + x, top);
          glVertex2f(left + x2, top);
          glVertex2f(left + x2, top + height);
          glVertex2f(left + x, top + height);
        glEnd();
        }
        x:= x2;
      end;
      Mesh.AddQuad2d(
        ThreadLoadKindColor[tlk_Sleeping, 0], ThreadLoadKindColor[tlk_Sleeping, 1], ThreadLoadKindColor[tlk_Sleeping, 2], 1,
        left + x, top, left + width, top + height,
        0,0,1,1);

      {
      glColor3fv(@ThreadLoadKindColor[tlk_Sleeping]);
      glBegin(GL_QUADS);
        glVertex2f(left + x, top);
        glVertex2f(left + width, top);
        glVertex2f(left + width, top + height);
        glVertex2f(left + x, top + height);
      glEnd();
      glColor4f(1,1,1,1);
      }
      Mesh.Render;
      glEnable(GL_TEXTURE_2D);
    end;

    procedure RenderThreadLoads;
    var
      i: integer;
      w: UnicodeString;
      fit2: TStringFitRec;
    begin
      FixedFont.FitStringIntoRectangle (
        FPS_COUNTER_MAX_FONT_SIZE, 3, 1,
          round(fit.ActualWidth * 0.2), 100,  @fit2);
      for i:= 0 to High(tLIRs) do with tLIRs[i] do begin
        if (Abra = 'gpp') or (Abra = 'LGu') then continue;
        w:= Abra;
        FixedFont.RenderString (@fit2, left, top, PUnicodeChar(w));
        RenderBarDiag(tLIRs[i], left + fit2.ActualWidth + 2, top + 3, fit.ActualWidth - fit2.ActualWidth - 4, fit.ActualHeight - 6);
        top-= fit.ActualHeight;
      end;
    end;
    procedure RenderThreadLoadsVerbose;
    var
      i, left2: integer;
      d: int64;
      w: UnicodeString;
      k: TThreadLoadKind;
      fit2: TStringFitRec;
    begin
      for i:= 0 to High(tLIRs) do with tLIRs[i] do begin
        w:= Abra + ': ';
        for k in TThreadLoadKind do begin
          d:= Math.min(999999, round(KLus[k] * 0.001));
          if d = 0
            then w+= '         '
            else w+= ThreadLoadKindAbbr[k] + format('%.6d',[d]);
          if k < High(TThreadLoadKind) then w+= ' ';
        end;
        FixedFont.FitStringIntoRectangle (
          FPS_COUNTER_MAX_FONT_SIZE, Length(w), 1,
          round(Mother^.Display.WindowClientRect.Width) - 10, 100,  @fit2);
        left2:= Mother^.Display.WindowClientRect.Width - fit2.ActualWidth - 4;
        FixedFont.RenderString (@fit2, left2, top, PUnicodeChar(w));
        RenderBarDiag(tLIRs[i], left2, top + fit2.ActualHeight - 3, fit2.ActualWidth - 4, 4);
        top-= fit2.ActualHeight;
      end;
    end;
  begin
    if not DoShowFps then exit;
    //FPS:xxx
    // qf:xxx
    FixedFont.FitStringIntoRectangle (
      FPS_COUNTER_MAX_FONT_SIZE, 9, 1,
      round(FPS_COUNTER_SIZE * (Mother^.Display.WindowClientRect.Width + Mother^.Display.WindowClientRect.Height) / 2),
      100,  @fit);
   // SetGlStates2d;
    FixedFont.SetRenderState (false);
    FixedFont.color4f(1, 1, 1, 1);
    left:= Mother^.Display.WindowClientRect.Width - fit.ActualWidth - 4;
    top:= Mother^.Display.WindowClientRect.Height - fit.ActualHeight;
    if Mother^.Timer.FPS > 999.0 then Mother^.Timer.FPS:=999.0;
    if Mother^.Timer.QF > QF_MAX then Mother^.Timer.QF:= QF_MAX;
    for par in TFpsCounterParam do
      if par in Mother^.Timer.FpsDisplayingParams then begin
        case par of
          fcp_ThreadLoadsVerbose: RenderThreadLoadsVerbose;
          fcp_QualityFactor: OutputString(format('  qf:%4d',[round(Mother^.Timer.QF)]));
          fcp_FramesPerSecond: OutputString(format('  FPS:%3d',[round(Mother^.Timer.FPS)]));
          fcp_RdtscFrequency: OutputString(format('cpu:%.3f',[Mother^.Timer.RDTSCFrequency / 1000000000]));
          fcp_SwapBuffersTime: OutputString(format('swb:%.3f',[Mother^.Timer.AverageSwapBuffersDuration / 1000]));
          fcp_ThreadLoads: RenderThreadLoads;
        else
          Die(MI_ERROR_PROGRAMMER_NO_BAKA, [
           'FPS counter render not implemented for '
            + GetEnumName(typeinfo(TFpsCounterParam), ord(par))]);
        end;
      end;
    if _dsay <> '' then FixedFont.RenderString (@fit, 5, top, PUnicodeChar(_dsay));
    _dsay:='';
  end;

