{
    This file is part of Chentrah,
    Copyright (C) 2004-2008 Anton Rzheshevski (chebmaster@mail.ru).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see http://www.gnu.org/licenses/

 **********************************************************************
 
    This file contains the OpenGL initialization and function loading
    routines

 **********************************************************************}



var
  LibGL, LibEGL: {$ifdef use_dynlibs} THandle {$else} Pointer {$endif} ;
  LibGLName, LibEGLName: TFileNameString;

  eFixedPipelineEmulationProgram: GLuint;

  function GetGLProcAddress2(ProcName, Suffix1, Suffix2: ansistring): Pointer;
  begin
    //Result:=_GetGLProcAddress(ProcName);
    //if not Assigned(Result)
    //  then
    Result:=_GetGLProcAddress(ProcName + Suffix1);
    if not Assigned(Result)
      then Result:=_GetGLProcAddress(ProcName + Suffix2);
    if not Assigned(Result)
      then Die(MI_ERROR_GL_FUNCTION_NOT_FOUND,
        [ProcName + Suffix1 + '() | ' + ProcName + Suffix2 + '()',
         utf8Decode(LibGlName), MessageContainer[MI_HINT_OPENGLTOOLOW]]);
  end;


  function GetGLProcAddress(ProcName: ansistring; Suffix: ansistring = ''): Pointer;
  var fn: string;
  begin
    Result:=_GetGLProcAddress(ProcName);
    if not Assigned(Result) and (Suffix <> '')
      then Result:=_GetGLProcAddress(ProcName + Suffix);
    {$ifdef windows}
      if not Assigned(Result) and (Suffix = '') and Assigned(wglGetProcAddress)
        then Result:=wglGetProcAddress(PAnsiChar(ProcName));
    {$endif}
    if not Assigned(Result) then begin
      fn:= ProcName + '()';
      if Suffix <>'' then fn:=fn + ' | ' + ProcName + Suffix + '()';
      Die(MI_ERROR_GL_FUNCTION_NOT_FOUND,
        [fn, '' + LibGlName, MessageContainer[MI_HINT_OPENGLTOOLOW]]);
    end;
  end;


  function exGetGLProcAddress(ProcName: PAnsiChar; Suffix: PAnsiChar = nil): Pointer; cdecl;
  begin
   try
    Result:=GetGlProcAddress(PCharToString(ProcName), PCharToString(Suffix));
   except
    ProcessGuardedException();
   end;
  end;

  function exGetGLProcAddress2(ProcName, Suffix1, Suffix2: PAnsiChar): Pointer; cdecl;
  begin
   try
    Result:=GetGlProcAddress2(PCharToString(ProcName), PCharToString(Suffix1), PCharToString(Suffix2));
   except
    ProcessGuardedException();
   end;
  end;



  function _GetGLProcAddress(ProcName: ansistring): Pointer;
  var
    baseaddr: pointer;
    exename: TFileNameString;
  begin
    Result:= nil;
    if Assigned(Framework) then Result:= Framework.GetOpenGLProcAddress(ProcName);
    if Assigned(Result) and Mother^.Debug.Verbose
      then AddLog('    %0() at %1 in %2', [ProcName, Result, GetExeFileName(Result)]);

    if not Assigned(Result) then begin
      {$ifndef use_dynlibs}
        Result:=dlsym(LibGL, PAnsiChar(ProcName));
      {$else}
        // in Windows, the extension functions are loaded differently from the
        // "common" functions. What a pain.
        {$ifdef windows}
          if Assigned(wglGetProcAddress)
            then Result:=wglGetProcAddress(PAnsiChar(ProcName));
        {$endif}
        if not Assigned(Result)
          then Result:=GetProcAddress(LibGL, PAnsiChar(ProcName));
      {$endif}
      if not Assigned(Result) and Assigned(eglGetProcAddress)
          then Result:= eglGetProcAddress(PAnsiChar(ProcName));
      if not Assigned(Result) and {$ifndef use_dynlibs} Assigned(LibEGL) {$else} (LibEGL <> 0){$endif}
        then begin
          Result:= GetEGLProcAddress(ProcName);
          if Assigned(Result) then Exit; //because the function above already logged it
        end;
      if Assigned(Result) and Mother^.Debug.Verbose then begin
        exename:= GetExeFileName(Result);
        if (LibGlName <> exename) and (LibGlName <> '')
          then AddLog('    %0() at %1 in %2 (wrapper %3)', [ProcName, Result, exename, LibGlName])
          else AddLog('    %0() at %1 in %2', [ProcName, Result, exename]);
      end;
    end;
  end;

  function GetEGLProcAddress(ProcName: ansistring): Pointer;
  var
    baseaddr: pointer;
    exename: TFileNameString;
  begin
    {$ifndef use_dynlibs}
      Result:=dlsym(LibEGL, PAnsiChar(ProcName));
    {$else}
      Result:=GetProcAddress(LibEGL, PAnsiChar(ProcName));
    {$endif}
      if not Assigned(Result) and Assigned(eglGetProcAddress)
         then Result:= eglGetProcAddress(PAnsiChar(ProcName));

    if Mother^.Debug.Verbose then begin
      if Assigned(Result) then begin
        exename:= GetExeFileName(Result);
        if (LibEGLName <> exename) and (LibEGLName <> '')
          then AddLog('    %0() at %1 in %2 (wrapper %3)', [ProcName, Result, exename, LibEGLName])
          else AddLog('    %0() at %1 in %2', [ProcName, Result, exename]);
      end else begin
        AddLog('    %0() NOT FOUND in %1', [ProcName, LibEGLName]);
      end;
    end;
  end;

  function GetGLProcAddressIgnoreMissing(ProcName: ansistring): Pointer;
  var
    baseaddr: pointer;
    exename: TFileNameString;
  begin
    if Assigned(Framework) then Result:= Framework.GetOpenGLProcAddress(ProcName);
    if Assigned(Result) and Mother^.Debug.Verbose
      then AddLog('  %0() at %1 in %2', [ProcName, Result, GetExeFileName(Result)]);

    if not Assigned(result) then begin
      {$ifndef use_dynlibs}
        if Assigned(LibGL)
          then Result:=dlsym(LibGL, PAnsiChar(ProcName));
      {$else}
        if LibGL <> 0
          then Result:=GetProcAddress(LibGL, PAnsiChar(ProcName));
      {$endif}
        if not Assigned(Result) and Assigned(eglGetProcAddress)
           then Result:= eglGetProcAddress(PAnsiChar(ProcName));

      if Assigned(Result) then begin
        if Mother^.Debug.Verbose then begin
          exename:= GetExeFileName(Result);
          if (LibGLName <> exename) and (LibGLName <> '')
            then AddLog('    %0() at %1 in %2 (wrapper %3)', [ProcName, Result, exename, LibGLName])
            else AddLog('    %0() at %1 in %2', [ProcName, Result, exename]);
        end;
      end else begin
        AddLog('    %0() NOT FOUND in %1', [ProcName, LibGLName]);
      end;
    end;
  end;

  Procedure CloseOpenGL;
  begin
    case Mother^.GAPI.Mode of
      gapi_DeprecatedImmediateGL,
      gapi_GL21: begin
       {$ifndef use_dynlibs}
         If Assigned(LibGL) then dlClose(LibGL);
       {$else}
         If LibGL <> 0 then FreeLibrary(LibGL);
       {$endif}
      end;
      gapi_GLES2: begin
       {$ifndef use_dynlibs}
         If Assigned(LibGL) then dlClose(LibGL);
         If Assigned(LibEGL) then dlClose(LibEGL);
       {$else}
         If LibGL <> 0 then FreeLibrary(LibGL);
         If LibEGL <> 0 then FreeLibrary(LibEGL);
       {$endif}
      end;
      else
        DieUnsupportedGLMode;
    end;
  end;


  procedure CheckGLVersion;
  var
    i: integer;
    report: TIniFile;
    name: UnicodeString;
  begin
    with Mother^ do begin
      GAPI.Renderer:= AnsiString(PAnsiChar(glGetString(GL_RENDERER)));
      GAPI.VersionString:= AnsiString(PAnsiChar(glGetString(GL_VERSION)));
      GAPI.Vendor:= AnsiString(PAnsiChar(glGetString(GL_VENDOR)));
      GAPI.Summary:= GAPI.VersionString + ' | ' + GAPI.Renderer + ' | ' + GAPI.Vendor;
      GAPI.RendererAlias:= SanateStringForFileName(GAPI.Vendor + '_' + GAPI.Renderer + '_opengl' + GAPI.VersionString);
      for i:=1 to length(GAPI.Renderer) do
        if GAPI.Renderer[i] in ['/', '\', '|', ',', '.'] then begin
          GAPI.Renderer:=Copy(GAPI.Renderer, 1, i - 1);
          break;
        end;
//      cge.RenderName:= GAPI.Renderer;
      case GAPI.Mode of
        gapi_GL21: name:= RuEn('OpenGL 2.1','OpenGL 2.1');
        gapi_DeprecatedImmediateGL: name:= RuEn('Древний, как говно мамонта, OpenGL Immediate Mode','Obsolete crap of OpenGL Immediate Mode');
        gapi_GLES2: name:= RuEn('OpenGL ES 2','OpenGL ES 2');
      else
        name:= GetEnumName(typeinfo(TGAPIModeEnum), Ord(GAPI.Mode));
      end;
      AddLog(MI_LOG_OGLINITIALIZED, [name,
                                     GAPI.Renderer,
                                     GAPI.Vendor,
                                     GAPI.VersionString]);
      GetOpenGLVersion(GAPI.VersionHi, GAPI.VersionMed);
      GAPI.Extensions:= PAnsiChar(glGetString(GL_EXTENSIONS));
      GAPI.NpotSupported:= Pos('GL_ARB_texture_non_power_of_two', GAPI.Extensions) > 0;
      glGetIntegerv(GL_MAX_TEXTURE_SIZE, @GAPI.TextureMax);
      {$ifndef glesonly}
        glGetIntegerv(GL_MAX_3D_TEXTURE_SIZE, @GAPI._3dTextureMax);
      {$endif}
      glGetIntegerv(GL_MAX_RENDERBUFFER_SIZE, @GAPI.FboMax);
      {$ifndef glesonly}
        glGetIntegerv(GL_MAX_TEXTURE_COORDS, @GAPI.MaxTextureCoords);
      {$endif}
      glGetIntegerv(GL_MAX_VERTEX_TEXTURE_IMAGE_UNITS, @GAPI.MaxVShTextureUnits);
      glGetIntegerv(GL_MAX_TEXTURE_IMAGE_UNITS, @GAPI.MaxTextureUnits);
      glGetFloatv(GL_MAX_TEXTURE_MAX_ANISOTROPY_EXT, @GAPI.MaxAnisotropy);

      if State.HomePath <> '' then begin
        //if the program starts normally, write a profile file for reporting purposes.
        //*** Implement sending it to the master server!
        if Debug.Verbose then AddLog('Saving opengl renderer profile');

        ValidateWritePath(State.HomePath + LogsDir);
        report:= TIniFile.Create(State.HomePath + LogsDir + 'opengl_'
          + SanateStringForFileName(
            AnsiString(PAnsiChar(glGetString(GL_RENDERER))) + '_' + GAPI.VersionString
          )
          + '.ini');
        report.WriteString('opengl', 'version', GAPI.VersionString);

        report.WriteInteger('opengl', 'max_texture_size', GAPI.TextureMax);
        report.WriteInteger('opengl', 'max_3d_texture_size', GAPI._3dTextureMax);
        report.WriteInteger('opengl', 'max_FBO_size', GAPI.FboMax);
        report.WriteBool('opengl', 'non-power-of-2_textures_supported', GAPI.NpotSupported);
        report.WriteInteger('opengl', 'fprg_max_texture_coords', GAPI.MaxTextureCoords);
        report.WriteInteger('opengl', 'fprg_max_texture_units', GAPI.MaxTextureUnits);
        report.WriteInteger('opengl', 'vprg_max_texture_units', GAPI.MaxVShTextureUnits);
        report.WriteInteger('opengl', 'max_anisotropy', round(GAPI.MaxAnisotropy));

        report.WriteString('opengl_extensions', 'string', GAPI.Extensions);

        report.UpdateFile;
        report.Free;
      end;


      if (GAPI.VersionHi < OGL_REQ_HI)
      or ((GAPI.VersionHi = OGL_REQ_HI) and (GAPI.VersionMed < OGL_REQ_MED)) then begin
       {$ifdef windows}
        if lowercase (GAPI.Renderer) = 'gdi generic'
          then
            AddStopMessage(MI_MICROSOFT_GDI,[])
          else
       {$endif}
            AddStopMessage(MI_OPENGL_VERSION_TOO_LOW,
                            [OGL_REQ_HI,
                             OGL_REQ_MED,
                             MessageContainer[MI_HINT_OPENGLTOOLOW],
                             GAPI.VersionString,
                             GAPI.Renderer,
                             GAPI.Vendor]);
      end;
    end;
  end;

  procedure CheckGLESVersion(majo, mino: EGLint);
  var
    i: integer;
    report: TIniFile;
  begin
    with Mother^ do begin
      GAPI.VersionHi:= majo;
      GAPI.VersionMed:= mino;
      GAPI.Renderer:= AnsiString(PAnsiChar(glGetString(GL_RENDERER)));
      GAPI.VersionString:= AnsiString(PAnsiChar(glGetString(GL_VERSION)));
      GAPI.Vendor:= AnsiString(PAnsiChar(glGetString(GL_VENDOR)));
      GAPI.Summary:= GAPI.VersionString + ' | ' + GAPI.Renderer + ' | ' + GAPI.Vendor;
      GAPI.RendererAlias:= SanateStringForFileName(GAPI.Vendor + '_' + GAPI.Renderer + '_opengl' + GAPI.VersionString);
      for i:=1 to length(GAPI.Renderer) do
        if GAPI.Renderer[i] in ['/', '\', '|', ',', '.'] then begin
          GAPI.Renderer:=Copy(GAPI.Renderer, 1, i - 1);
          break;
        end;

      AddLog(MI_LOG_OGLINITIALIZED, [GAPI.VersionString,
                                     GAPI.Renderer,
                                     GAPI.Vendor]);
    end;
  end;

  Procedure InitOpenGL;
  var
    cw: word;
    i: integer;
    gl_name, egl_name,
    gl_name0, egl_name0: TFileNameString;
    function LoadGLLibrary: boolean;
    begin
      if Mother^.Debug.Verbose then AddLog('  Loading %0...', [gl_name]);
     {$ifdef use_dynlibs}
      LibGL:= {$ifdef che_unicode}
                LoadLibraryW
              {$else}
                LoadLibrary
              {$endif}
                  (PFileNameChar(gl_name)); //Windows and MacOS X
      Result:= LibGL <> 0;
     {$else}
      LibGL := dlopen(PAnsiChar(WideToAnsi(FileNameToUnicode(gl_name))), RTLD_NOW); //Linux
      Result:= Assigned(LibGL);
     {$endif}
     if result then LibGlName:= GetDLLFileName(LibGL);
     if Mother^.Debug.Verbose then
       if not result then AddLogComment('FAILED.')
                     else AddLogComment('Ok, %0', [LibGlName]);

    end;
    function LoadEGLLibrary: boolean;
    begin
      if Mother^.Debug.Verbose then AddLog('  Loading %0...', [egl_name]);
     {$ifdef use_dynlibs}
      LibEGL:= {$ifdef che_unicode}
                LoadLibraryW
              {$else}
                LoadLibrary
              {$endif}
                  (PFileNameChar(egl_name));
      Result:= LibEGL <> 0;
     {$else}
      LibEGL := dlopen(PAnsiChar(WideToAnsi(FileNameToUnicode(egl_name))), RTLD_NOW); //Linux
      Result:= Assigned(LibEGL);
     {$endif}
     if result then LibEGLName:= GetDLLFileName(LibEGL);
     if Mother^.Debug.Verbose then
       if not result then AddLogComment('FAILED.')
                     else AddLogComment('Ok, %0', [LibEGLName]);

    end;
  begin
   if Assigned(Framework) and Assigned(Framework.GetOpenGLProcAddress('glTexImage2D')) then begin
     if Mother^.Debug.Verbose then AddLog('  Skipping the GAPI loading stage, gettng the procedure addresses from the framework instead...');
     InitGlProcAddresses;
     Exit;
   end;


   {$ifdef cgekernel}
    if Config.Bool[Mother^.State.OstID, 'use_gles_not_gl']
      then Mother^.GAPI.Mode:= gapi_GLES2
      else ;//Mother^.GAPI.Mode:= gapi_GL21;
   {$endif}

    case Mother^.GAPI.Mode of
      {$ifndef glesonly}
      gapi_DeprecatedImmediateGL,
      gapi_GL21: begin
        if Mother^.Debug.Enabled then AddLog('Init OpenGL...');

        gl_name0:= UnicodeToFilename(Config.Str['opengl', 'GL_' + SystemSuffix + '_0']);
        i:=0;
        repeat
          gl_name:= UnicodeToFilename(Config.Str['opengl', 'GL_' + SystemSuffix + '_' + IntToStr(i)]);
          if gl_name='' then Die(MI_ERROR_OPENGLDLLNOTFOUND, [
            FileNameToUnicode(gl_name0)]);
          if LoadGlLibrary() then break;
          inc(i);
        until false;

        //Disable FPU exceptions (it will ignore divisions by zero and such)
        //Without this, the program crashes in most unexpected places!
        {$ifdef cpui386}
        cw:= $133F;
        asm
          fldcw [cw]
        end;
        SetExceptionMask([exInvalidOp, exDenormalized, exZeroDivide,exOverflow, exUnderflow, exPrecision]);
        {$endif}


       {$ifdef unix}
         {$ifdef darwin}
           //AGL code here...? Probably no need to, as everything should be hard linked.
         {$else}
          //1.1
          pointer(glXQueryExtensionsString):= GetGLProcAddress('glXQueryExtensionsString','');
          pointer(glXChooseVisual):= GetGLProcAddress('glXChooseVisual','');
          pointer(glXGetConfig):= GetGLProcAddress('glXGetConfig','');
          pointer(glXCreateContext):= GetGLProcAddress('glXCreateContext','');
          pointer(glXDestroyContext):= GetGLProcAddress('glXDestroyContext','');
          pointer(glXMakeCurrent):=  GetGLProcAddress('glXMakeCurrent','');
          pointer(glXSwapBuffers):= GetGLProcAddress('glXSwapBuffers','');
          pointer(glXSwapIntervalSGI):= _GetGLProcAddress('glXSwapIntervalSGI');

          //1.3
          //pointer(glXGetVisualFromFBConfig):= GetGLProcAddress('glXGetVisualFromFBConfig', '');
          //pointer(glXChooseFBConfig):= GetGLProcAddress('glXChooseFBConfig', '');

         {$endif}
       {$else}
          LoadFiveQuestionableWglFunctions;

          pointer(wglGetProcAddress) := GetGLProcAddress ('wglGetProcAddress','');
          pointer(wglCreateContext):= GetGLProcAddress ('wglCreateContext','');
          pointer(wglDeleteContext):= GetGLProcAddress ('wglDeleteContext','');
          pointer(wglMakeCurrent):= GetGLProcAddress ('wglMakeCurrent','');
       {$endif}

        if Mother^.Debug.Verbose then AddLog('  Loading  basic OpenGL 1.1 functions...');
        InitGlProcAddresses;
      end;
      {$endif}
      gapi_GLES2: begin
        if Mother^.Debug.Enabled then AddLog('Init OpenGL ES...');

        egl_name0:= UnicodeToFilename(Config.Str['opengl', 'EGL_' + SystemSuffix + '_0']);
        i:=0;
        repeat
          egl_name:= UnicodeToFilename(Config.Str['opengl', 'EGL_' + SystemSuffix + '_' + IntToStr(i)]);
          if egl_name='' then Die(MI_ERROR_OPENGLDLLNOTFOUND, [
            FileNameToUnicode(egl_name0)]);
          if LoadEGLLibrary() then break;
          inc(i);
        until false;

        gl_name0:= UnicodeToFilename(Config.Str['opengl', 'GLES_' + SystemSuffix + '_0']);
        i:=0;
        repeat
          gl_name:= UnicodeToFilename(Config.Str['opengl', 'GLES_' + SystemSuffix + '_' + IntToStr(i)]);
          if gl_name='' then Die(MI_ERROR_OPENGLDLLNOTFOUND, [
            FileNameToUnicode(gl_name0)]);
          if LoadGlLibrary() then break;
          inc(i);
        until false;

        if Mother^.Debug.Verbose then AddLog('  Loading the procedure addresses from the EGL DLL ...');
        InitEGLProcAddresses;
        if Mother^.Debug.Verbose then AddLog('  Loading the procedure addresses from the GL ES DLL ...');
        InitGLESv2ProcAddresses;

      end;
      else
        DieUnsupportedGLMode;
    end;
  end;


  procedure GetOpenGLVersion(var hi, med: integer);
  var
    s: String[3];
    p: PAnsiChar;
  begin
    try
      p:=glGetString(GL_VERSION);
      s:='';
      while p^ in ['0'..'9'] do begin
        s:=s + p^;
        inc (p);
      end;
      hi:=StrToInt(s);
      inc(p);
      s:='';
      while p^ in ['0'..'9'] do begin
        s:=s + p^;
        inc (p);
      end;
      med:=StrToInt(s);
    except
      hi:=OGL_REQ_HI;
      med:=OGL_REQ_MED;
    end;
  end;

  procedure SwitchToFixedPipeline;
  begin
    case Mother^.GAPI.Mode of
      gapi_GL21, gapi_DeprecatedImmediateGL : begin
        if Assigned(glBindFramebuffer)
          then glBindFramebuffer(GL_FRAMEBUFFER, 0); //switch from FBO to the window framebuffer
        if Assigned(glUseProgram)
          then glUseProgram(0); //shut down the shaders
      end;
      gapi_GLES2: begin
        if Assigned(glBindFramebuffer)
          then glBindFramebuffer(GL_FRAMEBUFFER, 0); //switch from FBO to the window framebuffer
         if Assigned(glUseProgram)
          then glUseProgram(eFixedPipelineEmulationProgram);
      end;
      else
        DieUnsupportedGLMode;
    end;
  end;


  procedure CheckGLSLError(Handle: GLhandleARB; Param: GLenum; Explanation: UnicodeString;
                         source: RawByteString = ''; isShader: boolean = false);
  var
    slen, res: GLint;
    s: RawByteString;
  begin
    if (Assigned(glGetObjectParameterivARB)) then begin
      glGetObjectParameterivARB(Handle, Param, @res);
      if (res <> GL_TRUE) or true then begin //Mother^.Debug.Verbose then begin
        SetLength(s, 100000);
        glGetInfoLogARB(Handle, Length(s), @slen, PGLcharARB(@s[1]));
        SetLength(s, slen);
      end;
    end
      else
        if isShader then begin
          glGetShaderiv(Handle, Param, @res);
          if (res <> GL_TRUE) or true then begin //Mother^.Debug.Verbose then begin
            SetLength(s, 100000);
            glGetShaderInfoLog(Handle, Length(s), @slen, PGLchar(@s[1]));
            SetLength(s, slen);
          end;
        end
        else begin
          glGetProgramiv(Handle, Param, @res);
          if (res <> GL_TRUE) or true then begin //Mother^.Debug.Verbose then begin
            SetLength(s, 100000);
            glGetProgramInfoLog(Handle, Length(s), @slen, PGLchar(@s[1]));
            SetLength(s, slen);
          end;
        end;


    if res <> GL_TRUE then begin
      Mother^.State.Fatality:= Yes; //cuz opengl just hangs if we try to continue!
      if source <> ''
        then Die(RuEn(
          'Ошибка %0: %1'#10#13'Исходный код шейдера: %2',
          'Failed %0: %1'#10#13'Shader source: %2'), [Explanation, s, source])
        else Die(RuEn(
          'Ошибка %0: %1',
          'Failed %0: %1'), [Explanation, s]);
    end;

    if Mother^.Debug.Verbose then
      if source <> ''
        then AddLog('    Log of %0 : %1 ; Shader source: %2',[Explanation, s, source])
        else AddLog('    Log of %0 : %1',[Explanation, s]);
  end;


  procedure  CreateESFFPEmulationProgram;
  var
    F, V: GLuint;
    src: RawByteString;
  begin
//Die('Себе на заметку: ДОДЕЛАТЬ CreateESFFPEmulationProgram()');
    if Mother^.Debug.Verbose then AddLog('  Initializing FFP emulation shader for GLES 2...');

    eFixedPipelineEmulationProgram:= glCreateProgram();
     CheckGlError;

    src:= FileGetContents('*I/modules/chentrah/glsl/ES_2_FFP_emul.vert') + #0;
    V:= glCreateShader(GL_VERTEX_SHADER_ARB);
     CheckGlError;
    glShaderSource(V,  1, @(pointer(src)), NIL);//@len);
     CheckGlError;
    glCompileShader(V);
     CheckGLSLError(V, GL_OBJECT_COMPILE_STATUS_ARB, RuEn(
      'компиляции вершинного шейдера',
      'compiling vertex shader'), src, true);

    src:= FileGetContents('*I/modules/chentrah/glsl/ES_2_FFP_emul.frag') + #0;
    F:= glCreateShader(GL_FRAGMENT_SHADER_ARB);
     CheckGlError;
    glShaderSource(F,  1, @(pointer(src)), NIL);//@len);
     CheckGlError;
    glCompileShader(F);
     CheckGLSLError(F, GL_OBJECT_COMPILE_STATUS_ARB, RuEn(
      'компиляции фрагментного шейдера',
      'compiling fragment shader'), src, true);


    glAttachShader(eFixedPipelineEmulationProgram, V);
     CheckGlError;

    glAttachShader(eFixedPipelineEmulationProgram, F);
     CheckGlError;

    glLinkProgram(eFixedPipelineEmulationProgram);
     CheckGLSLError(eFixedPipelineEmulationProgram, GL_OBJECT_LINK_STATUS_ARB, RuEn(
      'сборки шейдеров в программу GLSL',
      'linking GLSL program'));
  end;




