{
Procedure ClearWarningQueue;
  begin
    WarningQueue.Clear;
  end;
  }
{
  Function IsWarningQueueEmpty: boolean;
  var
    pes: PMotherSehState;
  begin
    pes:= GetExceptionState();
    Result:= (Length(pes^.Warnings) = 0);
  end;

}

  Function PullWarnings: UnicodeString;
  var
    i: integer;
    r: UnicodeString;
  begin
    CollateWarningsFromAllThreads;
    r:='';
    For i:= High(Mother^.ExceptionState.Warnings) downto 0 do begin
      r:= r + Mother^.ExceptionState.Warnings[i];
      if i > 0 then r:=r + #10#13#10#13;
    end;
    SetLength(Mother^.ExceptionState.Warnings, 0);
    Result:=r;
  end;

  var
    CachedLog: array of UnicodeString;




  procedure WriteLogToFile(u: UnicodeString; EraseLF: boolean);
  var
    a: AnsiString;
    f: file;
    i: integer;
    var cl: array of UnicodeString;
  begin
    if u = '' then Exit;
   try
    if DontWriteALog or not Mother^.Initialized then begin
      if EraseLF and (Length(CachedLog) > 0)
        then CachedLog[High(CachedLog)]+= u
        else begin
          SetLength(CachedLog, Length(CachedLog) + 1);
          CachedLog[High(CachedLog)]:= u;
        end;
    end else begin

      //The second cached block (these are put backwards, guess why)
      if Length(CachedLog) > 0 then begin
        u:= #10#13
          + PervertedFormat(RuEn(
            'Выполнена запись кешированных %0 строк журнала в файл при первой возможности.',
            'Wrote %0 cached log lines to file as soon as it was possible.'
          ), [Length(CachedLog) + Length(CachedVerboseLog)])
          + #10#13#10#13
          + u;
        cl:= CachedLog;
        SetLength(CachedLog, 0);
        for i:= 0 to High(cl) do begin
          WriteLogToFile(cl[i], false);
          cl[i]:= '';
        end;
      end;

      //the first cached block
      if Mother^.Initialized and Mother^.Debug.Verbose and (Length(CachedVerboseLog) > 0) then begin
        cl:= CachedVerboseLog;
        CachedVerboseLog:= nil;
        for i:= 0 to High(cl) do begin
          WriteLogToFile(cl[i], false);
          cl[i]:= '';
        end;
        SetLength(cl, 0);
      end;

      if not Assigned(LogStream) then begin
        if Mother^.State.RestartRequested then begin
          LogStream:= TFileStream.Create(Mother^.State.LogFile, fmOpenReadWrite + fmShareDenyWrite);
          LogStream.Position:= LogStream.Size - 1;
        end
        else
          LogStream:= TFileStream.Create(Mother^.State.LogFile, fmCreate + fmShareDenyWrite);
        u:= #$FEFF + u;//unicode text signature (zero-width space)
      end;

{      AssignFile(f, Mother^.State.LogFile);
      if not fileexists(Mother^.State.LogFile)
        then begin
          Rewrite(f, 1);
          u:=#$FEFF + u;//unicode text signature (zero-width space)
        end
        else begin
          Reset(f, 1);
          if EraseLF then Seek(f, FileSize(f) - 2)
                     else Seek(f, FileSize(f) - 1)
        end;
}
      a:= Utf8Encode(u);
{
      BlockWrite(f, a[1], length(a));
      CloseFile(f);
}
      LogStream.Write(a[1], length(a));
    end;
   except
   end;
  end;

  var
    LastFrameWhenLogWasUsed: int64 = 0;

  Procedure AddLog(U: UnicodeString; Param: array of const);
  var uf: UnicodeString = '';
  begin
    EnterCriticalsectionWv(Mother^.CriticalSection);
      if LastFrameWhenLogWasUsed <> CurrentFrame then begin
        LastFrameWhenLogWasUsed:= CurrentFrame;
        uf:= #10#13 + PervertedFormat(RuEn(
          '[кадр №%0, %1]',
          '[frame #%0, %1]'),[CurrentFrame, FormatDateTime(
          'hh:nn:ss.zzz, ddd, dd.mm.yyyy', CurrentFrameDate)]) + #10#13;
      end;
      U:= PervertedFormat(U, Param);
     {$ifdef unix}
      WriteLn;
      if uf <> '' then WriteLn(Trim(Utf8Encode(StrReplace(uf, UnicodeString(#10#13), UnicodeString(#13)))));
      Write(TrimRight(Utf8Encode(StrReplace(U, UnicodeString(#10#13), UnicodeString(#13)))));
      {$else}
        {$ifdef debug_build}
          Writeln;
          if uf <> '' then WriteLn(Trim(WideToConsole(uf)));
          Write(TrimRight(WideToConsole(U)));
        {$endif}
     {$endif}

      if not DontWriteALog then begin //a special case. We don't want to ruin
                                    //the log of an already running instance.
        if Assigned(Console) then begin
          if uf <> '' then Console.Add(WideToCp1251(uf));
          Console.Add(WideToCp1251(U + #10#13));

    //       if Mother^.Module.CurrentlyExecutingInd = NOT_A_MODULE then _cgeHeartBeat;
        end;
        WriteLogToFile(#10#13 + uf + u, No);
      end;
    LeaveCriticalSection(Mother^.CriticalSection);
  end;

  Procedure AddLogComment(U: UnicodeString; Param: array of const);
  begin
    EnterCriticalsectionWv(Mother^.CriticalSection);
    U:= PervertedFormat(U, Param);
      {$ifdef unix}
      Write(TrimRight(Utf8Encode(StrReplace(U, UnicodeString(#10#13), UnicodeString(#13)))));
      {$else}
        {$ifdef debug_build}
          Write(TrimRight(WideToConsole(U)));
        {$endif}
      {$endif}

      if Assigned(Console) then begin
        Console.AddComment(WideToCp1251(U + #10#13));
  {      if Console.Count > 0
          then Console[Console.Count - 1]
            := Copy(Console[Console.Count - 1], 1, Length(Console[Console.Count - 1]) - 2) + WideToAnsi(U)
          else Console.Add(WideToAnsi(U));}
  //       if Mother^.Module.CurrentlyExecutingInd = NOT_A_MODULE then _cgeHeartBeat;
      end;
      if not DontWriteALog then WriteLogToFile(u, Yes);
    LeaveCriticalSection(Mother^.CriticalSection);
  end;
{
  Procedure AddLog(S: Ansistring; Param: array of const);
  begin
    AddLog(AnsiToWide(S), Param);
  end;
 }
  Procedure AddLog(mID: TMessageID; Param: array of const);
  begin
    AddLog(MessageContainer[mID], Param);
  end;

  Procedure AddLog(mID: TMessageID);
  begin
    AddLog(MessageContainer[mID], []);
  end;
{
  Procedure AddLogComment(S: AnsiString; Param: array of const);
  begin
    AddLogComment(AnsiToWide(S), Param);
  end;
}
  Procedure AddLogComment(mID: TMessageID; Param: array of const);
  begin
    AddLogComment(MessageContainer[mID], Param);
  end;
{
  Procedure AddLogComment(S: AnsiString);
  begin
    AddLogComment(S,[]);
  end;
}
  Procedure AddLogComment(U: UnicodeString);
  begin
    AddLogComment(U,[]);
  end;


  Procedure AddLogComment(mID: TMessageID);
  begin
    AddLogComment(mID,[]);
  end;


  Procedure AddLogOK; cdecl;
  begin
    AddLogComment(MessageContainer[MI_LOG_SUCCESS]);
  end;
{
  Procedure AddLog(S: AnsiString);
  begin
    AddLog(S, []);
  end;
}
  Procedure AddLog(U: UnicodeString);
  begin
    AddLog(U, []);
  end;

  procedure VerboseLog(mID: TMessageID; Param: array of const);
  begin
    if Mother^.Debug.Verbose then AddLog(mID, param);
  end;

  Procedure VerboseLogOk;
  begin
    if Mother^.Debug.Verbose then AddLogOk;
  end;

  procedure VerboseLog(mID: TMessageID);
  begin
    if Mother^.Debug.Verbose then AddLog(mID, []);
  end;


  Procedure VerboseLog(U: UnicodeString; Param: array of const);
  begin
    if Mother^.Initialized then begin
      if Mother^.Debug.Verbose then AddLog(U, Param);
    end
    else begin
      U:= PervertedFormat(U, Param);
     {$ifdef unix}
      WriteLn;
      Write(TrimRight(Utf8Encode(StrReplace(U, UnicodeString(#10#13), UnicodeString(#13)))));
     {$else}
       {$ifdef debug_build}
         Writeln;
         Write(TrimRight(WideToConsole(U)));
       {$endif}
     {$endif}
      U:=U + #10#13;

      SetLength(CachedVerboseLog, Length(CachedVerboseLog) + 1);
      CachedVerboseLog[High(CachedVerboseLog)]:= U;
    end;
  end;

  Procedure VerboseLog(U: UnicodeString);
  begin
    VerboseLog(U, []);
  end;



  function RectToStr(r: TWindowManagerRect) : UnicodeString;
  begin
    with r do
      Result:= PervertedFormat('%0x%1(%2,%3)',[width, height, left, top]);
  end;

  procedure _LogGoDown;
  begin
    AddLog(
      RuEn('Инициировано закрытие программы:'#10#13'  %0.'
          ,'Program shutdown initiated:'#10#13'  %0.')
          ,[Mother^.State.SuddenShutdownReason]);
  end;

  procedure LogGoDown(s: UnicodeString);
  begin
    Mother^.State.SuddenShutdownReason:= s;
    _LogGoDown;
  end;

  procedure LogGoDownWM(s: UnicodeString);
  begin
    Mother^.State.SuddenShutdownReason:= PervertedFormat(RuEn(
      'окно получило оповещение %0' ,
      'the window received %0 notification'), [s]);
    _LogGoDown;
  end;

  procedure LogGoDownM(s: UnicodeString);
  begin
    Mother^.State.SuddenShutdownReason:= PervertedFormat(RuEn(
      'получено сообщение %0',
      'a system notification received, %0'), [s]);
    _LogGoDown;
  end;

