{
    This file is part of Chentrah,
    Copyright (C) 2004-2008 Anton Rzheshevski (chebmaster@mail.ru).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see http://www.gnu.org/licenses/

 **********************************************************************

   This file contains the main function of the mother module

 **********************************************************************}


var
  name, selfname: AnsiString;

  procedure SafeInitFramework; cdecl; forward;


  procedure Run();
   var
    f: file;
    i: integer;
    t: TDateTime;
    w, olderrormessage: UnicodeString;
  begin
   {$ifdef windows}
    // 2013: it causes a cascade of exceptions at the program termination in FPC 2.6.0 and 2.6.2
    //winWideStringalloc:= false; //WideString becomes incompatible with COM but works faster
   {$endif}
    InitMotherState;
    For i:=1 to ParamCount do begin
      if (lowercase (Paramstr(i))= '--restart')
         or (copy(lowercase(ParamStr(i)), 1, 9) = '--lockup-')
        then Mother^.State.RestartRequested:= Yes;

      if copy(lowercase(ParamStr(i)), 1, 9) = '--server-' then begin
        Mother^.State.IsServer:= true;
        Mother^.State.ServerModuleName:= copy(ParamStr(i), 10, length(ParamStr(i)) - 9);
      end;
    end;
//    Sleep(100);
// ******** FIRST STAGE INIT: ***************************************
    Try
      CgeStartTime:=Now();
     {$ifdef Win32}
      CgeStartingTick:= GetTickCount();
     {$endif}
      if not ThisIsAnOnlyInstance and not Mother^.State.RestartRequested then begin
        AnotherInstanceIsRunning:=Yes;
        DontWriteALog:=Yes;
      end;

      InitUnicode;
      FormatSettings.DecimalSeparator:= '.';
      InitTranslit;

      MessageContainer:=TMessageContainer.Create;
      Config:=TConfigManager.Create;

      {$ifdef unix}
      if (Length(ParamStr(0)) >=8) and (copy(ParamStr(0), 1, 4) = '/tmp')
        then AddStopMessage(MI_NO_UPX_ALLOWED, [AppNick]);
      {$endif}
      // the "stop message" is that red screen which pops up later,
      //  after all the  initialization stages, and tells you (using the
      //  fullscreen OpenGL and fancy graphics) that no, this program won't run.
      
      {$ifdef windows}
        {$ifdef cpu32}
          LoadFunctionsUnsupportedInWindows98();
        {$endif}
      {$endif}

      DetectOSType;

      Randomize;
      DetermineInstallPath;

//      WarningQueue:= TUnicodeStringContainer.Create;
      EngineObject:= TEngineObjects.Create;

      if Mother^.State.InstallPath <> '' then begin

        CreateSplashWindow; //P.S. Did it. move it to after paths are determined if I want to re-use the old strategy of many splash variations in bmp files

        Config.LoadMainFile;
        
        SetCurrentDir(Mother^.State.InstallPath);

        with Mother^ do begin
          {$ifdef debug_build}
            //see cl_init_mother.inc
          {$else}
            Debug.Enabled:= Cge.Config.Bool['debugging', 'enabled'];
            Debug.Verbose:= Debug.Enabled and Cge.Config.Bool['debugging', 'verbose'];
            Debug.ShowFrameNumbersInLog:= Debug.Verbose or Cge.Config.Bool['debugging', 'log_frame_numbers'];
          {$endif}
          Debug.StopOnOpenGlErrors:= Cge.Config.Bool['debugging', 'stop_on_opengl_errors'];
          {$ifdef nodevmodewin32}
            State.DeveloperMode:= No;
          {$else}
            {$ifdef debug_build}
              State.DeveloperMode:= true;
            {$else}
              State.DeveloperMode:= Cge.Config.Bool['main','developer_mode'];
            {$endif}
          {$endif}
          Module.MultithreadingDisabled:= Cge.Config.Bool['debugging', 'disable_module_multithreading'];
        end;
        
        DetermineHomePath;

        SetCurrentDir(Mother^.State.HomePath);

        if Mother^.State.OS = ostWine then begin
          if not ValidateWritepath(Mother^.State.HomePath) or not ValidateFileIsWriteable(Mother^.State.HomePath + ConfigFilePath)
            //This is a *stub* for that pesky Wine bug
          then begin
            AddStopMessage(MI_WINE_GLITCHES_HOME_PATH_NOT_WRITEABLE,
              [AnsiToWide(Mother^.State.HomePath), CgeStringEn]);
            Mother^.State.HomePath:= '';
          end;
        end;

        if Mother^.State.HomePath <> '' then begin
          Mother^.State.CacheDir:= Mother^.State.HomePath + LOcalStoreStr + PathSlash;

          DisplayLastWill; //if any. Causes error dialog and Halt()

          Config.LoadUserFile;
        end;
        AppNick:=MyAppNick;//UpperCase(ChangeFileExt(ExtractFileName(ParamStr(0)),''));

        ModulesBasePath:= Config.Path['Modules','BasePath'];
        Mother^.State.SessionDir:=CgePath(Mother^.State.HomePath + 'sessions' + DS);

        MessageContainer.LoadFiles;
      end
      else
        DontWriteALog:=True; //there's nowhere to write it

      Mother^.State.LogFile:= Mother^.State.HomePath + 'LOG';
        if Mother^.State.IsServer then Mother^.State.LogFile+= '_server_' + Mother^.State.ServerModuleName;
        Mother^.State.LogFile+= '.txt';

      if not Mother^.State.RestartRequested and not DontWriteALog and FileExists(Mother^.State.LogFile)
      then try
        //AssignFile(f, Mother^.State.LogFile);
        //Erase(f);
      except
        DontWriteALog:= Yes;
        AddStopMessage(MI_CANNOT_DELETE_OLD_LOG,
          [Mother^.State.LogFile,
          (ExceptObject as Exception).Message, Mother^.State.UserNameUCS16]);
      end;

      Mother^.Initialized:= true;
      if not Mother^.Debug.Verbose then SetLength(CachedVerboseLog, 0);


      Console:= TConsole.Create;


//byte(nil^):=0;

    Except
      //raise exception and die...
      KillSplashWindow;
      PreInitDie(PIYInitStageCrash, [(ExceptObject as Exception).Message, CallStack(0, '  ')]);
    End;


// ******** SECOND STAGE INIT: ****************************
//    repeat
      Try
        SecondStageInitTime:=Now();
        if Mother^.State.RestartRequested then begin
         // while not ThisIsAnOnlyInstance do Sleep(500);
          DontWriteALog:=No;
        end;
        if AnotherInstanceIsRunning
          then Die(MI_ERROR_MULTIPLE_INSTANCES_NOT_ALLOWED,[MessageContainer[MI_CGE_TITLE]]);
(*
        {$ifdef nodevmodewin32}
        //I long since created a SEH hack to work around that
        if Config.Bool['main','developer_mode'] then AddStopMessage(RuEn(
          'Данная версия %0 (v%1),'#10#13+
          '  из-за изъяна в RTL Free Pascal 2.2.0,'#10#13+
          '  позволяет использовать режим разработчика'#10#13+
          '  только под Linux.',
          'This version of %0 (v%1),'#10#13+
          '  due to the bugs in the FreePascal 2.2.0 RTL,'#10#13+
          '  supports the developer mode only for Linux.'
          ), [
           MessageContainer[MI_CGE_TITLE],VersionToStr(VersionMajor, VersionMinor, BuildNumber)]);
        {$endif}

        {$ifdef buildmein}
        if Mother^.State.DeveloperMode then begin
          {$ifdef win32}
            GiveWarning('no_dev_mode_for_you-' + SystemSuffix,  MI_WARN_WARNING, MI_WARN_DEVELOPER_MODE_BUILTIN,
              [MessageContainer[MI_CGE_TITLE],VersionToStr(VersionMajor, VersionMinor, BuildNumber), MyInternalModuleName]);
          {$else}
            GiveWarning('no_dev_mode_for_you-' + SystemSuffix,  MI_WARN_WARNING, MI_WARN_DEVELOPER_MODE_WIN32_ONLY,
              [Mother^.State.OstID, {$ifdef windows}'Win64'{$else}{$ifdef darwin}'Mac State.OS X'{$else}'Linux'{$endif}{$endif}]);
          {$endif}
          Mother^.State.DeveloperMode:= false;
        end;
        {$endif}
*)
        if Mother^.State.DeveloperMode and not Mother^.State.ProgramDirIsWriteable
          then AddStopMessage(MI_ERROR_DEVELOPER_MODE_NO_CAN_DO,[
                 AnsiToWide(Mother^.State.InstallPath) + MyAppName + '.ini', AnsiToWide(Mother^.State.InstallPath)]);

        if Mother^.State.RestartRequested
          then begin
            AddLog(RuEn('ПЕРЕЗАПУСК...','RESTARTING...'));
            Mother^.State.RestartRequested:=No;
            NowRestarting:=Yes;
          end
          else begin
            if Mother^.Debug.Enabled then AddLog('First Stage Init took %0 seconds.',[DateTimeToP(Now() - CgeStartTime, 1)]);
          end;
        //AddLog(MessageContainer[MI_LOG_STARTED], [StrCurrentTime, MessageContainer.LID]);

        if Mother^.Debug.Verbose then begin
          AddLog('Install path: %0',[FileNameToUnicode(Mother^.State.InstallPath)]);
{          AddLog('  Default install path: %0',[AnsiToWide(StdLocation)]);
          AddLog('  Alternate install path: %0',[AnsiToWide(AltLocation)]); }
          AddLog('Home path: %0',[FileNameToUnicode(Mother^.State.HomePath)]);
          AddLog('  Default home path: %0',[FileNameToUnicode(Mother^.State.StdHomepath)]);
          AddLog('  cache path: %0',[FileNameToUnicode(Mother^.State.CacheDir)]);
          AddLog('  sessions path: %0',[FileNameToUnicode(Mother^.State.SessionDir)]);
        end;

        ValidateWritePath(Mother^.State.CacheDir);
        ValidateWritePath(Mother^.State.SessionDir);

        {$ifdef windows}SetLastError(0);{$endif}

        InitCTimer;

        SetCurrentDir(Mother^.State.HomePath);

        w:= PervertedFormat(MessageContainer[MI_VERSION_INFO], [MessageContainer[MI_CGE_TITLE],
            VersionToStr(VersionMajor, VersionMinor, BuildNumber), CompileDate, CompileTime, CompilerVersion, CompileTarget]);
        if Mother^.State.DeveloperMode
          then w+= #10#13 + RuEn('  (режим разработчика)', '  (developer mode on)');
        AddLog(w);

        LogOSVersion;

        if Mother^.Debug.Enabled then AddLog('Initializing the lockup guard...');

        EngineObject.AddObj(TLockupGuard.Create);  //менеджер автохаракири при зависании

        if not system.IsMultiThread then Die('Pascal error, IsMultiThread = false');

       {$ifdef windows}
        if Mother^.Debug.Verbose then AddLog('Error Mode was %0h', [IntToHex(clGetErrorMode(), 1)]);
        clSetErrorMode(SEM_NOGPFAULTERRORBOX or SEM_FAILCRITICALERRORS or SEM_NOOPENFILEERRORBOX);
        if Mother^.Debug.Verbose then AddLog('Error Mode now %0h', [IntToHex(clGetErrorMode(), 1)]);
        {$ifndef buildmein}
         if Mother^.Debug.Enabled then AddLog('Activating the exception handling hack.');
         InitSehHack;
        {$endif}
       {$endif}


       {$ifdef unix}
        SetSigtermHandler;
       {$endif}

       {$ifdef cpu32}
        if Mother^.Debug.Verbose then AddLog(RuEn('Загрузка отладочной информации...', 'Loading the self-debugging info...'));
        Mother^.State.DebugInfoPath:= Mother^.State.InstallPath + 'modules' + PathSlash + 'chentrah' + PathSlash + 'bin' + PathSlash;
        SetLength(LineInfoPaths, 1);
        LineInfoPaths[0]:= Mother^.State.DebugInfoPath;
        {if Mother^.State.DeveloperMode or Mother^.State.DebugMode
          then InitLineInfo(nil); //preload the self-debug info}

        if not InitLineInfo(@Run, true) then AddStopMessage(MI_DEBUG_INFO_NOT_FOUND, [_LineInfoError]);
       {$endif}

        GetHotKeysFromConfig;

//        Containers:=TStringList.Create;
//        TexContainer:=TStringList.Create;

        if Mother^.State.PhysMemory < OsMinMem[Mother^.State.OS]
          then AddStopMessage(MI_NOT_ENOUGH_PHYSICAL_MEMORY, [
            OSDisplayedName[Mother^.State.OS], OsMinMemRep[Mother^.State.OS],
            Mother^.State.PhysMemory, CgeString]);

//        if Mother^.Debug.Enabled then AddLog('Initializing resource manager...');
//        InitResourceManager;

        //EngineObject.AddObj(TMd5List.Create);
        //Md5List:= EngineObject.Last as TMd5List;

        if Mother^.Debug.Enabled then AddLog('Initializing Vampyre Imaging Library...');
        StartVampyre;

        Mother^.LockupGuard.Checkup;

        if Mother^.State.IsServer then begin
          AddLog(RuEn(
            'Режим сервера для модуля "%0"',
            'Running as a server for module "%0"'
            ),[Mother^.State.ServerModuleName]);
          Mother^.State.FrameworkName:= 'server';
        end
        else begin
          Mother^.State.FrameworkName:= Config.Str['framework_' + SystemSuffix, 'used'];
          if Mother^.State.FrameworkName = '' then begin
            Mother^.State.FrameworkName:= Config.Str[Mother^.State.OstID, 'default_framework'];
            if Mother^.Debug.Verbose then AddLog(
              'No framework name found in the config, using default "%0"'
              , [Mother^.State.FrameworkName]);
          end
          else
            if Mother^.Debug.Verbose then AddLog(
              'Framework to use, as of config: "%0"', [Mother^.State.FrameworkName]);
        end;

        olderrormessage:= StringReplace(
          Utf8Decode(Config.Str['session', 'LastWill']),
          '{CR/LF}', #10#13, [rfReplaceAll]);
        if olderrormessage <> '' then begin
          AddWarning(RuEn(
            'Прошлая сессия была завершена нештатно',
            'The previous session was not closed correctly'
            ), olderrormessage, []);
          Config.Str['session', 'LastWill']:= '';
        end;

        if Mother^.Debug.Enabled then AddLog(RuEn(
          'Инициализация фреймворка "%0"...',
          'Initializing the framework "%0"...'), [Mother^.State.FrameworkName]);
        Mother^.LockupGuard.Guard(@SafeInitFramework, luc_FrameworkStart);

(*  MOVED TO FRAMEWORK CLASSES


CgeEmptyCursor;

        Console.InitGL;

      {$ifdef windows}
        if Mother^.Debug.Enabled then AddLog('Starting the pen tablet manager...');
        WinTabManager:= TWinTabmanager.Create;
        EngineObject.AddObj(WinTabManager);
      {$else}
        //In Linux, the tablet manager is created and owned by the window manager,
        //  due to deep integration with the main event loop
      {$endif}

        EngineObject.AddObj(TGamepadManager.Create);

        Mother^.Display.BlackoutStart:= Now();

        Console.Draw(No, Yes);

        if Mother^.Debug.Enabled then AddLog('Loading the fixed font...');
        InitFixedFont;

        if Mother^.Debug.Enabled then AddLog('Loading the stub texture...');
        InitStubTexture;

        LoadQF;

        if Mother^.Debug.Enabled then AddLog('Initializing sound...');

        Mother^.LockupGuard.Guard(@SafeInitSound, luc_SoundSystemStart);
*)

//        if State.DebugMode then AddLog('Loading the module...');
        EngineObject.AddObj(TModule.Create);
        Module:=EngineObject.Last as TModule;

       {$ifdef windows}
        if Mother^.Debug.Enabled then AddLog('Deleting the splash screen...');
       {$endif}
        KillSplashWindow;


        InitFpsCounter;

        Mother^.Display.LogoAnimationPlaying:= //not Mother^.Debug.Verbose and
          not NoHardwareAcceleration and not NowRestarting;
        Mother^.Display.LogoAnimationStart:= Now();
        Mother^.Display.BeginBlackout;
      Except
        KillSplashWindow;

        EverythingIsOk:=No;
        if not(ExceptObject is EFake) then Try
            Die(MI_DIED_INITIALIZING, ['']);
          Except
          End;
      End;
      KillSplashWindow;

      // ******** MAIN LOOP:
      if Mother^.Debug.Enabled then AddLog('CGE initialization took %1:%0. Entering the main loop.',[DateTimeToP(Now() - CgeStartTime, 1), DateTimeToP(Now() - CgeStartTime, 0)]);

      if EverythingIsOk then begin
        Mother^.Safe:= true;

        Framework.MainThreadMainLoop;
      end;
(*
      try
        //if State.DebugMode then AddLog('Second Stage Init took %0 seconds.',[(Now() - SecondStageInitTime) * 86400.0]);
        if Mother^.Debug.Enabled then AddLog('CGE initialization took %1:%0. Entering the main loop.',[DateTimeToP(Now() - CgeStartTime, 1), DateTimeToP(Now() - CgeStartTime, 0)]);
        Mother^.State.ProgramInitialized:= Yes;

        Framework.MainThreadLoop;

       {$ifdef unix}
        if ReceivedSigterm then AddLog(RuEn(
          'Получен SIGTERM, закрываемся.',
          'SIGTERM received, performing shutdown.'));
       {$endif}
        if Mother^.Debug.Enabled then AddLog('...exited the main loop.');
      Except
          //fake exceptions are required due to the
          // rather paranoidal reaction of the X server to closing the window -
          // it yells "fatal error!" and tries to kill our app on the spot...
          // so we'll cheat and just won't return from the callback function,
          // by raising a marked exception.
        if not (XServerBully) then begin
          EverythingIsOk:=No;
          if not (ExceptObject is EDying) then Try
            Die(MI_ERROR_UNCAUGHT_EXCEPTION);
          Except
          End;
        end
        else XServerBully:=No;// allows error messages again
      End;
*)
//    until true;//not State.RestartRequested or not EverythingIsOk;

    PerformCgeShutdown;
  end;
  



  procedure SafeInitFramework; cdecl;
  begin
    Framework:= TAbstractFramework.GetFrameworkByName(Mother^.State.FrameworkName).Create;
    EngineObject.AddObj(Framework);
  end;


  procedure PerformCgeShutdown;
  var
    UnprocessedErrorMessages: UnicodeString;
  begin
    Mother^.Safe:= false;
    // ******** CLOSING SEQUENCE:
      CreateSplashWindow('_shutdown');

      While EngineObject.Length > 0 do begin
        name:='<Unknown>';
        Try
          name:=EngineObject.Last.ClassName;
          if Mother^.Debug.Enabled then AddLog('Destroying engine object #%0 (%1)', [EngineObject.Length, name]);
          if EngineObject.Last is TThread then begin
            (EngineObject.Last as TThread).Terminate;
            if Mother^.Debug.Enabled then AddLog('  waiting for the thread to terminate...');
            (EngineObject.Last as TThread).WaitFor;
          end;
          EngineObject.Last.Free;
        Except
          Try
            Die('The %0.Free destructor has crashed.', [name]);
          Except
          End;
        End;
        EngineObject.Delete(EngineObject.High);
      end;
//chbt_StartChecking();
//    pointer(LockupGuard):= nil; //it is released by now. Was needed for KillMyself().
    KillSplashWindow;

    {$ifdef windows}
    TerminateSehHack;
    {$endif}

    UnprocessedErrorMessages:= PullWarnings();
     //These should be empy if everything goes as designed
    if UnprocessedErrorMessages <> ''
      then AddLog(RuEn(
        'Оставались необработанные сообщения об ошибках:'#10#13#10#13'%0',
        'There were unprocessed error messages left:'#10#13#10#13'%0'
        ), [UnprocessedErrorMessages]);

    AddLog(MI_LOG_END, [MessageContainer[MI_CGE_TITLE]]);

    if Assigned(LogStream) then LogStream.Free;

    //CollateWarningsFromAllThreads; //These should be empy if everything goes as designed

    if
      //((not EverythingIsOk)
      //or Mother^.ExceptionState.NowDying)
     //and (Length(Mother^.ExceptionState.Warnings) > 0)
      UnprocessedErrorMessages <> ''
    then DisplayDyingYells(UnprocessedErrorMessages);

    //MessageContainer.Free;
    //Config.Free;

    //Mother^.State.Terminated:= true;
    //DoneCriticalsection(Mother^.CriticalSection);

    if Mother^.State.RestartRequested
      then RestartMyself('--restart');


    if Mother^.Module.ExitWithoutTryingToUnloadTheDLL then KillMyself();
    //***TODO: Investigate! Crashed down here (Oleauth64.dll and so on)
    //after going out of hibernation!

    halt(0);

//      if EverythingIsOk then Halt(0) else Halt(1);
//    TerminateProcess(
  end;

