{
    This file is part of Chentrah,
    Copyright (C) 2004-2014 Anton Rzheshevski (chebmaster@mail.ru).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see http://www.gnu.org/licenses/

 **********************************************************************

    This file contains the module class implementation

 **********************************************************************}

  procedure TModule.Load;
  begin

    f_Index:= f_IndexToLoad;
    ResetMotherErrorState();
    with Mother^.Module do begin
      MustUnload:= false;
      IsSavingSession:= false;
      ReadyToUnload:= false;
    end;
    Console.SetDefaultBg(dbTitle);
    if Mother^.Debug.Verbose then AddLog(' Loading the module %0 ...', [_FName]);
    try
      if f_Index > 0 then begin
        if IsDLLLoaded then Die(MI_ERROR_PROGRAMMER_NO_BAKA, [RuEn(
          'Попытка загрузить модуль когда модуль уже загружен',
          'Attempt to load a module when the module is already loaded')]);
        if not FileExists(_Fromdir + _FName)  then begin
          {if not Mother^.State.SSE2Available and FileExists(_Fromdir + sse2fname)
            then Die(MI_NO_NOSSE2)
            else} Die(MI_ERROR_MODULE_MISSING, [_Fromdir + _FName]);
        end;
        f_dll_age:= FileAge(_Fromdir + _FName);
        {$ifdef windows}
        if Mother^.State.DeveloperMode {and not RunningInWine} then begin

          chCopyFile(_Fromdir + _FName, _ToDir + _FName);

          {$if defined(cpu64) or defined(cpuarm)}
          if FileExists(_Fromdir + ChangeFileExt(_FName, '.zd2')) then
          {$endif}
            chCopyFile(_Fromdir + ChangeFileExt(_FName, '.zd2'), _ToDir + ChangeFileExt(_FName, '.zd2'));

          _dllname:= _ToDir + _Fname;
          // Windows locks executable files so that
          // they could not be changed while they are running,
          // thus making recompile-on-the-fly impossible.

           //To hack around this, we make a copy
          // of the original dll and load that copy, leaving the original file
          // alone so that the compiler can rewrite it.

          //Since Wine doesn't emulate this particular feature of msWindows,
          // we can omit this step.
          //EDIT: they've implemented the file locking in 0.9.56
        end
        else
        {$endif}
          _dllname:= _FromDir + _Fname;//in Linux we don't have to bother.

        Mother^.Module.ControlLogicThreadStackSize:= 1024 * 1024 * MList[f_Index].ControlLogicThreadStackSize;
        Mother^.Module.BackgroundThreadStackSize:= 1024 * 1024 * MList[f_Index].BackgroundThreadStackSize;
        _LoadMeDll(_dllname);

        _LoadAddresses;
      end
      else begin
        LoadBIM;
      end;
      AttemptNum:= -1; //Stop trying to reload it. The compilation is finished
      //  (as we succeeded in loading the procedure addressses) so it's a genuine failure.

      _Initialize;


      Mother^.Module.State:= ms_Loaded;
      SwitchedConsoleOff:=false;
      if Assigned(Framework) then begin
        Framework.SetWindowName(PUnicodeChar(
          MessageContainer[MI_CGE_TITLE] + ' - ' + MList[f_Index].Name));
      end;
    except
      ResetExecutionZoneFlags;
      Die(MI_ERROR_CANNOT_LOAD_MODULE, [_Fname]);
    end;
  end;

  procedure TModule.LoadBIM;
  begin
     if Mother^.Debug.Verbose then AddLog('   Attaching the built in control panel module.');
    pointer(M_Initialize):= @BimInitialize;
    pointer(M_Pulse):= @BIMPulse;
    pointer(M_Cleanup):= @BIMCleanup;
    pointer(PassUnhandledException):= @BimPassUnhandledException;
    _dllname:= _FromDir + '$BIM';
  end;

  procedure TModule.Unload;
  begin
    if not M_Cleanup() then begin
      DieReraisingExceptionScreenedByModule(
        MsgFormat(MI_CRASHED_UNLOADING_MODULE, [MList[f_Index].Name]));
    end;
    _UnloadMeDll;
  end;

  constructor TModule.Create;
  begin
    inherited Create;
    f_Index:= NOT_A_MODULE;
    f_IndexToLoad:= NOT_A_MODULE;
    _RebuildModulesList;
    ModuleCheckPeriod:=Config.IntChk['modules', 'CheckIfModulesAreRecompiledPeriod', 100, 60000];
    Reload(GetDefaultNum())
  end;

  Procedure TModule._Initialize;
  var DoubleError: boolean = false;
  begin
    Mother^.Module.PureName:= PureName;
    if Mother^.Debug.Verbose then AddLog('Initializing the module %0 ...', [ExtractFileName(_FName)]);
    Try
      Mother^.Module.CurrentlyExecutingInd:= f_Index;
      if not M_Initialize(Mother)
      then begin
        DoubleError:= Mother^.ExceptionState.ModuleIsAlreadyProcessingUnhandledException;
        ResetExecutionZoneFlags;
        try
          AddLog('  Trying to unload the invalid DLL...');
          _UnloadMeDll;
        except
          with ExceptObject as Exception do AddLog('    Screened %0  %1', [string(ClassName), Message]);
          Mother^.State.StateTrashedRestartRequired:= true;
        end;
        if DoubleError
          then Die(RuEn(
            'Модуль рухнул при попытке обработать ошибку %0',
            'The module crashed trying to process an error %0')
            ,[ExpExpAddress(pointer(Mother^.ExceptionState.ExceptionAddress))])
          else DieReraisingExceptionScreenedByModule(RuEn(
            'Модуль загрузился, но при инициализации произошёл сбой.',
            'The module had loaded but failed to initialize.'));
          //raise exception.Create('');
        end;
      ResetExecutionZoneFlags;
      if Mother^.Debug.Verbose then AddLogOk;
    Except
      ResetExecutionZoneFlags;
      Die(MI_ERROR_MODULE_CRASHED_AT_LOADING, [_FName]);
    End;
  end;

  destructor TModule.Destroy;
  var
    kresult: UnicodeString;
  begin
   { if Mother^.Module.ExitWithoutTryingToUnloadTheDLL then begin
      AddLog(RuEn(
        'Аварийное завершение - пытаюсь убить DLL модуля..',
        'Emergency shutdown, trying to kill the module DLL..'));
      try
        _UnloadMeDll;
        kresult:= RuEn('НЕУДАЧА!','FAILED!');
        {$ifdef unix}
          if DLL > 0 then FpKill(DLL, SIGKILL);
        {$else}

          if {$ifdef cpu32}Assigned(TerminateProcess) and {$endif}
            (DLL <> 0)
          then if TerminateProcess(DLL, 0)
            then kresult:= RuEn('успешно','success');
        {$endif}
        AddLog('  ..%0',[kresult]);
      Except
        With ExceptObject as Exception do AddLog(RuEn(
          '  ..НЕУДАЧА! %0 %1',
          '  ..FAILED! %0 %1'
        ), [string(ClassName), String(Message)]);
      end;
    end
    else} begin
      if Mother^.Module.State = ms_Loaded
      then begin
        if Mother^.Debug.Verbose then AddLog('Unloading the module %0 ...', [ExtractFileName(_FName)]);
        Mother^.Module.State:= ms_Unloading; //ms_Exiting;
        Mother^.Module.ApplicationExiting:= true;
        repeat
          Sleep(2);
          Pulse;
        until not (Mother^.Module.State in [ms_Unloading, ms_WaitingForModuleToFinish]);
        if Mother^.Debug.Verbose then AddLogOk;
      end;
    end;
    inherited;
  end;

  function TModule._GetProcAddress(ProcName: ansistring) : pointer;
  var
    exename: TFileNameString;
  begin
   {$IFDEF windows}
    Result:=Windows.GetProcAddress(DLL, PAnsiChar(ProcName));
   {$ELSE}
    Result:=dlsym(DLL, PAnsiChar(ProcName));
  {$ENDIF}
    if Assigned(Result) then begin
      exename:= GetExeFileName(Result);
      VerboseLog('    %0() at %1 in %2', [ProcName, Result, exename]);
    end
    else
      VerboseLog('    %0() not found!', [ProcName]);
  end;

  Procedure TModule._LoadAddresses;
  begin
    if f_Index > 0 then begin
      {$ifdef buildmein}
        pointer(M_Initialize):= @IM_Init;
        pointer(M_Pulse):= @IM_Pulse;
 //       pointer(M_Shutdown):= @IM_Shutdown;
        pointer(M_Cleanup):= @IM_Cleanup;
      {$else}
        pointer(M_Initialize):=_GetProcAddress(PAnsiChar(AnsiString('M_Init')));
        pointer(M_Pulse):=_GetProcAddress(PAnsiChar(AnsiString('M_Pulse')));
//        pointer(M_Shutdown):=_GetProcAddress(PChar('M_Shutdown'));
        pointer(M_Cleanup):=_GetProcAddress(PAnsiChar(AnsiString('M_Cleanup')));
       {$ifndef buildmein}
        pointer(PassUnhandledException):= _GetProcAddress(PAnsiChar(AnsiString('M_PassUnhandledException')));
       {$endif}
       //pointer(PassThreadDying):= _GetProcAddress(PChar('M_PassThreadDying'));

        if not Assigned (M_Initialize)
           or not Assigned (M_Pulse)
 //          or not Assigned (M_Shutdown)
           or not Assigned (M_Cleanup)
          {$ifndef buildmein}
           or not Assigned (PassUnhandledException)
          {$endif}
           then begin
              try
                AddLog('  Trying to unload the invalid DLL...');
                pointer(PassUnhandledException):= nil;
                _UnloadMeDll;
              except
                with ExceptObject as Exception do AddLog('    Screened %0  %1', [string(ClassName), string(Message)]);
                Mother^.State.StateTrashedRestartRequired:= true;
              end;
             Die(MI_ERROR_INVALID_MODULE, [_dllname]);

           end;
     {$endif}

    end
    else begin
      M_Initialize:= nil;
      M_Pulse:= nil;
//      M_Shutdown:= nil;
      M_Cleanup:= nil;
     {$ifndef buildmein}
      PassUnhandledException:= nil;
     {$endif}
     //PassThreadDying:= nil;
    end;
  end;

   function TModule._GetFname: TFileNameString;
   begin
     Result:=_ToDir + _FName;
   end;

   procedure TModule._LoadMeDll(n: TFileNameString);
   var
     repeatcount: integer = 0;
   begin
     {$ifdef unix}
       DLL:= dlopen(PAnsiChar(FileNameToUtf8(n)), RTLD_NOW);
       If not Assigned(DLL) then Die('dlopen() returned NULL for'#10#13'  %0', [n]);
     {$else}
       repeat
         if repeatcount > 0 then Sleep(50);
         SetLastError(0);
         DLL:=
            {$ifdef che_unicode}
              LoadLibraryW
            {$else}
              LoadLibrary
            {$endif}
                (PFileNameChar(n));
         inc(repeatcount);
       until (DLL <> 0) or (repeatcount > 13);
       if DLL = 0 then Die('LoadLibrary() returned zero for'#10#13'  %0', [n]);
     {$endif}
   end;

   procedure TModule._UnloadMeDll;
   begin
    {$ifndef buildmein}
     Mother^.Module.DllInitialized:= false;
     try
      {$ifdef unix}
       if DLL = nil then Exit;
       dlClose(DLL);
       DLL:=nil;
      {$else}
       if DLL = 0 then Exit;
       FreeLibrary(DLL);
       DLL:=0;
      {$endif}
     except
       try Die('Crashed trying to unload the module DLL') except end;
     end;
    {$endif}
   end;

  function TModule.IsDllLoaded: boolean;
  begin
    {$ifdef unix}
     Result:= DLL <> nil;
    {$else}
     Result:= DLL <> 0;
    {$endif}
  end;


  procedure TModule.Checkdate;
  var fa: longint;
  begin
    if f_Index < 1 then Exit;
    if Tick() < ReloadAt then Exit;
    fa:=FileAge(_Fromdir + _FName);
    if (fa > 0) and (f_dll_age = 0) then f_dll_age:=fa;
    if (fa > 0) and (fa <> f_dll_age)
    then begin
      PrevModuleDateCheckTick:=Tick;
      f_dll_age:=fa;
      if Mother^.Module.State = ms_Loaded then Mother^.Module.State:=ms_Unloading;
    end
    else ReloadAt:=Tick() + ModuleCheckPeriod;
  end;
  

  procedure TModule.Reload(i: integer);
  var
    ecap, etxt, oops: UnicodeString;
  begin
    AddLog(RuEn(
      'Приступаю к загрузке модуля "%0"',
      'Going to load module "%0"'
      ), [Mlist[i].Name]);

    if Mother^.State.StateTrashedRestartRequired then begin
      oops:= RuEn(
        'Данные приложения необратимо повреждены! Выполняю принудительный выход.',
        'Application data is trashed! Performing forced exit.');
      AddLog(oops);
      Mother^.Module.State:= ms_DoNothing;
      Mother^.State.ExitRequested:= true;
      Config.Str['session', 'MostRecentUsedModule']:= MList[0].PureN; //module selection menu
      GetModuleStopMessage(ecap, etxt);
      if ecap <> '' then etxt:= ecap + #10#13#10#13 + etxt;
      if etxt = '' then etxt:= oops;
      Config.Str['session', 'LastWill']:= StringReplace(etxt, #10#13, '{CR/LF}', [rfReplaceAll]);
      Mother^.Module.ExitWithoutTryingToUnloadTheDLL:= true;
      Exit;
    end;

    f_IndexToLoad:= i;


    CgeHourglassCursor;
//    Mother^.GAPI.SkipSwapBuffers:= true;
    Mother^.Display.BeginBlackout;
    Mother^.Display.fadeIn:= 0;

    { It can freeze trying to load the DLL, thus leaving the previous frame
        hanging on the screen. Here we remedy that by performing a guaranteed
        flip with back buffer filled with black: }
    Console.DrawBackground;
    Framework.SwapBuffers;
    Console.DrawBackground;
    Framework.SwapBuffers;

    case Mother^.Module.State of
      ms_Loaded, ms_Crashed: Mother^.Module.State:= ms_Unloading;
      ms_NotInitialized: Mother^.Module.State:= ms_Loading;
      ms_CrashedAndDone: begin
        _UnloadMeDll;
        Mother^.Module.State:= ms_Loading;
      end;
    else
      //do nothing
    end;


//    Mother^.Display.FadeIn:= 1.0;
    
    if i > 0 then begin
      _ToDir:= OptiPath(Mother^.State.CacheDir + 'bin' + PathSlash);
      MakeSurePathExists(_ToDir);
      CheckForGuardedException;
      _FName:= OptiFileName(Mother^.State.InstallPath + 'modules' + PathSlash + MList[i].Dll);
      _FromDir:= ExtractFilePath(_Fname);
      Mother^.State.ModuleBaseDir:= _FromDir; //_BaseDir:= _FromDir;
      PureName:= WideToAnsi(ChangeFileExt(ExtractFileName(_Fname), ''));
      Mother^.Module.NameW:= MList[i].Name;
//      Mother^.ModuleRequiresOpenGL2:= MList[i - 1].RequiresOpenGL2; //Error: identifier idents no member "ModuleRequiresOpenGL2"
      _FName:=  DllPrefix + ChangeFileExt(
        UnicodeToFileName(AnsiToWide(PureName))
        {$ifdef cpu64} + TFileNameString('-x86_64') {$endif}
        {$ifdef che_legacy} + TFileNameString('_legacy') {$endif}
        , DllExtension);
    end
    else begin
      _FromDir:= Mother^.State.InstallPath + 'modules' + PathSlash + 'chentrah' + PathSlash;
      _FName:= '$BIM';
      _ToDir:= '';
    end;
    Mother^.Module.FileName:= _FromDir + _Fname;

    ReloadAt:= Tick();
  end;
  
  procedure TModule.ReactToSpace;
  begin
//    if Mother^.Module.State = ms_Crashed then GoToModuleSelectionMenu;
    case Mother^.Module.State of
      ms_Crashed: AddLog(RuEn(
        'У модуля ещё остаются работающие потоки. Дождитесь их завершения или перезапустите всё приложение целиком.',
        'The module stil has running threads. Wait for their completion or restart the application as a whole.'));
      ms_CrashedAndDone: begin
        Mother^.Module.State:= ms_Loading;
        ReloadAt:=Tick();
      end;
    end;
  end;
  
  procedure TModule.ReactToEsc;
  begin
    if Mother^.Module.State in [ms_CrashedAndDone]
      then Mother^.State.ExitRequested:=Yes;
  end;
  
  function TModule.GetDefaultNum: integer;
  var
    i: integer;
    s: ansistring;
  begin
    Result:= 0;
    if Mother^.State.IsServer
       then s:= Mother^.State.ServerModuleName
    else begin
      s:= ParamStr(1);
      if (s= '') or (copy (s, 1, 2) = '--')
      then begin
        s:= Config.Str['session', 'MostRecentUsedModule'];
        if s = '' then s:= Config.Str['Main', 'DefaultModule'];
      end;
    end;
    s:=lowercase(s);
    if s > '' then
      for i:=0 to high(MList) do
        if MList[i].PureN = s then Exit(i);
    if Mother^.State.IsServer
       then Die(RuEn(
         'WARNING: "%0" нет в текущем списке dll модулей.',
         'WARNING: "%0" is not a currently known module dll name.'
         ), [s]);
    AddLog(RuEn(
      'WARNING: "%0" нет в текущем списке dll модулей. Вызываю меню выбора модуля.',
      'WARNING: "%0" is not a currently known module dll name. Invoking the module selection menu.'),
       [s]);
  end;
  

  function TModule.Pulse(): boolean;
  var
    success, pesets: boolean;
    a: pointer;

  begin
//dbgsay(GetEnumName(typeinfo(TModuleState), ord(Mother^.Module.State)));

//addlog('mother TModule.pulse, state is %0', [GetEnumName(TypeInfo(TModuleState), Ord(Mother^.Module.State))]);
    if Mother^.Module.State <> ms_Loaded then with Mother^.Input do begin
      TextInput:= false;
      GrabPointer:= false;
    end;

    if MessageContainer.CurrentLanguage >=0 then GiveWarnings();
    { A one-shot function that fills the warnings list,
      must be called only after the language is chosen.
      I'd wish I hadn't have to put this call here, but... oh, well :|    }
      
    if ((length (WarningMessage) > 0)
      or (Length(StopMessage) > 0)
      or (MessageContainer.CurrentLanguage < 0))
      and (Index <> 0) //not BimLoaded
      then Module.GoToModuleSelectionMenu; //go to the mother menu

    case Mother^.Module.State of
      ms_DoNothingUntilMotherSavesTheSession: begin
Die('не доделано');

      end;
      ms_Loading: begin
        if Mother^.Debug.Verbose then AddLog('Mother is loading the module #%0...',[f_indexToLoad]);
        Result:=true;
        try
          if (f_IndexToLoad > 0) or (bimState = bimsModuleMenu)
            then Config.Str['session', 'MostRecentUsedModule']:= MList[f_IndexToLoad].PureN;
          if Mother^.State.StateTrashedRestartRequired then begin
             Mother^.Module.State:= ms_DoNothing;
             AddLog(RuEn(
               'Нарушена целостность структуры данных приложения. Принудительный перезапуск приложения.',
               'Application state is trashed, forcing application restart.'));
             Mother^.State.RestartRequested:= Yes;
          end
          else begin
            Load;
            Mother^.GAPI.SkipSwapBuffers:= true; //to avoid flickering of an unrendered frame.
          end;
        except
          result:=false;
          ReloadAt:=Tick() + ModuleCheckPeriod;
          if (AttemptNum <= 0) or (AttemptNum > MAX_MODULE_LOAD_ATTEMPTS) then begin
            AttemptNum:= -1;
            if Mother^.Debug.Enabled or Mother^.State.DeveloperMode
            then begin
              AddLog(MI_DEVMODE_MODULE_UNLOADED, [StopDying()]);
              _BSOD;
            end
            else
              _SwitchToStopScreen(false);
            _UnloadMeDll;
            Exit;
          end
          else begin
            AddLog(RuEn(
              'Попытка #%0 провалилась'#10#13#10#13'%1'
             ,'Attempt #%0 failed.'#10#13#10#13'%1')
             ,[AttemptNum, StopDying()]);
            inc(AttemptNum);
          end;
        end;
      end;
      ms_Unloading: begin
        Result:=false;
        Mother^.GAPI.SkipSwapBuffers:= true;
        if Mother^.Debug.Verbose then AddLog('Mother is unloading the module...');
        if not IsDLLLoaded then begin
          if Mother^.Debug.Verbose then AddLog('  There is no DLL to unload!');
          if Mother^.Module.ApplicationExiting
            then Mother^.Module.State:= ms_DoNothing
            else Mother^.Module.State:= ms_Loading;
          Exit;
        end;
        Mother^.Module.CurrentlyExecutingInd:= f_Index;
        Mother^.Module.MustUnload:= true;
        success:= M_Pulse();
        Mother^.Module.CurrentlyExecutingInd:= NOT_A_MODULE;
        if success then begin
          ResetExecutionZoneFlags;
          if StopIfScreenedExceptionPresent then Exit(False);
          if not Mother^.Module.ReadyToUnload then begin
            if Mother^.Debug.Verbose then AddLog('The module is not ready to be unloaded, mother is switching to waiting mode.');
            Mother^.Module.State:= ms_WaitingForModuleToFinish;
            Mother^.Module.WaitingToFinishTimeoutMoment:= Now() + MODULE_STOP_TIMEOUT;
          end else begin
            ResetExecutionZoneFlags;
            Unload;
            if Mother^.Module.ApplicationExiting
              then Mother^.Module.State:= ms_DoNothing
              else Mother^.Module.State:= ms_Loading;
            AttemptNum:=1;
          end;
        end
        else begin
          if Mother^.State.Fatality then _ProcessFatality; //causes immediate restart
          if Mother^.State.DeveloperMode or Mother^.Debug.Enabled
          then begin
            AddLog(MI_ERROR_MODULE_CRASHED_UNLOADING, [StopDying()]);
            Mother^.State.StateTrashedRestartRequired:= Yes;
            _BSOD;
//          _UnloadMeDll;
          end
          else
            _SwitchToStopScreen(false);
          Result:=False;
        end;
      end;
      ms_WaitingForModuleToFinish: begin
        Mother^.Module.CurrentlyExecutingInd:= f_Index;
        try
          success:= M_Pulse(); //The module is required to stop all activity
            //except displaying the "Waiting for background tasks to terminate" screen.
          Mother^.Module.CurrentlyExecutingInd:= NOT_A_MODULE;
          SetLength(Mother^.Input.Events, 0);
          pesets:=false;
        except
          ResetExecutionZoneFlags;
          success:= false;
          pesets:= true;
         {$ifndef buildmein}
          try
            a:= ExceptAddr();
            Die(RuEn(
              'ФАТАЛЬНО:'#10#13'  непойманное исключение пересекло границу dll''ы!',
              'FATAL:'#10#13'  an uncaught exception crossed the dll boundary!'))
          except end;
         {$endif}
        end;
        if not success then begin
          Mother^.Module.RollBackTheSessionOnNextLoad:= False;
          Mother^.Module.CurrentlyExecutingInd:=NOT_A_MODULE;
          if Mother^.State.Fatality then _ProcessFatality; //causes immediate restart
          Mother^.State.StateTrashedRestartRequired:= Yes;
          if Mother^.State.DeveloperMode or Mother^.Debug.Enabled then begin
            AddLog(MI_DEVMODE_MODULE_UNLOADED, [StopDying()]);
            _BSOD;
            Result:=False;
          end
          else
            _SwitchToStopScreen(pesets);
        end
        else begin
 // ------  normal operation here ---------
          if StopIfScreenedExceptionPresent then Exit(False);
          if not Mother^.Module.ReadyToUnload then begin
            if
              //timeout (default 15 seconds) for terminating the background threads
              (Now() > Mother^.Module.WaitingToFinishTimeoutMoment)
              //the logic thread is the king. Wait until it finishes for however long it takes.
              and not Mother^.Module.SessionNotSaved
            then begin
              Mother^.Module.State:= ms_DoNothing;
              AddLog(RuEn(
                'Превышено время ожидания завершения модуля. Принудительный перезапуск приложения.',
                'Timeout. The module takes too long to terminate. Forcing application restart.'));
              Mother^.State.RestartRequested:= Yes;
            end;
          end else begin
            if Mother^.Debug.Verbose then AddLog('Mother has finished waiting for the module to shut down.');
            ResetExecutionZoneFlags;
            Unload;
            if Mother^.Module.ApplicationExiting
              then Mother^.Module.State:= ms_DoNothing
              else Mother^.Module.State:= ms_Loading;
            AttemptNum:=1;
          end;
        end;
      end;
{          ms_Exiting: begin
        {There's no need to care about the dll threads here,
          as the Chentrah process calls TerminateProcess() on itself as it finishes.}
        if {$ifdef unix} DLL <> nil {$else} DLL <> 0 {$endif} then begin
          Mother^.Module.CurrentlyExecutingInd:= f_Index;
{              Mother^.Module.MustUnload:= true;
          success:= M_Shutdown();
          Mother^.Module.CurrentlyExecutingInd:= NOT_A_MODULE;
          if not success then begin
            DieReraisingExceptionScreenedByModule(MsgFormat(MI_ERROR_MODULE_CRASHED_UNLOADING_SHORT, []));
          end;
}
          Unload;
        end;
        Result:=True;
      end;
}
      ms_Loaded: begin
        Mother^.Module.CurrentlyExecutingInd:= f_Index;
        try
          success:= M_Pulse();
          Mother^.Module.CurrentlyExecutingInd:= NOT_A_MODULE;
          SetLength(Mother^.Input.Events, 0);
          pesets:=false;
        except
          ResetExecutionZoneFlags;
          success:= false;
          pesets:= true;
          Mother^.State.StateTrashedRestartRequired:= true;
         {$ifndef buildmein}
          try
            a:= ExceptAddr();
            Die(RuEn(
              'ФАТАЛЬНО:'#10#13'  непойманное исключение пересекло границу dll''ы!',
              'FATAL:'#10#13'  an uncaught exception crossed the dll boundary!')
           {$ifndef unix}
             //call the dll internal self-debugger
           //  + #10#13'%0', [M_ReportDebugBacktrack(a)] //not necessary anymore
           {$endif}
          ) except end;
          {$endif}
        end;
        if not success then begin
          Mother^.Module.RollBackTheSessionOnNextLoad:= False;
          Mother^.Module.CurrentlyExecutingInd:=NOT_A_MODULE;
          if Mother^.State.Fatality then _ProcessFatality; //causes immediate restart
          Mother^.State.StateTrashedRestartRequired:= Yes;
          if Mother^.State.DeveloperMode or Mother^.Debug.Enabled then begin
            AddLog(MI_DEVMODE_MODULE_UNLOADED, [StopDying()]);
            Result:=False;
            _BSOD;
          end
          else begin
            Result:=False;
            _BSOD;
//            _SwitchToStopScreen(pesets);
            Exit;
          end;
        end
        else begin
          //Mother^.Module.CurrentlyExecutingInd:=NOT_A_MODULE;
          ResetExecutionZoneFlags;
          if not SwitchedConsoleOff then begin
            Console.Fade;
            SwitchedConsoleOff:=Yes;
          end;
          Result:=True;
          if StopIfScreenedExceptionPresent then Result:= False;
          AttemptNum:=0;
          CheckDate;
        end;
        if Mother^.Module.RequestToChooseModule then begin
           Mother^.Module.RequestToChooseModule:= No;
           GoToModuleSelectionMenu;
        end;
      end;
      ms_Crashed: begin
        Result:= false;
        ReloadAt:= 0;
        f_dll_age:= 0;
        Console.DrawBSODBackground;
        if not DoesModulehaveThreadsRunning
          then Mother^.Module.State:= ms_CrashedAndDone
      end;
      ms_CrashedAndDone: begin
        Result:= false;
        ReloadAt:= 0;
        f_dll_age:= 0;
        if Mother^.State.IsServer then begin
          if Mother^.Debug.Verbose then AddLog('Server mode: module crashed, restarting in 5 seconds.');
          Mother^.State.RestartRequested:= true;
          Sleep(5000);
        end
        else begin
          Console.DrawBSODBackground;
          if not (Mother^.State.DeveloperMode or Mother^.Debug.Enabled)
            then _SwitchToStopScreen(false);
        end;
      end;
    end;
  end;

  function TModule.StopIfScreenedExceptionPresent: boolean;
  begin
    Result:= Mother^.ExceptionState.NowDying;
    if Result then begin
      if Mother^.Debug.Verbose then AddLog('Found a screened exception, switching to stop screen...');
      _SwitchToStopScreen(false);
    end;
  end;

  procedure TModule._SwitchToStopScreen(itisfatal: boolean);
  var c, t: UnicodeString;
  begin
//dbgsays('b');
    f_IndexToLoad:= 0;

    Console.Fade;
    AddMStopMessage(itisfatal);
    cl_builtinmodule.prevState:= bimsInvalid;
    if itisfatal then
      cl_builtinmodule.bimState:= bimsModuleFatalError
    else begin
      cl_builtinmodule.bimState:= bimsModuleError;
    end;

    if Assigned(Framework) then begin
      Framework.PlayWav('crashed.wav');
    end;
    if Mother^.State.IsServer
      then begin
        cge.GetModuleStopMessage(c, t);
        Framework.YellAndExpire(c + #10#13#10#13 + t);
      end
      else Reload(0);

//dbgsays('c');
  end;

  procedure TModule._BSOD;
  var
    c, t, oops: UnicodeString;
  begin
    Mother^.Module.State:= ms_Crashed;
    if Assigned(Framework) then begin
      Framework.PlayWav('oops.wav');

      if Mother^.State.StateTrashedRestartRequired then oops:= RuEn(
        //red instead of blue, too B)
        ' - пришёл пушной полярный, только перезапуск спасёт отца русской демократии.',
        ' - it''s dead, Jim! Crashed hard! Only total application restart would help.')
      else oops:= RuEn(
        ' - BSOD',
        ' - blue screen of death');
      Framework.SetWindowName(PUnicodeChar(
        MessageContainer[MI_CGE_TITLE] + oops));
      if Mother^.State.IsServer then begin
        cge.GetModuleStopMessage(c, t);
        if (c = '') and (t = '') then begin
          c:= RuEn(
            'Крах.',
            'Crashed.');
          t:= RuEn(
            'Подробности см. в файле журнала.',
            'See details in the log file.');
        end;
        Framework.YellAndExpire(c + #10#13#10#13 + t);
      end;
    end;
  end;

  procedure TModule._ProcessFatality;
  var
//    F: TextFile;
    s: TFileStream;
    w: UnicodeString;
    u: Utf8String;
  begin
    w:= StopDying();
    AddLog(
       RuEn(
         'Обнаружена фаталити, пишу завещание и самоубиваюсь:#10#13#10#13%0',
         'State.Fatality detected, writing the last will and terminating myself:#10#13#10#13%0'),
       [w]
     );
    try
      MakeSurePathExists(UnmangleFileName('*H/logs'));
      s:= TFileStream.Create(UnmangleFileName('*H/logs/last-last-will.txt'), fmCreate);
//      AssignFile(F, _UnmangleFileName('*H/logs/last-last-will.txt'));
//      Rewrite(F);
      u:= Utf8Encode(w) + #10#13;
      s.Write(u[1], Length(u));
//      WriteLn(F, Utf8Encode(w));
//      CloseFile(F);
      s.Free;
      RestartMyself('--last-will');
    except
      AddLog(RuEn(
        'Не удалось записать файл с завещанием, перезаупуск невозможен!',
        'Failed to write the last will file, restart is not possible!'));
    end;
    KillMyself;
  end;

  procedure TModule.GoToModuleSelectionMenu;
  begin
    {f_IndexToLoad:= 0;
    case Mother^.Module.State of
      ms_Loaded, ms_Crashed: Mother^.Module.State:= ms_Unloading;
      ms_CrashedAndDone: begin
        _UnloadMeDll;
        Mother^.Module.State:= ms_Loading;
      end;
    else
      //do nothing, it will happen naturally
    end;}

    Console.Fade;
    cl_builtinmodule.prevState:= bimsInvalid;
    cl_builtinmodule.bimState:= bimsModuleMenu;

    Reload(0);
  end;

  procedure TModule.RollbackSession;
  begin
    if Mother^.Module.State <> ms_Loaded then begin
      AddLog(RuEn('Откат сессии не выполнен так как модуль не загружен.','Session rollback canceled: no module is loaded.'));
      Exit;
    end;
    Mother^.Module.RollBackTheSessionOnNextLoad:= Yes;
    Reload(f_indextoload);
  end;
  
  procedure TModule._RebuildModulesList;
  var
    F: TFileList;
    ini: TIniFile;
    i, cl: integer;
    w: UnicodeString = '';
    ModulesRequiringOGL2: UnicodeString = '';
  begin
    if Mother^.State.InstallPath = '' then begin
      SetLength(MList, 0);
      SetLength(AssetStorage, 0);
      ListAlreadyBuilt:= Yes;
    end;

    if ListAlreadyBuilt then Exit;

    F:= TFileList.Create(IncludeTrailingPathDelimiter(
                            Mother^.State.InstallPath + 'modules') + '*.module');
    if F.Count = 0 then Die('Не найдено ни одного модуля!','No modules found!', []);
    F.SortByDate;
    SetLength(MList, F.Count + 1);

    EnterCriticalSectionWv(Mother^.Core.AssetStorageCriticalSection);
    try
      for i:= 0 to High(AssetStorage) do begin
        if Assigned(AssetStorage[i]) then TMemoryStream(AssetStorage[i]).Free;
        AssetStorage[i]:= nil;
      end;
      SetLength(AssetStorage, F.Count + 1);
      for i:= 0 to High(AssetStorage) do AssetStorage[i]:= nil;
    finally
      LeaveCriticalSection(Mother^.Core.AssetStorageCriticalSection)
    end;

    FillChar(MList[0], SizeOf(MList[0]) * Length(MList), 0);
    
    if Mother^.Debug.Verbose then AddLog(RuEn('Найдено %0 модулей', '%0 modules found'), [F.Count]);
    with MessageContainer do
      for i:= F.Count - 1 downto 0 do begin
        MList[i + 1].PureN:= lowercase(ExtractFileName(ChangeFileExt(F.Names[i], '')));
        ini:= TIniFile.Create(F.Names[i]);
        if CurrentLanguage >=0 then begin
          cl:= CurrentLanguage;
          w:= Utf8Decode(ini.ReadString('module', 'name-' + LanguageId[CurrentLanguage],''));
        end
        else cl:= SubstituteLanguage;
        if w = '' then
          w:= Utf8Decode(ini.ReadString('module', 'name-' + LanguageId[SubstituteLanguage],''));
        if w = '' then Die(
          'В %0 не задано имя модуля,'#10#13'  как для языка %1, так и для языка %2.',
          'Undefined module name in %0'#10#13'  for both %1 and %2 languages.',
          [ExtractFileName(F.Names[i]),
           LanguageName(cl), LanguageName(SubstituteLanguage)]);
        MList[i + 1].Name:= w;
        MList[i + 1].Dll:= ini.readString('module', 'dll', '');
        if MList[i + 1].Dll = '' then Die(
          'В %0 не задана dll модуля.',
          'Undefined module dll in %0', [ExtractFileName(F.Names[i])]);
        MList[i + 1].RequiresOpenGL2:= 0 <> ini.readInteger('module', 'requires_opengl_2', 0);
        if MList[i + 1].RequiresOpenGL2 then  ModulesRequiringOGL2+= '  ' + MList[i].Name + #10#13;
        Mlist[i + 1].ControlLogicThreadStackSize:= ini.readInteger('module', 'control_logic_thread_stack_size_megabytes', 100);
        Mlist[i + 1].BackgroundThreadStackSize:= ini.readInteger('module', 'background_threads_stack_size_megabytes', 16);
        ini.Free;
      end;
    with MList[0] do begin
      PureN:= '$control_panel';
      Name:= RuEn(
        'Панель управления Чентра',
        'Chentrah control panel');
      Dll:= '$control_panel';
      RequiresOpenGL2:= false;
      ControlLogicThreadStackSize:= 16;
      BackgroundThreadStackSize:= 16;
    end;
    F.Free;

    ListAlreadyBuilt:= Yes;


    if ((Mother^.GAPI.VersionHi < OGL_REQ_HI2)
      or ((Mother^.GAPI.VersionHi = OGL_REQ_HI2) and (Mother^.GAPI.VersionMed < OGL_REQ_MED2)))
    and (ModulesRequiringOGL2 <> '')
      then GiveWarning('ogl_too_low_' + IniCompatibleStringForm(Mother^.GAPI.Summary),
        MI_WARN_WARNING, MI_WARNING_OPENGL_VERSION_TOO_LOW,
      [OGL_REQ_HI2, OGL_REQ_MED2,
        ModulesRequiringOGL2,
        Mother^.GAPI.VersionString,
        Mother^.GAPI.Renderer,
        Mother^.GAPI.Vendor]);
  end;

  procedure TModule.ResetExecutionZoneFlags;
  begin
   Mother^.Module.CurrentlyExecutingInd:= NOT_A_MODULE;
   Mother^.ExceptionState.ModuleIsAlreadyProcessingUnhandledException:= false;
  end;

  function TModule.GetAssetStorage: pointer;
  begin
   if Index > 0
     then Result:= AssetStorage[Index]
     else Result:= nil;
  end;

  procedure TModule.SetAssetStorage(p: pointer);
  begin
    AssetStorage[Index]:= p;
  end;
