{
    This file is part of Chentrah,
    Copyright (C) 2004-2015 Anton Rzheshevski (chebmaster@mail.ru).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see http://www.gnu.org/licenses/

 **********************************************************************

    Windows x86-64 only.

    This unit is to copensate for the bug
      http://bugs.freepascal.org/view.php?id=4605
      also known as
      http://bugs.freepascal.org/view.php?id=12974

    There is a hack for the Windows Structured Exception Handling
      mechanism allowing to catch AVs and other system exceptions
      in a DLL (typically these just ignore the try...except blocks
      not belonging to the main executable).

    This unit has some definitions borrowed from the System unit of the
      FreePascal RTL since these were defined in the implementation section.

 **********************************************************************}

{$if not defined(windows) or not defined(cpu64)}
  {$fatal This unit is only needed wor Win64}
{$endif}

{$rangechecks off}
{$overflowchecks off}

unit cl_seh_hack_win64;
interface
uses SysUtils, Classes, Windows;

  procedure InitSehHack;
  procedure TerminateSehHack;

implementation
  uses cge;     

//Copy-pasted from the System unit (FPC 2.6.4)--------------------------------\  

type
  M128A = record
    Low : QWord;
    High : Int64;
  end;

  PContext = ^TContext;
  TContext = record
    P1Home : QWord;
    P2Home : QWord;
    P3Home : QWord;
    P4Home : QWord;
    P5Home : QWord;
    P6Home : QWord;
    ContextFlags : DWord;
    MxCsr : DWord;
    SegCs : word;
    SegDs : word;
    SegEs : word;
    SegFs : word;
    SegGs : word;
    SegSs : word;
    EFlags : DWord;
    Dr0 : QWord;
    Dr1 : QWord;
    Dr2 : QWord;
    Dr3 : QWord;
    Dr6 : QWord;
    Dr7 : QWord;
    Rax : QWord;
    Rcx : QWord;
    Rdx : QWord;
    Rbx : QWord;
    Rsp : QWord;
    Rbp : QWord;
    Rsi : QWord;
    Rdi : QWord;
    R8 : QWord;
    R9 : QWord;
    R10 : QWord;
    R11 : QWord;
    R12 : QWord;
    R13 : QWord;
    R14 : QWord;
    R15 : QWord;
    Rip : QWord;
    Header : array[0..1] of M128A;
    Legacy : array[0..7] of M128A;
    Xmm0 : M128A;
    Xmm1 : M128A;
    Xmm2 : M128A;
    Xmm3 : M128A;
    Xmm4 : M128A;
    Xmm5 : M128A;
    Xmm6 : M128A;
    Xmm7 : M128A;
    Xmm8 : M128A;
    Xmm9 : M128A;
    Xmm10 : M128A;
    Xmm11 : M128A;
    Xmm12 : M128A;
    Xmm13 : M128A;
    Xmm14 : M128A;
    Xmm15 : M128A;
    VectorRegister : array[0..25] of M128A;
    VectorControl : QWord;
    DebugControl : QWord;
    LastBranchToRip : QWord;
    LastBranchFromRip : QWord;
    LastExceptionToRip : QWord;
    LastExceptionFromRip : QWord;
 end;

type
  PExceptionRecord = ^TExceptionRecord;
  TExceptionRecord = record
    ExceptionCode   : DWord;
    ExceptionFlags  : DWord;
    ExceptionRecord : PExceptionRecord;
    ExceptionAddress : Pointer;
    NumberParameters : DWord;
    ExceptionInformation : array[0..EXCEPTION_MAXIMUM_PARAMETERS-1] of Pointer;
  end;

  PExceptionPointers = ^TExceptionPointers;
  TExceptionPointers = packed record
    ExceptionRecord   : PExceptionRecord;
    ContextRecord     : PContext;
  end;

  TVectoredExceptionHandler = function (excep : PExceptionPointers) : Longint; stdcall;

function AddVectoredExceptionHandler(FirstHandler : DWORD;VectoredHandler : TVectoredExceptionHandler) : pointer;
        stdcall; external 'kernel32' name 'AddVectoredExceptionHandler';

function RemoveVectoredExceptionHandler(handler: pointer): dword;
        stdcall; external 'kernel32' name 'RemoveVectoredExceptionHandler';

const
  MaxExceptionLevel = 16;
  exceptLevel : Byte = 0;
  
function GetCurrentProcess : dword;
 stdcall;external 'kernel32' name 'GetCurrentProcess';

function ReadProcessMemory(process : dword;address : pointer;dest : pointer;size : dword;bytesread : pdword) :  longbool;
 stdcall;external 'kernel32' name 'ReadProcessMemory';

function is_prefetch(p : pointer) : boolean;
  var
    a : array[0..15] of byte;
    doagain : boolean;
    instrlo,instrhi,opcode : byte;
    i : longint;
  begin
    result:=false;
    { read memory savely without causing another exeception }
    if not(ReadProcessMemory(GetCurrentProcess,p,@a,sizeof(a),nil)) then
      exit;
    i:=0;
    doagain:=true;
    while doagain and (i<15) do
      begin
        opcode:=a[i];
        instrlo:=opcode and $f;
        instrhi:=opcode and $f0;
        case instrhi of
          { prefix? }
          $20,$30:
            doagain:=(instrlo and 7)=6;
          $60:
            doagain:=(instrlo and $c)=4;
          $f0:
            doagain:=instrlo in [0,2,3];
          $0:
            begin
              result:=(instrlo=$f) and (a[i+1] in [$d,$18]);
              exit;
            end;
          else
            doagain:=false;
        end;
        inc(i);
      end;
  end;
   
  
  procedure JumpToDllRaiseFunction; {$if FPC_FULLVERSION>=20800} noreturn; {$endif}
  begin
    SysResetFPU;

    Module.PassUnhandledException;

    repeat sleep(1000) until false; //lock up if unthinkable happens
  end;
  
  function GetModuleByAddr(addr: pointer): THandle;
  var
    Tmm: TMemoryBasicInformation;
  begin
    if VirtualQuery(addr, @Tmm, SizeOf(Tmm)) <> sizeof(Tmm)
      then Result:=0
      else Result:= THandle(Tmm.AllocationBase);
  end;

  var sehhhhandler: pointer = nil;

  function MyExceptionHandler(excep : PExceptionPointers) : Longint; stdcall;
  var
    pes: PMotherSehState;
  begin
//dbgsay('* MyExceptionFilter');
//dbgsay('   Exception code = '+ IntToHex(excep^.ExceptionRecord^.ExceptionCode,8)) ;
//dbgsay('   RIP = '+ IntToHex(excep^.ContextRecord^.Rip ,11)) ;
//dbgsay('   Addr = '+ IntToHex(qword(excep^.ExceptionRecord^.ExceptionAddress) ,11)) ;
//dbgsay('   Thread ID=' + IntToHex(GetCurrentTHreadID(), 8));
//dbgsay('   Main thread ID=' + IntToHex(Mother^.State.MainThreadId, 11)) ;

    
    pes:= GetExceptionState(); //per-thread!

//dbgsay('  ' + IntToHex(pes^.StackSegment, 4) + '  ' + IntToHex(excep^.ContextRecord^.SegSs, 4));
    if not Assigned(pes) or (pes^.StackSegment <> excep^.ContextRecord^.SegSs)
      then Exit(EXCEPTION_CONTINUE_SEARCH);
      
    //this case is copypasted from the System unit without understanding it.
	if (cardinal(excep^.ExceptionRecord^.ExceptionCode) = STATUS_ACCESS_VIOLATION)
       { Athlon prefetch bug? }
      and is_prefetch(pointer(excep^.ContextRecord^.rip))
    then begin
      { if yes, then retry }
      excep^.ExceptionRecord^.ExceptionCode := 0;
      Exit(EXCEPTION_CONTINUE_EXECUTION);
    end;    


//dbgsay('   pes='+ IntToHex(qword(pes), 11));
//dbgsay('   pes^.IsModuleThread='+ IntToStr(ord(pes^.IsModuleThread)));
//dbgsay('   pes^.ThreadTitle='+ pes^.ThreadTitle);
//dbgsay('   pes^.ModuleIsAlreadyProcessingUnhandledException='+ IntToStr(ord(pes^.ModuleIsAlreadyProcessingUnhandledException)));
//dbgsay('   Addr belongs to the EXE='+ IntToStr(Ord(Mother^.State.ExeHandle = GetModuleByAddr(pointer(excep^.ContextRecord^.rip)))));
//dbgsay('   Addr belongs to the DLL='+ IntToStr(Ord(Module.DllHandle = GetModuleByAddr(pointer(excep^.ContextRecord^.rip)))));

//save the bloody details
    pes^.ExceptionCode:= excep^.ExceptionRecord^.ExceptionCode;
    pes^.ExceptionAddress:= excep^.ContextRecord^.Rip;


    if cardinal(excep^.ExceptionRecord^.ExceptionCode) = STATUS_ACCESS_VIOLATION then begin
      pes^.AVMemoryOperation:= ptruint(excep^.ExceptionRecord^.ExceptionInformation[0]);
      pes^.AVMemoryAddress:= ptruint(excep^.ExceptionRecord^.ExceptionInformation[1]);
//dbgsay('   AV memory address = '+ IntToHex(pes^.AVMemoryAddress, 11));
//dbgsay('   AV memory operation = '+ IntToStr(pes^.AVMemoryOperation));

    end;

    if
      not Mother^.State.Terminated
      and Assigned(Module) and Assigned(Module.PassUnhandledException) //DLL is loaded
      and not pes^.ModuleIsAlreadyProcessingUnhandledException
      and (
        (
          (Module.DllHandle
            = GetModuleByAddr(pointer(excep^.ContextRecord^.Rip)))
          { Case 1. Exception is in the DLL code. Whatever the thread is,
            pass it to the module. }
        )
       or
        (
          (pes^.IsModuleThread)//ThreadId <> Mother^.State.MainThreadId)
          and (Mother^.State.ExeHandle
            <> GetModuleByAddr(pointer(excep^.ContextRecord^.Rip)))
          { Case 2. The thread belongs to the module. If the address is
            outside the EXE, pass it to the module even if the address
            is outside the DLL as well. Assume it is in some system dll
            the module DLL has called. }
         )
      )
    then begin
//dbgsay('   Passing to module.');
      pes^.ModuleIsAlreadyProcessingUnhandledException:= true;

      //cheat by changing the return address to the function
      //  that calls the module to raise its own language exception
      excep^.ContextRecord^.Rip := ptruint (@JumpToDllRaiseFunction);

      //tell Windows: we have corrected the problem, go on.
      excep^.ExceptionRecord^.ExceptionCode := 0;
      Result := EXCEPTION_CONTINUE_EXECUTION;
    end
    else begin
//dbgsay('  Calling the old handler.');

      Result:= EXCEPTION_CONTINUE_SEARCH;
    end;
  end;  


  procedure InitSehHack;
  begin
    Mother^.State.ExeHandle:= GetModuleByAddr (@InitSehHack);
    Mother^.State.MainThreadId:= GetCurrentThreadId;
    sehhhhandler:= AddVectoredExceptionHandler(1, @MyExceptionHandler)
  end;

  procedure TerminateSehHack;
  begin
    RemoveVectoredExceptionHandler(sehhhhandler);
  end;    

end.
 
