{$codepage utf-8}
{$longstrings on}
{$mode delphi} //NOT mode objfpc as the rest of the engine

{$if FPC_FULLVERSION>=20700}
  {$define che_unicode}    //must match un_typedefs.h
{$endif}

unit cl_windows_missing;

{$define use_dynlibs}

interface
  uses Windows;

{$packrecords C}

{$ifdef cpu32}
  procedure LoadFunctionsUnsupportedInWindows98;
{$endif}

type
  TMonitorInfo = record
    cbSize: DWORD;
    rcMonitor: RECT;
    rcWork: RECT;
    dwFlags: DWORD;
  end;
  PMonitorInfo = ^TMonitorInfo;


const
  MONITOR_DEFAULTTONULL    = $00000000;
  MONITOR_DEFAULTTOPRIMARY = $00000001;
  MONITOR_DEFAULTTONEAREST = $00000002;

  function timeBeginPeriod(x1: UINT): MMRESULT; stdcall; external 'winmm.dll' name 'timeBeginPeriod';
  function timeEndPeriod(x1: UINT): MMRESULT; stdcall;  external 'winmm.dll' name 'timeEndPeriod';

type
  MMRESULT = UINT;
const
  TIMERR_NOERROR = 0;
  TIMERR_NOCANDO = 96 + 1;

//Addendum: W98 support *dropped*. No need for the separation below.
//  Will correct later (it works, so no urgent need to fix)
type
  TDisplayDeviceW = record
    cb:DWORD;
    DeviceName:array[0..31] of UnicodeChar;
    DeviceString:array[0..127] of UnicodeChar;
    StateFlags:DWORD;
    DeviceID:array[0..127] of UnicodeChar;
    DeviceKey:array[0..127] of UnicodeChar;
  end;
  PDisplayDeviceW = ^TDisplayDeviceW;

const
      DISPLAY_DEVICE_ATTACHED_TO_DESKTOP = $00000001;
      DISPLAY_DEVICE_MULTI_DRIVER        = $00000002;
      DISPLAY_DEVICE_PRIMARY_DEVICE      = $00000004;
      DISPLAY_DEVICE_MIRRORING_DRIVER    = $00000008;
      DISPLAY_DEVICE_VGA_COMPATIBLE      = $00000010;
      DISPLAY_DEVICE_REMOVABLE           = $00000020;
      DISPLAY_DEVICE_MODESPRUNED         = $08000000;
      DISPLAY_DEVICE_REMOTE              = $04000000;
      DISPLAY_DEVICE_DISCONNECT          = $02000000;
      DISPLAY_DEVICE_ACTIVE              = $00000001;
      DISPLAY_DEVICE_ATTACHED            = $00000002;

  function EnumDisplayDevicesW(lpDeviceName: Pointer; iDevNum: DWORD;
     lpDisplayDevice: PDisplayDeviceW; dwFlags: DWORD): WINBOOL; stdcall;
     external 'user32' name 'EnumDisplayDevicesW';

  function CreateDCW(_para1:LPCSTR; _para2:LPCSTR; _para3:LPCSTR;
     _para4:pDEVMODE):HDC; stdcall; external 'gdi32' name 'CreateDCW';

{$ifdef cpu64}
  //static linking: 64 bits means Windows XP or higher.
  function MonitorFromRect (const lprc: RECT; dwFlags: DWORD): HMONITOR;
                                                         external 'user32.dll';

  function GetMonitorInfo (hMonitor: HMONITOR; lpmi: PMonitorInfo): BOOL;
                                  external 'user32.dll' name 'GetMonitorInfoA';

  function CreateWindowExW(dwExStyle:DWORD; lpClassName:LPCWSTR;
    lpWindowName:LPCWSTR; dwStyle:DWORD; X:longint;Y:longint;
    nWidth:longint; nHeight:longint;
    hWndParent:HWND; hMenu:HMENU;hInstance:HINST; lpParam:LPVOID):HWND;
                                                         external 'user32.dll';

  function SetWindowTextW(hWnd:HWND; lpString:LPCWSTR):WINBOOL;
                                                         external 'user32.dll';

  function TerminateProcess (hProcess:HANDLE; uExitCode:UINT):WINBOOL;
                                                       external 'kernel32.dll';

  function SHGetFolderPathA  (HWND:hwnd;csidl:dword;htoken:THandle;
                                      dwflags:dword;pszpath:PAnsiChar):HResult;
                                                        external 'shell32.dll';

  {$ifdef che_unicode}
    function SHGetFolderPathW  (HWND:hwnd;csidl:dword;htoken:THandle;
                                      dwflags:dword;pszpath:PWideChar):HResult;
                                                        external 'shell32.dll';
  {$endif}

{$else cpu64}
  //32 bits: must not link statically,
  //  to avoid losing compatibility with vanilla Windows 98 SE!

var
  MonitorFromRect: function (const lprc: RECT; dwFlags: DWORD): HMONITOR; stdcall;
  GetMonitorInfo: function (hMonitor: HMONITOR; lpmi: PMonitorInfo): BOOL; stdcall;

  TerminateProcess: function(hProcess:HANDLE; uExitCode:UINT):WINBOOL; stdcall;


  SHGetFolderPathA: function
//    (p1: pointer; p2: integer; p3: pointer; p4: dword; p5: pointer): HResult; stdcall;
    (HWND:hwnd;csidl:dword;htoken:THandle;dwflags:dword;pszpath:PAnsiChar):HResult; StdCall;

  {$ifdef che_unicode}
    SHGetFolderPathW: function (HWND:hwnd;csidl:dword;htoken:THandle;
                                      dwflags:dword;pszpath:PWideChar):HResult; StdCall;
  {$endif}

  CreateWindowExW: function(dwExStyle:DWORD; lpClassName:LPCWSTR;
    lpWindowName:LPCWSTR; dwStyle:DWORD; X:longint;Y:longint;
    nWidth:longint; nHeight:longint;
    hWndParent:HWND; hMenu:HMENU;hInstance:HINST; lpParam:LPVOID):HWND; stdcall;

  SetWindowTextW: function(hWnd:HWND; lpString:LPCWSTR):WINBOOL; stdcall;
{$endif cpu64}

  const
    SHGFP_TYPE_CURRENT = 0;
    CSIDL_PROGRAM_FILES = $26;
    CSIDL_LOCAL_APPDATA = $1c;
    CSIDL_APPDATA = $1a;
    CSIDL_FLAG_NO_ALIAS = $1000;
    CSIDL_FLAG_CREATE = $8000; { (force creation of requested folder if it doesn't exist yet)     }

implementation
{$ifdef cpu32}
  uses cge;

  function LoadProcAddress(hdll: THandle; ProcName: ansistring) : pointer;
  begin
    Result:= GetProcAddress(hdll, PAnsiChar(ProcName));
    if Assigned(Result) then begin
      VerboseLog('    %0() at %1 in %2', [ProcName, Result, GetExeFileName(Result)]);
    end
    else
      VerboseLog('    %0() not found!', [ProcName]);
  end;

  procedure LoadFunctionsUnsupportedInWindows98;
  var
    shell32, kernel32, user32: THandle;
  begin

    //lol this is called when the log and config aren't yet initialized

    //use *cached* verbose log (i.e. it is cached anyway but discarded later
    //  if there is no verbose option set in the config
    VerboseLog('Dynamically loading Win32 API functions to ensure compatibility with vanilla Windows 98 SE...');

    VerboseLog('  loading shell32.dll...');
    shell32:= LoadLibrary('shell32.dll');

    @SHGetFolderPathA:= LoadProcAddress(shell32, 'SHGetFolderPathA');
    Mother^.State.Windows98.DoesNotSupportUserDirectory:= not (
      Assigned(SHGetFolderPathA)
    );

    {$ifdef che_unicode}
      @SHGetFolderPathW:= LoadProcAddress(shell32, 'SHGetFolderPathW');
    {$endif}


//    FreeLibrary(shell32);

    VerboseLog('  loading kernel32.dll...');
    kernel32:= LoadLibrary('kernel32.dll');

    @TerminateProcess:= LoadProcAddress(kernel32, 'TerminateProcess');

//    FreeLibrary(kernel32);

    VerboseLog('  loading user32.dll...');
    user32:= LoadLibrary('user32.dll');

    @MonitorFromRect:= LoadProcAddress(user32, 'MonitorFromRect');
    @GetMonitorInfo:= LoadProcAddress(user32, 'GetMonitorInfoA');
    @CreateWindowExW:= LoadProcAddress(user32, 'CreateWindowExW');
    @SetWindowTextW:= LoadProcAddress(user32, 'SetWindowTextW');

//    pointer():= LoadProcAddress(user32, '');
//    pointer():= LoadProcAddress(user32, '');
//    pointer():= LoadProcAddress(user32, '');

//    FreeLibrary(user32);

  end;
{$endif}


end.
