{
    This file is part of Chentrah,
    Copyright (C) 2004-2015 Anton Rzheshevski (chebmaster@mail.ru).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see http://www.gnu.org/licenses/

 **********************************************************************

    This file contains the framework manager class header.

 **********************************************************************}

{
    Initially, the engine structure incorporated direct ties to low-level
    system functions: Win32 API for Windows and X11 / system calls for Linux.
    Later, I decided to incapsulate all this functionality in a class,
    so it would be simple to switch to a specific framework
    (GTK+, Ogre, KDE - whatever)

    So, "framework" in this case encompasses everything that deals with
    the application window, input, sound and so on.
    For example, the current (and only, as of 2014) WinApi framework
    is utterly unsuitable for working with multiple monitors.
    And it will stay that way. If I get to supporting OculusRift (I hope),
    that would be an independent, completely separate framework.

    Moreover, a server instance would use a stub framework called "Server"
    that does nothing.

    Note: OpenGL is mostly outside of this, as the only exception.
    The game module knows OpenGL and works with it directly.

    Any sound APIs, not so. The module mixes and applies any effects
    by rendering sounds via GLES shaders. The result is then fed to the
    NewSoundData() method.

}
type
  TWindowStateCommandedByTheOs = (wsco_Normal, wsco_NoFocus, wsco_Minimized, wsco_Maximized);



  TAbstractFramework = class;
  CAbstractFramework = class of TAbstractFramework;
  TAbstractFramework = class
  protected
    class var AvailableFrameworks: array of CAbstractFramework;
    class var AvailableFrameworkNames: array of UnicodeString;

    procedure HBeat; virtual; abstract;
    procedure ProcessNetworkTraffic; virtual; abstract;

    function DetailsToLog: UnicodeString; virtual; //must begin with newline

  public

    class procedure RegisterFramework(C: CAbstractFramework; name: UnicodeString);

    //to be used by main() from cl_main.inc only!
    //raises an error if requested framework not found, and resets the config to use default
    class function GetFrameworkByName(name: UnicodeString): CAbstractFramework;

    constructor Create; virtual;

     //tailored for SDL. By default returns nil meaning 'load normally'.
    function GetOpenGLProcAddress(ProcName: ansistring): Pointer; virtual;

    //must point to a 32x32 ABGR pixels array.
    //NIL causes to hide the pointer.
    procedure SetPointerImage(p: pointer; xhs, yhs: integer); virtual; abstract;

    // performs OpenGL's SwapBuffers with timing measurements
    procedure SwapBuffers;  virtual; abstract;

    procedure SetWindowName(n: PUnicodeChar); virtual; abstract;

    function MouseX: float; virtual; abstract;

    function MouseY: float; virtual; abstract;

    Procedure MainThreadMainLoop; virtual; abstract;

    procedure SetVolume(volume: float); virtual; abstract;

    procedure NewSoundData(begins: TDateTime; samples, rate: integer; data: pointer); virtual; abstract;

    procedure PlaySound(samples, rate: integer; data: pointer); virtual; abstract;

    procedure PlayWav(filename: TFileNameString); virtual; abstract;

    //needed for limiting rendering settings when laptop runs on battery
    //must fill Mother^.State.BatteryRemaining according to Windows standard:
    //0..100 mean battery percent remaining, 255 means on AC power or unknown
    //frameworks that do not support this must set
    //Mother^.State.BatteryRemaining:= 255 at initialization
    //because Chentrah auto-terminates at low battery power!
    procedure CheckBatteryStatus; virtual; abstract;

    procedure YellAndExpire(y: UnicodeString); virtual; abstract;
  end;



