{
    This file is part of Chentrah,
    Copyright (C) 2004-2016 Anton Rzheshevski (chebmaster@mail.ru).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see http://www.gnu.org/licenses/

 **********************************************************************

    This file contains the basics of frameworks

 **********************************************************************}



  //This one is unique in that it is called from corresponding unit's
  //  initialization section, long before CGE is initialized
  class procedure TAbstractFramework.RegisterFramework(C: CAbstractFramework; name: UnicodeString);
  var i: integer;
  begin
    i:= Length(AvailableFrameworks);
    SetLength(AvailableFrameworks, i + 1);
    SetLength(AvailableFrameworkNames, i + 1);
    AvailableFrameworks[i]:= C;
    AvailableFrameworkNames[i]:= Trim(LowerCase(name));
  end;

  //This one is called from main()
  class function TAbstractFramework.GetFrameworkByName(name: UnicodeString): CAbstractFramework;
  var i: integer;
  begin
    For i:=0 to High(AvailableFrameworks) do
      if AvailableFrameworkNames[i] = Trim(LowerCase(name))
        then Exit(AvailableFrameworks[i]);
    if Assigned(Config) then Config.Str['framework_' + SystemSuffix, 'used']:= AvailableFrameworkNames[0];
    GiveWarning(
      'unknown_framework_' + {$I %FPCTARGETOS%} + '_' + {$I %FPCTARGETCPU%} + '_' + name,
      RuEn('Неизвестный фреймворк «' + name + '»',
           'Unknown framework "' + name + '"'),
      RuEn(
        'Текущая версия поддерживает следующие фреймворки: %0.'
        + ' Настройки были сброшены на использование фреймворка по умолчанию, %1.',
       'The current version supports following frameworks: %0.'
        + 'The settings were reset to default, %1.'
      ), [name, Implode(', ', AvailableFrameworkNames), AvailableFrameworkNames[0]]);
    Result:= AvailableFrameworks[0];
  end;


  constructor TAbstractFramework.Create;
  var
    cn, n: UnicodeString;
  begin
    inherited Create;

    cn:= Self.ClassName;
    if cn = 'TServerFramework' then n:= RuEn(
      'серверный (без графики)',
      'server (no graphics)')
    else if cn = 'TSDL2Framework' then n:= RuEn(
      'на библиотеке SDL2',
      'using SDL2 library')
    else if cn = 'TX11Framework' then n:= RuEn(
      'нативный X11',
      'native X11')
    else if cn = 'TWinApiFramework' then n:= RuEn(
      'нативный Windows API',
      'native Windows API')
    else n:= cn;
    AddLog(RuEn(
      'Фреймворк: %0%1',
      'Framework: %0%1'),
      [n, Self.DetailsToLog()]);

    cge.Framework:= Self;
  end;

  function TAbstractFramework.DetailsToLog: UnicodeString;
  begin
    Result:= '';
  end;

  function TAbstractFramework.GetOpenGLProcAddress(ProcName: ansistring): Pointer;
  begin
    Result:= nil;
  end;

