{
    This file is part of Chentrah,
    Copyright (C) 2004-2016 Anton Rzheshevski (chebmaster@mail.ru).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see http://www.gnu.org/licenses/

 **********************************************************************

    This units defines a common base of the X11 and WinAPI frameworks.

 **********************************************************************}

{$longstrings on}
{$codepage utf-8}
{$include un_globaldefs.h}

unit framework_native;
interface
  uses SysUtils, classes
    {$ifdef windows}
      , windows
      , cl_windows_missing
    {$endif}


    , ctypes, math, cge, un_strings

    {$ifdef windows}
     , jedi_directsound8
    {$endif}

    , framework_basic

  ;

  {$include framework_native_sound.h}

  {$include framework_native_gamepad.h}


Const
  MaxKey = 255;
  MinimumPaddingToScreenEdges = 100;

type

  TNativeFramework = class (TBasicFramework)
  protected
    GLInitialized: boolean;

    minW, minH: integer;
    //InitialWidth, InitialHeight: integer;
//    HasFocus: Boolean;

    gcMx, gcMy, gcMxV, gcMyV, gcMxAcc, gcMyAcc : float;

    procedure InitDisplay; virtual; abstract;

    procedure InitWindowSubsystem;
    procedure ShutdownWindowSubsystem;

    procedure CreateWindow; virtual; abstract;
    procedure InitGLFirstStage; virtual; abstract; //choose visual
    procedure InitGL; virtual; abstract;
    procedure CloseGL; virtual; abstract;
    procedure CloseWindow; virtual;  abstract;
    procedure _UploadPointer; virtual; abstract;
    procedure _CheckCursorMode; virtual;

    procedure EmulateMouseViaGamepad; override;

  public

    GamepadManager: TGamepadManager;
    SoundManager: TMotherSoundMan;
    {$ifdef windows}
     WindowHandle: HWND;
    {$endif}
    {$ifdef unix}
     WindowHandle: GLXDrawable;
    {$endif}
    eDisplay: EGLDisplay;
    eConfig: EGLConfig;
    eSurface: EGLSurface;
    eContext: EGLContext;

    //implementing abstract functions that make up the framework API
    procedure SetVolume(volume: float); override;
    procedure NewSoundData(begins: TDateTime; samples, rate: integer; data: pointer); override;
    procedure PlaySound(samples, rate: integer; data: pointer); override;
    procedure PlayWav(filename: TFileNameString); override;
    procedure SetPointerImage(p: pointer; xhs, yhs: integer); override;

    //service functions shared by both
    procedure AddRightThumbToMouseMovement(x, y: float);

  end;


implementation



  procedure TNativeFramework.SetVolume(volume: float);
  begin
    if Assigned(SoundManager) then SoundManager.setVolume(volume);
  end;


  procedure TNativeFramework.NewSoundData(begins: TDateTime; samples, rate: integer; data: pointer);
  begin
    if Assigned(SoundManager) then SoundManager.NewData(begins, samples, rate, data);
  end;


  procedure TNativeFramework.PlaySound(samples, rate: integer; data: pointer);
  begin
    if Assigned(SoundManager) then SoundManager.PlaySound(samples, rate, data);
  end;


  procedure TNativeFramework.PlayWav(filename: TFileNameString);
  begin
    if Assigned(SoundManager) then SoundManager.PlayWav(filename);
  end;



  procedure TNativeFramework.AddRightThumbToMouseMovement(x,y: float);
  var
    a, accel, nx, ny
      //, decel
      : float;
  begin
    a:= sqrt(sqr(x) + sqr(y));
    if (a > 1) then begin
      x:= x / a;
      y:= y / a;
      a:= 1.0;
    end;
    nx:= x * Mother^.Gamepad.MouseSpeedMultiplier;
    ny:= y * Mother^.Gamepad.MouseSpeedMultiplier;
    gcMxV+= (nx - gcMx);
    gcMyV+= (ny - gcMy);
    gcMx:= nx;
    gcMy:= ny;
    if (a > (Mother^.Gamepad.MouseAccelThreshold / 100)) then begin
      accel:= sqr((a - (Mother^.Gamepad.MouseAccelThreshold / 100)) / (1.0 - (Mother^.Gamepad.MouseAccelThreshold / 100)));
      gcMxAcc:= accel * x/a * Mother^.Gamepad.MouseAccelMultiplier;
      gcMyAcc:= accel * y/a * Mother^.Gamepad.MouseAccelMultiplier;
    end else begin
      gcMxAcc:= 0;
      gcMyAcc:= 0;
      gcMxV:= gcMx;
      gcMyV:= gcMy;
    end;
  end;

  procedure TNativeFramework.EmulateMouseViaGamepad;
  var
    delta: double;
    ne: integer;
  begin
    delta:= Math.max(0.0, Math.min(0.1, UsecDelta(@PreviousCcToMTsc) / 1000000.0));
    gcMxV+= delta * gcMxAcc;
    gcMyV+= delta * gcMyAcc;
    if f_pointerGrabbed then begin
      CombinedMx:= 0; //only delta
      CombinedMy:= 0;
    end;
    CombinedMx+= delta * gcMxV;
    CombinedMy-= delta * gcMyV;
    if not f_pointerGrabbed then begin
      CombinedMx:= Math.max(0, Math.min(CombinedMx, Mother^.Display.WindowClientRect.Width));
      CombinedMy:= Math.max(0, Math.min(CombinedMy, Mother^.Display.WindowClientRect.Height));
    end;
    if (round(CombinedMx) <> gcMxPrevi) or (round(CombinedMy) <> gcMyPrevi) then begin
      gcMxPrevi:= round(CombinedMx);
      gcMyPrevi:= round(CombinedMy);
      //Mother^.Display.MouseIsBeyondOurWindow:= false;
      MouseIsGamepad:= true;
      ne:= NewInputEvent();
      with Mother^.Input.Events[ne] do begin
        if f_pointerGrabbed
          then EventType:= CGM_MOUSEMOVE
          else EventType:= CGM_MOUSEDELTA;
        mx:= CombinedMx;
        my:= CombinedMy;
      end;
    end;

    if Assigned(GamepadManager)
      then GamepadManager.Pulse;
  end;








  procedure TNativeFramework.InitWindowSubsystem;
  begin

    //init some global variables

    f_PointerVisible:= Yes;
    Mother^.Display.InFullScreenMode:= Config.Bool['video', 'fullscreen']; // and not Mother^.State.DeveloperMode;
    if Mother^.Display.InFullScreenMode
      then f_wsco:= wsco_Maximized
      else f_wsco:= wsco_Normal;
    f_state:= f_wsco;


    InitDisplay;

    Mother^.Display.ScreenRect:= GetScreenRect(Mother^.Display.InFullScreenMode);
    formerScreenRect:= Mother^.Display.ScreenRect;

    LoadWindowRectFromConfig;

    InitGLFirstStage;

    CreateWindow;

    if Mother^.Display.InFullScreenMode
      then f_wsco:= wsco_Maximized
      else f_wsco:= wsco_Normal;

    f_focus:=True;
    Mother^.Display.WindowVisible:= True;

    InitGL;

    if Mother^.Debug.Enabled then begin
      AddLog('OpenGL extensions: ' + PAnsiChar(glGetString(GL_EXTENSIONS)));
     {$ifdef unix}
{ //not necessary
      if Assigned(glXQueryExtensionsString)
        then AddLog('GLX extensions: ' + PAnsiChar(glXQueryExtensionsString(Display, Screen)));
}
     {$endif}
    end;
    f_window_subsystem_initialized:= Yes;
  end;


  procedure TNativeFramework.ShutdownWindowSubsystem;
  begin
    f_window_subsystem_initialized:= No;

    SaveQF;
    CloseGL;
    CloseWindow;
  end;

  procedure TNativeFramework._CheckCursorMode();
  begin
    if Mother^.Display.UseHardware1bitCursor <> f_prevUseHwCursor
    then begin
      f_prevUseHwCursor:= Mother^.Display.UseHardware1bitCursor;
      _UploadPointer;
    end;
  end;


  {$ifdef windows}
   procedure FlipBytes(b: Pbyte; size: integer);
   var
     i, j: integer;
     a: byte;
   begin
     For i:= 1 to Size do begin
       a:=0;
       for j:=0 to 7 do
         if (b^ and (1 shl j)) <> 0 then a:= a or (1 shl (7-j));
       b^:= a;
       inc(b);
     end;
   end;
  {$endif}

  procedure TNativeFramework.SetPointerImage(p: pointer; xhs, yhs: integer);
  var
    x, y, j: integer;
    wA, wL: dword;
    averageL, averageA, averageV, z: float;
  const
    lkoeff: array[0..2] of float = (0.3, 0.5, 0.2);
    ditherfactor: array [0..3, 0..3] of float = (
      ( 0.00, 0.45, 0.10,  0.33),
      ( 0.75, 1.00, 0.66,  0.90),
      ( 0.10, 0.33, 0.00,  0.45),
      ( 0.66, 0.90, 0.75,  1.00));
  begin
//***TODO optimize by comparing to the currently loaded image!
    if Assigned(p) then with PointerBitmap do begin
      Move(p^, Bitmap, 32 * 32 * 4);

      if (xhs <0) or (yhs <0) or (xhs> 31) or (yhs > 31) then Die(
        MI_ERROR_PROGRAMMER_NO_BAKA, ['Invalid mouse pointer hot spot coordinates ('
         +IntToStr(xhs)+':'+IntToStr(yhs)+')']);

      // now compose the two 1-bit masks
      averageL:= 0;
      averageV:= 0;
      for y:= 0 to 31 do
        for x:= 0 to 31 do begin
          for j:=0 to 2 do begin
            averageL += Bitmap[y, x, j] * lkoeff[j] * (Bitmap[y, x, 3]/255) / (32*32);
          end;
          averageV += Bitmap[y, x, 3] / (32*32*255)
        end;
      if averageV < 0.01 then averageV:= 0.01;
      averageL:= averageL / averageV;
      averageA:= 0;
      for y:= 0 to 31 do
        for x:= 0 to 31 do
          averageA += Bitmap[y, x, 3] / (32*32);
      for y:= 0 to 31 do begin
        wA:=0;
        wL:=0;
        for x:= 31 downto 0 do begin
          wA:= wA shl 1;
          wL:= wL shl 1;
          if Bitmap[y, x, 3] > averageA * HW_POINTER_ALPHA_THRESHOLD then begin
            wA:= wA or 1;
            z:= 0.0;
            for j:=0 to 2
              do z+= Bitmap[y, x, j] * lkoeff[j]  * (Bitmap[y, x, 3]/255);
            if z > (averageL * ( 1 + (ditherfactor[x and $3, y and $3]-0.5) * HW_POINTER_DITHER_FACTOR))
              then wL:= wL or 1;
          end;
        end;
        XorMask[y]:= wL;
        AndMask[y]:= not wA;
        {$ifdef windows}
          FlipBytes(@XorMask[y], 4);
          FlipBytes(@AndMask[y], 4);
        {$endif}
      end;
      PointerXHotSpot:= xhs;
      PointerYHotSpot:= yhs;
      f_PointerVisible:= Yes;
      MOVE(PointerBitmap, PointerCache, sizeof(PointerBitmap));
    end
    else begin
      With PointerBitmap do begin
        FillChar(Bitmap, sizeof(Bitmap), 0);
        FillChar(Andmask, sizeof(AndMask), $ff);
        FillChar(XorMask, sizeof(XorMask), 0);
      end;
      //PointerXHotSpot:= 1;
      //PointerYHotSpot:= 1;
      f_PointerVisible:= No;
    end;
    _UploadPointer;
  end;


    {$include framework_native_sound.inc}

  {$include framework_native_gamepad.inc}

end.
