

  procedure CheckHr(hr: HRESULT; n: UnicodeString);
  begin
    if hr <> DS_OK then Die('%0() returned %1', [n, DSErrorString(hr)]);
  end;


  constructor TDirectXSoundMan.Create;
  var
    hr: HRESULT;
    i: integer;
    bd: TDSBufferDesc;
    wf: TWaveFormatEx;
  begin
    try
      resampleFreq:= Config.Int['sound','directsound_force_resample'];
      InitDirectSound;
//      AddLog('DirectSound: ')
      hr:= DirectSoundCreate8(NIL, f_device, NIL);  CheckHr(hr, 'DirectSoundCreate8');
//      DirectSoundDevice.GetCaps(
      if Mother^.Debug.Verbose then AddLog('  Default device created.');
(* there is no need for this.
      hr:=DirectSoundDevice.SetCooperativeLevel(WindowManager.WindowHandle, DSSCL_PRIORITY);
      CheckHr(hr, 'DirectSoundDevice.SetCooperativeLevel');
      if Mother^.State.DebugMode then AddLog('  Cooperative level set to DSSCL_PRIORITY.');*)


      hr:= f_device.SetCooperativeLevel((Cge.Framework as TNativeFramework).WindowHandle, DSSCL_NORMAL);
      CheckHr(hr, 'IDirectSound8.SetCooperativeLevel');
      if Mother^.Debug.Verbose then AddLog('  Cooperative level set to DSSCL_NORMAL.');

      //Creating the stereo buffers for tripl buffering.
      with wf do begin
        wFormatTag:= 1; //WAVE_FORMAT_PCM;       { format type }
        nChannels:= 2;        { number of channels (i.e. mono, stereo, etc.) }
        nSamplesPerSec:= 44100;  { sample rate }
        nAvgBytesPerSec:= 44100 * 4; { for buffer estimation }
        nBlockAlign:= 4;      { block size of data }
        wBitsPerSample:= 16;   { number of bits per sample of mono data }
        cbSize:=0;           { the count in bytes of the size of }
      end;
      with bd do begin
        dwSize:= sizeof(TDSBufferDesc);
        dwFlags:= 0;  //specify hardware?
        dwBufferBytes:= snd_buffer_size*4;
        dwReserved:= 0;
        lpwfxFormat:= @wf;
//        guid3DAlgorithm:=
      end;
      for i:=0 to snd_buffers - 1 do begin
        hr:= f_device.CreateSoundBuffer(bd, f_buffer[i], NIL);
        CheckHr(hr, 'IDirectSound8.CreateSoundBuffer');
      end;

      //Play the intro sound;
      TestGong;
    except
      Die(MI_ERROR_SOUND_INIT);
    end;
    AddLog('DirectSound 8 initialized.');
  end;


  destructor TDirectXSoundMan.Destroy;
  var i: integer;
  begin
    f_b_buffer:= nil; //takes care of calling Release(). Pascal syntax is strange here.
    for i:=0 to snd_buffers - 1 do f_buffer[i]:= nil;
    DirectSoundDevice:= nil;
    CloseDirectSound;
  end;


  procedure TDirectXSoundMan.SetVolume(volume: float);
  begin

  end;

  procedure TDirectXSoundMan.NewData(begins: TDateTime; samples, rate: integer; data: pointer);
  var
    bd: TDSBufferDesc;
  begin
    if f_muted then Enable;

   //lock the next buffer
   //fill the next buffer
   //unlock the next buffer

   //get current buffer playing position and calculate the next buffer start position
   //stop current buffer
   //start the next buffer

   //cycle the current buffer id

    f_GotNewData:= Yes;
  end;

  procedure TDirectXSoundMan.PlaySound(samples, rate: integer; data: pointer);
  var
    bd: TDSBufferDesc;
    hr: HResult;
    wf: TWaveFormatEx;
    p1, p2: pointer;
    b1, b2: dword;
    dst_samples: integer;
i: integer;
  begin
    dst_samples:= samples;
    try
      if Assigned(f_b_buffer) then begin
        hr:= f_b_buffer.Stop; CheckHr(hr, 'IDirectSoundBuffer.Stop');
        f_b_buffer:= nil;
      end;
      with wf do begin
        wFormatTag:= 1; //WAVE_FORMAT_PCM;       { format type }
        nChannels:= 1;        { number of channels (i.e. mono, stereo, etc.) }
        if resampleFreq > 0 then begin
          nSamplesPerSec:= resampleFreq;
          nAvgBytesPerSec:= resampleFreq * 2;
          dst_samples:= Trunc((1.0 * samples * resampleFreq) / rate);
        end else begin
          nSamplesPerSec:= rate;  { sample rate }
          nAvgBytesPerSec:= rate * 2; { for buffer estimation }
        end;
        nBlockAlign:=2;      { block size of data }
        wBitsPerSample:= 16;   { number of bits per sample of mono data }
        cbSize:=0;           { the count in bytes of the size of }
      end;
      with bd do begin
        dwSize:= sizeof(TDSBufferDesc);
        dwFlags:=  {DSBCAPS_CTRLDEFAULT or} DSBCAPS_LOCDEFER;
        dwBufferBytes:= dst_samples*2;
        dwReserved:= 0;
        lpwfxFormat:= @wf;
//        guid3DAlgorithm:=
      end;
      hr:= f_device.CreateSoundBuffer(bd, f_b_buffer, NIL);    CheckHr(hr, 'IDirectSound8.CreateSoundBuffer');
      hr:= f_b_buffer.Lock(0, 0, @p1, @b1, @p2, @b2, DSBLOCK_ENTIREBUFFER);  CheckHr(hr, 'IDirectSoundBuffer.Lock');
      if resampleFreq > 0
        then _Resample(data, p1, samples, min(b1 div 2, dst_samples))
        else Move(data^, p1^, min(b1, samples*2));
//for i:=0 to dst_samples - 1 do AddLog('-- %0',[SmallInt((p1 + i*2)^)]);
      hr:= f_b_buffer.Unlock(p1, b1, p2, b2); CheckHr(hr, 'IDirectSoundBuffer.Unlock');
      hr:= f_b_buffer.Play(0, 100, DSBPLAY_LOCSOFTWARE); CheckHr(hr, 'IDirectSoundBuffer.Play');
    except
      AddLog('TDirectXSoundMan.PlaySound() failed: '#10#13'  %0', [StopDying()]);
    end;
  end;

  procedure TDirectXSoundMan.Cycle;
  begin
    if not f_GotNewData then begin
      //stop the current buffer if there weren't any new data in one second or if (Mother^.Module.State <> msLoaded)
    end;
    f_GotNewData:= No;
  end;

  procedure TDirectXSoundMan.Disable;
  begin
    if f_muted then Exit;

  end;

  procedure TDirectXSoundMan.Enable;
  begin
    if not f_muted then Exit;

  end;








  function NoDsHint: TMessageId;
  begin
    if Mother^.State.OS in WindowsEmulators
      then Result:= MI_HINT_NO_DSOUND8_WINE
      else
        if Mother^.State.OS in WindowsShouldHaveBuiltInOpenAL
          then Result:= MI_HINT_NO_DSOUND8_VISTAANDUP
          else Result:= MI_HINT_NO_DSOUND8_XPANDBELOW;
  end;


  function GetDSProcAddress(ProcName: ansistring): Pointer;
  var
    baseaddr: pointer;
    exename: UnicodeString;
  begin
    Result:= GetProcAddress(DirectSoundDll, PAnsiChar(ProcName));
    if Assigned(Result) and Mother^.Debug.Verbose then begin
      AddLog('  %0() at %1 in %2', [ProcName, Result, GetExeFileName(Result)]);
    end;
    if not Assigned(Result)
      then Die(MI_ERROR_DIRECTSOUND8_FUNCTION_NOT_FOUND,
        [ProcName + '()', DirectSoundDllName, MessageContainer[NoDsHint()]]);
  end;

 Procedure InitDirectSound;
    function LoadDSLibrary: boolean;
    begin
      DirectSoundDllName:= config.path['DirectX','DirectSound8dll'];
      if DirectSoundDllName = '' then DirectSoundDllName:= 'NULL';
      if Mother^.Debug.Verbose then AddLog('  Loading %0...', [DirectSoundDllName]);
      DirectSoundDll:=
              {$ifdef che_unicode}
                LoadLibraryW
              {$else}
                LoadLibrary
              {$endif}
                  (PFileNameChar(DirectSoundDllName));
      Result:= DirectSoundDll <> 0;
      if Mother^.Debug.Verbose then
        if not result then AddLogComment('FAILED.')
                      else AddLogComment('Ok.');
    end;
  begin
    if Mother^.Debug.Enabled then AddLog('Init DirectSound 8...');
    if not LoadDsLibrary()
      then Die(MI_ERROR_DIRECTSOUND8NOTFOUND,
                 [DirectSoundDllName, MessageContainer[NoDsHint()]]);

    if Mother^.Debug.Verbose then AddLog('  Loading the procedure addresses from the DLL...');
    //DirectSoundCreate:= GetDSProcAddress('');
    pointer(DirectSoundEnumerateW):= GetDSProcAddress('DirectSoundEnumerateW');
    //DirectSoundEnumerateA:= GetDSProcAddress('');
//    DirectSoundCaptureEnumerateW:=
    //DirectSoundCaptureEnumerateA:=
    pointer(DirectSoundCreate8):= GetDSProcAddress('DirectSoundCreate8');
    pointer(GetDeviceID):= GetDSProcAddress('GetDeviceID');
  end;

  Procedure CloseDirectSound;
  begin
    If DirectSoundDll <> 0 then FreeLibrary(DirectSoundDll);
  end;

  function DSErrorString(h: HRESULT): UnicodeString;
  begin
    case h of
      DS_OK, DS_NO_VIRTUALIZATION: Result:='';
      //cl_directsound_init.inc(68,22) Error: range check error while evaluating constants
      DSERR_ALLOCATED: Result:='DSERR_ALLOCATED: '#10#13'  The call failed because resources (such as a priority level)'#10#13'  were already being used by another caller';
      DSERR_CONTROLUNAVAIL: Result:='DSERR_CONTROLUNAVAIL: '#10#13'  The control (vol, pan, etc.) requested by the caller is not available';
      DSERR_INVALIDPARAM: Result:='DSERR_INVALIDPARAM: '#10#13'  An invalid parameter was passed to the returning function';
      DSERR_INVALIDCALL: Result:='DSERR_INVALIDCALL: '#10#13'  This call is not valid for the current state of this object';
      DSERR_GENERIC: Result:='DSERR_GENERIC: '#10#13'  An undetermined error occurred inside the DirectSound subsystem';
      DSERR_PRIOLEVELNEEDED: Result:='DSERR_PRIOLEVELNEEDED: '#10#13'  The caller does not have the priority level required for the function to'#10#13'  succeed';
      DSERR_OUTOFMEMORY: Result:='DSERR_OUTOFMEMORY: '#10#13'  Not enough free memory is available to complete the operation';
      DSERR_BADFORMAT: Result:='DSERR_BADFORMAT: '#10#13'  The specified WAVE format is not supported';
      DSERR_UNSUPPORTED: Result:='DSERR_UNSUPPORTED: '#10#13'  The function called is not supported at this time';
      DSERR_NODRIVER: Result:='DSERR_NODRIVER: '#10#13'  No sound driver is available for use';
      DSERR_ALREADYINITIALIZED: Result:='DSERR_ALREADYINITIALIZED: '#10#13'  This object is already initialized';
      DSERR_NOAGGREGATION: Result:='DSERR_NOAGGREGATION: '#10#13'  This object does not support aggregation';
      DSERR_BUFFERLOST: Result:='DSERR_BUFFERLOST: '#10#13'  The buffer memory has been lost, and must be restored';
      DSERR_OTHERAPPHASPRIO: Result:='DSERR_OTHERAPPHASPRIO: '#10#13'  Another app has a higher priority level, preventing this call from'#10#13'  succeeding';
      DSERR_UNINITIALIZED: Result:='DSERR_UNINITIALIZED: '#10#13' This object has not been initialized';
      DSERR_NOINTERFACE: Result:='DSERR_NOINTERFACE: '#10#13'  The requested COM interface is not available';
      DSERR_ACCESSDENIED: Result:='DSERR_ACCESSDENIED: '#10#13'  Access is denied';
      DSERR_BUFFERTOOSMALL: Result:='DSERR_BUFFERTOOSMALL: '#10#13'  Tried to create a DSBCAPS_CTRLFX buffer shorter than DSBSIZE_FX_MIN milliseconds';
      DSERR_DS8_REQUIRED: Result:='DSERR_DS8_REQUIRED: '#10#13'  Attempt to use DirectSound 8 functionality on an older DirectSound object';
      DSERR_SENDLOOP: Result:='DSERR_SENDLOOP: '#10#13'  A circular loop of send effects was detected';
      DSERR_BADSENDBUFFERGUID: Result:='DSERR_BADSENDBUFFERGUID: '#10#13'  The GUID specified in an audiopath file does not match a valid MIXIN buffer';
      DSERR_OBJECTNOTFOUND: Result:='DSERR_OBJECTNOTFOUND: '#10#13'  The object requested was not found';
      DSERR_FXUNAVAILABLE: Result:='DSERR_FXUNAVAILABLE: '#10#13'The effects requested could not be found on the system, or they were found'#10#13'  but in the wrong order, or in the wrong hardware/software locations.';
    else
      Result:= RuEn('Unknown error code ','Неизвестный код ошибки ') + IntToHex(h, 8) + 'h';
    end;
  end;




