{
    This file is part of Chentrah,
    Copyright (C) 2004-2015 Anton Rzheshevski (chebmaster@mail.ru).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see http://www.gnu.org/licenses/

 **********************************************************************}
{$longstrings on}
{$codepage utf-8}
{$include un_globaldefs.h}

unit framework_sdl2; // For both Windows and Linux

interface

uses SysUtils, classes, math, cge,
  SDL2, // resides at ../../../3rdparty/sdl2-fpc-master/SDL2/
        // path must be passed to compiler
  framework_basic;


type
  TSDL2Framework = class (TBasicFramework)
  protected
    procedure CrashAndBurn;
    function GetScreenRect(fullscreen: boolean): TWindowManagerRect; //if not fullscreen then return the working area (excluding the taskbar)
//    function GetWindowOuterRect(): TWindowManagerRect;
//    function GetWindowClientRect(): TWindowManagerRect;
  public
    Window: PSDL_Window;
    Context: SDL_GLContext;
    dispaly_index: longint; //a stub for future. Curently always 0.
    constructor Create; override;
    destructor Destroy; override;
    procedure SetPointerImage(p: pointer; xhs, yhs: integer); override;
    procedure SwapBuffers; override;
    function MouseX: float; override;
    function MouseY: float; override;
    procedure SetVolume(volume: float); override;
    procedure NewSoundData(begins: TDateTime; samples, rate: integer; data: pointer); override;
    procedure PlaySound(samples, rate: integer; data: pointer); override;
    procedure PlayWav(filename: TFileNameString); override;
    procedure CheckBatteryStatus; override;
    procedure SetWindowName(n: PUnicodeChar); override;
    function GetOpenGLProcAddress(ProcName: ansistring): Pointer; override;
  end;


implementation

  constructor TSDL2Framework.Create;
  var
    flags: Uint32;
    use_rect: TWindowManagerRect;
    i: integer;
  begin
    inherited;
    Cge.Framework:= Self; //other classes need to access it.

    if Mother^.Debug.Enabled then AddLog('Initializing SDL2...');

    //If the DLL is missing, Windows shows this error message box..

    //Запуск программы невозможен, так как на компьютере
    //отсутствует SDL2.dll. Попробуйте переустановть программу.

    // ..and kils our application. No save.

    if 0 <> SDL_Init(SDL_INIT_VIDEO) then CrashAndBurn;

    SDL_GL_SetAttribute( SDL_GL_CONTEXT_MAJOR_VERSION, 2 );
    SDL_GL_SetAttribute( SDL_GL_CONTEXT_MINOR_VERSION, 1 );
    SDL_GL_SetAttribute( SDL_GL_CONTEXT_PROFILE_MASK, SDL_GL_CONTEXT_PROFILE_COMPATIBILITY);
    SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, 1 );
    SDL_GL_SetAttribute( SDL_GL_ACCELERATED_VISUAL, 1 );

    if 0 <> SDL_GL_LoadLibrary(nil) then CrashAndBurn;

    //Create window
    Mother^.Display.InFullScreenMode:= Config.Bool['video', 'fullscreen'];

    Mother^.Display.ScreenRect:= GetScreenRect(Mother^.Display.InFullScreenMode);
    formerScreenRect:= Mother^.Display.ScreenRect;

    LoadWindowRectFromConfig;

    flags:= SDL_WINDOW_OPENGL or SDL_WINDOW_SHOWN ;

    if Mother^.Display.InFullScreenMode then begin
      flags:= flags or SDL_WINDOW_FULLSCREEN;
      use_rect:= Mother^.Display.ScreenRect;
    end
    else
      use_rect:= Mother^.Display.WindowRect;

    with use_rect do Window:= SDL_CreateWindow(
      PAnsiChar(WideToAnsi(MessageContainer[MI_CGE_TITLE])),
      left, top, width, height, flags);

    if not Assigned(Window) then CrashAndBurn;

    Context:= SDL_GL_CreateContext( Window );
    if not Assigned(Context) then CrashAndBurn;

    if 0 <> SDL_GL_MakeCurrent(Window, Context) then CrashAndBurn;

    InitOpenGL;
    LoadOpenGLFunctions;

  end;

  function TSDL2Framework.GetOpenGLProcAddress(ProcName: ansistring): Pointer;
  begin
    Result:= SDL_GL_GetProcAddress(PAnsiChar(ProcName));
  end;



  function TSDL2Framework.GetScreenRect(fullscreen: boolean): TWindowManagerRect;
  var
    m: TSDL_DisplayMode;
    r: TSDL_Rect;
  begin
    if fullscreen then with Result do begin
      if 0 <> SDL_GetCurrentDisplayMode(dispaly_index, @m) then CrashAndBurn;
      left:=0;
      top:=0;
      width:= m.w;
      height:= m.h;
    end
    else  with Result do begin
      if 0 <> SDL_GetDisplayBounds(dispaly_index, @r) then CrashAndBurn;
      left:= r.x;
      top:= r.y;
      width:= r.w;
      height:= r.h
    end;
  end;

  procedure TSDL2Framework.CrashAndBurn;
  begin
    Die(RuEn(
      'Сбой SDL: %0',
      'SDL eror: %0'),
      [SDL_GetError()]);
  end;

  destructor TSDL2Framework.destroy;
  begin

    inherited;
  end;



  procedure TSDL2Framework.CheckBatteryStatus;
  //needed this for limiting rendering settings when laptop runs on battery
  begin
  end;

  //must point to a 32x32 ABGR pixels array.
  //NIL causes to hide the pointer.
  procedure TSDL2Framework.SetPointerImage(p: pointer; xhs, yhs: integer);
  begin
  end;

  // performs OpenGL's SwapBuffers for the main framebuffer, nothing more.
  procedure TSDL2Framework.SwapBuffers;
  begin
  end;

  function TSDL2Framework.MouseX: float;
  begin
  end;


  function TSDL2Framework.MouseY: float;
  begin
  end;

  procedure TSDL2Framework.SetVolume(volume: float);
  begin
  end;


  procedure TSDL2Framework.NewSoundData(begins: TDateTime; samples, rate: integer; data: pointer);
  begin
  end;


  procedure TSDL2Framework.PlaySound(samples, rate: integer; data: pointer);
  begin
  end;


  procedure TSDL2Framework.PlayWav(filename: TFileNameString);
  begin
  end;

  procedure TSDL2Framework.SetWindowName(n: PUnicodeChar);
  begin
  end;





initialization
  TAbstractFramework.RegisterFramework(TSDL2Framework, 'SDL2');
finalization

end.

