{
    This file is part of Chentrah,
    Copyright (C) 2004-2015 Anton Rzheshevski (chebmaster@mail.ru).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see http://www.gnu.org/licenses/

 **********************************************************************}

// The server framework is stub framework. It does nothing.


{$longstrings on}
{$codepage utf-8}
{$include un_globaldefs.h}

unit framework_server;

interface

uses SysUtils, Math, cge;

type

  TServerFramework = class (TAbstractFramework)
  protected
    PreviousFpsCullTsc: qword;
    procedure HBeat; override;
    procedure ProcessNetworkTraffic; override;
  public
     constructor Create; override;
     destructor Destroy; override;
     procedure SetPointerImage(p: pointer; xhs, yhs: integer); override;
     procedure SwapBuffers; override;
     function MouseX: float; override;
     function MouseY: float; override;
     procedure SetVolume(volume: float); override;
     procedure NewSoundData(begins: TDateTime; samples, rate: integer; data: pointer); override;
     procedure PlaySound(samples, rate: integer; data: pointer); override;
     procedure PlayWav(filename: TFileNameString); override;
     procedure CheckBatteryStatus; override;
     Procedure MainThreadMainLoop; override;
     procedure SetWindowName(n: PUnicodeChar); override;
     procedure YellAndExpire(y: UnicodeString);  override;
  end;

implementation


  constructor TServerFramework.Create;
  begin
    inherited;
    Cge.Framework:= Self;
    CheckBatteryStatus;
  end;

  destructor TServerFramework.Destroy;
  begin
    pointer(Cge.Framework) := nil;
    inherited;
  end;

  procedure TServerFramework.CheckBatteryStatus;
  begin
    Mother^.State.BatteryRemaining:= 255;
  end;

  Procedure TServerFramework.MainThreadMainLoop;
  begin
    try
      Mother^.State.ProgramInitialized:= Yes;

      repeat

        HBeat;

        NowRestarting:= No;
      Until Mother^.State.ExitRequested or Mother^.State.RestartRequested;

      if Mother^.Debug.Enabled then AddLog('...exited the main loop.');
    Except
      EverythingIsOk:=No;
      if not (ExceptObject is EDying) then Try
          Die(MI_ERROR_UNCAUGHT_EXCEPTION);
      Except
      End;
    End;
  end;

 procedure TServerFramework.HBeat;
 var
   delta: double;
//qq: qword;
 begin
   if Assigned(Mother^.LockupGuard) then Mother^.LockupGuard.Checkup;

   Mother^.Display.WindowHasFocus:= true;
   Mother^.Display.CanRender:= false;
   Mother^.Timer.CullFrameRate:= FPS_CULL_DEFAULT;
   Mother^.Timer.SuspendSwapBuffersMeasurement:= true;

   ProcessNetworkTraffic;

   if Assigned(Module) then begin
     Module.CheckDate;
     Module.Pulse;
   end;

   MeasureFps;

   NoteThreadOccupation(tlk_Working);

   {$ifndef cpuarm}
   RecheckTSCFreq;
   {$endif}

   CheckBatteryStatus;

   delta:= UsecDelta(@PreviousFpsCullTsc) - 1000000.0 / max(FPS_CULL_DEFAULT, Mother^.Timer.CullFrameRate);

   delta:= 1.0 * Sign(delta) * Math.min(abs(delta), 1000000.0 / FPS_CULL_DEFAULT); {
     куллинг дельты - абсолютно необходимая вещь. Бывает, кадр занимает секунд по пять,
     при старте или если антивирус дрочит при запуске модуля.
   }
   if delta < 0 then begin
      Mother^.Timer.HiResSleep(-Round(delta / 1000.0));
      delta:= 0;
   end;
   UsecDelta(@PreviousFpsCullTsc);
   PreviousFpsCullTsc+= round(delta);
 end;

  procedure TServerFramework.ProcessNetworkTraffic;
  begin

  end;


  procedure TServerFramework.YellAndExpire(y: UnicodeString);
  begin
    Mother^.State.ExitRequested:= true;
    AddLog(RuEn(
      'Режим сервера, сообщение об ошибке: %0',
      'Server mode, error message: %0'),[y]);
    cge.DisplayDyingYells(y);
  end;

  procedure TServerFramework.SetPointerImage(p: pointer; xhs, yhs: integer);
  begin
  end;

  // performs OpenGL's SwapBuffers for the main framebuffer, nothing more.
  procedure TServerFramework.SwapBuffers;
  begin
  end;


  function TServerFramework.MouseX: float;
  begin
    Result:= 0;
  end;


  function TServerFramework.MouseY: float;
  begin
    Result:= 0;
  end;

  procedure TServerFramework.SetVolume(volume: float);
  begin
  end;


  procedure TServerFramework.NewSoundData(begins: TDateTime; samples, rate: integer; data: pointer);
  begin
  end;


  procedure TServerFramework.PlaySound(samples, rate: integer; data: pointer);
  begin
  end;


  procedure TServerFramework.PlayWav(filename: TFileNameString);
  begin
  end;

  procedure TServerFramework.SetWindowName(n: PUnicodeChar);
  begin
    //do nothing
  end;



initialization
  TAbstractFramework.RegisterFramework(TServerFramework, 'Server');
finalization

end.
