{
    This file is part of Chentrah,
    Copyright (C) 2004-2016 Anton Rzheshevski (chebmaster@mail.ru).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see http://www.gnu.org/licenses/

 **********************************************************************}
{$longstrings on}
{$codepage utf-8}
{$include un_globaldefs.h}
{$rangechecks off}

unit framework_winapi;


(*
 My thanks to Jan Horn
   for his OpenGL Delphi examples - http://www.sulaco.co.za/
   -- these allowed me to make my first step into OpenGL.
*)

interface

uses SysUtils, classes, Windows
    , cl_windows_missing

    //, jwawinuser
    , ctypes, math, cge, un_strings

    , framework_native
    , framework_winapi_wintab
  ;


type
  TWinApiFramework = class (TNativeFramework)
  protected
    icon_hwcursor : HCursor;
    pfd : TPIXELFORMATDESCRIPTOR;
    rect: TRect;
    ShowCursorCount: integer;
    dwStyle, dwExStyle: dword;

    f_prevMX, f_prevMY: integer; // for grabbed mode

    procedure _CheckCursorMode; override;

    //implementing the abstract service functions introduced by the native framework
    procedure InitDisplay; override;
    procedure CreateWindow; override;
    procedure InitGLFirstStage; override;
    procedure InitGL; override;
    procedure CloseGL; override;
    procedure CloseWindow; override;
    procedure _UploadPointer; override;
    function GetScreenRect(fullscreen: boolean): TWindowManagerRect; override;//if not fullscreen then return the working area (excluding the taskbar)
    function GetWindowOuterRect(): TWindowManagerRect; override;   //Ask directly from the OS
    function GetWindowClientRect(): TWindowManagerRect; override;  //Ask directly from the OS
    function GetIsWindowFullscreen(): boolean; override; //Ask directly from the OS
    function GetIsWindowMinimized(): boolean; override; //Ask directly from the OS
    procedure CorrectFullscreenWindowSize; override; //if screen size changes between frames: for example, laptop cover open
    procedure _SwapBuffers; override;
    procedure Heartbeat; override; //processing window messages and such

    procedure MaybeSnapCursorBackToCenter;
  public
    WinTabManager: TWinTabmanager;
    display: HDC;
    RenderingContext : HGLRC;
    constructor Create; override;
    destructor Destroy; override;

    //implementing abstract functions that make up the framework API
    procedure CheckBatteryStatus; override;
    procedure SetWindowName(n: PUnicodeChar); override;
  end;


implementation


  {$include framework_wgl.inc}

  constructor TWinApiFramework.Create;
  begin
    inherited;

    Cge.Framework:= Self; //other classes need to access it.

    if Mother^.Debug.Enabled then AddLog('Starting the window manager...');

    PrevBatteryRemaining:= -1;

    InitWindowSubsystem;

    //CgeEmptyCursor;
    Self.SetPointerImage(nil, 0, 0);

    Console.InitGL;

    if Mother^.Debug.Enabled then AddLog('Starting the pen tablet manager...');
    WinTabManager:= TWinTabmanager.Create;
    EngineObject.AddObj(WinTabManager);

    GamepadManager:= TGamepadManager.Create;
    GamepadManager.pointer_to_instance:= pointer(@Self.GamepadManager);
    EngineObject.AddObj(GamepadManager);

    Mother^.Display.BeginBlackout(INITIAL_BLACKOUT);

    Console.DrawBackground;
    Console.Draw;

    InitMotherAssets;

    LoadQF;

    if Mother^.Debug.Enabled then AddLog('Initializing sound...');

    try
      if Config.Bool[Mother^.State.OstID, 'use_directsound_not_openal']
        then SoundManager:= TDirectXSoundMan.Create
        else SoundManager:= TOpenALSoundman.Create;
      EngineObject.AddObj(SoundManager);
    except
      pointer(SoundManager):= nil;
      if Config.bool['sound', 'allow_running_without']
      then
        AddWarning (MI_CAPTION_NOSOUND, StopDying(),[])
      else
        AddStopMessage(StopDying()); //clear the error messages buffer by copying them into an error screen.
    end;
  end;

  procedure TWinApiFramework.InitDisplay;
  begin
    ShowCursorCount:= 1;
  end;

  destructor TWinApiFramework.destroy;
  begin
    ShutdownWindowSubsystem;
    pointer(Cge.Framework) := nil;
    inherited;
  end;

  procedure TWinApiFramework.CheckBatteryStatus;
  //needed this for limiting rendering settings when laptop runs on battery
  var
    ps: SYSTEM_POWER_STATUS;
//    F: TextFile;
    s: TFileStream;
    u: Utf8String;
  begin
    if Now() < BatteryStatusCheckMoment then Exit; //check every 10 seconds
    if not GetSystemPowerStatus(ps) then begin
      if Mother^.State.BatteryRemaining <> 255
        then AddLog(RuEn(
          'Не удалось получить от ОС состояние электропитания. Заряд батареи НЕ учитывается!',
          'Failed to get power status from the OS. The battery charge is NOT accounted for!'));
      Mother^.State.BatteryRemaining:= 255;
      BatteryStatusCheckMoment:= Now() + (5 * 60) / SecondsPerDay; //check again five minutes later
    end
    else if ps.ACLineStatus > 0 then begin
      BatteryStatusCheckMoment:= Now() + 10 / SecondsPerDay; //check often: what if it gets unplugged?
      Mother^.State.BatteryRemaining:= 255;
    end
    else begin
      Mother^.State.BatteryRemaining:= ps.BatteryLifePercent; //it's an unsigned byte, LOL.
      BatteryStatusCheckMoment:= Now() + 30 / SecondsPerDay;  //check two times a minute

      if Mother^.State.BatteryRemaining < 10 then begin
        MakeSurePathExists(UnmangleFileName('*H/logs'));
        s:= TFileStream.Create(UnmangleFileName('*H/logs/last-last-will.txt'), fmCreate);
//        AssignFile(F, _UnmangleFileName('*H/logs/last-last-will.txt'));
//        Rewrite(F);
//        WriteLn(F,
          u:= Utf8Encode(RuEn(
          'Аварийный выход: батарея ноутбука щас сдохнет!',
          'Emergency shutdown: laptop battery is dying!')) + #10#13;
          //);
        s.Write(u[1], Length(u));
        s.Free;
//        CloseFile(F);
        RestartMyself('--last-will');
      end;
    end;

    if (PrevBatteryRemaining <> Mother^.State.BatteryRemaining)
    then begin
      if Mother^.State.BatteryRemaining > 100
        then AddLog(RuEn('Питание от сети.', 'AC power.'))
        else AddLog(RuEn('Заряд батареи %0%.', 'Battery charge %0%.')
                    , [Mother^.State.BatteryRemaining]);
        PrevBatteryRemaining:= Mother^.State.BatteryRemaining;
    end;

    //Disable higher-resolution waits for better battery life
    if Mother^.State.BatteryRemaining <= 100 then begin
      if Mother^.Timer.ActiveTBP1 then begin
        if Mother^.Debug.Verbose then AddLog('Calling timeEndPeriod(1)...') ;
        if timeEndPeriod(1) = TIMERR_NOERROR then begin
          Mother^.Timer.ActiveTBP1:= false;
          if Mother^.Debug.Verbose then AddLogOk
        end else
          if Mother^.Debug.Verbose then AddLog('FAILED');
      end;
    end else begin
      if not Mother^.Timer.ActiveTBP1 then begin
        if Mother^.Debug.Verbose then AddLog('Calling timeBeginPeriod(1)...') ;
        if timeBeginPeriod(1) = TIMERR_NOERROR then begin
          Mother^.Timer.ActiveTBP1:= true;
          if Mother^.Debug.Verbose then AddLogOk
        end else
          if Mother^.Debug.Verbose then AddLog('FAILED');
      end;
    end;
   end;

  procedure TWinApiFramework._SwapBuffers;
  begin
    case  Mother^.GAPI.Mode of
      {$ifndef glesonly}
      gapi_GL21, gapi_DeprecatedImmediateGL : begin
        if Assigned(wglSwapBuffers) then wglSwapBuffers(display);
      end;
      {$endif}
      gapi_GLES2: begin
        if Assigned(eglSwapBuffers) then eglSwapBuffers(eDisplay, eSurface);
      end
    else
      DieUnsupportedGLMode;
    end;
  end;

  procedure TWinApiFramework._CheckCursorMode();
  begin
     //it's our task to restore the proper cursor each time Windows
     //  swithes it to something else
    if CursorCreated and not Mother^.Display.MouseIsBeyondOurWindow
      then SetCursor(icon_hwcursor); //winAPI call

    if Mother^.State.OS in OSNeedsMousePointerHack //namely Vista
    then begin
      if (not f_PointerVisible or Mother^.Display.UseHardware1bitCursor)
        //and not Mother^.Display.MouseIsBeyondOurWindow
      then begin
        if ShowCursorCount < 1 then begin
          ShowCursor(True); //winAPI call
          inc(ShowCursorCount);
        end;
      end
      else begin
        if ShowCursorCount > 0 then begin
          ShowCursor(False); //just setting a completely transparent cursor is not enough
          dec(ShowCursorCount);
        end;
      end;
    end;

    inherited;
  end;

  procedure TWinApiFramework.SetWindowName(n: PUnicodeChar);
  begin
    if
      {$ifdef cpu32}
        not Assigned(SetWindowTextW) or Mother^.State.Windows98.DoesNotSupportUnicode
      {$else}
        false
      {$endif}
    then begin
      SetWindowText(
        WindowHandle,
        PAnsiChar(WideToAnsi(n))
      );
    end
    else begin
      SetWindowTextW(
        WindowHandle,
        n
      );
    end;
   end;



















  const
    WindowedStyle: dword = WS_CAPTION or WS_SYSMENU or WS_SIZEBOX or WS_VISIBLE
      or WS_MAXIMIZEBOX or WS_MINIMIZEBOX or WS_CLIPCHILDREN;
    WindowedExStyle: dword =  WS_EX_APPWINDOW or WS_EX_WINDOWEDGE;
    FullscreenStyle: dword =  dword(WS_POPUP) or dword(WS_MAXIMIZE)
      or dword(WS_VISIBLE) or dword(WS_CLIPCHILDREN);
    FullscreenExStyle: dword =  WS_EX_APPWINDOW or WS_EX_TOPMOST;

  var gv_FramesSinceLeavingFullscreen: int64 = 0;

  procedure TWinApiFramework._UploadPointer;
  var
    //cursorInfo: TIconInfo;
    //cursor: THandle;
    OldCursor: HCursor;
    EmptyBitmap, EmptyMask: array[0..((32*32) div 8) - 1] of byte;
  begin
    _UploadGLcursor;
    OldCursor:= icon_hwcursor;

    if Mother^.Display.UseHardware1bitCursor
      then icon_hwcursor:= CreateCursor(
        system.Hinstance, PointerXHotSpot, PointerYHotSpot, 32, 32, @PointerBitmap.AndMask, @PointerBitmap.XorMask)
      else begin
        FillChar(EmptyMask, Sizeof(EmptyBitmap), $ff);
        FillChar(EmptyBitmap, Sizeof(EmptyBitmap), 0);
        icon_hwcursor:= CreateCursor(system.Hinstance, 1, 1, 32, 32, @EmptyMask, @EmptyBitmap);
      end;
    SetCursor(icon_hwcursor);

    if CursorCreated then DestroyCursor(OldCursor);
    CursorCreated:= Yes;
  end;

  // Эта процедура обрабатывает приходящие в адрес окна сообщения.
  function WndProc(hWnd: HWND; Msg: UINT;  wParam: WPARAM;  lParam: LPARAM): LRESULT; stdcall;
  begin
    if not (Assigned(Framework) and (Framework as TWinApiFramework).WindowSubsystemInitialized)
      then Exit(DefWindowProc(hWnd, Msg, wParam, lParam));
    case msg of
     WM_ACTIVATE: begin
       // cannot be processed in the main message loop:
       // arrives only here, in the window procedure
       Case LOWORD(wParam) of
        WA_ACTIVE, WA_CLICKACTIVE:
          if (Framework as TWinApiFramework).WindowStateCommandedByTheOs in [wsco_Minimized, wsco_NoFocus]
          then begin
            (Framework as TWinApiFramework).WindowStateCommandedByTheOs:= wsco_Normal;
            if Mother^.Debug.Verbose then AddLog('The application window is activated. Commanding the window manager to switch to the default windowed mode.');
          end;
        WA_INACTIVE:
          if ((Framework as TWinApiFramework).WindowStateCommandedByTheOs in [wsco_Normal, wsco_Maximized])
          then begin
            (Framework as TWinApiFramework).WindowStateCommandedByTheOs:= wsco_NoFocus;
            if Mother^.Debug.Verbose then AddLog('The application window is deactivated. Commanding the window manager to switch to the no-focus windowed mode.');
          end;
       end;
     end;
     WM_SYSCOMMAND: begin
       Case (wParam and $FFF0) of
        {MSDN: In WM_SYSCOMMAND messages, the four low-order bits of the wParam parameter
        are used internally by the system. To obtain the correct result when testing
        the value of wParam, an application must combine the value 0xFFF0 with the
        wParam value by using the bitwise AND operator.}
{         SC_HOTKEY, SC_TASKLIST:
           if (Assigned(Framework) and Assigned((Framework as TWinApiFramework)) and (Framework as TWinApiFramework).Focus and (Framework as TWinApiFramework).MustBeMaximized//Mother^.Display.InFullScreenMode
             then (Framework as TWinApiFramework).MustBeMaximized:= false;//Exit(-1);
}        SC_CLOSE: begin
           if Mother^.Debug.Verbose then AddLog('The OS commands the application window to be closed.Processing normally.');
         end;
         SC_MAXIMIZE: begin
            if ((Framework as TWinApiFramework).WindowStateCommandedByTheOs <> wsco_Maximized)
            then begin
              if gv_FramesSinceLeavingFullscreen < 2 then begin
                if Mother^.Debug.Verbose then AddLog('The OS commands the application window to be maximized. Since it was less than 2 frames ago since leaving fullscreen, simply blocking the maximization.');
              end
              else begin
                if Mother^.Debug.Verbose then AddLog('The OS commands the application window to be maximized. Preventing the OS from doing that. Commanding the window manager to switch to fullscreen instead.');
                (Framework as TWinApiFramework).WindowStateCommandedByTheOs:= wsco_Maximized;
              end;
            end
            else
              if Mother^.Debug.Verbose then AddLog('The OS commands the application window to be maximized. This makes no sense, it is already fullscreen. Simply rejecting the command and doing nothing.');
            Exit(1); //Do NOT let Windows to maximize the window.
         end;
         SC_RESTORE: begin
           if (Framework as TWinApiFramework).WindowStateCommandedByTheOs in [wsco_Minimized, wsco_Maximized] then begin
             if Mother^.Debug.Verbose then AddLog('The OS commands the application window to restore its size and posiztion. Commanding the window manager to switch to default windowed mode.');
             (Framework as TWinApiFramework).WindowStateCommandedByTheOs:= wsco_Normal;
           end
           else
             if Mother^.Debug.Verbose then AddLog('The OS commands the application window to restore its size and posiztion. Letting the OS process this..');
         end;
         SC_MINIMIZE: begin
           if ((Framework as TWinApiFramework).WindowStateCommandedByTheOs <> wsco_Minimized)
           then begin
             if Mother^.Debug.Verbose then AddLog('The OS commands the application window to be minimized. Commanding the window manager to switch to the minimized mode.');
           end
           else
             if Mother^.Debug.Verbose then AddLog('The OS commands the application window to be inimized. This makes no sense, it is already minimized.');
         end;
//else addlog('WM_SYSCOMMAND (%0, %1)',[IntToHex(wparam, 4), IntToHex(lparam, 8)]);
       end;

     end;
     WM_KILLFOCUS: begin
        (Framework as TWinApiFramework).WindowStateCommandedByTheOs:= wsco_NoFocus;
        if Mother^.Debug.Verbose then AddLog('The application window loses focus. Commanding the window manager to switch to the no-focus windowed mode.');
      end;
     WM_SETFOCUS: begin
       if ((Framework as TWinApiFramework).WindowStateCommandedByTheOs <> wsco_Maximized)
       then begin
        (Framework as TWinApiFramework).WindowStateCommandedByTheOs:= wsco_Normal;
        if Mother^.Debug.Verbose then AddLog('The application window receives focus. Commanding the window manager to switch to default windowed mode.');
      end
      else
        if Mother^.Debug.Verbose then AddLog('The application window receives focus. This makes no sense, we are already fullscreen.');
     end;
     WM_SYSKEYDOWN: begin
       case wparam of
         VK_F10: begin
           if Mother^.Debug.Verbose then AddLog('The application window received system command "F10 pressed" to activate menu bar. Blocking it to allow using the F10 key.');
           Exit(DefWindowProc(hWnd, WM_KEYDOWN, wParam, lParam));
         end;
         else
           if (Framework as TWinApiFramework).WindowStateCommandedByTheOs = wsco_Maximized
           then begin
             (Framework as TWinApiFramework).WindowStateCommandedByTheOs:= wsco_Normal;
             if Mother^.Debug.Verbose then AddLog('A system key was pressed! Commanding the window manager to switch to default windowed mode.');
           end;
       end;
//       if wparam <> VK_MENU then AddLog('Window received WM_SYSKEYDOWN (%0, %1)', [IntToHex(wparam, 4), IntToHex(lparam, 8)]);
     end;
     WM_SYSKEYUP: begin
       case wparam of
         VK_F10: begin
           if Mother^.Debug.Verbose then AddLog('The application window received system command "F10 released". Blocking it to allow using the F10 key.');
           Exit(DefWindowProc(hWnd, WM_KEYUP, wParam, lParam));
         end;

       end;

//       if wparam <> VK_MENU then AddLog('Window received WM_SYSKEYDOWN (%0, %1)', [IntToHex(wparam, 4), IntToHex(lparam, 8)]);

     end;

{ $200:;
$20:;
$7f:;
$84:;
$104:;
else addlog('WindowProc (, %0, %1, %2)',[IntToHex(msg, 8), IntToHex(wparam, 4), IntToHex(lparam, 8)]);
}
    end;
    Result := DefWindowProc(hWnd, Msg, wParam, lParam)    // Default result if nothing happens
  end;

  function TWinApiFramework.GetScreenRect(fullscreen: boolean): TWindowManagerRect;
  var
    rc: TRect;
    Monitor: HMONITOR;
    mi: TMonitorInfo;
  begin
{
    if Mother^.State.OS in [ostWin98, OstWin2k] then with Result do begin
      //Use legacy method fow windozes older than XP
      left:= 0;
      top := 0;
      width:= GetSystemMetrics(SM_CXSCREEN);
      height:= GetSystemMetrics(SM_CYSCREEN);
      //Assuming that taskbar is at the bottom and is 28 pixels high:
      if not Fullscreen then Height-= 28;
    end
    else begin
}
      if windowhandle = 0 then begin
        //Not created yet. Use default monitor.
         with rc do begin
           left:= 0;
           top:= 0;
           right:= 1;
           bottom:= 1;
         end;
         Monitor:= MonitorFromRect(rc, MONITOR_DEFAULTTOPRIMARY);
      end else begin
        //get monitor from window
        GetWindowRect(windowhandle, rc);
        Monitor:= MonitorFromRect(rc, MONITOR_DEFAULTTONEAREST);
      end;
      mi.cbSize:= sizeof(mi);
      GetMonitorInfo(Monitor, @mi);
      if fullscreen
        then rc:= mi.rcMonitor
        else rc:= mi.rcWork;
      Result.left:= rc.left;
      Result.top:= rc.top;
      Result.width:= rc.right - rc.left;
      Result.height:= rc.bottom - rc.top;
//     end;
  end;

  function WmRectToWindowsRect(r: TWindowManagerRect): TRect;
  begin
    Result.left:= r.left;
    Result.top:= r.top;
    Result.right:= r.left + r.width;
    Result.bottom:= r.top + r.height;
  end;

  function WindowsRectToWmRect(r: TRect): TWindowManagerRect;
  begin
    Result.left:= r.left;
    Result.top:= r.top;
    Result.width:= r.Right - r.left;
    Result.height:= r.Bottom - r.top;
  end;

  function TWinApiFramework.GetWindowOuterRect(): TWindowManagerRect;
  var r: TRect;
  begin
    GetWindowRect(WindowHandle, r);
    Result:= WindowsRectToWmRect(r);
  end;


  function TWinApiFramework.GetWindowClientRect(): TWindowManagerRect;
  var
    r: TRect;
    point: TPoint;
  begin
    point.x:= 0;
    point.y:= 0;
    with Cge.Framework as TWinApiFramework do begin
      GetClientRect(WindowHandle, r); //This suka always returns 0,0 for the top left corner!
      ScreenToClient(WindowHandle, point);
    end;
    Result:= WindowsRectToWmRect(r);
    Result.Top-= point.x;
    Result.Left-= point.y;
  end;


  procedure TWinApiFramework.CreateWindow;
  var
    wndClass : TWndClass;
    s: string;
  begin
    if Mother^.Debug.Verbose then AddLog('Screen rect is %0',[RectToStr(Mother^.Display.ScreenRect)]);

    if Mother^.Debug.Verbose then AddLog('Creating the window...');

    if Mother^.Display.InFullScreenMode then begin
      dwStyle:= FullscreenStyle;
      dwExStyle:= FullscreenExStyle;
    end else begin
      dwStyle:= WindowedStyle;
      dwExStyle:= WindowedExStyle;
    end;

    s:=AppNick+'_OpenGL_WND';

    // Заполняем бланк-формуляр, которым будем кормить процедуру создания класса окна
    ZeroMemory(@wndClass, SizeOf(wndClass));
    with wndClass do begin
      style         := //CS_HREDRAW or    // Перерисовывать окно целиком при изменении ширины
                       //CS_VREDRAW or    // ...то же, для высоты
                       CS_OWNDC;        // Окну нужен собственный контекст устройства
      lpfnWndProc   := @WndProc;        // Оконная процедура, которая будет ловить мессаги
      hInstance := system.HInstance;
      lpszClassName := PAnsiChar(s);
      hIcon:= LoadIcon(system.HInstance, PAnsiChar('MAINICON'));
    end;

    // Регистрируем класс окна (ну, вот нельзя так просто создать окно - ещё,
    //   види-те ли, класс нужен... Нет, явно Виндовс писали извращенцы.)
    if (0 = RegisterClass(wndClass)) then Die('RegisterClass() returned 0');

{    wmr:= GetScreenRect(true);
    Mother^.Display.ScreenWidth:= wmr.width;//GetSystemMetrics(SM_CXSCREEN);
    Mother^.Display.ScreenHeight:= wmr.height;//GetSystemMetrics(SM_CYSCREEN);

    if not Mother^.Display.InFullScreenMode then begin
     //adjust the initial width and height with regards to the window decorations
      _rect.left:=1;
      _rect.top:=1;
      _rect.right:=Mother^.Display.WindowClientWidth;
      _rect.bottom:=Mother^.Display.WindowClientHeight;
      AdjustWindowRectEx(_Rect, dwStyle, false, dwExStyle);
      Mother^.Display.WindowClientWidth:=_rect.right - _rect.left + 1;
      Mother^.Display.WindowClientHeight:=_rect.bottom - _rect.top + 1;
      w:= Mother^.Display.WindowClientWidth;
      h:= Mother^.Display.WindowClientHeight;
      l:= round(Mother^.Display.WindowedOriginX * Mother^.Display.ScreenWidth);
      t:= round(Mother^.Display.WindowedOriginY * Mother^.Display.ScreenHeight);
    end else begin
      w:= Mother^.Display.ScreenWidth;
      h:= Mother^.Display.ScreenHeight;
      l:=0;
      t:=0;
    end;
 }

    if Mother^.Debug.Verbose then AddLog('The window rect, as read from the config and clipped to the screen rect, should be %0',[RectToStr(Mother^.Display.WindowRect)]);

    // Создаём окно
    with Mother^.Display.WindowRect do begin
      if {$ifdef cpu32}not Assigned(CreateWindowExW) or {$endif}
        Mother^.State.Windows98.DoesNotSupportUnicode
      then begin
        WindowHandle:= CreateWindowEx(
          dwExStyle, PAnsiChar(s),
          PChar(WideToAnsi(MessageContainer[MI_CGE_TITLE])), dwStyle,
            left, top, width, height, 0, 0, hInstance, nil);
      end
      else begin
        WindowHandle:= CreateWindowExW(
          dwExStyle, PUnicodeChar(AnsiToWide(s)),
          PUnicodeChar(MessageContainer[MI_CGE_TITLE]), dwStyle,
            left, top, width, height, 0, 0, hInstance, nil);
      end;
    end;
    if WindowHandle = 0
      then Die('CreateWindowEx() returned 0');

    //SetForegroundWindow(Mother^.Display.WindowHandle);


    Mother^.Display.WindowRect:= GetWindowOuterRect();
    Mother^.Display.WindowClientRect:= GetWindowClientRect();
    formerClientRect:= Mother^.Display.WindowClientRect;
    Mother^.Display.WindowExists:= True;
    Mother^.Display.CanRender:= true;



    if Mother^.Debug.Verbose then AddLog('Window created. Client area %0, window %1, fullscreen=%2',[
      RectToStr(Mother^.Display.WindowClientRect), RectToStr(Mother^.Display.WindowRect),
      Mother^.Display.InFullScreenMode
    ]);
  end;

  procedure  TWinApiFramework.CloseWindow;
  begin
    if not Mother^.Display.WindowExists then begin
      AddLog(RuEn(
        'Не удалось удалить окно за неимением такового.',
        'Unable to destroy the window due to absence of it.'));
      Exit
    end
    else try
      if (WindowHandle <> 0)
        and IsWindow(WindowHandle)
        and not DestroyWindow(WindowHandle)
      then begin
        AddLog('DestroyWindow() call failed!');
        WindowHandle := 0;
      end
      else
        if Mother^.Debug.Enabled then AddLog('Window destroyed.');
    except
      AddLog(StopDying());
    end;

    // Снимаем с учёта класс окна
    if (not Windows.UnRegisterClass(PAnsiChar(AppNick+'_OpenGL_WND'), hInstance))
    then  AddLog('UnRegisterClass() call failed!');

    Mother^.Display.CanRender:= false;
    Mother^.Display.WindowExists:=False;

    if CursorCreated then DestroyCursor(icon_hwcursor);
  end;

  {$ifndef fpc}
  function GetKeyboardState(lpKeyState:PBYTE):WINBOOL; external 'user32' name 'GetKeyboardState';
  {$endif}

  const
    MAPVK_VK_TO_VSC = 0;
    MAPVK_VK_TO_VSC_EX = 4;
  function DivineScanCode(var msg: TMSG): integer; //as in "divination", not "divinity".
  begin
    { Note: Wine behaves differently.
        In the original Windows, bit 24 (extended keyboard key)
        is duplicated in the bit 23 of the scancode itself.
        Wine doesn't do that.}
    Result:= ((msg.lParam shr 16) and $FF) or ((msg.lParam shr 17) and $80);
{   if (Mother^.State.OS in WindowsEmulators) then case Result of
      ***TODO   ДОДЕЛАТЬ!
      $xx: Result:= KEY_LEFT_WINDOWS;
      $xx: Result:= KEY_RIGHT_WINDOWS;
      $xx: Result:= KEY_APP_MENU;
    end;}
  end;


  function TWinApiFramework.GetIsWindowFullscreen(): boolean;
  var s: ptruint;
  begin
    s:= {$ifdef cpu64}GetWindowLongPtr{$else}GetWindowLong{$endif}
      (WindowHandle, GWL_STYLE);
    Result:= (s and WS_POPUP) <> 0;
  end;

  function TWinApiFramework.GetIsWindowMinimized(): boolean;
  var s: ptruint;
  begin
    s:= {$ifdef cpu64}GetWindowLongPtr{$else}GetWindowLong{$endif}
      (WindowHandle, GWL_STYLE);
    Result:= (s and WS_MINIMIZE) <> 0;
  end;

  procedure TWinApiFramework.CorrectFullscreenWindowSize;
  begin
    with Mother^.Display.ScreenRect do
      if not SetWindowPos(WindowHandle, HWND_TOPMOST, left, top, width, height, 0)
        then Die('Failed to adjust window size!');
    MaybeSnapCursorBackToCenter;
  end;

  procedure TWinApiFramework.MaybeSnapCursorBackToCenter;
  var
    point: TPoint;
  begin
    if not f_PointerGrabbed then Exit;
    point.x:= Mother^.Display.WindowClientRect.Width div 2;
    point.y:= Mother^.Display.WindowClientRect.Height div 2;
    ClientToScreen(WindowHandle, point);
    SetCursorPos(point.x, point.y);
    GetCursorPos(point); //may not match. Frakkin' scaling!
    f_prevMX:= point.x; //delta is calculated in screen coordinates
    f_prevMY:= point.y;
  end;


  Procedure TWinApiFramework.Heartbeat;
  var
    msg: TMsg;
    i, oldmx, oldmy
      //, px, py
      : integer;
    point: TPoint;
    KeyboardState: array[0..255] of byte;
    buff: array[0..10] of UnicodeChar;
    abuff: array[0..10] of AnsiChar;
    numchars: integer;
    scancode: integer;
    cr: TWindowManagerRect;

    procedure GrabPointer;
    begin
      SetPointerImage(nil, 0, 0);
      f_PointerGrabbed:= True;
      MaybeSnapCursorBackToCenter;
      CombinedMx:= 0.0;
      CombinedMy:= 0.0;
      if Mother^.Debug.Verbose then AddLog('Mouse pointer grabbed.');
    end;


    procedure UngrabPointer;
    begin
      SetPointerImage(@PointerCache, PointerXHotSpot, PointerYHotSpot);
      f_PointerGrabbed:= False;
      if Mother^.Debug.Verbose then AddLog('Mouse pointer ungrabbed.');
    end;

    procedure CheckIfMouseMoved(per_cycle: boolean = false);
    var
      gbcond: boolean;
    begin
      gbcond:= Mother^.Input.GrabPointer and f_focus;
      if f_PointerGrabbed <> gbcond then begin
        if gbcond then GrabPointer else UngrabPointer;
      end;

      if f_wsco = wsco_Minimized then Exit;

      if f_pointerGrabbed then begin
        if (per_cycle) then begin
          GetCursorPos(point);
          f_MouseX:= point.x - f_prevMX;
          f_MouseY:= point.y - f_prevMY;
          MaybeSnapCursorBackToCenter;
          Mother^.Display.MouseIsBeyondOurWindow:= false;
          OnMouseMove; // this ancestor metod processes differently based on f_pointerGrabbed
        end
        else begin
          f_MouseX:= 0;
          f_MouseY:= 0;
        end;
      end
      else begin
        GetCursorPos(point);
        ScreenToClient(WindowHandle, point);
        if per_cycle then begin
          oldmx:=f_MouseX;
          oldmy:=f_MouseY;
        end;
        f_MouseX:=point.x;
        f_MouseY:=point.y;
        Mother^.Display.MouseIsBeyondOurWindow:= (f_MouseY < 0) or (f_MouseX < 0)
                                or (f_MouseY > Mother^.Display.WindowClientRect.Height)
                                or (f_MouseX > Mother^.Display.WindowClientRect.Width);
        if ((f_MouseX <> oldmx) or (f_MouseY <> oldmy)) and (per_cycle) then OnMouseMove;
      end
    end;


    procedure GoFullscreen;
    var
      //_rect: TRect;
      dwStyle, dwExStyle: dword;
    begin
      Mother^.Display.ScreenRect:= GetScreenRect(true); //Width:= GetSystemMetrics(SM_CXSCREEN);
      AddLog(RuEn(
        'Переключение в полноэкранный режим (%0х%1)...',
        'Switching to full screen mode (%0x%1)...'
        ), [Mother^.Display.ScreenRect.Width, Mother^.Display.ScreenRect.Height]
      );
      dwStyle:= dword(WS_POPUP) or dword(WS_MAXIMIZE) or dword(WS_VISIBLE) or dword(WS_CLIPCHILDREN);
      dwExStyle:= WS_EX_APPWINDOW or WS_EX_TOPMOST;
      with Mother^.Display.ScreenRect do

      if (0 = {$ifdef cpu64} SetWindowLongPtr {$else} SetWindowLong {$endif} (
        WindowHandle, GWL_STYLE, dwStyle))
        //causes overflow error, this female dog is declared as SetWindowLongA(hWnd:HWND; nIndex:longint; dwNewLong:LONG):LONG;
      or (0 = {$ifdef cpu64} SetWindowLongPtr {$else} SetWindowLong {$endif} (
        WindowHandle, GWL_EXSTYLE, dwExStyle))
      or not SetWindowPos(
        WindowHandle, HWND_TOPMOST, left, top, width, height, 0)
      then Die('Failed to change the window to fullscreen!');
      Mother^.Display.InFullScreenMode:= true;
      Mother^.Display.WindowRect:= GetWindowOuterRect;
      Mother^.Display.WindowClientRect:= GetWindowClientRect;
      Config.Bool['video', 'fullscreen']:= true;
      AddLogOk;
      MaybeSnapCursorBackToCenter;
    end;

    procedure LeaveFullScreen;
    //var _rect: TRect;
    begin
      Config.Bool['video', 'fullscreen']:= false;
      AddLog(RuEn(
        'Переключение из полноэкранного в оконный режим...',
        'Switching from fullscreen to windowed mode...'
      ));
      gv_FramesSinceLeavingFullscreen:= 0;
      dwStyle:= WindowedStyle;
      dwExStyle:= WindowedExStyle;
      Mother^.Display.ScreenRect:= GetScreenRect(false);
      LoadWindowRectFromConfig;
      with Mother^.Display.WindowRect do
      if (0 = {$ifdef cpu64} SetWindowLongPtr {$else} SetWindowLong {$endif} (
        WindowHandle, GWL_STYLE, dwStyle))
      or (0 = {$ifdef cpu64} SetWindowLongPtr {$else} SetWindowLong {$endif} (
        WindowHandle, GWL_EXSTYLE, dwExStyle))
      or not SetWindowPos(
        WindowHandle, HWND_NOTOPMOST,
         left, top, width, height, 0)
      then Die('Failed to change the window from fullscreen to windowed!');
      Mother^.Display.InFullScreenMode:= false;
      Mother^.Display.WindowRect:= GetWindowOuterRect;
      Mother^.Display.WindowClientRect:= GetWindowClientRect;
      if Mother^.Debug.Verbose then AddLog('  the client area is %0, the entire window including decorations is %1',
        [RectToStr(Mother^.Display.WindowClientRect), RectToStr(Mother^.Display.WindowRect)]);
      MaybeSnapCursorBackToCenter;
    end;

  begin
    if IsWindow(WindowHandle) then begin
      NoteThreadOccupation(tlk_PollingInput);
    { 2006: я уже задолбался
      отлавливать все возможные варианты закрытия окна,
      после которого ни о чём не догадывающаяся программа продолжает работать,
      не отображаясь на Панели задач, но грузя процессор и деловито
      отправляя в никуда поток команд OpenGL... :(
    }

      f_focus:= (f_wsco in [wsco_Normal, wsco_Maximized]) and (GetForegroundWindow() {GetFocus()}
        = WindowHandle);
      f_HadMessages:= false;
      if f_focus
        then f_LoseFocusFramesCount:=0
        else inc(f_LoseFocusFramesCount);
   (*
      if Mother^.Display.InFullScreenMode and not f_focus then begin
        LogGoDown(RuEn('окно потеряло фокус','the window lost focus'));
        Mother^.State.ExitRequested:=true;
      end
      else begin
   *)
      CheckWindowSize;
      if Assigned(WinTabManager) then begin
        WinTabmanager.Pulse;
        for i:=0 to high(WinTabManager.Events) do
          with WinTabManager.Events[i] do
            OnPen((x * Mother^.Display.FullscreenRect.width) - (Mother^.Display.FullscreenRect.left - Mother^.Display.WindowClientRect.left),
                  ((1-y) * Mother^.Display.FullscreenRect.Height) - (Mother^.Display.FullscreenRect.top - Mother^.Display.WindowClientRect.top),
                  pressure, reserved1, reserved2, reserved3);
      end;
      CheckIfMouseMoved(true);
      _CheckCursorMode();
      While PeekMessage(msg, 0, 0, 0, PM_REMOVE) do begin
        case msg.message of
          WM_KEYDOWN:
            begin
              f_HadMessages:= true;
              scancode:= DivineScanCode(msg);
              numchars:=0;
              if Mother^.Input.TextInput then begin
                GetKeyboardState(@KeyboardState[0]);
{ //Researching the Unicode support state in the vanilla Windows 98 SE:
addlog('ToUnicode: %0 %1',[msg.wparam, (msg.lParam shr 16) and $FF]);
addlog('-----------------------------------');
addlog('%0  %1',[msg.wparam, (msg.lparam shr 16) and $ff]);
for i:=0 to 255 do addlog('%0,',[KeyboardState[i]]);
addlog('------------------------');
}
                numchars:=ToUnicode(msg.wparam, (msg.lParam shr 16) and $FF, @KeyboardState[0], @buff[0], 10, 0);
                if (NumChars = 0) and (Mother^.State.OS = ostWin98) then begin
                  //Windows 98 doesn't have the Unicode support installed.
                  //  Assuming system locale to be Russian/Cyrillic CP1251
                  numchars:= ToAscii(msg.wparam, (msg.lParam shr 16) and $FF, @KeyboardState[0], @abuff[0], 0);
                  for i:=0 to numchars - 1 do
                    buff[i]:= Cp1251CharToUnicodeChar(abuff[i]);
                end;
              end;
              if ((msg.lParam and $FFFF ) = 1)
                or (Mother^.Input.TextInput and (TKey(scancode) in KeysAcceptingAutoRepeatInTextInputMode))
              then begin
                if ({(numchars < 1) and} (TKey(scancode) < KEY_MWHEEL_DOWN))
                  or (TKey(scancode) in KeysWorkingInTextInputMode)
                then begin
                  if (not f_KeyState[scancode])
                    or (Mother^.Input.TextInput and (TKey(scancode) in KeysAcceptingAutoRepeatInTextInputMode))
                  then begin
                    f_KeyState[scancode]:=true;
                    CheckIfMouseMoved();
                    OnPress(scancode);
                  end;
                end;
                for i:=0 to numchars - 1 do
                  if buff[i] >= ' ' //no special characters!
                    then begin
                      OnType(buff[i]);
                      OnPress(0); {notify the recipient of the updates,
                                   since OnType just adds to the internal buffer
                                   of the Chentrah module.
                                   The zero scan code is the mouse movement message}
                    end;
              end;
            end;
          WM_KEYUP:
            begin
              f_HadMessages:= true;
              scancode:= DivineScanCode(msg); //угадать скан-код на кофейной гуще
              if (scancode < ord(KEY_MWHEEL_DOWN))
                and f_KeyState[scancode]
              then begin
                f_KeyState[scancode]:=false;
                CheckIfMouseMoved();
                OnRelease(scancode);
              end;
            end;
          WM_SYSKEYDOWN:
            begin
//              if Mother^.Debug.Verbose then AddLog('Window received WM_SYSKEYDOWN (%0, %1)', [msg.wparam, msg.lparam]);
              f_HadMessages:= true;
            end;
          WM_SYSKEYUP:
            begin
//              if Mother^.Debug.Verbose then AddLog('Window received WM_SYSKEYUP (%0, %1)', [msg.wparam, msg.lparam]);
              f_HadMessages:= true; ;
            end;
          WM_LBUTTONDOWN: begin
            f_HadMessages:= true;
            if not f_KeyState[ord(KEY_MOUSE_LEFT)] then begin
              f_KeyState[ord(KEY_MOUSE_LEFT)]:=true;
              CheckIfMouseMoved();
              OnPress(ord(KEY_MOUSE_LEFT));
            end;
          end;
          WM_LBUTTONUP: begin
            f_HadMessages:= true;
            if f_KeyState[ord(KEY_MOUSE_LEFT)] then begin
              f_KeyState[ord(KEY_MOUSE_LEFT)]:=false;
              CheckIfMouseMoved();
              OnRelease(ord(KEY_MOUSE_LEFT));
            end;
          end;
          WM_RBUTTONDOWN:
            if not f_KeyState[ord(KEY_MOUSE_RIGHT)] then begin
              f_KeyState[ord(KEY_MOUSE_RIGHT)]:=true;
              CheckIfMouseMoved();
              OnPress(ord(KEY_MOUSE_RIGHT));
            end;
          WM_RBUTTONUP: begin
            f_HadMessages:= true;
            if f_KeyState[ord(KEY_MOUSE_RIGHT)] then begin
              f_KeyState[ord(KEY_MOUSE_RIGHT)]:=false;
              CheckIfMouseMoved();
              OnRelease(ord(KEY_MOUSE_RIGHT));
            end;
          end;
          WM_MBUTTONDOWN: begin
            f_HadMessages:= true;
            if not f_KeyState[ord(KEY_MOUSE_MIDDLE)] then begin
              f_KeyState[ord(KEY_MOUSE_MIDDLE)]:=true;
              CheckIfMouseMoved();
              OnPress(ord(KEY_MOUSE_MIDDLE));
            end;
          end;
          WM_MBUTTONUP: begin
            f_HadMessages:= true;
            if f_KeyState[ord(KEY_MOUSE_MIDDLE)] then begin
              f_KeyState[ord(KEY_MOUSE_MIDDLE)]:=false;
              CheckIfMouseMoved();
              OnRelease(ord(KEY_MOUSE_MIDDLE));
            end;
          end;
          WM_MOUSEMOVE: begin
            // the mouse movement is custom-handled now
            //f_HadMessages:= true; //probably won't work correctly in Win 2k & XP
          end;
          WM_MOUSEWHEEL: begin
              f_HadMessages:= true;
              i:=smallint(HIWORD(msg.wParam)) div 120;
              while i > 0 do begin
                CheckIfMouseMoved();
                OnPress(ord(KEY_MWHEEL_UP));
                dec(i);
              end;
              while i < 0 do begin
                CheckIfMouseMoved();
                OnPress(ord(KEY_MWHEEL_DOWN));
                inc(i);
              end;
            end;
(*
          WM_SYSCOMMAND: begin
            case msg.wParam of

              SC_CLOSE: begin
                LogGoDownWM('WM_SYSCOMMAND + SC_CLOSE');
                Mother^.State.ExitRequested:= Yes;
              end;

              SC_MAXIMIZE: begin
                if Mother^.Debug.Verbose then AddLog('Window received WM_SYSCOMMAND + SC_MAXIMIZE');
                f_maximized:= true;
                f_minimized:= false;
              end;
              SC_SCREENSAVE: begin
                if Mother^.Debug.Verbose then AddLog('Window received WM_SYSCOMMAND + SC_SCREENSAVE');
                f_maximized:= false;
              end;
              SC_MINIMIZE: begin
                if Mother^.Debug.Verbose then AddLog('Window received WM_SYSCOMMAND + SC_MINIMIZE');
                f_minimized:= true;
                f_maximized:= false;
                {  if Mother^.Display.InFullScreenMode then begin
                    LogGoDown(RuEn('окно потеряло фокус','the window lost focus'));
                    Mother^.State.ExitRequested:=true;
                  end;}
                end;
              SC_MONITORPOWER: begin
                if Mother^.Debug.Verbose then AddLog('Window received WM_SYSCOMMAND + SC_MONITORPOWER');
                f_maximized:= false;
                {if msg.lParam > 0 then begin
                  LogGoDown(RuEn('монитор выключается или переходит в спящий режим','your monitor shuts down or goes to sleep'));
                  Mother^.State.ExitRequested:=true;
                end}
              end;
              SC_RESTORE: begin
                if Mother^.Debug.Verbose then AddLog('Window received WM_SYSCOMMAND + SC_RESTORE');
                f_minimized:= false;
                f_maximized:= false;
              end;

            else
              if Mother^.Debug.Verbose then AddLog('Window received WM_SYSCOMMAND (wparam=%0, lparam=%1)', [intToHex(msg.wparam, 4), IntToHex(msg.lparam, 8)]);
            end;
          end;
*)
          WM_SIZE: begin
            case msg.wparam of
(*
              SIZE_MAXIMIZED: begin
                if Mother^.Debug.Verbose then AddLog('Window received WM_SIZE + SIZE_MAXIMIZED');
                f_minimized:= false;
                f_maximized:= true;
              end;
*)
              SIZE_RESTORED: begin
                if Mother^.Debug.Verbose then AddLog('Window received WM_SIZE + SIZE_RESTORED');
                f_HadMessages:= true;
//                  f_maximized:= false;
//                  f_minimized:= false;
//                UpdateWindowSize;
              end;
(*
              SIZE_MINIMIZED: begin
                if Mother^.Debug.Verbose then AddLog('Window received WM_SIZE + SIZE_MINIMIZED');
                f_minimized:= true;
                f_maximized:= false;
              end;
            else
              if Mother^.Debug.Verbose then AddLog('Window received WM_SIZE (wparam=%0, lparam=%1)', [intToHex(msg.wparam, 4), IntToHex(msg.lparam, 8)]);
*)
            end;
          end;
          WM_CLOSE:
            begin
              LogGoDownWM('WM_CLOSE');
              Mother^.State.ExitRequested:=true;
            end;
          WM_DESTROY:
            begin
              LogGoDownWM('WM_DESTROY');
              Mother^.State.ExitRequested:=true;
            end;
       //   $a0:;
       // else
       //   if Mother^.Debug.Verbose then AddLog('Window received: message=%0|%3, wparam=%1, lparam=%2', [IntToHex(msg.message, 4), IntToHex(msg.wparam, 4), IntToHex(msg.lparam, 8), msg.message]);
        end;
//AddLog('  message=%0|%3, wparam=%1, lparam=%2', [IntToHex(msg.message, 4), IntToHex(msg.wparam, 4), IntToHex(msg.lparam, 8), msg.message]);
        {if not msg.Message in [WM_QUIT, WM_CLOSE, WM_DESTROY] then} begin
          TranslateMessage(msg);
          DispatchMessage(msg);
        end;
      end;
//      end;
      if f_HadMessages then f_LoseFocusFramesCount:=0;

      if (f_state in [wsco_Normal, wsco_NoFocus])
        and (Mother^.State.OS = ostWine)
        and (gv_FramesSinceLeavingFullscreen < 3)
      then begin
        //The fuck...? I restore the window, then Wine suddenly moves it again,
        // to negative origin, so that the client area begins exactly at 0,0
        // and the window is overlapped with the XFCE taskbar
        cr:= ClipWindowToRect(Mother^.Display.WindowRect, Mother^.Display.ScreenRect);
        if (cr <> Mother^.Display.WindowRect) then begin
          if Mother^.Debug.Verbose then AddLog('Wine specific hack: our window happens to be outside the screen rect in less than 3 frames since leaving fullscreen. Correcting the window position.');
          if not MoveWindow(WindowHandle,
            cr.left, cr.top,cr.width,cr.height, false)
            then Die('MoveWindow() retuened false!');
        end;
      end;

      if f_state <> wsco_Maximized then inc(gv_FramesSinceLeavingFullscreen);
      if f_wsco <> f_state then begin
//addlog('*bp');
        case f_wsco of
          wsco_Maximized: begin
            Mother^.Display.InFullScreenMode:= true;
            GoFullscreen;
          end;
          wsco_Normal: begin
            case f_state of
              wsco_Maximized: begin
                Mother^.Display.InFullScreenMode:= false;
                LeaveFullScreen;
              end;
            end;
          end;
          wsco_NoFocus: begin
            case f_state of
              wsco_Maximized: begin
                Mother^.Display.InFullScreenMode:= false;
                LeaveFullScreen;
              end;
            end;
          end;
          wsco_Minimized: begin
            case f_state of
              wsco_Maximized: begin
                Mother^.Display.InFullScreenMode:= false;
                LeaveFullScreen;
              end;
            end;
          end;
        end;
        f_state:= f_wsco;
      end;
      NoteThreadOccupation(tlk_Working);
//addlog('* %0 %1',[ord(f_wsco), ord(f_state)]);
    end
    else begin // IsWindow(Mother^.Display.WindowHandle)
      LogGoDown(RuEn('окно приложения внезапно исчезло!','the application window has suddenly vanished!'));
      Mother^.Display.CanRender:= No;
      Mother^.Display.WindowExists:= No;
      Mother^.State.ExitRequested:= Yes;
    end;
  end;



initialization
  TAbstractFramework.RegisterFramework(TWinApiFramework, 'WinAPI');
finalization

end.

