{
    This file is part of Chentrah,
    Copyright (C) 2004-2016 Anton Rzheshevski (chebmaster@mail.ru).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see http://www.gnu.org/licenses/

 **********************************************************************}
{$longstrings on}
{$codepage utf-8}
{$include un_globaldefs.h}

unit framework_x11;


{
  Developed in 2007..2009 deliberately using outdated techniques for better compatibility
  Buried in 2012
  Raised in 2016 to shamble on
}

interface

uses
  SysUtils, classes
   , baseunix
   , unix
   , x
   , xlib
   , xutil
   , xatom //for XA_ATOM
   , dl
   , keysym

   , ctypes, math, cge, un_strings

   , framework_native
   , framework_x11_tablet
;


type
  TX11Framework = class (TNativeFramework)
  protected

    VisualInfo: PXVisualInfo;
    pnProtocols, fs_atoms : array [0..1] of TAtom;
    nWMProtocols : TAtom;
    icon_hwcursor: TCursor;
    Icon: TPixMap;
    TabletManager: TXinputTabletManager;
    attributes: TXSetWindowAttributes;
    f_width, f_height: integer;

    //implementing the abstract service functions introduced by the native framework
    procedure InitDisplay; override;
    procedure CreateWindow; override;
    procedure InitGLFirstStage; override;
    procedure InitGL; override;
    procedure CloseGL; override;
    procedure CloseWindow; override;
    procedure _UploadPointer; override;
    function GetScreenRect(fullscreen: boolean): TWindowManagerRect; override;//if not fullscreen then return the working area (excluding the taskbar)
    function GetWindowOuterRect(): TWindowManagerRect; override;   //Ask directly from the OS
    function GetWindowClientRect(): TWindowManagerRect; override;  //Ask directly from the OS
    function GetIsWindowFullscreen(): boolean; override; //Ask directly from the OS
    function GetIsWindowMinimized(): boolean; override; //Ask directly from the OS
    procedure CorrectFullscreenWindowSize; override; //if screen size changes between frames: for example, laptop cover open
    procedure _SwapBuffers; override;
    procedure Heartbeat; override; //processing window messages and such
    procedure UpdateWindowSize;
    //procedure GrabKeyboardAndMouseForFullscreen;
    //procedure UngrabKeyboardAndMouseForFullscreen;
    procedure UpdateX11WindowProperties;
    procedure GoFullscreen;
    procedure LeaveFullscreen;
    function AtomOrDie(Name: RawByteString; para2: system.boolean): TAtom;
  public
    display: PDisplay;
    screen: cint;
    RenderingContext: GLXContext;

    constructor Create; override;
    destructor Destroy; override;

    //implementing abstract functions that make up the framework API
    procedure SetWindowName(n: PUnicodeChar); override;
    procedure GrabPointer; override;
    procedure UngrabPointer; override;
  end;


implementation

 {$include framework_glx.inc}


  constructor TX11Framework.Create;
  begin
    inherited;

    Cge.Framework:= Self; //other classes need to access it.

    if Mother^.Debug.Enabled then AddLog('Starting the window manager...');


    InitWindowSubsystem;

    //CgeEmptyCursor;
    Self.SetPointerImage(nil, 0, 0);

    Console.InitGL;

    GamepadManager:= TGamepadManager.Create;
    GamepadManager.pointer_to_instance:= pointer(@Self.GamepadManager);
    EngineObject.AddObj(GamepadManager);

    Mother^.Display.BeginBlackout(INITIAL_BLACKOUT);

    Console.DrawBackground;
    Console.Draw;

    InitMotherAssets;

    LoadQF;

    if Mother^.Debug.Enabled then AddLog('Initializing sound...');

    try
      SoundManager:= TOpenALSoundman.Create;
      EngineObject.AddObj(SoundManager);
    except
      pointer(SoundManager):= nil;
      if Config.bool['sound', 'allow_running_without']
      then
        AddWarning (MI_CAPTION_NOSOUND, StopDying(),[])
      else
        AddStopMessage(StopDying()); //clear the error messages buffer by copying them into an error screen.
    end;


  end;

  function TX11Framework.AtomOrDie(Name: RawByteString; para2: system.boolean): TAtom;
  begin
    Result:= XInternAtom(Display, PChar(Name), para2);
    if Result = 0 then Die(RuEn(
      'X11 не может найти атом "%0"',
      'X11 cannot find atom "%0"'
      ), [Name]);
  end;

  procedure TX11Framework.InitDisplay;
  begin
    //Соединяемся с X сервером
    Display:= XOpenDisplay(nil);
    if not Assigned(Display) then Die(MI_CANT_CONNECT_TO_X_SERVER);
    screen:= XDefaultScreen(Display);
  end;

  destructor TX11Framework.destroy;
  begin
    ShutdownWindowSubsystem;
    pointer(Cge.Framework) := nil;
    inherited;
  end;


  procedure TX11Framework.SetWindowName(n: PUnicodeChar);
  var
    title: TXTextProperty;
    un: Utf8String;
  begin
    //  http://rus-linux.net/MyLDP/BOOKS/ProgrX/xwin-1-5.shtml

    un:= Utf8Encode(PCharToStr(n));
    FillChar(title, sizeof(title), 0);
    if 0 <>
      XUtf8TextListToTextProperty(Display, @(pointer(un)), 1, XUTF8StringStyle, @title)
        then Exit;

    XSetWMName(Display, WindowHandle, @title);
    XFlush(Display);
  end;

  procedure TX11Framework._SwapBuffers;
  begin
    case  Mother^.GAPI.Mode of
      {$ifndef glesonly}
      gapi_GL21, gapi_DeprecatedImmediateGL : begin
        if Assigned(glXSwapBuffers) then glXSwapBuffers(Display, WindowHandle);
      end;
      {$endif}
      gapi_GLES2: begin
        if Assigned(eglSwapBuffers) then eglSwapBuffers(eDisplay, eSurface);
      end
    else
      DieUnsupportedGLMode;
    end;
  end;


  function TX11Framework.GetWindowOuterRect(): TWindowManagerRect;   //Ask directly from the OS
  begin
    Result:= GetWindowClientRect(); //There is simply NO way.
    //Unlike WinAPI, X11 was simply not made with window decorations in mind
    //It operates with client rect only, leaving the window decorations to
    //tender mercies of whatever implementation of window manager.
  end;

  function TX11Framework.GetWindowClientRect(): TWindowManagerRect;  //Ask directly from the OS
  var
    wx, wy: cint;
    ww: TXID;
  begin
    //is received not from an API call but from window messages
    if not f_window_subsystem_initialized then begin
      //have to use hacks (the value read from the config) until window can receive messages
      Result:= Mother^.Display.WindowRect;
    end
    else begin
      XTranslateCoordinates(Display, WindowHandle, DefaultRootWindow(Display),
        0, 0, @wx, @wy, @ww);
      Result.left:= wx;
      Result.top:= wy;
      Result.width:= Mother^.Display.WindowClientRect.Width;//f_width;
      Result.height:= Mother^.Display.WindowClientRect.Height;//f_height;
    end;
  end;

  function TX11Framework.GetIsWindowFullscreen(): boolean; //Ask directly from the OS
  begin
    with Mother^.Display.WindowClientRect do begin
      Result:= (top = 0) and (left = 0) and (width = Mother^.Display.ScreenRect.width)
        and (height = Mother^.Display.ScreenRect.height)
    end;
  end;

  function TX11Framework.GetIsWindowMinimized(): boolean; //Ask directly from the OS
  begin
    Result:= false;
    //***TODO: implement it, too lazy now. Something about _NET_WM_STATE_HIDDEN or shit.
  end;

  procedure TX11Framework.CorrectFullscreenWindowSize;
  begin
    with Mother^.Display.ScreenRect do
      XMoveResizeWindow(Display, WindowHandle, left, top, width, height);
    XFlush(Display);
  end;








  {$include xunikey.inc}
     //ripped from the ptc/x11 package included with the FreePascal sources
     // do NOT open. Lovecraftian code inside.
     // http://www.gamedev.ru/code/forum/?id=90549&page=2#m21

  function GetXErrorString(dpy: PDisplay; code: integer): Utf8String;
  //XGetErrorText(display, code, buffer_return, length)
  //      Display *display;
  //      int code;
  //      char *buffer_return;
  //      int length;
  var buf: array[0..10000] of AnsiChar;
  begin
    XGetErrorText(dpy, code, @buf[0], 10000);
    Result:= PCharToStr(buf);
    if Result = '' then Result:= 'an unspecified error';
  end;

  procedure cge_XPanic(msg: UnicodeString); //dry up and die in a civilized manner
  begin
     Mother^.State.ExitRequested:= Yes;
     Mother^.Display.WindowExists:= No;
     Mother^.Display.CanRender:= No;
     Cge.XServerBully:= Yes;

     AddLog(MI_ERROR_X_SERVER_FATAL, [msg]);
     LogGoDown(RuEn(
       'X сервер неожиданно убил это приложение',
       'X server suddenly killed this application'));

     PerformCgeShutdown;
  end;

  //int (*XSetIOErrorHandler(handler))()
  //      int (*handler)(Display *);
  function _X_Fatal_Callback(para1:PDisplay):cint; cdecl; //never returns!
  begin
    //..if we return from this function, our app will die instantly.
    Mother^.Display.CanRender:= false;
    cge_XPanic('an unspecified I/O error');
  end;

  // int (*XSetErrorHandler(handler))()
  //       int (*handler)(Display *, XErrorEvent *)
  function _X_Error_Callback(dpy: PDisplay; Error: PXErrorEvent): cint; cdecl;
  begin
    Mother^.Display.CanRender:= false;
    cge_XPanic(GetXErrorString(dpy, Error^.error_code));
  end;

{
  const
    mask: array[0..1] of cardinal = (
      CWBackPixel or CWBorderPixel or CWEventMask
      ,
      CWOverrideRedirect or CWBackPixel or CWSaveUnder or CWBackingStore or CWEventMask
    );
}

  procedure TX11Framework.CreateWindow;
  var
   //attr: TXSetWindowAttributes;
   rep: integer;
   hints: TXWMHints;
   mask: culong;
  begin
   // InitGLFirstStage; //To get that f*** Visual up and running


    // would we not do that - our program would silently die at the first error,
    //   being unable to save the session.
    // One of the main reasons I did not use SDL, glut and the like.
    XSetIOErrorHandler(@_X_Fatal_Callback);
    XSetErrorHandler(@_X_Error_Callback);

    FillChar(attributes, sizeof(attributes), 0);
    attributes.colormap:= XCreateColormap( Display, DefaultRootWindow(Display), VisualInfo^.visual, AllocNone );
    attributes.event_mask:= ExposureMask or KeyPressMask
       or KeyReleaseMask or ButtonPressMask or ButtonReleaseMask
       or PointerMotionMask or SubstructureNotifyMask or FocusChangeMask
       or VisibilityChangeMask or StructureNotifyMask
       or EnterNotify or LeaveNotify;

    mask:= CWColormap or CWEventMask or CWBorderPixel or CWBackPixel or CWOverrideRedirect ;

    if Mother^.Debug.Verbose then AddLog('Screen rect is %0',[RectToStr(Mother^.Display.ScreenRect)]);

    if Mother^.Debug.Verbose then AddLog('The window rect, as read from the config and clipped to the screen rect, should be %0',[RectToStr(Mother^.Display.WindowRect)]);

    if Mother^.Debug.Verbose then AddLog('Creating the window...');


    with Mother^.Display.WindowRect do
//***TODO: Wrong rect is used! WindowRect stores size including decorations but here it is used to set client size!
      WindowHandle:= XCreateWindow(display,
        DefaultRootWindow(display),
        top, left, width, height,
        0,
        VisualInfo^.depth,//CopyFromParent,
        InputOutput,
        VisualInfo^.visual,//NIL, //CopyFromParent, //Oops, cannot assign 0 to apointer in objfpc mode
        mask,
        @attributes
      );

{
WindowHandle:= XCreateWindow(display, DefaultRootWindow(display),
     (InitialWidth - Mother^.Display.WindowClientWidth) div 2,
     (InitialHeight - Mother^.Display.WindowClientHeight) div 2,

     Mother^.Display.WindowClientWidth,Mother^.Display.WindowClientHeight,
     0,CopyFromParent,InputOutput,CopyFromParent,
     mask[ord(Mother^.Display.InFullScreenMode)], @attr);


}

    if WindowHandle = 0 then Die('XCreateWindow() returned 0');

   (* Выбираем события, обрабатываемые программой *)

    XSelectInput (
         Display, WindowHandle,
         ExposureMask or KeyPressMask
           or KeyReleaseMask or ButtonPressMask or ButtonReleaseMask
           or PointerMotionMask or FocusChangeMask
           or VisibilityChangeMask
           or EnterWindowMask or LeaveWindowMask
    );


   (* Чтобы ловило оповещения о закрытии *)

    pnProtocols[0]:= AtomOrDie('WM_DELETE_WINDOW', False);
    pnProtocols[1]:= XInternAtom(Display, 'WM_SAVE_YOURSELF', True);
                    //aw, chucks. Modern Linuxes don't have it :(
                    //AtomOrDie('WM_SAVE_YOURSELF', True);
    XSetWMProtocols(Display, WindowHandle, @pnProtocols[0], 2);

    nWMProtocols:= AtomOrDie('WM_PROTOCOLS', True);

   (* Показываем окно *)
    XMapWindow ( display, Windowhandle );
    XFlush(display);

    Mother^.Display.WindowExists:=True;

    if Mother^.Display.InFullScreenMode
      then GoFullscreen
      else LeaveFullscreen;

   (* Without this, the input won't work due to override_redirect *)
    //if Mother^.Display.InFullScreenMode then GrabKeyboardAndMouseForFullscreen;

       if Mother^.Debug.Verbose then AddLog('Starting the pen tablet manager...');
     TabletManager:= TXinputTabletManager.Create(display, pointer(WindowHandle));
  end;


  procedure TX11Framework.UpdateX11WindowProperties;
  var
    event: TXEvent;
    sizehints : TXSizeHints;
    rect: TWindowManagerRect;
  begin
    XSync( Display, 1 );


    FillChar( sizehints, SizeOf( TXSizeHints ), 0 );
    if Mother^.Display.InFullScreenMode then begin
      sizehints.flags    := PBaseSize or PWinGravity;
      with Mother^.Display.ScreenRect do begin
        sizehints.x          := left;
        sizehints.y          := top;
        sizehints.width      := width;
        sizehints.height     := height;
        sizehints.min_width  := width;
        sizehints.max_width  := width;
        sizehints.min_height := height;
        sizehints.max_height := height;
      end;
    end
    else begin
      sizehints.flags    := PPosition or PSize or PMinSize or PMaxSize;
      with Mother^.Display.WindowRect do begin
        sizehints.x          := left;
        sizehints.y          := top;
        sizehints.width      := width;
        sizehints.height     := height;
        sizehints.min_width  := 256;
        sizehints.max_width  := Mother^.Display.ScreenRect.Width;
        sizehints.min_height := 256;
        sizehints.max_height := Mother^.Display.ScreenRect.Height;
      end;
    end;
    XSetWMNormalHints( Display, WindowHandle, @sizehints );

    FillChar( event, SizeOf( TXEvent ), 0 );
    event._type                := ClientMessage;
    event.xclient._type        := ClientMessage;
    event.xclient.send_event   := 1;
    event.xclient.window       := WindowHandle;
    event.xclient.message_type := AtomOrDie('_NET_WM_STATE', FALSE );
    event.xclient.format       := 32;
    event.xclient.data.l[ 0 ]  := ord(Mother^.Display.InFullScreenMode);
    event.xclient.data.l[ 1 ]  := AtomOrDie('_NET_WM_STATE_FULLSCREEN', FALSE );
    XSendEvent( Display, DefaultRootWindow(Display), False, SubstructureRedirectMask or SubstructureNotifyMask, @event );

    XFlush(Display);
  end;

  procedure TX11Framework.GoFullscreen;
{  var
    attr: TXSetWindowAttributes;}
  begin
    //FRAK. I can't seem to be able to make it work :(...
    f_wsco:= wsco_Normal;
    AddLog(RuEn(
      'Фреймворк X11 в настоящее время поддерживает только оконный режим',
      'The X11 framework currently only supports windowed mode'
    ));
    Config.Bool['video', 'fullscreen']:= false;
    Exit;

    Mother^.Display.ScreenRect:= GetScreenRect(true);
    AddLog(RuEn(
      'Переключение в полноэкранный режим (%0х%1)...',
      'Switching to full screen mode (%0x%1)...'
      ), [Mother^.Display.ScreenRect.Width, Mother^.Display.ScreenRect.Height]
    );

    Mother^.Display.InFullScreenMode:= true;

    UpdateX11WindowProperties;

    with Mother^.Display.ScreenRect do XMoveResizeWindow(Display, WindowHandle, left, top, width, height);

 {   FillChar(attr, sizeof(attr), 0);
    attr.override_redirect:= 0;
    XChangeWindowAttributes ( Display, WindowHandle, CWOverrideRedirect, @attr);
}


//    UpdateX11WindowProperties;

    XFlush(Display);

    Mother^.Display.WindowRect:= GetWindowOuterRect;
    Mother^.Display.WindowClientRect:= GetWindowClientRect;
    Config.Bool['video', 'fullscreen']:= true;
    AddLogOk;
  end;

  procedure TX11Framework.LeaveFullScreen;
{
  var
    attr: TXSetWindowAttributes;
}
  begin
    Config.Bool['video', 'fullscreen']:= false;
    if (Mother^.Display.InFullScreenMode) then AddLog(RuEn(
      'Переключение из полноэкранного в оконный режим...',
      'Switching from fullscreen to windowed mode...'
    ));
    FramesSinceLeavingFullscreen:= 0;

    Mother^.Display.InFullScreenMode:= false;

    LoadWindowRectFromConfig;

{
    FillChar(attr, sizeof(attr), 0);
    attr.override_redirect:= 1;
    XChangeWindowAttributes ( Display, WindowHandle, CWOverrideRedirect, @attr);
}
    UpdateX11WindowProperties;

    with Mother^.Display.WindowRect do
      XMoveResizeWindow(Display, WindowHandle, left, top, width, height);
    XFlush(Display);

    Mother^.Display.WindowRect:= GetWindowOuterRect;
    Mother^.Display.WindowClientRect:= GetWindowClientRect;
    if Mother^.Debug.Verbose then AddLog('  the client area is %0, the entire window including decorations is %1',
      [RectToStr(Mother^.Display.WindowClientRect), RectToStr(Mother^.Display.WindowRect)]);
  end;

(*
  procedure TX11Framework.GrabKeyboardAndMouseForFullscreen;
  var
   rep: integer;
  begin
    rep:=0;
    repeat
      inc(rep);
      if rep > 10 then break;
      if rep > 0 then Sleep(50);
    until 0 = XGrabKeyboard(display, WindowHandle, true,
        GrabModeAsync, GrabModeAsync, CurrentTime);
    if rep > 10 then Die(RuEn(
      'Не удалось перехватить клавиатуру',
      'Unable to grab keyboard'));

    rep:=0;
    repeat
      inc(rep);
      if rep > 10 then break;
      if rep > 0 then Sleep(50);
    until 0 = XGrabPointer(display, WindowHandle, true,
        ButtonPressMask or ButtonReleaseMask or PointerMotionMask,
        GrabModeAsync, GrabModeAsync, WindowHandle, None, CurrentTime);
    if rep > 10 then
      Die(RuEn(
        'Не удалось перехватить мышь',
        'Unable to grab mouse'));
  end;

  procedure TX11Framework.UngrabKeyboardAndMouseForFullscreen;
  begin
    XUngrabPointer(Display, CurrentTime);
    XUngrabKeyboard(Display, CurrentTime);
  end;
*)
  procedure  TX11Framework.CloseWindow;
  begin
    TabletManager.Free;
    pointer(TabletManager):= nil;

    if not Mother^.Display.WindowExists then begin
      if Mother^.Debug.Verbose then AddLog('Cannot destroy the window because it doesn''t exist.');
      Exit
    end;

    if CursorCreated then begin
      XUndefineCursor(display, WindowHandle);
      XFlush(display);
      XFreeCursor(display, icon_hwcursor);
      XFlush(display);
    end;

    XDestroyWindow(Display, WindowHandle);

    XFlush(display); {without this line, the fullscreen window would
      freeze, not truly destroyed, hide the error message dialog,
      block any switching via alt-tab...
     resulting in one livid user who *ahem* doesn't like us too much
      after losing all his open documents, having to restart the X.}

    if Mother^.Debug.Verbose then AddLog('Window destroyed.');
    Mother^.Display.WindowExists:=False;
  end;


  function DivineScanCode(var event: TXEvent): integer;  {
    Outputs key codes in the IBM PC standard (necessary for
    uniformity, the game module must be as little platform dependent as possible)
  }
  var
    sym: TKeySym;
  begin
    sym:= XLookupKeysym(@event.xkey, 0);
    case sym of
      XK_Home               :Exit(ord(KEY_HOME));
      XK_Left               :Exit(ord(KEY_LEFT));
      XK_Up                 :Exit(ord(KEY_UP));
      XK_Right              :Exit(ord(KEY_RIGHT));
      XK_Down               :Exit(ord(KEY_DOWN));
      XK_Page_Up            :Exit(ord(KEY_PGUP));
      XK_Page_Down          :Exit(ord(KEY_PGDN));
      XK_End                :Exit(ord(KEY_END));
      XK_BackSpace          :Exit(ord(KEY_BACKSPACE));
      XK_Tab                :Exit(ord(KEY_TAB));
      XK_Return             :Exit(ord(KEY_ENTER));
      XK_Pause              :Exit(ord(KEY_PAUSE));
      XK_Scroll_Lock        :Exit(ord(KEY_SCROLL_LOCK));
      XK_Sys_Req            :Exit(ord(KEY_SYSRQ));
      XK_Escape             :Exit(ord(KEY_ESCAPE));
      XK_Delete             :Exit(ord(KEY_DELETE));
      XK_Insert             :Exit(ord(KEY_INSERT));
      XK_Menu               :Exit(ord(KEY_APP_MENU));
      XK_Num_Lock           :Exit(ord(KEY_NUM_LOCK));
      XK_KP_Enter           :Exit(ord(KEY_KP_ENTER));
      XK_KP_Home, XK_KP_7   :Exit(ord(KEY_KP_7));
      XK_KP_Left, XK_KP_4   :Exit(ord(KEY_KP_4));
      XK_KP_Up, XK_KP_8     :Exit(ord(KEY_KP_8));
      XK_KP_Right, XK_KP_6  :Exit(ord(KEY_KP_6));
      XK_KP_Down, XK_KP_2   :Exit(ord(KEY_KP_2));
      XK_KP_Page_Up, XK_KP_9  :Exit(ord(KEY_KP_9));
      XK_KP_Page_Down, XK_KP_3:Exit(ord(KEY_KP_3));
      XK_KP_End, XK_KP_1    :Exit(ord(KEY_KP_1));
      XK_KP_Insert, XK_KP_0 :Exit(ord(KEY_KP_0));
      XK_KP_Delete, XK_KP_Decimal :Exit(ord(KEY_KP_DECIMAL));
      XK_KP_Multiply        :Exit(ord(KEY_KP_STAR));
      XK_KP_Add             :Exit(ord(KEY_KP_PLUS));
      XK_KP_Subtract        :Exit(ord(KEY_KP_MINUS));
      XK_KP_Divide          :Exit(ord(KEY_KP_DIVIDE));

      XK_Shift_L            :Exit(ord(KEY_LEFT_SHIFT));
      XK_Shift_R            :Exit(ord(KEY_RIGHT_SHIFT));
      XK_Control_L          :Exit(ord(KEY_LEFT_CTRL));
      XK_Control_R          :Exit(ord(KEY_RIGHT_CTRL));
      XK_Caps_Lock          :Exit(ord(KEY_CAPS_LOCK));

      XK_Alt_L              :Exit(ord(KEY_LEFT_ALT));
      XK_Alt_R              :Exit(ord(KEY_RIGHT_ALT));
      XK_Super_L            :Exit(ord(KEY_LEFT_WINDOWS));
      XK_Super_R            :Exit(ord(KEY_RIGHT_WINDOWS));
    else
      Result:= (event.xKey.keycode - 8) and $FF;  // A hack!
    end;
  end;


  function TX11Framework.GetScreenRect(fullscreen: boolean): TWindowManagerRect;
    var
      _Screen: PScreen;
      wa: TXWindowAttributes;
  begin
    if fullscreen then begin
      _Screen:= XScreenOfDisplay(display, 0);
      Result.left:= 0;
      Result.top:= 0;
      Result.width:= _Screen^.width;
      Result.height:= _Screen^.height;
    end
    else begin
      XGetWindowAttributes(Display, DefaultRootWindow(Display), @wa);
      Result.left:= wa.x;
      Result.top:= wa.y;
      Result.width:= wa.width;
      Result.height:= wa.height;
    end;
  end;


  Procedure TX11Framework.Heartbeat;
  var
    Event, NextEvent: TXEvent;
    i, s: integer;
    sym: TKeySym;
    scancode: integer;
    ItsAutoRepeat, GotNext: boolean;

    procedure CheckIfMouseMoved(per_cycle: boolean = false);
    begin

    end;



  begin
    if GoingDownHard then Exit;

    NoteThreadOccupation(tlk_PollingInput);

    _CheckCursorMode();
    f_HadMessages:= false;

    CheckWindowSize;

    f_focus:= (f_wsco in [wsco_Normal, wsco_Maximized]);

    While XPending(Display) > 0 do begin
      XNextEvent ( Display, @Event );
      if Event._type = KeyRelease then begin
        scancode:= DivineScanCode(event);
        //defeat the auto-repeat feature!
        GotNext:= No;
        ItsAutoRepeat:= No;
        if XPending(Display) > 0 then begin
          GotNext:= Yes;
          XNextEvent ( Display, @NextEvent );
          ItsAutoRepeat:= (NextEvent._type = KeyPress) and (NextEvent.xkey.keycode = Event.xkey.keycode) and (NextEvent.xkey.time = Event.xkey.time);
        end;
        if f_KeyState[scancode] and not ItsAutoRepeat then begin
          f_KeyState[scancode]:= false;
          OnRelease(scancode);
        end;
        if GotNext then Event:= NextEvent;
      end;

      case Event._type of
        KeyPress: begin
          f_HadMessages:= true;
          scancode:= DivineScanCode(event);
          i:= -1;
          if Mother^.Input.TextInput then begin
            //http://www.gamedev.ru/code/forum/?id=90549&page=2#m18

            {Absolutely empiric! There are no docs on it,
              it just happens that the Russian keyboard layout is indicated
              as the bit 13 of the "state" field.
             Correspondingly, XLookUpKeysum converts key codes to Russian
              keysyms if the bit 1 of its "index" parameter is set. Go figure.
            }
            s:= (event.xkey.state and $1) or ((event.xkey.state shr (13 - 1)) and 2);
            sym:= XLookupKeysym(@event.xkey, s);
            if sym = 0
              then sym:= XLookupKeysym(@event.xkey, s and 1);

            //[Shmuck Bait] Look up the source of this function (xunikey.inc).
            //Your eyes will not bleed. Honest.
            i:= X11ConvertKeySymToUnicode(sym);

          end;
          if i > 0 then begin
            OnType(UnicodeChar(i));
            OnPress(0);
          end;
          if (not f_KeyState[scancode]) or (ItsAutoRepeat and Mother^.Input.TextInput
            and (TKey(scancode) in KeysAcceptingAutoRepeatInTextInputMode))
          then begin
            f_KeyState[scancode]:=true;
            if (i <= 0) or (TKey(scancode) in KeysWorkingInTextInputMode)
              then OnPress(scancode);
          end;
        end;

        ClientMessage: begin
          if (Event.xclient.message_type = nWMProtocols)
            and (Event.xclient.data.l[0] = pnProtocols[0])
          then begin
           { cge_XPanic('Received WM_DELETE_WINDOW message');} {immediate shutdown.
              I'm trying to hack around mysterious lockups here.
              The lockups happen if you close the window,
              it started after I added OpenAL support. Chentrah process gets
              stuck in the Zombie state, and even kill from root doesn't work,
              only the X server restart.
            P.S. Resolved, was a bug in the TModule incorrectly setting Mother^.Module.CurrentlyExecutingInd
              state which led to the array out-of-bounds error and trashed memory}

            LogGoDownWM('WM_DELETE_WINDOW');
            Mother^.State.ExitRequested:= Yes;
            break;
          end;
          if (Event.xclient.message_type = nWMProtocols)
            and (Event.xclient.data.l[0] = pnProtocols[1])
          then begin
            //cge_XPanic('Received WM_SAVE_YOURSELF message');
            LogGoDownWM('WM_SAVE_YOURSELF');
            Mother^.State.ExitRequested:=Yes;
            break;
          end;
        end;
        FocusIn: begin
          case f_wsco of
            wsco_Maximized:; //do nothing
          else
            f_wsco:= wsco_Normal;
          end;
        end;
        FocusOut: begin
          f_wsco:= wsco_NoFocus;
        end;
        ButtonPress: begin
          f_HadMessages:= true;
          f_MouseX:=event.xbutton.x;
          f_MouseY:=event.xbutton.y;
          case (Event.xbutton.button) of
            Button1: begin
              f_KeyState[ord(KEY_MOUSE_LEFT)]:=true;
              OnPress(ord(KEY_MOUSE_LEFT));
            end;
            Button3: begin
              f_KeyState[ord(KEY_MOUSE_RIGHT)]:=true;
              OnPress(ord(KEY_MOUSE_RIGHT));
            end;
            Button2: begin
              f_KeyState[ord(KEY_MOUSE_MIDDLE)]:=true;
              OnPress(ord(KEY_MOUSE_MIDDLE));
            end;
            Button4:
              OnPress(ord(KEY_MWHEEL_UP));
            Button5:
              OnPress(ord(KEY_MWHEEL_DOWN));
          end;
        end;
        VisibilityNotify: begin
          Mother^.Display.WindowVisible:= (Mother^.Display.InFullScreenMode or (event.xvisibility.state <> VisibilityFullyObscured));
        end;
        ButtonRelease: begin
          f_HadMessages:= true;
          f_MouseX:=event.xbutton.x;
          f_MouseY:=event.xbutton.y;
          case (Event.xbutton.button) of
            Button1: begin
              f_KeyState[ord(KEY_MOUSE_LEFT)]:=false;
              OnRelease(ord(KEY_MOUSE_LEFT));
            end;
            Button3: begin
              f_KeyState[ord(KEY_MOUSE_RIGHT)]:=false;
              OnRelease(ord(KEY_MOUSE_RIGHT));
            end;
            Button2: begin
              f_KeyState[ord(KEY_MOUSE_MIDDLE)]:=false;
              OnRelease(ord(KEY_MOUSE_MIDDLE));
            end;
          end;
        end;
        MotionNotify: begin
          f_HadMessages:= true;
          f_MouseX:=event.xmotion.x;
          f_MouseY:=event.xmotion.y;
          OnMouseMove;
        end;
        EnterNotify: begin
          f_HadMessages:= true;
          Mother^.Display.MouseIsBeyondOurWindow:= false;
          f_MouseX:=event.xcrossing.x;
          f_MouseY:=event.xcrossing.y;
          OnMouseMove;
        end;
        LeaveNotify: begin
          Mother^.Display.MouseIsBeyondOurWindow:= true;
        end;
        ResizeRequest: begin
          f_width:= event.xresizerequest.Width;
          f_height:= event.xresizerequest.Height;
        end;
      else
        if Assigned (TabletManager) and (Event._type = TabletManager.ValuatorEvent)
          then with PXDeviceMotionEvent(@Event)^ do begin
            if (axes_count <> length(TabletManager.AxisInfo))
            or (first_axis <> 0)
              then Die(RuEn('Абнормальный формат сообщения от графического планшета!','Abnornal pen tablet input message!'));
            OnPen(
              x - x_root + ((axis_data[0] - tabletManager.AxisInfo[0].min_value) * Mother^.Display.ScreenRect.Width) / (tabletManager.AxisInfo[0].max_value - tabletManager.AxisInfo[0].min_value + 1),
              y - y_root + ((axis_data[1] - tabletManager.AxisInfo[1].min_value) * Mother^.Display.ScreenRect.Height) / (tabletManager.AxisInfo[1].max_value - tabletManager.AxisInfo[1].min_value + 1),
              axis_data[2]  / tabletManager.AxisInfo[2].max_value, //pressure
              axis_data[3]  / tabletManager.AxisInfo[3].max_value, //reserved1
              axis_data[4]  / tabletManager.AxisInfo[4].max_value, //reserved2
              axis_data[5]  / tabletManager.AxisInfo[5].max_value //reserved3
            );
{addlog (' pen event: x=%0, y=%1, pressure=%2', [
              ((axis_data[0] - tabletManager.AxisInfo[0].min_value) * Mother^.Display.WindowClientWidth) / (tabletManager.AxisInfo[0].max_value - tabletManager.AxisInfo[0].min_value + 1),
              ((axis_data[1] - tabletManager.AxisInfo[1].min_value) * Mother^.Display.WindowClientHeight) / (tabletManager.AxisInfo[1].max_value - tabletManager.AxisInfo[1].min_value + 1),
              axis_data[2]  / tabletManager.AxisInfo[2].max_value
              ]);
}
          end;
      end;
    end;

    if f_focus or f_HadMessages
      then f_LoseFocusFramesCount:=0
      else inc(f_LoseFocusFramesCount);

    if f_wsco <> f_state then begin
      case f_wsco of
        wsco_Maximized: begin
          (*
          f_wsco:= wsco_Normal;
          AddLog(RuEn(
            'Фреймворк X11 в настоящее время поддерживает только оконный режим',
            'The X11 framework currently only supports windowed mode'
          ));
          //Mother^.Display.InFullScreenMode:= false;
          //Mother^.Display.InFullScreenMode:= true;
          *)
          GoFullscreen;
        end;
        wsco_Normal: begin
          case f_state of
            wsco_Maximized: begin
              Mother^.Display.InFullScreenMode:= false;
              LeaveFullScreen;
            end;
          end;
        end;
        wsco_NoFocus: begin
          case f_state of
            wsco_Maximized: begin
              Mother^.Display.InFullScreenMode:= false;
              LeaveFullScreen;
            end;
          end;
        end;
        wsco_Minimized: begin
          case f_state of
            wsco_Maximized: begin
              Mother^.Display.InFullScreenMode:= false;
              LeaveFullScreen;
            end;
          end;
        end;
      end;
      f_state:= f_wsco;
    end;

    UpdateWindowSize;

    NoteThreadOccupation(tlk_Working);
  end;

  procedure TX11Framework.UpdateWindowSize;
  var
    attr: TXWindowAttributes;
  begin
    if GoingDownHard then Exit;

    XGetWindowAttributes(Display, WindowHandle, @attr);
    with Mother^.Display.WindowClientRect do
      if (attr.width <> Width) or (attr.height <> Height)
        or (attr.x <> left) or (attr.y <> top)
      then begin
        Width:= attr.width;
        Height:= attr.height;
        left:= attr.x;
        top:= attr.y;
      end;
  end;


  procedure TX11Framework.GrabPointer;
  begin
    if f_PointerGrabbed then Exit;
    if 0 <> XGrabKeyboard(display, WindowHandle, true, GrabModeAsync, GrabModeAsync, CurrentTime)
    then Die(RuEn('Не удалось перехватить клавиатуру','Unable to grab keyboard'));
    if 0 <> XGrabPointer(display, WindowHandle, true, ButtonPressMask or ButtonReleaseMask or PointerMotionMask, GrabModeAsync, GrabModeAsync, WindowHandle, None, CurrentTime)
    then Die(RuEn('Не удалось перехватить мышь','Unable to grab mouse'));
    SetPointerImage(nil, 0, 0);
    f_PointerGrabbed:= True;
  end;


  procedure TX11Framework.UngrabPointer;
  begin
    if not f_PointerGrabbed then Exit;
    XUngrabPointer(display, CurrentTime);
    XUngrabKeyboard(display, CurrentTime);
    f_PointerGrabbed:= False;
    SetPointerImage(@PointerCache, PointerXHotSpot, PointerYHotSpot);
  end;



  procedure TX11Framework._UploadPointer;
  var
    PixmapF, PixmapB: TPixmap;
    cursor_fg, cursor_bg : TXColor;
    screen_colormap : TColormap;
    rc : TStatus;
    GC : TGC;
    rValues : TXGCValues;
    x, y: integer;
    OldCursor: TCursor;
  begin
    _UploadGLcursor;
    OldCursor:= icon_hwcursor;

    screen_colormap := XDefaultColormap(display, XDefaultScreen(display));//screen);
    (* выделяем черный и белый цвета *)
    if 0 = XAllocNamedColor(display, screen_colormap, 'black', @cursor_fg, @cursor_fg)
      then Die('XAllocNamedColor() failed: unable to alloc "black"');
    if 0 = XAllocNamedColor(display, screen_colormap, 'white', @cursor_bg, @cursor_bg)
      then Die('XAllocNamedColor() failed: unable to alloc "white"');
    PixmapF:= XCreatePixmap(display, WindowHandle, 32, 32, 1);
    PixmapB:= XCreatePixmap(display, WindowHandle, 32, 32, 1);

    //all this stuff only to fill the bitmap with zeroes
    rValues.foreground := 0;
    rValues.background := 0;
    GC:= XCreateGC(Display, PixmapF, GCForeground, @rValues);
    //create a blank transparent cursor at first
    XSetForeground(display, GC, 0);
    for y:=0 to 31 do
      for x:= 0 to 31 do
        XDrawPoint(Display, PixmapF, GC, x, y);
    if Mother^.Display.UseHardware1bitCursor then begin
      //fill it with the mask data
      XSetForeground(display, GC, 1);

      for y:=0 to 31 do
        with PointerBitmap do
          for x:= 0 to 31 do
            if (XorMask[y] and (1 shl x)) <> 0 //white pixel
              then XDrawPoint(Display, PixmapF, GC, x, y);
    end;
    XFreeGC(Display, GC);


    GC:= XCreateGC(Display, PixmapB, GCForeground, @rValues);
    XSetForeground(display, GC, 0);
    for y:=0 to 31 do
      for x:= 0 to 31 do
        XDrawPoint(Display, PixmapB, GC, x, y);
    if Mother^.Display.UseHardware1bitCursor then begin
      XSetForeground(display, GC, 1);
      for y:=0 to 31 do
        with PointerBitmap do
          for x:= 0 to 31 do
            if (AndMask[y] and (1 shl x)) = 0 // black pixel
                then XDrawPoint(Display, PixmapB, GC, x, y);
    end;
    XFreeGC(Display, GC);

    icon_hwcursor := XCreatePixmapCursor(
      display, PixmapF, PixmapB, @cursor_fg, @cursor_bg, PointerXHotSpot, PointerYHotSpot);
    XFreePixmap(display, PixmapF);
    XFreePixmap(display, PixmapB);
    XDefineCursor(display, WindowHandle, icon_hwcursor);
    if CursorCreated then begin
      XFreeCursor(display, OldCursor);
      XFlush(display);
    end;
    CursorCreated:= Yes;
  end;



initialization
  // A rare case where I use the initialization section.
  // For the main executable, it works. It only does not work in Linux DLLs.

  TAbstractFramework.RegisterFramework(TX11Framework, 'X11');
finalization

end.
