{
    This file is part of Chentrah,
    Copyright (C) 2004-2016 Anton Rzheshevski (chebmaster@mail.ru).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see http://www.gnu.org/licenses/

 **********************************************************************}

{$include mo_globaldefs.h}
{$typeinfo on}

unit mi_terrain;

interface
uses SysUtils, Classes {$ifdef windows}, Windows {$else} , baseunix, dl {$endif},
  mo_hub, chepersy, mo_classes, syncobjs, mo_threads, mo_assets, mo_texture;

{
  1. Terrain mesh represents a world chunk
  2. Terrain mesh is considered small enough to be rebuild with any terrain
    changes, there are no incremental optimizations (roughly 150..500 tris per
    mesh, may be much higher in rare instances)
  3. Light values are stored in a texture, can be updated without rebuilding
    the terrain mesh. Light values are per-vertex, four-component, RGB value and
    concaveness for AO. Glowing surfaces like lava store their brightness
    in the texturing parameters.
    Doors are tetrameshes, together with their doorframes, it's how light gets
    clipped by a closed door. Not touched here.
  4. Texturing, including decals, CAN be updated incrementally without
    rebuilding the terrain mesh
  5. Terrain meshes are grouped together for storage and rendering as their
    number is roughly estimated to be around 3000.

}

type
{
  TTerrainTextureKind = (ttk_);

  TTerrainTextureAtlas = class(TGenericAsset)
  protected
    f_init_task: TAbstractThreadedTask;
    function _GetActualized: boolean; override;
  public
    procedure RegisterFields; override;
    procedure OnModuleDestroy; override;
    function IsValidForServer: boolean; override;
    function Actualize: boolean; override;
    function Devour(c: TGenericAsset): boolean; override;
    procedure EraseContainer; override;
  end;}

  TMeshDataTexture = class(TTexture)
  protected

  public

  end;

  TMeshDataManager = class(TGenericAsset)
  protected

  public
    function FindSuitableTexture(size: integer); virtual;
  end;

  TTerrainMeshDataManager = class(TMeshDataManager)
  protected

  public

  end;

  TTerrainMesh = class(TGenericAsset)
  protected

  public
    DataTexture: TMeshDataTexture;
    Vertices: TVector3fArray;
    Light: TVercor4fArray;
    Tris: TWordArray;

    {
      The mesh does NOT store vertices }
    constructor Create ()
  end;

implementation

end.

