var
  gv_RetrievingFromMother: boolean = false;
  gv_StoringToMother: boolean = false;


  procedure TGenericAsset.RegisterFields;
  begin
    RegType(typeinfo(TAssetHash), sizeof(TAssetHash), ['Q1', 'Q2', 'Q3', 'Q4', typeinfo(qword)]);
    //RegType('*TAssetHash', typeinfo(dword), [0,3]); //the size is set in concrete.
    RegType(typeinfo(TAssetArray));//, typeinfo(TGenericAsset));
    RegType(typeinfo(TAssetPropertiesSet));//, typeinfo(TAssetPropertiesEnum));
    Inherited;
    ListFields([
      'properties', @f_props, typeinfo(TAssetPropertiesSet),
      'hash', @Hash, typeinfo(TAssetHash),//'*TAssetHash',
      '-InQueueSinceFrame', @InQueueSinceFrame, typeinfo(qword),
      '-NextInQueue', @NextInQueue, CPS_POINTER //do NOT save
    ]);
  end;

  procedure THandleAsset.RegisterFields;
  begin
    inherited;
    ListFields([
      '_handle', @_handle, typeinfo(ptruint)
    ]);
  end;

  //Much cooler than simply declaring a method virtual
  procedure TGenericAsset.DieBecauseStupidDeveloperForgotToOverrideMethod(s: UnicodeString);
  begin
    Die(RuEn(
      'В классе %0 забыли переопределить метод %1()',
      'Someone forgot to override the %1() method in class %0'
    ), [string(Self.ClassName), s]);
  end;

  function TGenericAsset._GetActualizedNilSafe: boolean;
  begin
    if not Assigned(Self) then Exit(false);
    Result:= _GetActualized;
  end;

  function TGenericAsset._GetActualized: boolean;
  begin
    Result:= ap_actualized in f_props;
  end;

  procedure TGenericAsset.MarkAsActualized;
  begin
    if IsHashZero(f_hash)
      then Die (MI_ERROR_PROGRAMMER_NO_BAKA, [
       PervertedFormat(RuEn(
        'Ресурс %0: попытка пометить как актуализированный при пустом хэше',
        'Asset %0: attempt to mark as actualized while the hash is empty'
        ), [string(Self.ClassName)])]);
    Include(f_props, ap_actualized);
  end;


  procedure TGenericAsset.OnZeroReferenceCount;
  begin
    if not Self.MustBeFreedFromMainThread()
      or (GetCurrentThreadId() = Mother^.State.MainThreadId)
    then inherited //cal Destroy
    else begin
      EnterWv(TAssetManager.ToDeleteInMainThreadCS); //wrapper to allow the thread load indicator profiling this
      if not Assigned(TAssetManager.ToDeleteInMainThread)
        then TAssetManager.ToDeleteInMainThread:= TAssetContainer.Create;
      TAssetManager.ToDeleteInMainThread.Add(Self); //refcount is 1 again
      TAssetManager.ToDeleteInMainThreadCS.Leave;
    end;
  end;

  function TGenericAsset.MustBeFreedFromMainThread: boolean;
  begin
    Result:= false
  end;

  function TGenericAsset.Actualize: boolean;
  begin
    DieBecauseStupidDeveloperForgotToOverrideMethod('Actualize');
  end;

  function THandleAsset.Actualize: boolean;
  begin
    GenResource;
    MarkAsActualized;
    Result:= true;
  end;

  procedure THandleAsset.OnGenResourceFailure(ErrorMessage: UnicodeString);
  begin
    Die(RuEn(
      '%0: Сбой генерации, %1',
      '%0: Failed to generate, %1'
      ), [string(Self.ClassName), ErrorMessage]);
  end;

  function TGenericAsset.IsNecessary: boolean;
  begin
    Result:= not (ap_optional in f_props);//f_optional;
  end;

  function TGenericAsset.IsValidForServer: boolean;
  begin
    Result:= false;
  end;

  procedure TGenericAsset.EraseContainer;
  begin
    //does NOT clear the hash as it is needed for further processing
    Exclude(f_props, ap_actualized);
  end;

  procedure THandleAsset.EraseContainer;
  begin
    Exclude(f_props, ap_actualized);
    _handle:= 0;
  end;


  procedure TGenericAsset.OnModuleDestroy;
  begin
    DieBecauseStupidDeveloperForgotToOverrideMethod('OnModuleDestroy');
  end;

  procedure THandleAsset.OnModuleDestroy;
  begin
    if (ap_actualized in f_props) and (_handle <> 0) and  not Mother^.Module.NeedToKeepAssets()
      then FreeResource;
  end;

  constructor TGenericAsset.Create(p: TAssetPropertiesSet = []);
  begin
    inherited Create;
    Self.CpsMask:= Self.CpsMask or ASSET_MASK;
    //UpdateHash; Each asset must take care of its own hash!
    f_props:= p;
    if not (ap_actualized in f_props) then MyLogClass.AddNonActualizedAsset(Self);
  end;

  procedure TGenericAsset.AfterLoading;
  var
    Counterpart: TGenericAsset;
  begin
    if gv_RetrievingFromMother then Exit; //loading stored assets
      //from module ci stream. Do nothing;

    if Assigned(TAssetManager.Retrieved)
      then Counterpart:= TAssetManager.Retrieved.GetByHash(@Hash)
      else Counterpart:= nil;
//if Assigned(Counterpart) then AddLog('** %0', [TellInvalidAsset(Counterpart)]);
//bgSay(PervertedFormat('* %0.AfterLoading :  Counterpart=%1', [string(Self.ClassName), pointer(counterpart)]));
//if Assigned(counterpart) then
//bgSay(PervertedFormat('                     Counterpart=%0', [string(Counterpart.ClassName)]));

    if Assigned(Counterpart)
      //Counterpart.Exists
      //and Counterpart.IsValid //Fuck! It could be *not* valid due to lacking
        //some non-saveable derivative objects (like TMotherStream and such.
        //Because of that, Counterpart.Actualized() returns false.
        //Imma fucking idiot :(
      and (ap_actualized in Counterpart.AssetProperties)

    then begin
      if Mother^.Debug.Verbose then AddLog('    %0.AfterLoading: devouring the retrieved counterpart %1 with hash %2', [string(ClassName), pointer(Counterpart), HashToHex(Hash)]);
      if Devour(Counterpart) //copy its fields *including* the ap_actualized flag!
      then begin
        f_props-= [ap_unclaimed];
        if Mother^.Debug.Verbose then AddLog('    ..devoured');
        Counterpart.EraseContainer;
      end
      else begin
        if Mother^.Debug.Verbose then AddLog('    ..failed');
        EraseContainer; //handles and such are
        //  always zeroed out when loading session. Previously, this was done via
        //  skippable fields. But since I switched to using Chepersy for storing
        //  entire assets in the mother stream, that is just plain not possible.

        //actualization is now *stored* in the session as it is a part of a set
        //  (previously it wasn't because it was a skipped field).
        //EraseContainer erases it too.
      end;
    end
    else begin
      if Mother^.Debug.Verbose then begin
        if Assigned(Counterpart)
          then AddLog('    %0.AfterLoading: the counterpart for hash %1 is not valid (%2)',[string(ClassName), HashToHex(Hash), TellInvalidAsset(Counterpart)])
          else AddLog('    %0.AfterLoading: no counterpart retrieved for hash %1.',[string(ClassName), HashToHex(Hash)]);

      end;
      EraseContainer;
    end;

{Old version for comparison:
      pr: PAssetRecord;

      pr:= Mother^.AssetKeeper.Claim(@Hash);
      AfterEfCheck; //throw an exception if there was an error
      if Assigned(pr) then begin
        //  The handle could have changed, for example the texture
        //    we created during the yesterday session from the same image
        //    (thus the same hash) had different index in OpenGL.
        //  So we reload it with the value received from the mother module.
        _handle:= pr^.Handle;
        FActualized:= true;
      end;
}
//    end;
//bgsay('     4');
//    end;


    //Actualized() can be checking a shitload of other things
    //  besides (ap_actualized in f_props), so beware.
    if not Actualized then MyLogClass.AddNonActualizedAsset(Self);
//bgsay('     5');
  end;


  function TGenericAsset.Devour(c: TGenericAsset): boolean;
  begin
    f_props:= c.AssetProperties;
    f_hash:= c.Hash;
    Result:= true;
  end;

  function THandleAsset.Devour(c: TGenericAsset): boolean;
  begin
    Self._handle:= (c as THandleAsset).Handle;
    Result:= inherited Devour(c);
  end;



  procedure TGenericAsset.BeforeSaving;
  begin
    if Self.Actualized and Mother^.Module.NeedToKeepAssets()
    and not gv_StoringToMother
      then TAssetManager.ToStore.Add(Self);

{    if Actualized and (_handle <> 0)
      and Mother^.Module.NeedToKeepAssets()
    then begin
      Mother^.AssetKeeper.Store(@_handle, @hash, self.ClassIndex);
      AfterEfCheck;
    end;
}
  end;

{
constructor TGenericAsset.CreateToDestroy(h: ptruint);
  begin
    try
      _handle:= h;
      FreeResource;
      _handle:= 0;
    except
      Die(RuEn('Сбой при удалении ничейного ресурса %0','Failed to free orphaned asset %0'), [string(Self.ClassName)]);
    end;
  end;
}

  function TGenericAsset.IsValid: boolean;
  begin
    Result:= Actualized;// and not Scraped;
  end;

  function THandleAsset.IsValid: boolean;
  begin
    Result:= Actualized
      //and not Scraped
      and (_handle <> 0);
  end;


  procedure TGenericAsset.UpdateHash;
  begin
    Mother^.AssetKeeper.GenHash(@f_hash);
    AfterEfCheck;
    f_hash.ClassNameMd5:= Self.ClassNameMd5();
  end;


{  function TGenericAsset.Valid: boolean;
  begin
    //WTF Self is always 0?
    Dbgsay('TGenericAsset.Valid: Self=' + IntToHex(ptruint(pointer(Self)), 8) );
    if not Assigned(Self)
      then Exit(false)
      else Result:= Self.IsValid;
  end;
}


  destructor THandleAsset.Destroy;
  begin
    if _handle <> 0 then FreeResource;
    inherited;
  end;


  function TGenericAsset.SortOrder: ffloat;
  begin
    if IsNecessary
      then Result:= -1
      else Result:= 0;
  end;

{
  function TGenericAsset.MakeSureThreadedTaskIsAdded(C: CAbstractThreadedTask): TAbstractThreadedTask;
    //To be used in Actualize, assuming only one task of each kind is needed.
    //if such task is found, returns it. Otherwise creates it.

    //Usage:   if not MakeSureThreadedTaskIsAdded(TYourThreadedTask).Completed then Exit(false);
    var i: integer;
  begin
    for i:= 0 to High(ThreadedTasks) do
      if Assigned(ThreadedTasks[i]) and (ThreadedTasks[i] is C)
        then Exit(ThreadedTasks[i]);
    if Mother^.Debug.Verbose then addlog('Asset %0 adds threaded task %1', [Self, string(C.ClassName)]);
    Result:= C.Create(Self);
    Self.ThreadedTaskCreateCallback(Result);
    Self.AddThreadedTask(Result);
    Result.Reschedule; //moved here from the constructor. Calls Module.AddThreadedTask().
  end;


  procedure TGenericAsset.ThreadedTaskCreateCallback(T: TAbstractThreadedTask);
  begin
    {Here the asset passes whatever parameters are necessary for the task
      (file name for load, dimensions for resize and so on)
      This is done by accessing specially crafted properties or methods
      of the task class (the "as" operator is a must).

     I know it's a bit convoluted, but this couldn't be put into the
       constructor calling parameters due to the constructor call being deferred
    }

    //by default do nothing
  end;


  procedure TGenericAsset.AddThreadedTask(T: TAbstractThreadedTask);
  begin
    SetLength(ThreadedTasks, Length(ThreadedTasks) + 1);
    ThreadedTasks[High(ThreadedTasks)]:= T;
  end;


  procedure TGenericAsset.FinishThreadedTask(T:  TAbstractThreadedTask);
  var i, j: integer;
  begin
    for i:= 0 to High(ThreadedTasks) do
      if ThreadedTasks[i] = T then begin
        T.Finish; //it's where it can safely write back to the asset object
        T.Free;
        for j:= i to High(ThreadedTasks) - 1 do ThreadedTasks[j]:= ThreadedTasks[j + 1];
        SetLength(ThreadedTasks, Length(ThreadedTasks) - 1);
        Exit;
      end;
  end;
 }

  procedure TGenericAsset.MarkAsUnclaimed;
  begin
    Include(f_props, ap_unclaimed);
  end;

  procedure TGenericAsset.MarkAsTemporary;
  begin
    Include(f_props, ap_temporary);
  end;

  procedure TGenericAsset._setHash(h: TAssetHash);
  begin
    if Actualized then Die(MI_ERROR_PROGRAMMER_NO_BAKA, [PervertedFormat(RuEn(
      'Попытка изменить хэш у актуализированного ресурса %0.'#10#13'  текущий хеш  %1, задаваемый %2.',
      'Attempt to set hash for an acualized asset %0.'#10#13'  current hash is %1, provided hash is %2.'
    ),[string(ClassName), HashToHex(f_hash), HashToHex(h)])]);

    f_hash:= h;
  end;

  constructor TAssetInnerMatroska.Create;
  begin
    inherited Create;
    Self.CpsMask:= Self.CpsMask or ASSET_MASK or VOLATILE_MASK;
  end;
{
procedure TGenericAsset.Scrape;
begin
  AddLog('%0.Scrape'#10#13'  %1', [string(Self.ClassName), ExplainCallStack(0,99)]);
  inherited;
end;
}

{
  function TGenericAsset.RequestMasterDimensions(var w, h: integer): boolean;
  begin
    Result:= true;
    w:= 256;
    h:= 256;
  end;

}

// **************************************************************

{   procedure TGenericDynamicAsset.AfterLoading;
   begin
     if f_runhash <> TAssetHash(Mother^.State.RunSessionHash)
       then begin //do not even try to claim it back
         _handle:= 0; //This is extraneouws. It *should* be zero anyway
         Self.Scrape(); //Commit suicide. The asset owner must check if it is scraped and replace it.
       end
       else inherited;
   end;
}
{   procedure TGenericDynamicAsset.BeforeSaving;
   begin
     Mother^.AssetKeeper.GenHash(@hash);
     AfterEfCheck;
     inherited;
   end;
}
{
   procedure TGenericDynamicAsset.RegisterFields;
   begin
     inherited;
     ListFields(['f_runhash', @f_runhash, typeinfo(TAssetHash)]);
   end;
 }
{
{$ifopt C+} //if assertions are turned on...

  function GetValidateDetailsFormatErrorMessage(details: array of const): UnicodeString;
  var j, i: integer;
  begin
    Result:= RuEn(
      'массив параметров'#10#13'    [',
      'the details array'#10#13'    [');
    for i:= 0 to High(details) do begin
      Result+= VarRecTypeName(details[i].Vtype);
      if i < High(details) then Result+= ', ';
    end;
    if Length(ValidateDetailsEtalon) > 1
      then Result+= RuEn(
        ']'#10#13'  не соответствует ни одному из эталонов:',
        ']'#10#13'  does not match any of the following etalons:')
      else  Result+= RuEn(
        ']'#10#13'  не соответствует эталону:',
        ']'#10#13'  does not match the etalon:');
    VarRecTypeName
    for j:= 0 to High(ValidateDetailsEtalon) do begin
      Result+= #10#13'    [';
      for i:= 0 to High(ValidateDetailsEtalon[j]) do begin
        Result+= VarRecTypeName(ValidateDetailsEtalon[j][i].Vtype);
        if i < High(ValidateDetailsEtalon[j]) then Result+= ', ';
      end;
      Result+= ']';
    end;
  end;

  //can be called several times with different etalons, joined by OR
  function TGenericAsset.ValidateDetailsFormat (details, etalon: array of const): boolean;
  var
    i: integer;
  begin
    Result:= true;
    if Length(etalon) <> Length(details) then Result:= false;
    for i:= 0 to High(etalon) do
      if etalon[i].Vtype <> details[i].Vtype then begin
        Result:= false;
        break;
      end;
    if Result then begin
      SetLength(ValidateDetailsEtalon, 0);
    end
    else begin
      SetLength(ValidateDetailsEtalon, Length(ValidateDetailsEtalon) + 1);
      ValidateDetailsEtalon[High(ValidateDetailsEtalon)]:= etalon;
    end;
  end;
{$endif}
}
