{
    This file is part of Chentrah,
    Copyright (C) 2004-2016 Anton Rzheshevski (chebmaster@mail.ru).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see http://www.gnu.org/licenses/

 **********************************************************************}

{$include mo_globaldefs.h}
{$typeinfo on}

unit mo_atlas;

interface
uses SysUtils, Classes {$ifdef windows}, Windows {$else} , baseunix, dl {$endif},
  mo_hub, chepersy, mo_classes, typinfo, syncobjs, mo_threads, mo_assets, mo_texture, md5;


type
  TAbstractAtlasChunk = class; //forward declaration
  TStaticAtlasChunk = class;
  TAtlasChunks = array of TAbstractAtlasChunk;


  TAbstractAtlas = class (TGenericAsset)
    function CompletedLoading: boolean; virtual; abstract;
    procedure Pulse; virtual; abstract;
  end;

  TStaticAtlasPropsEnum = (sapr_initialized, sapr_texturesLoaded, sapr_cacheExists, sapr_cacheLoaded, sapr_cacheSaved);
  TStaticAtlasPropsSet = set of TStaticAtlasPropsEnum;
  TStaticAtlasCacheSavingTask = class;

  TStaticAtlas = class (TAbstractAtlas)
  protected
    f_addrEnum: PTypeInfo;
    f_init_stage: integer;
    f_enumTypeName: RawByteString;
    f_path: UnicodeString;
    f_ownTexture: TTexture;
    f_chunks: TAtlasChunks;
    f_enumMd5: TMd5Digest;
    f_atlas_props: TStaticAtlasPropsSet;
    f_cacheSavingTask: TStaticAtlasCacheSavingTask;
    procedure InitStatic; virtual;//is used in constructor and in AfterLoading,
    // to connect to current code constants

  public
    property Texture: TTexture read f_ownTexture;

    {The algorithm for the file names is as following: it determines
       an enum value's name, then replaces its beginning up to
       and including the first underscore (_) character with the path
       and adds .$IMG extension, which the file manager transparently
       resolves into the first found .TGA .PNG .BMP .JPG and so on it can find.
     Both file names and enum names are forcibly converted to lower case. }
    constructor Create(addrEnum: PTypeInfo; Path: UnicodeString);

    procedure AfterLoading; override; //loading the cached & low-res versions
    procedure Pulse; override; //loading the hi-res version & updating cache

    //For generating close-proximity LOD meshes.
    //May return an actual texture, OR may return its own texture's chunk if
    //  the atlas had been loaded from cache while the texture is not ready yet.
    function GetTexture(const enumOrd: integer; out tex: TTexture;
            out ds, dt, zs, zt, clampAnisotropy: float): boolean; virtual;

    function GetChunkDetails(const enumOrd: integer; out tex: TTexture;
            out ds, dt, zs, zt, clampAnisotropy: float): boolean; virtual;

    //Copies all relevant properties of the other instance
    // and leaves it zeroed out, to be safely disposed of
    //Used a) on counterparts retrieved from mother
    //  b) on clones lazily updated in the background
    //  see TManagedFbo.Devour as an example: it's complicated!
    //A complex asset may refuse to devour if something doesn't match
    //  -- that's why checking Actualized afterwards is necessary.
    function Devour(c: TGenericAsset): boolean; override;

    function Actualize: boolean; override;


    procedure EraseContainer; override;
    procedure RegisterFields; override;
  end;

  TDynamicAtlas = class (TAbstractAtlas)


  end;


  TAbstractAtlasChunk = class (TAssetInnerMatroska) //does not get saved into session
  protected
    f_left, f_top, f_width, f_height, f_max_anisotropy: float;
    f_lod: longint;
    function _GetTexture: TTexture; virtual;
  public
    constructor Create(Left, Top, Width, Height: float; MaxAniso: float = 2.0; Lod: integer = 0);
    procedure RegisterFields; override;
    property Left: float read f_left;
    property Top: float read f_top;
    property Width: float read f_width;
    property Height: float read f_height;
    property MaxAnisotropy: float read f_max_anisotropy;
    property Lod: longint read f_lod;
    property Texture: TTexture read _GetTexture;
  end;



  TStaticAtlasChunk = class(TAbstractAtlasChunk)
  protected
    f_texture: TTexture;
    function _GetTexture: TTexture; override;
  public
    constructor Create(_Left, _Top, _Width, _Height: float; _MaxAniso: float = 2.0; _Lod: integer = 0; _Texture: TTexture = nil);
    destructor Destroy; override;
    procedure RegisterFields; override;
  end;

  TStaticAtlasInitTask = class(TAbstractThreadedTask)
  protected
  public

  end;

  TStaticAtlasBuildnUpdateCacheTask = class(TAbstractThreadedTask)
    //This one breaks the "all file operations in one thread" rule
    //  for simplicity's sake.
    //  Intended for making screenshots, saving atlas caches & stuff
  protected
    f_imghandle: pointer;
    f_file_name: TFileNameString;
    function Perform: boolean; override;
  public
    constructor Create(Imagehandle: pointer; FileName: TFileNameString; DeleteImage: boolean = false);
    function GetType: TThreadedTaskType; override;
  end;

implementation



  //Copies all relevant properties of the other instance
  // and leaves it zeroed out, to be safely disposed of
  //Used a) on counterparts retrieved from mother
  //  b) on clones lazily updated in the background
  //  see TManagedFbo.Devour as an example: it's complicated!
  //A complex asset may refuse to devour if something doesn't match
  //  -- that's why checking Actualized afterwards is necessary.
  function TStaticAtlas.Devour(c: TGenericAsset): boolean;
  begin


  end;


  procedure TStaticAtlas.AfterLoading;
  begin
    InitStatic;
    inherited;
  end;


  constructor TStaticAtlas.Create(addrEnum: PTypeInfo; Path: UnicodeString);
  begin
    inherited Create;
    f_enumTypeName:= chepersy.GetNameByTypeInfo(addrEnum);
    if f_enumTypeName = '' then Die(RuEn(
        'Невозможно создать %0: Чеперси неизвестен тип "%1"',
        'Unable to create %0: Chepersy does not know type "%1"'
      ), [RawByteString(Self.ClassName), f_enumTypeName]);
    f_path:= Path;
    InitStatic;
  end;


  procedure TStaticAtlas.InitStatic;
  begin
    f_addrEnum:= chepersy.GetTypeInfoByName(f_enumTypeName);
    if not Assigned(f_addrEnum) then Die(RuEn(
        'Чеперси неизвестен тип "%0"',
        'Chepersy does not know type "%0"'
      ), [f_enumTypeName]);

    f_atlas_props:= [];
  end;

  const
    i_init = 0;
    i_loading_textures = 1;
    i_loaded_textures = 2;
    i_cleaning_up_textures = 3;
    i_done = 4;

  function TStaticAtlas.Actualize: boolean;
  begin
    case f_init_stage of
      i_init: begin

      end;
      i_loading_textures: begin

      end;
      i_loaded_textures: begin

      end;
      i_cleaning_up_textures: begin

      end;
      i_done: begin




        MarkAsActualized;
        Result:= true;
      end;
    else
      Die('Unknown init stage %0', [f_init_stage]);
    end;
  end.

  procedure TAbstractAtlasChunk.RegisterFields;
  begin
    inherited;
    ListFields([
      @f_left, 'left', @f_top, 'top', @f_width, 'width', @f_height, 'height', @f_max_anisotropy, 'max_aniso', typeinfo(float),
      @f_lod, 'lod', typeinfo(longint)
    ]);
  end;

  procedure TStaticAtlasChunk.RegisterFields;
  begin
    inherited;
    ListFields([
      @f_texture, 'texture', typeinfo(TTexture)
    ]);
  end;

  function TAbstractAtlasChunk._GetTexture: TTexture;
  begin
    Result:= nil;
  end;

  function TStaticAtlasChunk._GetTexture: TTexture;
  begin
    Result:= f_texture;
  end;

  constructor TAbstractAtlasChunk.Create(Left, Top, Width, Height: float; MaxAniso: float = 2.0; Lod: integer = 0);
  begin
    inherited Create;
    f_left:= Left;
    f_top:= Top;
    f_width:= Width;
    f_height:= Height;
    f_max_anisotropy:= MaxAniso;
    f_lod:= Lod;
  end;

  constructor TStaticAtlasChunk.Create(_Left, _Top, _Width, _Height: float; _MaxAniso: float = 2.0; _Lod: integer = 0; _Texture: TTexture = nil);
  begin
    inherited Create(_Left, _Top, _Width, _Height, _MaxAniso, _Lod);
    f_texture:= _Texture;
  end;

  destructor TStaticAtlasChunk.Destroy;
  begin
    f_texture:= nil; //decrease its refcount
    inherited;
  end;


end.
