{
    This file is part of Chentrah,
    Copyright (C) 2004-2015 Anton Rzheshevski (chebmaster@mail.ru).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see http://www.gnu.org/licenses/

 **********************************************************************}


unit mo_camera;

{$include mo_globaldefs.h}

interface

uses
  SysUtils, Classes, math,
  chepersy, mo_hub, typinfo,
  mo_classes, mo_assets, mo_gmathbase,
  mo_3dbasics, mo_glsl, mo_fbo;

type
  TMultisampleShader = class;

  { TCamera }

  TCamera = class(TAbstractCamera)
  protected
    Multisam, Multisam2: TMultisampleShader;
    Fbo: TManagedFbo;
    renW, renH, CurrentMultiSample: longint;
    frW, frH: float;
    function _GetHdr(): float; virtual;
    procedure CheckRenderBuffersSizes; virtual;
    procedure CheckMultisampler; virtual;
    procedure ClearFbo; virtual; //used by render
    procedure RenderBufferToScreen; override;
    procedure RenderSceneToBuffer; override;
    procedure PrepareToRender; override;
  public
    procedure RegisterFields; override;
    procedure ViewportResize(ViewportX, ViewportY, ViewportWidth, ViewportHeight: float); override;
    procedure AfterLoading; override;
  end;

//  TFishEyeCamera = class(TCamera)
//  TStereoCamera = class(TCamera)

  TMultisampleShader = class (TGlslProgram)
  private
    f_samples: integer;
    f_hdr, frw, frh: float;
  public
    property GetFrW: float read frw;
    property GetFrH: float read frh;
    property HDR: float read f_hdr;
    property Samples: integer read f_samples;
    constructor Create (_samples: integer; _hdr, _frw, _frh: float; aprops: TAssetPropertiesSet = []);
    procedure RegisterFields; override;
    procedure GenerateShaderSource(var v, f: AnsiString); override;
  end;

implementation

  uses mo_module;

  { TCamera }

  function TCamera._GetHdr(): float;
  var chdr: float;
  begin
    if not Assigned(Multisam) then Exit(1.0);
    //as multisampler is lazily updated, its HDR factor may not always match our current HDR factor.
    Result:= Math.max (Multisam.Hdr, 1.0);
  end;

  procedure TCamera.AfterLoading;
  begin
    //if Assigned(Fbo) and Fbo.Scraped then pointer(Fbo):= nil;
  end;

  procedure TCamera.RegisterFields;
  begin
    RegClass(TMultisampleShader);
    RegClass(TmanagedFbo);
    inherited;
    ListFields([
      'msm', @Multisam, 'msm2', @Multisam2, 'fbo', @Fbo, CPS_MANAGED_OBJECT,
      'renW', @renW, 'renH', @renH, 'cms', @CurrentMultiSample, typeinfo(longint),
      'frw', @frW, 'frh', @frH, typeinfo(float)
    ]);

  end;

  procedure TCamera.PrepareToRender;
  begin
    CheckRenderBuffersSizes;
    CheckMultisampler;
  end;

  procedure TCamera.RenderSceneToBuffer;
  begin

    if not Assigned(Fbo) then exit; // this should not be happening!
    if not Fbo.Actualized then Exit; //this should NOT happen!

    Scene.Prepare(Self);
    ClearFbo;
    Fbo.Bind;
    glViewport(0, 0, min(renW, Fbo.Width), min(renH, Fbo.Height));
    Scene.Render(Self);
    CheckGlError;
    glBindFramebuffer(GL_FRAMEBUFFER_EXT, 0);
  end;

  procedure TCamera.RenderBufferToScreen;
  var rw, rh: float;
  begin

    SetGlStatesFor2dRender;

    glEnable(GL_ALPHA_TEST);
    glAlphaFunc(GL_GREATER, 0.01);
    glEnable(GL_BLEND);
    glDisable(GL_DEPTH_TEST);
    glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA);
    glEnable(GL_TEXTURE_2D);
    CheckGlError;

    if not FBO.Actualized then exit; //or not Multisam.Actualized then exit; //this should NOT happen!

    SetGlStatesFor2dRender;
    glEnable(GL_TEXTURE_2D);

    //glBindTexture(GL_TEXTURE_2D, Mother^.GAPI.StubTexture);
    FBO.texture.Bind;
    CheckGlError;

    //glViewport(round(f_prevX), round(f_prevY), round(f_prevWidth), round(f_prevHeight));

    if Assigned(Multisam) and Multisam.Actualized
      then Multisam.Use
      else SwitchToFixedPipeline;



    glColor4f(Mother^.Display.FadeIn, Mother^.Display.FadeIn, Mother^.Display.FadeIn, Mother^.Display.FadeIn);


    rW:= renW / max(1.0, Fbo.Width);
    rH:= renH / max(1.0, Fbo.Height);

//glBindTexture(GL_TEXTURE_2D, Mother^.GAPI.StubTexture);


    glBegin(GL_QUADS);
      glTexCoord2f(0, 0);
      glVertex2f(f_prevX, f_prevY); //bottom left corner
      glTexCoord2f(0, rH);
      glVertex2f(f_prevX, f_prevY + f_prevHeight);
      glTexCoord2f(rW, rH);
      glVertex2f(f_prevX + f_prevWidth, f_prevY + f_prevHeight);
      glTexCoord2f(rW, 0);
      glVertex2f(f_prevX + f_prevWidth, f_prevY);
    glEnd;

    glBindTexture(GL_TEXTURE_2D, Mother^.GAPI.StubTexture);

    SwitchToFixedPipeline;
    SetGlStatesFor2dRender;
  end;

  procedure TCamera.ViewportResize(ViewportX, ViewportY, ViewportWidth, ViewportHeight: float);
  var
    resized: boolean;
  begin
    resized:= (ViewportHeight <> f_prevHeight) or (ViewportWidth <> f_prevWidth);
    inherited ViewportResize(ViewportX, ViewportY, ViewportWidth, ViewportHeight);
    if resized then CheckRenderBuffersSizes;
  end;

  procedure TCamera.CheckRenderBuffersSizes;
  var
    needW, needH, roundmask: integer;
    qff, desiredW, desiredH: float;
  begin
{
if Assigned(Fbo2) and Fbo2.Actualized then begin
      Fbo.Free;
      Fbo:= Fbo2;
      pointer(Fbo2):= nil;
    end;
}
    qff:= sqrt(Mother^.Timer.QF / 100); //quality factor aka global detalization multiplier
    //QF is capped at 999, which gives max qff = 3.16
    // which is also the reasonable upper limit of such a crude method of FSAA.
    //P.S. at 2999, max qff = 5.48
    if (Mother^.Timer.QF >= 100) and (Mother^.Timer.QF < 120) then begin
      //limit to precise match because small increments over 100% only blur, making the image worse.
      renW:= trunc(f_prevWidth);
      renH:= trunc(f_prevHeight);
      desiredW:= f_prevWidth;
      desiredH:= f_prevHeight;
    end else begin
      desiredW:= f_prevWidth * qff;
      desiredH:= f_prevHeight * qff;
      renW:= trunc(desiredW);
      renH:= trunc(desiredH);
    end;

    if not Mother^.GAPI.NpotSupported then begin
      desiredW:= math.min(f_prevWidth, desiredW);
      desiredH:= math.min(f_prevHeight, desiredH);
    end;
    roundmask:= math.Max(256, GetHigherPowerOf2(round(desiredW / 4))) - 1;
    needW:= Math.min(Mother^.GAPI.FboMax, Math.max(256, ValidateTextureDimension(((round(desiredW) - 1) or roundmask) + 1))); //round up to nearest 256
    roundmask:= math.Max(256, GetHigherPowerOf2(round(desiredH / 4))) - 1;
    needH:= Math.min(Mother^.GAPI.FboMax, Math.max(256, ValidateTextureDimension(((round(desiredH) - 1) or roundmask) + 1))); //round up to nearest 256

    if not Assigned(Fbo) then Fbo:= TManagedFbo.Create(GL_RGBA8); //always defaults to 256x256 at the first creation

    Fbo.RequestDesiredDimensions(needW, needH); //we use smart FBO here

    Fbo.Pulse;

    renW:= min(renW, Fbo.Width);
    renH:= min(renH, Fbo.Height);

    //This is a crude stop-gap. After loading a session,
    //  Fbo size may return 0 so frW and frH become NaN,
    //  which then leaks into the shader source code making
    //  OpenGL fail compiling it.
    frW:= renW / Math.Max(256.0, Fbo.Width); //renW / Fbo.Width;
    frH:= renH / Math.Max(256.0, Fbo.Height); //renH / Fbo.Height;
//addlog('renw=%0, renh=%1, frw=%2, frh=%3',[renw, renh, frw, frh]);
  end;

  procedure TCamera.CheckMultisampler;
  var
    msm, cms: integer;
    msp: TMultisampleShader = nil;
    ap: TAssetPropertiesSet;
    mfrW, mfrH: float;
  begin
    cms:= CurrentMultiSample;
    if (Mother^.Timer.QF < 120) //or not Mother^.GAPI.NpotSupported
      then msm:= 1
      else msm:= 3 + trunc (Mother^.Timer.QF / 50); //5 minimum
      // Fuck, these are nor related! I can have 20 samples with 8 texture coords max.
      //min(Mother^.GAPI.MaxTextureCoords, 1 + trunc (Mother^.Timer.QF / 100));
                               //round(sqrt(Mother^.Timer.QF * sqr(2) / 120)));

    if Assigned(Multisam2) and Multisam2.Actualized then begin
      if (msm > 1) or (f_hdr > 1.0) then begin
        Multisam.Free;
        Multisam:= Multisam2;
      end;
      pointer(Multisam2):= nil;
    end;

    mfrW:= FrW / math.max(1.0, f_prevWidth);
    mfrH:= FrH / math.max(1.0, f_prevHeight);

    if ((msm <> CurrentMultiSample)
    or ( //it's existing, samples match, but something (fbo size or hdr factor) has sufficiently changed.
      (msm > 1) and (

      {lazy shader recreating (cuz passing uniform parameters is, you know,
         _hard_, so I hardcode them and rebuild the shader as necessary) }
      //(f_prevWidth > prevW * 1.1) or (f_prevWidth * 1.1 < prevW)
      //or (f_prevHeight > prevH * 1.1) or (f_prevHeight * 1.1 < prevH)
       (not Assigned(Multisam))// or Multisam.Scraped)
      or (Multisam.GetFrW * 1.1 < mfrW) or (Multisam.GetFrW > mfrW * 1.1)
      or (Multisam.GetFrH * 1.1 < mfrH) or (Multisam.GetFrH > mfrH * 1.1)
      or (Assigned(Multisam) and ((Multisam.HDR > f_hdr * 1.1) or (f_hdr > Multisam.HDR * 1.1)))
    ))
    and not Assigned(Multisam2)
    )
    then begin
      if (msm > 1) or (f_hdr > 1.0)
        then begin
          ap:= [];
          if not (f_hdr > 1.0) or Assigned(Multisam) then ap+= [ap_optional];
          msp:= TMultisampleShader.Create(
            msm, f_hdr, mfrW, mfrH
            , ap);
          if not Assigned(Multisam)
            then Multisam:= msp
            else Multisam2:= msp;
        end
        else begin
          if Assigned(Multisam) then begin
            //if not Multisam.Scraped then Multisam.Free;
            Multisam:= nil;
          end;
          if Assigned(Multisam2) then begin
            //Multisam2.Free;
            Multisam2:= nil;
          end;
        end;
      CurrentMultiSample:= msm;
    end;

  end;

  var
    SQNumCycles: double = 0.0;
    SQMicrosecondsTotal: double = 0.0;

  procedure TCamera.ClearFbo;
  var
    q: qword;
    d: double;
  begin
    if not Fbo.Actualized then Fbo.Actualize;
    Fbo.Bind;
    if OpenGLQuirkScissorFailed in Mother^.GAPI.Quirks then begin
      //the quirk is confirmed. Clear the FBO by rendering a quad
      glUseProgram(0);
      glViewport(0, 0, Math.min(renW + (renW div 16), Fbo.Width), Math.min(renH + (renH div 16), Fbo.Height));
      glMatrixMode(GL_PROJECTION);
      glLoadIdentity();
      glOrtho(0, 1, 0, 1, 1 , 0);
      glMatrixMode(GL_MODELVIEW);
      glLoadIdentity();
      glDisable(GL_BLEND);
      glDisable(GL_ALPHA_TEST);
      glDepthMask( GL_TRUE );
      glEnable(GL_DEPTH_TEST); glDepthFunc(GL_ALWAYS);
      glDisable (GL_TEXTURE_2D);
      glcolor4f(0, 0, 0, 0);
      glBegin(GL_QUADS);
        glVertex3f(0, 0, 0);
        glVertex3f(0, 1.05, 0);
        glVertex3f(1.05, 1.05, 0);
        glVertex3f(1.05, 0, 0);
      glEnd;
    end else begin
      Mother^.Timer.UsecDelta(@q);
      glScissor(0, 0, Math.min(renW + (renW div 16), Fbo.Width), Math.min(renH + (renH div 16), Fbo.Height));
      glEnable(GL_SCISSOR_TEST);
      glClear(GL_COLOR_BUFFER_BIT or GL_DEPTH_BUFFER_BIT);
      glDisable(GL_SCISSOR_TEST);
      SQMicrosecondsTotal+= Math.min(20000.0, Mother^.Timer.UsecDelta(@q));
        //clipping, bitches. Because shit happens, like system slowdown
        //or hibernation at the wrong moment
      SQNumCycles+= 1.0;
      if (SQNumCycles > 100.0) then begin//enough recurrences to confirm it
        if (SQMicrosecondsTotal / SQNumCycles > 10000.0) // there's no middle
          //ground. It is either negligible when everything is all right
          //  or horrific when the driver falls to software
        then begin
          Mother^.GAPI.Quirks += [OpenGLQuirkScissorFailed]; //gets saved in the config
          Mother^.GAPI.QuirksTested+= [OpenGLQuirkScissorFailed];
          AddLog(RuEn(
            'Драйвер OpenGL свалился в программную прорисовку при использовании glScissor + glClear.'
              + #10#13'  Вывих учтён, использую очистку FBO прорисовкой прямоугольника.',
            'OpenGL driver fell back to software while using glScissor + glClear.'
              + #10#13'  Quirk registered, using full-screen quad to clear FBO.'));
        end;
        SQMicrosecondsTotal:= 0.0;
        SQNumCycles:= 0.0;
      end;
      //if SQNumCycles > 0 then DbgSay(PervertedFormat('mspc %0',[round(SQMicrosecondsTotal / SQNumCycles)]));
    end;
    SwitchToFixedPipeline;
  end;

  constructor TMultisampleShader.Create (_samples: integer; _hdr, _frw, _frh: float; aprops: TAssetPropertiesSet = []);
  begin
    inherited Create(aprops);
    f_samples:= _samples;
    f_hdr:= _hdr;
    frw:= _frw;
    frh:= _frh;
  end;

  procedure TMultisampleShader.RegisterFields;
  begin
    inherited;
    ListFields([
      'samples', @f_samples, typeinfo(integer),
      'hdr', @f_hdr,
      'frw', @frw,
      'frh', @frh,
      typeinfo(float)
    ]);
  end;

  procedure TMultisampleShader.GenerateShaderSource(var v, f: AnsiString);
  var
    sampler_code: AnsiString = '';
    hdr_mul: AnsiString;
    sampler_rx, sampler_ry, a, adelta, dx, dy, sr, srdelta: float;
    i, steps, step, ssum, perstep: integer;
    function csign(f: float): AnsiString;
    begin
      if f < 0 then Result:= '' else Result:= '+';
    end;
  begin
    //use default verter shader emulating ffp
    inherited GenerateShaderSource(v, f);

    hdr_mul:= '';
    if f_hdr > 1.0 then hdr_mul:=
      'vec4(' + FloatToStrF(f_hdr, ffExponent, 15, 1)
        + ','+ FloatToStrF(f_hdr, ffExponent, 15, 1)
        + ',' + FloatToStrF(f_hdr, ffExponent, 15, 1)
        + ',1.0) * ';
    steps:= 1 + ((f_samples - 1) div 8);
    ssum:= 1;
    for step:=1 to steps do begin
      perstep:= math.min(8, f_samples - ssum);
      adelta:= 3.1416 * 2 / perstep;
      a:= 3.1416 * 1.0; //0.75; //45 degrees
      sampler_rx:= 0.7 * ((1 + steps - step)/steps) * frw;
      sampler_ry:= 0.7 * ((1 + steps - step)/steps) * frh;
      for i:= 1 to perstep do begin
        dx:= sin(a) * sampler_rx;
        dy:= cos(a) * sampler_ry;

//addlog('step=%0, i=%1, a=%2, dx=%3, dy=%4, frw=%5, frh=%6',[step, i, a, dx, dy, frw, frh]);
        sampler_code+= 'sum += texture2D(u_texture, vec2(gl_TexCoord[0].x '
           + csign(dx) + ' ' + FloatToStrF(dx, ffExponent, 15, 1) + ', gl_TexCoord[0].y '
           + csign(dy) + ' ' + FloatToStrF(dy, ffExponent, 15, 1) + ')) * '
           + FloatToStrF(1 / f_samples, ffExponent, 15, 1) + '; ';
        a+= adelta;
      end;
      ssum+= perstep;
    end;

    //compose fragment shader
    f:=
       'uniform sampler2D u_texture;'
     + 'void main(void) {'
         + 'vec4 sum = vec4(0.0);'
         + 'sum += texture2D(u_texture, vec2(gl_TexCoord[0].x, gl_TexCoord[0].y)) * '
         + FloatToStrF(1 / f_samples, ffExponent, 15, 1) + ';'
         + sampler_code
       //  + 'gl_FragColor = texture2D(u_texture, vec2(gl_TexCoord[0]));'
         + 'gl_FragColor = ' + hdr_mul + ' sum * gl_Color;'
     + '}';
  end;




end.
