{
    This file is part of Chentrah,
    Copyright (C) 2004-2014 Anton Rzheshevski (chebmaster@mail.ru).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see http://www.gnu.org/licenses/

 **********************************************************************

    This file contains the basic library of classes based on chepersy.

 **********************************************************************}

{$include mo_globaldefs.h}

//NOTE: ffloat differs.
//  For multiplayer, it's 16.16 fixed point,
//  for single player it's "single" aka "float"

unit mo_classes;
interface
  uses SysUtils, chepersy, mo_hub, typinfo, mo_gmathbase;

const
  VOLATILE_MASK_BIT = 20;
  VOLATILE_MASK = dword(1) shl VOLATILE_MASK_BIT;

type
  TCgeObject = class(chepersy.TChepersyObject)
  private
    function _GetVolatile(): boolean;
    procedure _SetVolatile(v: boolean);
  public
    function SortOrder: ffloat; virtual; //by default returns zero. Used by TSortedList.
    property Volatile: boolean read _GetVolatile write _SetVolatile;
      //volatile instances don't get saved into the session.
      //Will save to the mother ci stream, though
    procedure OnModuleDestroy; virtual; // by default does nothing
  end;
  CCgeObject = class of TCgeObject;

  TPreservableLink =  class(TCgeObject)
  private
    SessionGUID: TAssetHash;
    o: ptruint;
    procedure UpdateHash;
    function _GetObj(): TCgeObject;
    procedure _SetObj(obj: TCgeObject);
  public
    property Obj: TCgeObject read _GetObj write _SetObj;
  end;


  TArrayOfCgeObjects = array of TCgeObject;

  TSortedList = class(TCgeObject)
  private
    A: TArrayOfCgeObjects;
    function _ReadData(i: integer): TCgeObject;
    function _GetLast: TCgeObject;
    function _GetCount: integer;
    function _GetHigh: integer;
  public
    procedure Sort();
    procedure AfterLoading(); override;
    procedure RegisterFields; override;
    procedure Prune; //removes NIL and scraped elements
    function ExtractFirst: TCgeObject;
    function Pop: TCgeObject;
    function Add(item:  TCgeObject): integer;
    function Remove(i: integer): TCgeObject;
    property Data[index: integer]: TCgeObject  read _ReadData; default;
    property Count: integer read _GetCount;
    property High: integer read _GetHigh;
    property Last: TCgeObject read _GetLast;
  end;

{
  //TAsset takes its role, with its new functionality of treaded tasks

  TBackgroundTaskStatus = (bts_scheduled, bts_completed, bts_started, bts_volatile);
  TBackgroundTaskStatusSet = set of TBackgroundTaskStatus;
  TBackgroundTask = class(TCgeObject)
  private
    fStatus: TBackgroundTaskStatusSet;
    function ShouldTerminate: boolean;
    procedure RemoveFromList;
  public
    constructor Create;
    procedure AfterLoading; override;
    procedure RegisterFields; override;
    procedure Scrape; override; //additionally removes itself from the list
    property Status: TBackgroundTaskStatusSet read fStatus;
    function Execute: boolean; virtual; abstract;
  end;
}
var
  TestInstanceOfCgeObject: TCgeObject;

procedure InitCgeClasses;

implementation
  uses mo_module, mo_threads;

procedure InitCgeClasses;
begin
  TestInstanceOfCgeObject:= TCgeObject.Create;
end;

function TCgeObject.SortOrder: ffloat;
begin
  Result:= FF_ZERO;
end;

procedure TCgeObject.OnModuleDestroy;
begin
  //do nothing
end;

function TSortedList.Add(item: TCgeObject): integer;
var
  i, iHole: integer;
  so: ffloat;
begin
  //  http://en.wikipedia.org/wiki/Insertion_sort
  i:= length(A);
  SetLength(A, i + 1);
  so:= item.SortOrder;
  iHole:= i;
  while (iHole > 0) and (A[iHole - 1].SortOrder > so) do begin
    A[iHole] := A[iHole - 1];
    dec(iHole);
  end;
  A[iHole]:= item; //to avoid unnecessary writing. A good algorithm is a dead... I mean, a cache-friendly algorithm.
end;

procedure TSortedList.Sort();
var
  item: TCgeObject;
  i, iHole: integer;
  so: fixed32;
begin
  //  http://en.wikipedia.org/wiki/Insertion_sort
  for i:=0 to length(A) - 1 do begin
    item:= A[i];
    so:= item.SortOrder;
    iHole:= i;
    while (iHole > 0) and (A[iHole - 1].SortOrder > so) do begin
      A[iHole] := A[iHole - 1];
      dec(iHole);
    end;
    if (iHole <> i) then A[iHole]:= item; //to avoid unnecessary writing. A good algorithm is a dead... I mean, a cache-friendly algorithm.
  end;
end;

function TSortedList.ExtractFirst: TCgeObject;
begin
  if Self.High < 0 then exit(nil);
  Result:= Remove(0);
end;

function TSortedList._GetLast: TCgeObject;
begin
  Result:= A[High];
end;


function TSortedList.Pop: TCgeObject;
begin
  Result:= A[High];
  SetLength(A, length(A) - 1);
end;

function TSortedList.Remove(i: integer): TCgeObject;
var j: integer;
begin
  Result:= A[i];
  for j:= i to High - 1 do A[j]:= A[j + 1];
  SetLength(A, length(A) - 1);
end;



procedure TSortedList.AfterLoading();
begin
  inherited;
  Prune;
end;

procedure TSortedList.Prune();
var i,j,k,maxl: integer;
  n: array of integer;
begin
  SetLength(n, Length(A));
  j:= -1;
  for i:=0 to system.High(A) do begin
    if Assigned(A[i]) then begin
      Inc(j);
      n[j]:= i;
    end;
  end;
  if j <> system.High(A) then begin
    for i:= 0 to j do A[j]:= A[n[j]];
    SetLength(A, j +1);
  end;
  SetLength(n, 0);
{ I am scared at times by the Lovecraftian depths of my mind >_<
     How did I even... conceive this... this thing?

  i:=0;
  maxl:= length(A);
  while (i < maxl) do begin
    j:= i;
    while (j < length(A)) and (not Assigned(A[j]) or A[j].Scraped) do inc(j);
    if (i <> j) then begin
      for k:=0 to (maxl - i - j) do A[j + k]:= A[i + k];
      maxl-= (i - j)
    end
    else i+= 1;
  end;
  if maxl <> length(A) then SetLength(a, maxl);
}
end;

function TSortedList._ReadData(i: integer): TCgeObject;
begin
  Result:= A[i];
end;

function TSortedList._GetCount: integer;
begin
  Result:= length(A);
end;

function TSortedList._GetHigh: integer;
begin
  Result:= system.high(A);
end;

procedure TSortedList.RegisterFields;
begin
  RegClass(TCgeObject);
  RegType(typeinfo(TArrayOfCgeObjects));//, typeinfo(TCgeObject));
  inherited;
  ListFields(['A', @A, typeinfo(TArrayOfCgeObjects)]);
end;

function TCgeObject._GetVolatile(): boolean;
begin
  Result:= (Self.CpsMask and VOLATILE_MASK) > 0;
end;

procedure TCgeObject._SetVolatile(v: boolean);
begin
  if v then Self.CpsMask:= Self.CpsMask or VOLATILE_MASK
       else  Self.CpsMask:= Self.CpsMask and not dword(VOLATILE_MASK);

end;

procedure TPreservableLink.UpdateHash;
begin
  Mother^.AssetKeeper.GenHash(@SessionGUID);
  AfterEfCheck;
  SessionGUID.ClassNameMd5:= ClassNameMd5;
end;

function TPreservableLink._GetObj(): TCgeObject;
begin
  if SessionGUID <> Module.SessionHash then o:= 0;
  Result:= TCgeObject(pointer(o));
end;

procedure TPreservableLink._SetObj(obj: TCgeObject);
begin
  TCgeObject(pointer(o)):= obj;
  SessionGUID:= Module.SessionHash;
end;


//fStatus: TBackgroundTaskStatusSet;(bts_completed, bts_started, bts_volatile)
{
function TBackgroundTask.ShouldTerminate: boolean;
begin
  Result:= Threadmanager.StopRequested;
end;

constructor TBackgroundTask.Create;
begin
  fStatus:= [];
  inherited;
  Module.BackgroundTasks.Add(Self);
end;


procedure TBackgroundTask.RemoveFromList;
var i: integer;
begin
  for i:= 0 to Module.BackgroundTasks.High do
    if (Self = Module.BackgroundTasks[i]) then begin
      Module.BackgroundTasks.Remove(i);
      Exit;
    end;
end;

procedure TBackgroundTask.AfterLoading;
begin
  if bts_volatile in fStatus then self.RemoveFromList;
end;


procedure TBackgroundTask.RegisterFields;
begin
  RegType(typeinfo(TBackgroundTaskStatusSet));
  inherited;
  ListFields([@fStatus, 'status', typeinfo(TBackgroundTaskStatusSet)]);
end;

procedure TBackgroundTask.Scrape;
begin
  RemoveFromList;
  inherited;
end;
}




end.
