{
    This file is part of chepersy
    Copyright (c) 2004-2015 by Anton Rzheshevski (chebmaster@mail.ru),

    See the file COPYING.CPS, included in this distribution,
    for details about the copyright.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

 **********************************************************************}



{  Scenario format: <number of entries>, ([<tind>, <offset>, <properties>] | [0, <offset> <size>]) }

type
  TScenarioEntry = packed record
    sTind,
    sOffset: integer; // in bytes
    case integer of
      0: (
        sSize: integer; //in dwords ( 1 = 4 bytes), according the file
         // and buffer structure where most things are aligned to 32 bits
      );
      1: (
        sFProps: TFieldPropertiesSet;
      )
  end;
  PScenarioEntry = ^TScenarioEntry;

  TOneScenario = array of TScenarioEntry;
  POneScenario = ^TOneScenario;

var
  Scenario,
  CloningScenario
    : array of TOneScenario;

  procedure CpsBuildsavingScenario(Cind: integer);
  //also used for loading if types match
  var
    i, ii, sizeind: integer;
    ilastbinary: integer; //plastbinary: PScenarioEntry;
  begin
//addlog ('Saving scenario %0 >', [cind]);
    if High(Scenario) < Cind then SetLength(Scenario, Cind + 1);
    ilastbinary:= -1; //plastbinary:= nil;
    For i:=0 to Fields[Cind].High do
      With Fields[Cind][i] do begin
//addlog(' %0 %1 %2',[NameSpace[Types[tind].name], Types[tind].containsComplexTypes, Types[tind].containsChepersyObjects]);
        if fpro_Skipped in properties {Skip} then begin
          ilastbinary:= -1; //plastbinary:= nil;
          Continue;
        end;
        if Types[tind].containsComplexTypes or Types[tind].containsChepersyObjects
        then begin
          Assert(Assigned(Types[tind].proc), 'Processing procedure is not set!');
          //complex field:
          ilastbinary:= -1; //plastbinary:= nil;
          SetLength(Scenario[Cind], Length(Scenario[Cind]) + 1);
          with Scenario[Cind][High(Scenario[Cind])] do begin
            sTind:= tind;
            sOffset:= offset;
            sFProps:= properties;
//addlog('  new #%0, offset %1',[High(Scenario[Cind]), sOffset]);
          end;
        end
        else begin
          //binary field:
          if (ilastbinary >= 0)//Assigned(plastbinary)
          and (Fields[Cind][i - 1].offset mod AlignGranularity = 0) //предыдущее выровнено)
          and (offset mod AlignGranularity = 0)//(это выровнено)
          then begin //merge(!) with previous.
            inc (Scenario[Cind][ilastbinary].sSize, SizeToBufferIndex(Types[tind].size));//inc(plastbinary^.sSize, SizeToBufferIndex(Types[tind].size));
//addlog('  type %2, merge with #%0, +%1',[ilastbinary, 4 * SizeToBufferIndex(Types[tind].size)]);
          end
          else begin //add normally
            SetLength(Scenario[Cind], Length(Scenario[Cind]) + 1);
            ilastbinary:= High(Scenario[Cind]); //plastbinary:= @Scenario[Cind][High(Scenario[Cind])];
            with Scenario[Cind][High(Scenario[Cind])] do begin
              sTind:= 0;
              sOffset:= offset;
              sSize:= SizeToBufferIndex(Types[tind].size);
//addlog('  new #%0, offset %1, size %2',[High(Scenario[Cind]), sOffset, 4 * sSize]);
            end;
          end;
        end;
      end;
//addlog ('Saving scenario %0 <', [cind]);
  end;

  procedure CpsFindLoadingScenario(Cind, B_LocalClassInd: integer);
  var
    i: integer;
  begin
    if (Fields[Cind].High <> DH.B_Fields[B_LocalClassInd].High) //number of fields had changed

    or DH.TypeChanged[DH.NaToT[DH.Tnames[B_LocalClassInd]]]
    then begin
      DH.B_ScenarioIndex[B_LocalClassInd]:= -1; //no scenario could be used for this record
      BasketWarningsAdd(
               MsgFormat(MI_BASKET_CONVERSION,
                         [LocalClassNameStr(B_LocalClassInd),
                          FieldKindToCapStr(DH.B_RecKinds[B_LocalClassInd])]));
      Exit;
    end;
    For i:=0 to DH.B_Fields[B_LocalClassInd].High do
      With DH.B_Fields[B_LocalClassInd][i] do
        if (RFInd <> i) //the order of fields had changed
        or (DH.TypeChanged[Tind]) //the field type requires conversion
        or (NewTind <> 0) //the field has changed its type
        then begin
          DH.B_ScenarioIndex[Cind]:= -1; //no scenario could be used for this record
          BasketWarningsAdd(
               MsgFormat(MI_BASKET_CONVERSION,
                          [LocalClassNameStr(B_LocalClassInd),
                           FieldKindToCapStr(DH.B_RecKinds[B_LocalClassInd])]));
          Exit;
        end;

    //else use the existing saving scenario
    DH.B_ScenarioIndex[B_LocalClassInd]:= Cind;//tScenarioIndex[Cind];
  end;


  procedure CpsBuildCloningScenario(Cind: integer);
  var
    i: integer;
    head, tail: ptrint;
  begin
    if High(CloningScenario) < Cind then SetLength(CloningScenario, Cind + 1);

    if Fields[Cind].High < 0 then Exit;

    //copy the entire record as a single block
    head:= Fields[Cind][0].offset;
    with Fields[Cind][Fields[Cind].High] do tail:= offset + Types[tind].size;
    SetLength(CloningScenario[Cind], Length(CloningScenario[Cind]) + 1);
    with CloningScenario[Cind][High(CloningScenario[Cind])] do begin
      sTind:= 0;
      sOffset:= head;
      sSize:= SizeToBufferIndex(tail - head);
    end;

    //deal with managed type fields
    for i:=0 to Fields[Cind].High do
      with Fields[Cind][i] do begin
        if Types[tind].containsComplexTypes or Types[tind].containsChepersyObjects
        then begin
          SetLength(CloningScenario[Cind], Length(CloningScenario[Cind]) + 1);
          with CloningScenario[Cind][High(CloningScenario[Cind])] do begin
            sTind:= tind;
            sOffset:= offset;
            sFProps:= properties;
          end;
        end;
      end;
  end;


  procedure CloneRecord(base, newbase: pointer; Cind: integer; fprops: TFieldPropertiesSet);
  //if base is nil it means someone has already copied the content. Only run the post-cloning processing
  var
    i, j: integer;
    p: pointer;
    mysc: POneScenario;
  begin
    mysc:= @CloningScenario[Cind];

    //first, copy *all* fields as binary:

    i:= 0;
    while (i < High(mysc^)) and (mysc^[i].sTind = 0) do begin
      if Assigned(base) then with mysc^[i] do
        Move(pointer(ptruint(base) + ptruint(sOffset))^,
             pointer(ptruint(newbase) + ptruint(sOffset))^,
             4 * sSize);
      inc(i);
    end;

    //performing post-cloning action on the managed types:

    for j:= i to High(mysc^) do with mysc^[j] do begin
      Assert(Assigned(Types[sTind].Proc), 'Chepersy internal error! Type '
        + NameSpace[Types[sTind].Name] + ' has no processing procedure.');

      Types[sTind].Proc (
        pointer(ptruint(newbase) + ptruint(sOffset)),
        fio_Clone, sTind,
        fprops + sFProps
      );
    end;
  end;

  procedure SaveRecord(base: pointer; Cind: integer; fprops: TFieldPropertiesSet);
  var
    j: integer;
    mysc: POneScenario;
  begin
    mysc:= @Scenario[Cind];

    //Always by scenario

    For j:= 0 to High(mysc^) do with mysc^[j] do begin
      if sTind > 0
        then
          Types[sTind].Proc (
            pointer(ptruint(base) + ptruint(sOffset)),
            fio_Save, sTind,
            fprops + sFProps)
        else
          WriteToBuffer(pointer(ptruint(base) + ptruint(sOffset))^, 4 * sSize);
    end;
  end;
  
  procedure WalkRecord(base: pointer; Cind: integer; fprops: TFieldPropertiesSet);
  var
    j: integer;
    mysc: POneScenario;
  begin
    mysc:= @Scenario[Cind];

    //Always by scenario (the saving one)

    For j:=1 to High(mysc^) do with mysc^[j] do
      if (sTind > 0) and Types[sTind].containsChepersyObjects
        then Types[sTind].Proc (
          pointer(ptruint(base) + ptruint(sOffset)),
          fio_Walk, sTind,
          fprops + sFProps);
  end;

  procedure LoadRecord(base: pointer; op: TFieldOperation; B_ClassIndex: integer;
    fprops: TFieldPropertiesSet);
  var
    i, j: integer;
    mysc: POneScenario;
  begin
    i:= DH.B_ScenarioIndex[B_ClassIndex];

    if i >=0 then begin
      //Load by scenario
      mysc:= @Scenario[i];
//addlog('LoadRecord >  (base=%0, scenario index=%1',[mysc, i]);
      For j:=0 to High(mysc^) do with mysc^[j] do
        if sTind > 0 then
//begin
//addlog('   %3  Offset=%4 Tind=%0 (%2), Types[Tind].proc=%1', [sTind, pointer(Types[sTind].proc), NameSpace[Types[sTind].name], j, sOffset]);

           Types[sTind].proc (
              pointer(ptruint(base) + ptruint(sOffset)),
              op, DH.B_FromGtind[sTind],
              fprops + sFProps
           )
//end
        else
//begin
//addlog('   %0  offset=%1 Binary, size=%2', [j, sOffset, sSize * 4]);
          ReadFromBuffer(pointer(ptruint(base) + ptruint(sOffset))^, 4 * sSize);
//end;
//addlog('LoadRecord <');
    end
    else begin
//addlog('LoadRecord >  (base=%0, scenario NOT FOUND, B_ClassIndex=%1 ',[mysc, B_ClassIndex]);
      // Scenario cannot be used for this record.
      //   Either the field list has changed, or some enumerated type has changed.

      // Load using the old routine, featuring a total, per-field conversion.
      for i:=0 to DH.B_Fields[B_ClassIndex].High do begin
        with DH.B_Fields[B_ClassIndex][i] do begin
          if fpro_Skipped in properties {Skip} then Continue; //Skipped at saving. Nothing to do.
{addlog('--- %0 %1 rfind%2 %3 %4 offset%5',[i, B_NameSpace[Name], rfind, B_NameSpace[Tnames[tind]], TypeNameSpace[newtind]
,Fields[CNums[B_ClassIndex]][RFind].offset
]);
}
          if (RFind >= 0) and (op = fio_Load) then begin
          //load the field
            if NewTind <> 0
              then GetConverterProc(DH.B_ToGtind[tind], NewTind)
                 (pointer(base) + Fields[DH.CNums[B_ClassIndex]][RFind].offset, tind, NewTind)
            else with Types[DH.B_ToGtind[Tind]] do
              if Assigned(proc)
                 then
                   proc(pointer(ptruint(base) + Fields[DH.CNums[B_ClassIndex]][RFind].offset), fio_Load, tind, properties)
                 else
                   ReadFromBuffer(pointer(ptruint(base) + Fields[DH.CNums[B_ClassIndex]][RFind].offset)^, 4 * SizeToBufferIndex(Size));
          end
          else begin
            //Skip this field, it doesn't exist in current implementation.
            with Types[Tind] do
              if Assigned(proc)
                then proc(nil, fio_Skip, tind, properties)
                else ReadInt; //CpsStream.Position:= CpsStream.Position + 4 * SizeToBufferIndex(Size);
          end;
        end;
//addlog('    pos %0', [CpsStream.Position]);
      end;
//addlog('LoadRecord <  (base=%0, scenario NOT FOUND, B_ClassIndex=%1 ',[mysc, B_ClassIndex]);
    end;
  end;

