{
    This file is part of chepersy
    Copyright (c) 2004-2015 by Anton Rzheshevski (chebmaster@mail.ru),

    See the file COPYING.CPS, included in this distribution,
    for details about the copyright.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

 **********************************************************************}


{
    This file contains headers for the sparce object index, a key feature
    of the layered game physics.
}


{
    This stuff is made in a somehow non-optimal way. There should have been
    several virtual methods for accessing the data array and calculating the
    relocation index, with methods like Add() being non-virtual instead.
    But I sacrificed convenience and code readability for speed: these indexes
    are vital for the performance of a heavily layered game world.
}


  TSparceIndexAbstractChunk = class(TChepersyObject)
  protected
    f_count: Longint;
  public
    procedure CopyTo(otherchunk: TSparceIndexAbstractChunk); virtual; abstract;

    function Get(off: dword): TChepersyObject; virtual; abstract;
    function Has(off: dword): boolean; virtual; abstract;

    function Add(off: dword; o: TChepersyObject): TSparceIndexAbstractChunk;
      virtual; abstract; { May upgrade itself to a larger chunk }

    function Remove(off: dword): TSparceIndexAbstractChunk;
      virtual; abstract; { May demote itself to a smaller chunk or just call
                                                    Self.Free and return nil! }
    procedure FreeContained; virtual; abstract;
   // procedure ScrapeContained; virtual; abstract;

    property Count: Longint read f_count;

    procedure RegisterFields; override;
  end;


  TSparceIndexChunk3 = class(TSparceIndexAbstractChunk)
  protected
    type
      T64DwordArray = array[0..63] of dword; //array[0..1023] of <2 bits>;
      TSIC3MAOArray = array[1..3] of TChepersyObject;
      TSIC3BIArray = array[1..3] of word;
    var
      data: TSIC3MAOArray;
      bind: TSIC3BIArray;
      reloc: T64DwordArray;
  public
    procedure CopyTo(otherchunk: TSparceIndexAbstractChunk); override;
    function Get(off: dword): TChepersyObject; override;
    function Has(off: dword): boolean; override;
    function Add(off: dword; o: TChepersyObject): TSparceIndexAbstractChunk; override;
    function Remove(off: dword): TSparceIndexAbstractChunk; override;
    procedure RegisterFields; override;
    procedure FreeContained; override;
   // procedure ScrapeContained; override;

  end;


  TSparceIndexChunk15 = class(TSparceIndexAbstractChunk)
  protected
    type
      T128DwordArray = array[0..127] of dword; //array[0..1023] of <4 bits>;
      TSIC15MAOArray = array[1..15] of TChepersyObject;
      TSIC15BIArray = array[1..15] of word;
    var
      data: TSIC15MAOArray;
      bind: TSIC15BIArray;
      reloc: T128DwordArray;
  public
    procedure CopyTo(otherchunk: TSparceIndexAbstractChunk); override;
    function Get(off: dword): TChepersyObject; override;
    function Has(off: dword): boolean; override;
    function Add(off: dword; o: TChepersyObject): TSparceIndexAbstractChunk; override;
    function Remove(off: dword): TSparceIndexAbstractChunk; override;
    procedure RegisterFields; override;
    procedure FreeContained; override;
  //  procedure ScrapeContained; override;

  end;



  TSparceIndexChunk255 = class(TSparceIndexAbstractChunk)
  protected
    type
      TThousandByteArray = array[0..1023] of Byte;
      TSIC256BIArray = array[1..255] of word;
    var
      data: TArrayOfChepersyObjects; { The array only shrinks during
        defragmentation, which only happens before saving }
      bind: TSIC256BIArray;
      reloc: TThousandByteArray;
    procedure Defragment;
  public
    procedure CopyTo(otherchunk: TSparceIndexAbstractChunk); override;
    function Get(off: dword): TChepersyObject; override;
    function Has(off: dword): boolean; override;
    function Add(off: dword; o: TChepersyObject): TSparceIndexAbstractChunk; override;
    function Remove(off: dword): TSparceIndexAbstractChunk; override;
    procedure RegisterFields; override;
    procedure BeforeSaving; override;
    procedure FreeContained; override;
   // procedure ScrapeContained; override;
    destructor Destroy; override;
  end;



  TNotSoSparceIndexChunk = class(TSparceIndexAbstractChunk)
  protected
  type
    TThousandMAOArray = array[0..1023] of TChepersyObject;
  var
    data: TThousandMAOArray; { if more than one fourth of the chunk is filled,
      switch to full-sized storage to avoid fragmentation }
  public
    procedure CopyTo(otherchunk: TSparceIndexAbstractChunk); override;
    function Get(off: dword): TChepersyObject; override;
    function Has(off: dword): boolean; override;
    function Add(off: dword; o: TChepersyObject): TSparceIndexAbstractChunk; override;
    function Remove(off: dword): TSparceIndexAbstractChunk; override;
    procedure RegisterFields; override;
    procedure FreeContained; override;
   // procedure ScrapeContained; override;
  end;



  TSparceObjectArray = class(TChepersyObject)
  protected
    type
      TSparceIndexChunksArray = array of TSparceIndexAbstractChunk;
    var
      chunks: TSparceIndexChunksArray;
      chunks_offset: dword;
      f_count: longint;
    procedure ShrinkArray();
  public
    procedure Clear;
    procedure FreeContained;
    //procedure ScrapeContained;
    procedure Add(index: Longint; o: TChepersyObject);
    procedure Remove(index: Longint);
    function Get(index: Longint): TChepersyObject;
    function Has(index: Longint): boolean;
    procedure RegisterFields; override;
    procedure BeforeSaving; override;
    property Count: Longint read f_count;
    property Data[index: Longint]: TChepersyObject read Get write Add; default;
  end;

{
  generic TSparceMOArray<T> = class(TSparceObjectArray)
  public
    procedure Add(index: Longint; o: T);
    function Get(index: Longint): T;
  end;
}



