{
    This file is part of chepersy
    Copyright (c) 2004-2015 by Anton Rzheshevski (chebmaster@mail.ru),

    See the file COPYING.CPS, included in this distribution,
    for details about the copyright.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

 **********************************************************************}


{
    This file contains implementation of the sparce object index, a key feature
    of the layered game physics.
}

  procedure TSparceIndexAbstractChunk.RegisterFields;
  begin
    inherited;
    ListFields([
      'f_count', @f_count, typeinfo(Longint)
    ]);
  end;

  procedure TSparceIndexChunk3.RegisterFields;
  begin
    RegType(typeinfo(T64DwordArray), typeinfo(dword), [0, 63]);
    RegType(typeinfo(TSIC3MAOArray), typeinfo(TChepersyObject), [0, 2]);
    //chepersy 0.9.2 still cannot support low walue different than 0
    RegType(typeinfo(TSIC3BIArray), typeinfo(word), [0, 2]);
    inherited;
    ListFields([
      'data', @data, typeinfo(TSIC3MAOArray),
      'bind', @bind, typeinfo(TSIC3BIArray),
      'reloc', @reloc, typeinfo(T64DwordArray)
    ]);
  end;

  procedure TSparceIndexChunk15.RegisterFields;
  begin
    RegType(typeinfo(T128DwordArray), typeinfo(dword), [0, 127]);
    RegType(typeinfo(TSIC15MAOArray), typeinfo(TChepersyObject), [0, 14]);
    RegType(typeinfo(TSIC15BIArray), typeinfo(word), [0, 14]);
    inherited;
    ListFields([
      'data', @data, typeinfo(TSIC15MAOArray),
      'bind', @bind, typeinfo(TSIC15BIArray),
      'reloc', @reloc, typeinfo(T128DwordArray)
    ]);
  end;

  procedure TSparceIndexChunk255.RegisterFields;
  begin
    RegType(typeinfo(TThousandByteArray), typeinfo(Byte), [0, 1023]);
    RegType(typeinfo(TSIC256BIArray), typeinfo(word), [0, 254]);
    inherited;
    ListFields([
      'data', @data, typeinfo(TArrayOfChepersyObjects),
      'bind', @bind, typeinfo(TSIC256BIArray),
      'reloc', @reloc, typeinfo(TThousandByteArray)
    ]);
  end;

  procedure TNotSoSparceIndexChunk.RegisterFields;
  begin
    RegType(typeinfo(TThousandMAOArray), typeinfo(TChepersyObject), [0, 1023]);
    inherited;
    ListFields([
      'data', @data, typeinfo(TThousandMAOArray)
    ]);
  end;

  procedure TSparceIndexChunk3.CopyTo(otherchunk: TSparceIndexAbstractChunk);
  var
    i: integer;
  begin
    if f_count = 0 then Exit;
    for i:= 1 to f_count do otherchunk.Add(bind[i], data[i]);
  end;

  procedure TSparceIndexChunk3.FreeContained;
  var
    i: integer;
  begin
    for i:= 1 to f_count do data[i].Free;
  end;
{
  procedure TSparceIndexChunk3.ScrapeContained;
  var
    i: integer;
  begin
    for i:= 1 to f_count do data[i].Scrape;
  end;
 }
  procedure TSparceIndexChunk15.CopyTo(otherchunk: TSparceIndexAbstractChunk);
  var
    i: integer;
  begin
    Assert(f_count > 0, 'TSparceIndexChunk15.CopyTo: is empty!');
    for i:= 1 to f_count do otherchunk.Add(bind[i], data[i]);
  end;

  procedure TSparceIndexChunk15.FreeContained;
  var
    i: integer;
  begin
    for i:= 1 to f_count do data[i].Free;
  end;

{
  procedure TSparceIndexChunk15.ScrapeContained;
  var
    i: integer;
  begin
    for i:= 1 to f_count do data[i].Scrape;
  end;
}

  procedure TSparceIndexChunk255.CopyTo(otherchunk: TSparceIndexAbstractChunk);
  var
    i: integer;
  begin
    Assert(f_count > 0, 'TSparceIndexChunk255.CopyTo: is empty!');
    Defragment;
    for i:= 1 to f_count do otherchunk.Add(bind[i], data[i]);
  end;


  procedure TSparceIndexChunk255.FreeContained;
  var
    i: integer;
  begin
    for i:= 1 to f_count do data[i].Free;
  end;

{
  procedure TSparceIndexChunk255.ScrapeContained;
  var
    i: integer;
  begin
    for i:= 1 to f_count do data[i].Scrape;
  end;
}


  procedure TNotSoSparceIndexChunk.CopyTo(otherchunk: TSparceIndexAbstractChunk);
  var
    i: integer;
  begin
    Assert(f_count > 0, 'TNotSoSparceIndexChunk.CopyTo: is empty!');
    for i:= 0 to 1023 do
      if Assigned(data[i])
        then otherchunk.Add(i, data[i]);
  end;


  procedure TNotSoSparceIndexChunk.FreeContained;
  var
    i: integer;
  begin
    for i:= 0 to 1023 do if Assigned(data[i]) then data[i].Free;
  end;

{
  procedure TNotSoSparceIndexChunk.ScrapeContained;
  var
    i: integer;
  begin
    for i:= 0 to 1023 do if Assigned(data[i]) then data[i].Scrape;
  end;
}


  function TSparceIndexChunk3.Get(off: dword): TChepersyObject;
  var
    reloc_ind, rle, rlo, rlmask: dword;
  begin
    //16 2-bit indices packed into one dword
    rlo:= (off and $f) shl 1;
    rle:= off shr 4;
    rlmask:= 3;
    reloc_ind:= (reloc[rle] shr rlo) and rlmask;
    if reloc_ind = 0
      then Result:= nil
      else Result:= data[reloc_ind];
  end;

  function TSparceIndexChunk3.Has(off: dword): boolean;
  var
    rle, rlo, rlmask: dword;
  begin
    rlo:= (off and $f) shl 1;
    rle:= off shr 4;
    rlmask:= 3;
    Result:= ((reloc[rle] shr rlo) and rlmask) > 0;
  end;

  function TSparceIndexChunk3.Add(off: dword; o: TChepersyObject): TSparceIndexAbstractChunk;
  var
    reloc_ind, rle, rlo, rlmask: dword;
  begin
    Result:= Self;
    //16 2-bit indices packed into one dword
    rlo:= (off and $f) shl 1;
    rle:= off shr 4;
    rlmask:= 3;
    if f_count > 0
      then reloc_ind:= (reloc[rle] shr rlo) and rlmask
      else reloc_ind:= 0;
    if reloc_ind > 0 then begin
      //replace existing
      data[reloc_ind]:= o;
      Result:= Self;
    end
    else begin
      if f_count < 3 then begin
        Inc(f_count);
        reloc_ind:= f_count; //the data array here is indexed from 1!
        reloc[rle]:= reloc[rle] or (f_count shl rlo);
        data[reloc_ind]:= o;
        bind[reloc_ind]:= off;
        Result:= Self;
      end
      else begin
        //replace self with a bigger sized chunk
        Result:= TSparceIndexChunk15.Create;
        Self.CopyTo(Result);
        Result.Add(off, o);
        Self.Free;
      end;
    end;
  end;

  function TSparceIndexChunk3.Remove(off: dword): TSparceIndexAbstractChunk;
  var
    reloc_ind, rle, rlo, rlee, rloe, rlmask: dword;
  begin
    //16 2-bit indices packed into one dword
    rlo:= (off and $f) shl 1;
    rle:= off shr 4;
    rlmask:= 3;
    reloc_ind:= (reloc[rle] shr rlo) and rlmask;
    if reloc_ind = 0 then Exit(Self);
    reloc[rle]:= reloc[rle] and not (rlmask shl rlo); //reloc[off]:= 0;
    if f_count = 1 then begin
      Result:= nil;
      Self.Free;
    end
    else begin
      Result:= Self;
      if reloc_ind < f_count then begin
        //compact the array by moving the last element to the freeing space
        rloe:= (bind[f_count] and $f) shl 1;
        rlee:= bind[f_count] shr 4;
        data[reloc_ind]:= data[f_count];
        bind[reloc_ind]:= bind[f_count];
        reloc[rlee]:= (reloc[rlee] and not (rlmask shl rloe))
          or (reloc_ind shl rlo); //reloc[bind[f_count]]:= reloc_ind;
      end;
      data[f_count]:= nil;
      dec(f_count);
    end;
  end;




  function TSparceIndexChunk15.Get(off: dword): TChepersyObject;
  var
    reloc_ind, rle, rlo, rlmask: dword;
  begin
    //8 4-bit indices packed into one dword
    rlo:= (off and $7) shl 2;
    rle:= off shr 3;
    rlmask:= $f;
    reloc_ind:= (reloc[rle] shr rlo) and rlmask;
    if reloc_ind = 0
      then Result:= nil
      else Result:= data[reloc_ind];
  end;


  function TSparceIndexChunk15.Has(off: dword): boolean;
  var
    rle, rlo, rlmask: dword;
  begin
    rlo:= (off and $7) shl 2;
    rle:= off shr 3;
    rlmask:= $f;
    Result:= ((reloc[rle] shr rlo) and rlmask) > 0
  end;


  function TSparceIndexChunk15.Add(off: dword; o: TChepersyObject): TSparceIndexAbstractChunk;
  var
    reloc_ind, rle, rlo, rlmask: dword;
  begin
    Result:= Self;
    //8 4-bit indices packed into one dword
    rlo:= (off and $7) shl 2;
    rle:= off shr 3;
    rlmask:= $f;
    if f_count > 0
      then reloc_ind:= (reloc[rle] shr rlo) and rlmask
      else reloc_ind:= 0;
    if reloc_ind > 0 then begin
      //replace existing
      data[reloc_ind]:= o;
      Result:= Self;
    end
    else begin
      if f_count < 15 then begin
        Inc(f_count);
        reloc_ind:= f_count; //the data array here is indexed from 1!
        reloc[rle]:= reloc[rle] or (f_count shl rlo);
        data[reloc_ind]:= o;
        bind[reloc_ind]:= off;
        Result:= Self;
      end
      else begin
        //replace self with a bigger sized chunk
        Result:= TSparceIndexChunk255.Create;
        Self.CopyTo(Result);
        Result.Add(off, o);
        Self.Free;
      end;
    end;
  end;

  function TSparceIndexChunk15.Remove(off: dword): TSparceIndexAbstractChunk;
  var
    reloc_ind, rle, rlo, rlee, rloe, rlmask: dword;
  begin
    //8 4-bit indices packed into one dword
    rlo:= (off and $7) shl 2;
    rle:= off shr 3;
    rlmask:= $f;
    reloc_ind:= (reloc[rle] shr rlo) and rlmask;
    if reloc_ind = 0 then Exit(Self);
    reloc[rle]:= reloc[rle] and not (rlmask shl rlo); //reloc[off]:= 0;

    Result:= Self;
    if reloc_ind < f_count then begin
      //compact the array by moving the last element to the freeing space
      rloe:= (bind[f_count] and $7) shl 2;
      rlee:= bind[f_count] shr 3;
      data[reloc_ind]:= data[f_count];
      bind[reloc_ind]:= bind[f_count];
      reloc[rlee]:= (reloc[rlee] and not (rlmask shl rloe))
        or (reloc_ind shl rlo); //reloc[bind[f_count]]:= reloc_ind;
    end;
    data[f_count]:= nil;
    dec(f_count);

    if f_count <= 3 then begin
      Result:= TSparceIndexChunk3.Create;
      Self.CopyTo(Result);
      Self.Free;
    end
    else Result:= Self;

  end;


  function TSparceIndexChunk255.Get(off: dword): TChepersyObject;
  begin
    if reloc[off] > 0
      then Result:= data[reloc[off]]
      else Result:= nil;
  end;


  function TSparceIndexChunk255.Has(off: dword): boolean;
  begin
    Result:= reloc[off] > 0;
  end;


  function TSparceIndexChunk255.Add(off: dword; o: TChepersyObject): TSparceIndexAbstractChunk;
  var
    reloc_ind: dword;
  begin
    Result:= Self;
    if f_count > 0
      then reloc_ind:= reloc[off]
      else reloc_ind:= 0;
    if reloc_ind > 0 then begin
      //replace existing
      data[reloc_ind]:= o;
      Result:= Self;
    end
    else begin
      if f_count < 255 then begin
        Inc(f_count);
        reloc_ind:= f_count;
        reloc[off]:= f_count;
        if Length(data) < f_count + 1 then SetLength(data, Math.min(256, f_count + 1 + 50));
        data[reloc_ind]:= o;
        bind[reloc_ind]:= off;
        Result:= Self;
      end
      else begin
        //replace self with a bigger sized chunk
        Result:= TNotSoSparceIndexChunk.Create;
        Self.CopyTo(Result);
        Result.Add(off, o);
        Self.Free;
      end;
    end;
  end;

  function TSparceIndexChunk255.Remove(off: dword): TSparceIndexAbstractChunk;
  var
    reloc_ind: dword;
  begin
    reloc_ind:= reloc[off];
    if reloc_ind = 0 then Exit(Self);
    reloc[off]:= 0;

    Result:= Self;
    if reloc_ind < f_count then begin
      //compact the array by moving the last element to the freeing space
      data[reloc_ind]:= data[f_count];
      bind[reloc_ind]:= bind[f_count];
      reloc[bind[f_count]]:= reloc_ind;
    end;
    data[f_count]:= nil;
    dec(f_count);

    if f_count <= 10 then begin
      Result:= TSparceIndexChunk15.Create;
      Self.CopyTo(Result);
      Self.Free;
    end
    else Result:= Self;
  end;

  procedure TSparceIndexChunk255.Defragment;
  var
    reloc2: TThousandByteArray;
    bind2: TSIC256BIArray;
    data2: array[0..255] of TChepersyObject;
    i, m: integer;
  begin
    FillChar(bind2, sizeof(bind2), 0);
    FillChar(reloc2, sizeof(reloc2), 0);
    m:= 1;
    for i:=0 to 1023 do begin
      if reloc[i] = 0 then continue;
      reloc2[i]:= m;
      bind2[m]:= i;
      data2[m]:= data[reloc[i]];
      Inc(m);
    end;
    MOVE(reloc2, reloc, sizeof(reloc));
    MOVE(bind2, bind, sizeof(bind));
    MOVE(data2, data[0], (f_count + 1) * sizeof(pointer)); { Avoid memory
                                 reallocation, this is important for speed. }
  end;

  destructor TSparceIndexChunk255.Destroy;
  begin
    SetLength(data, 0);
    inherited;
  end;

  procedure TSparceIndexChunk255.BeforeSaving;
  begin
    Defragment;
  end;


  function TNotSoSparceIndexChunk.Get(off: dword): TChepersyObject;
  begin
    Result:= data[off];
  end;

  function TNotSoSparceIndexChunk.Has(off: dword): boolean;
  begin
    Result:= Assigned(data[off]);
  end;


  function TNotSoSparceIndexChunk.Add(off: dword; o: TChepersyObject): TSparceIndexAbstractChunk;
  begin
    if not Assigned(data[off]) then inc(f_count);
    data[off]:= o;
  end;

  function TNotSoSparceIndexChunk.Remove(off: dword): TSparceIndexAbstractChunk;
  begin
    if not Assigned(data[off]) then Exit(Self);
    data[off]:= nil;
    dec(f_count);
    if f_count <= 180 then begin
      Result:= TSparceIndexChunk255.Create;
      Self.CopyTo(Result);
      Self.Free;
    end
    else Result:= Self;
  end;


  procedure TSparceObjectArray.ShrinkArray();
  var
    first_nonempty_chunk, last_nonempty_chunk: integer;
  begin
    if Length(chunks) = 0 then Exit;
    first_nonempty_chunk:= 0;
    while (first_nonempty_chunk < High(chunks))
      and not Assigned(chunks[first_nonempty_chunk])
        do Inc(first_nonempty_chunk);
    last_nonempty_chunk:= High(chunks);
    while (last_nonempty_chunk > first_nonempty_chunk)
      and not Assigned(chunks[last_nonempty_chunk])
        do Dec(last_nonempty_chunk);
    if (last_nonempty_chunk = first_nonempty_chunk)
      and not Assigned(chunks[last_nonempty_chunk])
    then begin
      SetLength(chunks, 0);
      chunks_offset:= 0;
      Exit;
    end;
    chunks:= COPY(chunks, first_nonempty_chunk,
                               last_nonempty_chunk - first_nonempty_chunk + 1);
    chunks_offset+= first_nonempty_chunk;
  end;

  procedure TSparceObjectArray.Clear;
  var
    i: integer;
  begin
    for i:= 0 to High(chunks) do
      if Assigned(chunks[i])
        then chunks[i].Free;
    SetLength(chunks, 0);
    chunks_offset:= 0;
    f_count:= 0;
  end;

  procedure TSparceObjectArray.FreeContained;
  var
    i: integer;
  begin
    if Assigned(chunks) then
      for i:= 0 to High(chunks) do
        if Assigned(chunks[i]) then begin
          chunks[i].FreeContained;
          chunks[i].Free;
        end;
    SetLength(chunks, 0);
    chunks_offset:= 0;
    f_count:= 0;
  end;

{
  procedure TSparceObjectArray.ScrapeContained;
  var
    i: integer;
  begin
    if Assigned(chunks) then
      for i:= 0 to High(chunks) do
        if Assigned(chunks[i]) then begin
          chunks[i].ScrapeContained;
          chunks[i].Free;
        end;
    SetLength(chunks, 0);
    chunks_offset:= 0;
    f_count:= 0;
  end;
}
  procedure TSparceObjectArray.Add(index: Longint; o: TChepersyObject);
  var
    coff, i, j: integer;
    newchunks: TSparceIndexChunksArray;
    newchunk: TSparceIndexAbstractChunk;
    prevccount: Longint;
  begin
    if not Assigned(o) then begin
      Remove(index);
      Exit;
    end;
    Assert(index > 0, 'TSparceObjectArray.Add() : index must be positive');
    coff:= integer(dword(index) shr 10) - chunks_offset;
    if coff < 0 then begin
      SetLength(newchunks, Length(chunks) + (- coff));
      for i:= 0 to -coff do newchunks[i]:= nil;
      for i:= 0 to High(chunks) do newchunks[-coff + i]:= chunks[i];
      chunks_offset-= (- coff);
      coff:= 0;
    end;
    if coff > High(chunks) then begin
      j:= Length(chunks);
      SetLength(chunks, coff + 1);
      for i:= j to High(chunks) do chunks[i]:= nil;
    end;
    if not Assigned(chunks[coff]) then chunks[coff]:= TSparceIndexChunk3.Create;
    prevccount:= chunks[coff].Count;
    newchunk:= chunks[coff].Add(dword(index) and dword($000003ff), o); { The
      chunk may create a bigger one and suicide. This check here is to be more
      cache friendly, to not write to the array unnecessarily. }
    if newchunk <> chunks[coff] then chunks[coff]:= newchunk;
    if Assigned(newchunk)
      then f_count += newchunk.Count - prevccount
      else f_count -= prevccount;
  end;

  procedure TSparceObjectArray.Remove(index: Longint);
  var
    coff: Integer;
    newchunk: TSparceIndexAbstractChunk;
    prevccount: Longint;
  begin
    Assert(index > 0, 'TSparceObjectArray.Remove() : index must be positive');
    coff:= integer(dword(index) shr 10) - chunks_offset;
    if (coff < 0) or (coff > High(chunks)) or not Assigned(chunks[coff]) then Exit;
    prevccount:= chunks[coff].Count;
    newchunk:= chunks[coff].Remove(dword(index) and dword($000003ff));
    if newchunk <> chunks[coff] then chunks[coff]:= newchunk;
    if Assigned(newchunk)
      then f_count += newchunk.Count - prevccount
      else f_count -= prevccount;
    if f_count = 0 then begin
      SetLength(chunks, 0);
      chunks_offset:= 0;
    end;
  end;

  function TSparceObjectArray.Get(index: Longint): TChepersyObject;
  var
    coff: integer;
  begin
    Assert(index > 0, 'TSparceObjectArray.Get() : index must be positive');
    coff:= integer(dword(index) shr 10) - chunks_offset;
    if (coff < 0) or (coff > High(chunks)) or not Assigned(chunks[coff])
      then Result:= nil
      else Result:= chunks[coff].Get(dword(index) and dword($000003ff));
  end;

  function TSparceObjectArray.Has(index: Longint): boolean;
  var
    coff: integer;
  begin
    Assert(index > 0, 'TSparceObjectArray.Has() : index must be positive');
    coff:= integer(dword(index) shr 10) - chunks_offset;
    if (coff < 0) or (coff > High(chunks)) or not Assigned(chunks[coff])
      then Result:= false
      else Result:= chunks[coff].Has(dword(index) and dword($000003ff));
  end;


  procedure TSparceObjectArray.RegisterFields;
  begin
    RegClass(TSparceIndexAbstractChunk);
    RegClass(TSparceIndexChunk3);
    RegClass(TSparceIndexChunk15);
    RegClass(TSparceIndexChunk255);
    RegClass(TNotSoSparceIndexChunk);
    RegType(typeinfo(TSparceIndexChunksArray));//, typeinfo(TSparceIndexAbstractChunk));
    inherited;
    ListFields([
      'chunks', @chunks, typeinfo(TSparceIndexChunksArray),
      'chunks_offset', @chunks_offset, typeinfo(dword),
      'f_count', @f_count, typeinfo(Longint)
    ]);
  end;

  procedure TSparceObjectArray.BeforeSaving;
  begin
    ShrinkArray();
  end;

