{
    This file is part of chepersy
    Copyright (c) 2004-2015 by Anton Rzheshevski (chebmaster@mail.ru),

    See the file COPYING.CPS, included in this distribution,
    for details about the copyright.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

 **********************************************************************}


{
    This file contains implementation of the sparce longint array.
}

  procedure TSparceLongintArrayAbstractChunk.RegisterFields;
  begin
    inherited;
    ListFields([
      'f_count', @f_count, typeinfo(Longint)
    ]);
  end;

  procedure TSparceLongintArrayChunk3.RegisterFields;
  begin
    RegType(typeinfo(T64DwordArray), typeinfo(dword), [0, 63]);
    RegType(typeinfo(TSLAC3MAOArray), typeinfo(TChepersyObject), [0, 2]);
    //chepersy 0.9.2 still cannot support low walue different than 0
    RegType(typeinfo(TSLAC3BIArray), typeinfo(word), [0, 2]);
    inherited;
    ListFields([
      'data', @data, typeinfo(TSLAC3MAOArray),
      'bind', @bind, typeinfo(TSLAC3BIArray),
      'reloc', @reloc, typeinfo(T64DwordArray)
    ]);
  end;

  procedure TSparceLongintArrayChunk15.RegisterFields;
  begin
    RegType(typeinfo(T128DwordArray), typeinfo(dword), [0, 127]);
    RegType(typeinfo(TSLAC15MAOArray), typeinfo(TChepersyObject), [0, 14]);
    RegType(typeinfo(TSLAC15BIArray), typeinfo(word), [0, 14]);
    inherited;
    ListFields([
      'data', @data, typeinfo(TSLAC15MAOArray),
      'bind', @bind, typeinfo(TSLAC15BIArray),
      'reloc', @reloc, typeinfo(T128DwordArray)
    ]);
  end;

  procedure TSparceLongintArrayChunk255.RegisterFields;
  begin
    RegType(typeinfo(TThousandByteArray), typeinfo(Byte), [0, 1023]);
    RegType(typeinfo(TSLAC256BIArray), typeinfo(word), [0, 254]);
    RegType(typeinfo(TArrayOfLongint));//, typeinfo(Longint));
    inherited;
    ListFields([
      'data', @data, typeinfo(TArrayOfLongint),
      'bind', @bind, typeinfo(TSLAC256BIArray),
      'reloc', @reloc, typeinfo(TThousandByteArray)
    ]);
  end;

  procedure TNotSoSparceLongintArrayChunk.RegisterFields;
  begin
    RegType(typeinfo(TThousandLongintArray), typeinfo(TChepersyObject), [0, 1023]);
    inherited;
    ListFields([
      'data', @data, typeinfo(TThousandLongintArray)
    ]);
  end;

  procedure TSparceLongintArrayChunk3.CopyTo(otherchunk: TSparceLongintArrayAbstractChunk);
  var
    i: integer;
  begin
    if f_count = 0 then Exit;
    for i:= 1 to f_count do otherchunk.Add(bind[i], data[i]);
  end;


  procedure TSparceLongintArrayChunk15.CopyTo(otherchunk: TSparceLongintArrayAbstractChunk);
  var
    i: integer;
  begin
    Assert(f_count > 0, 'TSparceLongintArrayChunk15.CopyTo: is empty!');
    for i:= 1 to f_count do otherchunk.Add(bind[i], data[i]);
  end;


  procedure TSparceLongintArrayChunk255.CopyTo(otherchunk: TSparceLongintArrayAbstractChunk);
  var
    i: integer;
  begin
    Assert(f_count > 0, 'TSparceLongintArrayChunk255.CopyTo: is empty!');
    Defragment;
    for i:= 1 to f_count do otherchunk.Add(bind[i], data[i]);
  end;




  procedure TNotSoSparceLongintArrayChunk.CopyTo(otherchunk: TSparceLongintArrayAbstractChunk);
  var
    i: integer;
  begin
    Assert(f_count > 0, 'TNotSoSparceLongintArrayChunk.CopyTo: is empty!');
    for i:= 0 to 1023 do
      if data[i] <> 0
        then otherchunk.Add(i, data[i]);
  end;




  function TSparceLongintArrayChunk3.Get(off: dword): Longint;
  var
    reloc_ind, rle, rlo, rlmask: dword;
  begin
    //16 2-bit indices packed into one dword
    rlo:= (off and $f) shl 1;
    rle:= off shr 4;
    rlmask:= 3;
    reloc_ind:= (reloc[rle] shr rlo) and rlmask;
    if reloc_ind = 0
      then Result:= 0
      else Result:= data[reloc_ind];
  end;


  function TSparceLongintArrayChunk3.Add(off: dword; L: Longint): TSparceLongintArrayAbstractChunk;
  var
    reloc_ind, rle, rlo, rlmask: dword;
  begin
    Result:= Self;
    //16 2-bit indices packed into one dword
    rlo:= (off and $f) shl 1;
    rle:= off shr 4;
    rlmask:= 3;
    if f_count > 0
      then reloc_ind:= (reloc[rle] shr rlo) and rlmask
      else reloc_ind:= 0;
    if reloc_ind > 0 then begin
      //replace existing
      data[reloc_ind]:= L;
      Result:= Self;
    end
    else begin
      if f_count < 3 then begin
        Inc(f_count);
        reloc_ind:= f_count; //the data array here is indexed from 1!
        reloc[rle]:= reloc[rle] or (f_count shl rlo);
        data[reloc_ind]:= L;
        bind[reloc_ind]:= off;
        Result:= Self;
      end
      else begin
        //replace self with a bigger sized chunk
        Result:= TSparceLongintArrayChunk15.Create;
        Self.CopyTo(Result);
        Result.Add(off, L);
        Self.Free;
      end;
    end;
  end;

  function TSparceLongintArrayChunk3.Remove(off: dword): TSparceLongintArrayAbstractChunk;
  var
    reloc_ind, rle, rlo, rlee, rloe, rlmask: dword;
  begin
    //16 2-bit indices packed into one dword
    rlo:= (off and $f) shl 1;
    rle:= off shr 4;
    rlmask:= 3;
    reloc_ind:= (reloc[rle] shr rlo) and rlmask;
    if reloc_ind = 0 then Exit(Self);
    reloc[rle]:= reloc[rle] and not (rlmask shl rlo); //reloc[off]:= 0;
    if f_count = 1 then begin
      Result:= nil;
      Self.Free;
    end
    else begin
      Result:= Self;
      if reloc_ind < f_count then begin
        //compact the array by moving the last element to the freeing space
        rloe:= (bind[f_count] and $f) shl 1;
        rlee:= bind[f_count] shr 4;
        data[reloc_ind]:= data[f_count];
        bind[reloc_ind]:= bind[f_count];
        reloc[rlee]:= (reloc[rlee] and not (rlmask shl rloe))
          or (reloc_ind shl rlo); //reloc[bind[f_count]]:= reloc_ind;
      end;
      data[f_count]:= 0;
      dec(f_count);
    end;
  end;




  function TSparceLongintArrayChunk15.Get(off: dword): Longint;
  var
    reloc_ind, rle, rlo, rlmask: dword;
  begin
    //8 4-bit indices packed into one dword
    rlo:= (off and $7) shl 2;
    rle:= off shr 3;
    rlmask:= $f;
    reloc_ind:= (reloc[rle] shr rlo) and rlmask;
    if reloc_ind = 0
      then Result:= 0
      else Result:= data[reloc_ind];
  end;



  function TSparceLongintArrayChunk15.Add(off: dword; L: Longint): TSparceLongintArrayAbstractChunk;
  var
    reloc_ind, rle, rlo, rlmask: dword;
  begin
    Result:= Self;
    //8 4-bit indices packed into one dword
    rlo:= (off and $7) shl 2;
    rle:= off shr 3;
    rlmask:= $f;
    if f_count > 0
      then reloc_ind:= (reloc[rle] shr rlo) and rlmask
      else reloc_ind:= 0;
    if reloc_ind > 0 then begin
      //replace existing
      data[reloc_ind]:= L;
      Result:= Self;
    end
    else begin
      if f_count < 15 then begin
        Inc(f_count);
        reloc_ind:= f_count; //the data array here is indexed from 1!
        reloc[rle]:= reloc[rle] or (f_count shl rlo);
        data[reloc_ind]:= L;
        bind[reloc_ind]:= off;
        Result:= Self;
      end
      else begin
        //replace self with a bigger sized chunk
        Result:= TSparceLongintArrayChunk255.Create;
        Self.CopyTo(Result);
        Result.Add(off, L);
        Self.Free;
      end;
    end;
  end;

  function TSparceLongintArrayChunk15.Remove(off: dword): TSparceLongintArrayAbstractChunk;
  var
    reloc_ind, rle, rlo, rlee, rloe, rlmask: dword;
  begin
    //8 4-bit indices packed into one dword
    rlo:= (off and $7) shl 2;
    rle:= off shr 3;
    rlmask:= $f;
    reloc_ind:= (reloc[rle] shr rlo) and rlmask;
    if reloc_ind = 0 then Exit(Self);
    reloc[rle]:= reloc[rle] and not (rlmask shl rlo); //reloc[off]:= 0;

    Result:= Self;
    if reloc_ind < f_count then begin
      //compact the array by moving the last element to the freeing space
      rloe:= (bind[f_count] and $7) shl 2;
      rlee:= bind[f_count] shr 3;
      data[reloc_ind]:= data[f_count];
      bind[reloc_ind]:= bind[f_count];
      reloc[rlee]:= (reloc[rlee] and not (rlmask shl rloe))
        or (reloc_ind shl rlo); //reloc[bind[f_count]]:= reloc_ind;
    end;
    data[f_count]:= 0;
    dec(f_count);

    if f_count <= 3 then begin
      Result:= TSparceLongintArrayChunk3.Create;
      Self.CopyTo(Result);
      Self.Free;
    end
    else Result:= Self;

  end;


  function TSparceLongintArrayChunk255.Get(off: dword): Longint;
  begin
    if reloc[off] > 0
      then Result:= data[reloc[off]]
      else Result:= 0;
  end;



  function TSparceLongintArrayChunk255.Add(off: dword; L: Longint): TSparceLongintArrayAbstractChunk;
  var
    reloc_ind: dword;
  begin
    Result:= Self;
    if f_count > 0
      then reloc_ind:= reloc[off]
      else reloc_ind:= 0;
    if reloc_ind > 0 then begin
      //replace existing
      data[reloc_ind]:= L;
      Result:= Self;
    end
    else begin
      if f_count < 255 then begin
        Inc(f_count);
        reloc_ind:= f_count;
        reloc[off]:= f_count;
        if Length(data) < f_count + 1 then SetLength(data, Math.min(256, f_count + 1 + 50));
        data[reloc_ind]:= L;
        bind[reloc_ind]:= off;
        Result:= Self;
      end
      else begin
        //replace self with a bigger sized chunk
        Result:= TNotSoSparceLongintArrayChunk.Create;
        Self.CopyTo(Result);
        Result.Add(off, L);
        Self.Free;
      end;
    end;
  end;

  function TSparceLongintArrayChunk255.Remove(off: dword): TSparceLongintArrayAbstractChunk;
  var
    reloc_ind: dword;
  begin
    reloc_ind:= reloc[off];
    if reloc_ind = 0 then Exit(Self);
    reloc[off]:= 0;

    Result:= Self;
    if reloc_ind < f_count then begin
      //compact the array by moving the last element to the freeing space
      data[reloc_ind]:= data[f_count];
      bind[reloc_ind]:= bind[f_count];
      reloc[bind[f_count]]:= reloc_ind;
    end;
    data[f_count]:= 0;
    dec(f_count);

    if f_count <= 10 then begin
      Result:= TSparceLongintArrayChunk15.Create;
      Self.CopyTo(Result);
      Self.Free;
    end
    else Result:= Self;
  end;

  procedure TSparceLongintArrayChunk255.Defragment;
  var
    reloc2: TThousandByteArray;
    bind2: TSLAC256BIArray;
    data2: array[0..255] of Longint;
    i, m: integer;
  begin
    FillChar(bind2, sizeof(bind2), 0);
    FillChar(reloc2, sizeof(reloc2), 0);
    m:= 1;
    for i:=0 to 1023 do begin
      if reloc[i] = 0 then continue;
      reloc2[i]:= m;
      bind2[m]:= i;
      data2[m]:= data[reloc[i]];
      Inc(m);
    end;
    MOVE(reloc2, reloc, sizeof(reloc));
    MOVE(bind2, bind, sizeof(bind));
    MOVE(data2, data[0], (f_count + 1) * sizeof(pointer)); { Avoid memory
                                 reallocation, this is important for speed. }
  end;

  destructor TSparceLongintArrayChunk255.Destroy;
  begin
    SetLength(data, 0);
    inherited;
  end;

  procedure TSparceLongintArrayChunk255.BeforeSaving;
  begin
    Defragment;
  end;


  function TNotSoSparceLongintArrayChunk.Get(off: dword): Longint;
  begin
    Result:= data[off];
  end;

  function TNotSoSparceLongintArrayChunk.Add(off: dword; L: Longint): TSparceLongintArrayAbstractChunk;
  begin
    if data[off] = 0 then inc(f_count);
    data[off]:= L;
  end;

  function TNotSoSparceLongintArrayChunk.Remove(off: dword): TSparceLongintArrayAbstractChunk;
  begin
    if data[off] = 0 then Exit(Self);
    data[off]:= 0;
    dec(f_count);
    if f_count <= 180 then begin
      Result:= TSparceLongintArrayChunk255.Create;
      Self.CopyTo(Result);
      Self.Free;
    end
    else Result:= Self;
  end;


  procedure TSparceLongintArray.ShrinkArray();
  var
    first_nonempty_chunk, last_nonempty_chunk: integer;
  begin
    if Length(chunks) = 0 then Exit;
    first_nonempty_chunk:= 0;
    while (first_nonempty_chunk < High(chunks))
      and not Assigned(chunks[first_nonempty_chunk])
        do Inc(first_nonempty_chunk);
    last_nonempty_chunk:= High(chunks);
    while (last_nonempty_chunk > first_nonempty_chunk)
      and not Assigned(chunks[last_nonempty_chunk])
        do Dec(last_nonempty_chunk);
    if (last_nonempty_chunk = first_nonempty_chunk)
      and not Assigned(chunks[last_nonempty_chunk])
    then begin
      SetLength(chunks, 0);
      chunks_offset:= 0;
      Exit;
    end;
    chunks:= COPY(chunks, first_nonempty_chunk,
                               last_nonempty_chunk - first_nonempty_chunk + 1);
    chunks_offset+= first_nonempty_chunk;
  end;

  procedure TSparceLongintArray.Clear;
  var
    i: integer;
  begin
    for i:= 0 to High(chunks) do
      if Assigned(chunks[i])
        then chunks[i].Free;
    SetLength(chunks, 0);
    chunks_offset:= 0;
    f_count:= 0;
  end;


  procedure TSparceLongintArray.Add(index: Longint; L: Longint);
  var
    coff, i, j: integer;
    newchunks: TSparceLongintArrayChunksArray;
    newchunk: TSparceLongintArrayAbstractChunk;
    prevccount: Longint;
  begin
    Assert(index > 0, 'TSparceLongintArray.Add() : index must be positive');
    coff:= integer(dword(index) shr 10) - chunks_offset;
    if coff < 0 then begin
      if L = 0 then Exit;
      SetLength(newchunks, Length(chunks) + (- coff));
      for i:= 0 to -coff do newchunks[i]:= nil;
      for i:= 0 to High(chunks) do newchunks[-coff + i]:= chunks[i];
      chunks_offset-= (- coff);
      coff:= 0;
    end;
    if coff > High(chunks) then begin
      if L = 0 then Exit;
      j:= Length(chunks);
      SetLength(chunks, coff + 1);
      for i:= j to High(chunks) do chunks[i]:= nil;
    end;
    if not Assigned(chunks[coff]) then begin
      if L = 0 then Exit;
      chunks[coff]:= TSparceLongintArrayChunk3.Create;
    end;
    prevccount:= chunks[coff].Count;
    newchunk:= chunks[coff].Add(dword(index) and dword($000003ff), L); { The
      chunk may create a bigger one and suicide. This check here is to be more
      cache friendly, to not write to the array unnecessarily. }
    if newchunk <> chunks[coff] then chunks[coff]:= newchunk;
    if Assigned(newchunk)
      then f_count += newchunk.Count - prevccount
      else f_count -= prevccount;
    if f_count = 0 then begin
      SetLength(chunks, 0);
      chunks_offset:= 0;
    end;
  end;

  function TSparceLongintArray.Get(index: Longint): Longint;
  var
    coff: integer;
  begin
    Assert(index > 0, 'TSparceLongintArray.Get() : index must be positive');
    coff:= integer(dword(index) shr 10) - chunks_offset;
    if (coff < 0) or (coff > High(chunks)) or not Assigned(chunks[coff])
      then Result:= 0
      else Result:= chunks[coff].Get(dword(index) and dword($000003ff));
  end;


  procedure TSparceLongintArray.RegisterFields;
  begin
    RegClass(TSparceLongintArrayAbstractChunk);
    RegClass(TSparceLongintArrayChunk3);
    RegClass(TSparceLongintArrayChunk15);
    RegClass(TSparceLongintArrayChunk255);
    RegClass(TNotSoSparceLongintArrayChunk);
    RegType(typeinfo(TSparceLongintArrayChunksArray));//, typeinfo(TSparceLongintArrayAbstractChunk));
    inherited;
    ListFields([
      'chunks', @chunks, typeinfo(TSparceLongintArrayChunksArray),
      'chunks_offset', @chunks_offset, typeinfo(dword),
      'f_count', @f_count, typeinfo(Longint)
    ]);
  end;

  procedure TSparceLongintArray.BeforeSaving;
  begin
    ShrinkArray();
  end;

