{
    This file is part of chepersy
    Copyright (c) 2004-2008 by Anton Rzheshevski (chebmaster@mail.ru),

    See the file COPYING.CPS, included in this distribution,
    for details about the copyright.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

 **********************************************************************}


function TChepersyObject.QueryInterface(
  {$IFDEF FPC_HAS_CONSTREF}constref{$ELSE}const{$ENDIF} iid : tguid;out obj) : longint;{$IFNDEF WINDOWS}cdecl{$ELSE}stdcall{$ENDIF};

  begin
     if getinterface(iid,obj) then
       result:= S_OK
     else
       result:=longint(E_NOINTERFACE);
  end;

function TChepersyObject._AddRef : longint;{$IFNDEF WINDOWS}cdecl{$ELSE}stdcall{$ENDIF};
  begin
    if ((CpsMask and CPS_NO_INTERLOCK_MASK) > 0) and not ForceInterlock
    then begin
      Inc(f_RefCount);
      Result:= f_RefCount;
    end
    else
      Result:= interlockedincrement(f_RefCount)
  end;

function TChepersyObject._Release : longint;{$IFNDEF WINDOWS}cdecl{$ELSE}stdcall{$ENDIF};
  begin
    if ((CpsMask and CPS_NO_INTERLOCK_MASK) > 0) and not ForceInterlock
    then begin
      Dec(f_RefCount); Result:= f_RefCount;
    end
    else
      Result:= interlockeddecrement(f_RefCount);

    if Result = 0 then Self.OnZeroReferenceCount;
  end;

procedure TChepersyObject.OnZeroReferenceCount;
begin
  Self.Destroy;
end;

procedure TChepersyObject.AfterConstruction;
  begin
     { we need to fix the refcount we forced in newinstance }
     { further, it must be done in a thread safe way        }
     //declocked(frefcount);
    Dec(f_RefCount);
  end;

procedure TChepersyObject.BeforeDestruction;
  begin
     if f_RefCount <> 0 then Die(RuEn(
      '%0: ненулевой счётчик ссылок (%1) перед вызовом деструктора',
      '%0: non-zero reference count (%1) before destructor call'),
      [AnsiString(Self.ClassName), f_RefCount]);
  end;

class function TChepersyObject.NewInstance: TObject;
  begin
     Result:= inherited NewInstance;
     if Result <> nil then
       TChepersyObject(Result).f_RefCount:= 1;
  end;


{$ifdef fpc}
{ class function TObject.NewInstance : tobject;
  var
    p : pointer;
  begin
    getmem(p, InstanceSize);
    if p <> nil then
      InitInstance(p);
    NewInstance:=TObject(p);
  end;
 }
{  class function TChepersyObject.NewInstance : tobject;
  var
    p : pointer;
  begin
    getmem(p, InstanceSize);//ToDo: replace with custom memory manager.
   // if p <> nil then
    InitInstance(p); //ToDo: replace with custom procedure.
    NewInstance:=TObject(p);
  end; }
{$endif}



  procedure TChepersyObject.RegisterFields;
  begin
    ListFields([
      'CpsMask', @f_cpsmask, typeinfo(dword),
      '{Skipped,NeverClone}f_RefCount', @f_refCount, typeinfo(dword)//typeinfo(longint)
      //Can't use longint as it is not registered yet at this time
      //  while I have NO desire of touching the hard-coded type numbers used by
      //  the backward/forward compatibility mechanism to resolve unknown types

      ,'{Skipped,NeverClone}__CpsSomething', pointer(@f_refCount) + sizeof(f_refCount), typeinfo(dword)
    ]);
  end;



  procedure TChepersyObject.AfterLoading;
  begin
    //do nothing. This is NOT called if the method is not redefined (method address check)
  end;
  
  procedure TChepersyObject.BeforeSaving;
  begin
    //do nothing. This is NOT called if the method is not redefined (method address check)
  end;


(*
  //Intercept and pass the instance to the garbage collector
  procedure TChepersyObject.FreeInstance;
  begin
    if not Scraped then Scrape;
    // do NOT call inherited
  end;

   //the replacement
  procedure TChepersyObject.CpsFreeInstance;
  begin
    // a clone of FreeInstance from objpas.inc as of FPC 2.6.4
    // may break in future language versions

    CleanupInstance;
    FreeMem(Pointer(Self));
  end;
*)

  function TChepersyObject.Clone(): TChepersyObject;
  begin
    Result:= TChepersyObject(CChepersyObject(Self.ClassType).Generate);
    CloneRecord(pointer(Self), pointer(Result), Self.ClassIndex, []);
  end;

  function TChepersyObject.CloneAnother(another: TChepersyObject): boolean;
  begin
    if not Assigned(another) or not (another is Self.ClassType)
    then Result:= false
    else begin
      CloneRecord(pointer(another), pointer(Self), Self.ClassIndex, []);
      Result:= true;
    end;
  end;

  constructor TChepersyObject.Register;
  {
     To calculate the field offsets we need an instance first!
     Thus, this constructor: it creates a temporary instance
       and registers the fields. Called from RegClass() procedure.
  }
  begin
    Generate;
    RegBegin(Self);
    RegisterFields;
    RegEnd;
  end;

  //constructor TChepersyObject.Generate;
  class function TChepersyObject.Generate: pointer; //TChepersyObject;
  {
     We neeed to avoid the usual initialization routines
     when we load the instance from a file.
     So here is a special constructor for this. It does nothing,
     not even calls AfterConstruction.
  }
  begin
    GetMem(Result, InstanceSize());
    Result:= TChepersyObject(InitInstance(Result)); //a) clears memory
                                            //b) fills the Self pointer
                                            //c) initializes interfaces if any
  end;

//obsoleted in 0.9. It was a horrible, buggy idea to make it a destructor
//  destructor TChepersyObject.TechnicalDestroy;
//  begin
    //ReallyFreeInstance;
//  end;
  
  {stubs for the "class fields" implemented via fake virtual methods}
  {$ifdef oldhacky}
    procedure TChepersyObject._MyClassIndex(); begin end;
      //not empty and different: to prevent the compiler from using the default
      //  "abstract field" handler.
    procedure TChepersyObject._MyScenarioIndex(); begin end;
    procedure TChepersyObject._MyFieldsList(); begin end;
  {$endif}

  {$ifndef fpc}
    procedure RemoveMemoryWriteProtection(addr: pointer);
    var
      MBI: MEMORY_BASIC_INFORMATION;
      stub: dword;
    begin
      if not IsBadWritePtr(addr, sizeof(pointer) * 5) then Exit;
      // that's an WinAPI function, but we don't need to worry
      // because there is no such thing as Delphi for Linux.
      
      if Mother^.Debug.Verbose then AddLog(
        'Class VMT is write-protected. Removing the protection at %0...',[addr]);
      if (VirtualQuery(addr, MBI, SizeOf(MBI)) <> SizeOf(MBI))
      or not VirtualProtect(MBI.BaseAddress, MBI.RegionSize, PAGE_EXECUTE_READWRITE, @stub)
        then Die(PervertedFormat(RuEn(
          'Не удалось получить права на запись в память по адресу %0.',
          'Failed to gain write access rights for memory at address %0.'), [addr]));
      if Mother^.Debug.Verbose then AddLog('Success. Gained write access for %1 bytes at %0.',
        [MBI.AllocationBase, MBI.RegionSize]);
    end;
  {$endif}

 {$ifdef delphiworkaround}
  procedure DumbDelphiSetMyFieldsList(C: TClass; F: TObject);
  begin
    aFieldsList[CHash(C)]:=TFieldsList(F);
  end;
 {$else}
  class procedure TChepersyObject._SetMyFieldsList(F: TObject);//TFieldsList);
  begin
    {$ifdef oldhacky}
    TFieldsList(pointer(ptruint(Self) + vmtMyFieldsList)^):= TFieldsList(F);
    {$else}
    aFieldsList[CHash(ClassType())]:= TFieldsList(F);
    {$endif}
  end;
 {$endif}

{$ifdef oldhacky}
  class function TChepersyObject.FieldsList(): TObject;//TFieldsList;
  begin
    Result:=TFieldsList(pointer(Cardinal(Self)+vmtMyFieldsList)^)
  end;

  class function TChepersyObject.ClassIndex(): integer;
  begin
    Result:=integer(pointer(Cardinal(Self)+vmtMyClassIndex)^);
  // {$ifdef safeloading}
    if Result = InvalidClassIndex then Result:= 0;
  // {$endif}
  end;
{$else}
  {$ifdef fpc}
  class function TChepersyObject.ClassIndex(): integer;
  begin
    if not CFitsIndexArrays(ClassType()) then begin Result:=0; Exit end;
    Result:= aClassInd[CHash(ClassType())];
  end;
  {$endif}
{$endif}

  function TChepersyObject.ClassNameMd5(): TMd5Digest;
  var i: integer;
  begin
    i:= Self.ClassIndex();
    Result:= Types[Records[i]].TypeInfoMd5;
  end;

  function GetClassIndex(C: TClass): integer;
  begin
    if not CFitsIndexArrays(C) then begin
      Result:=0;
//writeln('--------- Notfits, ',C.ClassName);
      Exit
    end;
    Result:=aClassInd[CHash(C)];
  end;

  procedure SetClassIndex(C: TClass; index: integer);
  begin
    MakeSureClassFitsIndexArrays(C);
    aClassInd[CHash(C)]:=index;
  end;
  
{$ifndef delphiworkaround}
  class procedure TChepersyObject._RegisterMyIndex(index: integer);
  var p: pointer;
  begin
    {$ifdef oldhacky}
      p:=pointer(Cardinal(Self)+vmtMyClassIndex);
      {$ifndef fpc}
       RemoveMemoryWriteProtection(p);
      {$endif}
      integer(p^):= index;
    {$else}
      MakeSureClassFitsIndexArrays(ClassType());
      aClassInd[CHash(ClassType())]:=index;
    {$endif}
  end;
 {$endif}

 {$ifdef delphiworkaround}
  procedure DumbDelphiSetScenarioIndex(C:Tclass; index: integer);
  begin
    tScenarioIndex[GetClassIndex(C)]:= index;
  end;
 {$else}
{  class procedure TChepersyObject._SetScenarioIndex(index: integer);
  begin
    tScenarioIndex[
      {$ifdef oldhacky}
        longint(pointer(Cardinal(Self) + vmtMyClassIndex)^)
      {$else}
        aClassInd[CHash(ClassType())]
      {$endif}
                   ]:= index;
  end;
}
 {$endif}

