{
    This file is part of chepersy
    Copyright (c) 2004-2015 by Anton Rzheshevski (chebmaster@mail.ru),

    See the file COPYING.CPS, included in this distribution,
    for details about the copyright.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

 **********************************************************************}

(*

//Procedure template:

procedure RP_ (PField: pointer; OP: TFieldOperation;  Tind: integer;
  FieldProperties: TFieldPropertiesSet);  RP_CallingConvention;
begin
  case op of
    fio_Load:
    fio_Save:
    fio_Clone: // is called *after* the field was copied binarily, on the copy, to deal with managed types
    fio_Skip: // is used at loading, for fields that are not present in the current version
    fio_Walk:
  end;
end;

//Since v0.9, when loading tind receives the local type index, not the global one
*)

procedure RP_Pointer (PField: pointer; OP: TFieldOperation; Tind: integer;
  FieldProperties: TFieldPropertiesSet);  RP_CallingConvention;
begin
  //Pointer in Chentrah isn't a saveable type...
  case op of
    fio_Save: Die(MI_ERROR_PROGRAMMER_NO_BAKA, [
      RuEn('Попытка сохранить поле типа указатель - упущено при парсинге.',
           'Attempt to save a pointer type field - slipped the parsing.')]);
    fio_Skip, fio_Walk, fio_Clone:; //do nothing

    // The Mother^.Timer.Second role, as the Type #0, is
    //  to notify about the unloadable fields.
    fio_Load: Die(RuEn(
      'Встречено поле неизвестного неразруливаемого типа.',
      'Found a field of an unknown unresolvable type.'));
  end;
end;

procedure RP_DieInAgony (PField: pointer; OP: TFieldOperation; Tind: integer;
  FieldProperties: TFieldPropertiesSet);  RP_CallingConvention;
begin
  //  To die on the unknown unresolvable types
  case op of
    fio_Load: Die(RuEn(
      'Встречено поле неизвестного неразруливаемого типа.',
      'Found a field of an unknown unresolvable type.'));
    fio_Save, fio_Skip, fio_Walk, fio_Clone: //do nothing;
  end;
end;


procedure RP_Dyna (PField: pointer; OP: TFieldOperation; Tind: integer;
  FieldProperties: TFieldPropertiesSet);  RP_CallingConvention;
var
  o: TDyna;
begin
  case op of
    fio_Load: TDyna(PField^):= ReadDyna();
    fio_Save: WriteDyna(TDyna(PField^), FieldProperties);
    fio_Clone: Die(MI_ERROR_PROGRAMMER_NO_BAKA, [RuEn(
      'Клонирование полей устаревшего типа TDyna не поддерживается.',
      'Cloning fields of the obsolete TDyna type is not supported.')]);
    fio_Skip: begin
                o:=ReadDyna(); //load it and destroy it. The fate is cruel.
                o.Free;
              end;
    fio_Walk: WalkDyna(TDyna(PField^), FieldProperties);
  end;
end;

(*
procedure RP_CustomClass (PField: pointer; OP: TFieldOperation; Tind: integer); register;
var
  p: TCustomClassProc;
  o: TObject;
  i: longint;
  procedure ValidateIt(o: TObject);
  begin
    if ReadInt() <> 777 then Die(RuEn(
      'Пользовательская процедура класса %0'#10#13'  прочитала из потока неверное количество данных.',
      'The custom processing procedure of class %0'#10#13'  read a wrong amount of data from the stream.'),
      [o.ClassName]);
  end;
  procedure SafeCallCCProc(var oo: TObject; ptind: integer);
  begin
    try
      p:= Types[ptind].CustomClassProc;
//writeln('ooooo  ', IntToHex(cardinal(pointer(@p)),8));
      p(oo, OP);
    except
      Die(RuEn(
        'Крах пользовательской процедуры класса %0',
        'Error in custom processing procedure of class %0.'),
        [NameSpace[Types[ptind].Name]]);
    end;
  end;
  procedure CheckTindValid(cv: longint);
  begin
    if B_ToGtind[cv] <=0 then
      Die(RuEn(
        'Неизвестный пользовательский класс %0',
        'Unknown custom class %0'),
        [B_NameSpace[CNames[tind]]]);
  end;
begin
  case op of
    fio_Load: begin
      i:= ReadInt();
//writeln(',,,,,, ', i);//, '  ', FieldKindToStr(Types[Records[i]].Kind), '  ', ClassNameSpace[i]);
      if i=0
        then pointer(pfield^):= nil
        else begin
          if i > 0 then begin
            i:= B_RecBTinds[i];
            CheckTindValid(i);
            SafeCallCCProc(TObject(PField^), B_ToGtind[i]);
            ValidateIt(TObject(PField^));
          end
          else ReadPersistent(TChepersyObject(PField^));
        end;
    end;
    fio_Skip: begin
      i:= ReadInt();
      if i > 0 then begin
        i:= B_RecBTinds[i];
        CheckTindValid(i);
        SafeCallCCProc(o, B_ToGtind[i]);
        MemoryLeakSuspected:= true;
        SuspectsAdd(o);
        ValidateIt(o);
      end
      else
        if i < 0 then SkipPersistent;
    end;
    fio_Save: begin
      if not Assigned(pointer(PField^)) then WriteInt(0)
      else begin
        i:= GetClassIndex(TObject(PField^).ClassType);
        if i = 0 then Die(RuEn(
          'Класс %0 не зарегистрирован!',
          'Class %0 is not registered!'),
          [string(TObject(PField^).ClassName)]);
//writeln(',,,,,, ', i, '  ', FieldKindToStr(Types[Records[i]].Kind), '  ', ClassNameSpace[i]);
        if Types[Records[i]].Kind <> fk_class then begin
          WriteInt(i);
          SafeCallCCProc(TObject(PField^), Records[i]);
          WriteInt(777);
        end
        else begin
          WriteInt(-1);
          WritePersistent(TChepersyObject(PField^));
        end;
      end;
    end;
    fio_Walk: begin
      if Assigned(pointer(PField^)) then begin
        i:= GetClassIndex(TObject(PField^).ClassType);
        if i = 0 then Die(RuEn(
          'Класс %0 не зарегистрирован! (ожидался %1)!',
          'Class %0 is not registered! (expected %1)'),
          [string(TObject(PField^).ClassName), TypeNameSpace[tind]]);
        if Types[Records[i]].Kind = fk_class
          then WalkPersistent(TChepersyObject(PField^))
          else SafeCallCCProc(TObject(PField^), Records[i]);
      end;
    end;
  else
  end;
end;

  procedure CpsProcessObject(var o: TObject; op: TFieldOperation);
  begin
    RP_CustomClass(@o, op, 0);
  end;

*)

procedure RP_Metaclass (PField: pointer; OP: TFieldOperation; Tind: integer;
  FieldProperties: TFieldPropertiesSet);  RP_CallingConvention;
begin
  Assert(not (fpro_CloneByIncreasingRefCount in FieldProperties), 'This type does not support cloning-by-increasing-RefCount');
  case op of
    fio_Load: begin
      dword(PField^):= ReadInt();
      if Assigned(pointer(PField^)) then begin
         {$ifdef safeloading}
          if dword(PField^) >= Length(DH.Cnums) then DieInvalidContainer(RuEn('Данные запороты: индекс класса - за пределами!','Data corrupt: class index out of bounds!'));
         {$endif}
         CChepersyObject(PField^):= DH.B_Classes[dword(PField^)];
      end;
    end;
    fio_Clone:; //Do nothing. Any metaclass references are kept as is.
    fio_Save:
      if not Assigned(pointer(pfield^)) then WriteInt(0)
      else WriteInt(
          {$ifdef delphiworkaround}
            GetClassIndex(CTrulyPersistent(PField^))
          {$else}
            CChepersyObject(PField^).ClassIndex
          {$endif} );
    fio_Skip: ReadInt; //SkipDword;
    fio_Walk:;
  end;
end;

var
  rpp_sc: ptruint;
  rpp_count: integer = 0;
  rpp_crashed: boolean = false;
  rpp_i: integer;
procedure RP_Persistent (PField: pointer; OP: TFieldOperation; Tind: integer;
  FieldProperties: TFieldPropertiesSet);  RP_CallingConvention;
begin
{
  No use. It just doesn't fucking work :(
  I mean, it works, but then a crash happens anyway. Even though the recursion was rolled back and the stack freed
  {$ifdef safeloading}
    inc(rpp_count);
    //check stack
    {$ifdef cpu64}
      Asm
        pushf
        push rax
        mov rax, rsp
        sub rax, [CpsStackLimit]
        mov [rpp_sc], rax
        pop rax
        popf
      End;
    {$else}
      Asm
        pushf
        push eax
        mov eax, esp
        sub eax, [CpsStackLimit]
        mov [rpp_sc], eax
        pop eax
        popf
      End;
    {$endif}
//if (op = fio_Save) and (rpp_sc < 10000000) then addlog('- %0',);
   if rpp_sc < CpsCustomStackCheckPanicThreshold then begin
//addlog('*dying at %0 | %1', [rpp_sc, rpp_count]);
     rpp_crashed:= true;
     rpp_i:= rpp_count;
     //because, by some reason, invoking Die() at such depths causes a real stack overflow in the error somewhere in the processing routines >_<
     //using a postponed crash.
     if op = fio_Load then pointer(PField^):= nil;
     dec(rpp_count);
     Exit;
   end;
  {$endif}
 }
  {$ifdef bench_read_persistent}{$include un_bottleneck_end}{$endif} //Cuz it is only called *inside* an outer block in XXXPersistent
  case op of
    fio_Load:
      ReadPersistent(pointer(PField^), FieldProperties);
    fio_Save:
      WritePersistent(pointer(PField^), FieldProperties);
    fio_Clone:
      ClonePersistent(pointer(PField^), FieldProperties);
    fio_Skip:
      SkipPersistent;
    fio_Walk:
      WalkPersistent(pointer(PField^), FieldProperties);
  end;
  {$ifdef bench_read_persistent}{$include un_bottleneck_begin}{$endif}
{
  {$ifdef safeloading}
    dec(rpp_count);
    if rpp_crashed and (rpp_count < 2) then begin
      rpp_crashed:= false;
      rpp_count:= 0;
      Die(RuEn(
        'Переполнение стека вследствие тяжёлой рекурсии (%0 вхождений)',
        'Stack overflow due to heavy recursion (%0 loops)'),
      [rpp_i]);
    end;
  {$endif}
}
end;


procedure RP_RawByteString (PField: pointer; OP: TFieldOperation; Tind: integer;
  FieldProperties: TFieldPropertiesSet);  RP_CallingConvention;
begin
  case op of
    fio_Load: RawByteString(PField^):= ReadRawByteString();
    fio_Save: WriteRawByteString(RawByteString(PField^));
    fio_Skip: SkipRawByteString;
    fio_Clone: IncRawByteStringRefCount(pointer(PField^));
  end;
end;

procedure RP_Enum (PField: pointer; OP: TFieldOperation; Tind: integer;
  FieldProperties: TFieldPropertiesSet);  RP_CallingConvention;
begin
  case op of
    fio_Load: if DH.TypeChanged[DH.B_ToGtind[tind]]
                then LoadEnum(DH.B_ToGtind[Tind], PField)
                else dword(PField^):= ReadInt();
    fio_Save: WriteInt(dword(PField^));
    fio_Skip: ReadInt; //SkipDword;
    fio_Clone: Die('Internal error in Chepersy!');
  end;
end;

procedure RP_Record (PField: pointer; OP: TFieldOperation; Tind: integer;
  FieldProperties: TFieldPropertiesSet);  RP_CallingConvention;
begin
  //field properties are joined with the child record's field properties
  case op of
    fio_Load, fio_Skip: LoadRecord(PField, OP, DH.B_Lrecind[tind], FieldProperties);
    fio_Save: SaveRecord(PField, Types[Tind].RecordInd, FieldProperties);
    fio_Walk: WalkRecord(PField, Types[Tind].RecordInd, FieldProperties);
    fio_Clone: CloneRecord(nil, PField, Types[Tind].RecordInd, FieldProperties);//Die('Internal error in Chepersy!');
  end;
end;

procedure RP_GenericDynArray (PField: pointer; OP: TFieldOperation; Tind: integer;
  FieldProperties: TFieldPropertiesSet);  RP_CallingConvention;
//processes *any* types that have a processing procedure.
// exploits the fact that the dynamic arrays of any type
// have a similar layout in memory:
// reference count at offset -8 (we always set it to 1) and
// High value at -4 (length = High + 1)
var
  i, s, m, t: cardinal;
  p: TCustomTypeProcessingProc;
  psource: pointer;
begin
  {$ifdef safeloading}
 try
  {$endif}
  case op of
    fio_Load, fio_Skip: begin
      t:=DH.B_TBaseInds[Tind];
      s:=Types[DH.B_ToGtind[t]].Size;
      p:=Types[DH.B_ToGtind[t]].Proc;
    end
  else
    t:=Types[Tind].ElementTypeInd;
    s:=Types[t].Size;
    p:=Types[t].Proc;
  end;
  case op of
    fio_Save: begin
          if Assigned(pointer(PField^)) then begin
            m:= GetDynArrayLength(PField);
              //longint(pointer(cardinal(PField^) - 4)^) {$ifdef fpc} + 1{$endif}
              //the value at -4 contains Length() in Delphi
              //and High() (i.e. Length() -1 ) in FreePascal
            WriteInt(m);
//***TODO:CPS_OPTIMIZE Write binary blocks as binary
            for i:=0 to m - 1 do
              p(pointer(ptruint(PField^) + i * s), fio_Save, t, FieldProperties);
          end
          else WriteInt(0);
        end;
    fio_Load: begin
          m:= ReadInt;
//addlog(' -- %2 %1 %0',[ DH.B_NameSpace[DH.TNames[t]], DH.B_Namespace[DH.tnames[tind]], m]);
          if m > 0 then begin
            NewDynArray(PField, m, DH.B_ToGtind[t]);
            for i:=0 to m - 1 do
              p(pointer(ptruint(PField^) + i * s), fio_Load, t, FieldProperties);
          end;
        end;
    fio_Clone:
        //The cloning mechanism first copies all fields as binary,
        //  then runs the converter proc over the select fields that require it.
        //So, at this point PField points at the clone, not the original
        if Assigned(pointer(PField^)) then begin
          if fpro_CloneByIncreasingRefCount in FieldProperties then begin
            IncDynArrayRefCount(pfield);
          end
          else begin
            m:= GetDynArrayLength(PField);//longint(pointer(ptruint(PField^) - 4)^) + 1;
            psource:= pointer(PField^);
            NewDynArray(PField, m, t);
            MOVE(psource^, pointer(PField^)^, m * s);
          end;
          if Types[t].containsComplexTypes
          then
            for i:= 0 to m - 1 do
              p(pointer(ptruint(PField^) + i * s), fio_Clone, t, FieldProperties);

        end;
    fio_Skip: begin
          m:=ReadInt;
          if m > 0 then
            for i:=1 to m do
              p(nil, fio_Skip, t, FieldProperties);
        end;
    fio_Walk:
        if Assigned(pointer(PField^)) then
          for i:=0 to GetDynArrayLength(PField) - 1//longint(pointer(cardinal(PField^) - 4)^) {$ifndef fpc}- 1{$endif}
            do p(pointer(cardinal(PField^) + i*s), fio_Walk, t, FieldProperties);
  end;
  
  {$ifdef safeloading}
 except
    case op of
      fio_Skip, fio_Load:
        Die(RuEn(
            'При вызове RP_GenericDynArray(%0, %1, %2, [%3]).',
            'During a call to RP_GenericDynArray(%0, %1, %2, [%3]).')
        , [PField, GetEnumName(TypeInfo(TFieldOperation), ord(OP)), DH.B_Namespace[DH.tnames[tind]]
          , SetToString(dword(FieldProperties), typeinfo(TFieldPropertiesSet))]);
    else
        Die(RuEn(
            'При вызове RP_GenericDynArray(%0, %1, %2, [%3]).',
            'During a call to RP_GenericDynArray(%0, %1, %2, [%3]).')
        , [PField, GetEnumName(TypeInfo(TFieldOperation), ord(OP)), Namespace[types[tind].name]
          , SetToString(dword(FieldProperties), typeinfo(TFieldPropertiesSet))]);
    end;
 end;
  {$endif}
end;

procedure RP_BinaryStaticArray (PField: pointer; OP: TFieldOperation; Tind: integer;
  FieldProperties: TFieldPropertiesSet);  RP_CallingConvention;
begin
  case op of
    fio_Load: ReadUnaligned(PField, Types[DH.B_ToGtind[Tind]].Size);
    fio_Save: WriteToBuffer(PField^, Types[Tind].Size);
    fio_Skip: SkipUnaligned(Types[DH.B_ToGtind[Tind]].Size);
    fio_Clone: Die('Internal error in Chepersy!');
  end;
end;

procedure RP_GenericStaticArray (PField: pointer; OP: TFieldOperation; Tind: integer;
  FieldProperties: TFieldPropertiesSet);  RP_CallingConvention;
var
  i, s, t: cardinal;
  p: TCustomTypeProcessingProc;
begin
  case op of
    fio_Load, fio_Skip: begin
      t:=DH.B_TBaseInds[Tind];
      s:=Types[DH.B_ToGtind[t]].Size;
      p:=Types[DH.B_ToGtind[t]].Proc;
    end
  else
    t:=Types[Tind].ElementTypeInd;
    s:=Types[t].Size;
    p:=Types[t].Proc;
  end;
  case op of
    fio_Save, fio_Walk: begin
            for i:=0 to (Types[Tind].Size div s) - 1 do
              p(pointer(ptruint(PField) + i*s), OP, t, FieldProperties);
        end;
    fio_Load: begin
//addlog('rp_genericstatic %0  %1  %2  %3', [TSizes[Tind], s, (TSizes[Tind] div s) - 1, B_namespace[tnames[tind]]]);
            for i:=0 to (DH.TSizes[Tind] div s) - 1 do
              p(pointer(ptruint(PField) + i * s), fio_Load, t, FieldProperties);
        end;
    fio_Skip: begin
            for i:=1 to DH.TSizes[Tind] div s do
              p(nil, fio_Skip, t, FieldProperties);
        end;
    fio_Clone:
        //The cloning mechanism first copies all fields as binary,
        //  then runs the converter proc over the select fields that require it.
        //So, at this point PField points at an already copied array of the clone.
        if Types[t].containsComplexTypes or Types[t].containsChepersyObjects
        then
          for i:=0 to (Types[Tind].Size div s) - 1 do
            p(pointer(ptruint(PField) + i * s), fio_Clone, t, FieldProperties);
  end;
end;


procedure RP_Extended (PField: pointer; OP: TFieldOperation; Tind: integer;
  FieldProperties: TFieldPropertiesSet);  RP_CallingConvention;
begin
  case op of
    fio_Load: ReadUnaligned(PField, 10);
    fio_Save: WriteToBuffer(PField^, 10);
    fio_Skip: SkipUnaligned(10);
    fio_Clone: Die('Internal error in Chepersy!');
  end;
end;

procedure RP_Byte (PField: pointer; OP: TFieldOperation; Tind: integer;
  FieldProperties: TFieldPropertiesSet);  RP_CallingConvention;
begin
  case op of
    fio_Load: ReadUnaligned(PField, 1);
    fio_Save: WriteToBuffer(PField^, 1);
    fio_Skip: SkipUnaligned(1);
    fio_Clone: Die('Internal error in Chepersy!');
  end;
end;

procedure RP_ShortInt (PField: pointer; OP: TFieldOperation; Tind: integer;
  FieldProperties: TFieldPropertiesSet);  RP_CallingConvention;
begin
  case op of
    fio_Load: ReadUnaligned(PField, 1);
    fio_Save: WriteToBuffer(PField^, 1);
    fio_Skip: SkipUnaligned(1);
    fio_Clone: Die('Internal error in Chepersy!');
  end;
end;

procedure RP_Word (PField: pointer; OP: TFieldOperation; Tind: integer;
  FieldProperties: TFieldPropertiesSet);  RP_CallingConvention;
begin
  case op of
    fio_Load: ReadUnaligned(PField, 2);
    fio_Save: WriteToBuffer(PField^, 2);
    fio_Skip: SkipUnaligned(2);
    fio_Clone: Die('Internal error in Chepersy!');
  end;
end;

procedure RP_SmallInt (PField: pointer; OP: TFieldOperation; Tind: integer;
  FieldProperties: TFieldPropertiesSet);  RP_CallingConvention;
begin
  case op of
    fio_Load: ReadUnaligned(PField, 2);
    fio_Save: WriteToBuffer(PField^, 2);
    fio_Skip: SkipUnaligned(2);
    fio_Clone: Die('Internal error in Chepersy!');
  end;
end;

type
  TCBS = set of byte;


procedure LoadSet(Tind: integer; PField: pointer);
  var
    cba, cbb: TCBS;
  procedure Convert(Tind_b: integer);
  var
    b: integer;
  begin
    FillChar(cbb, sizeof(cbb), 0);
    for b:=0 to DH.EnumConvTable[Tind_b].High do
      if b in cba then begin
        cbb:=cbb + [DH.EnumConvTable[Tind_b][b]];
  //addlog('-- %0->%1',[b, EnumConvTable[Tind][b]]);
      end;
  end;
begin
  FillChar(cba, sizeof(cba), 0);
  ReadUnaligned(@cba, ReadInt());
  Convert(Types[Tind].ElementTypeInd);
  MOVE(cbb, PField^, Types[Tind].size);
end;

var rpls_i: integer;
procedure RP_Set (PField: pointer; OP: TFieldOperation; Tind: integer;
  FieldProperties: TFieldPropertiesSet);  RP_CallingConvention;
begin
  case op of
    fio_Load:
       if DH.TypeChanged[DH.B_ToGtind[tind]]
         then LoadSet(DH.B_ToGtind[Tind], PField)
         else ReadUnaligned(PField, ReadInt());
    fio_Save: begin
         rpls_i:=1 + (Types[Tind].SetLen - 1) div 8;
//addlog('type = "%0", rpls = %1', [Namespace[Types[Tind].name] ,rpls_i]);
         WriteInt(rpls_i);
         WriteToBuffer(PField^, rpls_i);
       end;
    fio_Skip:
       SkipUnaligned(ReadInt());
    fio_Clone:
       Die('Internal error in Chepersy!');
  end;
end;

procedure RP_UnknownSet (PField: pointer; OP: TFieldOperation; Tind: integer;
  FieldProperties: TFieldPropertiesSet);  RP_CallingConvention;
begin
  case op of
    fio_Skip:
       SkipUnaligned(ReadInt());
    fio_Load, fio_Save: Die(MI_ERROR_PROGRAMMER_NO_BAKA, [
      RuEn('Попытка сохранить поле типа chepersy_unknownset.',
           'Attempt to save a field of chepersy_unknownset.')]);
    fio_Clone:
      Die('Internal error in Chepersy!');
  end;
end;


procedure RP_WideString (PField: pointer; OP: TFieldOperation; Tind: integer;
  FieldProperties: TFieldPropertiesSet);  RP_CallingConvention;
begin
   case op of
    fio_Load:
//begin
    WideString(PField^):=ReadWideString();
//addlog(WideString(PField^));
//end;
    fio_Save: WriteWideString(WideString(PField^));
    fio_Skip: SkipWideString;
    fio_Clone: IncWideStringRefCount(pointer(PField^));
  end;
end;

procedure RP_UnicodeString (PField: pointer; OP: TFieldOperation; Tind: integer;
  FieldProperties: TFieldPropertiesSet);  RP_CallingConvention;
begin
   case op of
    fio_Load:
//begin
    UnicodeString(PField^):=ReadUnicodeString();
//addlog(WideString(PField^));
//end;
    fio_Save: WriteUnicodeString(UnicodeString(PField^));
    fio_Skip: SkipUnicodeString;
    fio_Clone: IncUnicodeStringRefCount(pointer(PField^));
  end;
end;

procedure RP_ArrayOfDWORD (PField: pointer; OP: TFieldOperation; Tind: integer;
  FieldProperties: TFieldPropertiesSet);  RP_CallingConvention;
// since it works with binary data, it should work
//   with *any* dynamic array whose base size is equal to 32 bits:
//   cardinal, integer, single, glUint, gl float, et cetera.
var
  s: integer;
  psource: pointer;
begin
  case op of
    fio_Load: begin
        s:=ReadInt();
        pointer(pfield^):=nil;
        SetLength(Tarrayofdword(PField^), s);
        if s > 0 then
     //need to check, since an empty dynamic array is in fact a nil pointer.
          ReadFromBuffer(pointer(PField^)^, s * 4);
      end;
    fio_Save: begin
        s:=Length(Tarrayofdword(PField^));
        WriteInt(s);
        if s > 0 then WriteToBuffer(pointer(PField^)^, s * 4);
      end;
    fio_Skip: begin
        SkipUnaligned(4 * ReadInt());//SkipBin(ReadInt());
      end;
    fio_Walk: Die('Internal error in Chepersy');
    fio_Clone: begin
        //The cloning mechanism first copies all fields as binary,
        //  then runs the converter proc over the select fields that require it.
        //So, at this point PField points at the clone, not the original
        if Assigned(pointer(PField^)) then begin
          if fpro_CloneByIncreasingRefCount in FieldProperties then begin
            IncDynArrayRefCount(PField);
          end
          else begin
            s:= Length(TArrayOfDword(PField^));
            psource:= pointer(PField^);
            pointer (PField^):= nil;
            SetLength(TArrayOfDword(PField^), s);
            MOVE(psource^, pointer(PField^)^, s * SizeOf(dword));
          end;
        end;
    end;
  end;
end;

procedure RP_ArrayOfQWORD (PField: pointer; OP: TFieldOperation; Tind: integer;
  FieldProperties: TFieldPropertiesSet);  RP_CallingConvention;
var
  s: integer;
  psource: pointer;
begin
  case op of
    fio_Load: begin
        s:=ReadInt();
        pointer(pfield^):=nil;
        SetLength(TarrayofQword(PField^), s);
        if s > 0
     //need to check, since an empty dynamic array is in fact a nil pointer.
          then ReadFromBuffer(pointer(PField^)^, s * 8);//ReadBin(pointer(PField^), s*2);
      end;
    fio_Save: begin
        s:=Length(TarrayofQword(PField^));
        WriteInt(s);
        if s > 0 then WriteToBuffer(pointer(PField^)^, s * 8);
      end;
    fio_Skip: begin
        SkipUnaligned(8 * ReadInt());//SkipBin(ReadInt()*2);
      end;
    fio_Walk: Die('Internal error in Chepersy');
    fio_Clone: begin
        //The cloning mechanism first copies all fields as binary,
        //  then runs the converter proc over the select fields that require it.
        //So, at this point PField points at the clone, not the original
        if Assigned(pointer(PField^)) then begin
          if fpro_CloneByIncreasingRefCount in FieldProperties then begin
            IncDynArrayRefCount(PField);
          end
          else begin
            s:= Length(TArrayOfQword(PField^));
            psource:= pointer(PField^);
            pointer (PField^):= nil;
            SetLength(TArrayOfQword(PField^), s);
            MOVE(psource^, pointer(PField^)^, s * SizeOf(qword));
          end;
        end;
    end;
  end;
end;

procedure RP_ArrayOfEnum (PField: pointer; OP: TFieldOperation; Tind: integer;
    FieldProperties: TFieldPropertiesSet);  RP_CallingConvention;
var
  s, t, i: integer;
begin
  case op of
    fio_Load: begin
        s:=ReadInt();
        pointer(pfield^):= nil;
        SetLength(Tarrayofdword(PField^), s);
        if s > 0 then begin
          t:= DH.B_ToGtind[DH.B_TBaseInds[Tind]];//Types[Tind].ElementTypeInd;
     //need to check, since an empty dynamic array is in fact a nil pointer.
          if DH.TypeChanged[t]
            then
              For i:=0 to s - 1 do
                LoadEnum(t, @(Tarrayofdword(PField^)[i]))
            else
              ReadFromBuffer(pointer(PField^)^, s * 4);
              //ReadBin(pointer(PField^), s);
        end;
      end;
    fio_Save: begin
        s:=Length(Tarrayofdword(PField^));
        WriteInt(s);
        if s > 0
          then WriteToBuffer(pointer(PField^)^, s * 4);//WriteBin(pointer(PField^), s);
      end;
    fio_Skip: begin
        SkipUnaligned(4 * ReadInt());//SkipBin(ReadInt());
      end;
    fio_Clone, fio_Walk: Die('Internal error in Chepersy');
  end;
end;

{$ifndef fpc}
//converting longbool from Delphi to FPC format
procedure RP_LONGBOOL (PField: pointer; OP: TFieldOperation; Tind: integer);  RP_CallingConvention;
begin
  case op of
    fio_Load: begin
        if ReadDword() = 0 then longbool(PField^):=false
                           else longbool(PField^):=true;
      end;
    fio_Save: begin
        if longbool(PField^) then WriteDword(1)
                             else WriteDword(0);
      end;
    fio_Skip:
        ReadDword();//ReadInt;
  end;
end;
{$endif}



