{
    This file is part of Chentrah,
    Copyright (C) 2004-2014 Anton Rzheshevski (chebmaster@mail.ru).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see http://www.gnu.org/licenses/

 **********************************************************************}


unit mo_fbo;

{$include mo_globaldefs.h}

interface

uses
  sysutils, Classes, math, md5,  chepersy, mo_hub, typinfo, mo_classes,
    mo_gmathbase, mo_assets, mo_texture;

procedure RegClasses;

{
MrShoor	Удалить	Постоялец	www	«»	12 янв. 2014	7:54	#7
Cheb
На всякий случай, у FBO есть еще одна проблема, в которую пока ты не вступил. Как повторить:
1. Создаем FBO и аттачим в него COLOR и DEPHT24_STENCIL8 текстуру.
2. Рисуем что-то
3. Деаттачим из этого FBO наш COLOR и DEPHT24_STENCIL8
4. Аттачим заново. Наблюдаем в DEPHT24_STENCIL8 мусор на многих ATI карточках.

а ведь DEPHT24_STENCIL8 иногда надо пропихнуть в другой FBO, чтобы он пошел в связке с другими COLOR-ами.
Решение - создавать один FBO, и в нем менять COLOR текстуры не трогая DEPHT24_STENCIL8


glTexImage2D(GL_TEXTURE_2D, 0, GL_DEPTH_COMPONENT, m_TexWidth, m_TexHeight, 0, GL_DEPTH_COMPONENT, GL_FLOAT, NULL)
}

type
  TManagedFbo = class; //forward declaration

  TManagedFboTexture = class(TTexture)
  protected
    f_master: TManagedFbo;
    procedure OnGenResourceFailure(ErrorMessage: UnicodeString); override;
  public
    property Master: TManagedFbo read f_master;
    constructor Create(_master: TManagedFbo;
      _format, _image_format: GLuint;
      p: TAssetPropertiesSet = []
     );
    procedure RegisterFields; override;
  end;

  TFboDepthRenderbuffer = class (TManagedFboTexture)
  protected
    procedure FreeResource; override;
    procedure GenResource; override;
  public
    constructor Create(_master: TManagedFbo; _format: GLuint; props: TAssetPropertiesSet);
    function DefaultFormat: GLuint; override; //GL_DEPTH_COMPONENT24
    function DefaultImageFormat: GLuint; override; //GL_FLOAT
  end;




  //This is not a true FBO but a smart object for use in viewports and such
  //  that supports background resizing
  //It creates and manages the color attachments automatically
  TManagedFbo = class (TTextureLikeAsset)
  protected
    f_texture,  f_depthbuffer: TManagedFboTexture;
    f_replacement: TManagedFbo; //the lazily initizliized replacement.
                              //See the RequestDesiredDimensions method

    desired_width, desired_height, desired_format: GLuint;
    procedure CreateBuffers; virtual;
    procedure Fallback; virtual;
    procedure OnGenResourceFailure(ErrorMessage: UnicodeString); override;
    function _GetActualized: boolean; override;
  public
    property texture: TManagedFboTexture read f_texture;
    property depthbuffer: TManagedFboTexture read f_depthbuffer;
    constructor Create( _format: GLuint; _width: integer = 256; _height: integer = 256; props: TAssetPropertiesSet = []);
    destructor Destroy; override;
    procedure RegisterFields; override;

//    function RequestMasterDimensions(var w, h: integer): boolean; override;
//    procedure MasterYourTextureIsTooBig(who: TGenericAsset; what: UnicodeString); override;

    procedure Bind; virtual;


    procedure FreeResource; override;
    procedure GenResource; override;
    function Actualize: boolean; override;
    function MustBeFreedFromMainThread: boolean; override;

    procedure AfterLoading; override;

    procedure RequestDesiredDimensions(w, h: integer; fmt: GLuint = 0); //Will update lazily during following frames. Or not.

    function TextureFormat: GLuint; virtual;
    function TextureImageFormat: GLuint; virtual; //the strict EXT version fails without setting it :(
    function DepthBufferFormat: GLuint; virtual;

    procedure MasterYourTextureIsTooBig(who: TGenericAsset; what: UnicodeString); virtual;

    procedure Pulse; { Call ONLY in the main thread }

  end;


implementation
  uses mo_module, mo_gl_quirks;

{ TManagedFboTexture }

constructor TManagedFboTexture.Create(_master: TManagedFbo;
  _format, _image_format: GLuint;
  p: TAssetPropertiesSet = []
 );
begin
  f_master:= _master;
  inherited Create(Master.Width, Master.Height, _format, _image_format, 0, [], p);
end;

procedure TManagedFboTexture.RegisterFields;
begin
  inherited;
  ListFields([
    'master', @f_master, CPS_MANAGED_OBJECT {TManagedFbo)
      Chepersy can't resolve circular class references.
      But it does not check class field types either (all class fields
      are treated  as TChepersyObject)
     }
  ]);
end;

procedure TManagedFboTexture.OnGenResourceFailure(ErrorMessage: UnicodeString);
begin
  f_master.MasterYourTextureIsTooBig(Self, ErrorMessage);
end;


{ TFboDepthRenderbuffer }

constructor TFboDepthRenderbuffer.Create(_master: TManagedFbo; _format: GLuint; props: TAssetPropertiesSet = []);
begin
  inherited Create(_master,_format,GL_FLOAT,props);
end;


procedure TFboDepthRenderbuffer.FreeResource;
begin
  glDeleteRenderbuffers(1, @_handle);
  CheckGlError;
  _handle:= 0;
end;

procedure TFboDepthRenderbuffer.GenResource;
var q: qword;
begin
  Assert(Assigned(@glGenRenderbuffers), '@glGenRenderbuffers = NIL');
  Mother^.Timer.UsecDelta(@q);
  try
    glGenRenderbuffers(1, @_handle);
    CheckGlError(true);
    glBindRenderbuffer(GL_RENDERBUFFER_EXT, _handle);
    CheckGlError(true);
    glRenderbufferStorage(GL_RENDERBUFFER_EXT, f_format, f_width, f_height);
    CheckGlError(true);
    if Mother^.Debug.Verbose then AddLog(
      '    Created a depth buffer %0x%1 (format %2) in %3 microseconds',[f_width, f_height, IntToHex(f_format, 4) + 'h', round(Mother^.Timer.UsecDelta(@q))]);
    UpdateHash;
  except
    OnGenResourceFailure(StopDying());
  end;
end;

function TFboDepthRenderbuffer.DefaultFormat: GLuint;
begin
  Result:= GL_DEPTH_COMPONENT32_ARB;//GL_DEPTH_COMPONENT24;
end;

function TFboDepthRenderbuffer.DefaultImageFormat: GLuint;
begin
  Result:= GL_FLOAT;
end;

{ TManagedFbo }

procedure TManagedFbo.AfterLoading;
begin
  inherited;
  if not Actualized then begin  //reset to lowest resolution, just in case. The owner would command it to increase resolution later.
    f_width:= 256;
    f_height:= 256;
    if Assigned(f_texture) then f_texture:= nil; //begin f_texture.Scrape; pointer(f_texture):= nil; end;
    if Assigned(f_depthbuffer) then f_depthbuffer:= nil;//begin f_depthbuffer.Scrape; pointer(f_depthbuffer):= nil; end;
  end;

end;

function TManagedFbo.MustBeFreedFromMainThread: boolean; begin Result:= true end;


procedure TManagedFbo.CreateBuffers;
var p: TAssetPropertiesSet;
begin
  if (desired_format = GL_RGBA8) and (OpenGLQuirkFboCantGL_RGBA8 in Mother^.GAPI.Quirks)
    then
       desired_format:= GL_RGBA16;
  f_texture:= TManagedFboTexture.Create(Self, desired_format, TextureImageFormat, f_props);
  f_depthbuffer:= TFboDepthRenderBuffer.Create(Self, DepthBufferFormat, f_props);
end;

procedure TManagedFbo.MasterYourTextureIsTooBig(who: TGenericAsset; what: UnicodeString);
var
  pes: PMotherSEHState;
begin
  Assert((who = f_texture) or (who = f_depthbuffer), 'TFbo.MasterYourTextureIsTooBig by asset that is not its own.');
  if (f_width = 256) and (f_height = 256) then begin
    pes:= GetExceptionState(); { otherwise, it logs call stack again
                                  in a case of genuine crash }
    pes^.NowDying:= true;
    pes^.CallStackLogged:= true;
    if who = f_texture
      then die(RuEn(
        'Не удалось создать текстуру фреймбуфера:',
        'Failed to create a texture for FBO:')
        + #10#13 + what)
      else Die(RuEn(
        'Не удалось создать буфер глубины для фреймбуфера:',
        'Failed to create a depth buffer for FBO:')
        + #10#13 + what);
  end
  else begin
    if who = f_texture
      then AddLog(RuEn(
          'Не удалось создать текстуру фреймбуфера %0x%1, попытка отката на 256x256'#10#13#10#13'%2',
          'Failed to create a framebuffer texture %0x%1, trying to fall back to 256x256'#10#13#10#13'%2'),
        [f_texture.width, f_texture.height, what])
      else AddLog(RuEn(
          'Не удалось создать текстуру фреймбуфера %0x%1, попытка отката на 256x256'#10#13#10#13'%2',
          'Failed to create a framebuffer texture %0x%1, trying to fall back to 256x256'#10#13#10#13'%2'),
        [f_depthbuffer.width, f_depthbuffer.height, what]);
    Fallback;
  end;
end;

{
function TFbo.RequestMasterDimensions(var w, h: integer): boolean;
begin
  w:= f_width;
  h:= f_height;
  Result:= true;
end;
}

constructor TManagedFbo.Create(_format: GLuint; _width: integer = 256; _height: integer = 256;
    props: TAssetPropertiesSet = []);
begin
  inherited Create(_width, _height, props);
  desired_format:= _format;
  CreateBuffers;
end;

function TManagedFbo.Actualize: boolean;
begin
  if not Assigned(f_texture) or not Assigned(f_depthbuffer)
    then CreateBuffers;

  if not f_texture.Actualized or not f_depthbuffer.Actualized
    then Exit(false); //satisfy dependencies

  Result:= inherited;
end;

procedure TManagedFbo.FreeResource;
begin
  if _handle = 0 then Exit;
  glBindFramebuffer(GL_FRAMEBUFFER_EXT, 0);
  CheckGlError;

  glDeleteFramebuffers(1, @_handle);
  CheckGlError;
  _handle:= 0;
end;

procedure TManagedFbo.Bind;
begin
  glBindTexture(GL_TEXTURE_2D, Mother^.GAPI.StubTexture); //in case the render texture is still bound
  CheckGlError;
  Assert(_handle <> 0, 'TManagedFbo.Bind called when handle is zero');
  glBindFramebuffer(GL_FRAMEBUFFER_EXT, _handle);
  CheckGlError;
end;

procedure TManagedFbo.GenResource;
var
  status: GLenum;
  w, h: integer;
  q: qword;
begin
  if not Assigned(f_texture) then CreateBuffers;
  CheckGlError;
  try
    Mother^.Timer.UsecDelta(@q);

    if Assigned(Module) and Assigned(Module.LoadBalancer) then Module.LoadBalancer.Suspend;
    glGenFramebuffers(1, @_handle);

    glBindFramebuffer(GL_FRAMEBUFFER_EXT, _handle);
    CheckGlError(true);

    //Attach 2D texture to this FBO
   { if not texture.IsResident then Die(RuEn(
      'Не удалось создать объект фреймбуфера, текстура для рендеринга нерезидентна',
      'Failed to create a framebuffer object, the render texture is not resident'));
    }


    glBindTexture(GL_TEXTURE_2D, Mother^.GAPI.StubTexture);
  //  f_texture.Bind;
    glFramebufferTexture2D(
      //GL_DRAW_FRAMEBUFFER doesn't work with ForceWare 93.71 (2008) for Gf FX 5200 (2003)
      GL_FRAMEBUFFER_EXT
      , GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, f_texture.Handle, 0);

    CheckGlError(true);



    glBindRenderbuffer(GL_RENDERBUFFER_EXT, f_depthbuffer.Handle);
    CheckGlError(true);

    glFramebufferRenderbuffer(GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT, GL_RENDERBUFFER_EXT, f_depthbuffer.Handle);
    CheckGlError(true);
  //addlog('TFbo.GenResource: %0 x %1',[texture.width, texture.height]);
  //addlog(ExplainCallStack(1));
    status:= glCheckFramebufferStatus(GL_FRAMEBUFFER_EXT);
    if status <> GL_FRAMEBUFFER_COMPLETE_EXT then Die  (
      RuEn(
        'Не удалось создать объект фреймбуфера %1x%2 (%0)',
        'Failed to create a framebuffer object %1x%2(%0)'),
      [GlErrorCodeToString(status), texture.width, texture.height]);

    f_width:= f_texture.Width;
    f_height:= f_texture.Height;

    if Mother^.Debug.Verbose then AddLog(
      '    Created a %0x%1 FBO in %2 microseconds',[texture.width, texture.height, round(Mother^.Timer.UsecDelta(@q))]);

    UpdateHash;
  except
    OnGenResourceFailure(StopDying());
  end;
  glBindFramebuffer(GL_FRAMEBUFFER_EXT, 0);
  CheckGlError;
end;

procedure TManagedFbo.OnGenResourceFailure(ErrorMessage: UnicodeString);
begin

  if (f_width = 256) and (f_height = 256) then Die(ErrorMessage); //we tried valiantly, but alas.

  Fallback;
end;

procedure TManagedFbo.Fallback;
begin
  if Assigned(f_texture) then f_texture.Free;
  if Assigned(f_depthbuffer) then f_depthbuffer.Free;
  f_width:= 256;
  f_height:= 256;
  desired_width:= 256;
  desired_height:= 256;
  if Assigned(Module) and Assigned(Module.LoadBalancer)
    then Module.LoadBalancer.EmergencyDrop; //Oh crap, the video card is on fire!
  CreateBuffers;
end;

function TManagedFbo.TextureFormat: GLuint;
begin
  Result:= GL_RGBA8;
  if OpenGLQuirkFboCantGL_RGBA8 in Mother^.GAPI.Quirks then Result:= GL_RGBA16;
  //GL_RGBA8;//GL_RGB8;//GL_RGBA16;//GL_RGBA8;
  //guess what, the ancient gF FX 5200 can't rgba8 but can rgba16 and rgb8 :\
  //*** TODO: make it a program-wide check, to determine best format automatically
end;

function TManagedFbo.TextureImageFormat: GLuint;
begin
  Result:= GL_RGBA;//GL_BGRA;
  //W.T.F ? Any other values result in GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT_EXT
end;

function TManagedFbo.DepthBufferFormat: GLuint;
begin
  Result:= GL_DEPTH_COMPONENT16;//GL_DEPTH24_STENCIL8;//GL_DEPTH_COMPONENT32; //GL_DEPTH_COMPONENT24;
end;



procedure TManagedFbo.RegisterFields;
begin
  inherited;
  ListFields([
    'texture', @f_texture, 'depth_buffer', @f_depthbuffer, typeinfo(TManagedFboTexture),
    'replacement', @f_replacement, typeinfo(TManagedFbo),

    //do NOT save these:
    '-dwidth', @desired_width, '-dheight', @desired_height, '-dfmt', @desired_format, typeinfo(GLuint)
  ]);
end;

destructor TManagedFbo.Destroy;
begin
  if Assigned(f_texture) then f_texture.Free;
  if Assigned(f_depthbuffer) then f_depthbuffer.Free;
  inherited;
end;


procedure TManagedFbo.RequestDesiredDimensions(w, h: integer; fmt: GLuint);
begin
  desired_width:= w;
  desired_height:= h;
  if fmt = 0 then begin
    if Assigned(f_texture)
      then desired_format:= f_texture.Format
      else desired_format:= Self.TextureFormat;
  end
  else
    desired_format:= fmt;
end;

procedure TManagedFbo.Pulse;
var
  isopt: boolean;
  rp: TAssetPropertiesSet;
begin
  if not Actualized then Exit;

  if ((desired_width = f_width) or (desired_width = 0))
    and ((desired_height = f_height) or (desired_height = 0))
    and ((desired_format = Self.TextureFormat) or (desired_format = 0))
  then begin
    if Assigned(f_replacement) then begin
      //in theory, the owning viewport COULD first request a huge replacement
      //  to be created, then change its mind and go back to the initial values
      //f_replacement.Scrape;
      //pointer(f_replacement):= nil;
      f_replacement:= nil;
    end;
    Exit;
  end;

  if f_replacement.Actualized then begin

    //Since the Devour method is designed for loading, with Chepsersy doing
    //  a part of the work, we have to perform the same actions manually:

    isopt:= ap_optional in f_props;

    Self.FreeResource;
    f_texture.FreeResource;
    f_depthbuffer.FreeResource;

    f_texture.Devour(f_replacement.texture);
    f_depthbuffer.Devour(f_replacement.depthbuffer);
    Self.Devour(f_replacement);

    f_replacement.texture.EraseContainer;
    f_replacement.depthbuffer.EraseContainer;
    f_replacement.EraseContainer;

    f_replacement.Free;

    if not isopt then f_props-= [ap_optional];

    pointer(f_replacement):= nil;
    desired_width:= 0;
    desired_height:= 0;
    desired_format:= 0;
    Exit;
  end;

  if not Assigned(f_replacement) then begin
    rp:= f_props;
    rp+= [ap_optional];
    rp-= [ap_actualized];
    f_replacement:= TManagedFbo.Create(
      desired_format, desired_width, desired_height, rp)
  end;
end;

function TManagedFbo._GetActualized: boolean;
begin
  Result:= inherited _GetActualized and f_texture.Actualized and f_depthbuffer.Actualized;
end;




procedure RegClasses;
begin
  RegClass(TFboDepthRenderbuffer);
  RegClass(TManagedFbo);
end;



end.
