{
    This file is part of Chentrah,
    Copyright (C) 2004-2014 Anton Rzheshevski (chebmaster@mail.ru).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see http://www.gnu.org/licenses/

 **********************************************************************}

//This is for _asynchronous_ file operations
//C

{$include mo_globaldefs.h}
{$typeinfo on}

unit mo_filemanager;
interface
uses SysUtils, Classes {$ifdef windows}, Windows {$else} , baseunix, dl {$endif},
  mo_hub, chepersy, mo_classes, syncobjs, mo_threads, mo_assets, mo_mstream,
   {$if defined(windows) and defined(win_unicode_fnames)}
     un_filelist_wide
   {$else}
     un_filelist
   {$endif}
  , mo_pakfiles

  ;//, un_unzip;

type


  { TAbstractFileTask }

  TAbstractFileTask = class (TAbstractThreadedTask)
    //The threaded tasks are NEVER registered, so there should be no need for RegisterFields
  protected
    f_content: pointer;
    f_contentSize: Longint; //NOT intended for working with files larger than a few megabytes
    f_fileName: TFileNameString;
  public
    property FileName: TFileNameString read f_fileName;
    destructor Destroy; override;
//    procedure Scrape; override;
    function GetType: TThreadedTaskType; override;
  end;


  { TFileLoadTask }

  TFileLoadTask = class(TAbstractFileTask)
  protected
    //Considers its owner a TFileRelatedAsset (e.g. TImage or TPrebaked3dModel)
    //  and relies on its FileName property
    f_buffer: TArrayOfByte;
    f_hash: TAssetHash;
    function Perform: boolean; override;
  public
    constructor Create(fname: TFileNameString);
    property PContent: pointer read f_content;
    property ContentSize: Longint read f_contentSize;
    property Hash: TAssetHash read f_hash;
    destructor Destroy; override; //procedure Scrape; override;
  end;

  { TFileSaveTask }

  TFileSaveTask = class(TAbstractFileTask)
  private
    function Perform: boolean; override;
  public
    constructor Create(fname: TFileNameString; Pcontent: pointer; ContentSize: longint);
  end;

  { TFileManagerInitTask }

  TFileManager = class; //forward declaration

  TFileManagerInitTask = class(TAbstractThreadedTask)
    (*
    Note: this doesn't load files into memory. It finds and opens
    pak files to list their content, then lists other, unpacked files.
    *)
  protected
    f_fm: TFileManager;
    f_list: TFileListArray;
    function Perform: boolean; override;
  public
    property List: TFileListArray read f_list;
    function GetType: TThreadedTaskType; override;
    constructor Create(fm: TFileManager);
  end;


//  TMemoryMappedFileArray = array of TMemoryMappedFile;

  { TFileManager }

  TFileManager = class(TGenericAsset)
  protected
    f_list: TFileListArray;
    f_paks: TArrayOfPackFiles;
    f_init_task: TFileManagerInitTask;
    function _GetActualized: boolean; override;
    class var sublist: array of array of TFilenameString;
    class procedure ValidateSubstitutionsList;
//b: boolean;
  public
    property PakFiles: TArrayOfPackFiles read f_paks;
    property List: TFileListArray read f_list;

    procedure RegisterFields; override;
    procedure OnModuleDestroy; override;
//    procedure GenResource; override;
//    procedure FreeResource; override;
    function IsValidForServer: boolean; override;

    procedure ClearCache;
    procedure Refresh; virtual;

    function Actualize: boolean; override;

    function FileExists(n: TFileNameString): boolean;
    function GetFileSize(n: TFileNameString): ptrint;

    function FindFile(fname: TFileNameString): PFileListEntry;

    //To be used by threaded tasks ONLY:
    //Checks the current thread ID
    function ReadFile(fname: TFileNameString; out fhash: TAssetHash): TArrayOfByte;
    function ReadFile(pf: PFileListEntry; out fhash: TAssetHash): TArrayOfByte;

    function GetHash(fname: TFileNameString): TAssetHash;
    function MakeCombinedHashOfManyFiles(
      _list: TFileListArray): TAssetHash; { in this case, hash consists of:
       NameMd5 -- md5 of all names joined sequentially, separated by commas
       Q3 - number of files
       Q4 - total size of files
    }

    function Devour(c: TGenericAsset): boolean; override;
    procedure EraseContainer; override;
    procedure CopyLists(out l: TFileListArray; out p: TArrayOfPackFiles); //on devoured
    procedure EraseLists; virtual; //on devoured
    procedure DumpIncompleteLists; virtual; {Usually called after devour
      if not actualized (one of the pak files changed on disk).
      Can be invoked by the pak files if it finds the file on disk changed
    }
    procedure RebuildListFromPaks;
  end;

implementation

uses //crc,
  mo_module;

{ TFileSaveTask }

function TFileSaveTask.Perform: boolean;
begin
  if not Module.Logic.FileManager.Actualized then Exit(false);
  Die(RuEn('Не доделано, блеать!','Not Implemented Yet!'))

end;


constructor TFileSaveTask.Create(fname: TFileNameString; Pcontent: pointer;
  ContentSize: longint);
begin
  inherited Create;
  f_fileName:= fname;
  f_content:= Pcontent;
  f_contentSize:= ContentSize;
  Reschedule;
end;
{ TFileLoadTask }

function TFileLoadTask.Perform: boolean;
begin
  if Mother^.Debug.Verbose then AddLog('   %0.Perform (%1)...', [string(Self.ClassName), f_fileName]);
  if not (Assigned(Module.Logic) and Module.Logic.FileManager.Actualized) then Exit(false);
  f_buffer:= Module.Logic.FileManager.ReadFile(f_fileName, f_hash);
  f_contentSize:= Length(f_buffer);
  if f_contentSize = 0
    then f_content:= nil
    else f_content:= @f_buffer[0];
  if Mother^.Debug.Verbose then AddLog('     %0.Perform end(%2 bytes from %1)...', [string(Self.ClassName), f_fileName, f_contentsize]);
end;

destructor TFileLoadTask.Destroy;
begin
  SetLength(f_buffer, 0);
  f_content:= nil;
  f_contentSize:= 0;
  inherited;
end;

constructor TFileLoadTask.Create(fname: TFileNameString);
begin
  inherited Create;
  f_fileName:= fname;
  Reschedule;
end;


{ TFileManagerInitTask }

constructor TFileManagerInitTask.Create(fm: TFileManager);
begin
  inherited Create;
  f_fm:= fm;
  Reschedule;
end;

function TFileManagerInitTask.Perform: boolean;
var
  paths: TFileNameStringArray;
  fl: TFullFileList;
  e: TFileListEntry;
  i, j, jr, b, n: integer;
  q: qword;
  us: double;
  exte: TFileNameString;
begin
  //paths:= Module.FileManagerPaths(); //by default, only one entry, '*M/'
  SetLength(paths, 1); paths[0]:= '*M/';
  for i:=0 to High(paths) do begin
    if Mother^.Debug.Verbose then AddLog('   building file list for path "%0"',[UnmangleFileName(paths[i])]);
    Mother^.Timer.UsecDelta(@q);
    fl:= TFullFileList.Create(UnmangleFileName(paths[i]));
    fl.SortByName;
    us:= Mother^.Timer.UsecDelta(@q);
    n:= 0;
    for j:= 0 to fl.Count - 1 do
      if fl.Size[j] > 0 then Inc(n); //also weeds directories out
    if Mother^.Debug.Verbose then AddLog('     directory tree traversed in %2µs, %0 entries, %1 of them non-zero',[fl.Count, n + 1, round(us)]);
    b:= Length(f_list);
    SetLength(f_list, b + n);
    jr:= 0;
    for j:= 0 to fl.Count - 1 do begin
      if fl.Size[j] <= 0 then continue;
      e.name:= MangleFileName(fl.Names[j]);

      { Ignore the sources and the DLL! Otherwise the hash would change
          with each compilation, leading to unnecessary reloading
          of file manager when loading older sessions. }
//addlog('& ' + e.name);
      if COPY(e.name, 1, 7) = '*M/src/' then continue;
      exte:= LowerCase(ExtractFileExt(fl.Names[j]));
      if (exte = '.dll') or (exte = '.so') or (exte = '.zd2')
        or (exte = '.ini') or (exte = '.module') then continue;
//addlog('  +');
      e.pak_file:= nil;
      //e.Hash.NameCRC:= crc32(0, @e.name[1], Length(e.name) * SizeOf(e.name[1]));
      e.Hash.NameMD5:= Md5String(e.name);
      e.Hash.FileSize:= fl.Size[j];
      //e.Hash.Age:= fl.MSDOSTimeStamp[j];
      e.Hash.FileAge:= FileDateTodateTime(fl.MSDOSTimeStamp[j]);
      f_list[b + jr]:= e;
      Inc(jr);
    end;
  end;
  Result:= true;
end;

function TFileManagerInitTask.GetType: TThreadedTaskType;
begin
  Result:= ttt_File;
end;

{ TAbstractFileTask }

destructor TAbstractFileTask.Destroy;
begin
  if Assigned(f_content) then begin
    FreeMem(f_content);
    f_content:= nil;
    f_contentSize:= 0;
  end;
  inherited Destroy;
end;

//procedure TAbstractFileTask.Scrape;
//begin
  //dispose of the memory right away, not waiting for garbage collector to deal with it
//  if Assigned(f_content) then begin
//    FreeMem(f_content);
 //   f_content:= nil;
 //   f_contentSize:= 0;
//  end;
  //inherited Scrape;
//end;

function TAbstractFileTask.GetType: TThreadedTaskType;
begin
  Result:= ttt_File;
end;


{ TFileManager }

procedure TFileManager.RegisterFields;
begin
  RegisterPakFileClasses;
//  RegType(typeinfo(TMemoryMappedFileArray), typeinfo(TMemoryMappedFile));
  inherited;
  ListFields([
    'list', @f_list, typeinfo(TFileListArray),
    'paks', @f_paks, typeinfo(TArrayOfPackFiles),
    '-ptr1', @f_init_task, CPS_POINTER
//,'b',@b, typeinfo(boolean)
  ]);
end;




procedure TFileManager.OnModuleDestroy;
begin
  ClearCache;
  inherited;
end;


procedure TFileManager.ClearCache;
begin

end;


procedure TFileManager.Refresh;
begin
  DumpIncompleteLists;
  Actualize;
end;

function TFileManager._GetActualized: boolean;
var i: integer;
begin
  Result:= inherited _GetActualized;

  //Check the child assets
  if Result then
    for i:=0 to High(f_paks) do
      if not Assigned(f_paks[i]) or not f_paks[i].Actualized
        then Exit(false);
end;

function TFileManager.Actualize: boolean;
var
  i, j: integer;
  es: TFileNameString;
begin


  if not Assigned(f_init_task) then f_init_task:= TFileManagerInitTask.Create(Self);

  if not f_init_task.Completed then Exit(false);

  if Length(f_list) = 0 then begin
    if Mother^.Debug.Verbose then AddLog('    The background task is completed, returning %0 files.',[Length(f_init_task.List)]);

    f_list:= f_init_task.List; //copy

    //They are *already* sorted by date descending, so the new ones are
    //  guaranteed to appear on top
    for i:=0 to High(f_list) do begin
      es:= LowerCase(ExtractFileExt(f_list[i].name));
      if (es = '.pk3') or (es = '.zip') then begin
        if Mother^.Debug.Verbose then AddLog('      Pak file %0 size=%1', [f_list[i].name, f_list[i].Hash.FileSize]);
        SetLength(f_paks, Length(f_paks) + 1);
        f_paks[High(f_paks)]:= TZipFile.Create(f_list[i].name);
      end
      else begin
        if Mother^.Debug.Verbose then AddLog('      %0 size=%1', [f_list[i].name, f_list[i].Hash.FileSize]);
      end;
    end;
    if Mother^.Debug.Verbose then AddLog('    Found %0 pak files', [Length(f_paks)]);
  end;

  //now we *wait* until the pak files initialize
  for i:= 0 to High(f_paks) do
    if not f_paks[i].Actualized then begin
      if Mother^.Debug.Verbose then AddLog('    One or more pak files aren''t actualized yet: %0',[f_paks[i].Name]);
      Exit(false);
    end;

  //all pak files are ready. Dump their lists into our own list.
  RebuildListFromPaks;

  f_hash:= Self.MakeCombinedHashOfManyFiles(f_list);

  //note we can't call inherited here, as this class doesn't have a GenResource method.
  //Result:= inherited; // this would cause an EAbstractError
  MarkAsActualized;
  Result:= true;
  f_init_task.Free;
  pointer(f_init_task):= nil;
end;

procedure TFileManager.DumpIncompleteLists;
//causes the file manager to be initialized anew
var i: integer;
begin
  for i:= 0 to High(f_paks) do
    if Assigned(f_paks[i])
      then f_paks[i]:= nil;
  SetLength(f_list, 0);
  SetLength(f_paks, 0);

  if Assigned (f_init_task) then begin //repeat the initialization threaded task
    f_init_task.Free;
    pointer(f_init_task):= nil;
  end;

  if ap_actualized in f_props
    then Exclude(f_props, ap_actualized)
    else MyLogClass.AddNonActualizedAsset(Self);
end;

procedure TFileManager.EraseContainer;
begin
  Include(f_props, ap_actualized); // a hack. See above.
  DumpIncompleteLists;
  Exclude(f_props, ap_actualized);
  inherited;
end;

procedure TFileManager.EraseLists;
begin
  //called on a devoured object.
  //Clear lists without deleting contents as these are already copied by the devourer.
  SetLength(f_list, 0);
  SetLength(f_paks, 0);
  Exclude(f_props, ap_actualized);
end;

procedure TFileManager.CopyLists(out l: TFileListArray; out p: TArrayOfPackFiles);
begin
  //Callede on a devoured object.
  l:= f_list;
  p:= f_paks;
end;

procedure TFileManager.RebuildListFromPaks;
var i, j, c : integer;
begin
  //1. Cut off the tail consisting of files listed as contained in the pak files
  i:= 0;
  while (i < High(f_list)) and not Assigned(f_list[i].pak_file) do inc(i);
  SetLength(f_list, i);

  //2. Combine the files from the pak files onto the list.

  //Note: no check is necessary, as the file read method goes down the list
  //  and uses the first instance it meets. So unpacked files take precedence
  //  over those in paks, and newer paks take precedence over older.

  //The small overhead of comparing a shitload strings each time is mitigated
  //  a) by Pascal being *fast* with it (string lengths are compared first)
  //  b) by it being performed in a treaded task anyway.

  for j:= 0 to High(f_paks) do
    with f_paks[j] do begin
      c:= Length(f_list);
      SetLength(f_list, Length(f_list) + Length(List));
      for i:= 0 to High(List) do begin
        f_list[c + i]:= List[i];
        f_list[c + i].pak_file:= f_paks[j];
      end;
    end;


end;

function TFileManager.FileExists(n: TFileNameString): boolean;
var i: integer;
begin
  for i:=0 to High(f_list) do
    if f_list[i].name = n then Exit(true);
  Result:= false;
end;

function TFileManager.GetFileSize(n: TFileNameString): ptrint;
var i: integer;
begin
  for i:=0 to High(f_list) do
    if f_list[i].name = n then Exit(f_list[i].Hash.FileSize);
  Result:= -1;
end;

class var sublist: array of array of TFilenameString;
class procedure TFileManager.ValidateSubstitutionsList;
begin
  if Length(sublist) > 0 then Exit;
  SetLength(sublist, 1);
  sublist[0]:= Explode('|', '.$img|.tga|.png|.bmp|.pcx|.jpg|.jpeg');
end;


function TFileManager.FindFile(fname: TFileNameString): PFileListEntry;
var
  i, j, subind: integer;
  ext, fex: TFileNameString;
  ematch: boolean;
begin
  Assert(Self.Actualized,
    'TFileManager.FindFile should only be called after the file manager is actualized!');
  Result:= nil;
  ValidateSubstitutionsList;
  subind:= -1;
  ext:= LowerCase(ExtractFileExt(fname));
  if Length(ext) > 0 then
    for j:= 0 to High(sublist) do
        if sublist[j][0] = ext then begin
          subind:= j;
          break
        end;
  if subind >=0 then begin
    //I do not give a flying fuck about performance here, it all happens in a
    //  background thread anyway
    for j:= 1 to High(sublist[subind]) do begin
      fname:= ChangeFileExt(fname, sublist[subind][j]);
      for i:=0 to High(f_list) do
        if f_list[i].name = fname then Exit(@f_list[i]);
    end;
  end
  else begin
    for i:=0 to High(f_list) do
      if f_list[i].name = fname then Exit(@f_list[i]);
  end;
end;



function TFileManager.ReadFile(fname: TFileNameString; out fhash: TAssetHash): TArrayOfByte;
var
  i: integer;
  pf: PFileListEntry = nil;
begin
  Assert(Assigned(Module.ThreadManager)
    and (
      Mother^.Module.MultithreadingDisabled
      or Module.ThreadManager.IsCurrentThreadTheFileThread),
    'TFileManager.ReadFile should only be called from one specific thread!');
  Assert(Self.Actualized,
    'TFileManager.ReadFile should only be called after the file manager is actualized!');
  pf:= FindFile(fname);
  if not Assigned(pf) then begin
    FillChar(fhash, sizeof(fhash), 0);
    Exit(nil);
  end;
  Result:= ReadFile(pf, fhash);
end;

function TFileManager.ReadFile(pf: PFileListEntry; out fhash: TAssetHash): TArrayOfByte;
var
  s: TFileStream;
  i: integer;
begin
  Assert(Assigned(Module.ThreadManager)
    and (
      Mother^.Module.MultithreadingDisabled
      or Module.ThreadManager.IsCurrentThreadTheFileThread),
    'TFileManager.ReadFile should only be called from one specific thread!');

  fhash:= pf^.Hash;
  SetLength(Result, pf^.Hash.FileSize);
  if Assigned(pf^.pak_file)
    then Result:= pf^.pak_file.ReadFile(pf^.name)
    else begin
      try
        s:= TFileStream.Create(UnmangleFileName(pf^.name) , fmOpenRead + fmShareDenyWrite);
        SetLength(Result, s.Size);
        i:= 0;
        while i < Length(Result) do i+= s.Read(Result[i], Length(Result) - i);
        s.Free;
      except
        Die(RuEn(
          'Сбой при загрузке файла'#10#13'  %0',
          'Failed to load file'#10#13'  %0'
         ), [UnmangleFileName(pf^.name)]);
      end;
    end;
end;


function TFileManager.GetHash(fname: TFileNameString): TAssetHash;
var
  I: integer;
begin
  for i:=0 to High(f_list) do
    if f_list[i].name = fname
      then Exit(f_list[i].Hash);
  ZeroHash(Result);
end;

function TFileManager.MakeCombinedHashOfManyFiles(_list: TFileListArray): TAssetHash;
var
  i: integer;
  allnames: TFileNameString;
begin
  //in this case, hash consists of:
  // Age -- equals to number of files
  // NameCRC -- crc32 of all names joined sequentially
  // FileSize -- combined size of all files
  ZeroHash(Result);
  //Result.Age:= Length(_list);
  Result.Q4:= 0;
  Result.Q3:= Length(_list);
  allnames:= '';
  for i:= 0 to High(_list) do begin
    Result.Q4+= _list[i].Hash.FileSize;
    allnames+= _list[i].name;
  end;
  if allnames = '' then allnames:= '*No files found!*';
  //Result.NameCRC:= crc32(0, @allnames[1], Length(allnames) * SizeOf(allnames[1]));
  Result.NameMd5:= Md5String(allnames);
end;

function TFileManager.IsValidForServer: boolean;
begin
  Result:= true;
end;

{
function TFileManager.Devour(c: TGenericAsset): boolean;
var i: integer;
begin
  with c as TFileManager do begin
    CopyLists(f_list, f_paks);
    EraseLists;
  end;
  for i:= 0 to High(f_paks) do
    if not f_paks[i].Actualized
      then Exit(false);

  RebuildListFromPaks; //as it would be pointing to wrong pak objects now!
  Result:= inherited Devour(c);
end;
}


function TFileManager.Devour(c: TGenericAsset): boolean;
var
  co: TFileManager;
  i: integer;
begin
  co:= c as TFileManager;
  if Mother^.Debug.Verbose then AddLog('   %0.Devour...', [string(ClassName)]);

  //make sure all the pak files match
  if Length(f_paks) <> Length(co.PakFiles) then begin
    if Mother^.Debug.Verbose then AddLog('    Number of paks doesn''t match (%0 vs %1). Aborting.', [Length(co.PakFiles), Length(f_paks)]);
    Exit(false);
  end;
  for i:= 0 to High(f_paks) do
    if f_paks[i].Hash <> co.PakFiles[i].Hash then begin
      if Mother^.Debug.Verbose then AddLog('    Hashes do not match for pak #%0 (%0 vs %1). Aborting.', [i, HashToHex(co.PakFiles[i].Hash), HashToHex(f_paks[i].Hash)]);
      Exit(false);
    end;

  //lists DO match as the hash guarantees it.

  f_list:= COPY(co.List);
  RebuildListFromPaks; //as it would be pointing to wrong pak objects now!

  Result:= inherited Devour(c);
  co:= nil;
end;


end.
