
  destructor TGameTickSchedule.Destroy;
  begin
    if Assigned(dyA) then SetLength(dyA, 0);
    inherited;
  end;

  procedure TGameTickSchedule.RegisterFields;
  begin
    RegType(typeinfo(Longint32Array), typeinfo(Longint), [0, 31]);
    inherited;
    ListFields([
      'stA', @stA, typeinfo(Longint32Array),
      'dyA', @dyA, typeinfo(TArrayOfLongint),
      'f_count', @f_count, typeinfo(Longint)
    ]);
  end;

  procedure TGameTickSchedule.Add(ind: Longint); { They are never removed, for
      simplicity's sake. If the object itself says it is scheduled for a
      latter tick, it just gets skipped.
    The per-tick data gets erased when that tick is passed, anyway.
    One index may be included several times or included in several STs. This
      is resolved by the game layer using object's own fields.
    }
  begin
    inc(f_count);
    if f_count <= 32 then stA[f_count - 1]:= ind
    else begin
      if Length(dyA) < (f_count - 32)
        then SetLength(dyA, Math.max(32, Length(dyA) * 2));
      dyA[f_count - 32 - 1]:= ind;
    end;
  end;

  function TGameTickSchedule.Get(num: Longint): Longint;
  begin
    Assert(num < f_count, 'TGameTickSchedule.Get: num >= Count!');
    if num < 32
      then Result:= stA[num]
      else Result:= dyA[num - 32];
  end;










(*

  procedure TGameTickSchedule.BeforeSaving;
  begin
    ShrinkArrays;
  end;

  constructor TGameTickSchedule.Create(CurrentST: SmallInt);
  begin
    inherited Create;
    reloc_offset:= CurrentST and MOSLTMASK;
  end;

  function TGameTickSchedule.GetFirst(st: SmallInt): Longint;
  var
    stro: Longint;
  begin
    {$push}
      {$rangecheck off}
      {$overflowcheck off}
      stro:= (st - reloc_offset) and MOSLTMASK;
    {$pop}
    if ((stro * 2 + 1) > High(reloc))
      or (reloc[stro * 2] < 0) then Exit(0);

    current_get_current:= reloc[stro * 2];
    Result:= A[current_get_current];
  end;

  function TGameTickSchedule.GetNext(): Longint;
  begin
    Inc(current_get_current);
    Result:= A[current_get_current];
    if Result < 0 then begin
      current_get_current:= -Result;
      Result:= A[current_get_current];
    end;
  end;

  procedure TGameTickSchedule.Add(ind: Longint; st: SmallInt);
  var
    stro, i, j, tail, lnk: Longint;
  begin
    Assert(index > 0, 'TGameTickSchedule.Add() : index must be positive');
    {$push}
      {$rangecheck off}
      {$overflowcheck off}
      stro:= (st - reloc_offset) and MOSLTMASK;
    {$pop}
    if (stro * 2 + 1) > High(reloc) then begin
      //grow the array if necessary needed
      j:= Length(reloc) div 2;
      SetLength(reloc, (stro + 1) * 2);
      for i:= j to stro do begin
        reloc[j * 2] := -1;
        reloc[j * 2 + 1] := -1;
      end;
    end;
    tail:= reloc[stro * 2 + 1];
    if tail < 0 then begin
      //the line is empty. Create a new one;
      tail:= AllocateChunk();
      reloc[stro * 2]:= tail; //remember the first index
      reloc[stro * 2 + 1]:= tail; //mark the last index
      A[tail]:= index;
      A[tail + 1]:= 0;
    end
    else begin
      Inc(tail);
      if (tail and $7) = $7 then begin
        lnk:= AllocateChunk();
        A[tail]:= -lnk;
        tail:= lnk;
      end;
      A[tail]:= index;
      A[tail + 1]:= 0;
    end;
  end;

  function TGameTickSchedule.AllocateChunk(): Longint;
  var
    i, j: integer;
  begin
    Result:= -1;
    for i:= 0 to High(freebitmask) do begin
      if freebitmask[i] <> $ffffffff then begin
        Result:= i * 32 + BsfDword(not freebitmask[i]);
        break;
      end;
    end;
    if Result > 0 then begin
      //Found an empty chunk :D
      freebitmask[Result div 32]:= freebitmask[Result div 32] and (dword(1) shl (Result and $1f));
    end
    else begin
      //Full :( Must enlarge the array.
      SetLength(freebitmask, Length(freebitmask) + 1);
      freebitmask[High(freebitmask)]:= 1;
      Result:= 32 * High(freebitmask);
      j:= Length(A);
      SetLength(A, Length(freebitmask) * 8 * 32);
      for i:= j to High(A) do A[i]:= 0;
    end;
  end;

  procedure TGameTickSchedule.EraseSt(st: SmallInt);
  var
    i, head, stro: Longint;
    procedure FreeCorrespondingChunk(ci: integer); inline;
    var
      j: Longint;
    begin
      for j:= 0 to 7 do A[ci + j]:= 0;
      freebitmask[ci div (8 * 32)]:= freebitmask[ci div (8 * 32)] and not (dword(1) shl ((ci div 8) and $1f) );
    end;
  begin
    {$push}
      {$rangecheck off}
      {$overflowcheck off}
      stro:= (st - reloc_offset) and MOSLTMASK;
    {$pop}
    if (stro * 2 + 1) > High(reloc) then Exit;
    head:= reloc[stro * 2];
    repeat
      i:= head;
      head:= - A[head + 7];
      FreeCorrespondingChunk(i);
    until head = 0;
    reloc[stro * 2]:= -1;
    reloc[stro * 2 + 1]:= -1;
  end;

  procedure TGameTickSchedule.ShrinkArrays();
  var
    i: integer;
  begin
    i:= High(reloc) div 2;
    while (i > 0) and (reloc[i * 2] < 0) do dec(i);
    SetLength(reloc, (i + 1) * 2);

    i:= High(freebitmask);
    while (i > 0) and (freebitmask[i] = 0) do dec(i);
    SetLength(freebitmask, i + 1);
    SetLength(A, (i + 1) * 8 * 32);
  end;
  *)
