{
    This file is part of Chentrah,
    Copyright (C) 2004-2015 Anton Rzheshevski (chebmaster@mail.ru).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see http://www.gnu.org/licenses/

 **********************************************************************}

unit mo_gl_quirks;

{$include mo_globaldefs.h}

interface
  uses mo_hub;

  //see also TGAPIQuirksEnum in un_typedefs.h

  //quirks are a per-videocard setting (device name + opengl driver name)
  // saved on mother exit and loaded at mother startup

  //Quirk #1 (glScissor + glClear causes fallback to software rendering)
  //could only be detected on the fly! E.g. Only after running at least
  //  10 seconds with an active TViewPort present and rendering.
  //Thus, it is never stated as tested until it is detected.

  //Quirk #2 : FBO is incompatible with GL_RGBA8 color buffer format
  procedure TestOpenGlQuirkNo2;

implementation

  procedure TestOpenGlQuirkNo2;
  //FBO supports GL_RGB8 and GL_RGBA16 but not GL_RGBA8
  //Found on: GeForce FX 5200
  //Detected by: failing to create a 256x256 FBO with a GL_RGBA8 color buffer
    function TryCreateFbo(cbf: GLuint; txt: string): boolean;
    var db, fbo, cb: GLuint;
    begin
      if Mother^.Debug.Verbose then AddLog('  Creating a FBO with %0 color buffer...',[txt]);
      glGenRenderbuffers(1, @db);
      CheckGlError;
      glBindRenderbuffer(GL_RENDERBUFFER_EXT, db);
      CheckGlError;
      glRenderbufferStorage(GL_RENDERBUFFER_EXT, GL_DEPTH_COMPONENT24, 256, 256);
      CheckGlError;
      glGenTextures(1, @cb);
      CheckGlError;
      glEnable(GL_TEXTURE_2D);
      glBindTexture(GL_TEXTURE_2D, cb);
      CheckGlError;
      glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
      glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
      glTexImage2D ( GL_TEXTURE_2D, 0, cbf, 256, 256, 0, GL_BGRA, GL_UNSIGNED_BYTE, NIL);
      CheckGlError;
      glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAX_LEVEL, 0);
      glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_BASE_LEVEL, 0);
      CheckGlError;
      glGenFramebuffers(1, @fbo);
      glBindFramebuffer(GL_FRAMEBUFFER_EXT, fbo);
      CheckGlError;
      glBindTexture(GL_TEXTURE_2D, Mother^.GAPI.StubTexture); //bind a safe non-used texture
      glFramebufferTexture2D(
        //GL_DRAW_FRAMEBUFFER
        GL_FRAMEBUFFER_EXT
        , GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, cb, 0);
      CheckGlError;
      glFramebufferRenderbuffer(GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT, GL_RENDERBUFFER_EXT, db);
      CheckGlError(true);
      Result:= GL_FRAMEBUFFER_COMPLETE_EXT = glCheckFramebufferStatus(GL_FRAMEBUFFER_EXT);
      glBindFramebuffer(GL_FRAMEBUFFER_EXT, 0);
      CheckGlError;
      glDeleteFramebuffers(1, @fbo);
      CheckGlError;
      glDeleteTextures(1, @cb);
      CheckGlError;
      glDeleteRenderbuffers(1, @db);
      CheckGlError;
      if Mother^.Debug.Verbose then AddLog('  ..result=%0',[Result]);
    end;
  begin
    if OpenGLQuirkFboCantGL_RGBA8 in Mother^.GAPI.QuirksTested then Exit; //already tested
    try
      if Mother^.Debug.Verbose then AddLog('This video card haven''t yet been tested for OpenGL quirk #2. Trying to detect it...');

      if TryCreateFbo(GL_RGBA16, 'GL_RGBA16') and not TryCreateFbo(GL_RGBA8, 'GL_RGBA8')
        then Mother^.GAPI.Quirks+= [OpenGLQuirkFboCantGL_RGBA8];

      if Mother^.Debug.Verbose then AddLog('The quirk is marked as checked. Detected=%0', [OpenGLQuirkFboCantGL_RGBA8 in Mother^.GAPI.Quirks]);
      Mother^.GAPI.QuirksTested+= [OpenGLQuirkFboCantGL_RGBA8];
    except
      Die(RuEn(
        'Сбой при проверке OpenGL на глюк №2 (FBO не может работать с 8-битным цветовым буфером с альфа-каналом)',
        'Failed to test for OpenGL quirk #2 (FBO cannot work with 8-bit RGBA color buffer)'
      ));
    end;
  end;

end.
