{
    This file is part of Chentrah,
    Copyright (C) 2004-2014 Anton Rzheshevski (chebmaster@mail.ru).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see http://www.gnu.org/licenses/

 **********************************************************************}

unit mo_gmathbase;

interface

uses
  sysutils, Classes, math, md5, chepersy, mo_hub, typinfo;

(*
{$ifdef dosmath}
 type
  ffloat = float; // no multiplayer here, d'uh. It's the goddamn MS-DOS!
 const
  FF_ZERO: float = 0.0;
  FF_ONE: float = 1.0;
  FF_MINUS_ONE: float = -1.0;
{$else}
*)
  //Windows

  {$ifdef fixedmath}  //multiplayer
    type
      ffloat = fixed32; //deterministic but inefficient
    const
     FF_ZERO = 0;
     FF_ONE = 1 shl 16;
     FF_MINUS_ONE = - (1 shl 16);
     FF_ONE_PERCENT = 655;
     FF_ONE_THOUSANDTH = 66;
  {$else}
{$fatal INVALIDATED SINCE JUNE 2015}
    type
      ffloat = float; //efficient but non-deterministic
    const
     FF_ZERO: float = 0.0;
     FF_ONE: float = 1.0;
     FF_MINUS_ONE: float = -1.0;
     FF_ONE_PERCENT: float = 0.01;
     FF_ONE_THOUSANDTH: float = 0.001;
  {$endif}
//{$endif}

 type
   TffVector = packed record
     x, y, z: fixed32;
   end;
   PffVector = ^TffVector;

  procedure RegisterGMathBaseTypes;
  function ValidateTextureDimension(i: GLint): GLint;
  function Validate3dTextureDimension(i: GLint): GLint;

  procedure CrossProduct(var v1, v2, cross: TffVector);
  function SquaredLength(var v: TffVector): ffloat;

implementation

procedure RegisterGMathBaseTypes;
begin
  RegType(typeinfo(TffVector), sizeof(TffVector), ['x', 'y', 'z', typeinfo(ffloat)]);

end;


(*

For what it's worth, here's a cross product function from Quake 3, with vec3_t defined as an array of three floats for x, y, and z:

void CrossProduct( const vec3_t v1, const vec3_t v2, vec3_t cross ) {
    cross[0] = v1[1]*v2[2] - v1[2]*v2[1];
    cross[1] = v1[2]*v2[0] - v1[0]*v2[2];
    cross[2] = v1[0]*v2[1] - v1[1]*v2[0];
}

*)

  procedure CrossProduct(var v1, v2, cross: TffVector);
  begin
    with cross do begin
      x:= v1.y*v2.z - v1.z*v2.y;
      y:= v1.z*v2.x - v1.x*v2.z;
      z:= v1.x*v2.y - v1.y*v2.x;
    end;
  end;

  function SquaredLength(var v: TffVector): ffloat;
  begin
    Result:= v.x*v.x + v.y*v.y + v.z*v.z;
  end;


function ValidateTextureDimension(i: GLint): GLint;
begin
  if (Mother^.GAPI.NpotSupported)
    then Result:= max(1, min(i, Mother^.GAPI.TextureMax))
    else Result:= max(1, min(GetHigherPowerOf2(i), Mother^.GAPI.TextureMax));
end;

function Validate3dTextureDimension(i: GLint): GLint;
begin
  if (Mother^.GAPI.NpotSupported)
    then Result:= max(1, min(i, Mother^.GAPI._3dTextureMax))
    else Result:= max(1, min(GetHigherPowerOf2(i), Mother^.GAPI._3dTextureMax));
end;


end.
