{
    This file is part of Chentrah,
    Copyright (C) 2004-2016 Anton Rzheshevski (chebmaster@mail.ru).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see http://www.gnu.org/licenses/

 **********************************************************************}

{
   This unit contains specialized memory manager to facilitate debugging
     of the layered game world.
   It uses protection mechansms (mprotect / VirtualProtect)
     to catch inter-layer leaks

   Keep in mind that this is *only* possible thanks to
     strictly single-threaded game logic.

   Also keep in mind that this manager is NOT memory-efficient, especially
     for blocks larger that about 2000 bytes.
}

{$include mo_globaldefs.h}
{$typeinfo on}

unit mo_go_mman;

interface
  uses SysUtils, Classes {$ifdef windows}, Windows {$else} , baseunix, dl {$endif}
    , chepersy;

  procedure InitWatchedMem(ChepersyReservedBits: dword);

  procedure CleanupWatchedMem; //on module DLL unloding, must free all pages
    //let's just say I don't trust initialization / finalization

  function GetWatchedMem(Size: longint): pointer;

  procedure FreeWatchedMem(p: pointer; ManageRefCount: boolean = false);

  function ReallocWatchedMem(p: pointer; Newsize: longint; ManageRefCount: boolean = false): pointer;

  function GetWatchedMemRefcountPtr(p: pointer): plongint;

  function GetWatchedMemFlagsPtr(p: pointer): pdword;

  procedure AllocateWatchedMemLayer: dword; // is a bit mask with exactly one bit set.
    //returns 0 if there are no free bits left.
    //please note that this mask is *shared* with Chepersy so that it is possible
    //to filter layers

  procedure MarkValidLayers(bitmask: dword);
    //any others will be marked as zombies AND forbidden (see below)

  procedure ProtectLayers(bitmask: dword); //only the listed layers will be writeable
    //for special cases only (normally you only want one layer at a time
    // to be writeable, see SwitchToWatchedMemLayer)

  procedure ForbidLayers(bitmask: dword);
  //all listed layers become protected from writing AND reading

  procedure ForcePurgeZombieLayers;

  procedure SwitchToWatchedMemLayer(bitmask: dword); //causes making that layer
    //writeable and all other valid layers to become protected.
    //Only one bit must be set in the mask.

implementation
  var
    PageSize: integer;

  procedure InitWatchedMem(ChepersyReservedBits: dword);
  begin

  end;




end.
