{
    This file is part of Chentrah,
    Copyright (C) 2004-2014 Anton Rzheshevski (chebmaster@mail.ru).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see http://www.gnu.org/licenses/

 **********************************************************************

    This file contains the main unit for the game modules.

 **********************************************************************}

{$include mo_globaldefs.h}


{$ifdef unix}
  { define extdecl := cdecl }
{$else}
  { define extdecl := stdcall }
{$endif}


unit mo_hub;

interface
uses
  SysUtils,
  Classes,
  typinfo,
  IniFiles
  {$ifdef unix}
   , variants
   , dl
  {$else}
   , windows
  {$endif}
   , md5
   , syncobjs
  ;

  {$include un_texsages.h}

   MI_CORE_MIDs_END
   //custom message IDs go here
   ,
   MI_OF_ASSET_LIST
  );
  
  {$include un_typedefs.h}

  {$ifdef fix_fpc3_unicode}
     {$include un_fpc_3-0-0_fix.h} // Use custom mods of TFileStream, TStringList and TIniFile
  {$endif}
  {$if defined(windows) and (FPC_FULLVERSION<30000) }
     {$include un_fpc_2-6-4_fix.h} // Use fixed TFileStream and copy-pasted everything that depends on it
  {$endif}
  
   procedure GrandInitialize(p: PMother; m,n,b: integer; logparams: array of const; sspl: integer);

  {$include un_lineinfo.h}

  {$include un_gles_from_fpc_2_6_4.h} //opengl es header
  {$ifndef glesonly} //for debugging purposes only (compatibility testing): cut off all support for regular opengl
    {$include un_gl_from_fpc_2_6_4.h} //opengl header
  {$endif}

  {$ifdef mo_uses_openal}
    { Normally, the game modules do not interact with OpenAL at all.
        The sound is rendered using FBO + GLSL (0.1 seconds worth pre-render)
        then passed to the mother module for buffering and output.  }
    {$include un_al.h}
  {$endif}
  
  {$ifdef notlaz} { A chunk of code blocked from the Lazarus IDE
     to avoid confusing its code tools}
    {$include mo_fpc_pl_chk.h} //checks if sizeof(pointer) = sizeof(dword)
  {$endif}

//  {$include un_interface.h}

  {$include un_unicode.h}

  {$include un_fileutil.h}
  
  {$include mo_gtypes.inc}

type
  TRegisterProc = procedure(n: integer; var p: pointer);
                                      {$ifdef notlaz}cdecl;{$endif}

  EMoDying = class(Exception);
  Tprocedure = procedure; cdecl;
  TTryWrapperProcedure = procedure(p: pointer); cdecl;

var
  Mother: PMother;
  SessionSpaceLimit: integer = 1;
  Config: TIniFile = nil;
  
  {$include un_api_common.h}

  function MsgFormat(M: TMessageId; Param: array of const): UnicodeString;
  function MsgRaw(M: TMessageId): UnicodeString;

//  function HashToStr(phash: pbyte): ansistring;

  function RuEn(Ru, En: UnicodeString): UnicodeString;
  function StrOrUndefined(U: UnicodeString): UnicodeString;

  function PCharToString(P: PAnsiChar): AnsiString;
  function PUnicodeCharToUnicodeString(P: PUnicodeChar): UnicodeString;
  function PFileNameCharToFileNameString(P: PFileNameChar): TFileNameString;

  Procedure Die(YellID :TMessageID; Param: array of const); overload;
  Procedure Die(AnsiYell: AnsiString); overload;
  Procedure Die(AnsiYell: AnsiString; param: array of const); overload;
  Procedure Die(YellID :TMessageID); overload;
  Procedure Die(u: UnicodeString; Param: array of const);  overload;
  Procedure Die(u: UnicodeString);  overload;
  Procedure Die(r,e: UnicodeString);  overload;
  Procedure DieBaka(u: AnsiString);
  procedure DieUnsupportedGLMode;
//  Procedure DieSilently;

  procedure DieBySEhHack;
  function SehHackTellExceptionAddress(a: ptruint): UnicodeString;
//  procedure PassMotherCallbackError(w: UnicodeString);
  function GetExceptionState(T: TThreadId = 0): PMotherSehState;
  function StopDying(): UnicodeString;


  Procedure AddLog(U: UnicodeString); overload;
  Procedure AddLog(S: AnsiString); overload;
  Procedure AddLog(mID: TMessageID); overload;
  Procedure AddLog(mID: TMessageID; Param: array of const); overload;
  Procedure AddLog(U: UnicodeString; Param: array of const); overload;
  Procedure AddLog(S: Ansistring; Param: array of const); overload;

  procedure DbgSay(w: UnicodeString);

  procedure SetConfigInt (section, id: AnsiString; Value: integer);
  function GetConfigInt (section, id: AnsiString): integer;
  
  function VersionToStr(major, minor, build: integer): AnsiString;
  function TellException(E: Exception): UnicodeString;
  function ExpExpAddress(addr: pointer): UnicodeString;
  function SehNameByCode(pes: PMotherSehState): UnicodeString;
  function ExplainCallStack(steps_to_ignore: integer; max_steps: integer = 9999): UnicodeString;
  function ShortExpAddr(addr: pointer): UnicodeString;
  function ToldMemRgn(ptr: pointer): UnicodeString;

  function HashToHex(h: TAssetHash): UnicodeString;
  function SetToStr(pset: pointer; BasicEnumTypeInfo: ptypeinfo): UnicodeString;

  function ParseHotKeyString(s: UnicodeString): TKeySetArray;
  function GetOneHotKeyName(var s: TKeySet): UnicodeString;
  function HotKeyStringDescription(def: TKeySetArray): UnicodeString;

//  function UnmangleFileName(s: TFileNameString): TFileNameString;
//  function MangleFileName(s: TFileNameString): TFileNameString;

  operator = (a, u: TAssetHash) b: System.boolean;
  
//Exception handling:
//  procedure TryWrap(p: Tprocedure); cdecl;
  procedure AfterEfCheck; cdecl;
//  procedure MySafeCallErrorProc(error : HResult; addr : pointer);
  var
    OldSafeCallErrorProc: TSafeCallErrorProc = nil;
    //SafeCallExceptionCaught: boolean = false;
    gv_ThreadsyncCS: TCriticalSection;

//OpenGL:
  procedure InitGLProcAddresses(Load2: system.boolean = false);
  procedure LoadOpenGLFunctions;
  procedure SwitchToFixedPipeline;

  procedure SetGlStatesFor2dRender;

//OpenAL:
  {$ifdef mo_uses_openal} //normally, the game modules do NOT use OpenAL.
   procedure InitALProcAddresses;
  {$endif}
  
//Timer
{$ifdef windows}
  //function Now(): TDateTime; //more precise version, with real resolution 1 ms
{$endif}
//  procedure Sleep(miliseconds: float);

  procedure PlaySound(wawfilename: ansistring);
const
  MAX_STACK_FRAMES_DUMP = 15;

  function GetChepersyErrorLog: UnicodeString;
  function GetChepersyWarnings: UnicodeString;

  function IsHashZero(h: TAssetHash): boolean;
  procedure ZeroHash(var h: TAssetHash);

  {$ifdef cpu64}
  function max(a, b: qword): qword; inline; overload;
  {without this, compiler stops with
     Error: Can't determine which overloaded function to call
     as there is no qword version of the function}
  {$endif}

  procedure ScreenDllEntryPointException;

 {$include un_threadloadind.h}

implementation

uses
  {$ifndef cpu32}
    lineinfo,
  {$endif}
  math, zstream, chepersy, un_strings, mo_module, mo_classes, mo_threads, mo_gl_quirks;

var
  dwarf: pointer;
  dwarfsize: integer;  //debugging line info extracted from the dll itself
  gv_LogCS: TCriticalSection; //TRTLCriticalSection;


  {$define impl}

  {$include un_api_common.inc}
  
  {$ifndef glesonly} //for debugging purposes only (compatibility testing): cut off all support for regular opengl
    {$include un_gl_from_fpc_2_6_4.inc}
  {$endif}
  {$include un_gles_from_fpc_2_6_4.inc}

  {$include un_talesteller.inc}

  {$include un_fileutil.inc}

  {$include mo_hub_1.inc}

  {$include un_typedefs.inc}

  {$include mo_hub_die.inc}

  {$include un_unicode.inc}
  
  {$include mo_timer.inc}

  {$ifdef windows}
    {$include un_errmode.inc}
  {$endif}

  {$include un_lineinfo.inc}

  {$include un_threadloadind.inc}

  {$ifdef fix_fpc3_unicode}
     {$include un_fpc_3-0-0_fix.inc} // Use custom mods of TFileStream, TStringList and TIniFile
  {$endif}
  {$if defined(windows) and (FPC_FULLVERSION<30000) }
     {$include un_fpc_2-6-4_fix.inc} // Use fixed TFileStream and copy-pasted everything that depends on it
  {$endif}


  procedure GrandInitialize(p: PMother; m,n,b: integer; logparams: array of const; sspl: integer);
  var
  {$ifdef fpc}
    Mm: TMemoryManager;
  {$endif}  
    dwlname: string;
    pes: PMotherSehState;
  begin
    Mother:= p;
    SessionSpaceLimit:= sspl;
   {$ifdef windows}
    //БЯКА! Устаревшая. winUnicodeStringalloc:= false; //UnicodeString becomes COM - incompatible but works faster
   {$endif}
//byte(nil^):=0;
   {$ifdef fpc}
 //    GetMemoryManager(Mm);
 //    Mm.NeedLock:=true;   //does it help?.. I do not use any threads...
  //   SetMemoryManager(Mm);

   {$endif}

   FormatSettings.DecimalSeparator:= '.';
   //DecimalSeparator:= '.'; //correceted jul 2014

    InitUnicode;

    InitTimer;

    gv_LogCS:= TCriticalSection.Create;
    gv_ThreadsyncCS:= TCriticalSection.Create;

    try
      Mother^.Core.ValidateVersion(m, n, b); AfterEfCheck;
      Mother^.Core.ValidateTypesets(sizeof(PFileNameChar^), 6); AfterEfCheck;
      Mother^.Core.ValidateTypesets(sizeof(Mother^), 5); AfterEfCheck;
      Mother^.Core.ValidateTypesets(ord(MI_CORE_MIDs_END), 0); AfterEfCheck;
      Mother^.Core.ValidateTypesets(ord(Re_OnUnload){Re_Stub)}, 1); AfterEfCheck;
//      Mother^.Core.ValidateTypesets(sizeof(TiCharArray), 2); AfterEfCheck;
      Mother^.Core.ValidateTypesetsLe('TOsType', ord(Mother^.State.OS), ord(high(TOSType))); AfterEfCheck;
//      Mother^.Core.ValidateTypesetsLe('TMotherState', sizeof(TMother), Mother^.dwSize); AfterEfCheck;

      AddLog(MI_VERSION_INFO, logparams);
      
      {$ifdef cpu32}
        SetLength(LineInfoPaths, 1);
        LineInfoPaths[0]:= PFileNameCharToFileNameString(Mother^.State.DebugInfoPath);
        if true //Mother^.State.DeveloperMode
        then begin
        //  AddLog(RuEn('Загрузка информации для самоотладки...','Loading the self-debugging info...'));
//          if not InitLineInfo(nil, true) then Die(MI_DEBUG_INFO_NOT_FOUND, [_LineInfoError]);
  {        if not InitLineInfo(nil)
              then Die(MI_ERROR_PROGRAMMER_NO_BAKA, [RuEn(
                 'Информация о номерах строк в формате dwarf 2 не найдена.'#10#13
                +'  (' + LineInfoError + ')'#10#13
                +' Убедитесь, что модуль был откомпилирован с ключом -gw.',
                 'The line number info in dwarf 2 format not found'#10#13
                +'  (' + LineInfoError + ').'#10#13
                +' Make sure the module is compiled with the -gw key.')]);}
        end;
      {$endif}
      {$ifdef windows}
        if Mother^.Debug.Verbose then AddLog('Error Mode = %0 before calling SetErrorMode(SEM_NOGPFAULTERRORBOX or SEM_FAILCRITICALERRORS or SEM_NOOPENFILEERRORBOX)'
                                                              , [IntToHex(clGetErrorMode(), 1) + 'h']);
        clSetErrorMode(SEM_NOGPFAULTERRORBOX or SEM_FAILCRITICALERRORS or SEM_NOOPENFILEERRORBOX);
        if Mother^.Debug.Verbose then AddLog('Error Mode = %0', [IntToHex(clGetErrorMode(), 1) + 'h']);
      {$endif}

//Moved to the control logic thread. See TLogic.OnCreate()
//      InitCgeClasses;
//      ClassesRegistrationStart;

      if Mother^.State.IsServer then begin
        if Mother^.Debug.Verbose then AddLog('Server mode, skipping initialization of OpenGL.');
      end
      else begin
        if ((Mother^.GAPI.VersionHi < 2)
          or ((Mother^.GAPI.VersionHi = 2) and (Mother^.GAPI.VersionMed < 0)))
        //and Mother^.ModuleRequiresOpenGL2
          then Die(MI_OPENGL_VERSION_TOO_LOW,
                    [2,
                     0,
                     MsgRaw(MI_HINT_OPENGLTOOLOW),
                     Mother^.GAPI.VersionString,
                     Mother^.GAPI.Renderer,
                     Mother^.GAPI.Vendor]);

        InitGLProcAddresses(true);//Mother^.ModuleRequiresOpenGL2);
        LoadOpenGLFunctions;
        {$ifdef mo_uses_openal}
          InitALProcAddresses;
        {$endif}
      end;

      if Mother^.Debug.Verbose then AddLog('Opening or creating config file... ');
      Config:= TIniFile.Create(string(Mother^.State.HomePath) + 'conf/' + string(Mother^.Module.PureName) + '.ini');
      if Mother^.Debug.Verbose then Mother^.Core.AddLogOK;

      if Mother^.Display.CanRender then TestOpenGlQuirkNo2;
    except
      pes:= GetExceptionState();
      if pes^.NowDying or not (ExceptObject is Exception)
        then Die ('Крах в GrandInitialize()', 'Crashed in GrandInitialize()')
        else Die(RuEn(
                'Крах в GrandInitialize(): непойманное исключение,'#10#13'  %0',
                'Crashed in GrandInitialize() with an uncaught exception,'#10#13'  %0')
                , [TellException(ExceptObject as Exception)]);
    end;
  end;

  function MsgRaw(M: TMessageId): UnicodeString;
  var
    AlreadyDying: boolean = false;
  begin
    if M > MI_CORE_MIDs_END then begin

      //Ask the local message manager. NOT IMPLEMENTED YET.

    end
    else begin
      AlreadyDying:= GetExceptionState()^.NowDying;

      Result:= PUnicodeCharToUnicodeString(Mother^.Core.MsgToString(M));
      if not AlreadyDying then AfterEfCheck; //this function can be a part of error processing
    end;
  end;

  function MsgFormat(M: TMessageId; Param: array of const): UnicodeString;
  begin
    Result:=PervertedFormat(MsgRaw(M), param);
  end;


  function PCharToString(P: PAnsiChar): AnsiString;
  var
    i: integer;
    p2: PAnsiChar;
  begin
    if not Assigned(p) then Result:=''
    else begin
      p2:=p;
      i:=0;
      While p2^ <> #0 do begin
        inc(p2);
        inc(i);
      end;
      SetLength(Result, i);
      Move(p^, Result[1], i);
    end;
  end;

  function PUnicodeCharToUnicodeString(P: PUnicodeChar): UnicodeString;
  var
    i: integer;
    p2: PUnicodeChar;
  begin
    if not Assigned(p) then Result:=''
    else begin
      p2:=p;
      i:=0;
      While p2^ <> #0 do begin
        inc(p2);
        inc(i);
      end;
      SetLength(Result, i);
      Move(p^, Result[1], i*2);
    end;
  end;

  function PFileNameCharToFileNameString(P: PFileNameChar): TFileNameString;
  var
    i: integer;
    p2: PFileNameChar;
  begin
    if not Assigned(p) then Result:=''
    else begin
      p2:=p;
      i:=0;
      While p2^ <> #0 do begin
        inc(p2);
        inc(i);
      end;
      SetLength(Result, i);
      Move(p^, Result[1], i * SizeOf(p^));
    end;
  end;
  
{  function HashToString(phash: pbyte): ansistring;
  var i: integer;
  begin
    Result:='';
    For i:=0 to 15 do begin
      Result:=Result + IntToHex(phash^, 2);
      if i < 15 then Result:=Result + '-';
      phash:=phash + 1;
    end;
  end;
}
  procedure SetConfigInt (section, id: AnsiString; Value: integer);
  begin
    Mother^.Config.SetInt(PAnsiChar(section), PAnsiChar(id), value);
  end;
  
  function GetConfigInt (section, id: AnsiString): integer;
  begin
    Result:=Mother^.Config.GetInt(PAnsiChar(section), PAnsiChar(id));
  end;

  function VersionToStr(major, minor, build: integer): AnsiString;
  begin
    Result:=format('%d.%.2d.%.4d' ,[Major, minor, build]);
  end;

  function RuEn(Ru, En: UnicodeString): UnicodeString;
  begin
    if Mother^.State.IsRussian then Result:=Ru else Result:=En;
  end;



  {$ifdef windows}
//Worked around in 2010 -- see cl_seh_hack.pp
(*  procedure TrybackWrapper(p: pointer); cdecl;
  begin
    {Backwrapper is called by the mother module's wrapper
      to filter out the game module's language exceptions
      that would confuse the mother module wrapper.
     The OS exceptions cannot be caught in a DLL
       (see bug #4605 at http://bugs.freepascal.org/view.php?id=4605 ),
       so they'll go straight throgh tis try  block
       to be caught by the mother module wrapper.  }
    try
      TProcedure(p)();
    except
      if not Mother^.State.NowDying then begin
        _dyell:='';
        CheckForGenericDyingYells;
        if Mother^.State.DyingAfterTrueException
          then _dyell:= PervertedFormat(RuEn('При вызове %0','During call to %0'),
                                          [ExpExpAddress(p)]) +#10#13#10#13 +  _dyell;
        Mother^.Core.DyingYellW(PUnicodeChar(_dyell));
        Mother^.Core.SetDying(yes);
      end;
    end;
  end; *)

{$endif}

(*
  procedure TryWrap(p: Tprocedure); cdecl;
  begin
    CgeTryWrap(@TrybackWrapper, @p);
    if Mother^.State.NowDying and Mother^.State.DyingAfterTrueException
      then Die(RuEn('При вызове %0',  'During call to %0'), [ExpExpAddress(@p)]);
  end;
  {$else}
  procedure TryWrap(p: Tprocedure); cdecl;
  begin
    try
      p
    except
      if not Mother^.State.NowDying
        then Die(RuEn('При вызове %0', 'During call to %0'),  [ExpExpAddress(@p)]);
    end
  end;
  {$endif} *)

 // GET RID OF THIS OBSOLETE STUFF!  ***************************************************

 { procedure TryWrap(p: Tprocedure); cdecl;
  begin
    try
      p
    except
      if not Mother^.State.NowDying
        then Die(RuEn('При вызове %0', 'During call to %0'),  [ExpExpAddress(@p)]);
    end
  end; }
{
function UnmangleFileName(s: TFileNameString): TFileNameString;
begin
  Result:= PFileNameCharToFileNameString(Mother^.FileManager.UnmangleFileName(PFileNameChar(s)));
end;

function MangleFileName(s: TFileNameString): TFileNameString;
begin
  Result:= PFileNameCharToFileNameString(Mother^.FileManager.MangleFileName(PFileNameChar(s)));
end;
}

operator = (a, u: TAssetHash) b: System.boolean;
begin
  b:=
     (a.Q1 = u.Q1) and (a.Q2 = u.Q2) and (a.Q3 = u.Q3) and (a.Q4 = u.Q4)
end;


function ParseHotKeyString(s: UnicodeString): TKeySetArray;
var
  a, b: UnicodeStringArray;
  v, u: UnicodeString;
  i, j, n: integer;
  k: TKey;
  fuck: boolean;
begin
  Result:= nil;
  a:= Explode(',', s, true);
  for j:= 0 to High(a) do begin
    if a[j] = '' then begin
      if Mother^.Debug.Verbose then AddLog('NOTICE: Empty position #%0 in the hotkey definition "%1"',[j, s]);
      continue;
    end;
    SetLength(Result, Length(Result) + 1);
    b:= Explode('+', a[j], true);
    for i:= 0 to high(b) do begin
      v:= UpperCase(b[i]);
      fuck:= true;
      for k in TKey do begin
        u:= uppercase(GetEnumName(typeinfo(TKey), ord(k)));
        if (v = u) or ('KEY_' + v = u) then begin
          Include(Result[High(Result)], k);
          fuck:= false;
          break;
        end;
      end;
      if fuck then begin
        Die(RuEn(
          '"%0" - нет такой клавиши.'#10#13'  (в строке настроек "%1")'#10#13'Все доступные значения перечислены в  файле'#10#13'  %2modules/chentrah/src/un_keys.h',
          'There is no such key as "%0"'#10#13'  (in config string "%1")'#10#13'All valid key names are listed in the file'#10#13'  %2modules/chentrah/src/un_keys.h'
          ),[v, s, Mother^.State.InstallPath]);
       // Die(MI_INVALID_HOST_HOTKEY_NAME, [u, Mother^.State.InstallPath]);
      end;
    end;
  end;
end;

function GetOneHotKeyName(var s: TKeySet): UnicodeString;
var
  k: TKey;
  a: UnicodeStringArray;
  w: UnicodeString;
begin
  Result:='';
  for k in s do begin
    if Mother^.State.IsRussian and (KeyNames[k, 0] <> '') then w:= KeyNames[k, 0]
    else begin
      if KeyNames[k, 1] <> ''
        then w:= KeyNames[k, 1]
        else w:= COPY(GetEnumName(typeinfo(TKey), ord(k)), 5, MaxInt);
    end;
    SetLength(a, Length(a) + 1);
    a[High(a)]:= w;
  end;
  Result:= Implode('+', a);
end;


function HotKeyStringDescription(def: TKeySetArray): UnicodeString;
var
  i: integer;
  w: UnicodeStringArray;
begin
  if Length(def) < 2 then Exit(GetOneHotkeyName(def[0]));
  SetLength(w, Length(def));
  for i:= 0 to High(def) do w[i]:= GetOneHotkeyName(def[i]);
  Result:= Implode(RuEn(' или ', ' or '), w);
end;

function GetChepersyErrorLog: UnicodeString;
var i: integer;
begin
  Result:='';
  for i:= CpsError.Count - 1 downto 0 do
    Result+= Utf8Decode(CpsError[i]);
end;

function GetChepersyWarnings: UnicodeString;
var i: integer;
begin
  Result:='';
  for i:= CpsParserWarnings.Count - 1 downto 0 do
    Result+= Utf8Decode(CpsParserWarnings[i]);
end;

 procedure SetGlStatesFor2dRender;
 begin
   glDisable(GL_TEXTURE_2D);
   glEnable(GL_ALPHA_TEST);
   glEnable(GL_BLEND);
   glAlphaFunc(GL_GREATER, 0);
   glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
   glDisable(GL_CULL_FACE);
   glDisable(GL_DEPTH_TEST);
   glViewport(0, 0, Mother^.Display.WindowClientRect.Width, Mother^.Display.WindowClientRect.Height);
   glMatrixMode(GL_PROJECTION);
   glLoadIdentity();
   glMatrixMode(GL_MODELVIEW);
   glLoadIdentity();
   //origin in the upper left corner!!
   glOrtho(0, Mother^.Display.WindowClientRect.Width, Mother^.Display.WindowClientRect.Height, 0, 0 ,1);
 end;

 var eFixedPipelineEmulationProgram: GLuint;

 procedure SwitchToFixedPipeline;
 begin
   case Mother^.GAPI.Mode of
     gapi_GL21, gapi_DeprecatedImmediateGL : begin
       if Assigned(glBindFramebuffer)
         then glBindFramebuffer(GL_FRAMEBUFFER, 0); //switch from FBO to the window framebuffer
       if Assigned(glUseProgram)
         then glUseProgram(0); //shut down the shaders
     end;
     gapi_GLES2: begin
       if Assigned(glBindFramebuffer)
         then glBindFramebuffer(GL_FRAMEBUFFER, 0); //switch from FBO to the window framebuffer
        if Assigned(glUseProgram)
         then glUseProgram(eFixedPipelineEmulationProgram);
     end;
     else
       DieUnsupportedGLMode;
   end;
 end;

  {$ifdef cpu64}
  function max(a, b: qword): qword;
  {without this, FPC 2.6.4 stops with
     Error: Can't determine which overloaded function to call
     as there is no qword version of the function}
  begin
    if a > b then Result:= a else Result:= b;
  end;
  {$endif}

  procedure ScreenDllEntryPointException;
  var pes: PMotherSehState;
  begin
    pes:= GetExceptionState;
    if Assigned(pes) then pes^.NowDying:= false;
  end;

end.


