{
    This file is part of Chentrah,
    Copyright (C) 2004-2015 Anton Rzheshevski (chebmaster@mail.ru).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see http://www.gnu.org/licenses/

 **********************************************************************}


  {$include mo_globaldefs.h}
  {$typeinfo on}

unit mo_logic;

interface
uses
  SysUtils, Classes, math, {$ifdef windows} Windows, {$endif}
   typinfo, chepersy, mo_classes, mo_hub,
    mo_assets, mo_filemanager, mo_texture,
  {mo_game,} mo_gui,
  mo_sessionstack{, mo_indexer, mo_text}
    ,zstream
//,mo_sortedchain
//, mo_rollback
, mo_threads
, mo_loba
  , mo_viewport
  , mo_camera
  , mo_image // for screenshot making task class

//, mo_ga_basics
   , syncobjs
 ;  
 
  procedure RegisterStandardLogicClasses;

type
  { Is reset to the lowest value before the Pulse. When set by the
    NextPulseWhen property, they increase in an ascending order (never go down).
    Thus, *any* subroutine could meddle with this setting during the pulse.
  }
  TLogicNextPulseWhenEnum = (
    lnf_WaitForRender,
    lnf_Steady100Hz,
    lnf_ASAPRenderAt30
  );

  TKeyboardState = array[TKey] of longbool;

  TPenState = packed record
    //x, y, //No need to. Uses mousex, mousey.
    pressure,
    xTilt, yTilt, Rotation: float;
  end;

  TJoystickPosition = packed record
    x, y, dx, dy: float;
  end;
  TJoystickState = array of TJoystickPosition;
  TJoysticksStates = array[TJoystickType] of TJoystickState;
  TJoyType = record
    case integer of
    0: (trec: TJoystickType);
    1: (tnum: integer);
  end;       
   
  TLogic = class (TCgeObject)
  protected
    f_firstRun: boolean;
    f_FrameNum: int64;
    f_runhash: TAssetHash;
    f_FrameTime: float;
    DeleteOnUnload: TArrayOfChepersyObjects;
    InputEvents: array of PInputEvent;
    InputCS, InterlockCS: TCriticalSection;

    procedure _set_When(v: TLogicNextPulseWhenEnum); inline;

    function _get_pressed_keys: TKeySet;
    procedure RegisterJoystickMovement(t: TJoystickType; _x, _y, _dx, _dy: float);

    class var
      // see mo_assets.pas for more detailed comments
      NecessaryAssetsList,
      LastNecessaryAsset: TGenericAsset;
      OptionalAssets: TSortedList; (* different approaches. Each optional asset
        is tried only once per frame, so any assets dependent on others
        will be carried over to the next frame even if there is enough time
        to process them.

        The necessary assets, on the other hand, would be kept processing
        until timeout (or until each remaining is processed 5 times).
      *)

      InitialAssetsNum, //Only the necessary assets are counted here
      CurrentAssetsNum: integer;
      
    class function ActualizeAssets(TimeLimitMicroseconds: integer = 10000): boolean;
      { This function is called repeatedly each pulse
        While it returns false, the logic displays a loading scren instead of
        its normal operation, and the pulse is not performed.
       }
   public
    mousex, mousey, //GUI pointer coordinates  (not grabbed)
    mousefpsdx, mousefpsdy //First person delta (grabbed)
      : float;
    DisplayWidth, DisplayHeight: integer;
    InsertMode,
    RestorePressedKeysStateAfterLoading: boolean;

    TextInput: UnicodeString; //the text input messages add to this
    PenState: TPenState;
    PrevFrameTSC: qword; //here be fields aligned to the 64-bits boundary
    AccumulatedLag: double;
    JoysticksStates: TJoysticksStates;
	
    KeyDown: TKeyboardState;

 	  ThreadedTasks: TThreadedTasks;
    FileManager: TFileManager;

    Control,   { The game world is connected and invoked using a
      common control like a whole-screen viewport, somewhere down the chain. }

    MainMenu, { The main menu functions even if Control is not
      (mainly, during the necessary assets actualization stage which is
      called "Loading") }

    AlternateControl { The loading progress screen or the waiting for background
      tasks to terminate screen.
      If this is not NIL, Control is frozen.
      If this is not TLoadingScreen, MainMenu is frozen too
        (frozen means also not rendering)
      This field is skipped, so classes used here do not need to be registered }
      : TControl;

    SessionId: AnsiString;

    Game: TChepersyObject;//TAbstractGame;
  protected
    f_Crashed: boolean;
    f_NextPulseWhen: TLogicNextPulseWhenEnum;
    f_saved: boolean;

    f_textInput, f_grabPointer: boolean;

    procedure SetTextInput(state: boolean);
    procedure SetGrabPointer(state: boolean);
  public
    property NextPulseWhen: TLogicNextPulseWhenEnum read f_NextPulseWhen write _set_When;
    property TextInputMode: boolean read f_textInput write SetTextInput;
    property GrabPointer: boolean read f_grabPointer write SetGrabPointer;


    property Crashed: boolean read f_Crashed;

    constructor Create; virtual;
    class procedure CreateOrLoad; virtual; //calls constructor or  loads session
	  destructor Destroy; override;

    procedure RegisterFields; override;
    procedure AfterLoading; override;
    procedure BeforeSaving; override;     	

    procedure InvokeMenu(mclacc: CControl); virtual;      
    class procedure AddNonActualizedAsset(v: TGenericAsset); //Can be called *before* the global "Module" variable is set
    procedure AddThreadedTask(task: TAbstractThreadedTask);
    function GetThreadedTask(kinds: TThreadedTaskTypes): TAbstractThreadedTask; //to be used by thread manager only
    function IsLoadingAssets: boolean; //to suppress specific reactions of some controls resizing (like TViewport)
    property FrameNum: int64 read f_FrameNum;
	
    procedure PushInputEvent(e: PInputEvent);
    function UnshiftInputEvent(): PInputEvent; //the caller is responsible for freeing the event
    procedure ProcessMessages();

    procedure OnCreate; virtual;
    procedure OnDestroy; virtual;

    function FileManagerPaths: TFileNameStringArray; virtual;  //invoked by TFileManager.Actualize

    procedure SwitchSession(newId: AnsiString);
    procedure ForkSession(newId: AnsiString);
    procedure SaveSession;
    procedure EraseCurrentSession; //causes a switch to session 'default';

    procedure RollbackSession; //causes the module to restart
    function GetSessionsList: TArrayOfAnsiString; {will crash gleefully
                           if it finds a folder with a name it doesn't like >:)
      Not to mention it's slow, parses the session root folder
                                        for the current module for folder names
    }
       
    procedure OnPress(key: TKey); virtual;
    procedure OnRelease(key: TKey); virtual; { These two are called
      by ProcessMessages(), they dispatch events to MainMenu and Control. }


    property PressedKeys: TKeySet read _get_pressed_keys;
    function CheckHotkey(k: Tkey; def: TKeySetArray): boolean;

	  function Render: boolean;//is called from the main thread
		  //thus synchronizing the threads -- see TThreadManager.CallLogicForRender
		  //return true if succeeded, false causes the module to draw a black "loading" screen.
		  //NOTE: assets are actualized inside this, meaning they are *always* actualized
		  //in the main thread!

	  procedure Pulse;//is always called from the logic thread.

	  property Saved: boolean read f_saved;

    procedure RenderBusyScreenInMainThread; virtual;

    function MicrosecondsLaggingBehind: double;
    procedure FromMainThreadStallIfGoingTooFast;
      //used by the dummy thread manager when lnf_Steady100Hz

    procedure EmergencySaveSession;

 end;
   
   CLogic = class of TLogic;
 
 
implementation
  uses mo_module;
  
  procedure RegisterStandardLogicClasses;
  begin
    {$define impl}
    {$include mo_gtypes.inc}

    RegClass(MyLogClass);
    RegisterAssetClasses;
    RegClass(TViewPort);
    RegClass(TCamera);

    RegisterStandardMenuClasses;
  end;        

 {$typedaddress on}

  procedure OnUnloadWalkProc(o: pointer);//TChepersyObject);
  begin
    if (TObject(o) is TCgeObject) then //imma idiot. without this, EInvalidCast: Invalid type cast
      if @((TChepersyObject(o) as TCgeObject).OnModuleDestroy) <> @(TestInstanceOfCgeObject.OnModuleDestroy)
        then (TChepersyObject(o) as TCgeObject).OnModuleDestroy;
    //if not o.Scraped then o.Scrape; //it's a total flush. The garbage collector is then forced to flush all
  end;    	
	
	destructor TLogic.Destroy;
	var i: integer;
	begin
		if Mother^.Debug.Verbose then AddLog('  Walking the graph to notify the objects of module destruction...');
		CpsWalkGraph(Self, @OnUnloadWalkProc);
		OnDestroy;
		if Module.Logic = Self then begin
		  //if Mother^.Debug.Verbose then AddLog('  Forcing garbage collector flush...');
		  //GarbageCollector.ForceFlush;
		  Module.Logic:= nil;
		end;
		try InputCS.Free; InterlockCS.Free; except end;
		inherited;
	end;

  procedure TLogic.InvokeMenu(mclacc: CControl);
  begin
    if Assigned(MainMenu) then MainMenu.Kill;
    MainMenu:= mclacc.Create;
  end;

  procedure TLogic.RegisterFields;
  begin
    RegisterStandardLogicClasses;

    RegClass(TControl);
//    RegClass(TStatic2dTexture);
//    RegClass(TDynamic2dTexture);
    RegClass(TTexture);
    RegClass(TModuleMainMenu);
    RegClass(TSortedList);
    RegClass(TFilemanager);
//    RegClass(TGuiSkin);
    RegType(typeinfo(TKeyboardState), TypeInfo(longbool), [0, high(TKeyboardState)]);
    RegType(TypeInfo(TPenState), SizeOf(TPenState), [
      //'x','y',
      'pressure','xtilt','ytilt','rotation',TypeInfo(float)]);
    RegType(TypeInfo(TJoystickType));
    RegType(TypeInfo(TJoystickPosition), Sizeof(TJoystickPosition), ['x','y','dx','dy',TypeInfo(float)]);
    RegType(TypeInfo(TJoystickState));//, TypeInfo(TJoystickPosition));
    RegType(TypeInfo(TJoysticksStates), TypeInfo(TJoystickState), TypeInfo(TJoystickType));
    RegType(TypeInfo(TThreadedTasks), typeinfo(pointer), typeinfo(TThreadedTaskType));//TypeInfo(TThreadedTaskType));
    RegType(TypeInfo(TLogicNextPulseWhenEnum));

//    RegClass(TAbstractGame);

    inherited;
    ListFields([
      'f_firstrun', @f_firstrun, TypeInfo(boolean),
      'f_FrameNum', @f_FrameNum, TypeInfo(qword),
      'f_runhash', @f_runhash, typeinfo(TAssetHash),
      '-f_FrameTime', @f_FrameTime, TypeInfo(float), //no saving amidst the cycle
      'DeleteOnUnload', @DeleteOnUnload, typeinfo(TArrayOfChepersyObjects),
      '-InputEvents', @InputEvents, CPS_POINTER,
      '-InputCS', @InputCS, CPS_POINTER,
      '-InterlockCS', @InterlockCS, CPS_POINTER,
      'mousex', @mousex,
      'mousey', @mousey,
      'mousefpsdx', @mousefpsdx,
      'mousefpsdy', @mousefpsdy, TypeInfo(float),
      'DisplayWidth', @DisplayWidth,
      'DisplayHeight', @DisplayHeight,  TypeInfo(integer),
      'InsertMode', @InsertMode,
      'RestorePressedKeysStateAfterLoading', @RestorePressedKeysStateAfterLoading, TypeInfo(boolean),
      'TextInput', @TextInput, typeinfo(UnicodeString),
      'PenState', @PenState, typeinfo(TPenState),
      '-PrevFrameTSC', @PrevFrameTSC, typeinfo(qword),
      '-AccumulatedLag', @AccumulatedLag, typeinfo(double),
      'JoyStates', @JoysticksStates, typeinfo(TJoysticksStates),
      'KeyDown', @KeyDown, '*TKeyboardState',
      '-ThreadedTasks', @ThreadedTasks, TypeInfo(ThreadedTasks),
      'FileManager', @FileManager, TypeInfo(TFilemanager),
      'Control', @Control, TypeInfo(TControl),
      'MainMenu', @MainMenu, typeinfo(TControl),
      '-AlternateControl', @AlternateControl, typeinfo(TControl),
//      'GuiSkin', @GuiSkin, TypeInfo(TGuiSkin),
//      'TextureIndex', @TextureIndex, TypeInfo(TIndexer),


      'session_id', @SessionId, typeinfo(AnsiString),
      'game', @Game, CPS_MANAGED_OBJECT,
      '-f_Crashed', @f_Crashed, typeinfo(Boolean),
      '-f_NextPulseWhen', @f_NextPulseWhen, typeinfo(TLogicNextPulseWhenEnum),
      '-f_saved', @f_saved,
      'f_textInput', @f_textInput, 'f_grabPointer', @f_grabPointer, typeinfo(boolean)
     ]);
  end; 





  function TLogic.IsLoadingAssets: boolean;
  begin
    Result:= Assigned(NecessaryAssetsList);
  end;

  procedure TLogic.BeforeSaving;
  begin

  end;

  function TLogic.FileManagerPaths: TFileNameStringArray;
  begin
    SetLength(Result, 1);
    Result[0]:= '*M/'; //the module base path
  end;

  procedure TLogic.OnCreate;
  begin
    Self.f_runhash:= Module.SessionHash;
	  InputCS:= TCriticalSection.Create;
    InterlockCS:= TCriticalSection.Create;
  end;

  procedure TLogic.OnDestroy;
  begin

  end;

  constructor TLogic.Create;
  begin
    inherited;
    FileManager:= TFileManager.Create;
  end;

  class procedure TLogic.CreateOrLoad;
  var
    filename: AnsiString;
    fs: TFileStream;
    ds: TDecompressionStream;
    i: integer;
    w: UnicodeString;
	begin
    Module.NewSessionHash();

    if not HaveASessionToLoad then begin
      AddLog(RuEn(
        'Сооздаю сессию с нуля!',
        'Creating the session from scratch!'),[]);

      Module.Logic:= TLogic(MyLogClass.Create());

    end
    else begin

      try
        filename:= GetSessionFileName();
        if Mother^.Debug.Verbose then AddLog('Loading session: file "%0"',[filename]);
        RetrieveAssets;
        fs:= TFileStream.Create(filename, fmOpenRead + fmShareDenyWrite);
       // ds:= TDecompressionStream.Create(fs);
        w:= Ruen('сессии', 'session');

        Module.Logic:= TLogic(CpsLoad(w, fs)); {must write directly there
          because some objects assume to find it there in their AfterLoading()
          methods }

        if not Assigned(Module.Logic) then Die(GetChepersyErrorLog());
        if not (Module.Logic is TLogic) then Die(RuEn(
          'Корневой объект загруженной сессии (%0) не является потомком TLogic',
          'The root object of the loaded session (%0) is not a TLogic descendant'),
          [string(Module.Logic.ClassName)]
        );
        //ds.Free;
        fs.Free;

        if CpsParserWarnings.Count > 0 then AddLog(RuEn(
          '~ ПРЕДУПРЕЖДЕНИЯ ПРИ ЗАГРУЗКЕ СЕССИИ: ~'#10#13'%0'#10#13'~ КОНЕЦ ПРЕДУПРЕЖДЕНИЙ ~',
          '~ WARNINGS AT SESSION LOADING: ~'#10#13'%0'#10#13'~ END WARNINGS ~'),
          [GetChepersyWarnings()]);
        FinalizeSessionRollback;
        if not Assigned(Module) then Die(GetChepersyErrorLog());

        Module.Logic.SessionID:= mo_sessionstack.SessionId;
      except
        Die(RuEn(
          'Не удалось загрузить сессию',
          'failed to load the session'));
      end;
    end;


	end;


  procedure TLogic.AfterLoading;
  begin
    Module.Logic:= Self;
    Module.LoadedHash:= Self.f_runhash;

    OnCreate;

    //Not possible now, wrong thread. Moved this to the ActualizeAssets method
    //DealWithUnclaimedAssets; //DeleteUnclaimedAssets;

    if not RestorePressedKeysStateAfterLoading
      then FillChar(KeyDown, sizeof(KeyDown), 0); //set all keys' pressed state to false

    Mother^.Input.TextInput:= f_textInput;
    Mother^.Input.GrabPointer:= f_grabPointer;


    if Assigned(MainMenu) then begin
      MainMenu.Kill;
      //InvokeMenu;
    end;
  end;



  procedure TLogic.AddThreadedTask(task: TAbstractThreadedTask);
  var
    kind: TThreadedTaskType;
  begin
    EnterWv(gv_ThreadsyncCS);
    try
      //addlog('TModule.AddThreadedTask %0',[GetEnumName(TypeInfo(TThreadedTaskType), Ord(task.GetType()))])  ;
      kind:= task.GetType;
      if not Assigned(ThreadedTasks[kind]) then ThreadedTasks[kind]:= TSortedList.Create;
      ThreadedTasks[kind].Add(task);
      //addlog('TModule.AddThreadedTask end')  ;

    finally
      gv_ThreadsyncCS.Leave;
    end;
  end;

  //to be used by thread manager only
  //WARNING: this function executes NOT in the logic thread but in whichever thread wants to get a new task.
  function TLogic.GetThreadedTask(kinds: TThreadedTaskTypes): TAbstractThreadedTask;
  var kind: TThreadedTaskType;
//tt: TThreadedTaskType;
  begin
    EnterWv(gv_ThreadsyncCS);
//WTF AddLog here causes a silent crash-to-desktop with about 3% probability :(
//    for tt in kinds do addlog('TModule.GetThreadedTask %0',[GetEnumName(TypeInfo(TThreadedTaskType), Ord(tt))])  ;
    Result:= nil;

    try
      for kind in kinds do begin
        if not Assigned(ThreadedTasks[kind]) then continue;
        if ThreadedTasks[kind].Count = 0 then continue;
        //aaaand, crash in ntdll.dll
        //then begin
        //  ThreadedTasks[kind].Scrape;
        //  pointer(ThreadedTasks[kind]):= nil;
        //  continue;
        //end;
        Result:= (ThreadedTasks[kind].ExtractFirst) as TAbstractThreadedTask;
      end;
  //addlog('TModule.GetThreadedTask: result %0',[Assigned(result)]) ;
    finally
      gv_ThreadsyncCS.Leave;
    end;
  end;

  //executes in: the main thread, during synchronization.
  class function TLogic.ActualizeAssets(TimeLimitMicroseconds: integer = 10000): boolean;
  type
    PGenericAsset = ^TGenericAsset;
  var
    q: qword;
    MicrosecondsPassed: double = 0.0;
    i, initialCAN: integer;
    L, T: PGenericAsset;
    a: TGenericAsset;
    oto: TThreadLoadKind;

    procedure Reschedule(c: TGenericAsset); inline; //moves the asset all the way to the back of the queue
    begin
      if Assigned(T^) then begin
        T^.NextInQueue:= c;
        T^:= c;
      end;
      if not Assigned(L^) then L^:= c; //it's the only asset
    end;

    function Pull: TGenericAsset; inline;
    begin
      Result:= L^;
      L^:= Result.NextInQueue;
      Result.NextInQueue:= nil;
      if not Assigned(L^) then T^:= nil;
    end;

    procedure ActualizeOptional; inline;
    begin
      //The optional assets are sorted by priority.
      i:= 0;
      Mother^.Timer.UsecDelta(@q);
      Self.OptionalAssets.Prune;
      Self.OptionalAssets.Sort; //They float up in priority the longer they wait
      // (see Module.FrameNum and TGenericAsset.InQueueSinceFrame)
      while (i < Self.OptionalAssets.Count) and (MicrosecondsPassed < TimeLimitMicroseconds) do begin
        a:= Self.OptionalAssets[i] as TGenericAsset;
        if not a.Actualized then try
          if Mother^.Debug.Verbose then addlog('  actualizing #%0/%2, %1',[i + 1, string(a.classname), Self.OptionalAssets.Count]);
          if Mother^.State.IsServer and not a.IsValidForServer then begin
            if Mother^.Debug.Verbose then addlog('    ..not valid in dedicated server mode, scraping.');
            Self.OptionalAssets.Remove(i);
            a:= nil;//.Scrape;
          end
          else begin
            if a.Actualize then begin
              Self.OptionalAssets.Remove(i);
              if Mother^.Debug.Verbose then addlog('    ..asset actualized');
            end else begin
              inc(i);
              //If the asset refuses to actualize, it means it is waiting
              //  for something it depends on.
              // For optional ones, it gets postponed to the next frame.
              if Mother^.Debug.Verbose then addlog('    ..asset refused.');
            end;
          end;
          MicrosecondsPassed+= Mother^.Timer.UsecDelta(@q);
        except
          Die(RuEn('Сбой при актуализации ресурса №%0 из %1 (%2)',
                   'Failed actualizing asset #%0 of %1 (%2)')
              ,[i + 1, Self.OptionalAssets.Count, string(a.ClassName)]);
        end
        else begin
          Self.OptionalAssets.Remove(i);
          if Mother^.Debug.Verbose then addlog('  asset %1 (#%0/%2) is already actualized',[i + 1, string(a.classname), Self.OptionalAssets.Count]);
        end;
      end;
    end;

    function ActualizeNecessary: boolean; inline;
    begin
      //There is no such thing as priority for the necessary assets, all are equally vital.
      Mother^.Timer.UsecDelta(@q);
      i:= 0;
      while Assigned(L^) and (MicrosecondsPassed < TimeLimitMicroseconds) and (i < initialCAN * 5 )
      do begin
        a:= Pull;
        inc(i);
//        if a.Scraped then begin
//          if Mother^.Debug.Verbose then addlog('  asset %1 (#%0/%2) is SCRAPED.',[i, string(a.classname), InitialAssetsNum]);
//          Dec(CurrentAssetsNum);
//          continue;
//        end;
        if a.Actualized then begin
          if Mother^.Debug.Verbose then addlog('  asset %1 (#%0/%2) is already actualized.',[i, string(a.classname), InitialAssetsNum]);
          Dec(CurrentAssetsNum);
          continue;
        end;
        if Mother^.State.IsServer and not a.IsValidForServer then begin
          if Mother^.Debug.Verbose then addlog('  asset %1 (#%0/%2) is not valid in dedicated server mode, scraping it.',[i, string(a.classname), InitialAssetsNum]);
          a:= nil;//.Scrape;
          Dec(CurrentAssetsNum);
          continue;
        end;
        try
          if Mother^.Debug.Verbose then addlog('  actualizing #%0/%2, %1',[i, string(a.classname), InitialAssetsNum]);

          if not a.Actualize
            then begin
              if Mother^.Debug.Verbose then addlog('    ..asset refused, rescheduling.');
              Reschedule(a);
            end
            else begin
              if Mother^.Debug.Verbose then addlog('    ..asset actualized');
              Dec(CurrentAssetsNum)
            end;
          //If the asset refuses to actualize, it means it is waiting
          // for something it depends on.
          // It coulsd be either a threaded task, or an another child asset.
          // anyway, we shuffle it to the back of the queue.

          //The worst case: all assets are depending on threaded tasks, so they would
          //  get rescheduled in an endless cycle until the designated time runs
          //  out, thus creating some wasted CPU load.
          //That's what the rule of "no more than five times the number of assets
          //  in the queue" is for.

          MicrosecondsPassed+= Mother^.Timer.UsecDelta(@q);
        except
          Die(RuEn('Сбой при актуализации ресурса №%0 из %1 (%2)',
                   'Failed actualizing asset #%0 of %1 (%2)')
              ,[i, InitialAssetsNum, string(a.ClassName)]);
        end
      end;
      Result:= not((L = @NecessaryAssetsList) and Assigned(L^));
      //false: not finished, continue. true: finished, remove the loading screen.
    end;
  begin
    oto:= NoteThreadOccupation(tlk_ActualizingAssets);
    if Mother^.Debug.Verbose then addlog('TLogic.ActualizeAssets...');
    if Assigned(NecessaryAssetsList) then begin
      if Mother^.Debug.Verbose then addlog('  using the list of necessary, remaining=%0',[CurrentAssetsNum]);
      L:= @NecessaryAssetsList;
      T:= @LastNecessaryAsset;
      initialCAN:= CurrentAssetsNum;
      Result:= false; //cause the loading screen
      if not Assigned(L) then begin
        if Mother^.Debug.Verbose then addlog('  The queue is empty.');
        Exit(true);
      end;
      Result:= ActualizeNecessary;
    end
    else begin
      if Mother^.Debug.Verbose then addlog('  using the list of optional');
      Result:= true; // do not cause the loading screen
      ActualizeOptional;
    end;
    if Mother^.Debug.Verbose then addlog('TLogic.ActualizeAssets end.');
    NoteThreadOccupation(oto);
  end;


  //executes in: the logic thread OR the main thread during synchronization
  class procedure TLogic.AddNonActualizedAsset(v: TGenericAsset);
  begin;
    if Mother^.Debug.Verbose then addlog('TLogic.AddNonActualizedAsset %0 (%1)...', [string(v.classname), v.IsNecessary]);
    if Assigned(v.NextInQueue) then Die('Attempt to add an %0 that is already in queue!', [string(v.classname)]);
    if v.IsNecessary then begin
      if not Assigned(NecessaryAssetsList) then begin
        NecessaryAssetsList:= v;
        LastNecessaryAsset:= v;
        InitialAssetsNum:= 0;
        CurrentAssetsNum:= 0;
      end
      else begin
        LastNecessaryAsset.NextInQueue:= v;
        LastNecessaryAsset:= v;
      end;
      Inc(InitialAssetsNum);
      Inc(CurrentAssetsNum);
    end
    else begin
      if not Assigned(OptionalAssets) then OptionalAssets:= TSortedList.Create;
      if Assigned(Module.Logic)
        then v.InQueueSinceFrame:= Module.Logic.FrameNum
        else v.InQueueSinceFrame:= 0;
      OptionalAssets.Add(v);
    end;
  end;



	function TLogic.Render: boolean; //is called from the main thread
		//when synchronizing the threads -- see TThreadManager.CallLogicForRender
		//return true if succeeded, false causes the module to draw a black "loading" screen.
		//NOTE: assets are actualized inside this, meaning they are *always* actualized
		//in the main thread!
    //Bottom line: this routine must be kept as short as possible!
  var
    HaveNonActualizedAssets: boolean;
    NeedsResize: boolean;
    oto: TThreadLoadKind;
    pes: PMotherSEHState;
	begin

    if Crashed then Exit(false);

    oto:= NoteThreadOccupation(tlk_MainThreradRender);
    //Assets are actualized here, as per conversion from the old paradigm
    //  where the logic was running in the main thread.
    //***TODO research possibility to move them out.
    //         Most probably taht would be  too costly.

    try
      try
        EnterWv(InterlockCS, tlk_WaitingSync);  //as a safety precaution.
        Module.ThreadManager.RenderDoneLogicMayProceed.ResetEvent;

        TAssetManager.Pulse;

        NeedsResize:= false;
        if Mother^.Display.CanRender
          and((DisplayWidth <> Mother^.Display.WindowClientRect.Width)
           or (DisplayHeight <> Mother^.Display.WindowClientRect.Height))
        then begin
          NeedsResize:= true;
          DisplayWidth:= Mother^.Display.WindowClientRect.Width;
          DisplayHeight:= Mother^.Display.WindowClientRect.Height;
        end;


        if Mother^.Module.State = ms_WaitingForModuleToFinish then begin
          //waiting for background threads to terminate before unloading
          if not Module.IsAlreadyWaitingForThreadsToTerminate then begin
            if Assigned(AlternateControl)
              and not (AlternateControl is TModuleStopingIndicator)
              then AlternateControl:= nil;//.Scrape;
            AlternateControl:= TModuleStopingIndicator.Create(); //everything is reset, utterly breaking the controls chain
            AlternateControl.Resize; //cannot include this in constructor, as it resizes the entire chain propagating down it
            Module.IsAlreadyWaitingForThreadsToTerminate:= Yes;
            //Threre is no way back from this state. The DLL is doomed to be unloaded.
          end;
          if Assigned(AlternateControl) and NeedsResize then AlternateControl.Resize;
          if Assigned(AlternateControl) and Mother^.Display.CanRender then AlternateControl.Render;
        end else begin
          if Assigned(AlternateControl)
            and (AlternateControl is TModuleStopingIndicator)
            then begin
              //AlternateControl.Scrape;
              AlternateControl:= NIL;
              //They maay have missed their resize
              if Assigned(MainMenu) then MainMenu.Resize;
              if Assigned(Control) then Control.Resize;
            end;
          HaveNonActualizedAssets:= false;

          if Assigned(NecessaryAssetsList) then begin
            if not Assigned(AlternateControl) or not (AlternateControl is TLoadingScreen) then begin
              AlternateControl:= TLoadingScreen.Create();
              AlternateControl.Resize;
              if Mother^.Debug.Verbose then AddLog('Created loading screen.');
            end;
          end;

          if NeedsResize then begin
            if Assigned(AlternateControl)
              then AlternateControl.Resize;

            //The main menu is allowed to work during loading
            if Assigned(MainMenu) and (not Assigned(AlternateControl)
              or (AlternateControl is TLoadingScreen))
              then MainMenu.Resize;

            if Assigned(Control)
              and not Assigned(AlternateControl)
              then Control.Resize;
          end;

          if Assigned(NecessaryAssetsList) then
            //between the pulse and render, on the off chance they will actualize in time
            //to avoid flickering the loading screen to life
            HaveNonActualizedAssets:= ActualizeAssets({***TODO: add microiseconds limit depending on QF});

          if Assigned(NecessaryAssetsList) then begin
            if not Assigned(AlternateControl)
              or not (AlternateControl is TLoadingScreen)
            then begin
              if Assigned(AlternateControl) then AlternateControl:= nil;//.Scrape;
              AlternateControl:= TLoadingScreen.Create();
            end;
            //(Control as TLoadingScreen).Progress:= InitialAssetsNum - CurrentAssetsNum / Math.max(1, InitialAssetsNum);
          end;

          if Mother^.Display.CanRender
            then try
              if Assigned(Control) and not Assigned(AlternateControl)
                then Control.Render; //the one who spawns the non-actualized assets

              if Assigned(AlternateControl)
                then AlternateControl.Render;

              if Assigned(MainMenu) and (not Assigned(AlternateControl)
                or (AlternateControl is TLoadingScreen))
                then MainMenu.Render;

              CheckGlError;
            except
              Die(RuEn(
                'Сбой в подпрограмме прорисовки',
                'Rendering subroutine failure'));
            end;
          if not Assigned(NecessaryAssetsList) then begin
            //return old control into action
            if Assigned(AlternateControl)
              and (AlternateControl is TLoadingScreen)
            then begin
              AlternateControl:= nil;//.Scrape;
              if Mother^.Debug.Verbose then AddLog('Destroyed the loading screen. The top control is now %0', [Control]);
              //They maay have missed their resize
              if Assigned(MainMenu) then MainMenu.Resize;
              if Assigned(Control) then Control.Resize;
            end;
            if Assigned(OptionalAssets) and (OptionalAssets.Count > 0) then ActualizeAssets;
          end;
        end;
      except
        Module.ThreadManager.RenderDoneLogicMayProceed.SetEvent; { Must avoid
          deadlock, as the module is likely to stop pulsing after this! }
        Die(RuEN(
          'Сбой в процедуре синхронизации потоков для прорисовки',
          'Failure in threads synchronization for render routine'));
      end;
    finally
      pes:= GetExceptionStateIfExists();

      Module.ThreadManager.LogicDoneRenderMayProceed.ResetEvent; { Otherwise,
        it may sqeeze one more frame before the next logic pulse if frames are
        short enough! Resulting in FPS > 100 which should NOT be possible.
      }

      InterlockCS.Leave;

      if f_NextPulseWhen <> lnf_WaitForRender
        then Module.ThreadManager.RenderDoneLogicMayProceed.SetEvent;
    end;
    NoteThreadOccupation(oto);
    Result:= true;
  end;

  procedure TLogic.EmergencySaveSession;
  begin
    try
      EnterWv(InterlockCS);
      SaveSession;
    finally
      InterlockCS.Leave;
      Mother^.Module.SessionNotSaved:= false;
    end;
  end;

  procedure TLogic.Pulse; //is always called from the logic thread.
  begin
    if Crashed then Exit;

    Mother^.Timer.UsecDelta(@PrevFrameTSC);

    f_NextPulseWhen:= Low(TLogicNextPulseWhenEnum);

    try
      EnterWv(InterlockCS);
      Module.ThreadManager.LogicDoneRenderMayProceed.ResetEvent;

      if Mother^.Module.SessionNotSaved and Mother^.Module.MustUnload
      then begin
        SaveSession;
        Exit;
      end;

        NoteThreadOccupation(tlk_PollingInput);
      try
        ProcessMessages; //calls OnKey for MainMenu and Control
      except
        Die(RuEn(
          'Сбой при рассылке сообщений ввода элементам интерфейса',
          'Error sending input events to GUI elements'));
      end;

        NoteThreadOccupation(tlk_Logic);
      if Mother^.State.IsServer then begin
        //Server mode: steady 100 Hz
        try
//          if Assigned(Game) then Game.Pulse(10);
        except
          f_Crashed:= true;
          Mother^.Module.SessionNotSaved:= false;
          Die(RuEn(
            'Крах игровой логики. Всё погибло.',
            'The game logic crashed. Everything is lost.'));
        end;
      end
      else begin
        //Client mode: the game is buried as one of the controls
        with Mother^.Display.Screenshot do if (request = 2) and Assigned(pdata) then begin
          if Mother^.Debug.Verbose then AddLog('Suddenly found screenshot data prepared by mother. Adding a background task to save it to a file.');
          TScreenshotMakeTask.Create;
          request:= -2;
        end;


        if Assigned(AlternateControl) then try
          AlternateControl.Cycle;
        except
          AlternateControl:= nil; { drop it, the pulse will be called again to save
            the session }
          Die(RuEn(
            'Крах логики служебного индикатора',
            'The service indicator logic crashed'));
        end;

        if not Mother^.Module.MustUnload then begin
          if Assigned(MainMenu) and (not Assigned(AlternateControl)
            or (AlternateControl is TLoadingScreen))
          then try
             MainMenu.Cycle;
          except
            MainMenu:= nil; { drop it, the pulse will be called again to save
              the session }
            Die(RuEn(
              'Крах логики главного меню',
              'The main menu logic crashed'));
          end;

          if Assigned(Control) then try
            Control.Cycle; { The game world sits in one
            of the controls down the chain }
            Mother^.Module.SessionNotSaved:= true;
            Inc(f_FrameNum);
          except
            f_Crashed:= true;
            Control:= nil;
            Mother^.Module.SessionNotSaved:= false;
            Die(RuEn(
              'Крах управляющей логики. Всё погибло.',
              'The control logic crashed. Everything is lost.'));
          end;

        end;

      end;
    finally
      Module.ThreadManager.LogicDoneRenderMayProceed.SetEvent;
      InterlockCS.Leave;
    end;
  end;

  procedure TLogic.RenderBusyScreenInMainThread;
  begin
    Mother^.Display.RenderBackground;
    AfterEfCheck;
  end;

  {$include mo_logic_sess.inc}

  {$include mo_logic_input.inc}

  procedure TLogic._set_When(v: TLogicNextPulseWhenEnum); inline;
  begin
    if v > f_NextPulseWhen then f_NextPulseWhen:= v;
  end;

  function TLogic.MicrosecondsLaggingBehind: double; //Only for "steady100Hz"
  var
    SaveTsc: qword;
  begin
    SaveTsc:= PrevFrameTSC;
    Result:= AccumulatedLag + Math.Min(1000000.0, Math.Max(-1000000.0,
      Mother^.Timer.UsecDelta(@PrevFrameTSC) - 10000.0));
    PrevFrameTSC:= SaveTsc;
  end;

  procedure TLogic.FromMainThreadStallIfGoingTooFast;
  var
    MilisecondsToSleep: integer;
  begin
    MilisecondsToSleep:= Math.Max(0, round(- MicrosecondsLaggingBehind / 1000));
    if MilisecondsToSleep > 2 then Mother^.Timer.HiResSleep(MilisecondsToSleep);
  end;

  procedure TLogic.SetTextInput(state: boolean);
  begin
    f_textInput:= state;
    Mother^.Input.TextInput:= state;
  end;

  procedure TLogic.SetGrabPointer(state: boolean);
  begin
    f_grabPointer:= state;
    Mother^.Input.GrabPointer:= state;
  end;

end.  
