  procedure TLogic.SwitchSession(newId: AnsiString);
  begin
    Module.SetConfigStr('SessionId', newId); //will not have effect until the next time the module is loading
    AddLog(RuEn('Переключение сессии с %0 на %1.','Switching the session from %0 to %1.'),[mo_sessionstack.SessionId, newID]);
    Module.Restart; //will eventually lead to current session being saved.
  end;

  procedure TLogic.ForkSession(newId: AnsiString);
  begin
    Module.SetConfigStr('SessionId', newId); //will not have effect until the next time the module is loading
    AddLog(RuEn('Сессия %0 форкнута как %1.','Session %0 forked as %1.'),[mo_sessionstack.SessionId, newID]);
    mo_sessionstack.SessionId:= newID;
    Module.Restart; // to actualize creating the folder
  end;

  procedure TLogic.EraseCurrentSession;
  begin
    AddLog(RuEn('Текущая сессия %0 помечена для удаления.','The current session %0 is marked for deletion.'),[mo_sessionstack.SessionId]);
    Module.GoingToEraseSession:= Yes; //causes the normal session saving routine to erase the session instead.
    SwitchSession('default');
  end;


  procedure TLogic.SaveSession;
{
***TODO:
  1. Save to a mother memory stream
  2. Have *the mother* dump that memory stream into a compressed file in a separate thread
  3. Profit from delay-less saving of even large data sets
}

  var
    tempfilename: ansistring;
    ms: TMemoryStream;
    ws: TStream;
    fs: TFileStream;
    nu: integer;
    cs: TCompressionStream;
    q: qword;
    sf, sm: integer;
  begin
    if not Mother^.Module.SessionNotSaved then begin
      AddLog(RuEn(
        'Сохранение сессии не нужно.',
        'There is no need to save the session.'));
      Exit;
    end;

    if Assigned(TAssetManager.ToStore)
      then TAssetManager.ToStore.ClearAndForget
      else TAssetManager.ToStore:= TAssetContainer.Create;
    if Mother^.Module.RollBackTheSessionOnNextLoad
    then begin
      AddLog(RuEn(
        'Сессия НЕ сохраняется, так как был запрошен её откат.',
        'Session is NOT saved because its rollback was requested.'), []);
      if Mother^.Module.NeedToKeepAssets() then begin
        WalkSessionToListAssetsToStore;
        StoreAssets;
      end;
    end
    else begin
      if Module.GoingToEraseSession then begin
        AddLog(RuEn(
                  'Стираю текущую сессию "%0".',
                  'Erasing the current session "%0".'), [mo_sessionstack.SessionId]);
        if Mother^.Module.NeedToKeepAssets() then begin
          WalkSessionToListAssetsToStore;
          StoreAssets;
        end;
        EraseSession(mo_sessionstack.SessionId);
        Mother^.Module.SessionNotSaved:= false;
        PlaySound('crashed.wav');
      end
      else try
        Mother^.State.SavingTheSessionNow:= true;
        tempfilename:= SessionFileNameWoExt + 'temporary.$$$';
        if Mother^.Debug.Verbose then AddLog('Saving session: temporary file name "%0"',[tempfilename]);
        MakeSurePathExists(ExtractFilePath(tempfilename));

//        if Mother^.Module.NeedToKeepAssets()
//          then

//            StoreResourceClassesList;
            //2014: weeded out a few *horrible* bugs in it

            {2008: After introducing OpenAL this critter
               started crashing craptacularly at the program exit, managing to
               lock the sound card and hang the chentrah process in a zombie state
               so that even sudo kill won't help.
              The only way to get everything working ahgain is restarting the X server.
              I suspect there's something wrong with the cross-module exception
               handling mechanism in the Linux version.}
        //ms:= TMemoryStream.Create;
        fs:= TFileStream.Create(tempfilename, fmCreate);

        if not CpsStore(
          Module.Logic, RuEn('сессии', 'session'), fs, VOLATILE_MASK, VOLATILE_MASK) //skip only volatiles
        then begin
          f_Crashed:= true;
          fs.Free;
          Mother^.State.StateTrashedRestartRequired:= Yes;
          Mother^.Core.SetAssetStorageMemoryStream(NIL); //erase the stored assets list
          Die(GetChepersyErrorLog());
        end;
        //Mother^.Timer.UsecDelta(@q);


        //
       // ms.SaveToStream(fs);
        //cs.Free;
       // sf:= fs.Size;
       // sm:= ms.Size;
        fs.Free;
       // ms.Free;
        if Mother^.Module.NeedToKeepAssets() then begin
          StoreAssets;
        end;
        PutSavedSessionOntoStack(tempfilename);
       { if Mother^.Debug.Verbose then AddLog('Dumped %0 kilobytes to file in %1 microseconds',// (%2 kilobytes, compression ratio %3%)',
          [round(sm/1024), round(Mother^.Timer.UsecDelta(@q))
          //, round(sf/1024), round(100.0 * sf/max(1,sm))
          ]);
        }
      except
        Mother^.Module.SessionNotSaved:= false;
        Mother^.State.SavingTheSessionNow:= false;
        Mother^.State.StateTrashedRestartRequired:= Yes;
        Mother^.Module.EraseAssetsList;
        Die(RuEn(
          'Крах при сохранении сессии!',
          'Crashed saving the session!'));
      end;
      Mother^.State.SavingTheSessionNow:= false;
    end;
    Mother^.Module.SessionNotSaved:= false;
  end;

  procedure TLogic.RollbackSession;
  begin
    Mother^.Module.RollBackTheSessionOnNextLoad:= Yes;
    Mother^.Module.State:= ms_Unloading;
    Mother^.Display.BeginBlackout(1.0);
  end;

  function TLogic.GetSessionsList: TArrayOfAnsiString;
  var
    sl: TSessionList;
    i: integer;
  begin
    sl:= TSessionList.Create(Self.SessionId);
    SetLength(Result, sl.Count);
    for i:=0 to sl.Count - 1 do Result[i]:= sl.SID[i];
  end;



