{
    This file is part of Chentrah,
    Copyright (C) 2004-2014 Anton Rzheshevski (chebmaster@mail.ru).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see http://www.gnu.org/licenses/

 **********************************************************************

    This file contains the wrapper for the game module for mother module's TStream.

 **********************************************************************}

unit mo_mstream;
interface
  uses SysUtils, classes, mo_hub;


type

TMotherStream = class(TStream)
protected
  f_ms: pointer;
  f_delete: boolean;
  function  GetSize: Int64; override;
  procedure SetSize64(const NewSize: Int64); override;
  function  GetPosition: Int64; override;
  procedure SetPosition(const Pos: Int64); override;
public
  property Handle: pointer read f_ms; //this is in fact a class instance of a different executable
  constructor CreateFromHandle(h:pointer; FreeMotherStreamAtDestroy: boolean = true);
  constructor CreateAsFileStream(FileName: TFileNameString; mode: word; FreeMotherStreamAtDestroy: boolean = true);
  constructor CreateAsMemoryStream(FreeMotherStreamAtDestroy: boolean = true);
  destructor Destroy; override;
  function Read(var Buffer; Count: Longint): Longint; override;
  function Write(const Buffer; Count: Longint): Longint; override;
  function Seek(const Offset: Int64; Origin: TSeekOrigin): Int64; override;
end;


implementation

  constructor TMotherStream.CreateFromHandle(h:pointer; FreeMotherStreamAtDestroy: boolean);
  begin
    inherited Create;
    f_ms:= h;
    f_delete:= FreeMotherStreamAtDestroy;
  end;

  constructor TMotherStream.CreateAsFileStream(FileName: TFileNameString; mode: word; FreeMotherStreamAtDestroy: boolean);
  begin
    inherited Create;
    f_ms:= Mother^.Stream.CreateOfFile(PFileNameChar(FileName), mode);
    AfterEfCheck;
    f_delete:= FreeMotherStreamAtDestroy;
  end;

  constructor TMotherStream.CreateAsMemoryStream(FreeMotherStreamAtDestroy: boolean);
  begin
    f_ms:= Mother^.Memory.CreateStream();
    AfterEfCheck;
    f_delete:= FreeMotherStreamAtDestroy;
  end;

  destructor TMotherStream.Destroy;
  begin
    if f_delete then begin
      Mother^.Stream.Free(f_ms);
      AfterEfCheck;
    end;
    inherited;
  end;

  function  TMotherStream.GetSize: Int64;
  begin
    Result:= Mother^.Stream.GetSize(f_ms);
    AfterEfCheck;
  end;

  procedure TMotherStream.SetSize64(const NewSize: Int64);
  begin
    Mother^.Stream.SetSize(f_ms, NewSize);
    AfterEfCheck;
  end;

  function  TMotherStream.GetPosition: Int64;
  begin
    Result:= Mother^.Stream.GetPosition(f_ms);
    AfterEfCheck;
  end;

  procedure TMotherStream.SetPosition(const Pos: Int64);
  begin
    Mother^.Stream.SetPosition(f_ms, Pos);
    AfterEfCheck;
  end;


  function TMotherStream.Read(var Buffer; Count: Longint): Longint;
  begin
    Result:= Mother^.Stream.Read(f_ms, @Buffer, Count);
    AfterEfCheck;
  end;

  function TMotherStream.Write(const Buffer; Count: Longint): Longint;
  begin
    Result:= Mother^.Stream.Write(f_ms, @Buffer, Count);
    AfterEfCheck;
  end;

  function TMotherStream.Seek(const Offset: Int64; Origin: TSeekOrigin): Int64;
  begin
    case Origin of
      soBeginning: SetPosition(Offset);
      soCurrent: SetPosition(GetPosition + Offset);
      soEnd: SetPosition(GetSize - Offset);
    end;
  end;

end.
