{
    This file is part of Chentrah,
    Copyright (C) 2004-2015 Anton Rzheshevski (chebmaster@mail.ru).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see http://www.gnu.org/licenses/

 **********************************************************************}


unit mo_viewport;

{$include mo_globaldefs.h}

interface

uses
  SysUtils, Classes, math, chepersy, mo_hub, typinfo, mo_classes, mo_assets,
  mo_glsl, mo_fbo, mo_3dbasics,
  mo_gui, mo_gmathbase;

type

  { TViewPort }

  //TViewport is a GUI object that knows nothing of the scene or the framebuffers.
  //  It is needed to integrate Camera into the GUI grid.

  //, and thus is subject to usual resizing dictated by GUI grid
  //How does it cope with that, and how does it limit the rendered ara
  //  is decided by each specialized class.

  //Stereo viewport uses two child viewports, for example

  TViewPort = class (TRectControl)
  protected
    f_camera : TAbstractCamera;
  public
    procedure RegisterFields; override;
    procedure Cycle; override;

    procedure Render; override;

    property Camera: TAbstractCamera read f_camera;

    function OnKey(key: TKey): boolean; override;

  end;

//  https://www.opengl.org/wiki/Framebuffer_Object_Examples
//  https://www.opengl.org/wiki/Rectangle_Texture
//  http://nehe.gamedev.net/article/glsl_an_introduction/25007/
//  http://www.gamedev.ru/code/articles/GLSL?page=4


implementation

  uses mo_module;


  procedure TViewPort.RegisterFields;
  begin
    inherited;
    ListFields([
      'camera', @camera, CPS_MANAGED_OBJECT
    ])
  end;

  function TViewport.OnKey(key: TKey): boolean;
  begin
    if not Assigned(f_camera)
      then Result:= True
      else Result:= f_camera.OnKey(key,
        //ranges 0.0..1.0 in viewport.
        (Module.Logic.mousex - Self.x) * Self.Width.Current / Math.max(1, Mother^.Display.WindowClientRect.Width),
        (Module.Logic.mousey - Self.y) * Self.Height.Current / Math.Max(1, Mother^.Display.WindowClientRect.Height)
        );
  end;


(*
  procedure TViewPort.CreateMultisamplePresentationProgram(newmls: integer);
  var
    sampler_code: AnsiString = '';
    hdr_mul: AnsiString;
    sampler_rx, sampler_ry, a, adelta, dx, dy, sr, srdelta: float;
    i, steps, step, ssum, perstep: integer;
    Multisampler: TGlslProgram;
    function csign(f: float): AnsiString;
    begin
      if f < 0 then Result:= '' else Result:= '+';
    end;
  begin
    if Assigned(Multisam2)
      or (Assigned(Multisam) and not Multisam.Actualized)
    then Exit;
    hdr_mul:= '';
    if f_hdr > 1.0 then hdr_mul:= FloatToStr(f_hdr) + ' * ';
    steps:= 1 + ((newmls - 1) div 8);
    ssum:= 1;
    for step:=1 to steps do begin
      perstep:= math.min(8, newmls - ssum);
      adelta:= 3.1416 * 2 / perstep;
      a:= 3.1416 * 1.0; //0.75; //45 degrees
      sampler_rx:= 0.7 * ((1 + steps - step)/steps) * FrW / math.max(1.0, Width.Current);
      sampler_ry:= 0.7 * ((1 + steps - step)/steps) * FrH / math.max(1.0, Height.Current);
      for i:= 1 to perstep do begin
        dx:= sin(a) * sampler_rx;
        dy:= cos(a) * sampler_ry;
        sampler_code+= 'sum += texture2D(u_texture, vec2(gl_TexCoord[0].x '
           + csign(dx) + ' ' + FloatToStr(dx) + ', gl_TexCoord[0].y '
           + csign(dy) + ' ' + FloatToStr(dy) + ')) * '
           + FloatToStr(1 / newmls) + '; ';
        a+= adelta;
      end;
      ssum+= perstep;
    end;
    Multisampler:= TGlslProgram.Create(
      TVertexShader.Create( //emulate FFP
        'void main() {'
         + 'gl_Position = gl_ModelViewProjectionMatrix * gl_Vertex;'
         + 'gl_FrontColor = gl_Color;'
         + 'gl_TexCoord[0] = gl_MultiTexCoord0;'
       +'}'
      ),
      TFragmentShader.Create(
         'uniform sampler2D u_texture;'
       + 'void main(void) {'
         + 'vec4 sum = vec4(0.0);'
         + 'sum += texture2D(u_texture, vec2(gl_TexCoord[0].x, gl_TexCoord[0].y)) * ' + FloatToStr(1 / newmls) + ';'
         + sampler_code
       //  + 'gl_FragColor = texture2D(u_texture, vec2(gl_TexCoord[0]));'
         + 'gl_FragColor = ' + hdr_mul + ' sum * gl_Color;'
       + '}'
      ),
      not Assigned(Multisam)
    );
    if Assigned(Multisam)
      then Multisam2:= Multisampler
      else Multisam:= Multisampler;
  end;
*)

  procedure TViewPort.Cycle;
  begin
    inherited;
    if Assigned(f_camera) then f_camera.Cycle;
  end;


  procedure TViewPort.Render;
  begin
    inherited;
    if Assigned(Camera) then begin
      Camera.ViewportResize(x, y, width.Current, height.Current);
      Camera.Render;
    end;
  end;



end.
