
  function BoolChars(b: boolean): UnicodeString;
  begin
    if b then Result:='True' else Result:='False';
  end;

  function VarRecToWide(V: TVarRec): UnicodeString;
  begin
    Case V.Vtype of
      vtInteger:    Result := IntToStr(V.VInteger);
      vtBoolean:    Result := AnsiToWide(BoolChars(V.VBoolean));
      vtChar:   Result := AnsiToWide(V.VChar);
      vtWideChar:   Result:=V.VWideChar;
      vtExtended:   Result := FloatToStr(V.VExtended^);
      vtAnsiString: Result := AnsiToWide(AnsiString(V.VAnsiString));
      vtWideString: Result := WideString(V.VWideString);
      vtPChar:  Result := AnsiToWide(PCharToString(V.VPChar));
      vtPWideChar:  Result := PUnicodeCharToUnicodeString(V.VPWideChar);
      vtObject:     Result := V.VObject.ClassName;
      vtClass:      Result := V.VClass.ClassName;
      vtPointer: begin
        if Assigned(V.VPointer)
          then Result:= IntToHex(ptruint(V.VPointer), MaxPointerHexChars) + 'h' //Format('%Ph',[V.VPointer])
          else Result:='NIL';
      end;

      //not supported in the FreePascal 1.0.6
      vtCurrency:   Result := CurrToStr(V.VCurrency^);
      vtVariant:    Result := string(V.VVariant^);

      vtInt64:      Result := IntToStr(V.VInt64^);
      vtQword:      Result:= IntToStr(V.VQword^);
      vtUnicodeString: Result := UnicodeString(V.VUnicodeString);
    else
      Result:='?unknown VarRec type ' + IntToStr(V.Vtype) + '?';
    end;
  end;



  function VarRecTypeName(vt: Integer): UnicodeString;
  begin
    case vt of
      vtInteger    : Result:=RuEn('целый','integer');
      vtBoolean    : Result:=RuEn('булев','boolean');
      vtChar       : Result:= 'Char';
      vtExtended   : Result:=RuEn('вещественный','extended');
      vtString     : Result:=RuEn('строка','string');
      vtPointer    : Result:=RuEn('указатель','pointer');
      vtPChar      : Result:= 'PChar';
      vtObject     : Result:=RuEn('объект','object');
      vtClass      : Result:=RuEn('класс','class');
      vtWideChar   : Result:= 'UnicodeChar';
      vtPWideChar  : Result:= 'PUnicodeChar';
      vtAnsiString : Result:=RuEn('строка','string');
      vtCurrency   : Result:= 'currency';
      vtVariant    : Result:=RuEn('вариант','variant');
      vtInterface  : Result:=RuEn('интерфейс','interface');
      vtWideString : Result:=RuEn
                            {$ifdef windows}
                              ('строка (COM BSTR / utf16)','wide string')
                            {$else}
                              ('юникодная строка (utf16)','unicode string')
                            {$endif};
      vtUnicodeString : Result:=RuEn('юникодная строка (utf16)','unicode string');
      vtInt64      : Result:=RuEn('64-битный целый','64-bit integer');
     {$ifdef fpc}
      vtQWord      : Result:=RuEn('64-битный беззнаковый целый','64-bit unsigned integer');
     {$endif}
    else
      Result:=RuEn('неизвестный','unknown');
    end;
  end;



  function PervertedFormat(U: UnicodeString; P: array of const): UnicodeString; //OVERLOAD;
  var
    j: integer;
    b, e: UnicodeString;
  begin
    e:='';
    For j:=0 to High(p) do begin
      b:=VarRecToWide(P[j]);
      if WidePos('%' + IntToStr(j), U) < 1
        then e:=e + '  [' + b + ']  '
        else u:=WideReplace(u, '%' + IntToStr(j), b);
    end;
    if e <> '' then begin
      e:=' +FORMAT ERROR!! ' + e;
    end;
    Result:=u + e;
  end;

  function PervertedFormat(U: UnicodeString; P: array of UnicodeString): UnicodeString; //OVERLOAD;
  var
    j: integer;
    e: UnicodeString;
  begin
    e:='';
    For j:=0 to High(p) do begin
      if WidePos('%' + IntToStr(j), U) < 1
        then e:=e + '  [' + P[j] + ']  '
        else u:=WideReplace(u, '%' + IntToStr(j), P[j]);
    end;
    if e <> '' then begin
      e:=' +FORMAT ERROR!! ' + e;
    end;
    Result:=u + e;
  end;




 function SetToString(bitmask: dword; ti: ptypeinfo): UnicodeString;
 var
   e: array of UnicodeString;
   i: integer;
   D: PTypeData;
 begin
   if not (ti^.Kind in [tkEnumeration, tkInteger, tkSet])
     then Die(MI_ERROR_PROGRAMMER_NO_BAKA, [IntToHex(bitmask, 8) + RuEn(
       ' приведённая TypeInfo не принадлежит множеству или перечислимому типу',
       ' the TypeInfo provided os not of a set nor of an enumerated type')
       + '(' + AnsiString(ti^.Name) + ')']);
   if ti^.Kind = tkSet then ti:= GetTypeData(ti)^.CompType;
   D:= GetTypeData(ti);
   if D^.MinValue <> 0
     then Die(MI_ERROR_PROGRAMMER_NO_BAKA, [RuEn(
      'не рассчитано на поддиапазонные типы',
      'not designed for subrange types')]);
   if D^.MaxValue > 31
     then Die(MI_ERROR_PROGRAMMER_NO_BAKA, [RuEn(
      'рассчитано только на перечислимые типы, множества от которых приводимы к dword (не более 32 элементов)',
      'only designed for enums with derivative sets type-castable to dword (no more than 32 entries)')]);
   for i:= 0 to D^.MaxValue do begin
     if (bitmask and (dword(1) shl i)) = 0 then continue;
     SetLength(e, Length(e) + 1);
     if ti^.Kind = tkEnumeration
       then e[High(e)]:= GetEnumName(ti, i)
       else e[High(e)]:= IntToStr(i);
   end;
   Result:= Implode(',', e);
 end;
