
{$ifndef cgekernel}
  type
    ffEnum = (ffSize, ffLeft, ffTop, ffGridWidth, ffGridHeight,
                     ffGlyphWidth, ffGlyphHeight, ffHspacing,
                     ffSsize, ffTsize);
    TffSet = array [ffEnum] of integer;
    Pffset = ^Tffset;

  const
    ffm_builtin: Tffset =
      (13, 14, 9,  14,  24,
          10,  22,  10,
          256,  512);
    ffm_normal: Tffset =
      (31, 22, 7,  36,  55,
          24,  54,  18,
          1024, 1024);
    ffzoomx: array[0..1] of float = (1.0, 0.85);
  var
    ffm: array[0..1] of PffSet = (@ffm_normal, @ffm_builtin);
{$endif}

  class procedure FixedFont.SizeToFace(size: integer; face: Pinteger; zoom: PGLfloat );
  begin
    zoom^:= size / ffm[0]^[ffSize];
  end;


  class function FixedFont.GetStringWidth(Size, numChars: integer): GLfloat;
  var
    face: integer;
    zoom: GLfloat;
  begin
    SizeToFace(Size, @face, @zoom);
    Result:=max(1, numChars) * round(zoom * ffm[0]^[ffHspacing]);
  end;

  class function FixedFont.GetStringHeight(Size, numLines: integer): GLfloat;
  var
    face: integer;
    zoom: GLfloat;
  begin
    SizeToFace(max(1, Size), @face, @zoom);
    Result:= max(1, numLines) * round(zoom * ffm[0]^[ffGlyphHeight]);
  end;



  class procedure FixedFont.FitStringIntoRectangle
                  (Size, numChars, numLines, Width, Height: integer;
                                                  fit: PStringFitRec);
  var
    w, h: float;//integer;
  begin
    Width:=max(1, Width);
    Height:=max(1, Height);
    numChars:=max(1, numChars);
    numLines:=max(1, numLines);

    SizeToFace(Size, @fit^.face, @fit^.zoomX);
    fit^.zoomY:= fit^.zoomX;
    w:={round}(fit^.zoomX * ffm[0]^[ffHspacing]);
    h:={round}(fit^.zoomY * ffm[0]^[ffGlyphHeight]);
    if (w * numChars <= Width) and (h * numLines <= Height ) then begin
      fit^.ActualWidth:= round(w * numChars);
      fit^.ActualHeight:= round(h * numLines);
    end
    else begin
//    if face^ = ffMaxFace then begin
      //zoom it down until it fits or zoom falls below 19/40
      if w * numChars > Width then begin
        w:= Width / {div} numChars;
        fit^.zoomX:= w / ffm[0]^[ffHspacing];
      end;
      if h * numLines > Height then begin
        h:= Height / {div} numLines;
        fit^.zoomY:= h / ffm[0]^[ffGlyphHeight];
      end;

      fit^.ActualWidth:= round(numChars * {round}(fit^.zoomX * ffm[0]^[ffHspacing]));
      fit^.ActualHeight:= round(numLines * {round}(fit^.zoomY * ffm[0]^[ffGlyphHeight]));
    end;
  end;


  var
    prevMipX: integer = 0;
    prevMipY: integer = 0;

  class procedure FixedFont.SetRenderState (noshadow: boolean);
  begin
    glGetError;

    noshadow:= noshadow or (Mother^.Text.FixedFontQuality = 1);
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D, Mother^.Text.BuiltinFont[Mother^.Text.FixedFontQuality].Texture);
    glEnable(GL_BLEND);
    glDisable(GL_DEPTH_TEST);
    glDisable(GL_CULL_FACE);
    if noshadow
      then glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_COLOR)
      else glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA);

    case  Mother^.GAPI.Mode of
     {$ifndef glesonly}
      gapi_GL21, gapi_DeprecatedImmediateGL :begin
        glDisable(GL_ALPHA_TEST);
      end;
     {$endif glesonly}
      gapi_GLES2: begin
      end;
    else
      DieUnsupportedGLMode;
    end;

  end;

  class procedure FixedFont.RenderString (fit: PStringFitRec; Left, Top: glFloat; w: PUnicodeChar);
  var
    cn, i: integer;
    sw, gw, gh: float;//integer;
    Mesh: TDumbUniMesh;
    ptc: PVector4f;
  begin
    if not Assigned(w) then exit;

    glGetError;

    Mesh:= TDumbUniMesh.Create;
    TGAPIAbstractionLayer.SetGLStatesForGUI;

    gh:={round}(ffm[0]^[ffGlyphHeight] * fit^.zoomY);
    gw:={round}(ffm[0]^[ffGlyphWidth] * fit^.zoomX)
      * ffzoomx[Mother^.Text.FixedFontQuality];
    sw:={round}(ffm[0]^[ffHspacing] * fit^.zoomX);

    glBindTexture(GL_TEXTURE_2D, Mother^.Text.BuiltinFont[Mother^.Text.FixedFontQuality].Texture);

    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAX_ANISOTROPY_EXT, 4);



//    glBegin(GL_QUADS);

    while w^ <> #0 do begin
      if (w^ <> ' ') //no rendering spaces!
         //and (ord(w^) <= Mother^.Text.BuiltinFont[Mother^.Text.FixedFontQuality].CharMapHigh)
      then begin
        cn:= Mother^.Text.BuiltinFont[Mother^.Text.FixedFontQuality].CharMap[ord(w^)];

        ptc:= PVector4f(Mother^.Text.BuiltinFont[Mother^.Text.FixedFontQuality].TexCoords) + cn;

        Mesh.AddQuad2d(
          ffcolor[0], ffcolor[1], ffcolor[2], ffcolor[3] * Mother^.Display.FadeIn,
          Left, Top, Left + gw, Top + gh,
            ptc^[0], ptc^[1], ptc^[2], ptc^[3] );

//        glTexCoord2f(fftexcoord[cn, 0, 0, Mother^.Text.FixedFontQuality], fftexcoord[cn, 0, 1, Mother^.Text.FixedFontQuality]);
//        glVertex2f(Left, Top);

//        glTexCoord2f(fftexcoord[cn, 1, 0, Mother^.Text.FixedFontQuality], fftexcoord[cn, 0, 1, Mother^.Text.FixedFontQuality]);
//        glVertex2f(Left + gw, Top);

//        glTexCoord2f(fftexcoord[cn, 1, 0, Mother^.Text.FixedFontQuality], fftexcoord[cn, 1, 1, Mother^.Text.FixedFontQuality]);
//        glVertex2f(Left + gw, Top + gh);

//        glTexCoord2f(fftexcoord[cn, 0, 0, Mother^.Text.FixedFontQuality], fftexcoord[cn, 1, 1, Mother^.Text.FixedFontQuality]);
//        glVertex2f(Left, Top + gh);
      end;
      Left += sw;
      inc(w);
    end;

    Mesh.Render;
    Mesh.Free;

//    glEnd();
  end;


  class procedure FixedFont.Color3f(r, g, b: GLfloat);
  begin
    ffcolor[0]:= r;
    ffcolor[1]:= g;
    ffcolor[2]:= b;
    ffcolor[3]:= 1;
  end;

  class procedure FixedFont.Color4f(r, g, b, a: GLfloat);
  begin
    ffcolor[0]:= r;
    ffcolor[1]:= g;
    ffcolor[2]:= b;
    ffcolor[3]:= a;
  end;

  class procedure FixedFont.Color3fv(p: PGLfloat);
  begin
    ffcolor[0]:= p^;
    ffcolor[1]:= (p + SizeOf(GLfloat))^;
    ffcolor[2]:= (p + 2 * SizeOf(GLfloat))^;
    ffcolor[3]:= 1;
  end;

  class procedure FixedFont.Color4fv(p: PGLfloat);
  begin
    ffcolor[0]:= p^;
    ffcolor[1]:= (p + 1)^;
    ffcolor[2]:= (p + 2)^;
    ffcolor[3]:= (p + 3)^;
  end;

