{
  This is a permanent fix for the broken RTL of FPC 2.6.4
  See http://bugs.freepascal.org/view.php?id=27221

  This fix is required to get Platinum level of Wine compatibility
}

{$ifndef windows}
  {$fatal Windows ONLY}
{$endif}


type


  TFileStream = class(classes.TFileStream)
  public
    destructor Destroy; override;
  end;

  //The rest of the classes are dragged along by the need to base them on the fixed TFileStream


  TStringList = class(classes.TStringList)
  public
    procedure LoadFromFile(const FileName: AnsiString);
    procedure SaveToFile(const FileName: AnsiString);
  end;

  TIniFileKey = class
  Private
    FIdent: Utf8String;
    FValue: Utf8String;
  public
    constructor Create(const AIdent, AValue: Utf8String);
    property Ident: Utf8String read FIdent write FIdent;
    property Value: Utf8String read FValue write FValue;
  end;

  TIniFileKeyList = class(TList)
  private
    function GetItem(Index: integer): TIniFileKey;
    function KeyByName(const AName: Utf8String; CaseSensitive : Boolean): TIniFileKey;
  public
    destructor Destroy; override;
    procedure Clear; override;
    property Items[Index: integer]: TIniFileKey read GetItem; default;
  end;

  TIniFileSection = class
  private
    FName: Utf8String;
    FKeyList: TIniFileKeyList;
  public
    Function Empty : Boolean;
    constructor Create(const AName: Utf8String);
    destructor Destroy; override;
    property Name: Utf8String read FName;
    property KeyList: TIniFileKeyList read FKeyList;
  end;

  TIniFileSectionList = class(TList)
  private
    function GetItem(Index: integer): TIniFileSection;
    function SectionByName(const AName: Utf8String; CaseSensitive : Boolean): TIniFileSection;
  public
    destructor Destroy; override;
    procedure Clear;override;
    property Items[Index: integer]: TIniFileSection read GetItem; default;
  end;

  { TCustomIniFile }

  TCustomIniFile = class
  Private
    FFileName: Utf8String;
    FSectionList: TIniFileSectionList;
    FEscapeLineFeeds: boolean;
    FCaseSensitive : Boolean;
    FStripQuotes : Boolean;
  public
    constructor Create(const AFileName: AnsiString; AEscapeLineFeeds : Boolean = False); virtual;
    destructor Destroy; override;
    function SectionExists(const Section: Utf8String): Boolean; virtual;
    function ReadString(const Section, Ident, Default: Utf8String): Utf8String; virtual; abstract;
    procedure WriteString(const Section, Ident, Value: String); virtual; abstract;
    function ReadInteger(const Section, Ident: Utf8String; Default: Longint): Longint; virtual;
    procedure WriteInteger(const Section, Ident: Utf8String; Value: Longint); virtual;
    function ReadInt64(const Section, Ident: Utf8String; Default: Int64): Longint; virtual;
    procedure WriteInt64(const Section, Ident: Utf8String; Value: Int64); virtual;
    function ReadBool(const Section, Ident: Utf8String; Default: Boolean): Boolean; virtual;
    procedure WriteBool(const Section, Ident: Utf8String; Value: Boolean); virtual;
    function ReadDate(const Section, Ident: Utf8String; Default: TDateTime): TDateTime; virtual;
    function ReadDateTime(const Section, Ident: Utf8String; Default: TDateTime): TDateTime; virtual;
    function ReadFloat(const Section, Ident: Utf8String; Default: Double): Double; virtual;
    function ReadTime(const Section, Ident: Utf8String; Default: TDateTime): TDateTime; virtual;
    function ReadBinaryStream(const Section, Name: Utf8String; Value: TStream): Integer; virtual;
    procedure WriteDate(const Section, Ident: Utf8String; Value: TDateTime); virtual;
    procedure WriteDateTime(const Section, Ident: Utf8String; Value: TDateTime); virtual;
    procedure WriteFloat(const Section, Ident: Utf8String; Value: Double); virtual;
    procedure WriteTime(const Section, Ident: Utf8String; Value: TDateTime); virtual;
    procedure WriteBinaryStream(const Section, Name: Utf8String; Value: TStream); virtual;
    procedure ReadSection(const Section: Utf8String; Strings: TStrings); virtual; abstract;
    procedure ReadSections(Strings: TStrings); virtual; abstract;
    procedure ReadSectionValues(const Section: Utf8String; Strings: TStrings); virtual; abstract;
    procedure EraseSection(const Section: Utf8String); virtual; abstract;
    procedure DeleteKey(const Section, Ident: String); virtual; abstract;
    procedure UpdateFile; virtual; abstract;
    function ValueExists(const Section, Ident: Utf8String): Boolean; virtual;
    property FileName: Utf8String read FFileName;
    property EscapeLineFeeds: boolean read FEscapeLineFeeds;
    Property CaseSensitive : Boolean Read FCaseSensitive Write FCaseSensitive;
    Property StripQuotes : Boolean Read FStripQuotes Write FStripQuotes;
  end;

  { TIniFile }

  TIniFile = class(TCustomIniFile)
  Private
    FStream: TStream;
    FCacheUpdates: Boolean;
    FDirty : Boolean;
    FBOM : String;
    procedure FillSectionList(AStrings: TStrings);
    Procedure DeleteSection(ASection : TIniFileSection);
    Procedure MaybeDeleteSection(ASection : TIniFileSection);
    procedure SetCacheUpdates(const AValue: Boolean);
  protected
    procedure MaybeUpdateFile;
    property Dirty : Boolean Read FDirty;
  public
    constructor Create(const AFileName: AnsiString; AEscapeLineFeeds : Boolean = False); overload; override;
    constructor Create(AStream: TStream; AEscapeLineFeeds : Boolean = False); overload;
    destructor Destroy; override;
    function ReadString(const Section, Ident, Default: Utf8String): Utf8String; override;
    procedure WriteString(const Section, Ident, Value: String); override;
    procedure ReadSection(const Section: Utf8String; Strings: TStrings); override;
    procedure ReadSectionRaw(const Section: Utf8String; Strings: TStrings);
    procedure ReadSections(Strings: TStrings); override;
    procedure ReadSectionValues(const Section: Utf8String; Strings: TStrings); override;
    procedure EraseSection(const Section: Utf8String); override;
    procedure DeleteKey(const Section, Ident: String); override;
    procedure UpdateFile; override;
    property Stream: TStream read FStream;
    property CacheUpdates : Boolean read FCacheUpdates write SetCacheUpdates;
  end;

  TMemIniFile = class(TIniFile)
  public
    constructor Create(const AFileName: AnsiString; AEscapeLineFeeds : Boolean = False); overload; override;
    procedure Clear;
    procedure GetStrings(List: TStrings);
    procedure Rename(const AFileName: Utf8String; Reload: Boolean);
    procedure SetStrings(List: TStrings);
  end;


