class procedure TGAPIAbstractionLayer.OrthoApplicationWindow;
begin
  case  Mother^.GAPI.Mode of
    {$ifndef glesonly}
    gapi_GL21, gapi_DeprecatedImmediateGL :begin
      glMatrixMode(GL_PROJECTION);
      glLoadIdentity();
      glMatrixMode(GL_MODELVIEW);
      glLoadIdentity();
      //origin in the upper left corner!
      glOrtho(0, Mother^.Display.WindowClientRect.Width, Mother^.Display.WindowClientRect.Height, 0, 0 ,1);
      //gluOrtho2d(0, WindowManager.Display.WindowClientWidth, 0, WindowManager.Display.WindowClientHeight);
    end;
    {$endif glesonly}
    gapi_GLES2: begin
      Self.Ortho(0, Mother^.Display.WindowClientRect.Width, Mother^.Display.WindowClientRect.Height, 0, 0, 1);
    end;
  else
    DieUnsupportedGLMode;
  end;
end;


class procedure TGAPIAbstractionLayer.SetGLStatesForGUI;
begin
  glEnable(GL_TEXTURE_2D);
  glEnable(GL_BLEND);
  glDisable(GL_SCISSOR_TEST);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  glDisable(GL_CULL_FACE);
  glDisable(GL_DEPTH_TEST);
  glDepthMask(GL_TRUE);
  glViewport(0, 0, Mother^.Display.WindowClientRect.Width, Mother^.Display.WindowClientRect.Height);

  case  Mother^.GAPI.Mode of
   {$ifndef glesonly}
    gapi_GL21, gapi_DeprecatedImmediateGL :begin
      glEnable(GL_ALPHA_TEST);
      glAlphaFunc(GL_GREATER, 0);
    end;
   {$endif glesonly}
    gapi_GLES2: begin
    end;
  else
    DieUnsupportedGLMode;
  end;

end;


//substitute for glOrtho unavalilable in GL ES
class procedure TGAPIAbstractionLayer.Ortho(left, right, bottom, top,  nearVal, farVal: GLfloat);
begin
  //https://www.opengl.org/sdk/docs/man2/xhtml/glOrtho.xml
{  matrix[0, 0]:= 2 / (right - left);
  matrix[0, 1]:= 0;
  matrix[0, 2]:= 0;
  matrix[0, 3]:= 0;

  matrix[1, 0]:= 0;
  matrix[1, 1]:= 2 / (top - bottom);
  matrix[1, 2]:= 0;
  matrix[1, 3]:= 0;

  matrix[2, 0]:= 0;
  matrix[2, 1]:= 1;
  matrix[2, 2]:= -2 / (farVal - nearVal);
  matrix[2, 3]:= 0;

  matrix[3, 0]:= - (right + left) / (right - left);
  matrix[3, 1]:= - (top + bottom) / (top - bottom);
  matrix[3, 2]:= - (farVal + nearVal) / (farVal - nearVal);
  matrix[3, 3]:= 1;
}
end;

//reminder: here "top" denotes smaller Y values as my GUI uses y axis downwards
procedure TDumbUniMesh.AddQuad2d(r, g, b, a, left, top, right, bottom, txleft, txtop, txright, txbottom: GLfloat);
var
  firstindex, firstvertex, newlength, i: GLint;
begin
  firstindex:= maxIndex + min(1, maxIndex); //because are initialized to 0, not -1
  firstvertex:= maxVertex + min (1, maxVertex);

  while Length(indices) < maxIndex + 6
    do Setlength(indices, max(16, Length(indices) * 2));

  indices[firstindex + 0]:= firstvertex + 0;
  indices[firstindex + 1]:= firstvertex + 1;
  indices[firstindex + 2]:= firstvertex + 2;
  indices[firstindex + 3]:= firstvertex + 2;
  indices[firstindex + 4]:= firstvertex + 3;
  indices[firstindex + 5]:= firstvertex + 0;

  while Length(normals) < maxVertex + 6 do begin
    newlength:= max(16, Length(normals) * 2 );
    SetLength(texcoords, newlength);
    SetLength(colors, newlength);
    SetLength(vertices, newlength);
    Setlength(normals, newlength);
  end;

  for i:= firstvertex to firstvertex + 3 do begin
    normals[i][0]:= 0;
    normals[i][1]:= 0;
    normals[i][2]:= 1.0;
  end;

  for i:= firstvertex to firstvertex + 3 do begin
    colors[i][0]:= r;
    colors[i][1]:= g;
    colors[i][2]:= b;
    colors[i][3]:= a;
  end;

  texcoords[firstvertex + 0][0]:= txleft;
  texcoords[firstvertex + 0][1]:= txtop;
  texcoords[firstvertex + 0][2]:= 0;

  texcoords[firstvertex + 1][0]:= txright;
  texcoords[firstvertex + 1][1]:= txtop;
  texcoords[firstvertex + 1][2]:= 0;

  texcoords[firstvertex + 2][0]:= txright;
  texcoords[firstvertex + 2][1]:= txbottom;
  texcoords[firstvertex + 2][2]:= 0;

  texcoords[firstvertex + 3][0]:= txleft;
  texcoords[firstvertex + 3][1]:= txbottom;
  texcoords[firstvertex + 3][2]:= 0;


  for i:= firstvertex to firstvertex + 3 do vertices[i][3]:= 1.0;

  vertices[firstvertex + 0][0]:= left;
  vertices[firstvertex + 0][1]:= top;
  vertices[firstvertex + 0][2]:= 0;

  vertices[firstvertex + 1][0]:= right;
  vertices[firstvertex + 1][1]:= top;
  vertices[firstvertex + 1][2]:= 0;

  vertices[firstvertex + 2][0]:= right;
  vertices[firstvertex + 2][1]:= bottom;
  vertices[firstvertex + 2][2]:= 0;

  vertices[firstvertex + 3][0]:= left;
  vertices[firstvertex + 3][1]:= bottom;
  vertices[firstvertex + 3][2]:= 0;

  maxIndex:= firstindex + 5;
  maxVertex:= firstvertex + 3;
end;

procedure TDumbUniMesh.Render;
var i: integer;
begin
  if Length(indices) <= 0 then Exit;
  if maxIndex < High(indices) then maxIndex:= High(indices);
  case  Mother^.GAPI.Mode of
   {$ifndef glesonly}
    gapi_DeprecatedImmediateGL, gapi_GL21: begin
      {
      glBegin(GL_TRIANGLES);
      for i:=0 to maxIndex do begin
        glColor4fv(@colors[indices[i]]);
        glNormal3fv(@normals[indices[i]]);
        glTexCoord3fv(@texcoords[indices[i]]);
        glVertex4fv(@vertices[indices[i]]);
      end;
      glEnd;
      }
     // https://www.opengl.org/wiki/Client-Side_Vertex_Arrays

glGetError(); //***TODO this is a hack for an error state provoked somewhere else. Note to self: find and eliminate

      glEnableClientState(GL_VERTEX_ARRAY);
      glVertexPointer(4, GL_FLOAT, 0, @vertices[0]);
      glEnableClientState(GL_COLOR_ARRAY);
      glColorPointer(4, GL_FLOAT, 0, @colors[0]);
      glClientActiveTexture(GL_TEXTURE0);
      glEnableClientState(GL_TEXTURE_COORD_ARRAY);
      glTexCoordPointer(3, GL_FLOAT, 0, @texcoords[0]);
      glEnableClientState(GL_NORMAL_ARRAY);
      glNormalPointer(GL_FLOAT, 0, @normals[0]);

      glDrawElements(GL_TRIANGLES, maxIndex + 1, GL_UNSIGNED_SHORT, @indices[0]);

      CheckGLError;
    end;
   {$endif glesonly}
    gapi_GLES2: begin
Die('***NotImplementedYet');
    end;
  else
    DieUnsupportedGLMode;
  end;
end;
