



function ExplainLineInfo(addr: pointer): string;

type
  TLineInfoRec = record
    exe, //returns '<unknown executable>' if it cannot detect it
        //path is always stripped.
    src, //path is always stripped
    func, //only when using the RTL lineinfo. Cheb's lineinfo doesn't support
         //  this, always sets to empty string.
    why_not // instead of the global variable LineInfoError
      : UnicodeString;
    line, // -1 means 'unknown'
    column  //only when using Cheb's lineinfo. The RTL lineinfo doesn't support this.
            // usually is set to -1 indicating 'unknown'
      :integer;
  end;

procedure GetLInfo(addr: pointer; var rec: TLineInfoRec);
  {
    The format of returned information:
    "exe" *always* receives the full name of the executable file
      (the main exe or one of dlls it uses) the addr belongs to.
      In Linux, it returns the real file name, with all symlinks
      resolved.
    "line" can be negative, which means no line info has been found
      for this address. See LineInfoError (below) for details.
    "src" returns the source file name. It either doesn't or does
      contain a full path. If the source was in the same directory
      as the program itself, there will be no path. If the source
      was in the different directory, there will be a full path
      (for the moment when the program was compiled, NOT for the
      current location of that source).
    "column" is positive ONLY when there is more than one address
      stored for the same source line. FreePascal generates this
      on VERY rare occasions, mostly for the arithmetic formulas
      spanning several lines. So most of the time column will
      receive -1.
  }

function InitLineInfo(someaddr: pointer; usecriticalsection: longbool = true): longbool;
  {
    This function is called by GetLInfo() anyway if it doesnt't
      find a loaded line info for the executable to which the
      requested addres belongs.
    Also installs the custom BackTraceStr handler.

    Input:
    someaddr is adress of any function that belongs to the executable
      you want to pre-load the line info for. For example, a function
      exported from a particular dll.
    If you pass NIL, it will load the line info for the executable
      yhis module was compiled with.

    Output:
    Returns false if it failed to load the line info for the particular
      executable. In this case look LineInfoError for explanation
    Returns true if the line info for the particular executable is loaded ok.
    Returns true and does nothing if line info for that executable is
      already loaded.
  }

  procedure CleanupLineInfo();
  { deletes critical section
  }

  function GetExeBaseAddr(addr: pointer): pointer;
  function GetExeFileName(addr: pointer): TFileNameString;

  {
    This function allows you to know which executable (i.e. the main exe
      or one of the dlls loaded by it) owns this part of the virtual
      addres space.
    baseaddr receives the exe/dll base address
      (always NIL for the main exe in Linux).
    The mechnaism is made with possibility of a DLL relocation
      in mind, but that particular feature is untested.
    This function is used by GetLInfo() to determine which executable
      to load line the info from.
  }

  var
    _LineInfoError: UnicodeString = '';
    LineInfoPaths: array of TFileNameString = nil;
    ExtractDwarfLineInfoError: UnicodeString = '';


  var
    lics: syncobjs.TCriticalSection = nil;
