
{$push}
  {$ifdef che_unicode}
    {$warn IMPLICIT_STRING_CAST off}
    {$warn IMPLICIT_STRING_CAST_LOSS off}
    {$warn EXPLICIT_STRING_CAST off}
    {$warn EXPLICIT_STRING_CAST_LOSS off}
  {$endif}

  function ExplainLineInfo(addr: pointer): string;
  var
    Lir: TLineInfoRec;
  begin
    GetLInfo(addr, Lir);
    Result:= '';
    if Lir.line < 0 then Result+= ' (no debug info: ' + Lir.why_not + ')'
    else begin
      Result+= ', line ' + IntToStr(Lir.line);
      if Lir.column >=0 then Result+= ', column ' + IntToStr(Lir.column);
      Result += ' of ' + Lir.src;
    end;
    if Lir.func <> '' then Result+= ', function ' + Lir.func;
    Result+= ' in ' + Lir.exe;
  end;
{$pop}

  function GetExeBaseAddr(addr: pointer): pointer;
  {$ifdef unix}
  var
    dlinfo: dl_info;
  begin
    FillChar(dlinfo, sizeof(dlinfo), 0);
    dladdr(addr, @dlinfo);
    Result:= dlinfo.dli_fbase;
    if ExtractFileName(Utf8String(dlinfo.dli_fname)) = ExtractFileName(ParamStr(0))
      then Result:= nil;
  {$else unix}
  var
    Tmm: TMemoryBasicInformation;
  begin
    if VirtualQuery(addr, @Tmm, SizeOf(Tmm)) <> sizeof(Tmm)
      then begin
        _LineInfoError:= 'The VirtualQuery() call failed, unable to determine allocation base.';
        Result:= NIL;
      end
      else
        Result:= Tmm.AllocationBase;
  {$endif unix}
  end;

  function GetExeFileName(addr: pointer): TFileNameString;
  {$ifdef unix}
  var
    dlinfo: dl_info;
  begin
    FillChar(dlinfo, sizeof(dlinfo), 0);
    dladdr(addr, @dlinfo);
    Result:= Utf8ToFileName(Utf8String(dlinfo.dli_fname));
  {$else unix}
  var
    TST: {$ifdef che_legacy}
           AnsiString
         {$else}
           UnicodeString
         {$endif};
  begin
    SetLength(TST, Max_Path + 2);
    SetLength(TST, {$ifdef che_legacy}
                     GetModuleFileName
                   {$else}
                     GetModuleFileNameW
                   {$endif} (
      THandle(GetExeBaseAddr(addr)), { a dirty trick that ALWAYS works
                Basically, Windows uses the allocation base as a handle. }
      @TST[1], Length(TST) - 1));
    Result:= UnicodeToFileName(TST);
  {$endif unix}
  end;


  procedure GetModuleByAddr(addr: pointer; var baseaddr: pointer; var filename: UnicodeString);
  {$ifdef unix}
  var
    dlinfo: dl_info;
  begin
    FillChar(dlinfo, sizeof(dlinfo), 0);
    dladdr(addr, @dlinfo);
    baseaddr:= dlinfo.dli_fbase;
    filename:= String(dlinfo.dli_fname);
    if ExtractFileName(filename) = ExtractFileName(ParamStr(0))
      then baseaddr:= nil;
//    if filename = BacktrackSymlink(ParamStr(0)) then baseaddr:= nil; //doesn't work!
//addlog ('----------'#10#13'  %0'#10#13'  %1',[BacktrackSymlink(ParamStr(0)), filename]);
  {$else unix}

  var
    Tmm: TMemoryBasicInformation;
    TST: {$ifdef che_legacy}
           AnsiString
         {$else}
           UnicodeString
         {$endif};
  begin
    if VirtualQuery(addr, @Tmm, SizeOf(Tmm)) <> sizeof(Tmm)
      then raise Exception.Create('The VirtualQuery() call failed.');
    baseaddr:=Tmm.AllocationBase;
    SetLength(TST, Max_Path + 2);
    SetLength(TST, {$ifdef che_legacy}
                     GetModuleFileName
                   {$else}
                     GetModuleFileNameW
                   {$endif} (
      THandle(Tmm.AllocationBase), { a dirty trick that ALWAYS works
                Basically, Windows uses the allocation base as a handle. }
      @TST[1], Length(TST) - 1));
    filename:= TST;
  {$endif unix}
  end;


{$ifdef cpu32}
  // original chelinfo. sadly, never ported to x86_64.
  {$include un_lineinfo_che.inc}

{$else}
 // use wrappers for the standard lineinfo
   procedure GetLInfo(addr: pointer;  var rec: TLineInfoRec);
   var
     func,source: ShortString; //the same string as RTL uses!
   begin
     if Assigned(lics) then lics.Enter;
     FillChar(rec, sizeof(rec), 0);
     rec.column:= -1;
     if not Assigned(addr) then begin
       rec.line:= -1;
       rec.func:= RuEn(
         'по нулевому адресу',
         'at the NIL address');
     end
     else begin

     rec.exe:= ExtractFileName(GetExeFileName(addr));
     //function GetLineInfo(addr:ptruint;var func,source:string;var line:longint) : boolean;

     if GetLineInfo(ptruint(addr), func, source, rec.line) then begin
      {$push}
        {$ifdef che_unicode}
          {$warn IMPLICIT_STRING_CAST off}
          {$warn IMPLICIT_STRING_CAST_LOSS off}
          {$warn EXPLICIT_STRING_CAST off}
          {$warn EXPLICIT_STRING_CAST_LOSS off}
        {$endif}
      rec.func:= AnsiToWide(func);
      rec.src:= ExtractFileName(TFileNameString(source));
      {$pop}
     end
     else begin
       rec.line:= -1;
       rec.why_not:= RuEn(
        'функция RTL не смогла получить информацию о строках',
        'the RTL function failed to retrieve line info');
     end;
     end;
     if Assigned(lics) then lics.Leave;
   end;

  function InitLineInfo(someaddr: pointer; usecriticalsection: longbool = true): longbool;
  begin
    if not Assigned(lics) then lics:= syncobjs.TCriticalSection.Create;
  end;

  procedure CleanupLineInfo();
  begin
    if Assigned(lics) then lics.Free;
  end;


{$endif}
