{$push}
  {$rangechecks off}  // Мать :(
  {$overflowchecks off}

  function ChelinfoBackTraceStr(addr : Pointer) : ShortString;
  var
    Lir: TLineInfoRec;
    Store  : TBackTraceStrFunc;
  begin
    { reset to prevent infinite recursion if problems inside the code }
    Store := BackTraceStrFunc;
    BackTraceStrFunc := @SysBackTraceStr;
    GetLInfo(addr, Lir);
    { create string }
    Result:=' $' + HexStr(ptrint(addr), {$ifdef cpu64} 11 {$else} 8 {$endif});
    if Lir.line < 0 then Result+= ' (no debug info: ' + Lir.why_not + ')'
    else begin
      Result+= ', line ' + IntToStr(Lir.line);
      if Lir.column >=0 then Result+= ', column ' + IntToStr(Lir.column);
      Result += ' of ' + Lir.src;
    end;
    Result+= ' in ' + Lir.exe;
    BackTraceStrFunc := Store;
  end;

var
    LineInfoInitialized: boolean = false;

  function ExtractDwarfLineInfo(
    ExeFileName: TFileNameString;
    var _dlnfo: pointer;
    var _dlnfoSize: integer;
    var Imagebase: cardinal
  ): longbool; forward;
  {
  Reads the dwarf line info from an executable.
    In case of error, see ExtractDwarfLineInfoError for details.
  ImageBase is nil for unix DLLs
    in all other cases the value it receives must be substracted
    from the addresses in the dwarf line info (and then the real
    base address added, to account for the possible relocation)
  NOTE: currently in unix it is also NIL for the main executable,
    corresponding the GetModuleByAddr() in un_lineinfo
    also returning NIL as the base address for the main executable.
  }

const
  PEDlin: string = '/30';  // The block of a windows EXE containing the dwarf information
  ELFDlin: string = '.debug_line'; // The block of a Linux executable containing the dwarf information

{$push}
  {$packrecords default}

  const
    MAX_RANDOM_OFFSET_TO_TRY = 1000; // a hack to fix broken dwarf generation in FPC 2.2

  { DWARF 2 default opcodes}
  const
    { Extended opcodes }
    DW_LNE_END_SEQUENCE = 1;
    DW_LNE_SET_ADDRESS = 2;
    DW_LNE_DEFINE_FILE = 3;
    { Standard opcodes }
    DW_LNS_COPY = 1;
    DW_LNS_ADVANCE_PC = 2;
    DW_LNS_ADVANCE_LINE = 3;
    DW_LNS_SET_FILE = 4;
    DW_LNS_SET_COLUMN = 5;
    DW_LNS_NEGATE_STMT = 6;
    DW_LNS_SET_BASIC_BLOCK = 7;
    DW_LNS_CONST_ADD_PC = 8;
    DW_LNS_FIXED_ADVANCE_PC = 9;
    DW_LNS_SET_PROLOGUE_END = 10;
    DW_LNS_SET_EPILOGUE_BEGIN = 11;
    DW_LNS_SET_ISA = 12;

  type
    { state record for the line info state machine }
    TMachineState = record
      address : cardinal;
      file_id : DWord;
      line : QWord;
      column : DWord;
      is_stmt : Boolean;
      basic_block : Boolean;
      end_sequence : Boolean;
      prolouge_end : Boolean;
      epilouge_begin : Boolean;
      isa : DWord;
      append_row : Boolean;
    end;


  { DWARF line number program header preceding the line number program, 32 bit version }
    TLineNumberProgramHeader32 = packed record
      unit_length : DWord;
      version : Word;
      length : DWord;
      minimum_instruction_length : Byte;
      default_is_stmt : byte;//Bool8;
      line_base : ShortInt;
      line_range : Byte;
      opcode_base : Byte;
    end;

    TDwarfChunk = packed record
      addr: pointer;
      line: integer;
      end_sequence: Boolean;
      column, fileind: smallint; // it is sooo unlikely for them
    end;                         // to go beyond the 32767 limit...
    TFileInfo = packed record
      name: ansistring;
      dirind: integer;
    end;
    TDwarftable = array of TDwarfChunk;
    TCompilationUnit = record
      dTable: TDwarftable;
      dtable_length: integer;
      Files: array of TFileInfo;
      Dirs: array of ansistring;
      minAddr, maxAddr: ptruint;
    end;
    TExecutableUnit = record
      name: string;
      CompilationUnit: array of TCompilationUnit;
    end;
  var
    base_addr: pointer = nil;
    ExecutableUnit: array of TExecutableUnit;


{$packrecords c}

{ ELF Header structures types}
type
  Elf32_Half = Word;
  Elf64_Half = Word;
  { Types for signed and unsigned 32-bit quantities.   }
  Elf32_Word = DWord;
  Elf32_Sword = Longint;
  Elf64_Word = DWord;
  Elf64_Sword = Longint;
  { Types for signed and unsigned 64-bit quantities.   }
  Elf32_Xword = QWord;
  Elf32_Sxword = Int64;
  Elf64_Xword = QWord;
  Elf64_Sxword = Int64;
  { Type of addresses.   }
  Elf32_Addr = DWord;
  Elf64_Addr = QWord;
  { Type of file offsets.   }
  Elf32_Off = DWord;
  Elf64_Off = QWord;
  { Type for section indices, which are 16-bit quantities.   }
  Elf32_Section = Word;
  Elf64_Section = Word;
  { Type for version symbol information.   }
  Elf32_Versym = Elf32_Half;
  Elf64_Versym = Elf64_Half;
{ some constants from the corresponding header files }
const
  El_NIDENT = 16;
  { some important indices into the e_ident signature of an ELF file }
  EI_MAG0 = 0;
  EI_MAG1 = 1;
  EI_MAG2 = 2;
  EI_MAG3 = 3;
  EI_CLASS = 4;
  { the first byte of the e_ident array must be of this value }
  ELFMAG0 = $7f;
  { the second byte of the e_ident array must be of this value }
  ELFMAG1 = Byte('E');
  { the third byte of the e_ident array must be of this value }
  ELFMAG2 = Byte('L');
  { the fourth byte of the e_ident array must be of this value }
  ELFMAG3 = Byte('F');

  { the fifth byte specifies the bitness of the header; all other values are invalid }
  ELFCLASS32 = 1;
  ELFCLASS64 = 2;

  ELFCLASS = {$IFDEF CPU32}ELFCLASS32{$ENDIF}{$IFDEF CPU64}ELFCLASS64{$ENDIF};

type
   { The ELF file header.  This appears at the start of every ELF file, 32 bit version }
  TElf32_Ehdr = record
    e_ident : array[0..El_NIDENT-1] of Byte; { file identification }
    e_type : Elf32_Half; { file type }
    e_machine : Elf32_Half; { machine architecture }
    e_version : Elf32_Word; { ELF format version }
    e_entry : Elf32_Addr; { entry point }
    e_phoff : Elf32_Off; { program header file offset }
    e_shoff : Elf32_Off; { section header file offset }
    e_flags : Elf32_Word; { architecture specific flags }
    e_ehsize : Elf32_Half; { size of ELF header in bytes }
    e_phentsize : Elf32_Half; { size of program header entry }
    e_phnum : Elf32_Half; { number of program header entries }
    e_shentsize : Elf32_Half; { size of section header entry }
    e_shnum : Elf32_Half; { number of section header entry }
    e_shstrndx : Elf32_Half; { section name strings section index }
  end;

  { ELF32 Section header }
  TElf32_Shdr = record
    sh_name : Elf32_Word; { section name }
    sh_type : Elf32_Word; { section type }
    sh_flags : Elf32_Word; { section flags }
    sh_addr : Elf32_Addr; { virtual address }
    sh_offset : Elf32_Off; { file offset }
    sh_size : Elf32_Word; { section size }
    sh_link : Elf32_Word; { misc info }
    sh_info : Elf32_Word; { misc info }
    sh_addralign : Elf32_Word; { memory alignment }
    sh_entsize : Elf32_Word; { entry size if table }
  end;

  { The ELF file header.  This appears at the start of every ELF file, 64 bit version }
  TElf64_Ehdr = record
    e_ident : array[0..El_NIDENT-1] of Byte;
    e_type : Elf64_Half;
    e_machine : Elf64_Half;
    e_version : Elf64_Word;
    e_entry : Elf64_Addr;
    e_phoff : Elf64_Off;
    e_shoff : Elf64_Off;
    e_flags : Elf64_Word;
    e_ehsize : Elf64_Half;
    e_phentsize : Elf64_Half;
    e_phnum : Elf64_Half;
    e_shentsize : Elf64_Half;
    e_shnum : Elf64_Half;
    e_shstrndx : Elf64_Half;
  end;

  { ELF64 Section header }
  TElf64_Shdr = record
    sh_name : Elf64_Word;
    sh_type : Elf64_Word;
    sh_flags : Elf64_Xword;
    sh_addr : Elf64_Addr;
    sh_offset : Elf64_Off;
    sh_size : Elf64_Xword;
    sh_link : Elf64_Word;
    sh_info : Elf64_Word;
    sh_addralign : Elf64_Xword;
    sh_entsize : Elf64_Xword;
  end;

  TElf_Shdr = {$ifdef cpu32}TElf32_Shdr{$endif}{$ifdef cpu64}TElf64_Shdr{$endif};
  TElf_Ehdr = {$ifdef cpu32}TElf32_Ehdr{$endif}{$ifdef cpu64}TElf64_Ehdr{$endif};


  {$packrecords default}

var
  ExeFileName: ansistring;
  header : TElf_Ehdr;
  strtab_header : TElf_Shdr;
  cursec_header : TElf_Shdr;

  buf : array[0..20] of char;


type
  tdosheader = packed record
     e_magic : word;
     e_cblp : word;
     e_cp : word;
     e_crlc : word;
     e_cparhdr : word;
     e_minalloc : word;
     e_maxalloc : word;
     e_ss : word;
     e_sp : word;
     e_csum : word;
     e_ip : word;
     e_cs : word;
     e_lfarlc : word;
     e_ovno : word;
     e_res : array[0..3] of word;
     e_oemid : word;
     e_oeminfo : word;
     e_res2 : array[0..9] of word;
     e_lfanew : longint;
  end;
  tpeheader = packed record
     PEMagic : longint;
     Machine : word;
     NumberOfSections : word;
     TimeDateStamp : longint;
     PointerToSymbolTable : longint;
     NumberOfSymbols : longint;
     SizeOfOptionalHeader : word;
     Characteristics : word;
     Magic : word;
     MajorLinkerVersion : byte;
     MinorLinkerVersion : byte;
     SizeOfCode : longint;
     SizeOfInitializedData : longint;
     SizeOfUninitializedData : longint;
     AddressOfEntryPoint : longint;
     BaseOfCode : longint;
     BaseOfData : longint;
     ImageBase : longint;
     SectionAlignment : longint;
     FileAlignment : longint;
     MajorOperatingSystemVersion : word;
     MinorOperatingSystemVersion : word;
     MajorImageVersion : word;
     MinorImageVersion : word;
     MajorSubsystemVersion : word;
     MinorSubsystemVersion : word;
     Reserved1 : longint;
     SizeOfImage : longint;
     SizeOfHeaders : longint;
     CheckSum : longint;
     Subsystem : word;
     DllCharacteristics : word;
     SizeOfStackReserve : longint;
     SizeOfStackCommit : longint;
     SizeOfHeapReserve : longint;
     SizeOfHeapCommit : longint;
     LoaderFlags : longint;
     NumberOfRvaAndSizes : longint;
     DataDirectory : array[1..$80] of byte;
  end;
  tcoffsechdr=packed record
    name     : array[0..7] of char;
    vsize    : longint;
    rvaofs   : longint;
    datalen  : longint;
    datapos  : longint;
    relocpos : longint;
    lineno1  : longint;
    nrelocs  : word;
    lineno2  : word;
    flags    : longint;
  end;

  var
    dosheader  : tdosheader;
    peheader   : tpeheader;
    coffsec    : tcoffsechdr;
{$pop}


  procedure GetLInfo(addr: pointer;  var rec: TLineInfoRec);
    //var exe, src: ansistring; var line, column: integer);
  var
    i,j,k, ei: integer;
    b: boolean;
  begin
//addlog(' GetLInfo(%0)', [addr]);
    if not Assigned(lics) then lics:= syncobjs.TCriticalSection.Create;
    lics.Enter;

    with rec do begin
      src:='';
      exe:='';
      line:= -1;
      column:= -1;
      why_not:= _LineInfoError; //LineInfoError:= '';

      if not LineInfoInitialized then
        if not InitLineInfo(addr, false) then begin
          lics.Leave;
          exit;
        end;

      try
        exe:= ExtractFileName(GetExeFileName(addr));
//addlog('  exe is %0',[exe]);
        ei:= -1;
        for i:=0 to high(ExecutableUnit) do
          if ExtractFileName(ExecutableUnit[i].name) = exe then ei:=i;
        if ei < 0 then begin
          if not InitLineInfo(addr, false) then begin
            lics.Leave;
            exit;
          end;
          ei:= high(ExecutableUnit); // it gets added at the end, of course
        end;
//addlog('  ei is %0 (%1)',[ei, ExecutableUnit[ei].name]);
        with ExecutableUnit[ei] do
          for j:=0 to high(CompilationUnit) do
            with ExecutableUnit[ei].CompilationUnit[j] do begin

              if (ptruint(addr) < minAddr)
                or (ptruint(addr) > maxAddr)
              then
                continue;

              for i:=0 to high(dtable) do begin
//addlog('--- %4 | %0 %1 %2 (%7) %3 < %5 - %6 > ', [dtable[i].addr, dtable[i].line, Files[dtable[i].fileind].name, dtable[i].end_sequence, addr,(ptruint(addr) >= ptruint(dtable[i].addr)), (i < high(dtable)) and  (ptruint(addr) < ptruint(dtable[i + 1].addr)), dtable[i].fileind]);

                if (addr = dtable[i].addr)
                  or ((i < high(dtable))
                    and (ptruint(addr) >= ptruint(dtable[i].addr))
                    and (ptruint(addr) < ptruint(dtable[i + 1].addr))
                    and not dtable[i].end_sequence //(cardinal(addr) - cardinal(dtable[i].addr) < 1000)//
                  )
                then begin
//addlog('=============== OK =========================');
//addlog('            %4 | %0 %1 %2 (%7) %3 < %5 - %6 > ', [dtable[i].addr, dtable[i].line, Files[dtable[i].fileind].name, dtable[i].end_sequence, addr,(ptruint(addr) >= ptruint(dtable[i].addr)), (i < high(dtable)) and  (ptruint(addr) < ptruint(dtable[i + 1].addr)), dtable[i].fileind]);
//addlog('            CU#%1, high(files) = %0', [high(files), j]);
                  src:= //IncludeTrailingPathDelimiter(Dirs[Files[dtable[i].fileind].dirind])
                        //+
                        Files[dtable[i].fileind].name;
                  line:= dtable[i].line;
                  column:= dtable[i].column;
                  //now check if the same line appears twice with different columns.
                  // if not, then reset column to -1.
                  b:= false;
                  for k:=0 to high(dtable) do
                    b:= b or
                        (     (dtable[k].line = line)
                          and (dtable[k].column <> column)
                          and (dtable[k].fileind = dtable[i].fileind));
                  if not b then column:= -1;

                  why_not:= '';
                  lics.Leave;
                  Exit;
                end;
              end;
            end;
        why_not:= RuEn(
          'не охватывает указанный адрес',
          'doesn''t include this address range'
        );
      except
        why_not:= (ExceptObject as Exception).Message;
      end;
    end;
    lics.Leave;
  end;

 procedure CleanupLineInfo();
 begin
   if Assigned(lics) then lics.Free;
 end;

  function DlnNameByExename(exename: TFileNameString): TFileNameString;
  begin
    {$ifdef unix}
     if ExtractFileExt(exename) = ''
       then Result:= ChangeFileExt(exename, '.elf.zd2')
       else
    {$endif}
            Result:= ChangeFileExt(exename, '.zd2');
  end;

  function CheckMd5Sums(d2stream: TStream; Exename: ansistring): boolean;
  var
    buf: array[0..2047] of byte;
    d1r, d2r, d1c, d2c: TMD5Digest;
    fs: TFileStream;
    procedure CalcMd5Sum(i: TStream; len: integer; var d: TMD5Digest);
    var
      a: integer;
      c: TMd5Context;
    begin
      MD5Init(c);
      repeat
        a:= min(2048, len);
        i.Read(buf[0], a);
        MD5Update(c, buf[0], a);
        dec(len, a);
      until len = 0;
      Md5Final(c, d);
     end;
     function HashesMatch(var d1, d2: TMd5Digest): boolean;
     var j: integer;
     begin
       Result:= true;
       for j:=0 to sizeof(TMd5Digest)  - 1 do
         if d1[j] <> d2[j] then Exit(false);
     end;
  begin
    d2stream.Position:= 0;
    CalcMd5Sum(d2stream, d2stream.Size - 2 * sizeof(TMD5Digest), d2c);
    d2stream.Read(d1r, sizeof(d1r));
    d2stream.Read(d2r, sizeof(d2r));
    fs:= TFileStream.Create(exename, fmOpenRead or fmShareDenyNone);
    CalcMd5Sum(fs, fs.Size, d1c);
    fs.Free;
    Result:= HashesMatch(d1r, d1c) and HashesMatch(d2r, d2c)
  end;

  function InitLineInfo(someaddr: pointer; usecriticalsection: longbool = true): longbool;
  var
    dwarfsize: integer;
    dli, dc, ts: TStream;
    unit_length, unit_base, next_base: dword;
    header_length: SizeInt;
    header : TLineNumberProgramHeader32;
    state : TMachineState;
    numoptable : array[1..255] of Byte;
    i, din: integer;
    s: ansistring;

    opcode, extended_opcode : Byte;
    extended_opcode_length : Integer;
    adjusted_opcode : Int64;
    addrIncrement, lineIncrement: Integer;
    _dwarf: pointer;
    ExeImageBase: ptruint;

    filename, exname: TFileNameString;

    { Reads an unsigned LEB encoded number from the input stream }
    function ReadULEB128() : QWord;
    var
      shift : Byte;
      data : Integer;
      val : QWord;
    begin
      shift := 0;
      result := 0;
      dli.Read (data, 1);
      while (data <> -1) do begin
        val := data and $7f;
        result := result or (val shl shift);
        inc(shift, 7);
        if ((data and $80) = 0) then
          break;
        dli.Read (data, 1);
      end;
    end;


    { Reads a signed LEB encoded number from the input stream }
    function ReadLEB128() : Int64;
    var
      shift : Byte;
      data : Integer;
      val : Int64;
    begin
      shift := 0;
      result := 0;
      dli.Read (data, 1);
      while (data <> -1) do begin
        val := data and $7f;
        result := result or (val shl shift);
        inc(shift, 7);
        if ((data and $80) = 0) then
          break;
        dli.Read (data, 1);
      end;
      { extend sign. Note that we can not use shl/shr since the latter does not
        translate to arithmetic shifting for signed types }
      result := (not ((result and (1 shl (shift-1)))-1)) or result;
    end;

    procedure SkipLEB128();
    var temp : int64;
    begin
      temp := ReadLEB128();
      //DEBUG_ADDLOG('Skipping LEB128 : %0',[temp]);
    end;

    function CalculateAddressIncrement(_opcode : Byte) : Int64;
    begin
      result := (Int64(_opcode) - header.opcode_base) div header.line_range * header.minimum_instruction_length;
    end;


    function ReadString(): ansistring;
    var a: ansichar;
    begin
      Result:= '';
      while (true) do begin
        dli.Read(a, 1);
        if a = #0 then Exit;
        Result+= a;
      end;
    end;

    { initializes the line info state to the default values }
    procedure InitStateRegisters(partial: boolean = false);
    begin
      with state do begin
        if not partial then begin
          //a hack for DW_LNE_SET_ADDRESS with zero address
          address := 0;
          file_id := 1;
        end;
        line := 1;
        column := 0;
        is_stmt := (header.default_is_stmt <> 0);
        basic_block := false;
        end_sequence := false;
        prolouge_end := false;
        epilouge_begin := false;
        isa := 0;
        append_row := false;
      end;
    end;

    function ParseCompilationUnit(var CompilationUnit: array of TCompilationUnit;
      FirstPass: boolean = false): boolean;
    var
      j: integer;
      q: qword;
//      lastgood: ptruint = 0;
      stuba: cardinal;
      saved_state : TMachineState;
      procedure AddChunk;
      begin
        With CompilationUnit[high(CompilationUnit)] do begin
          inc(dtable_length);
          //SetLength(dtable, length(dtable) + 1); //This crap was re-allocating on the order of 800 Megabytes.
                                                   //It was killing my "never really deallocate" heap tracer
          if FirstPass then Exit;
          with dtable[dtable_length - 1{high(dtable)}] do begin
            //account for thepossible relocation (in 99% cases ExeImagebase = base_addr)

//            if ptruint(state.address) < ExeImagebase //WTF ??? :(
//              then begin
//                addr:= pointer(ptruint(state.address) + lastgood - ExeImagebase + ptruint(base_addr))
//              end
//              else begin
                addr:= pointer(ptruint(int64(state.address) + (int64(base_addr) - int64(ExeImagebase))));
//                lastgood:= state.address;
//              end;

//addlog ('                                                    *** %4 %3 ][ %0 %1 %2', [pointer(state.address), pointer(ExeImagebase), base_addr, addr, state.line]);
            line:= state.line; // should we add 1 here ?
            column:= state.column;
            fileind:= state.file_id - 1; //drats, caught overflow error here :(
            end_sequence:= state.end_sequence;
//            {$ifdef cge}
//              if {$ifdef cgemodule}Mother^.Debug.Verbose {$else}Mother^.State.DebugMode {$endif}//  or true
//                then
//                  if (fileind < 0) or (fileind > high(files))
//                    then AddLog('dwrflnfo: %0 %1:%2, INVALID fileind %3!', [addr, line, column, fileind])
//                    else AddLog('dwrflnfo: %0 %1:%2 %3', [addr, line, column, files[fileind].name]);
//            {$else}
//              DEBUG_WRITE('dwrflnfo: ', IntToHex(cardinal(addr),8), ' ', line,' ', column, ' ',files[fileind].name);
//            {$endif}
            if minAddr = 0 then minAddr:= ptruint(addr) else minAddr:= Math.min(minAddr, ptruint(addr));
            maxAddr:= Math.max(maxAddr, ptruint(addr));
          end;
        end;
      end;


    begin //ParseCompilationUnit
      Result:= true;
//addlog(' > ParseCompilationUnit %0',[FirstPass]);

//writeln('a ', FirstPass);

      if not FirstPass then begin
        //parse it *twice* just to determine the array size and only set it once

        fillchar(CompilationUnit[high(CompilationUnit)], sizeof(TCompilationUnit), 0);

        MOVE(state, saved_state, sizeof(state)); //annoyingly, some units
          //continue from the previous one, signaling this by trying to
          // set address to NIL. So, state must be saved between units
        ParseCompilationUnit(CompilationUnit, true); //2-level recursion
        With CompilationUnit[high(CompilationUnit)] do begin
          SetLength(dtable, dtable_length);
          dtable_length:= 0;
        end;
        MOVE(saved_state, state, sizeof(state));
      end;
//writeln('b');

      With CompilationUnit[high(CompilationUnit)] do begin
        minAddr:= 0;
        maxAddr:= 0;
      end;

     // a hack: the next compilation unit can have an unpredictable ofset,
     // so we try finding it by checking the most common values of the header
      j:= 0;
      repeat
        if unit_base + j + sizeof(header) + 2 > dli.Size then begin
          //DEBUG_ADDLOG('The stream end reached, no more units.');
          Exit(false);
        end;
        dli.position:= unit_base + j;
        dli.Read(header, sizeof(header));
        with header do begin
          if (version = 2) and (line_range = 255) and (opcode_base = 13) then begin
            unit_base+= j;
//ADDLOG('rand_offset=%8, p=%9, unit_length %0 version %1  length %2  min_instr_leng %3 def_is_stmt %4 line_base %5 line_range %6 opcode_base %7',        [unit_length, version, length, minimum_instruction_length, default_is_stmt, line_base, line_range, opcode_base, j, pointer(unit_base)]);
            header_length := sizeof(header.length) + sizeof(header.version)
                          + sizeof(header.unit_length)
                          + header.length;
            Break;
          end;
        end;
        inc(j);
      until false;

      next_base:= unit_base + header.unit_length + sizeof(header.unit_length);

      fillchar(numoptable, sizeof(numoptable), #0);
      dli.Read(numoptable, header.opcode_base - 1);
      //DEBUG_ADDLOG('Opcode parameter count table');
      for j := 1 to header.opcode_base - 1 do begin
        //DEBUG_ADDLOG('Opcode[%0] - %1 parameters', [j, numoptable[j]]);
      end;

      With CompilationUnit[high(CompilationUnit)] do begin
        SetLength(dirs, 1);
        dirs[0]:= ''; //the project directory
        while (true) do begin
          s:= ReadString();
          if (s = '') then break;
          SetLength(dirs, length(dirs) + 1);
          dirs[high(dirs)]:= s;
          //DEBUG_ADDLOG('Dir %0: %1',[high(dirs), AnsiToWide(s)]);
        end;


        while (true) do begin
          s:= ReadString;
          if (s = '') then break;
          SetLength(files, length(files) + 1);
          with files[high(files)] do begin
            name:= s;
            dirind:= ReadLEB128(); { the directory index for the file }
//aDDLOG('CU#%3, File %0 (dir %2): %1',[high(files), name, dirind, high(CompilationUnit)]);
          end;
          SkipLEB128(); { skip last modification time for file }
          SkipLEB128(); { skip length of file }
        end;


        dli.Position:= header_length + unit_base;
        unit_length:= header.unit_length;

        InitStateRegisters(true);
        state.file_id:= 1;

        while (dli.Position - unit_base) < unit_length - 2{ dli.Size - 1} do begin
          dli.Read(opcode, 1);
//if not FirstPass then ADDLOG('Next opcode: %0  (stream pos. %1 ( %2 / %3 )',[opcode, dli.position, dli.position - unit_base, unit_length]);

          case (opcode) of
            { extended opcode }
            0 : begin
              extended_opcode_length := ReadULEB128();
              dli.Read(extended_opcode, 1);
              case (extended_opcode) of
                DW_LNE_END_SEQUENCE : begin
                  state.end_sequence := true;
                  state.append_row := true;
                  AddChunk;
//if not FirstPass then ADDLOG('DW_LNE_END_SEQUENCE');
                  InitStateRegisters(true);
                  { Oops. Sometimes this one is followed
                    with DW_LNE_SET_ADDRESS to 0 - and kaboom
                    Created a hack for it.
                  }
                end;
                DW_LNE_SET_ADDRESS : begin
                  //dli.Read(state.address, 4);
                  dli.Read(stuba, 4); // a bug in dwarf generator?
                  if stuba <> 0 then state.address:= stuba;
//if not FirstPass then ADDLOG('DW_LNE_SET_ADDRESS (%0)', [pointer(stuba)]);
                end;
                DW_LNE_DEFINE_FILE : begin
                  {$ifdef DEBUG_DWARF_PARSER}s := {$endif}ReadString();
                  SkipLEB128();
                  SkipLEB128();
                  SkipLEB128();
//if not FirstPass then ADDLOG('DW_LNE_DEFINE_FILE (' + s + ')');
                end;
                else begin
//if not FirstPass then ADDLOG('Unknown extended opcode (opcode %0 length %1)', [extended_opcode, extended_opcode_length]);
                  dli.Position:= dli.Position + extended_opcode_length - 1;
                end;
              end;
            end;
            DW_LNS_COPY : begin
              state.basic_block := false;
              state.prolouge_end := false;
              state.epilouge_begin := false;
              state.append_row := true;
//if not FirstPass then ADDLOG('DW_LNS_COPY');
              AddChunk;
            end;
            DW_LNS_ADVANCE_PC : begin
              q:= ReadULEB128() * header.minimum_instruction_length;
//if not FirstPass then ADDLOG('DW_LNS_ADVANCE_PC (+' + IntToStr(q) + ')'); //+ hexstr(state.address, sizeof(state.address)*2) + ')');
              inc(state.address, q);
            end;
            DW_LNS_ADVANCE_LINE : begin
              inc(state.line, ReadLEB128());
//if not FirstPass then ADDLOG('DW_LNS_ADVANCE_LINE (%0)', [state.line]);
             // AddChunk;
            end;
            DW_LNS_SET_FILE : begin
              state.file_id := ReadULEB128();
              state.line:= 1;
//if not FirstPass then aDDLOG('DW_LNS_SET_FILE (%0|%1)', [state.file_id, Files[state.file_id - 1].name]);
            end;
            DW_LNS_SET_COLUMN : begin
              state.column := ReadULEB128();
//if not FirstPass then ADDLOG('DW_LNS_SET_COLUMN (%0)', [state.column]);
            end;
            DW_LNS_NEGATE_STMT : begin
              state.is_stmt := not state.is_stmt;
//if not FirstPass then ADDLOG('DW_LNS_NEGATE_STMT (%0)',[state.is_stmt]);
            end;
            DW_LNS_SET_BASIC_BLOCK : begin
              state.basic_block := true;
//if not FirstPass then ADDLOG('DW_LNS_SET_BASIC_BLOCK');
            end;
            DW_LNS_CONST_ADD_PC : begin
              inc(state.address, CalculateAddressIncrement(255));
//if not FirstPass then aDDLOG('DW_LNS_CONST_ADD_PC (' + hexstr(state.address, sizeof(state.address)*2) + ')');
            end;
            DW_LNS_FIXED_ADVANCE_PC : begin
              dli.Read(state.address, sizeof(state.address));
//if not FirstPass then ADDLOG('DW_LNS_FIXED_ADVANCE_PC (' + hexstr(state.address, sizeof(state.address)*2) + ')');
            end;
            DW_LNS_SET_PROLOGUE_END : begin
              state.prolouge_end := true;
//if not FirstPass then aDDLOG('DW_LNS_SET_PROLOGUE_END');
            end;
            DW_LNS_SET_EPILOGUE_BEGIN : begin
              state.epilouge_begin := true;
//if not FirstPass then ADDLOG('DW_LNS_SET_EPILOGUE_BEGIN');
            end;
            DW_LNS_SET_ISA : begin
              state.isa := ReadULEB128();
//if not FirstPass then ADDLOG('DW_LNS_SET_ISA (%0)', [state.isa]);
            end;
            else begin { special opcode }
              if (opcode < header.opcode_base) then begin
//if not FirstPass then ADDLOG('Unknown standard opcode $' + hexstr(opcode, 2) + '; skipping');
                for j := 1 to numoptable[opcode] do
                  SkipLEB128();
              end else begin
                adjusted_opcode := integer(opcode) - header.opcode_base;
                addrIncrement := CalculateAddressIncrement(opcode);
                inc(state.address, addrIncrement);
                lineIncrement := header.line_base + (adjusted_opcode mod header.line_range);
                inc(state.line, lineIncrement);
//if not FirstPass then ADDLOG('Special opcode $' + hexstr(opcode, 2) + ' address increment: %0 new line: %1', [addrIncrement, lineIncrement]);
                state.basic_block := false;
                state.prolouge_end := false;
                state.epilouge_begin := false;
                state.append_row := true;
                AddChunk;
              end;
            end;
          end; //case
        end; //while
      end; //with
      Result:= true;
    end; //ParseCompilationUnit


  begin //InitLineInfo
    if not Assigned(lics) then lics:= syncobjs.TCriticalSection.Create;
    if usecriticalsection then lics.Enter;
    if someaddr = nil then someaddr:= @system.Pos;
    base_addr:= GetExeBaseAddr(someaddr);
    filename:= GetExeFileName(someaddr);
    din:= -1;
    for i:=0 to high(ExecutableUnit) do
      if ExecutableUnit[i].name = filename then din:=i;
    if din < 0 then begin
      SetLength(ExecutableUnit, length(ExecutableUnit) + 1);
      din:= high(ExecutableUnit); // it gets added at the end, of course
      ExecutableUnit[din].name:= filename;
      ExecutableUnit[din].CompilationUnit:= nil;
    end
    else begin
      if usecriticalsection then lics.Leave;
      Exit(true); //already initialized for this exe/dll
    end;
    BackTraceStrFunc := @ChelinfoBacktraceStr;

    if Mother^.Debug.Verbose then begin
      AddLog('  %2 is %1, base %0', [base_addr, filename, someaddr]);
    end;


    try
      Result:= false;
     {  First, try the external file with line information.
        Failing that, try to parse the executable itself }

      exname:= DlnNameByExename(filename);
      i:= -1;
      repeat
        if Mother^.Debug.Verbose then addlog('  ..file "%0", exists=%1..', [exname, FileExists(exname)]);
        if FileExists(exname)
          then break
          else exname:='';
        inc(i);
        if i > high(LineInfoPaths) then break;
        exname:= LineInfopaths[i] + DlnNameByExename(ExtractFileName(filename));
      until false;

      if exname <> ''
        //and (FileAge(filename) <= FileAge(DlnNameByExename(filename)))
      then begin
        //the compression streams are unable to seek,
        //so we decompress to a memory stream first.
        ts:=TFileStream.Create(exname, fmOpenRead);
        dc:=TDecompressionStream.Create(ts);
        dli:= TMemoryStream.Create;
        dc.Read(DwarfSize, 4);
        dc.Read(ExeImageBase, 4);
        dli.CopyFrom(dc, DwarfSize);
        dc.Free;
        if not CheckMd5Sums(ts, filename) then begin
            _LineInfoError:=
              {$ifndef cgemodule}
                PervertedFormat(MessageContainer[MI_DEBUG_INFO_MISMATCH], [ExtractFileName(filename)]);
              {$else}
                MsgFormat(MI_DEBUG_INFO_MISMATCH, [ExtractFileName(filename)]);
              {$endif}
          ts.Free;
          if usecriticalsection then lics.Leave;
          Exit(false);
        end;
        ts.Free;
      end
      else begin
        if Mother^.Debug.Verbose then addlog('  External file not found, trying to parse self: "%0"..', [filename]);
        //DEBUG_WRITE('  External file not found, trying to parse self.');
        if not ExtractDwarfLineInfo(filename, _dwarf, DwarfSize, ExeImageBase) then begin
          _LineInfoError:= ExtractDwarfLineInfoError;
          lics.Leave;
          Exit(false);
        end;
        dli:= TMemoryStream.Create;
        dli.Write(_dwarf^, DwarfSize);
        FreeMem(_dwarf);
      end;

      dli.Position:= 0;
      next_base:= 0;

      if Mother^.Debug.Verbose then addlog('  dwarf2 line info: %0 bytes..', [dli.size]);
      //DEBUG_WRITE('dwarf2 line info: ',dli.size,' bytes.');

      InitStateRegisters(false);

      With ExecutableUnit[din] do
        while true {next_base < dli.Size - sizeof(header) - 5} do begin
          SetLength(CompilationUnit, length(CompilationUnit) + 1);
          unit_base:= next_base;

          if not ParseCompilationUnit(CompilationUnit) then begin
            SetLength(CompilationUnit, length(CompilationUnit) - 1);

            Break;
          end;
        end;
    except
      Die(RuEn('Крах при парсинге отладочной информации.','Crashed while parsing the self-debugging info.'))
    end;
    if {Result and} (length(ExecutableUnit[din].CompilationUnit) > 0)
      then begin
        LineInfoInitialized:= true;
        if Mother^.Debug.Verbose then
          AddLog(RuEn('  найдено %0 блоков.','  found %0 units.'), [length(ExecutableUnit[din].CompilationUnit)])
      end
      else begin
        if Mother^.Debug.Verbose then AddLog(RuEn(
          'Не найдена отладочная информация для'#10#13'  %0.',
          'No debug info found for'#10#13'  %0.'),
            [ExecutableUnit[din].name]);
        _LineInfoError:= 'no compilation units found.';
        Result:=false;
      end;
    dli.Free;
    Result:= True;
    if usecriticalsection then lics.Leave;
  end; //InitLineInfo

  function cntostr(cn: pAnsichar): Ansistring;
  var
    i: integer = 0;
  begin
    Result:='';
    repeat
      if cn^ = #0 then break;
      Result+= cn^;
      inc(i);
      inc(cn);
    until i = 8;
  end;



function ExtractDwarfLineInfo(
  ExeFileName: TFileNameString; var _dlnfo: pointer; var _dlnfoSize: integer;
    var Imagebase: cardinal): longbool;

var
  DwarfOffset : int64;
  DwarfSize : SizeInt;
  i : Integer;
  f : TFileStream;
  IsCompressed: boolean = false;
  DC: TDecompressionStream;

begin
  //DEBUG_ADDLOG('Reading dwarf line info from %0', [ExeFileName]);
  Result := false;

  if (ExeFileName = '') or not fileexists(ExeFileName) then begin
      ExtractDwarfLineInfoError:='Invalid executable file name, ' + ExeFileName;
      Exit(false);
  end;

  f:= TFileStream.Create(ExeFileName, fmOpenRead or fmShareDenyNone);
  DwarfOffset:= -1;
  DwarfSize:= -1;

  {$ifdef unix}

    if (f.read(header, sizeof(header)) <> sizeof(header)) then begin
      ExtractDwarfLineInfoError:='Could not read the ELF header!';
      f.Free;
      Exit(false);
    end;

    { more paranoia checks }
    if ((header.e_ident[EI_MAG0] <> ELFMAG0) or (header.e_ident[EI_MAG1] <> ELFMAG1) or
      (header.e_ident[EI_MAG2] <> ELFMAG2) or (header.e_ident[EI_MAG3] <> ELFMAG3)) then begin
      ExtractDwarfLineInfoError:='Invalid ELF magic header.';
      f.Free;
      Exit(false);
    end;

    if (header.e_ident[EI_CLASS] <> ELFCLASS) then begin
      ExtractDwarfLineInfoError:='Invalid ELF header bitness.';
      f.Free;
      Exit(false);
    end;

    { seek to the start of section headers }

    { first get string section header }
    f.Position:= header.e_shoff + (header.e_shstrndx * header.e_shentsize);
    if (f.read(strtab_header, sizeof(strtab_header)) <> sizeof(strtab_header)) then begin
      ExtractDwarfLineInfoError:='Could not read string section header';
      f.Free;
      Exit(false);
    end;

    for i := 0 to (header.e_shnum-1) do begin
      f.Position:= header.e_shoff + (i * header.e_shentsize);
      if (f.Read(cursec_header, sizeof(cursec_header)) <> sizeof(cursec_header)) then begin
        ExtractDwarfLineInfoError:='Could not read next section header';
        f.Free;
        Exit(false);
      end;
      { paranoia TODO: check cursec_header.e_shentsize }

      f.Position:= strtab_header.sh_offset + cursec_header.sh_name;
      if (f.Read(buf, sizeof(buf)) <> sizeof(buf)) then begin
        ExtractDwarfLineInfoError:='Could not read section name';
        Exit(false);
      end;
      buf[sizeof(buf)-1] := #0;

      //DEBUG_ADDLOG('This section is "%0", offset %1 size %2', [pAnsichar(@buf[0]), cursec_header.sh_offset, cursec_header.sh_size]);
      if (pAnsichar(@buf[0]) = ELFDlin) then begin
        //DEBUG_ADDLOG(ELFDlin + ' section found');
        DwarfOffset := cursec_header.sh_offset;
        DwarfSize := cursec_header.sh_size;
        { more checks }
        //DEBUG_ADDLOG(' offset %0,  size %1', [DwarfOffset, DwarfSize]);
        Result := (DwarfOffset >= 0) and (DwarfSize > 0);
        break;
      end;
    end;

    Imagebase:= 0;

 {$else}

    { read and check header }
    if f.Size < sizeof(dosheader) then begin
        ExtractDwarfLineInfoError:= 'Could not read the PE header';
        f.Free;
        Exit(false);
      end;
    f.Read(dosheader, sizeof(tdosheader));
    f.Position:= dosheader.e_lfanew;
    f.Read(peheader, sizeof(tpeheader));
    if peheader.pemagic<>$4550 then begin
        ExtractDwarfLineInfoError:= 'Not a valid Portable Executable';
        f.Free;
        Exit(false);
      end;
//writeln('--base, ', IntToHex(peheader.Imagebase,8));

    { read section info }
    for i:=1 to peheader.NumberOfSections do
     begin
       f.Read(coffsec, sizeof(tcoffsechdr));
       //DEBUG_ADDLOG(coffsec.name);
//writeln(coffsec.name, '  ', coffsec.datalen);

       if cntostr(@coffsec.name) = PEDlin then begin
         DwarfOffset:= coffsec.datapos;
         DwarfSize:= coffsec.datalen;
         break;
       end;

     end;
    Result:= (DwarfOffset > 0);
    ImageBase:= peheader.Imagebase;

 {$endif}

  if Result then begin
    if IsCompressed then begin
      f.Position:= DwarfOffset;
      DC:= TDecompressionStream.Create(f);
      DC.Read(DwarfSize, sizeof(DwarfSize));
      DC.Read(ImageBase, sizeof(ImageBase));
      _dlnfoSize:= DwarfSize;
      GetMem(_dlnfo, DwarfSize);
      DC.Read(_dlnfo^, DwarfSize);
      DC.Free;
    end
    else begin
      GetMem(_dlnfo, DwarfSize);
      _dlnfoSize:= DwarfSize;
      f.Position:= DwarfOffset;
      f.Read(_dlnfo^, DwarfSize);
    end;
  end
  else
    ExtractDwarfLineInfoError:=
    {$ifdef unix}
      'The line info section not found in the ELF file.'
    {$else}
      'The line info section not found in the PE file.'
    {$endif}
    ;
  f.Free;
end;


{$pop}

