{
    This file is part of Chentrah,
    Copyright (C) 2004-2008 Anton Rzheshevski (chebmaster@mail.ru).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see http://www.gnu.org/licenses/

 **********************************************************************

    This file contains various routines for forming comprehensive
      exception reports and other error messages.

 **********************************************************************}
 
 
 // Requires the "chelinfo" unit ! and -gw to include the dwarf 2 debugging information!

  function LimitMessage(m: ansistring; limit1, limit: integer; startspace: ansistring): ansistring;
  var i, j, L: integer;
  begin
    Result:=m;
    if length(Result) <= limit1 then Exit;
    limit-= length(startspace);
    L:= limit1;
    j:=1;
    repeat
      i:=0;
      while (Result[L - i] <> ' ') and ((L - i) > (j + 10)) do inc(i);
      L-= i;
      Result:=Copy(Result, 1, L) + #10#13 + startspace + Copy(Result, L + 1, Length(Result) - L);
      j:= L + 2 + length(startspace);
      L:= j + limit;
    until L >= length(Result);
  end;

  function _ExpExpAddress(addr: pointer; brief: boolean): UnicodeString;
  var
 {$ifdef windows}
    Tmm: TMemoryBasicInformation;
 {$endif}
    Lir: TLineInfoRec;
    err: UnicodeString;
    a: pointer;
  begin
    if not Assigned(addr) then Exit (RuEn(
      'попытка выполнить код по нулевому адресу',
      'attempt to execute code at null address'));

    GetLInfo(addr, Lir);

    if (Lir.line < 1) then begin
      if (Lir.exe = '') then begin
        {$ifdef windows}
          if (VirtualQuery(addr, @Tmm, SizeOf(Tmm)) = SizeOf(Tmm))
            and (Tmm.State <> MEM_COMMIT)
          then Lir.func:= RuEn(
            'в невыделенном адресном пространстве',
            'in unallocated address space')
          else
        {$else}
            Lir.exe:= RuEn(
              '<неизвестном исполняемом модуле>',
              '<unknown executable module>')
        {$endif}
      end;
      err:= Lir.why_not;
      if Length(err) > 30 then err:= '  '#10#13 + err;
      if Lir.func <> '' then Lir.func:= ' (' + Lir.func + ') ';
      if brief then  Result:= PervertedFormat(RuEn(
          '%0%2 в %1',
          '%0%2 in %1')
          , [addr, Lir.exe, Lir.func])
      else Result:= PervertedFormat(RuEn(
          'по адресу %0%3 в %1'#10#13'  (нет отладочной информации: %2)',
          'at %0%3 in %1'#10#13'  (no debug info available: %2)')
          , [addr, Lir.exe, err, Lir.func])
    end
    else begin
      if Lir.func <> '' then Lir.func:= ' (' + Lir.func + ') ';
      if Lir.column < 1
      then Result:= PervertedFormat(RuEn(
        '%0:%1%3 в %2',
        '%0:%1%3 in %2'),
        [Lir.src, Lir.line, Lir.exe, Lir.func])
      else Result:= PervertedFormat(RuEn(
        '%0:%1:%3%4 в %2',
        '%0:%1:%3%4 in %2'),
        [Lir.src, Lir.line, Lir.exe, Lir.column, Lir.func]);
    end;
  end;

  function ExpExpAddress(addr: pointer): UnicodeString;
  begin
    Result:= _ExpExpAddress(addr, false)
  end;

  function ShortExpAddr(addr: pointer): UnicodeString;
  begin
    Result:= _ExpExpAddress(addr, true)
  end;



  {$ifdef cgekernel}
    function MsgFormat(M: TMessageId; Param: array of const): UnicodeString;
    begin
      Result:= PervertedFormat(MessageContainer[M], Param);
    end;

    function MsgRaw(M: TMessageID): UnicodeString;
    begin
      Result:= MessageContainer[M];
    end;
  {$endif}
  

  function TellException(E: Exception): UnicodeString;
  var
    m, n: UnicodeString;
  begin
   if not (E is Exception) then Exit ('Not an Exception');
   try
     if E is EAbstractERROR then Exit(MsgRaw(MI_EABSTRACT_EXPLAIN));
     if E is EAccessViolation then begin
       if cardinal(ExceptAddr) < 1024
         then Exit(MsgRaw(MI_ET_ACCESS_VIOLATION_CALL_TO_0))
 //        else Exit(MsgFormat(MI_ET_ACCESS_VIOLATION, [ExpExpAddress()]))
     end;
   //  Result:= MsgFormat(MI_SIMPLE_EXCEPTION_TALE, [trim(E.ClassName), AnsiToWide(E.Message)]);
     n:= trim(E.ClassName);
     if E.Message = 'Unknown Run-Time error : 210'
       then m:= RuEn(
         'Вызов виртуального метода у неинициализированного объекта',
         'Call of a virtual metod on an uninitialized object')
         + ' (Run-Time error 210)'
{
     else if E.Message = 'Unknown Run-Time error : xxx'
       then m:= RuEn(
         '',
         '')
         + ' (Run-Time error xxx)'
}

       else m:= AnsiToWide(LimitMessage(E. Message, 60 - length(n), 60, '  '));
     Result:= PervertedFormat(RuEn(
        '%0: %1'#10#13'  %2',
        '%0: %1'#10#13'  %2'),
        [n, m, ExpExpAddress(ExceptAddr())]);
   except
     Result:='Crashed trying to explain exception:'#10#13'  ' + (ExceptObject as Exception).Message;
   end;
  end;

{$ifdef windows}
 function SehHackTellExceptionAddress(a: ptruint): UnicodeString;
 begin
  Result:= ExpExpAddress(pointer(a));
 end;
{$endif}




  function ToldMemRgn(ptr: pointer): UnicodeString;

 {$ifndef windows}
  begin
    Result:='';//AnsiToWide('ChebLib''s tale-telling memory regions is implemented only for 32-bit Micro$oft Windows.');
  end;
 {$else windows}
  var
    S: UnicodeString;
    mm, ml, mr: _MEMORY_BASIC_INFORMATION;
    A: ptruint;
    function DescMemProp(AB, Protect, State, _Type: ptruint): UnicodeString;
    var TST: array[0..Max_Path + 1] of {$ifdef che_unicode} UnicodeChar {$else} AnsiChar {$endif};
    begin
      Case State of
       MEM_RESERVE: Result:=RuEn('зарезервирована', 'reserved');
       MEM_FREE: Result:= RuEn('свободна', 'free');
       MEM_COMMIT: begin
          if {$ifdef che_unicode}GetModuleFileNameW{$else}GetModuleFileNameA{$endif}
                        (THandle(AB), @TST, SizeOf(TST) div SizeOf(TST[0])) <> 0
            then Result:= PervertedFormat(RuEn('выделена %0','committed by %0'), [ExtractFileName(PFileNameCharToFileNameString(@TST))])
            else Result:= PervertedFormat(RuEn('выделена (база %0)','committed (base %0)'), [pointer(AB)]);

          If (_Type <> MEM_PRIVATE) then
            Result+= RuEn(', общая',', shared');
          If (Protect and PAGE_NOCACHE > 0) then
            Result+= RuEn(', некэшируемая',', non-cached');
          If (Protect and PAGE_GUARD > 0) then
            Result+= RuEn(', охраняемая',', guarded');
          Result+= ', ';//RuEn(', доступ: ',', access: ');
          Protect:= Protect and not (PAGE_NOCACHE + PAGE_GUARD);
          case Protect of
            PAGE_READONLY: Result+= RuEn('только чтение','read only');
            PAGE_READWRITE: Result+= RuEn('чтение/запись','read/write');
            PAGE_WRITECOPY: Result+= RuEn('копируется при записи','copy-on-write');
            PAGE_EXECUTE: Result+= RuEn('только выполнение кода','execute only');
            PAGE_EXECUTE_READ: Result+= RuEn('чтение и выполнение кода','read and execute');
            PAGE_EXECUTE_READWRITE: Result+= RuEn('полный','full');
            PAGE_EXECUTE_WRITECOPY: Result+= RuEn('выполнение кода/копируется при записи','execute/copy-on-write');
            PAGE_NOACCESS: Result+= RuEn('запрещён','denied');
          else
            result+= RuEn('неизвестен','unknown');
          end;
        end;
      end;
    end;
  begin
    Result:='';
    A:=ptruint(ptr);
    if VirtualQuery(Pointer(A), @mm, SizeOf(mm)) <> SizeOf(mm) then Exit;
//    if (cardinal(mm.BaseAddress) >= 0) then
      if VirtualQuery(Pointer(ptruint(mm.BaseAddress)-1), @ml, SizeOf(ml))
      = SizeOf(ml)
      then
        Result:=PervertedFormat('    %0..%1 (%2K): %3;'#10#13,
          [ml.BaseAddress, ml.BaseAddress + ml.RegionSize-1, ml.RegionSize div 1024,
           DescMemProp(ptruint(ml.AllocationBase), ml.Protect, ml.State, ml._Type)]);
    Result+= PervertedFormat('    %0..%1 (%2K): %3',
      [mm.BaseAddress, mm.BaseAddress + mm.RegionSize - 1, mm.RegionSize div 1024,
        DescMemProp(ptruint(mm.AllocationBase),mm.Protect, mm.State, mm._Type)]);
    if (ptruint(mm.BaseAddress) - 1 + mm.RegionSize  < {$ifdef cpu64}$FFFFFFFFFFE{$else}$FFFFFFFE{$endif}) then
      if VirtualQuery(Pointer(ptruint(mm.BaseAddress) + mm.RegionSize), @mr, SizeOf(mr)) = SizeOf(mr)
      then begin
        Result+= PervertedFormat(';'#10#13'    %0..%1 (%2K): %3',
          [mr.BaseAddress, mr.BaseAddress + mr.RegionSize, mr.RegionSize div 1024,
           DescMemProp(ptruint(mr.AllocationBase), mr.Protect,mr.State,mr._Type)]);
      end;
    Result+= '.';
  end;
 {$endif windows}


