procedure EnterCriticalSectionWv(var cs: TRtlCriticalSection); inline;
var
  oto: TThreadLoadKind;
begin
  oto:= NoteThreadOccupation(tlk_WaitingSync);
  EnterCriticalSection(cs);
  NoteThreadOccupation(oto);
end;

procedure EnterWv(cs: TCriticalSection; CountLoadAs: TThreadLoadKind = tlk_WaitingSync); inline;
var
  oto: TThreadLoadKind;
begin
  oto:= NoteThreadOccupation(CountLoadAs);
  cs.Enter;
  NoteThreadOccupation(oto);
end;

procedure TProfiledCriticalSection.Enter(CountLoadAs: TThreadLoadKind = tlk_WaitingSync);
var
  oto: TThreadLoadKind;
begin
  oto:= NoteThreadOccupation(CountLoadAs);
  inherited Enter;
  NoteThreadOccupation(oto);
end;

function NoteThreadOccupation(o: TThreadLoadKind): TThreadLoadKind;
var
  pes: PMotherSehState;
 {$ifdef cpuarm}
   oldc: timeval;
 {$endif}
begin
  if not Assigned(Mother) or not Mother^.Safe then Exit(tlk_Working);
  pes:= GetExceptionStateIfExists();
  if not Assigned(pes) then Exit(tlk_Working);
  {$ifdef cpuarm}
    {$push}
    {$overflowchecks off}
    {$rangechecks off}
    oldc:= pes^.tlitsc;
    fpgettimeofday(@(pes^.tlitsc), NIL);
    pes^.tli[pes^.tlic]+=
      (int64(1000000) * int64(pes^.tlitsc.tv_sec - oldc.tv_sec))
      + int64(pes^.tlitsc.tv_usec - oldc.tv_usec);
    {$pop}
  {$else}
    {$ifdef cpu64}
      asm
        mov rsi, [pes]
        mov rcx, qword[rsi + TMotherSEHState.tlitsc]
        mov eax, 0
        rdtsc
        mov dword[rsi + TMotherSEHState.tlitsc], eax;
        mov dword[rsi + 4 + TMotherSEHState.tlitsc], edx;
        shl rdx, 32
        add rdx, rax
        sub rdx, rcx
        xor rcx, rcx
        mov ecx, dword[rsi + TMotherSEHState.tlic]
        add qword[rsi + rcx * 8 + TMotherSEHState.tli], rdx
      end ['rax', 'rcx', 'rdx', 'rsi'];
    {$else}
      asm
        mov esi, [pes]
        mov ecx, dword[esi + TMotherSEHState.tlitsc]
        mov edi, dword[esi + 4 + TMotherSEHState.tlitsc]
        rdtsc
        mov dword[esi + TMotherSEHState.tlitsc], eax;
        mov dword[esi + 4 + TMotherSEHState.tlitsc], edx;
        sub eax, ecx
        sbb edx, edi
        mov ecx, dword[esi + TMotherSEHState.tlic]
        add dword[esi + ecx * 8 + TMotherSEHState.tli], eax
        adc dword[esi + ecx * 8 + 4 + TMotherSEHState.tli], edx
      end ['eax', 'ecx', 'edx', 'esi', 'edi'];
    {$endif}
  {$endif}
  Result:= pes^.tlic;
  pes^.tlic:= o;
end;

function GetExceptionStateIfExists(): PMotherSehState;
var
  pes: PMotherSehState;
  T: TThreadId;
begin
  Result:= pesThreadVar;
  if not Assigned(Result) then begin
    T:= GetCurrentThreadId();
    pes:= @Mother^.ExceptionState;
    while Assigned(pes) do begin
      if T = pes^.ThreadId then begin
        Result:= pes;
        break;
      end;
      pes:= pes^.Next;
    end;
    pesThreadVar:= Result;
  end;
end;

