{
    This file is part of Chentrah,
    Copyright (C) 2004-2015 Anton Rzheshevski (chebmaster@mail.ru).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see http://www.gnu.org/licenses/

 **********************************************************************}

const
  VersionMajor = 0;
  VersionMinor = 20;
  CgeStringEn = 'Chentrah';
  CgeStringRu = 'Чентра';
  CopyrightEn = '2004-2016, Anton Rzheshevski';
  CopyrightRu = '2004-2016, Антон Ржешевский';
  CopyrightLibsRu =
    'библиотеку Vampyre Imaging, © 2007-2015 Marek Mauder,';
  CopyrightLibsEn =
    'Vampyre Imaging Library, © 2007-2015 Marek Mauder';
//  SplashScreenNumVariants = 11;
//  SplashScreenPath = 'modules/cge/splash/';
  MyAppNick = 'chentrah';
  MyAppName = 'chentrah';

type

{$include un_globaldefs.h}

{$ifdef che_unicode}
  {$define che_unicode}  //please note that this is duplicated in some units due to circular reference

  {$ifdef che_legacy}
    {$fatal Compile the legacy version using FPC 2.6.4!}
  {$endif}

  {$if FPC_FULLVERSION<30000}
    {$fatal Impossible to use this compiler version: RTL is not unicode} //The sad truth
  {$endif}

  { modeswitch unicodestrings} //is in the main project file

  { warn IMPLICIT_STRING_CAST Error}
  {$warn IMPLICIT_STRING_CAST_LOSS Error}
  { warn EXPLICIT_STRING_CAST Error}
  {$warn EXPLICIT_STRING_CAST_LOSS Error}
  { if FPC_FULLVERSION<=40000} //***TODO Replace with version in which it is fixed
     {$define fix_fpc3_unicode} // Use custom mods of TFileStream, TStringList and TIniFile
  { endif}

  AnsiString1251 = type AnsiString(1251);
  TFileNameString = UnicodeString;
  PFileNameChar = PUnicodeChar;

{$else}
  //Legacy Free Pascal below 3.0 (2.6.0 to 2.6.4)
  RawByteString = AnsiString;
  AnsiString1251 = AnsiString;
  {$ifdef windows}
    //no unicode support for file names
    TFileNameString = AnsiString;
    PFileNameChar = PAnsiChar;
  {$else}
    TFileNameString = Utf8String;
    PFileNameChar = PAnsiChar;
  {$endif}
{$endif}


  float = single;
  boolean = longbool;

  TIntegerArray = array of integer;
  TByteArray = array of byte;

  type
    TFileNameStringArray = array of TFileNameString;

const
  Yes = True;
  No = False;

  floaterror = 1e-5;

  MAX_MODULE_THREADS =
     {$ifdef cpu32}
       3 //The fuckers exhaust 32-bit virtual address space
         // because Chepersy dictates humongous stack sizes!
     {$else}
       5 //overkill much
     {$endif}
     ;

//  {$ifdef chepersy_align_4}
    AlignGranularity = 4;
    BasicUnitSize = 4;
(*  {$endif}
  {$ifdef chepersy_align_8}
    {$fatal 64-bit alignment support not implemented yet!}
    //AlignGranularity = 8;
    //BasicUnitSize = 4;
  {$endif}
*)

     MaxPointerHexChars = {$ifdef cpu32} 8 {$else} {$ifdef windows} 11 {$else} 16 {$endif} {$endif};
     //Windows 7 has a 8 Tb address space limit, I do not know about Linux

//for Cheb's Game Engine only:
{$ifdef cge}

{$include un_keys.h}

type
  //only add new ones at the end, to avoid breaking backward compatibility
  // with the existing modules. Never remove any.
  TOsType = (
    ostLinux,
    ostWine,
    ostDarwine, //not supported
    ostWinXP,
    ostWin2k,
    ostWin98, //NOT supported, left here to avoid changing lots of code
    ostVista,
    ostWin7,
    ostMacOsX, //NOT supported
    ostWin8,
    ostWin81,
    ostWin10,
    ostLinuxARM //Raspbian
  );

const
  //Note to self: I added and debugged this part back
  //  when a PC with RAM size of of 0.25 Gb was not uncommon
  //  and 0.5 Gb was the norm. Should I remove this altogether?
  OsMinMem: array[TOSType] of integer = (
    // if you have less physical RAM than this, Chentrah will refuse to run.
    // in megabytes.
    200, //lin
    200, //wine
    0,   //darwine
    450, //xp
    200, //2k
    200, //98
    900, //vista
    0,   //7
    0,   //mac
    0,   //8
    0,   //8.1
    0,    //10
    0    //Raspbian
  );
  OsMinMemRep: array[TOSType] of integer = (
    // as it is shown to the user
    256, //lin
    256, //wine
    0,   //darwine
    512, //xp
    256, //2k
    256, //98
    1024,//vista
    0,   //7
    0,   //mac
    0,   //8
    0,   //8.1
    0,    //10
    0  //Raspbian
  );

  OsReqMem: array[TOSType] of integer = (
    // if you have less physical RAM than this,
    // Chentrah will whine about not having enough memory.
    450, //lin
    450, //wine
    0,   //darwine
    900, //xp
    450, //2k
    450, //98
    1800,//vista
    0,   //7
    0,   //mac
    0,   //8
    0,   //8.1
    0,    //10
    0    //Raspbian
  );

  OsReqMemRep: array[TOSType] of integer = (
    //as it is shown to the user
    512, //lin
    512, //wine
    0,   //darwine
    1024,//xp
    512, //2k
    512, //98
    2048,//vista
    0,   //7
    0,   //mac
    0,   //8
    0,   //8.1
    0,    //10
    0    //Raspbian
  );


  OsDisplayedName: array[TOsType] of string = (
    'Linux',
    'Wine',
    'Darwine',
    'Windows XP',
    'Windows 2000',
    'Windows 98/Me',
    'Windows Vista',
    'Windows 7',
    'MacOS X',
    'Windows 8',
    'Windows 8.1',
    'Windows 10',
    'Linux (ARM)' //Raspbian
  );

  //If the current OS is not in this list, give a scary warning at startup:
  //Temporarily not listing Windows 98 as supported. Sorry about this
  SupportedOSes: set of TOSType = [
    ostLinux,
    ostLinuxARM, //Raspbian
    //{$ifdef che_unicode}
      ostWine, //Legacy is buggy due to http://bugs.freepascal.org/view.php?id=27221
      //P.S. Had built layers of hackarounds for that. Support is platinum now.
    //{$endif}
    //ostWin98, //As soon as I find time to test under it :(
                 //P.S. R.I.P. Support dropped forever.
    ostWinXP,
    ostWin7,
    {$ifdef nowin10}
       //the special version optimized for Windows XP
       ostWin8, ostWin81
    {$else}
       ostWin10
    {$endif}
   ];
  {
    After the outrage of Kde 4, Gnome 3 and Ubuntu 11.10 (where pen tablet still
    glitches making it not usable) I'm *very* tempted to just drop the native
    Linux support. Wine is enough.

    The lack of a sane universal API does that to you.
    Maybe later I'll rework the framework to QT, but as of 2012,
                                                        Linux is NOT supported.
  }

  //just setting a completely transparent cursor is not enough,
  // we need to actually employ ShowCursor(False);
  OSNeedsMousePointerHack: set of TOSType = [ostVista, ostWin7, ostWin8, ostWin81, ostWin10];
  
  WindowsEmulators: set of TOSType = [ostWine, ostDarwine];

  WindowsShouldHaveBuiltInOpenAL: set of TOSType = [];
                         //WTF, they don't? Microsoft lied.  ostVista, ostWin7];

  UnableToDisplayRussianTextInDialogs: set of TOSType =
                                     [ostLinux, ostLinuxARM, ostWine, ostDarwine, ostMacOsX];
  //I mean MessageBoxA and xmessage here. The eternal character set problem.

  CGM_PRESS = 1; //(scancode, mx, my)
  CGM_RELEASE = 2; //(scancode, mx, my)
  CGM_MOUSEMOVE = 3; //(mx, my) - is used for GUI, in window client pixels
  CGM_MOUSEDELTA = 4; //(mx, my) - is used when grabbed (first person 3d),
  CGM_TYPE = 5; //(char, mx, my)
  CGM_PEN = 6; //x, y, pressure (all floats), reserved1, reserved2, reserved3
  CGM_JOYSTICK = 7; //type (integer), x, y, dx, dy (floats). x & y are ranged -1.0..1.0 (or 0.0..1.0 depending on the type), dx & dy measured in 1/second

  { Quality Factor is a global setting controlling the load balance
      100 is considered "average", with FBO size equalling the viewport size.

      The algorithms controlling it consider mostly the averaged SwapBuffers
      duration, but also events of large textures being not resident as they are
      bound, and so on.
      Also, the more factual FPS deviates from desired FPS, the more weight
      it has in the formula.
  }
  QF_MIN = 10; { Below 20, the emergency low-res font (built in into the executable)
    is used instead of the normal font }
  QF_MAX = 2999; { When laptop is running on battery,
      a hard-coded upper limit of 100 is used instead }

  FPS_CULL_DEFAULT = 30; //also, values of Mother^.Timer.CullFrameRate below this are ignored

  SECONDS_PER_DAY = 86400.0;

type

  TModuleState = (ms_NotInitialized, ms_Loading, ms_Unloading, //ms_Exiting,
    ms_Loaded, ms_Crashed, ms_WaitingForModuleToFinish, ms_DoNothing,
    ms_DoNothingUntilMotherSavesTheSession, ms_CrashedAndDone);

  //displays them bottom to top in this order.
  TFpsCounterParam = ( fcp_ThreadLoadsVerbose, fcp_QualityFactor,
    fcp_FramesPerSecond, fcp_RdtscFrequency, fcp_SwapBuffersTime,
    fcp_ThreadLoads);
  TFpsCounterParams = set of TFpsCounterParam;
  ShortString3 = String[3];


type
  TJoystickType = (JoytypePrimary, JoytypeRightThumb, JoytypeLeftTrigger, JoytypeRightTrigger);
  PInputEvent = ^TInputEvent;
  TInputEvent = packed record
    case EventType: dword of
    0: (
      Stub: array[1..9] of dword;
    );
    CGM_PRESS {also CGM_RELEASE and CGM_MOUSEMOVE} : (
      key: TKey;
      mx, my: float; //formerly integer. Must allow for subpixel precision if necessary.
    );
    CGM_TYPE: (
       character: UnicodeChar; //also mx and my from the previous part. Must not overlap.
    );
    CGM_PEN: (
      PenX, PenY, PenPressure, PenReserved1, PenReserved2, PenReserved3: float;
    );
    CGM_JOYSTICK: (
      JoyType: TJoystickType;
      JoyX, JoyY, JoydX, JoydY: float;
    );
  end;

  TGAPIQuirksEnum = (
    OpenGLQuirkScissorFailed,
      //glScissor + glClear caused fallback to software rendering
      //Use rendering a quad instead to clear a part of a FBO
      //Found on: GeForce FX 5200
      //Detected by: timing.

    OpenGLQuirkFboCantGL_RGBA8
      //FBO supports GL_RGB8 and GL_RGBA16 but not GL_RGBA8
      //Found on: GeForce 7025, Ubuntu 10 + wine 1.0
      //Detected by: failing to create a 256x256 FBO with a GL_RGBA8 color buffer
  );
  TGAPIQuirksSet = set of TGAPIQuirksEnum;

  GLenum     = Cardinal;      PGLenum     = ^GLenum;
  GLboolean  = Byte;          PGLboolean  = ^GLboolean;
  GLbitfield = Cardinal;      PGLbitfield = ^GLbitfield;
  GLbyte     = ShortInt;      PGLbyte     = ^GLbyte;
  GLshort    = SmallInt;      PGLshort    = ^GLshort;
  GLint      = Integer;       PGLint      = ^GLint;
  GLsizei    = Integer;       PGLsizei    = ^GLsizei;
  GLubyte    = Byte;          PGLubyte    = ^GLubyte;
  GLushort   = Word;          PGLushort   = ^GLushort;
  GLuint     = Cardinal;      PGLuint     = ^GLuint;
  GLfloat    = Single;        PGLfloat    = ^GLfloat;
  GLclampf   = Single;        PGLclampf   = ^GLclampf;
  GLdouble   = Double;        PGLdouble   = ^GLdouble;
  GLclampd   = Double;        PGLclampd   = ^GLclampd;
{ GLvoid     = void; }        PGLvoid     = Pointer;
                              PPGLvoid    = ^PGLvoid;

  TGLenum     = GLenum;
  TGLboolean  = GLboolean;
  TGLbitfield = GLbitfield;
  TGLbyte     = GLbyte;
  TGLshort    = GLshort;
  TGLint      = GLint;
  TGLsizei    = GLsizei;
  TGLubyte    = GLubyte;
  TGLushort   = GLushort;
  TGLuint     = GLuint;
  TGLfloat    = GLfloat;
  TGLclampf   = GLclampf;
  TGLdouble   = GLdouble;
  TGLclampd   = GLclampd;

  PPVector2f = ^Pvector2f;
  PVector2f = ^Tvector2f;
  TVector2f = packed array[0..1] of GLfloat;
  PPVector3f = ^PVector3f;
  PVector3f = ^TVector3f;
  TVector3f = packed array [0..2] of GLfloat;
  PPVector4f = ^PVector4f;
  PVector4f = ^TVector4f;
  TVector4f = packed array [0..3] of GLfloat;
  PMatrix4f = ^TMatrix4f;
  TMatrix4f = packed array [0..3] of TVector4f;
  glFloatArray = packed array of glFloat;
  glFloat2dArray = array of glFloatArray;


  TThreadLoadKind = (tlk_Sleeping,  tlk_Working,
    tlk_ActualizingAssets, tlk_MainThreradRender, tlk_MainThreadBackgroundTasks,
    tlk_PollingInput, tlk_ConsoleRender, tlk_SwapBuffers,
    tlk_Logic, tlk_GamePhysics, tlk_ExtraAI,
    tlk_WaitingSync);
const
  ThreadLoadKindAbbr: array[TThreadLoadKind] of ShortString3 = (
     'slp', 'wrk',
     'AAs', 'mtR', 'mtb',
     'inp', 'coR', 'SWB',
     'LOG', 'Gph', '+AI',
     'wsy'
  );
  ThreadLoadKindColor: array[TThreadLoadKind] of TVector3f = (
    (0.55,0.55,0.55 {is never really drawn}), (0.3, 1, 0.3),
    (1,0.7,0.2),(1,0.4,0.7),(0.6,0,0.3),
    (0.7,0.7,0),(0.6,0.4,0),(1,0.1,0.1),
    (0.3,0.6,1),(0,0.8,0.8),(0.6,0.3,1),
    (0.9, 0.9, 0.9)
  );

type
  TCbMessage = (Re_OnDestroy, Re_OnCycle, Re_OnUnload, //the module knows only these three
                       Re_OnGetFocus, Re_OnLoseFocus, Re_Stub);

  PMd5Digest = ^TMd5Digest;
  PAssetHash = ^TAssetHash;
  TAssetHash = packed record //128 bit
    //The hash uniqueness: with TSC counting at 2..5 GHz since power-up
    // and DateTime increasing at 18 times per second with the passage of time,
    // each "randomly generated" hash is guaranteed to be unique (if hash
    // generation uses locking)
    // "random" hash vs file date/size: with file size cleverly using the same
    // memory slot as the date of the "random" hash, and with the upper 32 bits
    // of a reasonably sized file (below 4Gb) guaranteed to be zero and upper
    // 32 bits of Tdatetime guaranteed to be non-zero, there is no risk
    // that a random hash would ever match a file hash
    case integer of
      0: //pseudorandom hash of guaranteed uniqueness
      (
        ClassNameMd5: TMd5Digest;
        Date: double; //TDateTime
        TSC: QWord;
      );
      1: //file date & size hash
      (
        NameMd5: TMd5Digest;
        FileSize: int64;
        FileAge: double;
      );
      3: (
        Q1, Q2, Q3, Q4: QWord;
      );
  end;
{  PAssetRecord = ^TAssetRecord;
  TAssetRecord = packed record
    Hash: TAssetHash;
    Handle,
    owner,
    cpsInd //required for deleting lost resources,
         // to know a vertex buffer from a vampyre image record
      : ptrint; //a raw metaclass pointer of the module instance is stored here
    Unclaimed: longbool;
  end;
}

  TThreadOccupationKindEnum = (tokIdle, tokGenericPulse, tokGenericRender,
    tokSleeping, tokWaitingForSwapBuffers,
    tokActualizingAssets, tokGamePhysics, tokAISimulationFrames, tokMultiplayerFloatingFrames,
    tokImaging, tokFileManager,
    tokRenderingToFbo
  );

  TThreadOccupationChart = array[TThreadOccupationKindEnum] of Double;
  PThreadOccupationChart = ^TThreadOccupationChart;

  //***TODO: Deprecate this crap! (text rendering)
  PStringFitRec = ^TStringFitRec;
  TStringFitRec = record
    face,
    ActualWidth,
    ActualHeight: integer;
    zoomX,
    zoomY: float;
  end;

  
  TMotherCPU = packed record
    Count: integer; //uses heuristics to guesstimate the real core count instead of cores*hyperthreading multiplier
    TSCInvariance: boolean; { if the Time Stamp Counter runs on a fixed frequency,
       independed of the real currlen CPU clock frequency. See more in Intel or AMD manuals for CPUID 80000007h
       Note: you need to find a *really* rare CPU for this to matter. Most
       older ones either don't support clock scaling, or do support this feature.
     }
    L2CacheSize: integer; //in kilobytes. May be 0, meaning "unknown"
                              //This is a *guesstimated* value: if it is < 512 on Intel processors,
                              //then cpu count gets divided by 2 while this value gets multiplied by 2.
    L2CacheLineSize: integer; //in bytes. May be 0.
    Name: {$ifdef cgekernel} String {$else} PAnsiChar {$endif};
    //Error: Data types which require initialization/finalization can't be used in variant records
    //FFFFFfffffuuuuu....

    ReservedStub: array[0..15] of ptruint;
  end;
  
  TMotherWindows98 = packed record
    DoesNotSupportUserDirectory: boolean; //there is no SHGetFolderPathA in the vanilla Windows 98. Needs a workaround.
    DoesNotSupportUnicode: boolean;
    ReservedStub: array[0..3] of ptruint;
  end;

  PPUnicodeChar = ^PUnicodeChar;


  PMotherSehState = ^TMotherSEHState;
  TMotherSEHState = packed record
    // see cge.ResetMotherErrorState() and cl_die.GetExceptionState()

    ThreadId: TThreadId;

    ExceptionCode, //true exception codes in Windows, not used in Linux (see cl_seh_hack.inc)
    ExceptionAddress: ptruint;

    NowDying,
    DyingAfterTrueException, {if the exception handling chain was set off
       by a real exception like an AV and not by a controlled call to Die() }
    CallStackLogged,
    ModuleIsAlreadyProcessingUnhandledException,
    IsModuleThread, {The current thread is created by the module and should be
      processed differently}
    ThreadFinished
      : boolean;

    AbbrTitleForIndicator: ShortString3;
    ThreadTitle: {$ifdef cgekernel} UnicodeString {$else} PUnicodeChar {$endif};

    Warnings: {$ifdef cgekernel} array of UnicodeString {$else} PPUnicodeChar {$endif};

    StackSegment: Word;

    {$ifdef cgekernel}
      _StopDyingString: UnicodeString;
    {$else}
      stub1: ptruint;
    {$endif}

    AVMemoryAddress,
    AVMemoryOperation: ptruint; //0 - read, 1 - write, 8-Data Execution Prevention. See MSDN.

    //thread load details are stored here: I did not find a better place
    tlic: TThreadLoadKind;
    tlitsc: {$ifdef cpuarm} timeval {$else} qword {$endif};
    tli: array [TThreadLoadKind] of int64; //accumulates RDTSC cycles


    ReservedStub: array[0..3] of ptruint;
    Next: PMotherSehState;
  end;
  
  TMotherState_ = packed record
    PhysMemory: integer; //in megabytes. Note: it often reports not all of it, e.g. 1010 when you have 1024
    IsRussian, //if the currently selected language is Russian.
    DeveloperMode,
    IsServer
      : boolean;
    UserName: {$ifdef cgekernel} AnsiString {$else} PAnsiChar {$endif};
   {$ifdef windows}
    CSIDL_AppDataDir: {$ifdef cgekernel} TFileNameString {$else} PFileNameChar {$endif};
   {$endif}

    OS: TOSType;
    OstID: {$ifdef cgekernel} AnsiString {$else} PAnsiChar {$endif}; //text representation of the OS enum value
    OSName: {$ifdef cgekernel} UnicodeString {$else} PUnicodeChar {$endif};


    UserNameUCS16: {$ifdef cgekernel} UnicodeString {$else} PUnicodeChar {$endif};
    ProgramDirIsWriteable: boolean;


    ProgramInitialized: boolean;

    Windows98: TMotherWindows98;

    ExePath,
    StdHomepath,
    DebugInfoPath,
    HomePath,
    InstallPath,
    CacheDir,
    SessionDir,
    ModuleBaseDir,
    LogFile:
      {$ifdef cgekernel} TFileNameString {$else} PFileNameChar {$endif}; //System 8-bit code page. Important!
        //it's utf-8 (forced) for Linux and cp1251 (usually) for windows
    FrameworkName,
    ServerModuleName:
      {$ifdef cgekernel} AnsiString {$else} PAnsiChar {$endif};

    RunSessionHash: TAssetHash; {some module resources (e.g.
        FBOs and textures) should _not_ be kept between different program runs.}
    IsPortable: boolean; //this is "portable" version of the program (see the main *.ini file)
    Fatality: boolean; {if set, the following Die bubbling up from the module
      causes to write the error  details to a file, restart self and display
      thse details in a dialog box}
    BatteryRemaining: integer; //Windows only. 0..100 or 255 when on AC/unknown

//    SSE2Available: boolean; //The standalone mother is always compiled *without* sse support
      //Tests availability by trying to execute an asm command and catching exception

    DriveZIsRoot: boolean; //in wine and darwine. It means "filesystem root" aka "/".
                           //This is used by Windows executable to determine Linux version
                           //  when running in Wine (by parsing the file Z:\etc\issue )


    ExeHandle: {$ifdef windows} THandle {$else} pointer {$endif}; //is filled by InitSehHack() in cl_seh_hack.pp
    ProcessID: SizeUint; //pid for linux
    MainThreadId: TThreadId; //is filled by InitSehHack()


    StateTrashedRestartRequired: boolean; //is set to signal that any attempt to load a module should cause program restart
    Terminated: boolean; //To curb broblems with various hacks that keep executing even after the call to TerminateProcess() (W.T.F.?)
    ExitRequested: boolean; //is used to initiate program shutdown


    RestartRequested: boolean;



    SuddenShutdownReason: // this one is filled in the call to LogGoDownXX().
      {$ifdef cgekernel} UnicodeString {$else} PUnicodeChar {$endif};

   // ModuleIsAlreadyProcessingUnhandledException: boolean; //to avoid an infinite loop
      //in the mechanism that passes the unhandled exceptions back to the module for error reporting

    NumMotherThreads,
    NumModuleThreads: integer;

    SavingTheSessionNow: boolean;

    ReservedStub: array[0..15] of ptruint;
  end;

  TMotherDebug = packed record
    Enabled,
    Verbose,
    StopOnOpenGlErrors,
    ShowFrameNumbersInLog: boolean;

    ReservedStub: array[0..15] of ptruint;
  end;
  
  TMotherCore = packed record
    //to be used by the game module ONLY!
    ValidateVersion: procedure (m, n, b: integer); cdecl;
    ValidateTypesets: procedure (i, n: integer); cdecl;
    ValidateTypesetsEq: procedure (n: PAnsiChar; v1, v2: integer); cdecl;
    ValidateTypesetsLe: procedure (n: PAnsiChar; v1, v2: integer); cdecl;
    AllocateThreadExceptionState: procedure(thread_title: PUnicodeChar); cdecl; //to be called by module threads when they begin their Execute()
    RequestExit: procedure (); cdecl;
    IsGoingDownHard: function (): boolean; cdecl;
    MainLoopHeartBeat: procedure (); cdecl;
//    StopDying: function (): PUnicodeChar; cdecl;  //M: StopDying
    DbgSay: procedure (Yell: PUnicodeChar); cdecl;
    DbgSayCriticalSection: TRtlCriticalSection;
//    SetDying: procedure (d: boolean); cdecl;
    LogW: procedure (txt: PUnicodeChar); cdecl;
    MsgToString: function  (M: TMessageId): PUnicodeChar; cdecl;
    AddLogOK:  procedure (); cdecl;
    GetAssetStorageMemoryStream: function (): pointer; cdecl;//This is a per-module setting!
    SetAssetStorageMemoryStream: procedure (s: pointer); cdecl;
    AssetStorageCriticalSection: TRtlCriticalSection;
    DisplaySessionFailedErrorMessage: procedure(w: PUnicodeChar); cdecl; //for EXTRA fatal cases like a stack overflow
    ReservedStub: array[0..15] of ptruint;
  end;

  PMotherGamepad = ^TMotherGamepad;
  TMotherGamepad = packed record
    Next: PMotherGamepad;
    Present: boolean;
    Name,
    AbsenceReason: {$ifdef cgekernel} UnicodeString {$else} PUnicodeChar {$endif};
    Rumble: array[0..3] of float; {
      This array is usually filled by the module. It is monitored by the gamepad polling thread.
      0 - left (low-frequency) vibrator amplitude (0.0..1.0)
      1 - left vibrator activity duration, seconds
      2 - right (high-frequency) vibrator amplitude
      3 - right vibrator activity duration
      The gamepad polling thread has its own duration counters that are reset by changes
        in the requested amplitudes (array entries 0 and 2). Durations cannot exceed
        limit set by the GAMEPAD_RUMBLE_DURATION_LIMIT constant (default 5.0 seconds)
      Setting the same amplitude won't reset the internal counters, so the vibrators will stop if there's no change.
    }
    {
      Xbox 360 game controller's right thumb emulating mouse
      There's no dead zone, all effects are quadratic!
      Works surprisingly well (at least using an expensive Razer tournament controller) }
    MouseSpeedMultiplier, //pixels per second for maximum position, clipped to 100..3000
    MouseAccelThreshold, //percent of the maximum position, clipped to 10..90
    MouseAccelMultiplier //in pixels per second^2 for maximum, clipped to 0..3000 (when thumb is pushed to the edge)
      : longint;
    ReservedStub: array[0..15] of ptruint;
  end;
  
  TMotherInput_ = packed record
    TextInput,
    GrabPointer: boolean;
    Events: {$ifdef cgekernel} array of TInputEvent {$else} PInputEvent {$endif};
//    GetEventElement: function (): integer; cdecl; //reads the message queue;
    ReservedStub: array[0..15] of ptruint;
  end; 
  
  TMotherConfig = packed record //For use by the game module only!
    SetInt: procedure (section, id: PAnsiChar; Value: integer); cdecl;
    GetInt: function (section, id: PAnsiChar): integer; cdecl;
    GetIntWithCheck: function (section, id: PAnsiChar; min, max: integer): integer; cdecl;
    SetString: procedure (section, id, Value: PAnsiChar); cdecl; //utf8
    GetString: function (section, id: PAnsiChar): PAnsiChar; cdecl; //utf8
    ReservedStub: array[0..15] of ptruint;
  end;
  
  TMotherModule_ = packed record
    State: TModuleState;

    MustUnload, //set by mother to command the module to save the session and prepare for unloading
    SessionNotSaved, //set by the module. No force-restart allowed until it clears this.
    IsSavingSession, //set by the module when it is saving the session
    ReadyToUnload, //set by the module when it is ready
    ApplicationExiting //set inside by mother TModule destructor
      : boolean;

    NameW: {$ifdef cgekernel} UnicodeString {$else} PUnicodeChar {$endif};
    RequestToChooseModule: boolean;
    FileName,
    PureName: //oops. Must NOT be used in cl_module for internal workings
     {$ifdef cgekernel} AnsiString {$else} PAnsiChar {$endif}; //System 8-bit code page. Important!
    NeedToKeepAssets: function (): boolean; cdecl;
    EraseAssetsList: procedure(); cdecl;
    WaitingToFinishTimeoutMoment: TDateTime; //to manage timeout and force restart on waiting for too long.

    ControlLogicThreadStackSize,
    BackgroundThreadStackSize: PtrUInt; //Is filled by the mother module at module loading, according its INI file
    RollBackTheSessionOnNextLoad: boolean;
    FloatToFixedAssignmentsDisabled: boolean;//if an attempt made to assign a floating-point
     // value to a fixed-point variable, assertion exception happens. This is to avoid leaking
     // non-deterministic data into deterministic data

    ActiveInd,

    CurrentlyExecutingInd: integer; //NOTE: applies to main thread only!
    (* This variable is vital,
      it controls exception handling and self-debugging.
      It is set to active module index whenever mother calls the module
        and gets reset back to NOT_A_MODULE afterwards
        See cl_seh_hack_xxx.JumpToDllRaiseFunction() and TModule.ResetExecutionZoneFlags()
    *)

    DllInitialized: boolean;

    MultithreadingDisabled: boolean;

    StopDying: function(pes: PMotherSehState = nil): PUnicodeChar; cdecl;
    StartDying: procedure (txt: PUnicodeChar); cdecl;

    ExitWithoutTryingToUnloadTheDLL: boolean;

    ReservedStub: array[0..15] of ptruint;
  end;

  PMotherBuiltinFont = ^TMotherBuiltinFont;
  TMotherBuiltinFont = packed record
    Texture: GLuint;
    MaxChars: integer;
    TexCoords: PVector4f;
    CharMap: PLongint; //65536 elements, addressed by Ord(UnicodeChar)
    CharMapHigh: Integer; //ord
    Monospace: boolean;
    KerningDelta: PShortInt; //MaxChars * MaxChars matrix. Always null for monospace fonts.
    ReservedStub: array[0..3] of ptruint;
  end;

  {March 2016, paradigm shift: now Mother only supplies modules with textures
    and basic data about glyph placement. The relevant routines are repeated
    in universal includes, less headache with Mother API and more flexibility
    this way. }
  TMotherText = packed record
    NumBuiltinFonts: integer;
    BuiltinFont: PMotherBuiltInFont; { нагло пользуюсь тем,
      что к указателям можно адресоваться как к массивам, C-style }

    FixedFontQuality: integer; //read-only, autoregulated! 0 = hq font, 1 = built-in lq font
    FixedFontEmergencyDownsample: boolean;

//    StringHeightToFontFace: procedure (size: integer; face: Pinteger; zoom: PGLfloat); cdecl;
//    FitStringIntoRectangle: procedure (Size, numChars, numLines, Width, Height: integer; fit: PStringFitRec);  cdecl;
//    GetStringWidth: function (Size, numChars: integer): GLfloat; cdecl;
//    GetStringHeight: function (Size, numLines: integer): GLfloat; cdecl;

    //the caller is responsible for switching to FFP and setting the ortho projection!
//    SetRenderState: procedure (noshadow: boolean); cdecl;
//    RenderString: procedure (fit: PStringFitRec; Left, Top: glFloat; w: PUnicodeChar); cdecl;
    ReservedStub: array[0..15] of ptruint;
  end;  
  
  TMotherSound = packed record
    BufferAhead: longint; //in samples (44100/second).
    GetOPenALProcAddress: function (ProcName: PAnsiChar): Pointer; cdecl; //M: GetALProcAddress

    //the data format is always 16 bits stereo.
    NewData: procedure (begins: TDateTime; samples, rate: integer; data: pointer); cdecl;
    SetVolume: procedure (volume: float); cdecl;
    PlayFromFile: procedure (Wavfilename: PAnsiChar); cdecl;
    ReservedStub: array[0..15] of ptruint;
  end; 
  

  TMotherPenTablet = packed record
    Present: boolean;
    Name,
    AbsenceReason: {$ifdef cgekernel} UnicodeString {$else} PUnicodeChar {$endif};
    ActivityDetected: boolean;
    PressureResolution: integer;

    ReservedStub: array[0..15] of ptruint;
  end;   
  
 { TMotherFileManager = packed record
    MakeSurePathExists: procedure (path: PFileNameChar); cdecl;
    UnmangleFileName,
    MangleFileName: function (m: PFileNameChar): PFileNameChar; cdecl;
    GetFileHash: procedure (name: PFileNameChar; phash: PAssetHash); cdecl;
    GetFileDateTime: function (name: PFileNameChar): TDateTime; cdecl;
    GetIsHashvalid: function (phash: pMd5Digest): boolean; cdecl;
    ReservedStub: array[0..15] of ptruint;
  end;  
}

  TMotherAssetKeeper = packed record
    GenHash: procedure (phash: PAssetHash); cdecl;
//   Store: procedure (phandle: pDWORD; phash: PAssetHash; ChepersyClassIndex: integer); cdecl;
//    Claim: function (phash: PAssetHash): PAssetRecord; cdecl;
//    GetFirst: function (): PAssetRecord; cdecl;
//    GetNext: function (): PAssetRecord; cdecl;
//    ClearList: procedure (); cdecl;
//    HashesMatch: function (a, b: PAssetHash): boolean; cdecl;
    ReservedStub: array[0..15] of ptruint;
  end;   

  TMotherMemory = packed record
    CreateStream: function (): pointer; cdecl;
//    DeleteStream: procedure (s: pointer); cdecl;
//    SeekStream: procedure (s: pointer; pos: integer); cdecl;
//    WriteToStream: procedure (s: pointer; buf: pointer; len: integer); cdecl;
//    ReadFromStream: function (s: pointer; buf: pointer; len: integer): integer; cdecl;
//    WriteIntToStream: procedure (s: pointer; i: integer); cdecl;
//    ReadIntFromStream: function (s: pointer): integer; cdecl;
    GetStreamMemory: function (s: pointer): pointer; cdecl;
//    GetStreamSize: function (s: pointer): integer; cdecl;
//    StreamLoadFromFile: procedure (s: pointer; name: PAnsiChar); cdecl;
//    StreamSaveToFile: procedure (s: pointer; name: PAnsiChar); cdecl;
    Get: function (size: integer): pointer; cdecl;
    Free: procedure (p: pointer); cdecl;
    Realloc: function (p: pointer; newsize: integer): pointer; cdecl;

    ReservedStub: array[0..15] of ptruint;
  end;

  TMotherStream_ = packed record
    CreateOfFile: function(name: PFileNameChar; Mode: Word): pointer; cdecl;
    GetSize: function(s:pointer) : Int64; cdecl;
    SetSize: procedure (s:pointer; const NewSize: Int64); cdecl;
    Read: function (s:pointer; Buf: pointer; Count: Longint): Longint; cdecl;
    Write: function (s:pointer; Buf: pointer; Count: Longint): Longint; cdecl;
    GetPosition: function(s:pointer): Int64; cdecl;
    SetPosition: procedure(s:pointer; const Pos: Int64); cdecl;
    Free: procedure(var s: pointer) cdecl;

    ReservedStub: array[0..15] of ptruint;
  end;

  TMotherTimer = packed record
    FPS,
    QF: single;
    StartupDateTime: TDateTime;
    RdtscFrequency, //gets updated dynamically, a few times per second. Thus, it's somewhat noisy.
    LastSwapBuffersDuration: double; //seconds
    FpsDisplayingParams: TFpsCounterParams;
    AverageSwapBuffersDuration: double;
    QF_ManualOverride: boolean;
    SuspendSwapBuffersMeasurement: boolean; { This is set by the module:
      signaling the mother to suspend measurement
      during operations like creating FBOs to avoid a feedback loop}
    CullFrameRate: integer; {
      This field gets set to 30 before each call to the module pulse function
        (or to 60 if a mouse movement event was received).
      Unless the module sets it back to a higher value, there will be Sleep() called
        by the mother module after return from the module pulse function, to cull
        frame rate at the value determined here. (implemented in cl_window)
      This way only high-performance 3d rendering modules would hog
        all the video card & cpu resources,
        and only when they specifically request it.
      Setting it higher than 300 disables culling.

      Aug. 2015, addendum:
        This value sets the *higher* limit for the frame duration. When the
        module has main thread background tasks, it sets CullFrameRate 50% higher
        than intended and performs its own work/wait cycle.
        This is done for simplicity, to minimize changes to existing code
        during paradigm shift.
    }
    Second: function (): double; cdecl;
    UsecDelta: function (ptsc: pqword): double; cdecl;
    {$ifdef windows}
    ActiveTBP1: boolean; //to monitor calls to timeBeginPeriod(1)
    //  Highrer resolution waits are used for arbitrary FPS stabilization
    //  but must be deactivated when laptop is running on battery as this
    //  OS-wide mode prevents CPU from power saving
    {$endif}
    HiResSleep: procedure(ms: longint); cdecl;
    ReservedStub: array[0..15] of ptruint;
  end;

  TGAPIModeEnum = (gapi_DeprecatedImmediateGL, gapi_GL21, gapi_GLES2);

  TMotherGAPI = packed record
    Initialized: boolean;
    Mode: TGAPIModeEnum;
    VersionHi,
    VersionMed: integer;
    VersionString,
    Renderer,
    Vendor,
    Summary:  {$ifdef cgekernel} UnicodeString {$else} PUnicodeChar {$endif};
    Extensions: {$ifdef cgekernel} UnicodeString {$else} PUnicodeChar {$endif};
    NpotSupported: boolean; //I own some real museum pieces that do not support
                            //NPOT textures. I love testing Chentrah on them :)
    TextureMax,
    _3dTextureMax,
    FboMax: ptrint;
    MaxTextureCoords,
    MaxTextureUnits,
    MaxVShTextureUnits: ptrint;
    Quirks,
    QuirksTested: TGAPIQuirksSet;
    RendererAlias: {$ifdef cgekernel} AnsiString {$else} PAnsiChar {$endif};//suitable for using in file names and INI section names
    MaxAnisotropy: float;

    SkipSwapBuffers: boolean; { Serves to avoid HOM effect in cases where
      rendering was not performed by some reason.
      Is reset after each SwapBuffers call.   }

    SwapBuffers: procedure (); cdecl;
    BuildAlphaMipmaps: procedure (Width, Height: integer; Data: pointer); cdecl; //RGBA8 only. An ancient relic written in assembly language to utilize MMX
    GetProcAddress: function (ProcName, Suffix: PAnsiChar): Pointer; cdecl; //M: GetGLProcAddress
    GetProcAddress2: function (ProcName, Suffix1, Suffix2: PAnsiChar): Pointer; cdecl; // M: GetGLProcAddress2

  //  GetHigherPowerOf2: function (i: integer): integer; cdecl;
  //  GetLowerPowerOf2:  function (i: integer): integer; cdecl;

    StubTexture: ptruint; //glEnum

    ReservedStub: array[0..15] of ptruint;
  end;

  TMotherMousePointer = packed record
    //checked.: check where it is called from since Vampyre is NOT thread-safe!
    //LoadFromFile: procedure (FileName: PAnsiChar; xhotspot, yhotspot: integer); cdecl;
    //LoadFromMemory: procedure (pfile: pointer; size, xhotspot, yhotspot: integer); cdecl;
    //Remember, kids! 32x32 only!
    LoadFromRawPixmap: procedure (pixels: pointer; xhotspot, yhotspot: integer); cdecl;
    LoadEmpty: procedure (); cdecl;
    LoadDefault: procedure (); cdecl;
    LoadHourglass: procedure (); cdecl;
    ReservedStub: array[0..15] of ptruint;
  end;

  TWindowManagerRect = record left, top, width, height: longint end;
  operator = (a, u: TWindowManagerRect) b: System.boolean;
  operator <> (a, u: TWindowManagerRect) b: System.boolean;

type
  PMotherScreenshot = ^TMotherScreenshot;
  TMotherScreenshot = packed record
    Request: integer; // module sets this,
      //0 - module can request a screenshot
      //1 - module asks mother to make screenshot,
      //2 - mother signals module it's done
      //-2 - module signals itself that the background task is running
      //-1 - completion, module signals mother it can delete the buffer
    Width, height: integer;
    Format: GLuint; //always GL_RGB8
    pdata: pointer; //not nil indicates that a screenshot has been made.
    ReservedStub: array[0..15] of ptruint;
  end;

  PMotherDisplay = ^TMotherDisplay;
  TMotherDisplay = packed record

    {
      All subroutines that render something or address anything graphical
      *MUST* check if this field is True.
      Formerly, this was done using WindowExists. But there could be
       situations where window exists, but rendering is not advisable. }
    CanRender: boolean;

    Next: PMotherDisplay; //a stub for potential multi-monitor support
(*
    //Dec 2015: moved this to respective framework classes.
    {$ifdef unix}
     {$ifdef darwin}
       //MacosX support was never finished :(
       RenderingContext: pointer;
     {$else}
       //Linux
       display: PDisplay;
       screen: integer;
       RenderingContext: GLXContext;
       WindowHandle: GLXDrawable;
     {$endif}
    {$else}
     //Windows
     display: HDC;
     RenderingContext : HGLRC;
     WindowHandle: HWND;
    {$endif}
*)
    InFullScreenMode,
    WindowExists,
    WindowVisible,
    WindowHasFocus,
    WindowMinimized: boolean;
    MouseIsBeyondOurWindow : boolean;
    RequestedFullScreenMode: integer; //0: do not care; -1: windowed; 1: fullscreen

    //replace all of the fields below
    ScreenRect, //in windowed mode, this is screen *client* rect, i.e. screen minus taskbar
    WindowClientRect, //To be used for rendering
    WindowRect, //including window decorations. Only needed to process mouse hovering over window borders
    FullscreenRect //required for pen tablet
      : TWindowManagerRect;
    //WindowOriginX, //TODO: add support for Linux! Currently it's always zero.
    //WindowOriginY,
    //ScreenWidth, //desktop size. again, gets updated each frame in both in Linux and Windows, via system call.
    //ScreenHeight,
    //WindowClientWidth, //actual window client area size at any time. OpenGL renderer uses this.
    //WindowClientHeight: integer; //gets updated every frame
    //These 4 keep values to return to from the fulscreen mode.
    //All in percentages of the screen size
    //WindowedOriginX,
    //WindowedOriginY,
    //WindowedWidth,
    //WindowedHeight: float;
    UseHardware1bitCursor: boolean; //is set automatically each frame if fps is below certain limit
    FadeIn: float; //for menu/startup transition effects
    LogoAnimationPlaying: boolean;
    LogoAnimationStart: TDateTime;
    BlackoutBegin: TDateTime; //fade to black is instaneous, then fade in for about one second.

    //*** TODO: Check where this is called from, Vampyre not thread-safe!
    LoadBackgroundFromFile: procedure (FileName: PFileNameChar); cdecl;
    LoadBackgroundFromMemoryMappedFile: procedure (p: pointer; size: integer); cdecl;
    RenderBackground:procedure  (); cdecl;
    GetWindowExists: function (): boolean; cdecl;
    SetWindowName: procedure (n: PUnicodeChar); cdecl;
    MousePointer: TMotherMousePointer;
    BeginBlackout: procedure (delta: double = 0.0); cdecl;
    Screenshot: TMotherScreenshot;
    ReservedStub: array[0..15] of ptruint;
  end;

  TMotherImaging = packed record
    (*
      This API must be called from one thread only
        (the library used does NOT support multi-threading)
      This means that after initial initialization is over and the mother begins
        loading the module, the mother can't use these anymore.
      One consequence is that any additional fonts, pointer images and such
        should be loaded by the module.
    *)
    CorrectThreadId: TThreadId;
    Delete: procedure (name: pointer); cdecl;
    Load: function (FileName: PFileNameChar): pointer; cdecl;
    LoadAs: function (FileName: PFileNameChar; format: GLenum): pointer; cdecl;
    LoadFromMem: function (p: pointer; size: integer): pointer; cdecl;
    LoadFromMemAs: function (p: pointer; size: integer; format: GLenum): pointer; cdecl;
    Create: function (width, height: integer; format: GLuint): pointer; cdecl;
    Save: procedure (FileName: PFileNameChar; name: pointer); cdecl;
    Resize: procedure (name: pointer; newwidth, newheight: integer); cdecl;
    Resample: procedure (name: pointer; newwidth, newheight: integer; wrap: boolean); cdecl;
    Clone: function (name1: pointer): pointer; cdecl;
    CloneResample: function (name1: pointer; newwidth, newheight: integer; wrap: boolean; newformat: GLuint = 0): pointer; cdecl;
    GetFmt: function (name: pointer): GLuint; cdecl;
    SetFmt: procedure(name: pointer; format: GLuint); cdecl;
    GetPixels: function (name: pointer): pointer; cdecl;
    GetWidth: function (name: pointer): integer; cdecl;
    GetHeight: function (name: pointer): integer; cdecl;
    GetBytes: function (name: pointer): integer; cdecl;
    ReservedStub: array[0..15] of ptruint;
  end;

  {$ifdef cgekernel}
  TLockupGuard = class;
//  TGamepadManager = class;
  {$endif}


  PMother = ^TMother;
  TMother = packed record
    Initialized,
    Safe: boolean;
    Validate: function(sizes: array of const): boolean; cdecl;
    Criticalsection,
    SehCriticalSection: TRTLCriticalSection;
    State: TMotherState_;
    Core: TMotherCore;
    ExceptionState
      : TMotherSehState; //contains a pointer-chain of states for other threads
    CPU: TMotherCPU;
    Timer: TMotherTimer;
    Config: TMotherConfig;
//    FileManager: TMotherFileManager;
    AssetKeeper: TMotherAssetKeeper;
    Memory: TMotherMemory;
    Stream: TMotherStream_;
    Module: TMotherModule_;
    GAPI: TMotherGAPI;
    Sound: TMotherSound;
    Text: TMotherText;
    Display: TMotherDisplay;
    Image: TMotherImaging;
	  Input: TMotherInput_;
    Gamepad: TMotherGamepad;
    PenTablet: TMotherPenTablet;
    {$ifdef cgekernel}
     LockupGuard: TLockupGuard;
//    GamepadManager: TGamepadManager;
    {$else}
     stub1: ptruint;
//    stub2: ptruint;
    {$endif}
    Debug: TMotherDebug;
    ReservedStub: array[0..10000] of ptruint;
  end;


const
  
  {$ifdef cgekernel}
   {$typeinfo off}
  {$endif}

 {$ifdef fpc}
  CompileDate = {$I %DATE%};
  CompileTime = {$I %TIME%};
  CompilerVersion = 'Free Pascal ' + {$I %FPCVERSION%};
  CompileTarget = {$I %FPCTARGETOS%} + '-' + {$I %FPCTARGETCPU%};
 {$else}
  {$fatal Chentrah source code is no longer compatible with Delphi!}
  <<<<<
  CompileDate = 'n/a';
  CompileTime = 'n/a';
  CompilerVersion = 'Delphi';
  CompileTarget = 'n/a';
 {$endif}

 {$ifdef unix}
  {$ifdef darwin}
   {$fatal Chentrah is not ported to MacOS X yet!}
   DS = '/';
   PathSlash = '/';
   SystemSuffix = 'macosx';
   DllExtension = '.dylib';
   DllPrefix = 'lib';
  {$else}
   //fatal Linux support is *temporarily* dropped as of Jan 2012!}
   DS = '/';
   PathSlash = '/';
   SystemSuffix = 'linux';
   DllExtension = '.so';
   DllPrefix = 'lib';
  {$endif}
 {$else}
  DS = '\';
  PathSlash = '\';
  SystemSuffix = 'win32';
  DllExtension = '.dll';
  DllPrefix = '';
 {$endif}



  Slash = PathSlash;


  {$ifdef fpc}
    //it is assumed that Delphi supports only win32 on i386
    {$ifndef ENDIAN_LITTLE}
      {$fatal POWERPC ARCHITECTURE NOT SUPPORTED DUE TO BINARY FORMAT INCOMPATIBILITIES}
    {$endif}

    {$ifndef CPU32}
      //fatal ONLY 32-BIT PLATFORMS ARE SUPPORTED YET!}
    {$endif}
  {$endif}

  Const
    SecondsPerDay = 86400.0;
{$endif}


