{
    This file is part of Chentrah,
    Copyright (C) 2004-2008 Anton Rzheshevski (chebmaster@mail.ru).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see http://www.gnu.org/licenses/

 **********************************************************************}


{$if FPC_FULLVERSION>=20700}

  {$push}
  {$warn IMPLICIT_STRING_CAST off}
  {$warn IMPLICIT_STRING_CAST_LOSS off}
  {$warn EXPLICIT_STRING_CAST off}
  {$warn EXPLICIT_STRING_CAST_LOSS off}

   procedure InitUnicode;
   begin
     //nothing to do
   end;

   function Cp1251CharToUnicodeChar(a: AnsiChar): UnicodeChar;
   var
     s: AnsiString1251;
     u: UnicodeString;
   begin
     SetLength(s, 1);
     byte(s[1]):= byte(a);
     u:= s;
     Result:= u[1];
     u:= '';
     s:= '';
   end;

   function FileNameToUnicode(f: TFileNameString): UnicodeString;
   begin
     Result:= f;
   end;

   function UnicodeToFileName(u: UnicodeString): TFileNameString;
   begin
     Result:= u;
   end;

   function WideToCp1251(u: UnicodeString): AnsiString1251;
   begin
     Result:= u;
   end;

   function Cp1251ToWide(a: AnsiString1251): UnicodeString;
   begin
     Result:= a;
   end;

   function WideToAnsi(u: UnicodeString): AnsiString;
   begin
     Result:= u;
   end;

   function WideToConsole(u: UnicodeString): RawByteString;
   begin
     Result:= u;
   end;


   function AnsiToWide(a: RawByteString): UnicodeString;
   begin
     Result:= a;
   end;

   function WideToFileName(u: UnicodeString): TFileNameString; overload;
   begin
     Result:= u;
   end;

   function FileNameToWide(f: TFileNameString): UnicodeString;
   begin
     Result:= f;
   end;

   function FileNameToUtf8(f: TFileNameString): Utf8String;
   begin
     Result:= Utf8Encode(f);
   end;

   function Utf8ToFileName(u: Utf8String): TFileNameString;
   begin
     Result:= Utf8Decode(u);
   end;

  {$pop}
{$else}
   // Legacy Free Pascal with no encoding support


  {$ifdef windows}
    //no unicode support for file names
     function FileNameToUtf8(f: AnsiString): Utf8String;
     begin
       Result:= Utf8Encode(AnsiToWide(f));
     end;

     function Utf8ToFileName(u: Utf8String): AnsiString;
     begin
       Result:= WideToAnsi(Utf8Decode(u));
     end;
  {$else}
     //utf-8 file names
     function FileNameToUtf8(f:  TFileNameString): Utf8String;
     begin
       Result:= f;
     end;

     function Utf8ToFileName(u:  Utf8String): TFileNameString;
     begin
       Result:= u;
     end;
  {$endif}

   procedure InitUnicode;
   var
     j: integer;
     a: AnsiChar;
   begin
     For j:=0 to 127 do _UniToAnsiTable[j]:=chr(j);
     For j:=128 to  $FFFF do _UniToAnsiTable[j]:='?'; //everything else is undefined at first.
     For a:=#0 to #127 do _AnsiToUnitable[a]:={$ifdef fpc}a;{$else}UnicodeChar(ord(a));{$endif}//Dumb Delphi! :(

     //Russian alphabet:
     // "a" to "ya", capitals then smalls, 32 + 32 characters
     For j:=$0410 to $044F do _UniToAnsiTable[j]:=chr($C0 + j - $0410);

     //33st character, "yo":
     _UniToAnsiTable[$0401]:=chr($A8); //capital
     _UniToAnsiTable[$0451]:=chr($B8); //small

     _UniToAnsiTable[ord('№')]:=chr($B9);
     _UniToAnsiTable[ord('«')]:=chr($AB);
     _UniToAnsiTable[ord('»')]:=chr($BB);

     For a:=#128 to #$DF do _AnsiToUnitable[a]:='?';
     //Russian alphabet again...
     For a:=#$C0 to #$FF do _AnsiToUnitable[a]:= UnicodeChar($0410 + ord(a) - $C0);
     //"yo":
     _AnsiToUnitable[#$A8]:=UnicodeChar($0401);
     _AnsiToUnitable[#$B8]:=UnicodeChar($0451);
   end;

   function FileNameToUnicode(f: TFileNameString): UnicodeString;
   begin
     Result:= AnsiToWide(f);
   end;

   function UnicodeToFileName(u: UnicodeString): TFileNameString;
   begin
     Result:= WideToAnsi(u);
   end;

   function Cp1251CharToUnicodeChar(a: AnsiChar): UnicodeChar;
   begin
     Result:= _AnsiToUniTable[a];
   end;

   function WideToCp1251(u: UnicodeString): AnsiString1251;
   var
     j: integer;
   begin
     SetLength(Result, Length(u));
     For j:=1 to Length(u) do Result[j]:= _UniToAnsiTable[ord(u[j])];
   end;

   function Cp1251ToWide(a: AnsiString1251): UnicodeString;
   var
     j: integer;
   begin
     SetLength(Result, Length(a));
     For j:=1 to Length(a) do Result[j]:= _AnsiToUniTable[a[j]];
   end;


   function WideToAnsi(u: UnicodeString): AnsiString;
   {$ifndef unix}
   var
     i: integer;
     b: longbool;
     c: ansichar;
   {$endif}
   begin
     {$ifdef unix} //utf-8 assumed
      Result:=Utf8Encode(u);
     {$else}
      if u='' then Result:=''
      else begin
        i:=WideCharToMultiByte(CP_ACP, WC_COMPOSITECHECK or WC_DISCARDNS, @u[1], length(u), nil, 0, nil, nil);
        SetLength(Result, i);
        b:=True;
        c:='?';
        WideCharToMultiByte(CP_ACP, WC_COMPOSITECHECK or WC_DISCARDNS, @u[1], length(u), @Result[1], i, @c, @b);
      end;
     {$endif}
   end;

   function WideToConsole(u: UnicodeString): AnsiString;
   {$ifndef unix}
   var
     i: integer;
     b: longbool;
     c: ansichar;
   {$endif}
   begin
     {$ifdef unix} //utf-8 assumed
      Result:=Utf8Encode(u);
     {$else}
      if u='' then Result:=''
      else begin
        i:=WideCharToMultiByte(GetConsoleOutputCP(), WC_COMPOSITECHECK or WC_DISCARDNS, @u[1], length(u), nil, 0, nil, nil);
        SetLength(Result, i);
        b:=True;
        c:='?';
        WideCharToMultiByte(GetConsoleOutputCP(), WC_COMPOSITECHECK or WC_DISCARDNS, @u[1], length(u), @Result[1], i, @c, @b);
      end;
     {$endif}
   end;


   function AnsiToWide(a: AnsiString): UnicodeString;
   {$ifndef unix}
   var
     i: integer;
   {$endif}
   begin
     {$ifdef unix} //utf-8 assumed
      Result:=Utf8Decode(a);
     {$else}
      if a='' then Exit('')
      else begin
        i:=MultiByteToWideChar(CP_ACP, MB_PRECOMPOSED, @a[1], length(a), nil, 0);
        if i > 0 then begin
        SetLength(Result, i);
        MultiByteToWideChar(CP_ACP, MB_PRECOMPOSED, @a[1], length(a), @Result[1], i);
        end
        else
          if Mother^.State.OS = ostWin98 then begin
            //No Unicode support installed, assuming Russian/Cyrillic (CP 1251)
            Result:= CP1251ToWide(a);
          end;
      end;
     {$endif}
   end;

  function WideToFileName(u: UnicodeString): TFileNameString; overload;
  begin
    {$ifndef unix}
      Result:= Utf8Encode(u);
    {$else}
      {$ifdef win_unicode_fnames}
        Result:= u; //both are UnicodeString
      {$else}
        Result:= WideToAnsi(u);
      {$endif}
    {$endif}
  end;

  function FileNameToWide(f: TFileNameString): UnicodeString;
  begin
    {$ifndef unix}
      Result:= Utf8Decode(f);
    {$else}
      {$ifdef  win_unicode_fnames}
        Result:= f; //both are UnicodeString
      {$else}
        Result:= AnsiToWide(f);
      {$endif}
    {$endif}
  end;

{$endif}

 function TrimUnicodeString (u: UnicodeString): UnicodeString;
 var j, a, b: integer;
 begin
   a:=1;
   While (a < length(u)) and (ord(u[a]) <= 32) do inc(a);
   b:=length(u);
   While (b > 1) and (ord(u[a]) <= 32) do dec(b);
   if a > b then exit('');
   Result:= Copy(u, a, b - a + 1);
 end;


function WideCOPY(s: UnicodeString; pos, len: integer): UnicodeString;
var i: integer;
begin
  Result:='';
  For i:=pos to pos + len - 1 do begin
    if i > Length(s) then Exit;
    Result:=Result + s[i];
  end;
end;

function WideReplace(u, a, b: UnicodeString): UnicodeString;
                            {$if FPC_FULLVERSION < 20700} OVERLOAD; {$endif}
var p: integer;
    s: UnicodeString;
begin
  if a = b then Exit(u);
  s:=u;
  p:= WidePos(a, b);
  if p > 0 then  b:= WideReplace(b, a, '<!>'); //or else... >:(
  p:=WidePos(a, s);
  While p > 0 do begin
    s:=WideCOPY(s, 1, p - 1) + b + WideCOPY(s, p + Length(a), Length(s));
    p:=WidePos(a, s);
  end;
  Result:=s;
end;

{$if FPC_FULLVERSION < 20700}

 function WideReplace(u: UnicodeString; a, b: AnsiString): UnicodeString; OVERLOAD;
 begin
   Result:=WideReplace(u, AnsiToWide(a), AnsiToWide(b));
 end;

 function WideReplace(u: UnicodeString; a: AnsiString; b: UnicodeString): UnicodeString; OVERLOAD;
 begin
   Result:=WideReplace(u, AnsiToWide(a), b);
 end;

{$endif}

 function WidePos(a, u: UnicodeString): integer;
 var
   j, i: integer;
 begin
   Result:=-1;
   For j:=1 to length(u) - length(a) + 1 do begin
     i:=1;
     While (i < Length(a) + 1) and (a[i] = u[i + j - 1]) do inc(i);
     if i = (Length(a) + 1) then begin
       Result:=j;
       Exit;
     end;
   end;
 end;

  function Explode(separator, s: UnicodeString; TrimIt: system.boolean): UnicodeStringArray; overload;
  var i, a: integer;
  begin
    Result:= nil;
    if s = '' then exit;
    if separator = '' then begin
      SetLength(Result, length(s));
      for i:= 0 to High(Result) do Result[i]:= COPY(s, i+1, 1);
      exit;
    end;
    i:= 1;
    a:= 1;
    repeat
      if (s[i] = separator[1]) and (COPY(s, i, length(separator)) = separator) then begin
        SetLength(Result, Length(Result) + 1);
        Result[High(Result)]:= COPY(s, a, i - a);
        if TrimIt then Result[High(Result)]:= TrimUnicodeString(Result[High(Result)]);
        i+= length(separator);
        a:= i;
      end
      else i+= 1;
    until i > length(s);
    if (a <= length(s)) then begin
      SetLength(Result, Length(Result) + 1);
      Result[High(Result)]:= COPY(s, a, length(s) - a + 1);
    end;
  end;

  function Explode(separator, s: RawByteString; TrimIt: system.boolean): RawByteStringArray; overload;
  var i, a: integer;
  begin
    Result:= nil;
    if s = '' then exit;
    if separator = '' then begin
      SetLength(Result, length(s));
      for i:= 0 to High(Result) do Result[i]:= COPY(s, i+1, 1);
      exit;
    end;
    i:= 1;
    a:= 1;
    repeat
      if (s[i] = separator[1]) and (COPY(s, i, length(separator)) = separator) then begin
        SetLength(Result, Length(Result) + 1);
        Result[High(Result)]:= COPY(s, a, i - a);
        if TrimIt then Result[High(Result)]:= Trim(Result[High(Result)]);
        i+= length(separator);
        a:= i;
      end
      else i+= 1;
    until i > length(s);
    if (a <= length(s)) then begin
      SetLength(Result, Length(Result) + 1);
      Result[High(Result)]:= COPY(s, a, length(s) - a + 1);
      if TrimIt then Result[High(Result)]:= Trim(Result[High(Result)]);
    end;
  end;


  function Implode(separator: RawByteString; a: RawByteStringArray): RawByteString; overload;
  var i, sl, al, l: integer;
  begin
    //a friendly reminder: strings in Pascal index from 1, not 0
    Result:= '';
    sl:= Length(separator);
    l:= 0;
    for i:=0 to High(a) do begin
      if i > 0 then l+= sl;
      l+= Length(a[i]);
    end;
    if Length(a) > 0 then Result:= COPY(a[0], 1, 1); //assigns string's character set in FPC 3
    SetLength(Result, l);
    l:= 0;
    for i:=0 to High(a) do begin
      if i > 0 then begin
        if sl > 0 then MOVE(separator[1], Result[l + 1], sl + SizeOf(separator[1])); //copying as a binary blob
        l+= sl;
      end;
      al:= Length(a[i]);
      if al > 0 then MOVE(a[i][1], Result[l + 1], al * Sizeof(separator[1]));
      l+= al;
    end;
  end;

  function Implode(separator: UnicodeString; a: UnicodeStringArray): UnicodeString; overload;
  var i, sl, al, l: integer;
  begin
    Result:= '';
    sl:= Length(separator);
    l:= 0;
    for i:=0 to High(a) do begin
      if i > 0 then l+= sl;
      l+= Length(a[i]);
    end;
    SetLength(Result, l);
    l:= 0;
    for i:=0 to High(a) do begin
      if i > 0 then begin
        if sl > 0 then MOVE(separator[1], Result[l + 1], sl + SizeOf(separator[1]));
        l+= sl;
      end;
      al:= Length(a[i]);
      if al > 0 then MOVE(a[i][1], Result[l + 1], al * Sizeof(separator[1]));
      l+= al;
    end;
{
    Result:= '';
    for i:=0 to High(a) do begin
      if i > 0 then Result+= separator;
      Result+= a[i];
    end;
}
  end;

//this code is so old it smells :( Of Turbo Pascal, no less :(
Function StrReplace (S, zamenjaemoe, zamenjajushee: RawByteString): RawByteString;
var p: integer;
    s2: RawByteString;
begin
  p:=Pos(zamenjaemoe, s);
  While p > 0 do begin
    s:=COPY(s, 1, p - 1) + zamenjajushee + COPY(s, p + Length(zamenjaemoe), Length(s));
    p:=Pos(zamenjaemoe, s);
  end;
  StrReplace:=s;
end;

Function StrReplace (S, zamenjaemoe, zamenjajushee:UnicodeString):UnicodeString;
var p: integer;
    s2: UnicodeString;
begin
  p:=Pos(zamenjaemoe, s);
  While p > 0 do begin
    s:=COPY(s, 1, p - 1) + zamenjajushee + COPY(s, p + Length(zamenjaemoe), Length(s));
    p:=Pos(zamenjaemoe, s);
  end;
  StrReplace:=s;
end;


